(PackageInfo named: 'Collections') postscript: 'CharacterSet allInstancesDo: #size'!
SystemOrganization addCategory: #'Collections-Abstract'!
SystemOrganization addCategory: #'Collections-Arrayed'!
SystemOrganization addCategory: #'Collections-Cache'!
SystemOrganization addCategory: #'Collections-Exceptions'!
SystemOrganization addCategory: #'Collections-Sequenceable'!
SystemOrganization addCategory: #'Collections-Stack'!
SystemOrganization addCategory: #'Collections-Streams'!
SystemOrganization addCategory: #'Collections-Strings'!
SystemOrganization addCategory: #'Collections-Support'!
SystemOrganization addCategory: #'Collections-Text'!
SystemOrganization addCategory: #'Collections-Unordered'!
SystemOrganization addCategory: #'Collections-Weak'!
SystemOrganization addCategory: #'Collections-Heap'!

Notification subclass: #ProgressNotification
	instanceVariableNames: 'amount done extra'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Exceptions'!

!ProgressNotification commentStamp: '<historical>' prior: 0!
Used to signal progress without requiring a specific receiver to notify. Caller/callee convention could be to simply count the number of signals caught or to pass more substantive information with #signal:.!

!ProgressNotification class methodsFor: 'exceptionInstantiator' stamp: 'ajh 1/22/2003 23:51'!
signal: signalerText extra: extraParam
	"TFEI - Signal the occurrence of an exceptional condition with a specified textual description."

	| ex |
	ex := self new.
	ex extraParam: extraParam.
	^ex signal: signalerText! !

!ProgressNotification methodsFor: 'accessing' stamp: 'ar 3/2/2001 20:11'!
amount
	^amount! !

!ProgressNotification methodsFor: 'accessing' stamp: 'ar 3/2/2001 20:12'!
amount: aNumber
	amount := aNumber! !

!ProgressNotification methodsFor: 'accessing' stamp: 'ar 3/2/2001 20:11'!
done
	^done! !

!ProgressNotification methodsFor: 'accessing' stamp: 'ar 3/2/2001 20:12'!
done: aNumber
	done := aNumber! !

!ProgressNotification methodsFor: 'accessing' stamp: 'ar 3/2/2001 20:12'!
extraParam
	^extra! !

!ProgressNotification methodsFor: 'accessing' stamp: 'ar 3/2/2001 20:12'!
extraParam: anObject
	extra := anObject! !

!BlockClosure methodsFor: '*Collections-Support-sorting' stamp: 'nice 3/12/2014 23:12'!
ascending
	"Return a SortFunction around the receiver. If the receiver is a 2 arg block, it is assumed it will do the collation directly itself, returning -1, 0, or 1. If the receiver is a one arg block, it will be evaluated for each a and b and of the sort input, and the result of sending <=> to those will be used."

	| function |
	function := SortFunction ascend.
	self numArgs = 1 ifTrue: [function monadicBlock: self].
	self numArgs = 2 ifTrue: [function collator: self].
	^function! !

!BlockClosure methodsFor: '*Collections-Support-sorting' stamp: 'nice 3/12/2014 23:12'!
asSortFunction

	^self ascending! !

!BlockClosure methodsFor: '*Collections-Support-sorting' stamp: 'nice 3/12/2014 23:12'!
descending
	"Opposite direction as ascending."

	^self ascending toggleDirection! !

Object subclass: #Ascii85Converter
	instanceVariableNames: 'dataStream stringStream number85 tupleSize binary'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!Ascii85Converter commentStamp: 'topa 7/10/2016 20:57' prior: 0!
I convert between binary data and an Ascii85 text representation of PostScript and PDF fame.
I am a little bit more efficient (~25% overhead) than Base64 (~30% overhead).

Instance Variables
	binary:		<Boolean>	- Tells whether to decode to binary data or characters
	dataStream:		<PositionableStream> - The data stream to decode to or encode from (Typically does not contain Ascii85)
	stringStream:		<PositionableStream> - The text stream to decode from or encode to (Always contains Ascii85)
	number85:		<Integer> - Decoder state, accumulated read number in base85
	tupleSize:		<SmallInteger> - Number of bytes read into number85 already
			
'Hello, World' ascii85Encoded.

'<~87cURD]htuF_+us~>' ascii85Decoded.
!

!Ascii85Converter class methodsFor: 'convenience' stamp: 'topa 7/9/2016 12:11'!
decodeToBytes: aStringOrStream
	" Analogous to Base64MimeConverter>>#mimeDecodeToBytes:"
	| expectedSize |
	expectedSize := aStringOrStream size * 4 // 5.
	^ (ByteArray new: expectedSize streamContents:
		[:stream |
			(self stringStream: aStringOrStream readStream dataStream: stream)
				decodeToByteArray]) readStream! !

!Ascii85Converter class methodsFor: 'convenience' stamp: 'topa 7/9/2016 12:12'!
decodeToChars: aStringOrStream
	" Analogous to Base64MimeConverter>>#mimeDecodeToChars:"
	| expectedSize |
	expectedSize := aStringOrStream size * 4 // 5.
	^ (String new: expectedSize streamContents:
		[:stream | self decode: aStringOrStream to: stream]) readStream! !

!Ascii85Converter class methodsFor: 'convenience' stamp: 'topa 7/9/2016 11:48'!
decode: aStringOrStream as: contentsClass

	^ contentsClass streamContents:
		[:out | self decode: aStringOrStream to: out]! !

!Ascii85Converter class methodsFor: 'convenience' stamp: 'topa 7/9/2016 12:09'!
decode: aStringOrStream to: outStream

	^ (self stringStream: aStringOrStream readStream dataStream: outStream)
		decode! !

!Ascii85Converter class methodsFor: 'convenience' stamp: 'topa 7/9/2016 11:49'!
encode: aCollectionOrStream

	^ String streamContents:
		[:out | self encode: aCollectionOrStream to: out]! !

!Ascii85Converter class methodsFor: 'convenience' stamp: 'topa 7/9/2016 12:09'!
encode: aCollectionOrStream to: outStream

	^ (self stringStream: outStream dataStream: aCollectionOrStream readStream)
		encode! !

!Ascii85Converter class methodsFor: 'instance creation' stamp: 'topa 7/9/2016 12:08'!
stringStream: aStream dataStream: anotherStream

	^ self new
		stringStream: aStream;
		dataStream: anotherStream;
		yourself! !

!Ascii85Converter methodsFor: 'accessing' stamp: 'topa 7/10/2016 16:02'!
ascii

	binary := false.! !

!Ascii85Converter methodsFor: 'accessing' stamp: 'topa 7/10/2016 16:02'!
binary

	binary := true.! !

!Ascii85Converter methodsFor: 'accessing' stamp: 'topa 7/9/2016 13:10'!
dataStream

	^dataStream! !

!Ascii85Converter methodsFor: 'accessing' stamp: 'topa 7/10/2016 00:53'!
dataStream: anObject

	dataStream := anObject.! !

!Ascii85Converter methodsFor: 'conversion' stamp: 'topa 7/10/2016 16:07'!
decode

	self readBOD ifFalse: [^ self dataStream].

	[self stringStream atEnd] whileFalse: [
		self stringStream skipSeparators.
		self readEOD ifTrue: [^ self endDecode]. "<--- End of data"

		self decodeChar: self stringStream next ifFail: [^ self dataStream]].
	"actually, should not reach, but our failure condition is returning the stream, anyway"		
	^ self dataStream! !

!Ascii85Converter methodsFor: 'conversion' stamp: 'topa 7/10/2016 16:12'!
decodeChar: char ifFail: failBlock

	char = $z ifTrue: [^ self decodeZIfFail: failBlock].
	
	(char between: $!! and: $u) ifFalse: [^ failBlock value].

	self incrementTupleSize.
	self incrementNumber85: char asInteger - 33 * (self pow85 at: self tupleSize).
	self tupleSize = 5 ifTrue:
		[self isBinary
			ifTrue: [self writeBytes255: self number85 atMax: 4]
			ifFalse: [self writeChars255: self number85 atMax: 4].
		self resetDecoderState]! !

!Ascii85Converter methodsFor: 'conversion' stamp: 'topa 7/10/2016 16:03'!
decodeToByteArray
	
	self binary.
	^ self decode! !

!Ascii85Converter methodsFor: 'conversion' stamp: 'topa 7/10/2016 16:16'!
decodeZIfFail: failBlock

	self tupleSize ~= 0 ifTrue: [^ failBlock value].
	self dataStream next: 4 put: (self isBinary ifTrue: [0] ifFalse: [Character null]).
! !

!Ascii85Converter methodsFor: 'conversion' stamp: 'topa 7/10/2016 01:19'!
encode

	| lineLength  |
	
	lineLength := 0.
	self stringStream nextPutAll: '<~'.
	[self dataStream atEnd] whileFalse: [
		| raw data out |
		lineLength >= 74 ifTrue: [self stringStream cr.  lineLength := 0].
		out := 5.
		raw := (self dataStream next: 4) asByteArray.
		raw size < 4 ifTrue:
			[out := raw size + 1.
			raw := raw, (self padOfSize: 4 - raw size)].
		data := raw unsignedLongAt: 1 bigEndian: true.
		data = 0
			ifTrue: [self stringStream nextPut: $z. lineLength := lineLength + 1]
			ifFalse: [self write85: data atMax: out. lineLength := lineLength + out]].
	self stringStream nextPutAll: '~>'.
	^ self stringStream
! !

!Ascii85Converter methodsFor: 'private' stamp: 'topa 7/10/2016 20:43'!
endDecode

	self tupleSize  > 0 ifTrue: 
		[self incrementNumber85: (self pow85 at: self tupleSize).
		self isBinary
			ifTrue: [self writeBytes255: self number85 atMax: self tupleSize - 1]
			ifFalse: [self writeChars255: self number85 atMax: self tupleSize - 1]].
	^ self dataStream! !

!Ascii85Converter methodsFor: 'private' stamp: 'topa 7/10/2016 15:56'!
incrementNumber85: aNumber

	number85 := number85 + aNumber.

	! !

!Ascii85Converter methodsFor: 'private' stamp: 'topa 7/10/2016 15:57'!
incrementTupleSize

	tupleSize := tupleSize + 1.
	! !

!Ascii85Converter methodsFor: 'initialize-release' stamp: 'topa 7/10/2016 16:02'!
initialize

	super initialize.
	self ascii.
	self resetDecoderState.! !

!Ascii85Converter methodsFor: 'testing' stamp: 'topa 7/10/2016 16:02'!
isBinary

	^ binary! !

!Ascii85Converter methodsFor: 'accessing' stamp: 'topa 7/10/2016 15:54'!
number85

	^ number85! !

!Ascii85Converter methodsFor: 'private' stamp: 'topa 7/9/2016 18:51'!
padOfSize: aNumber

	aNumber = 1 ifTrue: [^ #(0)].
	aNumber = 2 ifTrue: [^ #(0 0)].
	aNumber = 3 ifTrue: [^ #(0 0 0)].
	self error: 'Should not reach'.! !

!Ascii85Converter methodsFor: 'private' stamp: 'topa 7/10/2016 16:00'!
pow85
	
	^ #(52200625 614125 7225 85 1) "{85*85*85*85 . 85*85*85 . 85*85 . 85. 1}"! !

!Ascii85Converter methodsFor: 'private' stamp: 'topa 7/10/2016 00:03'!
readBOD

	self stringStream skipSeparators.
	self stringStream peek = $< ifFalse: [^ false] ifTrue:
		[self stringStream next. "ignore"
		self stringStream peek = $~ ifFalse: [^ false] ifTrue:
			[self stringStream next "ignore"]].
	^ true! !

!Ascii85Converter methodsFor: 'private' stamp: 'topa 7/10/2016 00:04'!
readEOD

	self stringStream skipSeparators.
	self stringStream peek = $~ ifFalse: [^ false] ifTrue:
		[self stringStream next. "ignore"
		self stringStream peek = $> ifFalse: [^ false] ifTrue:
			[self stringStream next "ignore"]].
	^ true! !

!Ascii85Converter methodsFor: 'private' stamp: 'topa 7/10/2016 15:56'!
resetDecoderState

	number85 := 0.
	tupleSize := 0.
	! !

!Ascii85Converter methodsFor: 'accessing' stamp: 'topa 7/9/2016 11:22'!
stringStream

	^ stringStream! !

!Ascii85Converter methodsFor: 'accessing' stamp: 'topa 7/9/2016 13:10'!
stringStream: anObject

	stringStream := anObject.! !

!Ascii85Converter methodsFor: 'accessing' stamp: 'topa 7/10/2016 15:54'!
tupleSize

	^ tupleSize! !

!Ascii85Converter methodsFor: 'writing' stamp: 'topa 7/9/2016 23:46'!
write85: anInteger atMax: aNumber

	| c1 c2 c3 c4 c5 remain |
	remain := anInteger.
	c5 := (remain \\ 85 + 33) asCharacter. remain := remain // 85.
	c4 := (remain \\ 85 + 33) asCharacter. remain := remain // 85.
	c3 := (remain \\ 85 + 33) asCharacter. remain := remain // 85.
	c2 := (remain \\ 85 + 33) asCharacter. remain := remain // 85.
	c1 := (remain \\ 85 + 33) asCharacter. remain := remain // 85.
	aNumber > 0 ifTrue: [self stringStream nextPut: c1.
	aNumber > 1 ifTrue: [self stringStream nextPut: c2.
	aNumber > 2 ifTrue: [self stringStream nextPut: c3.
	aNumber > 3 ifTrue: [self stringStream nextPut: c4.
	aNumber > 4 ifTrue: [self stringStream nextPut: c5]]]]].
	! !

!Ascii85Converter methodsFor: 'writing' stamp: 'topa 7/10/2016 20:41'!
writeBytes255: anInteger atMax: aNumber

	(aNumber between: 1 and: 4) ifFalse: [^ self error: 'Unexpected byte count'].
	4 to: (5 - aNumber) by: -1 do:
		[:index | self dataStream nextPut: (anInteger digitAt: index)].
! !

!Ascii85Converter methodsFor: 'writing' stamp: 'topa 7/10/2016 01:09'!
writeChars255: anInteger atMax: aNumber

	(aNumber between: 1 and: 4) ifFalse: [^ self error: 'Unexpected byte count'].
	4 to: (5 - aNumber) by: -1 do:
		[:index | self dataStream nextPut: (anInteger digitAt: index) asCharacter].
! !

Object subclass: #Collection
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Abstract'!

!Collection commentStamp: '<historical>' prior: 0!
I am the abstract superclass of all classes that represent a group of elements.!

Collection subclass: #Bag
	instanceVariableNames: 'contents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Unordered'!

!Bag commentStamp: '<historical>' prior: 0!
I represent an unordered collection of possibly duplicate elements.
	
I store these elements in a dictionary, tallying up occurrences of equal objects. Because I store an occurrence only once, my clients should beware that objects they store will not necessarily be retrieved such that == is true. If the client cares, a subclass of me should be created.!

!Bag class methodsFor: 'instance creation' stamp: 'nk 3/17/2001 09:52'!
contentsClass
	^Dictionary! !

!Bag class methodsFor: 'instance creation' stamp: 'sma 5/12/2000 13:31'!
new
	^ self new: 4! !

!Bag class methodsFor: 'instance creation' stamp: 'sma 5/12/2000 17:17'!
newFrom: aCollection 
	"Answer an instance of me containing the same elements as aCollection."

	^ self withAll: aCollection

"Examples:
	Bag newFrom: {1. 2. 3. 3}
	{1. 2. 3. 3} as: Bag
"! !

!Bag class methodsFor: 'instance creation' stamp: 'nk 3/17/2001 09:52'!
new: nElements
	^ super new setContents: (self contentsClass new: nElements)! !

!Bag methodsFor: 'adding' stamp: 'sma 5/12/2000 17:18'!
add: newObject 
	"Include newObject as one of the receiver's elements. Answer newObject."

	^ self add: newObject withOccurrences: 1! !

!Bag methodsFor: 'adding' stamp: 'sma 5/12/2000 17:20'!
add: newObject withOccurrences: anInteger 
	"Add newObject anInteger times to the receiver. Answer newObject."

	contents at: newObject put: (contents at: newObject ifAbsent: [0]) + anInteger.
	^ newObject! !

!Bag methodsFor: 'converting' stamp: 'sma 5/12/2000 14:34'!
asBag
	^ self! !

!Bag methodsFor: 'converting' stamp: 'nice 10/20/2009 19:37'!
asSet
	"Answer a set with the elements of the receiver."

	^ contents keys asSet! !

!Bag methodsFor: 'accessing' stamp: 'sma 5/12/2000 17:23'!
at: index 
	self errorNotKeyed! !

!Bag methodsFor: 'accessing' stamp: 'sma 5/12/2000 17:23'!
at: index put: anObject 
	self errorNotKeyed! !

!Bag methodsFor: 'accessing' stamp: 'tao 1/5/2000 18:25'!
cumulativeCounts
	"Answer with a collection of cumulative percents covered by elements so far."
	| s n |
	s := self size / 100.0. n := 0.
	^ self sortedCounts asArray collect:
		[:a | n := n + a key. (n / s roundTo: 0.1) -> a value]! !

!Bag methodsFor: 'enumerating' stamp: ''!
do: aBlock 
	"Refer to the comment in Collection|do:."

	contents associationsDo: [:assoc | assoc value timesRepeat: [aBlock value: assoc key]]! !

!Bag methodsFor: 'testing' stamp: ''!
includes: anObject 
	"Refer to the comment in Collection|includes:."

	^contents includesKey: anObject! !

!Bag methodsFor: 'math functions' stamp: 'jcg 11/16/2008 23:03'!
max
	"Answer the maximum value in the collection.  This optimized version only looks at each unique value once."
	^contents keys inject: contents keys anyOne into: [:max :each | max max: each]
! !

!Bag methodsFor: 'math functions' stamp: 'jcg 11/16/2008 23:03'!
min
	"Answer the minimum value in the collection.  This optimized version only looks at each unique value once."
	^contents keys inject: contents keys anyOne into: [:min :each | min min: each]
! !

!Bag methodsFor: 'enumerating' stamp: 'ul 3/8/2016 17:19'!
occurrencesOf: anObject
	"Answer how many of the receiver's elements are equal to anObject. Optimized version."

	^contents at: anObject ifAbsent: 0! !

!Bag methodsFor: 'copying' stamp: 'nice 10/5/2009 08:54'!
postCopy
	super postCopy.
	contents := contents copy! !

!Bag methodsFor: 'removing' stamp: 'nice 9/14/2009 20:28'!
removeAll
	"Implementation Note: as contents will be overwritten, a shallowCopy of self would be modified.
	An alternative implementation preserving capacity would be to create a new contents:
	self setContents: (self class contentsClass new: contents size)."
	
	contents removeAll! !

!Bag methodsFor: 'removing' stamp: 'sma 5/12/2000 14:32'!
remove: oldObject ifAbsent: exceptionBlock 
	"Refer to the comment in Collection|remove:ifAbsent:."

	| count |
	count := contents at: oldObject ifAbsent: [^ exceptionBlock value].
	count = 1
		ifTrue: [contents removeKey: oldObject]
		ifFalse: [contents at: oldObject put: count - 1].
	^ oldObject! !

!Bag methodsFor: 'private' stamp: 'sma 5/12/2000 14:49'!
setContents: aDictionary
	contents := aDictionary! !

!Bag methodsFor: 'accessing' stamp: 'sma 5/12/2000 11:35'!
size
	"Answer how many elements the receiver contains."

	| tally |
	tally := 0.
	contents do: [:each | tally := tally + each].
	^ tally! !

!Bag methodsFor: 'accessing' stamp: 'ul 9/13/2009 01:19'!
sortedCounts
	"Answer with a collection of counts with elements, sorted by decreasing
	count."

	^(Array new: contents size streamContents: [ :stream |
		contents associationsDo: [ :each |
			stream nextPut: each value -> each key ] ])
		sort: [:x :y | x >= y ];
		yourself! !

!Bag methodsFor: 'accessing' stamp: 'ul 9/12/2009 03:14'!
sortedElements
	"Answer with a collection of elements with counts, sorted by element."

	^contents associations
		sort;
		yourself! !

!Bag methodsFor: 'math functions' stamp: 'ul 11/7/2010 01:46'!
sum
	"Faster than the superclass implementation when you hold many instances of the same value (which you probably do, otherwise you wouldn't be using a Bag)."
	
	| sum first |
	first := true.
	contents keysAndValuesDo: [ :value :count |
		first 
			ifTrue: [ sum := value * count. first := false ]
			ifFalse: [ sum := sum + (value * count) ] ].
	first ifTrue: [ self errorEmptyCollection ].
	^sum! !

!Bag methodsFor: 'accessing' stamp: 'md 1/20/2006 15:58'!
valuesAndCounts

	^ contents! !

!Bag methodsFor: 'comparing' stamp: 'md 10/17/2004 16:09'!
= aBag
	"Two bags are equal if
	 (a) they are the same 'kind' of thing.
	 (b) they have the same size.
	 (c) each element occurs the same number of times in both of them"

	(aBag isKindOf: Bag) ifFalse: [^false].
	self size = aBag size ifFalse: [^false].
	contents associationsDo: [:assoc|
		(aBag occurrencesOf: assoc key) = assoc value
			ifFalse: [^false]].
	^true

! !

Bag subclass: #IdentityBag
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Unordered'!

!IdentityBag commentStamp: '<historical>' prior: 0!
Like a Bag, except that items are compared with #== instead of #= .

See the comment of IdentitySet for more information.
!

!IdentityBag class methodsFor: 'instance creation' stamp: 'nk 3/17/2001 09:53'!
contentsClass
	^IdentityDictionary! !

Collection subclass: #Bitset
	instanceVariableNames: 'bytes tally'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Support'!

!Bitset commentStamp: 'ul 8/22/2015 00:52' prior: 0!
I implement Bitsets, which are dictionary-like data structures mapping 0-1 values to integers between 0 and capacity-1, or in another way they are set-like data structures which can include values between 0 and capacity-1.
I implement three different kind of APIs, each corresponding to a way of thinking about this data structure:
- A Set-like API with #add:, #remove: and #includes:
- A Dictionary-like API with #at:, #at:put:
- And a bit-manipulation API with #bitAt:, #clearBitAt: and #setBitAt:.

Instance Variables
	bytes:		<ByteArray>
	tally:		<Integer>

bytes
	- a ByteArray which holds the values for each integer key. Each byte holds 8 values.

tally
	- the number of objects in this set, or the number or 1 values in this dictionary.
!

!Bitset class methodsFor: 'instance creation' stamp: 'ul 8/19/2015 23:46'!
new

	self error: 'Use #new: instead.'! !

!Bitset class methodsFor: 'instance creation' stamp: 'ul 8/19/2015 23:45'!
new: capacity

	^self basicNew initialize: capacity! !

!Bitset methodsFor: 'adding' stamp: 'ul 8/22/2015 00:56'!
add: anInteger
	"Add anInteger to this set. Return anInteger."

	self setBitAt: anInteger.
	^anInteger! !

!Bitset methodsFor: 'accessing' stamp: 'ul 8/19/2015 23:37'!
at: anInteger

	^self bitAt: anInteger
		! !

!Bitset methodsFor: 'accessing' stamp: 'ul 8/19/2015 23:48'!
at: anInteger put: aBit

	^self bitAt: anInteger put: aBit
		! !

!Bitset methodsFor: 'bit manipulation' stamp: 'ul 8/22/2015 00:53'!
bitAt: anInteger
	"Return the bit corresponding to anInteger."

	^((bytes at: (anInteger bitShift: -3) + 1) bitShift: 0 - (anInteger bitAnd: 7)) bitAnd: 1
		! !

!Bitset methodsFor: 'bit manipulation' stamp: 'ul 8/22/2015 00:55'!
bitAt: anInteger put: aBit
	"Set the value corresponding to anInteger to aBit. Return the new value."

	aBit caseOf: {
		[ 0 ] -> [ self clearBitAt: anInteger ].
		[ 1 ] -> [ self setBitAt: anInteger ] }.
	^aBit
		
		! !

!Bitset methodsFor: 'private' stamp: 'ul 8/20/2015 02:14'!
bytes

	^bytes! !

!Bitset methodsFor: 'accessing' stamp: 'ul 8/22/2015 00:57'!
capacity
	"Return the highest integer this collection can store plus one."

	^bytes size * 8! !

!Bitset methodsFor: 'bit manipulation' stamp: 'ul 8/22/2015 00:54'!
clearBitAt: anInteger
	"Set the value corresponding to anInteger to 0. Return true if the value wasn't 0."

	| index value mask newValue |
	index := (anInteger bitShift: -3) + 1.
	value := bytes at: index.
	mask := 1 bitShift: (anInteger bitAnd: 7).
	(newValue := (value bitOr: mask) - mask) = value ifTrue: [ ^false ].
	bytes at: index put: newValue.
	tally := tally - 1.
	^true
! !

!Bitset methodsFor: 'enumerating' stamp: 'ul 8/22/2015 01:26'!
do: aBlock
	"Evaluate aBlock with each integer which has its bit set to 1."

	| byte byteOffset lowBits remainingBits |
	remainingBits := tally.
	lowBits := Integer lowBitPerByteTable.
	1 to: bytes size do: [ :index |
		1 <= remainingBits ifFalse: [ ^self ].
		(byte := bytes at: index) = 0 ifFalse: [
			byteOffset := (index bitShift: 3) - 9. "- 8 - 1 to make it -1 based."
			[
				aBlock value: (lowBits at: byte) + byteOffset. "byteOffset is -1 based, lowBits is 1-based."
				remainingBits := remainingBits - 1.
				"Eliminate the low bit and loop if there're any remaning bits set."
				(byte := byte bitAnd: byte - 1) = 0 ] whileFalse ] ]! !

!Bitset methodsFor: 'comparing' stamp: 'ul 8/22/2015 00:59'!
hash
	"#hash is implemented, because #= is implemented."

	^(self species hash bitXor: tally hashMultiply) bitXor: bytes hash! !

!Bitset methodsFor: 'testing' stamp: 'ul 8/22/2015 01:54'!
includes: anInteger

	anInteger isInteger ifFalse: [ ^false ].
	-1 < anInteger ifFalse: [ ^false ].
	anInteger < self capacity ifFalse: [ ^false ].
	^(self bitAt: anInteger) = 1! !

!Bitset methodsFor: 'private' stamp: 'ul 8/22/2015 01:04'!
initialize: capacity
	"Capacity is expected to be a non-negative, multiple-of-eight integer."

	bytes := ByteArray new: capacity // 8.
	tally := 0! !

!Bitset methodsFor: 'enumerating' stamp: 'ul 3/9/2016 21:54'!
occurrencesOf: anObject
	"Answer how many of the receiver's elements are equal to anObject. Optimized version."

	(self includes: anObject) ifTrue: [ ^1 ].
	^0! !

!Bitset methodsFor: 'copying' stamp: 'ul 8/22/2015 01:00'!
postCopy
	"Copy bytes as well."

	bytes := bytes copy! !

!Bitset methodsFor: 'removing' stamp: 'ul 8/22/2015 01:08'!
removeAll

	tally = 0 ifTrue: [ ^self ].
	bytes atAllPut: 0. "Unlike most #removeAll implementations, we don't allocate a new ByteArray here, because this is a bit more efficient. The VM would have to fill the new array with zeroes anyway."
	tally := 0! !

!Bitset methodsFor: 'removing' stamp: 'ul 8/20/2015 01:42'!
remove: anInteger ifAbsent: absentBlock

	(self clearBitAt: anInteger) ifTrue: [ ^anInteger ].
	^absentBlock value! !

!Bitset methodsFor: 'bit manipulation' stamp: 'ul 8/22/2015 00:54'!
setBitAt: anInteger
	"Set the value corresponding to anInteger to 1. Return true if the value wasn't 1."

	| index value newValue |
	index := (anInteger bitShift: -3) + 1.
	value := bytes at: index.
	(newValue := (1 bitShift: (anInteger bitAnd: 7)) bitOr: value) = value ifTrue: [ ^false ].
	bytes at: index put: newValue.
	tally := tally + 1.
	^true! !

!Bitset methodsFor: 'accessing' stamp: 'ul 8/22/2015 00:56'!
size
	"Return the number of 1 values in this collection."

	^tally! !

!Bitset methodsFor: 'comparing' stamp: 'ul 8/20/2015 02:14'!
= anObject

	self species == anObject species ifFalse: [ ^false ].
	anObject size = tally ifFalse: [ ^false ].
	^anObject bytes = bytes! !

Collection subclass: #CharacterSet
	instanceVariableNames: 'map tally'
	classVariableNames: 'CrLf NonSeparators Separators'
	poolDictionaries: ''
	category: 'Collections-Support'!

!CharacterSet commentStamp: '<historical>' prior: 0!
A set of characters.  Lookups for inclusion are very fast.!

!CharacterSet class methodsFor: 'instance creation' stamp: 'ls 8/17/1998 20:42'!
allCharacters
	"return a set containing all characters"

	| set |
	set := self empty.
	0 to: 255 do: [ :ascii | set add: (Character value: ascii) ].
	^set! !

!CharacterSet class methodsFor: 'initialize-release' stamp: 'ul 4/1/2015 22:53'!
cleanUp: aggressive

	CrLf := NonSeparators := Separators := nil! !

!CharacterSet class methodsFor: 'accessing' stamp: 'ul 4/1/2015 22:53'!
crlf

	^CrLf ifNil: [ CrLf := self with: Character cr with: Character lf ]! !

!CharacterSet class methodsFor: 'instance creation' stamp: 'nk 8/3/2004 06:54'!
empty
 	"return an empty set of characters"
	^self new! !

!CharacterSet class methodsFor: 'instance creation' stamp: 'ls 1/3/1999 12:52'!
newFrom: aCollection
	| newCollection |
	newCollection := self new.
	newCollection addAll: aCollection.
	^newCollection! !

!CharacterSet class methodsFor: 'accessing' stamp: 'ul 4/1/2015 19:41'!
nonSeparators
	"return a set containing everything but the whitespace characters"

	^NonSeparators ifNil: [
		NonSeparators := self separators complement ]! !

!CharacterSet class methodsFor: 'accessing' stamp: 'ul 4/2/2015 00:01'!
separators
	"return a set containing just the whitespace characters"

	^Separators ifNil: [ Separators := self newFrom: Character separators ]! !

!CharacterSet class methodsFor: 'instance creation' stamp: 'ul 4/1/2015 23:30'!
withAll: aCollection
	"Create a new CharacterSet containing all the characters from aCollection."

	^self newFrom: aCollection! !

!CharacterSet methodsFor: 'collection ops' stamp: 'ul 8/23/2015 23:57'!
add: aCharacter
	"I automatically become a WideCharacterSet if you add a wide character to myself"
	
	| index |
	(index := aCharacter asInteger + 1) <= 256 ifFalse: [
		| wide |
		wide := WideCharacterSet new.
		wide addAll: self.
		wide add: aCharacter.
		self becomeForward: wide.
		^aCharacter ].
	(map at: index) = 1 ifFalse: [
		map at: index put: 1.
		tally := tally + 1 ].
	^aCharacter! !

!CharacterSet methodsFor: 'conversion' stamp: 'ar 10/25/2010 18:54'!
asString
	"Convert the receiver into a String"

	^String new: self size streamContents:[:s|
		self do:[:ch| s nextPut: ch].
	].! !

!CharacterSet methodsFor: 'private' stamp: 'ls 8/17/1998 20:35'!
byteArrayMap
	"return a ByteArray mapping each ascii value to a 1 if that ascii value is in the set, and a 0 if it isn't.  Intended for use by primitives only"
	^map! !

!CharacterSet methodsFor: 'conversion' stamp: 'nice 3/23/2007 02:28'!
byteComplement
	"return a character set containing precisely the single byte characters the receiver does not"
	
	| set |
	set := CharacterSet allCharacters.
	self do: [ :c | set remove: c ].
	^set! !

!CharacterSet methodsFor: 'conversion' stamp: 'nice 11/20/2007 00:19'!
complement
	"return a character set containing precisely the characters the receiver does not"
	
	^CharacterSetComplement of: self copy! !

!CharacterSet methodsFor: 'collection ops' stamp: 'ul 8/23/2015 23:58'!
do: aBlock
	"evaluate aBlock with each character in the set"

	| index |
	tally >= 128 ifTrue: [ "dense"
		index := 0.
		[ (index := index + 1) <= 256 ] whileTrue: [
			(map at: index) = 1 ifTrue: [
				aBlock value: (Character value: index - 1) ] ].
		^self ].
	"sparse"
	index := 0.
	[ (index := map indexOf: 1 startingAt: index + 1) = 0 ] whileFalse: [
		aBlock value: (Character value: index - 1) ].
	! !

!CharacterSet methodsFor: 'collection ops' stamp: 'nice 12/9/2009 15:29'!
findFirstInByteString: aByteString startingAt: startIndex
	"Double dispatching: since we know this is a ByteString, we can use a superfast primitive using a ByteArray map with 0 slots for byte characters not included and 1 for byte characters included in the receiver."
	^ByteString
		findFirstInString: aByteString
		inSet: self byteArrayMap
		startingAt: startIndex! !

!CharacterSet methodsFor: 'comparison' stamp: 'ls 8/17/1998 20:46'!
hash
	^self byteArrayMap hash! !

!CharacterSet methodsFor: 'testing' stamp: 'nice 5/9/2006 23:23'!
hasWideCharacters
	^false! !

!CharacterSet methodsFor: 'collection ops' stamp: 'ul 3/10/2016 14:07'!
includes: anObject

	| index |
	anObject isCharacter ifFalse: [ ^false ].
	(index := anObject asInteger + 1) > 256 ifTrue: [ ^false ].
	^(map at: index) > 0! !

!CharacterSet methodsFor: 'private' stamp: 'ul 8/23/2015 22:05'!
initialize

	map := ByteArray new: 256.
	tally := 0! !

!CharacterSet methodsFor: 'enumerating' stamp: 'ul 3/9/2016 21:54'!
occurrencesOf: anObject
	"Answer how many of the receiver's elements are equal to anObject. Optimized version."

	(self includes: anObject) ifTrue: [ ^1 ].
	^0! !

!CharacterSet methodsFor: 'copying' stamp: 'nice 10/5/2009 08:52'!
postCopy
	super postCopy.
	map := map copy! !

!CharacterSet methodsFor: 'removing' stamp: 'ul 8/23/2015 22:19'!
removeAll

	map atAllPut: 0.
	tally := 0! !

!CharacterSet methodsFor: 'collection ops' stamp: 'ul 8/23/2015 22:16'!
remove: aCharacter

	^self remove: aCharacter ifAbsent: aCharacter! !

!CharacterSet methodsFor: 'collection ops' stamp: 'ul 8/23/2015 23:58'!
remove: aCharacter ifAbsent: aBlock

	| index |
	(index := aCharacter asciiValue + 1) <= 256 ifFalse: [ ^aBlock value ].
	(map at: index) = 0 ifTrue: [ ^aBlock value ].
	map at: index put: 0.
	tally := tally - 1.
	^aCharacter! !

!CharacterSet methodsFor: 'collection ops' stamp: 'ul 8/23/2015 23:58'!
size

	^tally! !

!CharacterSet methodsFor: 'comparison' stamp: 'tk 7/5/2001 21:57'!
species
	^CharacterSet! !

!CharacterSet methodsFor: 'private' stamp: 'nice 5/9/2006 23:22'!
wideCharacterMap
	"used for comparing with WideCharacterSet"
	
	| wide |
	wide := WideCharacterSet new.
	wide addAll: self.
	^wide wideCharacterMap! !

!CharacterSet methodsFor: 'comparison' stamp: 'ul 8/23/2015 23:56'!
= anObject
	
	self species == anObject species ifFalse: [ ^false ].
	anObject size = tally ifFalse: [ ^false ].
	^self byteArrayMap = anObject byteArrayMap! !

Collection subclass: #CharacterSetComplement
	instanceVariableNames: 'absent byteArrayMapCache'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Support'!

!CharacterSetComplement commentStamp: 'nice 8/31/2008 14:53' prior: 0!
CharacterSetComplement is a space efficient implementation of (CharacterSet complement) taking care of WideCharacter (code > 255)

However, it will maintain a byteArrayMap for character <= 255 in a cache keeping 

instance variables:
	absent <CharacterSet> contains character that are not in the set (i.e. my complement)
	byteArrayMapCache <ByteArray | nil> cache this information because it has to be used in tight loops where efficiency matters!

!CharacterSetComplement class methodsFor: 'instance creation' stamp: 'nice 9/16/2011 21:26'!
of: aCharacterSet
	"answer the complement of aCharacterSet"
	
	^ self new complement: aCharacterSet! !

!CharacterSetComplement methodsFor: 'collection ops' stamp: 'nice 8/31/2008 14:56'!
add: aCharacter 
	"a character is present if not absent, so adding a character is removing it from the absent"
	
	(absent includes: aCharacter)
		ifTrue:
			[byteArrayMapCache := nil.
			absent remove: aCharacter].
	^ aCharacter! !

!CharacterSetComplement methodsFor: 'private' stamp: 'nice 8/31/2008 14:28'!
byteArrayMap
	"return a ByteArray mapping each ascii value to a 1 if that ascii value is in the set, and a 0 if it isn't.  Intended for use by primitives only"

	^byteArrayMapCache ifNil: [byteArrayMapCache := absent byteArrayMap collect: [:i | 1 - i]]! !

!CharacterSetComplement methodsFor: 'converting' stamp: 'nice 3/23/2007 02:08'!
complement
	"return a character set containing precisely the characters the receiver does not"
	
	^absent copy! !

!CharacterSetComplement methodsFor: 'initialize-release' stamp: 'nice 8/31/2008 14:56'!
complement: aCharacterSet
	"initialize with the complement"
	
	byteArrayMapCache := nil.
	absent := aCharacterSet.
	! !

!CharacterSetComplement methodsFor: 'collection ops' stamp: 'nice 3/23/2007 02:11'!
do: aBlock
	"evaluate aBlock with each character in the set.
	don't do it, there are too many..."

	self shouldNotImplement! !

!CharacterSetComplement methodsFor: 'collection ops' stamp: 'nice 12/11/2009 09:03'!
findFirstInByteString: aByteString startingAt: startIndex
	"Double dispatching: since we know this is a ByteString, we can use a superfast primitive using a ByteArray map with 0 slots for byte characters not included and 1 for byte characters included in the receiver."
	^ByteString
		findFirstInString: aByteString
		inSet: self byteArrayMap
		startingAt: startIndex! !

!CharacterSetComplement methodsFor: 'comparing' stamp: 'nice 11/1/2007 19:39'!
hash
	^ absent hash bitXor: self class hash! !

!CharacterSetComplement methodsFor: 'testing' stamp: 'nice 3/23/2007 02:12'!
hasWideCharacters
	"This is a guess that absent is not holding each and every possible wideCharacter..."
	
	^true! !

!CharacterSetComplement methodsFor: 'collection ops' stamp: 'ul 3/10/2016 10:44'!
includes: anObject

	anObject isCharacter ifFalse: [ ^false ].
	(absent includes: anObject) ifTrue: [ ^false ].
	^true! !

!CharacterSetComplement methodsFor: 'enumerating' stamp: 'ul 3/9/2016 21:55'!
occurrencesOf: anObject
	"Answer how many of the receiver's elements are equal to anObject. Optimized version."

	(self includes: anObject) ifTrue: [ ^1 ].
	^0! !

!CharacterSetComplement methodsFor: 'copying' stamp: 'nice 10/5/2009 08:52'!
postCopy
	super postCopy.
	absent := absent copy! !

!CharacterSetComplement methodsFor: 'printing' stamp: 'nice 11/19/2007 23:54'!
printOn: aStream
	"Print a description of the complement rather than self.
	Rationale: self would be too long to print."
	
	aStream nextPut: $(.
	absent printOn: aStream.
	aStream nextPut: $); space; nextPutAll: #complement.! !

!CharacterSetComplement methodsFor: 'collection ops' stamp: 'nice 3/23/2007 02:15'!
reject: aBlock
	"Implementation note: rejecting present is selecting absent"
	
	^(absent select: aBlock) complement! !

!CharacterSetComplement methodsFor: 'collection ops' stamp: 'ul 9/28/2011 14:04'!
removeAll

	self becomeForward: CharacterSet new! !

!CharacterSetComplement methodsFor: 'collection ops' stamp: 'nice 8/31/2008 14:54'!
remove: aCharacter
	"This means aCharacter is now absent from myself.
	It must be added to my absent."
	
	byteArrayMapCache := nil.
	^absent add: aCharacter! !

!CharacterSetComplement methodsFor: 'collection ops' stamp: 'nice 12/10/2009 19:20'!
remove: aCharacter ifAbsent: aBlock
	(self includes: aCharacter) ifFalse: [^aBlock value].
	^self remove: aCharacter! !

!CharacterSetComplement methodsFor: 'collection ops' stamp: 'nice 3/23/2007 02:15'!
select: aBlock
	"Implementation note: selecting present is rejecting absent"
	
	^(absent reject: aBlock) complement! !

!CharacterSetComplement methodsFor: 'collection ops' stamp: 'nice 3/23/2007 02:10'!
size
	"Is this 2**32-absent size ?"
	
	^self shouldNotImplement! !

!CharacterSetComplement methodsFor: 'printing' stamp: 'nice 11/19/2007 23:55'!
storeOn: aStream
	"Store a description of the elements of the complement rather than self."
	
	aStream nextPut: $(.
	absent storeOn: aStream.
	aStream nextPut: $); space; nextPutAll: #complement.! !

!CharacterSetComplement methodsFor: 'comparing' stamp: 'nice 3/23/2007 02:19'!
= anObject
	"Implementation note: we do not test if equal to a WideCharacterSet,
	because it is unlikely that WideCharacterSet is as complete as self"
	
	^self class == anObject class and: [
		absent = anObject complement ]! !

!Collection class methodsFor: 'class initialization' stamp: 'ul 5/5/2015 01:01'!
initialize
	"Set up a Random number generator to be used by atRandom when the 
	user does not feel like creating his own Random generator."

	Smalltalk addToStartUpList: self! !

!Collection class methodsFor: 'instance creation' stamp: 'apb 10/15/2000 22:05'!
ofSize: n
	"Create a new collection of size n with nil as its elements.
	This method exists because OrderedCollection new: n creates an
	empty collection,  not one of size n."
	^ self new: n! !

!Collection class methodsFor: 'instance creation' stamp: 'sma 5/6/2000 20:07'!
withAll: aCollection
	"Create a new collection containing all the elements from aCollection."

	^ (self new: aCollection size)
		addAll: aCollection;
		yourself! !

!Collection class methodsFor: 'instance creation' stamp: 'sma 5/6/2000 19:58'!
with: anObject 
	"Answer an instance of me containing anObject."

	^ self new
		add: anObject;
		yourself! !

!Collection class methodsFor: 'instance creation' stamp: 'sma 5/6/2000 20:01'!
with: firstObject with: secondObject 
	"Answer an instance of me containing the two arguments as elements."

	^ self new
		add: firstObject;
		add: secondObject;
		yourself! !

!Collection class methodsFor: 'instance creation' stamp: 'sma 5/6/2000 20:03'!
with: firstObject with: secondObject with: thirdObject 
	"Answer an instance of me containing the three arguments as elements."

	^ self new
		add: firstObject;
		add: secondObject;
		add: thirdObject;
		yourself! !

!Collection class methodsFor: 'instance creation' stamp: 'sma 5/6/2000 20:06'!
with: firstObject with: secondObject with: thirdObject with: fourthObject 
	"Answer an instance of me, containing the four arguments as the elements."

	^ self new
		add: firstObject;
		add: secondObject;
		add: thirdObject;
		add: fourthObject;
		yourself! !

!Collection class methodsFor: 'instance creation' stamp: 'sma 5/6/2000 20:06'!
with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject
	"Answer an instance of me, containing the five arguments as the elements."

	^ self new
		add: firstObject;
		add: secondObject;
		add: thirdObject;
		add: fourthObject;
		add: fifthObject;
		yourself! !

!Collection class methodsFor: 'instance creation' stamp: 'sma 5/6/2000 20:06'!
with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject with: sixthObject
	"Answer an instance of me, containing the six arguments as the elements."

	^ self new
		add: firstObject;
		add: secondObject;
		add: thirdObject;
		add: fourthObject;
		add: fifthObject;
		add: sixthObject;
		yourself! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 15:51'!
abs
	"Absolute value of all elements in the collection"
	^ self collect: [:a | a abs]! !

!Collection methodsFor: 'adapting' stamp: 'di 11/6/1998 13:34'!
adaptToCollection: rcvr andSend: selector
	"If I am involved in arithmetic with another Collection, return a Collection of
	the results of each element combined with the scalar in that expression."

	rcvr isSequenceable & self isSequenceable ifFalse:
		[self error: 'Only sequenceable collections may be combined arithmetically'].
	^ rcvr with: self collect:
		[:rcvrElement :myElement | rcvrElement perform: selector with: myElement]! !

!Collection methodsFor: 'adapting' stamp: 'mk 10/27/2003 21:48'!
adaptToComplex: rcvr andSend: selector
	"If I am involved in arithmetic with a scalar, return a Collection of
	the results of each element combined with the scalar in that expression."

	^ self collect: [:element | rcvr perform: selector with: element]! !

!Collection methodsFor: 'adapting' stamp: 'di 11/9/1998 12:16'!
adaptToNumber: rcvr andSend: selector
	"If I am involved in arithmetic with a scalar, return a Collection of
	the results of each element combined with the scalar in that expression."

	^ self collect: [:element | rcvr perform: selector with: element]! !

!Collection methodsFor: 'adapting' stamp: 'di 11/6/1998 13:37'!
adaptToPoint: rcvr andSend: selector
	"If I am involved in arithmetic with a scalar, return a Collection of
	the results of each element combined with the scalar in that expression."

	^ self collect: [:element | rcvr perform: selector with: element]! !

!Collection methodsFor: 'adapting' stamp: 'di 11/6/1998 13:37'!
adaptToString: rcvr andSend: selector
	"If I am involved in arithmetic with a String, convert it to a Number."
	^ rcvr asNumber perform: selector with: self! !

!Collection methodsFor: 'adding' stamp: 'ul 4/12/2015 23:52'!
addAllFirstTo: anOrderedCollection
	"Add all of my elements to the beginning of anOrderedCollection"

	self do: [ :each | anOrderedCollection addFirst: each ]! !

!Collection methodsFor: 'adding' stamp: 'sma 5/12/2000 17:26'!
addAll: aCollection 
	"Include all the elements of aCollection as the receiver's elements. Answer 
	aCollection. Actually, any object responding to #do: can be used as argument."

	aCollection do: [:each | self add: each].
	^ aCollection! !

!Collection methodsFor: 'adding' stamp: 'sma 5/12/2000 17:23'!
addIfNotPresent: anObject
	"Include anObject as one of the receiver's elements, but only if there
	is no such element already. Anwser anObject."

	(self includes: anObject) ifFalse: [self add: anObject].
	^ anObject! !

!Collection methodsFor: 'adding' stamp: ''!
add: newObject 
	"Include newObject as one of the receiver's elements. Answer newObject. 
	ArrayedCollections cannot respond to this message."

	self subclassResponsibility! !

!Collection methodsFor: 'adding' stamp: 'ul 4/12/2010 22:33'!
add: newObject withOccurrences: anInteger
	"Add newObject anInteger times to the receiver. Do nothing if anInteger is less than one. Answer newObject."

	anInteger timesRepeat: [self add: newObject].
	^ newObject! !

!Collection methodsFor: 'enumerating' stamp: 'sma 4/30/2000 11:17'!
allSatisfy: aBlock
	"Evaluate aBlock with the elements of the receiver.
	If aBlock returns false for any element return false.
	Otherwise return true."

	self do: [:each | (aBlock value: each) ifFalse: [^ false]].
	^ true! !

!Collection methodsFor: 'accessing' stamp: 'sma 5/12/2000 11:33'!
anyOne
	"Answer a representative sample of the receiver. This method can
	be helpful when needing to preinfer the nature of the contents of 
	semi-homogeneous collections."

	self emptyCheck.
	self do: [:each | ^ each]! !

!Collection methodsFor: 'enumerating' stamp: 'sma 4/30/2000 11:17'!
anySatisfy: aBlock
	"Evaluate aBlock with the elements of the receiver.
	If aBlock returns true for any element return true.
	Otherwise return false."

	self do: [:each | (aBlock value: each) ifTrue: [^ true]].
	^ false! !

!Collection methodsFor: 'math functions' stamp: 'raok 10/22/2002 00:20'!
arcCos
	^self collect: [:each | each arcCos]! !

!Collection methodsFor: 'math functions' stamp: 'raok 10/22/2002 00:20'!
arcSin
	^self collect: [:each | each arcSin]! !

!Collection methodsFor: 'math functions' stamp: 'raok 10/22/2002 00:20'!
arcTan
	^self collect: [:each | each arcTan]! !

!Collection methodsFor: 'converting' stamp: 'sma 5/6/2000 20:22'!
asArray
	"Answer an Array whose elements are the elements of the receiver.
	Implementation note: Cannot use ''Array withAll: self'' as that only
	works for SequenceableCollections which support the replacement 
	primitive."

	| array index |
	array := Array new: self size.
	index := 0.
	self do: [:each | array at: (index := index + 1) put: each].
	^ array! !

!Collection methodsFor: 'converting' stamp: 'sma 5/6/2000 20:10'!
asBag
	"Answer a Bag whose elements are the elements of the receiver."

	^ Bag withAll: self! !

!Collection methodsFor: 'converting' stamp: 'sma 5/6/2000 20:22'!
asByteArray
	"Answer a ByteArray whose elements are the elements of the receiver.
	Implementation note: Cannot use ''ByteArray withAll: self'' as that only
	works for SequenceableCollections which support the replacement 
	primitive."

	| array index |
	array := ByteArray new: self size.
	index := 0.
	self do: [:each | array at: (index := index + 1) put: each].
	^ array! !

!Collection methodsFor: 'converting' stamp: 'sma 5/6/2000 20:26'!
asCharacterSet
	"Answer a CharacterSet whose elements are the unique elements of the receiver.
	The reciever should only contain characters."

	^ CharacterSet newFrom: self! !

!Collection methodsFor: 'converting' stamp: 'brp 5/9/2005 15:58'!
asCollection

	^ self! !

!Collection methodsFor: 'printing' stamp: 'gk 1/14/2005 08:19'!
asCommaString
	"Return collection printed as 'a, b, c' "

	^String streamContents: [:s | self asStringOn: s delimiter: ', ']
		! !

!Collection methodsFor: 'printing' stamp: 'gk 1/14/2005 08:20'!
asCommaStringAnd
	"Return collection printed as 'a, b and c' "

	^String streamContents: [:s | self asStringOn: s delimiter: ', ' last: ' and ']
		! !

!Collection methodsFor: 'converting' stamp: 'ar 9/22/2000 10:12'!
asIdentitySet
	^(IdentitySet new: self size) addAll: self; yourself! !

!Collection methodsFor: 'converting' stamp: 'sma 5/12/2000 17:43'!
asOrderedCollection
	"Answer an OrderedCollection whose elements are the elements of the
	receiver. The order in which elements are added depends on the order
	in which the receiver enumerates its elements. In the case of unordered
	collections, the ordering is not necessarily the same for multiple 
	requests for the conversion."

	^ self as: OrderedCollection! !

!Collection methodsFor: 'converting' stamp: 'sma 5/6/2000 20:29'!
asSet
	"Answer a Set whose elements are the unique elements of the receiver."

	^ Set withAll: self! !

!Collection methodsFor: 'converting' stamp: ''!
asSortedArray
	"Return a copy of the receiver in sorted order, as an Array.  6/10/96 sw"

	^ self asSortedCollection asArray! !

!Collection methodsFor: 'converting' stamp: 'ul 3/22/2010 17:26'!
asSortedCollection
	"Answer a SortedCollection whose elements are the elements of the 
	receiver. The sort order is the default less than or equal. 
	Use #sorted: if you don't really need a SortedCollection, but a sorted collection!!"
	
	^self as: SortedCollection! !

!Collection methodsFor: 'converting' stamp: 'ul 3/22/2010 17:26'!
asSortedCollection: aSortBlock 
	"Answer a SortedCollection whose elements are the elements of the 
	receiver. The sort order is defined by the argument, aSortBlock.
	Use #sorted: if you don't really need a SortedCollection, but a sorted collection!!"

	| aSortedCollection |
	aSortedCollection := SortedCollection new: self size.
	aSortedCollection sortBlock: aSortBlock.
	aSortedCollection addAll: self.
	^ aSortedCollection! !

!Collection methodsFor: 'printing' stamp: 'gk 1/14/2005 08:27'!
asStringOn: aStream delimiter: delimString
	"Print elements on a stream separated
	with a delimiter String like: 'a, b, c'
	Uses #asString instead of #print:."

	self do: [:elem | aStream nextPutAll: elem asString]
		separatedBy: [aStream nextPutAll: delimString]! !

!Collection methodsFor: 'printing' stamp: 'gk 1/14/2005 08:27'!
asStringOn: aStream delimiter: delimString last: lastDelimString
	"Print elements on a stream separated
	with a delimiter between all the elements and with
	a special one before the last like: 'a, b and c'.
	Uses #asString instead of #print:

	Note: Feel free to improve the code to detect the last element."

	| n sz |
	n := 1.
	sz := self size.
	self do: [:elem |
		n := n + 1.
		aStream nextPutAll: elem asString]
	separatedBy: [
		aStream nextPutAll: (n = sz ifTrue: [lastDelimString] ifFalse: [delimString])]! !

!Collection methodsFor: 'enumerating' stamp: ''!
associationsDo: aBlock
	"Evaluate aBlock for each of the receiver's elements (key/value 
	associations).  If any non-association is within, the error is not caught now,
	but later, when a key or value message is sent to it."

	self do: aBlock! !

!Collection methodsFor: 'accessing' stamp: 'ul 12/17/2014 14:40'!
atRandom
	"Answer a random element of the receiver.  Uses the process-local random number generator. Causes an error if self has no elements."

	^self atRandom: ThreadSafeRandom value

"Examples:
	#('one' 'or' 'the' 'other') atRandom
	(1 to: 10) atRandom
	'Just pick one of these letters at random' atRandom
	#(3 7 4 9 21) asSet atRandom		(just to show it also works for Sets)
"! !

!Collection methodsFor: 'accessing' stamp: 'ul 3/15/2010 02:13'!
atRandom: aGenerator
	"Answer a random element of the receiver. Uses aGenerator which
	should be kept by the user in a variable and used every time. Use
	this instead of #atRandom for better uniformity of random numbers because 
	only you use the generator. Causes an error if self has no elements."

	| randomIndex index |
	self emptyCheck.
	randomIndex := aGenerator nextInt: self size.
	index := 1.
	self do: [ :each |
		index = randomIndex ifTrue: [ ^each ].
		index := index + 1 ]! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 15:57'!
average
	^ self sum / self size! !

!Collection methodsFor: 'accessing' stamp: 'sma 5/12/2000 11:41'!
capacity
	"Answer the current capacity of the receiver."

	^ self size! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 15:51'!
ceiling
	^ self collect: [:a | a ceiling]! !

!Collection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 11:45'!
collect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into a collection like the receiver. Answer  
	the new collection."

	| newCollection |
	newCollection := self species new.
	self do: [:each | newCollection add: (aBlock value: each)].
	^ newCollection! !

!Collection methodsFor: 'enumerating' stamp: 'ul 11/21/2009 01:16'!
collect: aBlock as: aClass
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into an instance of aClass. Answer the resulting collection."

	^(aClass new: self size) fillFrom: self with: aBlock! !

!Collection methodsFor: 'enumerating' stamp: 'ul 11/21/2009 01:16'!
collect: aBlock into: aCollection
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into aCollection. Answer aCollection."

	^aCollection fillFrom: self with: aBlock! !

!Collection methodsFor: 'enumerating' stamp: 'dgd 9/13/2004 23:42'!
collect: collectBlock thenDo: doBlock 
	"Utility method to improve readability."
	^ (self collect: collectBlock) do: doBlock! !

!Collection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 11:51'!
collect: collectBlock thenSelect: selectBlock
	"Utility method to improve readability."

	^ (self collect: collectBlock) select: selectBlock! !

!Collection methodsFor: 'testing' stamp: ''!
contains: aBlock
	"VW compatibility"
	^self anySatisfy: aBlock! !

!Collection methodsFor: 'filter streaming' stamp: 'sma 5/12/2000 12:07'!
contents
	^ self! !

!Collection methodsFor: 'copying' stamp: 'ar 2/11/2001 01:55'!
copyWithDependent: newElement
	"Answer a new collection with newElement added (as last
	element if sequenceable)."
	^self copyWith: newElement! !

!Collection methodsFor: 'copying' stamp: 'sma 5/12/2000 18:08'!
copyWithoutAll: aCollection
	"Answer a copy of the receiver that does not contain any elements 
	equal to those in aCollection."

	^ self reject: [:each | aCollection includes: each]! !

!Collection methodsFor: 'copying' stamp: 'sma 5/12/2000 14:43'!
copyWithout: oldElement 
	"Answer a copy of the receiver that does not contain any
	elements equal to oldElement."

	^ self reject: [:each | each = oldElement]

"Examples:
	'fred the bear' copyWithout: $e
	#(2 3 4 5 5 6) copyWithout: 5
"! !

!Collection methodsFor: 'copying' stamp: 'sma 5/12/2000 14:41'!
copyWith: newElement
	"Answer a new collection with newElement added (as last
	element if sequenceable)."

	^ self copy
		add: newElement;
		yourself! !

!Collection methodsFor: 'math functions' stamp: 'raok 10/22/2002 00:20'!
cos
	^self collect: [:each | each cos]! !

!Collection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 11:52'!
count: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Answer the number of elements that answered true."

	| sum |
	sum := 0.
	self do: [:each | (aBlock value: each) ifTrue: [sum := sum + 1]].
	^ sum! !

!Collection methodsFor: 'math functions' stamp: 'raok 10/22/2002 00:20'!
degreeCos
	^self collect: [:each | each degreeCos]! !

!Collection methodsFor: 'math functions' stamp: 'raok 10/22/2002 00:21'!
degreeSin
	^self collect: [:each | each degreeSin]! !

!Collection methodsFor: 'enumerating' stamp: 'nice 12/27/2009 03:10'!
detectMax: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Answer the element for which aBlock evaluates to the highest magnitude.
	If collection empty, return nil.  This method might also be called elect:."

	| maxElement maxValue |
	self do: [:each | | val | 
		maxValue == nil
			ifFalse: [
				(val := aBlock value: each) > maxValue ifTrue: [
					maxElement := each.
					maxValue := val]]
			ifTrue: ["first element"
				maxElement := each.
				maxValue := aBlock value: each].
				"Note that there is no way to get the first element that works 
				for all kinds of Collections.  Must test every one."].
	^ maxElement! !

!Collection methodsFor: 'enumerating' stamp: 'nice 12/27/2009 03:10'!
detectMin: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Answer the element for which aBlock evaluates to the lowest number.
	If collection empty, return nil."

	| minElement minValue |
	self do: [:each | | val | 
		minValue == nil
			ifFalse: [
				(val := aBlock value: each) < minValue ifTrue: [
					minElement := each.
					minValue := val]]
			ifTrue: ["first element"
				minElement := each.
				minValue := aBlock value: each].
				"Note that there is no way to get the first element that works 
				for all kinds of Collections.  Must test every one."].
	^ minElement! !

!Collection methodsFor: 'enumerating' stamp: ''!
detectSum: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Return the sum of the answers."
	| sum |
	sum := 0.
	self do: [:each | 
		sum := (aBlock value: each) + sum].  
	^ sum! !

!Collection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 11:20'!
detect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Answer the first element for which aBlock evaluates to true."

	^ self detect: aBlock ifNone: [self errorNotFound: aBlock]! !

!Collection methodsFor: 'enumerating' stamp: 'cmm 2/26/2011 22:52'!
detect: aBlock ifFound: foundBlock ifNone: exceptionBlock 
	"foundBlock takes one argument, the found object."
	self 
		do: [ :element | (aBlock value: element) ifTrue: [ ^foundBlock value: element ] ].
	^exceptionBlock value
! !

!Collection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 11:52'!
detect: aBlock ifNone: exceptionBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Answer the first element for which aBlock evaluates to true. If none  
	evaluate to true, then evaluate the argument, exceptionBlock."

	self do: [:each | (aBlock value: each) ifTrue: [^ each]].
	^ exceptionBlock value! !

!Collection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 17:52'!
difference: aCollection
	"Answer the set theoretic difference of two collections."

	^ self reject: [:each | aCollection includes: each]! !

!Collection methodsFor: 'enumerating' stamp: ''!
do: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument."

	self subclassResponsibility! !

!Collection methodsFor: 'enumerating' stamp: 'ar 12/31/2009 19:49'!
do: aBlock displayingProgress: aStringOrBlock
	"Enumerate aBlock displaying progress information. 
	If the argument is a string, use a static label for the process. 
	If the argument is a block, evaluate it with the element to retrieve the label.
		Smalltalk allClasses 
			do:[:aClass| (Delay forMilliseconds: 1) wait]
			displayingProgress: 'Processing...'.
		Smalltalk allClasses 
			do:[:aClass| (Delay forMilliseconds: 1) wait]
			displayingProgress:[:aClass| 'Processing ', aClass name].
	"
	^self do: aBlock displayingProgress: aStringOrBlock every: 20! !

!Collection methodsFor: 'enumerating' stamp: 'ul 6/17/2011 12:26'!
do: aBlock displayingProgress: aStringOrBlock every: msecs
	"Enumerate aBlock displaying progress information. 
	If the argument is a string, use a static label for the process. 
	If the argument is a block, evaluate it with the element to retrieve the label.
	The msecs argument ensures that updates happen at most every msecs.
	Example:
		Smalltalk allClasses 
			do:[:aClass| (Delay forMilliseconds: 1) wait]
			displayingProgress:[:aClass| 'Processing ', aClass name]
			every: 0.
		Smalltalk allClasses 
			do:[:aClass| (Delay forMilliseconds: 1) wait]
			displayingProgress:[:aClass| 'Processing ', aClass name]
			every: 100.
	"
	| size labelBlock count oldLabel lastUpdate |
	labelBlock := aStringOrBlock isString 
		ifTrue:[[:item| aStringOrBlock]] 
		ifFalse:[aStringOrBlock].
	oldLabel := nil.
	count := lastUpdate := 0.
	size := self size.
	'' displayProgressFrom: 0 to: size during:[:bar |
		self do:[:each| | newLabel |
			"Special handling for first and last element"
			(count = 0 or:[count+1 = size 
				or:[(Time millisecondsSince: lastUpdate) >= msecs]]) ifTrue:[
					bar value: count.
					oldLabel = (newLabel := (labelBlock value: each) ifNil:[oldLabel]) ifFalse:[
					ProgressNotification signal: '' extra: (oldLabel := newLabel).
				].
				lastUpdate := Time millisecondClockValue.
			].
			aBlock value: each.
			count := count + 1.
	]]! !

!Collection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 11:57'!
do: elementBlock separatedBy: separatorBlock
	"Evaluate the elementBlock for all elements in the receiver,
	and evaluate the separatorBlock between."

	| beforeFirst | 
	beforeFirst := true.
	self do:
		[:each |
		beforeFirst
			ifTrue: [beforeFirst := false]
			ifFalse: [separatorBlock value].
		elementBlock value: each]! !

!Collection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 11:59'!
do: aBlock without: anItem 
	"Enumerate all elements in the receiver. 
	Execute aBlock for those elements that are not equal to the given item"

	^ self do: [:each | anItem = each ifFalse: [aBlock value: each]]! !

!Collection methodsFor: 'private' stamp: ''!
emptyCheck

	self isEmpty ifTrue: [self errorEmptyCollection]! !

!Collection methodsFor: 'private' stamp: ''!
errorEmptyCollection

	self error: 'this collection is empty'! !

!Collection methodsFor: 'private' stamp: 'ar 3/9/2010 22:08'!
errorKeyNotFound: key
	"Signal KeyNotFound error"
	^(KeyNotFound key: key) signal! !

!Collection methodsFor: 'private' stamp: ''!
errorNoMatch

	self error: 'collection sizes do not match'! !

!Collection methodsFor: 'private' stamp: 'nice 11/19/2013 23:26'!
errorNotFound: anObject
	"anObject was not found in this collection, raise appropriate error."

	^(NotFound object: anObject) signal! !

!Collection methodsFor: 'private' stamp: 'yo 6/29/2004 13:14'!
errorNotKeyed

	self error: ('Instances of {1} do not respond to keyed accessing messages.' translated format: {self class name})
! !

!Collection methodsFor: 'math functions' stamp: 'raok 10/22/2002 00:21'!
exp
	^self collect: [:each | each exp]! !

!Collection methodsFor: 'private' stamp: 'ul 11/21/2009 01:14'!
fillFrom: aCollection with: aBlock
	"Evaluate aBlock with each of aCollections's elements as the argument.  
	Collect the resulting values into self. Answer self."

	aCollection do: [ :each |
		self add: (aBlock value: each) ]! !

!Collection methodsFor: 'enumerating' stamp: 'nice 12/9/2009 15:37'!
findFirstInByteString: aByteString startingAt: start
	"Find the index of first character starting at start in aByteString that is included in the receiver.
	Default is to use a naive algorithm.
	Subclasses might want to implement a more efficient scheme"

	start to: aByteString size do:
		[:index |
		(self includes: (aByteString at: index)) ifTrue: [^ index]].
	^ 0! !

!Collection methodsFor: 'filter streaming' stamp: 'sma 5/12/2000 12:08'!
flattenOnStream: aStream 
	^ aStream writeCollection: self! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 15:51'!
floor
	^ self collect: [:a | a floor]! !

!Collection methodsFor: 'enumerating' stamp: 'ul 11/7/2010 01:48'!
fold: binaryBlock
	"Evaluate the block with the first two elements of the receiver,
	 then with the result of the first evaluation and the next element,
	 and so on.  Answer the result of the final evaluation. If the receiver
	 is empty, raise an error. If the receiver has a single element, answer
	 that element."
	"#('if' 'it' 'is' 'to' 'be' 'it' 'is' 'up' 'to' 'me') fold: [:a :b | a, ' ', b]"

	^self reduce: binaryBlock! !

!Collection methodsFor: 'enumerating' stamp: 'ab 9/30/2002 19:26'!
gather: aBlock
	^ Array streamContents:
		[:stream |
		self do: [:ea | stream nextPutAll: (aBlock value: ea)]]! !

!Collection methodsFor: 'enumerating' stamp: 'ul 5/4/2015 16:16'!
groupBy: keyBlock
	"Like in SQL operation - Split the recievers contents into collections of elements for which keyBlock returns the same results, and return them."

	| result |
	result := Dictionary new.
	self do: [ :each | 
		| key |
		key := keyBlock value: each.
		(result at: key ifAbsentPut: [ OrderedCollection new ]) 
			add: each ].
	^result! !

!Collection methodsFor: 'enumerating' stamp: 'ul 5/4/2015 16:17'!
groupBy: keyBlock having: selectBlock 
	"Like in SQL operation - Split the recievers contents into collections of elements for which keyBlock returns the same results, and return those collections allowed by selectBlock."

	^(self groupBy: keyBlock) select: selectBlock! !

!Collection methodsFor: 'comparing' stamp: 'SqR 8/3/2000 13:36'!
hash
	"Answer an integer hash value for the receiver such that,
	  -- the hash value of an unchanged object is constant over time, and
	  -- two equal objects have equal hash values"

	| hash |

	hash := self species hash.
	self size <= 10 ifTrue:
		[self do: [:elem | hash := hash bitXor: elem hash]].
	^hash bitXor: self size hash! !

!Collection methodsFor: 'testing' stamp: 'ls 3/27/2000 17:25'!
identityIncludes: anObject 
	"Answer whether anObject is one of the receiver's elements."

	self do: [:each | anObject == each ifTrue: [^true]].
	^false! !

!Collection methodsFor: 'testing' stamp: 'ar 11/21/2009 00:34'!
ifEmpty: aBlock
	"Evaluate the block if I'm empty"

	self isEmpty ifTrue: [ ^aBlock value ]! !

!Collection methodsFor: 'testing' stamp: 'ul 11/11/2009 16:24'!
ifEmpty: emptyBlock ifNotEmptyDo: notEmptyBlock
	"Evaluate emptyBlock if I'm empty, notEmptyBlock otherwise"
	"Evaluate the notEmptyBlock with the receiver as its argument"

	self isEmpty ifTrue: [ ^emptyBlock value ].
	^notEmptyBlock value: self! !

!Collection methodsFor: 'testing' stamp: 'ul 11/11/2009 16:22'!
ifEmpty: emptyBlock ifNotEmpty: notEmptyBlock
	"Evaluate emptyBlock if I'm empty, notEmptyBlock otherwise"
	" If the notEmptyBlock has an argument, eval with the receiver as its argument"

	self isEmpty ifTrue: [ ^emptyBlock value ].
	^notEmptyBlock valueWithPossibleArgument: self! !

!Collection methodsFor: 'testing' stamp: 'ar 11/21/2009 00:36'!
ifNotEmptyDo: aBlock
	"Evaluate the given block with the receiver as its argument."

	self isEmpty ifFalse: [^ aBlock value: self].
! !

!Collection methodsFor: 'testing' stamp: 'ul 11/11/2009 16:25'!
ifNotEmptyDo: notEmptyBlock ifEmpty: emptyBlock
	"Evaluate emptyBlock if I'm empty, notEmptyBlock otherwise
	Evaluate the notEmptyBlock with the receiver as its argument"

	self isEmpty ifFalse: [ ^notEmptyBlock value: self ].
	^emptyBlock value! !

!Collection methodsFor: 'testing' stamp: 'ar 11/21/2009 00:36'!
ifNotEmpty: aBlock
	"Evaluate the given block unless the receiver is empty.

      If the block has an argument, eval with the receiver as its argument,
      but it might be better to use ifNotEmptyDo: to make the code easier to
      understand"

	self isEmpty ifFalse: [^ aBlock valueWithPossibleArgument: self].
! !

!Collection methodsFor: 'testing' stamp: 'ul 11/11/2009 16:25'!
ifNotEmpty: notEmptyBlock ifEmpty: emptyBlock
	"Evaluate emptyBlock if I'm empty, notEmptyBlock otherwise
	 If the notEmptyBlock has an argument, eval with the receiver as its argument"

	self isEmpty ifFalse: [ ^notEmptyBlock valueWithPossibleArgument: self ].
	^emptyBlock value! !

!Collection methodsFor: 'testing' stamp: ''!
includesAllOf: aCollection 
	"Answer whether all the elements of aCollection are in the receiver."
	aCollection do: [:elem | (self includes: elem) ifFalse: [^ false]].
	^ true! !

!Collection methodsFor: 'testing' stamp: ''!
includesAnyOf: aCollection 
	"Answer whether any element of aCollection is one of the receiver's elements."
	aCollection do: [:elem | (self includes: elem) ifTrue: [^ true]].
	^ false! !

!Collection methodsFor: 'testing' stamp: 'ul 8/14/2015 18:01'!
includesSubstringAnywhere: testString
	"Answer whether the receiver includes, anywhere in its nested structure, a string that has testString as a substring"
	self do:
		[:element |
			(element isString)
				ifTrue:
					[(element includesSubstring: testString) ifTrue: [^ true]].
			(element isCollection)
				ifTrue:
					[(element includesSubstringAnywhere: testString) ifTrue: [^ true]]].
	^ false

"#(first (second third) ((allSentMessages ('Elvis' includes:)))) includesSubstringAnywhere:  'lvi'"! !

!Collection methodsFor: 'testing' stamp: 'sma 5/12/2000 14:07'!
includes: anObject 
	"Answer whether anObject is one of the receiver's elements."

	^ self anySatisfy: [:each | each = anObject]! !

!Collection methodsFor: 'enumerating' stamp: ''!
inject: thisValue into: binaryBlock 
	"Accumulate a running value associated with evaluating the argument, 
	binaryBlock, with the current value of the argument, thisValue, and the 
	receiver as block arguments. For instance, to sum the numeric elements 
	of a collection, aCollection inject: 0 into: [:subTotal :next | subTotal + 
	next]."

	| nextValue |
	nextValue := thisValue.
	self do: [:each | nextValue := binaryBlock value: nextValue value: each].
	^nextValue! !

!Collection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 17:52'!
intersection: aCollection
	"Answer the set theoretic intersection of two collections."

	^ self select: [:each | aCollection includes: each]! !

!Collection methodsFor: 'testing' stamp: 'ar 8/17/1999 19:43'!
isCollection
	"Return true if the receiver is some sort of Collection and responds to basic collection messages such as #size and #do:"
	^true! !

!Collection methodsFor: 'testing' stamp: ''!
isEmpty
	"Answer whether the receiver contains any elements."

	^self size = 0! !

!Collection methodsFor: 'testing' stamp: 'bf 3/10/2000 09:29'!
isEmptyOrNil
	"Answer whether the receiver contains any elements, or is nil.  Useful in numerous situations where one wishes the same reaction to an empty collection or to nil"

	^ self isEmpty! !

!Collection methodsFor: 'testing' stamp: 'di 11/6/1998 09:16'!
isSequenceable
	^ false! !

!Collection methodsFor: 'testing' stamp: 'dgd 4/4/2004 12:14'!
isZero
	"Answer whether the receiver is zero"
	^ false! !

!Collection methodsFor: 'math functions' stamp: 'raok 10/22/2002 00:21'!
ln
	^self collect: [:each | each ln]! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 15:52'!
log
	^ self collect: [:each | each log]! !

!Collection methodsFor: 'math functions' stamp: 'jcg 11/16/2008 23:01'!
max
	"Answer the maximum value in the collection.  The collection must be non-empty and contain 'compatible' Magnitudes (eg: don't try this with a collection containing both Dates and Characters)."
	^ self inject: self anyOne into: [:max :each | max max: each]! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 16:00'!
median
	^ self asSortedCollection median! !

!Collection methodsFor: 'math functions' stamp: 'jcg 11/16/2008 23:02'!
min
	"Answer the minimum value in the collection.  The collection must be non-empty and contain 'compatible' Magnitudes (eg: don't try this with a collection containing both Dates and Characters)."
	^ self inject: self anyOne into: [:min :each | min min: each]! !

!Collection methodsFor: 'math functions' stamp: 'klc 3/7/2010 18:50'!
minMax
	"Scans for minimum and maximum in one pass returning the results as a two-element array"
	| min max |
	min := max := self anyOne.
	self do: [ :each |
		min := min min: each.
		max := max max: each ].
	^ Array with: min with: max! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 15:52'!
negated
	"Negated value of all elements in the collection"
	^ self collect: [:a | a negated]! !

!Collection methodsFor: 'enumerating' stamp: 'gh 9/18/2001 15:59'!
noneSatisfy: aBlock
	"Evaluate aBlock with the elements of the receiver.
	If aBlock returns false for all elements return true.
	Otherwise return false"

	self do: [:item | (aBlock value: item) ifTrue: [^ false]].
	^ true! !

!Collection methodsFor: 'testing' stamp: 'sma 5/12/2000 17:49'!
notEmpty
	"Answer whether the receiver contains any elements."

	^ self isEmpty not! !

!Collection methodsFor: 'enumerating' stamp: ''!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject."

	| tally |
	tally := 0.
	self do: [:each | anObject = each ifTrue: [tally := tally + 1]].
	^tally! !

!Collection methodsFor: 'printing' stamp: 'apb 4/21/2006 09:37'!
printElementsOn: aStream
	"The original code used #skip:, but some streams do not support that,
	 and we don't really need it."

	aStream nextPut: $(.
	self do: [:element | aStream print: element] separatedBy: [aStream space].
	aStream nextPut: $)! !

!Collection methodsFor: 'printing' stamp: 'sma 6/1/2000 09:41'!
printNameOn: aStream
	super printOn: aStream! !

!Collection methodsFor: 'printing' stamp: 'sma 6/1/2000 09:41'!
printOn: aStream 
	"Append a sequence of characters that identify the receiver to aStream."

	self printNameOn: aStream.
	self printElementsOn: aStream! !

!Collection methodsFor: 'printing' stamp: 'gk 1/14/2005 08:01'!
printOn: aStream delimiter: delimString
	"Print elements on a stream separated
	with a delimiter String like: 'a, b, c' "

	self do: [:elem | aStream print: elem] separatedBy: [aStream print: delimString]
		! !

!Collection methodsFor: 'printing' stamp: 'fbs 1/14/2005 10:54'!
printOn: aStream delimiter: delimString last: lastDelimString
	"Print elements on a stream separated
	with a delimiter between all the elements and with
	a special one before the last like: 'a, b and c'

	Note: Feel free to improve the code to detect the last element."

	| n sz |
	n := 1.
	sz := self size.
	self do: [:elem |
		n := n + 1.
		aStream print: elem]
	separatedBy: [
		n = sz
			ifTrue: [aStream print: lastDelimString]
			ifFalse: [aStream print: delimString]]! !

!Collection methodsFor: 'arithmetic' stamp: 'raok 10/22/2002 00:17'!
raisedTo: arg

	^ arg adaptToCollection: self andSend: #raisedTo:! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 16:00'!
range
	^ self max - self min! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 15:53'!
reciprocal
	"Return the reciever full of reciprocated elements"
	^ self collect: [:a | a reciprocal]! !

!Collection methodsFor: 'enumerating' stamp: 'ul 11/7/2010 01:47'!
reduce: binaryBlock
	"Apply the argument, binaryBlock cumulatively to the elements of the receiver.
	For sequenceable collections the elements will be used in order, for unordered
	collections the order is unspecified."

	| first nextValue |
	first := true.
	self do: [ :each |
		first
			ifTrue: [ nextValue := each. first := false ]
			ifFalse: [ nextValue := binaryBlock value: nextValue value: each ] ].
	first ifTrue: [ self errorEmptyCollection ].
	^nextValue! !

!Collection methodsFor: 'enumerating' stamp: ''!
reject: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Collect into a new collection like the receiver only those elements for 
	which aBlock evaluates to false. Answer the new collection."

	^self select: [:element | (aBlock value: element) == false]! !

!Collection methodsFor: 'enumerating' stamp: 'dgd 9/13/2004 23:42'!
reject: rejectBlock thenDo: doBlock 
	"Utility method to improve readability."
	^ (self reject: rejectBlock) do: doBlock! !

!Collection methodsFor: 'removing' stamp: 'nice 9/14/2009 20:30'!
removeAll
	"Remove each element from the receiver and leave it empty.
	ArrayedCollections cannot respond to this message.
	There are two good reasons why a subclass should override this message:
	1) the subclass does not support being modified while being iterated
	2) the subclass provides a much faster way than iterating through each element"

	self do: [:each | self remove: each].! !

!Collection methodsFor: 'removing' stamp: 'sma 5/12/2000 11:16'!
removeAllFoundIn: aCollection 
	"Remove each element of aCollection which is present in the receiver 
	from the receiver. Answer aCollection. No error is raised if an element
	isn't found. ArrayedCollections cannot respond to this message."

	aCollection do: [:each | self remove: each ifAbsent: []].
	^ aCollection! !

!Collection methodsFor: 'removing' stamp: 'sma 5/12/2000 11:19'!
removeAllSuchThat: aBlock 
	"Evaluate aBlock for each element and remove all that elements from
	the receiver for that aBlock evaluates to true.  Use a copy to enumerate 
	collections whose order changes when an element is removed (i.e. Sets)."

	self copy do: [:each | (aBlock value: each) ifTrue: [self remove: each]]! !

!Collection methodsFor: 'removing' stamp: 'nice 1/10/2009 00:01'!
removeAll: aCollection 
	"Remove each element of aCollection from the receiver. If successful for 
	each, answer aCollection. Otherwise create an error notification.
	ArrayedCollections cannot respond to this message."

	aCollection == self ifTrue: [^self removeAll].
	aCollection do: [:each | self remove: each].
	^ aCollection! !

!Collection methodsFor: 'removing' stamp: 'sma 5/12/2000 11:22'!
remove: oldObject 
	"Remove oldObject from the receiver's elements. Answer oldObject 
	unless no element is equal to oldObject, in which case, raise an error.
	ArrayedCollections cannot respond to this message."

	^ self remove: oldObject ifAbsent: [self errorNotFound: oldObject]! !

!Collection methodsFor: 'removing' stamp: 'sma 5/12/2000 11:14'!
remove: oldObject ifAbsent: anExceptionBlock 
	"Remove oldObject from the receiver's elements. If several of the 
	elements are equal to oldObject, only one is removed. If no element is 
	equal to oldObject, answer the result of evaluating anExceptionBlock. 
	Otherwise, answer the argument, oldObject. ArrayedCollections cannot 
	respond to this message."

	self subclassResponsibility! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 15:53'!
rounded
	^ self collect: [:a | a rounded]! !

!Collection methodsFor: 'math functions' stamp: 'nk 12/30/2003 15:47'!
roundTo: quantum
	^self collect: [ :ea | ea roundTo: quantum ]! !

!Collection methodsFor: 'enumerating' stamp: ''!
select: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Collect into a new collection like the receiver, only those elements for 
	which aBlock evaluates to true. Answer the new collection."

	| newCollection |
	newCollection := self species new.
	self do: [:each | (aBlock value: each) ifTrue: [newCollection add: each]].
	^newCollection! !

!Collection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 11:59'!
select: selectBlock thenCollect: collectBlock
	"Utility method to improve readability."

	^ (self select: selectBlock) collect: collectBlock! !

!Collection methodsFor: 'enumerating' stamp: 'dgd 9/13/2004 23:42'!
select: selectBlock thenDo: doBlock 
	"Utility method to improve readability."
	^ (self select: selectBlock) do: doBlock! !

!Collection methodsFor: 'math functions' stamp: 'raok 10/22/2002 00:23'!
sign
	^self collect: [:each | each sign]! !

!Collection methodsFor: 'math functions' stamp: 'raok 10/22/2002 00:22'!
sin
	^self collect: [:each | each sin]! !

!Collection methodsFor: 'accessing' stamp: 'sma 5/12/2000 11:34'!
size
	"Answer how many elements the receiver contains."

	| tally |
	tally := 0.
	self do: [:each | tally := tally + 1].
	^ tally! !

!Collection methodsFor: 'sorting' stamp: 'ul 3/22/2010 12:16'!
sorted
	"Return a new sequenceable collection which contains the same elements as self but its elements are sorted in ascending order using the #'<=' operator."

	^self sorted: nil! !

!Collection methodsFor: 'sorting' stamp: 'ul 3/22/2010 12:13'!
sorted: aSortBlockOrNil
	"Return a new sequenceable collection which contains the same elements as self but its elements are sorted by aSortBlockOrNil. The block should take two arguments and return true if the first element should preceed the second one. If aSortBlock is nil then <= is used for comparison."

	^self asArray sort: aSortBlockOrNil! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 15:53'!
sqrt
	^ self collect: [:each | each sqrt]! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 15:53'!
squared
	^ self collect: [:each | each * each]! !

!Collection methodsFor: 'printing' stamp: ''!
storeOn: aStream 
	"Refer to the comment in Object|storeOn:."

	| noneYet |
	aStream nextPutAll: '(('.
	aStream nextPutAll: self class name.
	aStream nextPutAll: ' new)'.
	noneYet := true.
	self do: 
		[:each | 
		noneYet
			ifTrue: [noneYet := false]
			ifFalse: [aStream nextPut: $;].
		aStream nextPutAll: ' add: '.
		aStream store: each].
	noneYet ifFalse: [aStream nextPutAll: '; yourself'].
	aStream nextPut: $)! !

!Collection methodsFor: 'math functions' stamp: 'ar 5/4/2010 08:38'!
sum
	"Compute the sum of all the elements in the receiver"

	^self reduce:[:a :b| a + b]! !

!Collection methodsFor: 'math functions' stamp: 'raok 10/22/2002 00:22'!
tan
	^self collect: [:each | each tan]! !

!Collection methodsFor: 'converting' stamp: 'hg 12/26/2001 23:53'!
topologicallySortedUsing: aSortBlock 
	"Answer a SortedCollection whose elements are the elements of the 
	receiver, but topologically sorted. The topological order is defined 
	by the argument, aSortBlock."

	| aSortedCollection |
	aSortedCollection := SortedCollection new: self size.
	aSortedCollection sortBlock: aSortBlock.
	self do: [:each | aSortedCollection addLast: each].	"avoids sorting"
	^ aSortedCollection sortTopologically
! !

!Collection methodsFor: 'math functions' stamp: 'TAG 11/6/1998 15:54'!
truncated
	^ self collect: [:a | a truncated]! !

!Collection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 17:54'!
union: aCollection
	"Answer the set theoretic union of two collections."

	^ self asSet addAll: aCollection; yourself! !

!Collection methodsFor: 'filter streaming' stamp: 'sma 5/12/2000 12:07'!
write: anObject 
	^ self add: anObject! !

!Collection methodsFor: 'arithmetic' stamp: 'di 11/6/1998 13:53'!
* arg

	^ arg adaptToCollection: self andSend: #*! !

!Collection methodsFor: 'arithmetic' stamp: 'di 11/6/1998 13:53'!
+ arg

	^ arg adaptToCollection: self andSend: #+! !

!Collection methodsFor: 'copying' stamp: 'brp 5/9/2005 15:59'!
, aCollection

	^ self copy addAll: aCollection asCollection; yourself! !

!Collection methodsFor: 'arithmetic' stamp: 'di 11/6/1998 13:53'!
- arg

	^ arg adaptToCollection: self andSend: #-! !

!Collection methodsFor: 'arithmetic' stamp: 'di 11/6/1998 13:53'!
/ arg

	^ arg adaptToCollection: self andSend: #/! !

!Collection methodsFor: 'arithmetic' stamp: 'di 11/6/1998 13:54'!
// arg

	^ arg adaptToCollection: self andSend: #//! !

!Collection methodsFor: 'arithmetic' stamp: 'di 11/6/1998 13:54'!
\\ arg

	^ arg adaptToCollection: self andSend: #\\! !

Collection subclass: #HashedCollection
	instanceVariableNames: 'tally array'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Abstract'!

!HashedCollection commentStamp: 'ul 4/12/2010 22:37' prior: 0!
I am an abstract collection of objects that implement hash and equality in a consitent way. This means that whenever two objects are equal, their hashes have to be equal too. If two objects are equal then I can only store one of them. Hashes are expected to be integers (preferably SmallIntegers). I also expect that the objects contained by me do not change their hashes. If that happens, hash invariants have to be re-established, which can be done by #rehash.

Since I'm abstract, no instances of me should exist. My subclasses should implement #scanFor:, #fixCollisionsFrom: and #noCheckNoGrowFillFrom:.

Instance Variables
	array:		<ArrayedCollection> (typically Array or WeakArray)
	tally:		<Integer> (non-negative)

array
	- An array whose size is a prime number, it's non-nil elements are the elements of the collection, and whose nil elements are empty slots. There is always at least one nil. In fact I try to keep my "load" at 75% or less so that hashing will work well.

tally
	- The number of elements in the collection. The array size is always greater than this.

Implementation details:
I implement a hash table which uses open addressing with linear probing as the method of collision resolution. Searching for an element or a free slot for an element is done by #scanFor: which should return the index of the slot in array corresponding to it's argument. When an element is removed #fixCollisionsFrom: should rehash all elements in array between the original index of the removed element, wrapping around after the last slot until reaching an empty slot. My maximum load factor (75%) is hardcoded in #atNewIndex:put:, so it can only be changed by overriding that method. When my load factor reaches this limit I replace my array with a larger one (see #grow) ensuring that my load factor will be less than or equal to 50%. The new array is filled by #noCheckNoGrowFillFrom: which should use #scanForEmptySlotFor: instead of #scanFor: for better performance. I do not shrink.
!

HashedCollection subclass: #Dictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Unordered'!

!Dictionary commentStamp: 'nice 8/26/2010 22:30' prior: 0!
A Dictionary is an unordered collection of values which are indexed by arbitrary keys.
A Dictionary is accessed via #at: and #at:put: messages like a SequenceableCollection, but instead of being Integer, the keys can be any object that responds to =.
		
The = message is used to test for the existence of the key in the Dictionary.
If the key is absent, #at:put: adds a new entry to the Dictionary.

	(Dictionary new)
		at: 'foo' put: 1;
		at: 'bar' put: 8;
		yourself.

Each key is unique: storing another value with #at:put: at an already used key overwrites previously associated value.

	(Dictionary new)
		at: 'bar' put: 4;
		at: 'bar' put: 8;
		at: 'bar'.
		
The values are not necessarily unique, thus a Dictionary can also be seen as a sort of Bag with this respect.

	(Dictionary new)
		at: 'foo' put: 8;
		at: 'bar' put: 8;
		yourself.

If the key is absent, #at: raises an Error. An alternate block of code can be executed and its value returned in this case using #at:ifAbsent:.
See also #at:ifAbsentPut:.

	(Dictionary new) at: 'foo' ifAbsent: [nil].

Dictionary is implemented as a HashedCollection of Association (a value is associated to its key et vice et versa).
Being a HashedCollection enables fast random access indexed by keys.
Consequently, keys must respond to #hash (see super).

BEWARE: as for every HashedCollection, it is better to not modify an object after it is used as a Dictionary key. The reason is that this might change the #hash code, making the Dictionary unable to find corresponding entry, or make two keys equal violating uniqueness. It's progammer responsibility to take care to not modify the keys, or eventually to send #rehash to the Dictionary if they ever happen to change.

It is possible to grow or shrink a Dictionary using the messages #add: and #remove: with an Association parameter, however the prefered way to do so is using #at:put: and #removeKey:.
BEWARE: as for super, space reserved in internal storage array can grow but never shrink.

For conveniency, it is also possible to create a Dictionary out of a Collection of associations, as for example in:

    {'foo' -> 1. 'bar' -> 8} as: Dictionary.
    Dictionary withAll: {'foo' -> 1. 'bar' -> 8}.
    Dictionary new addAll: {'foo' -> 1. 'bar' -> 8}; yourself.

BEWARE: though a Sequence collection can be considered as a sequence of values with Integer keys (see #keysAndValuesDo: ), it cannot be converted into a Dictionary using these keys, and following message will fail:

	#('foo' 'bar') as: Dictionary.

Enumerating a Dictionary with #do: will only enumerate the values, not the keys.
Remember, the order of evaluation is arbitrary and can change when you grow or shrink a Dictionary.

    ({'foo' -> 1. 'bar' -> 8} as: Dictionary) do: [:each | Transcript cr; show: each printString].

For enumerating keys and values, use #keysAndValuesDo:, or use #associationsDo: to enumerate the associations.
#select: #reject: #collect: will operate on values while preserving the keys and thus answer a new Dictionary.

    ({'foo' -> 1. 'bar' -> 8} as: Dictionary) collect: [:each | each squared].

The keys and values of a Dictionary can be extracted by sending #keys and #values message.
Though the keys are theoretically a Set and values a Bag, for efficiency reasons, these messages will both return an Array of keys
and an Array of values. A neat feature is that these messages are preserving the arbitrary storage order - in other words, (aDictionary values at: 3) is the value associated to key (aDictionary keys at: 3).!

!Dictionary class methodsFor: 'instance creation' stamp: 'fbs 9/9/2011 11:50'!
newFromPairs: anArray 

	"Answer an instance of me associating (anArray at:i) to (anArray at: i+1)
	 for each odd i.  anArray must have an even number of entries."

	| newDictionary |

	newDictionary := self new: (anArray size/2).
	1 to: (anArray size-1) by: 2 do: [ :i|
		newDictionary at: (anArray at: i) put: (anArray at: i+1).
	].
	^ newDictionary

	"  Dictionary newFromPairs: {'Red' . Color red . 'Blue' . Color blue . 'Green' . Color green}. "! !

!Dictionary class methodsFor: 'instance creation' stamp: 'cbc 4/13/2016 15:21'!
newFrom: aDict 
	"Answer an instance of me containing the same associations as aDict.
	 Error if any key appears twice."
	| newDictionary |
	newDictionary := self new: aDict size.
	aDict associationsDo:
		[:x |
		(newDictionary includesKey: x key)
			ifTrue: [self error: 'Duplicate key: ', x key printString]
			ifFalse: [newDictionary add: x copy]].
	^ newDictionary

"	Dictionary newFrom: {1->#a. 2->#b. 3->#c}
	{1->#a. 2->#b. 3->#c} as: Dictionary
	Dictionary newFrom: {1->#a. 2->#b. 1->#c}
	{1->#a. 2->#b. 1->#c} as: Dictionary
"! !

!Dictionary methodsFor: 'adding' stamp: 'ar 8/24/2010 20:19'!
addAll: aCollection
	"Include all the associations of aCollection as the receiver's elements. Answer 
	aCollection. Any collection containing associations can be used as argument."

	aCollection == self ifFalse: [
		aCollection associationsDo: [:assoc| self add: assoc].
	].
	^aCollection! !

!Dictionary methodsFor: 'adding' stamp: 'ul 1/25/2010 18:26'!
add: anAssociation

	| index |
	index := self scanFor: anAssociation key.
	(array at: index)
		ifNil: [ self atNewIndex: index put: anAssociation ]
		ifNotNil: [ :element |	element value: anAssociation value ].
	^anAssociation! !

!Dictionary methodsFor: 'accessing' stamp: 'ar 11/20/2007 14:47'!
associationAt: key 
	^ self associationAt: key ifAbsent: [self errorKeyNotFound: key]! !

!Dictionary methodsFor: 'accessing' stamp: 'ul 11/20/2009 16:59'!
associationAt: key ifAbsent: aBlock 
	"Answer the association with the given key.
	If key is not found, return the result of evaluating aBlock."

	^(array at: (self scanFor: key)) ifNil: [ aBlock value ]
! !

!Dictionary methodsFor: 'accessing' stamp: 'ul 9/27/2015 23:13'!
associationClass

	^Association! !

!Dictionary methodsFor: 'accessing' stamp: 'cwp 1/18/2014 10:06'!
associationDeclareAt: aKey
	"Return an existing association, or create and return a new one.  Needed as a single message by ImageSegment.prepareToBeSaved."
	self flag: #environments.
	
	^ self associationAt: aKey ifAbsent: [
		| existing |
		(Undeclared includesKey: aKey)
			ifTrue: 
				[existing := Undeclared associationAt: aKey.
				Undeclared removeKey: aKey.
				self add: existing]
			ifFalse: 
				[self add: aKey -> false]]! !

!Dictionary methodsFor: 'accessing' stamp: 'ul 9/12/2009 17:29'!
associations
	"Answer a Collection containing the receiver's associations."
	
	^Array new: self size streamContents: [ :stream |
		self associationsDo: [ :each | stream nextPut: each ] ]! !

!Dictionary methodsFor: 'enumerating' stamp: 'ul 1/25/2010 18:22'!
associationsDo: aBlock 
	"Evaluate aBlock for each of the receiver's elements (key/value 
	associations)."

	tally = 0 ifTrue: [ ^self].
	1 to: array size do: [ :index |
		(array at: index) ifNotNil: [ :element |
			aBlock value: element ] ]! !

!Dictionary methodsFor: 'enumerating' stamp: 'ul 1/25/2010 18:27'!
associationsSelect: aBlock 
	"Evaluate aBlock with each of my associations as the argument. Collect
	into a new dictionary, only those associations for which aBlock evaluates
	to true."

	| newCollection |
	newCollection := self copyEmpty.
	self associationsDo: [ :each | 
		(aBlock value: each) ifTrue: [ newCollection add: each ] ].
	^newCollection! !

!Dictionary methodsFor: 'accessing' stamp: 'ar 11/20/2007 14:47'!
at: key 
	"Answer the value associated with the key."

	^ self at: key ifAbsent: [self errorKeyNotFound: key]! !

!Dictionary methodsFor: 'accessing' stamp: 'sma 5/12/2000 15:01'!
at: key ifAbsentPut: aBlock 
	"Return the value at the given key. 
	If key is not included in the receiver store the result 
	of evaluating aBlock as new value."

	^ self at: key ifAbsent: [self at: key put: aBlock value]! !

!Dictionary methodsFor: 'accessing' stamp: 'ul 11/20/2009 17:09'!
at: key ifAbsent: aBlock 
	"Answer the value associated with the key or, if key isn't found,
	answer the result of evaluating aBlock."

	^((array at: (self scanFor: key)) ifNil: [ aBlock ]) value "Blocks and Associations expect #value"! !

!Dictionary methodsFor: 'accessing' stamp: 'di 3/7/2001 15:29'!
at: key ifPresentAndInMemory: aBlock
	"Lookup the given key in the receiver. If it is present, answer the value of evaluating the given block with the value associated with the key. Otherwise, answer nil."

	| v |
	v := self at: key ifAbsent: [^ nil].
	v isInMemory ifFalse: [^ nil].
	^ aBlock value: v
! !

!Dictionary methodsFor: 'accessing' stamp: 'jm 5/15/1998 07:20'!
at: key ifPresent: aBlock
	"Lookup the given key in the receiver. If it is present, answer the value of evaluating the given block with the value associated with the key. Otherwise, answer nil."

	| v |
	v := self at: key ifAbsent: [^ nil].
	^ aBlock value: v
! !

!Dictionary methodsFor: 'accessing' stamp: 'ul 9/27/2015 23:15'!
at: key ifPresent: oneArgBlock ifAbsentPut: absentBlock
	"Lookup the given key in the receiver. If it is present, answer the value of
	 evaluating oneArgBlock with the value associated with the key. Otherwise
	 add the value of absentBlock under the key, and answer that value."

	| index value |
	index := self scanFor: key.
	(array at: index) ifNotNil:
		[:element|
		 ^oneArgBlock value: element value].
	value := absentBlock value.
	self atNewIndex: index put: (self associationClass key: key value: value).
	^value! !

!Dictionary methodsFor: 'accessing' stamp: 'ar 3/4/2010 20:33'!
at: key ifPresent: oneArgBlock ifAbsent: absentBlock
	"Lookup the given key in the receiver. If it is present, answer the value of evaluating the oneArgBlock with the value associated with the key, otherwise answer the value of absentBlock."
	self at: key ifPresent:[:v| ^oneArgBlock value: v].
	^absentBlock value! !

!Dictionary methodsFor: 'accessing' stamp: 'ul 9/27/2015 23:13'!
at: key put: anObject 
	"Set the value at key to be anObject.  If key is not found, create a
	new entry for key and set is value to anObject. Answer anObject."

	| index |
	index := self scanFor: key.
	(array at: index)
		ifNil: [ self atNewIndex: index put: (self associationClass key: key value: anObject) ]
		ifNotNil: [ :association | association value: anObject ].
	^anObject! !

!Dictionary methodsFor: 'enumerating' stamp: 'ul 9/30/2009 14:14'!
collect: aBlock 
	"Evaluate aBlock with each of my values as the argument.  Collect the resulting values into a collection that is like me. Answer with the new collection."
	
	| newCollection |
	newCollection := self species new: self size.
	self associationsDo: [ :each |
		newCollection at: each key put: (aBlock value: each value) ].
	^newCollection! !

!Dictionary methodsFor: 'adding' stamp: ''!
declare: key from: aDictionary 
	"Add key to the receiver. If key already exists, do nothing. If aDictionary 
	includes key, then remove it from aDictionary and use its association as 
	the element of the receiver."

	(self includesKey: key) ifTrue: [^ self].
	(aDictionary includesKey: key)
		ifTrue: 
			[self add: (aDictionary associationAt: key).
			aDictionary removeKey: key]
		ifFalse: 
			[self add: key -> nil]! !

!Dictionary methodsFor: 'enumerating' stamp: 'ul 11/20/2009 17:21'!
do: aBlock

	self valuesDo: aBlock! !

!Dictionary methodsFor: 'private' stamp: ''!
errorValueNotFound

	self error: 'value not found'! !

!Dictionary methodsFor: 'private' stamp: 'ul 11/21/2009 01:15'!
fillFrom: aCollection with: aBlock
	"Evaluate aBlock with each of aCollections's elements as the argument.  
	Collect the resulting values into self. Answer self."

	aCollection keysAndValuesDo: [ :key :value |
		self at: key put: (aBlock value: value) ]! !

!Dictionary methodsFor: 'private' stamp: 'ul 2/4/2010 04:06'!
fixCollisionsFrom: start
	"The element at start has been removed and replaced by nil.
	This method moves forward from there, relocating any entries
	that had been placed below due to collisions with this one."

	| element index |
	index := start.
	[ (element := array at: (index := index \\ array size + 1)) == nil ] whileFalse: [
		| newIndex |
		(newIndex := self scanFor: element key) = index ifFalse: [
			array 
				at: newIndex put: element;
				at: index put: nil ] ]! !

!Dictionary methodsFor: 'filter streaming' stamp: 'MPW 1/4/1901 08:33'!
flattenOnStream:aStream
	^aStream writeDictionary:self.
! !

!Dictionary methodsFor: 'testing' stamp: 'ab 9/17/2004 00:39'!
includesAssociation: anAssociation
  ^ (self   
      associationAt: anAssociation key
      ifAbsent: [ ^ false ]) value = anAssociation value
! !

!Dictionary methodsFor: 'testing' stamp: 'sw 2/14/2000 14:34'!
includesIdentity: anObject
	"Answer whether anObject is one of the values of the receiver.  Contrast #includes: in which there is only an equality check, here there is an identity check"

	self do: [:each | anObject == each ifTrue: [^ true]].
	^ false! !

!Dictionary methodsFor: 'testing' stamp: 'mt 1/15/2015 14:35'!
includesKey: key 
	"Answer whether the receiver has a key equal to the argument, key."
	
	 ^(array at: (self scanFor: key)) notNil! !

!Dictionary methodsFor: 'testing' stamp: 'md 8/11/2005 16:49'!
isDictionary
	^true! !

!Dictionary methodsFor: 'accessing' stamp: 'ar 2/13/1999 21:16'!
keyAtIdentityValue: value 
	"Answer the key that is the external name for the argument, value. If 
	there is none, answer nil.
	Note: There can be multiple keys with the same value. Only one is returned."

	^self keyAtIdentityValue: value ifAbsent: [self errorValueNotFound]! !

!Dictionary methodsFor: 'accessing' stamp: 'ar 2/13/1999 21:16'!
keyAtIdentityValue: value ifAbsent: exceptionBlock
	"Answer the key that is the external name for the argument, value. If 
	there is none, answer the result of evaluating exceptionBlock.
	Note: There can be multiple keys with the same value. Only one is returned."
 
	self associationsDo: 
		[:association | value == association value ifTrue: [^association key]].
	^exceptionBlock value! !

!Dictionary methodsFor: 'accessing' stamp: ''!
keyAtValue: value 
	"Answer the key that is the external name for the argument, value. If 
	there is none, answer nil."

	^self keyAtValue: value ifAbsent: [self errorValueNotFound]! !

!Dictionary methodsFor: 'accessing' stamp: 'tk 2/18/97'!
keyAtValue: value ifAbsent: exceptionBlock
	"Answer the key that is the external name for the argument, value. If 
	there is none, answer the result of evaluating exceptionBlock.
	: Use =, not ==, so stings like 'this' can be found.  Note that MethodDictionary continues to use == so it will be fast."
 
	self associationsDo: 
		[:association | value = association value ifTrue: [^association key]].
	^exceptionBlock value! !

!Dictionary methodsFor: 'accessing' stamp: 'nice 10/20/2009 23:20'!
keys
	"Answer an Array containing the receiver's keys."
	
	^Array new: self size streamContents: [:s| self keysDo: [:key| s nextPut: key]]! !

!Dictionary methodsFor: 'enumerating' stamp: 'ar 7/11/1999 08:04'!
keysAndValuesDo: aBlock
	^self associationsDo:[:assoc|
		aBlock value: assoc key value: assoc value].! !

!Dictionary methodsFor: 'removing' stamp: 'di 4/4/2000 11:47'!
keysAndValuesRemove: keyValueBlock
	"Removes all entries for which keyValueBlock returns true."
	"When removing many items, you must not do it while iterating over the dictionary, since it may be changing.  This method takes care of tallying the removals in a first pass, and then performing all the deletions afterward.  Many places in the sytem could be simplified by using this method."

	| removals |
	removals := OrderedCollection new.
	self associationsDo:
		[:assoc | (keyValueBlock value: assoc key value: assoc value)
			ifTrue: [removals add: assoc key]].
 	removals do:
		[:aKey | self removeKey: aKey]! !

!Dictionary methodsFor: 'enumerating' stamp: ''!
keysDo: aBlock 
	"Evaluate aBlock for each of the receiver's keys."

	self associationsDo: [:association | aBlock value: association key]! !

!Dictionary methodsFor: 'accessing' stamp: 'cmm 5/13/2013 10:27'!
keysSortedSafely
	"Answer a sorted Array containing the receiver's keys."
	^ self keys sort:
		[ : x : y | x compareSafely: y ]! !

!Dictionary methodsFor: 'private' stamp: 'ul 1/25/2010 18:58'!
noCheckNoGrowFillFrom: anArray
	"Add the elements of anArray except nils to me assuming that I don't contain any of them, they are unique and I have more free space than they require."

	1 to: anArray size do: [ :index |
		(anArray at: index) ifNotNil: [ :association |
			array
				at: (self scanForEmptySlotFor: association key)
				put: association ] ]! !

!Dictionary methodsFor: 'copying' stamp: 'ul 11/19/2009 01:46'!
postCopy
	"Must copy the associations, or later store will affect both the
original and the copy"

	super postCopy.
	array := array collect: [ :association |
		association ifNotNil: [ association copy ] ]! !

!Dictionary methodsFor: 'printing' stamp: 'apb 7/14/2004 12:48'!
printElementsOn: aStream 
	aStream nextPut: $(.
	self size > 100
		ifTrue: [aStream nextPutAll: 'size '.
			self size printOn: aStream]
		ifFalse: [self keysSortedSafely
				do: [:key | aStream print: key;
						 nextPutAll: '->';				
						 print: (self at: key);
						 space]].
	aStream nextPut: $)! !

!Dictionary methodsFor: 'removing' stamp: 'ar 11/20/2007 14:47'!
removeKey: key 
	"Remove key from the receiver.
	If key is not in the receiver, notify an error."

	^ self removeKey: key ifAbsent: [self errorKeyNotFound: key]! !

!Dictionary methodsFor: 'removing' stamp: 'ul 11/20/2009 17:25'!
removeKey: key ifAbsent: aBlock 
	"Remove key (and its associated value) from the receiver. If key is not in 
	the receiver, answer the result of evaluating aBlock. Otherwise, answer 
	the value externally named by key."

	| index association |
	index := self scanFor: key.
	association := (array at: index) ifNil: [ ^aBlock value ].
	array at: index put: nil.
	tally := tally - 1.
	self fixCollisionsFrom: index.
	^association value! !

!Dictionary methodsFor: 'removing' stamp: ''!
removeUnreferencedKeys   "Undeclared removeUnreferencedKeys"

	^ self unreferencedKeys do: [:key | self removeKey: key].! !

!Dictionary methodsFor: 'removing' stamp: ''!
remove: anObject ifAbsent: exceptionBlock

	self shouldNotImplement! !

!Dictionary methodsFor: 'private' stamp: 'cmm 5/7/2013 20:58'!
scanFor: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or raise an error if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."

	| index start |
	index := start := anObject hash \\ array size + 1.
	[ 
		| element |
		((element := array at: index) == nil or: [ anObject = element key ])
			ifTrue: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!Dictionary methodsFor: 'enumerating' stamp: 'nice 10/5/2009 10:16'!
select: aBlock 
	"Evaluate aBlock with each of my values as the argument. Collect into a new dictionary, only those associations for which aBlock evaluates to true."

	| newCollection |
	newCollection := self copyEmpty.
	self associationsDo: [ :each |
		(aBlock value: each value) ifTrue: [
			newCollection add: each copy ] ].
	^newCollection! !

!Dictionary methodsFor: 'printing' stamp: ''!
storeOn: aStream
	| noneYet |
	aStream nextPutAll: '(('.
	aStream nextPutAll: self class name.
	aStream nextPutAll: ' new)'.
	noneYet := true.
	self associationsDo: 
			[:each | 
			noneYet
				ifTrue: [noneYet := false]
				ifFalse: [aStream nextPut: $;].
			aStream nextPutAll: ' add: '.
			aStream store: each].
	noneYet ifFalse: [aStream nextPutAll: '; yourself'].
	aStream nextPut: $)! !

!Dictionary methodsFor: 'removing' stamp: 'ul 7/17/2011 15:11'!
unreferencedKeys
	"| uk | (Time millisecondsToRun: [uk := TextConstants unreferencedKeys]) -> uk"

	^'Scanning for references . . .' 
		displayProgressFrom: 0
		to: Smalltalk classNames size * 2
		during: [ :bar |
			| currentClass n associations referencedAssociations |
			currentClass := nil.
			n := 0.
			associations := self associations asIdentitySet.
			referencedAssociations := IdentitySet new: associations size.
			self systemNavigation allSelectorsAndMethodsDo: [ :behavior :selector :method |
				behavior == currentClass ifFalse: [
					currentClass := behavior.
					 bar value: (n := n + 1) ].
				method literalsDo: [ :literal |
					(literal isVariableBinding and: [ associations includes: literal ]) ifTrue: [
						referencedAssociations add: literal ] ] ].
			(associations reject: [ :assoc | referencedAssociations includes: assoc ]) collect: [ :assoc| assoc key ] ]! !

!Dictionary methodsFor: 'private' stamp: 'ul 9/27/2015 23:15'!
valueAtNewKey: aKey put: anObject atIndex: index declareFrom: aDictionary 
	"Support for coordinating class variable and global declarations
	with variables that have been put in Undeclared so as to
	redirect all references to the undeclared variable."

	(aDictionary includesKey: aKey)
		ifTrue: 
			[self atNewIndex: index 
				put: ((aDictionary associationAt: aKey) value: anObject).
			aDictionary removeKey: aKey]
		ifFalse: 
			[self atNewIndex: index put: (self associationClass key: aKey value: anObject)]! !

!Dictionary methodsFor: 'accessing' stamp: 'ul 9/22/2009 04:02'!
values
	"Answer a Collection containing the receiver's values."

	^Array new: self size streamContents: [ :stream |
		self valuesDo: [ :value | stream nextPut: value] ]! !

!Dictionary methodsFor: 'enumerating' stamp: 'dtl 2/17/2003 09:48'!
valuesDo: aBlock 
	"Evaluate aBlock for each of the receiver's values."

	self associationsDo: [:association | aBlock value: association value]! !

!Dictionary methodsFor: 'comparing' stamp: 'ul 10/13/2015 23:23'!
= anObject
	"Two dictionaries are equal if
	 (a) they are the same 'kind' of thing.
	 (b) they have the same set of keys.
	 (c) for each (common) key, they have the same value"

	self == anObject ifTrue: [ ^true ].
	self species == anObject species ifFalse: [ ^false ].
	self size = anObject size ifFalse: [ ^false ].
	self associationsDo: [ :association |
		(anObject at: association key ifAbsent: [ ^false ]) = association value
			ifFalse: [ ^false ] ].
	^true! !

Dictionary subclass: #IdentityDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Unordered'!

!IdentityDictionary commentStamp: 'ls 06/15/02 22:35' prior: 0!
Like a Dictionary, except that keys are compared with #== instead of #= .

See the comment of IdentitySet for more information.!

!IdentityDictionary methodsFor: 'accessing' stamp: 'ul 1/25/2010 19:02'!
keyAtValue: value ifAbsent: exceptionBlock
	"Answer the key that is the external name for the argument, value. If 
	there is none, answer the result of evaluating exceptionBlock."
 
	^super keyAtIdentityValue: value ifAbsent: exceptionBlock! !

!IdentityDictionary methodsFor: 'private' stamp: 'ul 12/1/2009 04:27'!
scanForEmptySlotFor: anObject
	"Scan the key array for the first slot containing an empty slot (indicated by a nil). Answer the index of that slot. This method will be overridden in various subclasses that have different interpretations for matching elements."
	
	| index start |
	index := start := anObject scaledIdentityHash \\ array size + 1.
	[ 
		(array at: index) ifNil: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!IdentityDictionary methodsFor: 'private' stamp: 'ul 12/1/2009 04:28'!
scanFor: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or raise an error if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."

	| index start |	
	index := start := anObject scaledIdentityHash \\ array size + 1.
	[ 
		| element |
		((element := array at: index) == nil or: [ element key == anObject ])
			ifTrue: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

Dictionary subclass: #OrderedDictionary
	instanceVariableNames: 'order'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Sequenceable'!

!OrderedDictionary commentStamp: 'mt 1/16/2015 10:42' prior: 0!
I am an ordered dictionary. I have an additional index (called 'order') to keep track of the insertion order of my associations.

The read access is not affected by the additional index.

The index is updated in O(1) [time] when inserting new keys. For present keys, that insertion involves actions in O(n) to move the respective element to the end of the order.

The growth operation compacts the index and takes O(n) additional time.

NOTE: This is still no instance of SequenceableCollection. Having this, some protocols are missing and may require working on #associations, which is an Array and thus sequenceable.!

!OrderedDictionary methodsFor: 'enumerating' stamp: 'ul 1/19/2015 19:54'!
associationsDo: aBlock
	"Iterate over the order instead of the internal array."

	order from: 1 to: tally do: aBlock! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'ul 1/19/2015 20:06'!
atIndex: integer

	integer > tally ifTrue: [ self error: 'indices are out of bounds' ].
	^order at: integer! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'ul 1/19/2015 20:05'!
atIndex: integer ifAbsent: exceptionBlock
	"As we are sequenceable, provide index-based access."

	integer > tally ifTrue: [ ^exceptionBlock value ].
	^order at: integer ifAbsent: exceptionBlock! !

!OrderedDictionary methodsFor: 'private' stamp: 'ul 1/19/2015 19:53'!
atNewIndex: index put: anObject

	super atNewIndex: index put: anObject.
	order at: tally put: anObject
	! !

!OrderedDictionary methodsFor: 'copying' stamp: 'mt 1/20/2015 08:08'!
copyFrom: startIndex to: endIndex 
	"Answer a copy of the receiver that contains elements from position
	startIndex to endIndex."

	^ self shallowCopy postCopyFrom: startIndex to: endIndex! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'mt 1/16/2015 10:31'!
eighth
	"Answer the eighth element of the receiver.
	Raise an error if there are not enough elements."

	^ self atIndex: 8! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'mt 1/16/2015 10:33'!
fifth
	"Answer the fifth element of the receiver.
	Raise an error if there are not enough elements."

	^ self atIndex: 5! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'mt 1/16/2015 10:34'!
first
	"Answer the first element of the receiver"

	^ self atIndex: 1! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'mt 1/19/2015 10:09'!
first: n
	"Answer the first n elements of the receiver.
	Raise an error if there are not enough elements."

	^ self copyFrom: 1 to: n! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'mt 1/16/2015 10:33'!
fourth
	"Answer the fourth element of the receiver.
	Raise an error if there are not enough elements."

	^ self atIndex: 4! !

!OrderedDictionary methodsFor: 'private' stamp: 'ul 1/19/2015 19:53'!
growTo: anInteger

	| oldOrder |
	super growTo: anInteger.
	oldOrder := order.
	"Grow only to 75%. See #atNewIndex:put: in HashedCollection."
	order := self class arrayType new: anInteger + 1 * 3 // 4.
	order
		replaceFrom: 1
		to: tally
		with: oldOrder
		startingAt: 1! !

!OrderedDictionary methodsFor: 'private' stamp: 'ul 1/19/2015 19:53'!
initialize: n

	super initialize: n.
	order := self class arrayType new: n + 1 * 3 // 4! !

!OrderedDictionary methodsFor: 'sorting' stamp: 'ul 1/19/2015 19:51'!
isSorted
	"Return true if the receiver is sorted by #<=."
	
	^ order
		isSortedBetween: 1
		and: tally! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'mt 1/19/2015 10:10'!
last: n
	"Answer the last n elements of the receiver.  
	Raise an error if there are not enough elements."

	| size |
	size := self size.
	^ self copyFrom: size - n + 1 to: size! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'mt 1/16/2015 10:31'!
ninth
	"Answer the ninth element of the receiver.
	Raise an error if there are not enough elements."

	^ self atIndex: 9! !

!OrderedDictionary methodsFor: 'copying' stamp: 'ul 1/19/2015 20:08'!
postCopy
	"We must not copy associations again but retrieve them from the array, which is already a copy. See super."

	super postCopy.
	order := order copy.
	1 to: tally do: [ :index |
		order at: index put: (array at: (self scanFor: (order at: index) key)) ]! !

!OrderedDictionary methodsFor: 'copying' stamp: 'mt 1/20/2015 08:14'!
postCopyFrom: startIndex to: endIndex
	"Adapted from SequenceableCollection and OrderedCollection."

	| oldOrder |	
	oldOrder := order.
	array := self class arrayType
		new: (self class goodPrimeAtLeast: endIndex - startIndex + 1 * 4 // 3). "fill 75% to 100%"
	order := self class arrayType
		new: array size * 3 // 4. "remove 25%"

	startIndex to: endIndex do: [:index | | element |
		element := (oldOrder at: index) copy.
		order at: index - startIndex + 1 put: element.
		array at: (self scanFor: element key) put: element].

	tally := endIndex - startIndex + 1.! !

!OrderedDictionary methodsFor: 'removing' stamp: 'ul 1/19/2015 19:53'!
removeKey: key ifAbsent: aBlock

	| result |
	result := super removeKey: key ifAbsent: [ ^aBlock value ].
	(self scanOrderFor: key) ifNotNil: [ :index |
		order 
			replaceFrom: index
			to: tally
			with: order
			startingAt: index + 1 ].
.	order at: tally + 1 put: nil.
	^result! !

!OrderedDictionary methodsFor: 'private' stamp: 'ul 1/19/2015 19:53'!
scanOrderFor: anObject

	1 to: tally do: [ :index |
		(order at: index) key = anObject ifTrue: [ ^index ] ].
	^nil! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'mt 1/16/2015 10:34'!
second
	"Answer the second element of the receiver.
	Raise an error if there are not enough elements."

	^ self atIndex: 2! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'mt 1/16/2015 10:33'!
seventh
	"Answer the seventh element of the receiver.
	Raise an error if there are not enough elements."

	^ self atIndex: 7! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'mt 1/16/2015 10:33'!
sixth
	"Answer the sixth element of the receiver.
	Raise an error if there are not enough elements."

	^ self atIndex: 6! !

!OrderedDictionary methodsFor: 'sorting' stamp: 'ul 1/19/2015 20:02'!
sort

	self sort: nil! !

!OrderedDictionary methodsFor: 'sorting' stamp: 'mt 1/16/2015 10:19'!
sorted: aSortBlockOrNil

	^ self copy sort: aSortBlockOrNil! !

!OrderedDictionary methodsFor: 'sorting' stamp: 'ul 1/19/2015 19:54'!
sort: aSortBlock
	"Like in OrderedCollection, sort the associations according to the sort block."

	tally <= 1 ifTrue: [ ^self ].
	order
		mergeSortFrom: 1
		to: tally
		by: aSortBlock! !

!OrderedDictionary methodsFor: 'accessing' stamp: 'mt 1/16/2015 10:33'!
third
	"Answer the third element of the receiver.
	Raise an error if there are not enough elements."

	^ self atIndex: 3! !

Dictionary subclass: #PluggableDictionary
	instanceVariableNames: 'hashBlock equalBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Unordered'!

!PluggableDictionary commentStamp: '<historical>' prior: 0!
Class PluggableDictionary allows the redefinition of hashing and equality by clients. This is in particular useful if the clients know about specific properties of the objects stored in the dictionary. See the class comment of PluggableSet for an example.

Instance variables:
	hashBlock	<BlockContext>	A one argument block used for hashing the elements.
	equalBlock	<BlockContext>	A two argument block used for comparing the elements.
!

!PluggableDictionary class methodsFor: 'instance creation' stamp: 'dvf 6/10/2000 18:13'!
integerDictionary
	^ self new hashBlock: [:integer | integer hash \\ 1064164 * 1009]! !

!PluggableDictionary methodsFor: 'enumerating' stamp: 'ul 1/25/2010 18:40'!
collect: aBlock 
	"Evaluate aBlock with each of my values as the argument.  Collect the resulting values into a collection that is like me. Answer with the new collection."
	
	| newCollection |
	newCollection := (self species new: self size)
		hashBlock: hashBlock;
		equalBlock: equalBlock;
		yourself.
	self associationsDo: [ :each |
		newCollection at: each key put: (aBlock value: each value) ].
	^newCollection

! !

!PluggableDictionary methodsFor: 'copying' stamp: 'ul 11/20/2009 17:36'!
copyEmpty

	^super copyEmpty
		hashBlock: hashBlock;
		equalBlock: equalBlock;
		yourself! !

!PluggableDictionary methodsFor: 'accessing' stamp: 'ar 11/12/1998 18:46'!
equalBlock
	"Return the block used for comparing the elements in the receiver."
	^equalBlock! !

!PluggableDictionary methodsFor: 'accessing' stamp: 'ar 11/27/1998 23:55'!
equalBlock: aBlock
	"Set a new equality block. The block must accept two arguments and return true if the argumets are considered to be equal, false otherwise"
	equalBlock := aBlock.! !

!PluggableDictionary methodsFor: 'accessing' stamp: 'ar 11/12/1998 18:46'!
hashBlock
	"Return the block used for hashing the elements in the receiver."
	^hashBlock! !

!PluggableDictionary methodsFor: 'accessing' stamp: 'ar 11/12/1998 18:46'!
hashBlock: aBlock
	"Set a new hash block. The block must accept one argument and must return the hash value of the given argument."
	hashBlock := aBlock.! !

!PluggableDictionary methodsFor: 'private' stamp: 'ul 9/30/2009 14:19'!
scanForEmptySlotFor: anObject
	"Scan the key array for the first slot containing an empty slot (indicated by a nil). Answer the index of that slot. This method will be overridden in various subclasses that have different interpretations for matching elements."
	
	| index start |
	index := start := (hashBlock
		ifNil: [ anObject hash ]
		ifNotNil: [ hashBlock value: anObject ]) \\ array size + 1.
	[ 
		(array at: index) ifNil: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!PluggableDictionary methodsFor: 'private' stamp: 'ul 9/30/2009 14:23'!
scanFor: anObject 
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or raise an error if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."
	
	| index start |
	index := start := (hashBlock
		ifNil: [ anObject hash ]
		ifNotNil: [ hashBlock value: anObject ]) \\ array size + 1.
	[ 
		| element |
		((element := array at: index) == nil or: [
			equalBlock
				ifNil: [ element key = anObject ]
				ifNotNil: [ equalBlock value: element key value: anObject ] ])
			ifTrue: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!PluggableDictionary methodsFor: 'comparing' stamp: 'ul 10/13/2015 23:24'!
= anObject
	"Two dictionaries are equal if
	 (a) they are the same 'kind' of thing.
	 (b) they have the same set of keys.
	 (c) for each (common) key, they have the same value"

	self == anObject ifTrue: [ ^true ].
	self species == anObject species ifFalse: [ ^false ].
	hashBlock = anObject hashBlock ifFalse: [ ^false ].
	equalBlock = anObject equalBlock ifFalse: [ ^false ].
	self size = anObject size ifFalse: [ ^false ].
	self associationsDo: [ :association |
		(anObject at: association key ifAbsent: [ ^false ]) = association value
			ifFalse: [ ^false ] ].
	^true! !

Dictionary subclass: #WeakKeyDictionary
	instanceVariableNames: 'finalizer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Weak'!

!WeakKeyDictionary commentStamp: '<historical>' prior: 0!
I am a dictionary holding only weakly on my keys. This is a bit dangerous since at any time my keys can go away. Clients are responsible to register my instances by WeakArray such that the appropriate actions can be taken upon loss of any keys.

See WeakRegistry for an example of use.
!

WeakKeyDictionary subclass: #WeakIdentityKeyDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Weak'!

!WeakIdentityKeyDictionary commentStamp: '<historical>' prior: 0!
This class represents an identity dictionary with weak keys.!

!WeakIdentityKeyDictionary methodsFor: 'private' stamp: 'ul 12/1/2009 04:35'!
scanForEmptySlotFor: anObject
	"Scan the key array for the first slot containing an empty slot (indicated by a nil). Answer the index of that slot. This method will be overridden in various subclasses that have different interpretations for matching elements."
	
	| index start |
	index := start := anObject scaledIdentityHash \\ array size + 1.
	[ 
		(array at: index) ifNil: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!WeakIdentityKeyDictionary methodsFor: 'private' stamp: 'ul 12/1/2009 04:36'!
scanFor: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or raise an error if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."

	| index start |
	index := start := anObject scaledIdentityHash \\ array size + 1.
	[ 
		| element |
		((element := array at: index) == nil or: [ element key == anObject ])
			ifTrue: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

WeakIdentityKeyDictionary subclass: #WeakKeyToCollectionDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Weak'!

!WeakKeyToCollectionDictionary commentStamp: '<historical>' prior: 0!
This class represents an identity dictionary with weak keys, whose values are collections. 
Keys not in the dictionary are mapped to the empty collection.  Conversely, if a collection becomes empty, the mapping can be removed to save time and space.  However, because this requires re-hashing, it does not pay to do this to eagerly.!

!WeakKeyToCollectionDictionary methodsFor: 'private' stamp: 'ul 1/25/2010 20:01'!
noCheckNoGrowFillFrom: anArray
	"Add the elements of anArray except nils and associations with empty collections (or with only nils) to me assuming that I don't contain any of them, they are unique and I have more free space than they require."

	tally := 0.
	1 to: anArray size do: [ :index |
		(anArray at: index) ifNotNil: [ :association |
			association key ifNotNil: [ :key | "Don't let the key go away"
				| cleanedValue |
				(cleanedValue := association value copyWithout: nil) isEmpty 
					ifFalse: [
						association value: cleanedValue.
						array
							at: (self scanForEmptySlotFor: key)
							put: association.
						tally := tally + 1 ] ] ] ]! !

!WeakKeyDictionary methodsFor: 'adding' stamp: 'ar 3/21/98 16:00'!
add: anAssociation
	self at: anAssociation key put: anAssociation value.
	^ anAssociation! !

!WeakKeyDictionary methodsFor: 'accessing' stamp: 'ul 9/27/2015 23:13'!
associationClass

	^WeakKeyAssociation! !

!WeakKeyDictionary methodsFor: 'enumerating' stamp: 'ul 12/3/2014 03:50'!
associationsDo: aBlock 
	"Evaluate aBlock for each of the receiver's elements (key/value 
	associations)."

	tally = 0 ifTrue: [ ^self].
	1 to: array size do: [ :index |
		(array at: index) ifNotNil: [ :association |
			association key ifNotNil: [ :key | "Don't let the key go away."
				aBlock value: association ] ] ]! !

!WeakKeyDictionary methodsFor: 'accessing' stamp: 'ul 9/27/2015 23:14'!
at: key put: anObject 
	"Set the value at key to be anObject.  If key is not found, create a new
	entry for key and set is value to anObject. Answer anObject."
	
	key ifNil: [ ^anObject ].
	^super at: key put: anObject! !

!WeakKeyDictionary methodsFor: 'private' stamp: 'ul 4/12/2010 22:59'!
compact
	"Reduce the size of array so that the load factor will be ~75%."
	
	| newCapacity |
	newCapacity := self class goodPrimeAtLeast: self slowSize * 4 // 3.
	self growTo: newCapacity! !

!WeakKeyDictionary methodsFor: 'accessing' stamp: 'ul 2/3/2010 06:22'!
finalizer: aValueable

	finalizer := aValueable! !

!WeakKeyDictionary methodsFor: 'finalization' stamp: 'ul 6/12/2010 18:38'!
finalizeValues
	"Remove and finalize all elements which have nil key"
	
	|  cleanUpAfterRemove |
	tally = 0 ifTrue: [ ^self ].
	cleanUpAfterRemove := false.
	1 to: array size do: [ :index |
		(array at: index) 
			ifNil: [ cleanUpAfterRemove := false ]
			ifNotNil: [ :element |
				element key
					ifNil: [
						finalizer ifNotNil: [ finalizer value: element value ].
						array at: index put: nil.
						tally := tally - 1.
						cleanUpAfterRemove := true ]
					ifNotNil: [ :key |
						cleanUpAfterRemove ifTrue: [
							| newIndex |
							(newIndex := self scanFor: key) = index ifFalse: [
								array 
									at: newIndex put: element;
									at: index put: nil ] ] ] ] ].
	cleanUpAfterRemove ifTrue: [ "Continue rehashing at the front of array"
		self fixCollisionsFrom: array size ]
					! !

!WeakKeyDictionary methodsFor: 'private' stamp: 'ul 2/17/2010 13:37'!
fixCollisionsFrom: start
	"The element at start has been removed and replaced by nil.
	This method moves forward from there, relocating any entries
	that had been placed below due to collisions with this one."

	| element index |
	index := start.
	[ (element := array at: (index := index \\ array size + 1)) == nil ] whileFalse: [
		element key
			ifNil: [ 
				finalizer ifNotNil: [ finalizer value: element value ].
				array at: index put: nil.
				tally := tally - 1 ]
			ifNotNil: [ :key | "Don't let the key go away"
				| newIndex |
				(newIndex := self scanFor: key) = index ifFalse: [
					array 
						at: newIndex put: element;
						at: index put: nil ] ] ]! !

!WeakKeyDictionary methodsFor: 'private' stamp: 'ul 2/5/2010 04:35'!
growSize
	"Answer what my next table size should be.
	Note that, it can be less than the current."
	
	^self class goodPrimeAtLeast: self slowSize * 2 + 2! !

!WeakKeyDictionary methodsFor: 'private' stamp: 'ul 2/4/2010 01:02'!
noCheckNoGrowFillFrom: anArray
	"Add the elements of anArray except nils and flag to me assuming that I don't contain any of them, they are unique and I have more free space than they require."

	tally := 0.
	1 to: anArray size do: [ :index |
		(anArray at: index) ifNotNil: [ :association |
			association key 
				ifNil: [ finalizer ifNotNil: [ finalizer value: association value ] ]
				ifNotNil: [ :key | "Don't let the key go away"
					array
						at: (self scanForEmptySlotFor: key)
						put: association.
					tally := tally + 1 ] ] ]! !

!WeakKeyDictionary methodsFor: 'public' stamp: 'ul 2/4/2010 00:59'!
slowSize
	"Careful!! Answer the maximum amount
	of elements in the receiver, not the
	exact amount"

	| count |
	count := 0.
	1 to: array size do: [ :index |
		(array at: index) ifNotNil: [ :object |
			object key ifNotNil: [
				count := count + 1 ] ] ].
	^count! !

Dictionary subclass: #WeakValueDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Weak'!

!WeakValueDictionary commentStamp: '<historical>' prior: 0!
I am a dictionary holding only weakly on my values. Clients may expect to get a nil value for any object they request.!

!WeakValueDictionary methodsFor: 'adding' stamp: 'ar 3/21/98 16:02'!
add: anAssociation
	self at: anAssociation key put: anAssociation value.
	^ anAssociation! !

!WeakValueDictionary methodsFor: 'accessing' stamp: 'ul 9/27/2015 23:13'!
associationClass

	^WeakValueAssociation! !

!HashedCollection class methodsFor: 'private' stamp: 'cmm 11/29/2011 21:23'!
arrayType
	^ Array! !

!HashedCollection class methodsFor: 'initialize-release' stamp: 'ul 1/31/2011 00:05'!
cleanUp: aggressive
	"Rehash all instances when cleaning aggressively"

	aggressive ifTrue: [ self compactAll ]
! !

!HashedCollection class methodsFor: 'initialize-release' stamp: 'ul 4/12/2010 23:49'!
compactAll
	"HashedCollection compactAll"	
		
	self allSubclassesDo: #compactAllInstances! !

!HashedCollection class methodsFor: 'initialize-release' stamp: 'ul 10/17/2010 03:32'!
compactAllInstances
	"Do not use #allInstancesDo: because #compact may create new instances."

	self allInstances do: [ :each | each compact ]! !

!HashedCollection class methodsFor: 'sizing' stamp: 'ul 1/25/2010 22:45'!
goodPrimeAtLeast: lowerLimit
	"Answer the next good prime >= lowerlimit.
	If lowerLimit is larger than the largest known good prime,
	just make it odd."
	
	| primes low mid high prime |
	primes := self goodPrimes.
	low := 1.
	high := primes size.
	lowerLimit > (primes at: high) ifTrue: [
		^lowerLimit bitOr: 1 ].
	[ high - low <= 1 ] whileFalse: [
		mid := high + low // 2.
		prime := primes at: mid.
		prime = lowerLimit ifTrue: [ ^prime ].
		prime < lowerLimit
			ifTrue: [ low := mid ]
			ifFalse: [ high := mid ] ].
	(primes at: low) >= lowerLimit ifTrue: [ ^primes at: low ].
	^primes at: high! !

!HashedCollection class methodsFor: 'sizing' stamp: 'ul 4/7/2010 00:17'!
goodPrimes
	"Answer a sorted array of prime numbers less than one billion that make good
	hash table sizes. Should be expanded as needed.  See comments below code"
	
	^#(
		5 11 17 23 31 43 59 79 107 149 199 269 359 479 641 857 1151 1549 2069
		2237 2423 2617 2797 2999 3167 3359 3539 3727 3911
		4441 4787 5119 5471 5801 6143 6521 6827 7177 7517 7853
		8783 9601 10243 10867 11549 12239 12919 13679 14293 15013 15731
		17569 19051 20443 21767 23159 24611 25847 27397 28571 30047 31397
		35771 38201 40841 43973 46633 48989 51631 54371 57349 60139 62969
		70589 76091 80347 85843 90697 95791 101051 106261 111143 115777 120691 126311
		140863 150523 160969 170557 181243 190717 201653 211891 221251 232591 242873 251443
		282089 300869 321949 341227 362353 383681 401411 422927 443231 464951 482033 504011
		562621 605779 647659 681607 723623 763307 808261 844709 886163 926623 967229 1014617
		1121987 1201469 1268789 1345651 1429531 1492177 1577839 1651547 1722601 1800377 1878623 1942141 2028401
		2242727 2399581 2559173 2686813 2836357 3005579 3144971 3283993 3460133 3582923 3757093 3903769 4061261
		4455361 4783837 5068529 5418079 5680243 6000023 6292981 6611497 6884641 7211599 7514189 7798313 8077189
		9031853 9612721 10226107 10745291 11338417 11939203 12567671 13212697 13816333 14337529 14938571 15595673 16147291
		17851577 18993941 20180239 21228533 22375079 23450491 24635579 25683871 26850101 27921689 29090911 30153841 31292507 32467307
		35817611 37983761 40234253 42457253 44750177 46957969 49175831 51442639 53726417 55954637 58126987 60365939 62666977 64826669
		71582779 76039231 80534381 84995153 89500331 93956777 98470819 102879613 107400389 111856841 116365721 120819287 125246581 129732203
		143163379 152076289 161031319 169981667 179000669 187913573 196826447 205826729 214748357 223713691 232679021 241591901 250504801 259470131
		285162679 301939921 318717121 335494331 352271573 369148753 385926017 402603193 419480419 436157621 453034849 469712051 486589307 503366497 520043707 
		570475349 603929813 637584271 671138659 704693081 738247541 771801929 805356457 838910803 872365267 905919671 939574117 973128521 1006682977 1040137411 
		1073741833)

"The above primes past 2069 were chosen carefully so that they do not interact badly with 1664525 (used by hashMultiply), and so that gcd(p, (256^k) +/- a) = 1, for 0<a<=32 and 0<k<=8.  See Knuth's TAOCP for details."

"The above primes also try to map the values of ((0 to: 4095) collect: [ :each | each << 18 \\ prime ]) sort to an equidistant sequence of numbers. This helps to avoid the collision of chains in identity-based hashed collections. To do that  they were chosen to return a low value when the following block is evaluated with them as argument:
 [ :prime |
	| slots cost optimalDistance previous |
	slots := Array new: 4097.
	0 to: 4095 do: [ :ea | slots at: ea + 1 put: ea *  262144 \\ prime ].
	slots at: 4097 put: prime.
	slots sort.
	cost := 0.
	optimalDistance := prime // 4096.
	2 to: 4097 do: [ :index |
		| newCost |
		newCost := optimalDistance - ((slots at: index) - (slots at: index - 1)).
		newCost > cost ifTrue: [ cost := newCost ] ].
	cost ]."! !

!HashedCollection class methodsFor: 'instance creation' stamp: 'ul 11/19/2009 01:07'!
new
	^ self basicNew initialize: 5! !

!HashedCollection class methodsFor: 'instance creation' stamp: 'ul 11/19/2009 01:07'!
new: nElements
	"Create a Set large enough to hold nElements without growing"
	^ self basicNew initialize: (self sizeFor: nElements)! !

!HashedCollection class methodsFor: 'initialize-release' stamp: 'ul 1/25/2010 19:55'!
rehashAll
	"HashedCollection rehashAll"	
		
	self allSubclassesDo: #rehashAllInstances! !

!HashedCollection class methodsFor: 'initialize-release' stamp: 'ul 10/17/2010 03:33'!
rehashAllInstances
	"Do not use #allInstancesDo: because #rehash may create new instances."

	self allInstances do: [ :each | each rehash ]! !

!HashedCollection class methodsFor: 'sizing' stamp: 'ul 11/25/2009 04:24'!
sizeFor: nElements
	"Large enough prime (or odd if too large) size to hold nElements with some slop (see fullCheck)"
	
	nElements < 4 ifTrue: [ ^5 ].
	^self goodPrimeAtLeast: nElements + 1 * 4 // 3! !

!HashedCollection methodsFor: 'adding' stamp: 'ul 4/12/2010 22:38'!
add: newObject withOccurrences: anInteger
	"Add newObject anInteger times to the receiver. Do nothing if anInteger is less than one. Answer newObject."
	
	anInteger < 1 ifTrue: [ ^newObject ].
	^self add: newObject "I can only store an object once."
	! !

!HashedCollection methodsFor: 'private' stamp: 'ul 11/19/2009 00:58'!
array
	^ array! !

!HashedCollection methodsFor: 'private' stamp: 'ul 12/14/2009 18:09'!
atNewIndex: index put: anObject

	array at: index put: anObject.
	tally := tally + 1.
	"Keep array at least 1/4 free for decent hash behavior"
	array size * 3 < (tally * 4) ifTrue: [ self grow ]! !

!HashedCollection methodsFor: 'accessing' stamp: 'ul 11/19/2009 00:47'!
capacity
	"Answer the current capacity of the receiver."

	^ array size! !

!HashedCollection methodsFor: 'objects from disk' stamp: 'ul 10/17/2010 03:51'!
comeFullyUpOnReload: smartRefStream
	"Symbols have new hashes in this image."

	self compact! !

!HashedCollection methodsFor: 'private' stamp: 'ul 4/12/2010 22:53'!
compact
	"Reduce the size of array so that the load factor will be ~75%."
	
	| newCapacity |
	newCapacity := self class goodPrimeAtLeast: tally * 4 // 3.
	self growTo: newCapacity! !

!HashedCollection methodsFor: 'copying' stamp: 'ul 11/19/2009 00:52'!
copyEmpty
	"Answer an empty copy of this collection"
	
	"Note: this code could be moved to super"
	
	^self species new! !

!HashedCollection methodsFor: 'enumerating' stamp: 'ul 11/19/2009 00:53'!
doWithIndex: aBlock2
	"Support Set enumeration with a counter, even though not ordered"
	| index |
	index := 0.
	self do: [:item | aBlock2 value: item value: (index := index+1)]! !

!HashedCollection methodsFor: 'private' stamp: 'ul 11/19/2009 00:58'!
errorNoFreeSpace

	self error: 'There is no free space in this collection!!'! !

!HashedCollection methodsFor: 'compatibility' stamp: 'ar 1/8/2010 09:02'!
findElementOrNil: anObject
	"This method has been superseeded by #scanFor:
	It is here for compatibility with external packages only."
	^self scanFor: anObject! !

!HashedCollection methodsFor: 'private' stamp: 'ul 11/19/2009 00:59'!
fixCollisionsFrom: start
	"The element at start has been removed and replaced by nil.
	This method moves forward from there, relocating any entries
	that had been placed below due to collisions with this one."
	
	self subclassResponsibility! !

!HashedCollection methodsFor: 'compatibility' stamp: 'ul 1/25/2010 18:07'!
fullCheck
	"This is a private method, formerly implemented in Set, that is no longer required.
	It is here for compatibility with external packages only."
	"Keep array at least 1/4 free for decent hash behavior"
	
	array size * 3 < (tally * 4) ifTrue: [ self grow ]! !

!HashedCollection methodsFor: 'private' stamp: 'ul 11/25/2009 04:01'!
grow
	"Grow the elements array and reinsert the old elements"
	
	self growTo: self growSize! !

!HashedCollection methodsFor: 'private' stamp: 'ul 11/25/2009 04:00'!
growSize
	"Answer what my next higher table size should be"
	
	^self class goodPrimeAtLeast: array size * 3 // 2 + 2! !

!HashedCollection methodsFor: 'private' stamp: 'cmm 11/29/2011 21:24'!
growTo: anInteger
	"Grow the elements array and reinsert the old elements"
	
	| oldElements |
	oldElements := array.
	array := self class arrayType new: anInteger.
	self noCheckNoGrowFillFrom: oldElements! !

!HashedCollection methodsFor: 'private' stamp: 'cmm 11/29/2011 21:24'!
initialize: n
	"Initialize array to an array size of n"
	array := self class arrayType new: n.
	tally := 0! !

!HashedCollection methodsFor: 'private' stamp: 'ul 11/19/2009 01:01'!
noCheckNoGrowFillFrom: anArray
	"Add the elements of anArray except nils to me assuming that I don't contain any of them, they are unique and I have more free space than they require."
	
	self subclassResponsibility! !

!HashedCollection methodsFor: 'private' stamp: 'ul 11/19/2009 01:05'!
rehash
	
	self growTo: array size! !

!HashedCollection methodsFor: 'removing' stamp: 'ul 11/19/2009 00:57'!
removeAll
	"remove all elements from this collection.
	Preserve the capacity"
	
	self initialize: self capacity! !

!HashedCollection methodsFor: 'private' stamp: 'ul 11/19/2009 01:05'!
scanForEmptySlotFor: anObject
	"Scan the key array for the first slot containing an empty slot (indicated by a nil). Answer the index of that slot. This method will be overridden in various subclasses that have different interpretations for matching elements."
	
	| index start |
	index := start := anObject hash \\ array size + 1.
	[ 
		(array at: index) ifNil: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!HashedCollection methodsFor: 'private' stamp: 'ul 11/19/2009 01:06'!
scanFor: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or raise an error if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."

	self subclassResponsibility! !

!HashedCollection methodsFor: 'accessing' stamp: 'ul 11/19/2009 00:51'!
size
	^ tally! !

!HashedCollection methodsFor: 'accessing' stamp: 'ul 11/19/2009 00:50'!
someElement

	self deprecated: 'Use #anyOne'.
	^self anyOne! !

!HashedCollection methodsFor: 'enumerating' stamp: 'ul 11/19/2009 00:54'!
union: aCollection
	"Answer the set theoretic union of the receiver and aCollection, using the receiver's notion of equality and not side effecting the receiver at all."

	^ self copy addAll: aCollection; yourself

! !

HashedCollection subclass: #Set
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Unordered'!

!Set commentStamp: 'nice 8/26/2010 22:14' prior: 0!
I represent a set of objects without duplicates.  I can hold anything that responds to
#hash and #=, except for nil.  My instances will automatically grow, if necessary.
Note that I rely on #=, not #==.  If you want a set using #==, use IdentitySet.

Instance structure:

  array	An array whose non-nil elements are the elements of the set,
		and whose nil elements are empty slots.  There is always at least one nil.
		In fact I try to keep my "load" at 75% or less so that hashing will work well.

  tally	The number of elements in the set.  The array size is always greater than this.

The core operation is #scanFor:, which either finds the position where an
object is stored in array, if it is present, or finds a suitable position holding nil, if
its argument is not present in array.

I can include an UndefinedObject (nil) thanks to a special implementation using a wrapper (see message #asSetElement and class SetElement).
Indeed, a nil entry in the storage array means vacuity, it cannot mean I contain nil.!

Set subclass: #IdentitySet
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Unordered'!

!IdentitySet commentStamp: 'nice 8/26/2010 22:00' prior: 0!
The same as a Set, except that items are compared using #== instead of #=.

Almost any class named IdentityFoo is the same as Foo except for the way items are compared.  In Foo, #= is used, while in IdentityFoo, #== is used.  That is, identity collections will treat items as the same only if they have the same identity.

For example, note that copies of a string are equal:

	('abc' copy) = ('abc' copy)

but they are not identical:

	('abc' copy) == ('abc' copy)

A regular Set will only include equal objects once:

	| aSet |
	aSet := Set new.
	aSet add: 'abc' copy.
	aSet add: 'abc' copy.
	aSet


An IdentitySet will include multiple equal objects if they are not identical:

	| aSet |
	aSet := IdentitySet new.
	aSet add: 'abc' copy.
	aSet add: 'abc' copy.
	aSet
!

!IdentitySet methodsFor: 'converting' stamp: 'ar 9/22/2000 10:13'!
asIdentitySet
	^self! !

!IdentitySet methodsFor: 'private' stamp: 'ul 12/1/2009 04:38'!
scanForEmptySlotFor: anObject
	"Scan the key array for the first slot containing an empty slot (indicated by a nil). Answer the index of that slot. This method will be overridden in various subclasses that have different interpretations for matching elements."
	
	| index start |
	index := start := anObject scaledIdentityHash \\ array size + 1.
	[ 
		(array at: index) ifNil: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!IdentitySet methodsFor: 'private' stamp: 'ar 2/1/2010 19:53'!
scanFor: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or raise an error if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."

	| index start |
	index := start := anObject scaledIdentityHash \\ array size + 1.
	[ 
		| element |
		((element := array at: index) == nil or: [ element enclosedSetElement == anObject ])
			ifTrue: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

Set subclass: #KeyedSet
	instanceVariableNames: 'keyBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Unordered'!

!KeyedSet commentStamp: '<historical>' prior: 0!
Like Set except a key of every element is used for hashing and searching instead of the element itself.  keyBlock gets the key of an element.!

KeyedSet subclass: #KeyedIdentitySet
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Unordered'!

!KeyedIdentitySet commentStamp: 'nice 8/26/2010 22:01' prior: 0!
A KeyedIdentitySet is like a Bag, except that items are compared with #== instead of #= .

See the comment of IdentitySet for more information.!

!KeyedIdentitySet methodsFor: 'private' stamp: 'ul 12/1/2009 04:30'!
scanForEmptySlotFor: anObject
	"Scan the key array for the first slot containing an empty slot (indicated by a nil). Answer the index of that slot. This method will be overridden in various subclasses that have different interpretations for matching elements."
	
	| index start |
	index := start := anObject scaledIdentityHash \\ array size + 1.
	[ 
		(array at: index) ifNil: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!KeyedIdentitySet methodsFor: 'private' stamp: 'ar 2/1/2010 21:19'!
scanFor: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or raise an error if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."

	| index start |
	index := start := anObject scaledIdentityHash \\ array size + 1.
	[ 
		| element |
		((element := array at: index) == nil or: [ (keyBlock value: element enclosedSetElement) == anObject ])
			ifTrue: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!KeyedSet class methodsFor: 'instance creation' stamp: 'ajh 10/23/2000 23:16'!
keyBlock: oneArgBlock
	"Create a new KeySet whose way to access an element's key is by executing oneArgBlock on the element"

	^ self new keyBlock: oneArgBlock! !

!KeyedSet methodsFor: 'adding' stamp: 'ajh 12/4/2001 05:27'!
addAll: aCollection 
	"Include all the elements of aCollection as the receiver's elements"

	(aCollection respondsTo: #associationsDo:)
		ifTrue: [aCollection associationsDo: [:ass | self add: ass]]
		ifFalse: [aCollection do: [:each | self add: each]].
	^ aCollection! !

!KeyedSet methodsFor: 'adding' stamp: 'ar 2/1/2010 21:18'!
add: newObject
	"Include newObject as one of the receiver's elements, but only if
	not already present. Answer newObject."

	| index |
	index := self scanFor: (keyBlock value: newObject).
	(array at: index) ifNil: [self atNewIndex: index put: newObject asSetElement].
	^ newObject! !

!KeyedSet methodsFor: 'accessing' stamp: 'ar 11/20/2007 14:47'!
at: key 
	"Answer the value associated with the key."

	^ self at: key ifAbsent: [self errorKeyNotFound: key]! !

!KeyedSet methodsFor: 'accessing' stamp: 'ajh 12/10/2000 15:42'!
at: key ifAbsentPut: aBlock 
	"Answer the value associated with the key or, if key isn't found,
	add the result of evaluating aBlock to self"

	^ self at: key ifAbsent: [self add: aBlock value]! !

!KeyedSet methodsFor: 'accessing' stamp: 'ar 2/1/2010 21:18'!
at: key ifAbsent: aBlock 
	"Answer the value associated with the key or, if key isn't found,
	answer the result of evaluating aBlock."

	^(array at: (self scanFor: key)) ifNil: [ aBlock value ] ifNotNil:[:obj| obj enclosedSetElement]! !

!KeyedSet methodsFor: 'accessing' stamp: 'ajh 9/5/2000 03:58'!
at: key ifPresent: aBlock
	"Lookup the given key in the receiver. If it is present, answer the value of evaluating the given block with the value associated with the key. Otherwise, answer nil."

	| v |
	v := self at: key ifAbsent: [^ nil].
	^ aBlock value: v
! !

!KeyedSet methodsFor: 'copying' stamp: 'ul 11/20/2009 18:37'!
copyEmpty

	^super copyEmpty
		keyBlock: keyBlock;
		yourself! !

!KeyedSet methodsFor: 'private' stamp: 'ul 2/4/2010 04:07'!
fixCollisionsFrom: start
	"The element at start has been removed and replaced by nil.
	This method moves forward from there, relocating any entries
	that had been placed below due to collisions with this one."

	| element index |
	index := start.
	[ (element := array at: (index := index \\ array size + 1)) == nil ] whileFalse: [
		| newIndex |
		(newIndex := self scanFor: (keyBlock value: element enclosedSetElement)) = index ifFalse: [
			array 
				at: newIndex put: element;
				at: index put: nil ] ]! !

!KeyedSet methodsFor: 'testing' stamp: 'mt 1/15/2015 14:35'!
includesKey: key

	^ (array at: (self scanFor: key)) notNil! !

!KeyedSet methodsFor: 'testing' stamp: 'ul 2/12/2011 01:08'!
includes: anObject 
	
	(array at: (self scanFor: (keyBlock value: anObject))) ifNil: [ ^false ] ifNotNil: [ ^true ]! !

!KeyedSet methodsFor: 'private' stamp: 'ul 1/25/2010 22:11'!
initialize: n

	super initialize: n.
	keyBlock := #key
! !

!KeyedSet methodsFor: 'initialize' stamp: 'ajh 9/5/2000 03:36'!
keyBlock: oneArgBlock
	"When evaluated return the key of the argument which will be an element of the set"

	keyBlock := oneArgBlock! !

!KeyedSet methodsFor: 'accessing' stamp: 'nice 10/20/2009 23:20'!
keys
	"Answer an Array containing the receiver's keys."
	
	^Array new: self size streamContents: [:s| self keysDo: [:key| s nextPut: key]]! !

!KeyedSet methodsFor: 'accessing' stamp: 'ajh 7/3/2004 17:54'!
keysDo: block

	self do: [:item | block value: (keyBlock value: item)]! !

!KeyedSet methodsFor: 'accessing' stamp: 'ul 1/25/2010 21:04'!
keysSorted

	^self keys sort! !

!KeyedSet methodsFor: 'accessing' stamp: 'ul 2/19/2010 15:35'!
like: anObject
	"Answer an object in the receiver that is equal to anObject,
	nil if no such object is found. Relies heavily on hash properties"

	^(array at: (self scanFor: (keyBlock value: anObject)))
		ifNotNil: [ :element | element enclosedSetElement]! !

!KeyedSet methodsFor: 'accessing' stamp: 'ul 2/19/2010 15:35'!
like: anObject ifAbsent: aBlock
	"Answer an object in the receiver that is equal to anObject,
	or evaluate the block if not found. Relies heavily on hash properties"

	^(array at: (self scanFor: (keyBlock value: anObject)))
		ifNil: [ aBlock value ]
		ifNotNil: [ :element | element enclosedSetElement ]! !

!KeyedSet methodsFor: 'adding' stamp: 'ar 2/1/2010 21:17'!
member: newObject
	"Include newObject as one of the receiver's elements, if already exists just return it"

	| index |
	index := self scanFor: (keyBlock value: newObject).
	(array at: index) ifNotNil: [ :element | ^element enclosedSetElement].
	self atNewIndex: index put: newObject asSetElement.
	^ newObject! !

!KeyedSet methodsFor: 'private' stamp: 'ar 2/1/2010 21:18'!
noCheckNoGrowFillFrom: anArray
	"Add the elements of anArray except nils to me assuming that I don't contain any of them, they are unique and I have more free space than they require."

	1 to: anArray size do: [ :index |
		(anArray at: index) ifNotNil: [ :object |
			array
				at: (self scanForEmptySlotFor: (keyBlock value: object enclosedSetElement))
				put: object ] ]! !

!KeyedSet methodsFor: 'removing' stamp: 'nice 12/30/2008 19:01'!
removeAll
	"See super."
	
	| tmp |
	tmp := keyBlock.
	super removeAll.
	keyBlock := tmp! !

!KeyedSet methodsFor: 'removing' stamp: 'ar 11/20/2007 14:47'!
removeKey: key 

	^ self removeKey: key ifAbsent: [self errorKeyNotFound: key]! !

!KeyedSet methodsFor: 'removing' stamp: 'ar 2/1/2010 21:18'!
removeKey: key ifAbsent: aBlock

	| index obj |
	index := self scanFor: key.
	(obj := array at: index) == nil ifTrue: [ ^ aBlock value ].
	array at: index put: nil.
	tally := tally - 1.
	self fixCollisionsFrom: index.
	^ obj enclosedSetElement! !

!KeyedSet methodsFor: 'removing' stamp: 'ul 9/23/2009 21:13'!
remove: oldObject ifAbsent: aBlock

	| index |
	index := self scanFor: (keyBlock value: oldObject).
	(array at: index) == nil ifTrue: [ ^ aBlock value ].
	array at: index put: nil.
	tally := tally - 1.
	self fixCollisionsFrom: index.
	^ oldObject! !

!KeyedSet methodsFor: 'private' stamp: 'ar 2/1/2010 21:17'!
scanFor: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or raise an error if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."

	| index start |
	index := start := anObject hash \\ array size + 1.
	[ 
		| element |
		((element := array at: index) == nil or: [ (keyBlock value: element enclosedSetElement) = anObject ])
			ifTrue: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

Set subclass: #PluggableSet
	instanceVariableNames: 'hashBlock equalBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Unordered'!

!PluggableSet commentStamp: 'nice 3/25/2010 23:02' prior: 0!
PluggableSets allow the redefinition of hashing and equality by clients. This is in particular useful if the clients know about specific properties of the objects stored in the set which in turn can heavily improve the performance of sets and dictionaries.

Instance variables:
	hashBlock	<BlockContext>	A one argument block used for hashing the elements.
	equalBlock	<BlockContext>	A two argument block used for comparing the elements.

Example: Adding 1000 integer points in the range (0@0) to: (100@100) to a set.

	| rnd set max pt |
	set := Set new: 1000.
	rnd := Random new.
	max := 100.
	Time millisecondsToRun:[
		1 to: 1000 do:[:i|
			pt := (rnd next * max) truncated @ (rnd next * max) truncated.
			set add: pt.
		].
	].

The above is way slow since the default hashing function of points leads to an awful lot of collisions in the set. And now the same, with a somewhat different hash function:

	| rnd set max pt |
	set := PluggableSet new: 1000.
	set hashBlock:[:item| (item x bitShift: 16) + item y].
	rnd := Random new.
	max := 100.
	Time millisecondsToRun:[
		1 to: 1000 do:[:i|
			pt := (rnd next * max) truncated @ (rnd next * max) truncated.
			set add: pt.
		].
	].
!

!PluggableSet class methodsFor: 'instance creation' stamp: 'dvf 6/10/2000 18:13'!
integerSet
	^self new hashBlock: [:integer | integer hash \\ 1064164 * 1009]! !

!PluggableSet methodsFor: 'copying' stamp: 'ul 11/20/2009 17:36'!
copyEmpty

	^super copyEmpty
		hashBlock: hashBlock;
		equalBlock: equalBlock;
		yourself! !

!PluggableSet methodsFor: 'accessing' stamp: 'ar 11/12/1998 18:43'!
equalBlock
	"Return the block used for comparing the elements in the receiver."
	^equalBlock! !

!PluggableSet methodsFor: 'accessing' stamp: 'ar 11/27/1998 23:55'!
equalBlock: aBlock
	"Set a new equality block. The block must accept two arguments and return true if the argumets are considered equal, false otherwise"
	equalBlock := aBlock.! !

!PluggableSet methodsFor: 'accessing' stamp: 'ar 11/12/1998 18:43'!
hashBlock
	"Return the block used for hashing the elements in the receiver."
	^hashBlock! !

!PluggableSet methodsFor: 'accessing' stamp: 'ar 11/12/1998 19:02'!
hashBlock: aBlock
	"Set a new hash block. The block must accept one argument and return the hash value of the given argument."
	hashBlock := aBlock.! !

!PluggableSet methodsFor: 'private' stamp: 'ul 9/30/2009 14:19'!
scanForEmptySlotFor: anObject
	"Scan the key array for the first slot containing an empty slot (indicated by a nil). Answer the index of that slot. This method will be overridden in various subclasses that have different interpretations for matching elements."
	
	| index start |
	index := start := (hashBlock
		ifNil: [ anObject hash ]
		ifNotNil: [ hashBlock value: anObject ]) \\ array size + 1.
	[ 
		(array at: index) ifNil: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!PluggableSet methodsFor: 'private' stamp: 'ar 2/1/2010 20:17'!
scanFor: anObject 
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or raise an error if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."
	
	| index start |
	index := start := (hashBlock
		ifNil: [ anObject hash ]
		ifNotNil: [ hashBlock value: anObject ]) \\ array size + 1.
	[ 
		| element |
		((element := array at: index) == nil or: [
			equalBlock
				ifNil: [ element enclosedSetElement = anObject ]
				ifNotNil: [ equalBlock value: element enclosedSetElement value: anObject ] ])
			ifTrue: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!PluggableSet methodsFor: 'as yet unclassified' stamp: 'ul 10/13/2015 23:29'!
= anObject
	"Two sets are equal if
	 (a) they are the same 'kind' of thing.
	 (b) they have the same set of keys.
	 (c) for each (common) key, they have the same value"

	self == anObject ifTrue: [ ^true ].
	self species == anObject species ifFalse: [ ^false ].
	hashBlock = anObject hashBlock ifFalse: [ ^false ].
	equalBlock = anObject equalBlock ifFalse: [ ^false ].
	self size = anObject size ifFalse: [ ^false ].
	^self allSatisfy: [ :each | anObject includes: each ]! !

!Set class methodsFor: 'instance creation' stamp: ''!
newFrom: aCollection 
	"Answer an instance of me containing the same elements as aCollection."
	| newCollection |
	newCollection := self new: aCollection size.
	newCollection addAll: aCollection.
	^ newCollection
"
	Set newFrom: {1. 2. 3}
	{1. 2. 3} as: Set
"! !

!Set class methodsFor: 'initialization' stamp: 'ul 11/19/2009 01:15'!
quickRehashAllSets
	
	self deprecated: 'Use HashedCollection >> #rehashAll'.	
	HashedCollection rehashAll! !

!Set class methodsFor: 'initialization' stamp: 'ul 11/19/2009 01:15'!
rehashAllSets
	
	self deprecated: 'Use HashedCollection >> #rehashAll'.	
	HashedCollection rehashAll! !

!Set methodsFor: 'adding' stamp: 'cmm 10/13/2014 13:27'!
addNewElement: anObject 
	"Ensure anObject is part of the receiver.  Answer whether its membership was newly acquired."
	| index |
	index := self scanFor: anObject.
	^ (array at: index)
		ifNil:
			[ self
				atNewIndex: index
				put: anObject asSetElement.
			true ]
		ifNotNil: [ false ]! !

!Set methodsFor: 'adding' stamp: 'Igor.Stasenko 11/13/2009 06:59'!
add: newObject
	"Include newObject as one of the receiver's elements, but only if
	not already present. Answer newObject."

	| index |
	index := self scanFor: newObject.
	(array at: index) ifNil: [self atNewIndex: index put: newObject asSetElement].
	^ newObject! !

!Set methodsFor: 'converting' stamp: 'ar 11/20/1998 16:34'!
asSet
	^self! !

!Set methodsFor: 'enumerating' stamp: 'ul 11/20/2009 17:33'!
collect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into a collection like the receiver. Answer  
	the new collection."

	| newSet |
	newSet := Set new: self size.
	self do: [ :each | newSet add: (aBlock value: each) ].
	^newSet! !

!Set methodsFor: 'copying' stamp: 'sma 5/12/2000 14:45'!
copyWithout: oldElement 
	"Answer a copy of the receiver that does not contain any
	elements equal to oldElement."

	^ self copy
		remove: oldElement ifAbsent: [];
		yourself! !

!Set methodsFor: 'enumerating' stamp: 'ar 2/1/2010 19:55'!
do: aBlock 

	tally = 0 ifTrue: [ ^self ].
	1 to: array size do: [ :index |
		(array at: index) ifNotNil: [ :element |
			aBlock value: element enclosedSetElement] ]! !

!Set methodsFor: 'private' stamp: 'ul 2/4/2010 04:06'!
fixCollisionsFrom: start
	"The element at start has been removed and replaced by nil.
	This method moves forward from there, relocating any entries
	that had been placed below due to collisions with this one."

	| element index |
	index := start.
	[ (element := array at: (index := index \\ array size + 1)) == nil ] whileFalse: [
		| newIndex |
		(newIndex := self scanFor: element enclosedSetElement) = index ifFalse: [
			array 
				at: newIndex put: element;
				at: index put: nil ] ]! !

!Set methodsFor: 'testing' stamp: 'ul 3/10/2016 15:54'!
includes2: anObject 
	
	^((array at: (self scanFor: anObject)) == nil) not! !

!Set methodsFor: 'testing' stamp: 'ul 2/12/2011 01:05'!
includes: anObject 
	
	(array at: (self scanFor: anObject)) ifNil: [ ^false ] ifNotNil: [ ^true ]! !

!Set methodsFor: 'accessing' stamp: 'ar 2/1/2010 20:09'!
like: anObject
	"Answer an object in the receiver that is equal to anObject,
	nil if no such object is found. Relies heavily on hash properties"

	^(array at: (self scanFor: anObject)) ifNotNil:[:obj| obj enclosedSetElement]! !

!Set methodsFor: 'accessing' stamp: 'ul 2/19/2010 14:49'!
like: anObject ifAbsent: aBlock
	"Answer an object in the receiver that is equal to anObject,
	or evaluate the block if not found. Relies heavily on hash properties"
	
	^(array at: (self scanFor: anObject))
		ifNil: [ aBlock value ]
		ifNotNil: [ :element | element enclosedSetElement ]! !

!Set methodsFor: 'private' stamp: 'ar 2/1/2010 19:56'!
noCheckNoGrowFillFrom: anArray
	"Add the elements of anArray except nils to me assuming that I don't contain any of them, they are unique and I have more free space than they require."

	1 to: anArray size do: [ :index |
		(anArray at: index) ifNotNil: [ :object |
			array
				at: (self scanForEmptySlotFor: object enclosedSetElement)
				put: object ] ]! !

!Set methodsFor: 'enumerating' stamp: 'ul 3/9/2016 21:48'!
occurrencesOf: anObject
	"Answer how many of the receiver's elements are equal to anObject. Optimized version."

	(self includes: anObject) ifTrue: [ ^1 ].
	^0! !

!Set methodsFor: 'copying' stamp: 'ul 11/19/2009 01:45'!
postCopy
	super postCopy.
	array := array copy! !

!Set methodsFor: 'removing' stamp: 'ul 9/25/2009 00:38'!
remove: oldObject ifAbsent: aBlock

	| index |
	index := self scanFor: oldObject.
	(array at: index) ifNil: [ ^ aBlock value ].
	array at: index put: nil.
	tally := tally - 1.
	self fixCollisionsFrom: index.
	^ oldObject! !

!Set methodsFor: 'private' stamp: 'cmm 7/8/2013 17:56'!
scanFor: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or raise an error if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."

	| index start |
	index := start := anObject hash \\ array size + 1.
	[ 
		| element |
		((element := array at: index) == nil or: [ anObject = element enclosedSetElement ])
			ifTrue: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!Set methodsFor: 'enumerating' stamp: 'nice 5/22/2008 14:56'!
select: aBlock 
	"Use copyEmpty instead of self species new to give subclasses a chance to initialize additional inst vars."

	"Note: this code could be moved to super"
	
	| newCollection |
	newCollection := self copyEmpty.
	self do: [:each | (aBlock value: each) ifTrue: [newCollection add: each]].
	^newCollection! !

!Set methodsFor: 'comparing' stamp: 'ul 10/13/2015 23:29'!
= anObject
	"Two sets are equal if
	 (a) they are the same 'kind' of thing.
	 (b) they have the same set of keys.
	 (c) for each (common) key, they have the same value"

	self == anObject ifTrue: [ ^true ].
	self species == anObject species ifFalse: [ ^false ].
	self size = anObject size ifFalse: [ ^false ].
	^self allSatisfy: [ :each | anObject includes: each ]! !

Set subclass: #WeakSet
	instanceVariableNames: 'flag'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Weak'!

!WeakSet commentStamp: 'ul 11/20/2009 22:51' prior: 0!
I'm like Set but my instances only hold weakly to their elements.

Instance Variables:
	flag:		an Object which marks the empty slot in this instance. This object shouldn't be used anywhere else in the system. Every WeakSet has a different flag.

Differences from Set:
array is a WeakArray filled with flag initially. flag marks the empty slots, because elements which become garbage will be replaced with nil by the garbage collector. Besides nil, flag cannot be added to my instances.!

!WeakSet class methodsFor: 'as yet unclassified' stamp: 'ul 1/4/2016 14:54'!
arrayType

	^WeakArray! !

!WeakSet methodsFor: 'adding' stamp: 'ar 2/1/2010 21:31'!
add: newObject
	"Include newObject as one of the receiver's elements, but only if
	not already present. Answer newObject"

	| index element |
	index := self scanFor: newObject.
	((element := array at: index) == flag or: [ element == nil ])
		ifTrue: [self atNewIndex: index put: newObject asSetElement].
	^newObject! !

!WeakSet methodsFor: 'enumerating' stamp: 'ar 2/1/2010 21:32'!
collect: aBlock

	| newSet |
	newSet := self species new: self size.
	tally = 0 ifTrue: [ ^newSet ].
	1 to: array size do: [ :index |
		(array at: index) ifNotNil: [ :object |
			object == flag ifFalse: [
				newSet add: (aBlock value: object enclosedSetElement) ] ] ].
	^newSet! !

!WeakSet methodsFor: 'private' stamp: 'ul 4/12/2010 22:59'!
compact
	"Reduce the size of array so that the load factor will be ~75%."
	
	| newCapacity |
	newCapacity := self class goodPrimeAtLeast: self slowSize * 4 // 3.
	self growTo: newCapacity! !

!WeakSet methodsFor: 'enumerating' stamp: 'ul 9/4/2013 20:43'!
do: aBlock

	tally = 0 ifTrue: [ ^self ].
	1 to: array size do: [ :index |
		(array at: index) ifNotNil: [ :object |
			object == flag ifFalse: [
				aBlock value: object enclosedSetElement] ] ]! !

!WeakSet methodsFor: 'public' stamp: 'ul 9/12/2010 03:37'!
do: aBlock after: anElement

	| index endIndex |
	tally = 0 ifTrue: [ ^self ].
	anElement
		ifNil: [ index := 0 ]
		ifNotNil: [ 
			index := self scanFor: anElement.
			(array at: index) == flag ifTrue: [
				index := 0 ] ].
	endIndex := array size.
	[ (index := index + 1) <= endIndex ] whileTrue: [
		(array at: index) ifNotNil: [ :object |
			object == flag ifFalse: [
				aBlock value: object enclosedSetElement] ] ]! !

!WeakSet methodsFor: 'private' stamp: 'ul 2/5/2010 04:10'!
fixCollisionsFrom: start
	"The element at start has been removed and replaced by flag.
	This method moves forward from there, relocating any entries
	that had been placed below due to collisions with this one."

	| element index |
	index := start.
	[ (element := array at: (index := index \\ array size + 1)) == flag ] whileFalse: [
		element 
			ifNil: [ "This object is gone"
				array at: index put: flag.
				tally := tally - 1 ]
			ifNotNil: [
				| newIndex |
				(newIndex := self scanFor: element enclosedSetElement) = index ifFalse: [
					array 
						at: newIndex put: element;
						at: index put: flag ] ] ]
! !

!WeakSet methodsFor: 'private' stamp: 'ul 2/5/2010 04:35'!
growSize
	"Answer what my next table size should be.
	Note that, it can be less than the current."
	
	^self class goodPrimeAtLeast: self slowSize * 2 + 2! !

!WeakSet methodsFor: 'private' stamp: 'ul 1/4/2016 14:56'!
growTo: anInteger
	"Grow the elements array and reinsert the old elements"

	| oldElements |
	oldElements := array.
	array := self class arrayType new: anInteger withAll: flag.
	self noCheckNoGrowFillFrom: oldElements! !

!WeakSet methodsFor: 'testing' stamp: 'ul 2/12/2011 01:08'!
includes: anObject 
	
	(array at: (self scanFor: anObject))
		ifNil: [ ^false ]
		ifNotNil: [ :object | ^object ~~ flag ]! !

!WeakSet methodsFor: 'private' stamp: 'ul 1/4/2016 14:57'!
initialize: n
	"Initialize array to an array size of n"

	super initialize: n.
	flag := Object new.
	array atAllPut: flag! !

!WeakSet methodsFor: 'accessing' stamp: 'ar 2/1/2010 21:30'!
like: anObject
	"Answer an object in the receiver that is equal to anObject,
	nil if no such object is found. Relies heavily on hash properties"

	| element |
	^(element  := array at: (self scanFor: anObject)) == flag
		ifFalse: [ element enclosedSetElement]! !

!WeakSet methodsFor: 'accessing' stamp: 'ul 2/19/2010 14:56'!
like: anObject ifAbsent: aBlock
	"Answer an object in the receiver that is equal to anObject,
	or evaluate the block if not found. Relies heavily on hash properties"

	| element |
	((element  := array at: (self scanFor: anObject)) == flag or: [ element == nil ])
		ifTrue: [ ^aBlock value ]
		ifFalse: [ ^element enclosedSetElement ]! !

!WeakSet methodsFor: 'private' stamp: 'ar 2/1/2010 21:32'!
noCheckNoGrowFillFrom: anArray
	"Add the elements of anArray except nils and flag to me assuming that I don't contain any of them, they are unique and I have more free space than they require."

	tally := 0.
	1 to: anArray size do: [ :index |
		(anArray at: index) ifNotNil: [ :object |
			object == flag ifFalse: [ 
				array
					at: (self scanForEmptySlotFor: object enclosedSetElement)
					put: object.
				tally := tally + 1 ] ] ]! !

!WeakSet methodsFor: 'copying' stamp: 'nice 10/5/2009 08:59'!
postCopy
	| oldFlag |
	super postCopy.
	oldFlag := flag.
	flag := Object new.
	array replaceAll: oldFlag with: flag.! !

!WeakSet methodsFor: 'removing' stamp: 'ul 9/23/2009 21:14'!
remove: oldObject ifAbsent: aBlock

	| index |
	index := self scanFor: oldObject.
	(array at: index) == flag ifTrue: [ ^ aBlock value ].
	array at: index put: flag.
	tally := tally - 1.
	self fixCollisionsFrom: index.
	^oldObject! !

!WeakSet methodsFor: 'private' stamp: 'ul 9/30/2009 14:20'!
scanForEmptySlotFor: anObject
	"Scan the key array for the first slot containing an empty slot (indicated by flag or a nil). Answer the index of that slot. This method will be overridden in various subclasses that have different interpretations for matching elements."
	
	| index start |
	index := start := anObject hash \\ array size + 1.
	[ 
		| element |
		((element := array at: index) == flag or: [ element == nil ]) ifTrue: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!WeakSet methodsFor: 'private' stamp: 'ar 2/1/2010 21:30'!
scanForLoadedSymbol: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements"

	| element start finish |

	start := (anObject hash \\ array size) + 1.
	finish := array size.

	"Search from (hash mod size) to the end."
	start to: finish do:
		[:index | ((element := array at: index) == flag or: [element enclosedSetElement asString = anObject asString])
			ifTrue: [^ index ]].

	"Search from 1 to where we started."
	1 to: start-1 do:
		[:index | ((element := array at: index) == flag or: [element enclosedSetElement asString = anObject asString])
			ifTrue: [^ index ]].

	^ 0  "No match AND no empty slot"! !

!WeakSet methodsFor: 'private' stamp: 'ar 2/1/2010 21:30'!
scanFor: anObject
	"Scan the key array for the first slot containing either flag (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or raise an error if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."

	| index start |
	index := start := anObject hash \\ array size + 1.
	[ 
		| element |
		((element := array at: index) == flag or: [ element enclosedSetElement = anObject ])
			ifTrue: [ ^index ].
		(index := index \\ array size + 1) = start ] whileFalse.
	self errorNoFreeSpace! !

!WeakSet methodsFor: 'accessing' stamp: 'SqR 8/23/2000 15:12'!
size
	"Careful!! Answer the maximum amount
	of elements in the receiver, not the
	exact amount"

	^tally! !

!WeakSet methodsFor: 'public' stamp: 'ul 2/4/2010 01:10'!
slowSize
	"Careful!! Answer the maximum amount
	of elements in the receiver, not the
	exact amount"

	| count |
	count := 0.
	1 to: array size do: [ :index |
		(array at: index) ifNotNil: [ :object |
			object == flag ifFalse: [
				count := count + 1 ] ] ].
	^count! !

Collection subclass: #Heap
	instanceVariableNames: 'array tally sortBlock indexUpdateBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Heap'!

!Heap commentStamp: 'ul 6/2/2016 23:45' prior: 0!
I implement a special data structure called Binary Heap [ https://en.wikipedia.org/wiki/Binary_heap ], which is the most commonly used variant of the Heap data structure [ https://en.wikipedia.org/wiki/Heap_%28data_structure%29 ].
A Heap is a kind of binary tree stored in a linear array - see details after the instance variables description.

Heaps are good at handling priority queues because:
1) greatest priority element according to the sort block will be stored in first position and thus accessed in O(1) operations
2) worse time for inserting or removing an element is in O(log n) operations, where n is the size of the Heap
	Insertion/Removal times are more efficient than above upper bound, provided that:
	a) Elements are only removed at the beginning
	b) Elements are added with arbitrary sort order.
3) there is no need to fully sort the Heap, which makes it more efficient than a SortedCollection

The heap can be fully sorted by sending the message #sort.
Worse time for fully sorting the Heap is in O(n log n) operations, but this is rarely used a feature.
Remind that the Heap does not fully sort the collection if you don't ask.
Thus don't expect #do: and other iterators to enumerate elements according to the sortBlock order.

Instance variables:
	array		<Array>	The data repository
	tally		<Integer>	The number of elements in the heap
	sortBlock	<Block|nil>	A two-argument block defining the sort order,
							or nil in which case the default sort order is
								[:element1 :element2| element1 <= element2]
	indexUpdateBlock 	<Block|nil> 
							A two-argument block of the form [:data :index | ... ]
							which allows an application object to keep track of its
							index within the heap.  Useful for quick heap update
							when object's sort value changes (for example, when an
							object in a priority queue has its priority increased
							by an external event, you don't want to have to search
							through the whole heap to find the index before fixing
							the heap).  No update occurs if nil.
									
The Heap can be viewed as a binary tree (every node in the tree has at most two children).
The root is stored in first slot of internal array.
The children are stored in next two slots.
The children of children in next four slots.
etc...
For a node A of index i (1 based), the two children B1 and B2 are thus stored in indices (2*i) and (2*i+1).
Of course, the children indices must be less than the tally otherwise they are considered inexistent.

The Heap does arrange to preserve the following invariant:
For any children B of a node A, A is sorted before B, in other words, (sortBlock value: A value: B) = true
This implies that the root is always the first element according to sort order.

!

!Heap class methodsFor: 'instance creation' stamp: 'ar 7/1/1999 04:20'!
new
	^self new: 10! !

!Heap class methodsFor: 'instance creation' stamp: 'ar 7/1/1999 04:20'!
new: n
	^super new setCollection: (Array new: n)! !

!Heap class methodsFor: 'instance creation' stamp: 'ul 5/26/2016 19:58'!
on: anArray
	"Create a new heap using anArray as the internal array"
	
	^self on: anArray size: anArray size sortBlock: nil! !

!Heap class methodsFor: 'instance creation' stamp: 'ul 5/26/2016 19:57'!
on: anArray size: size
	"Create a new heap using the first size elements of anArray as the internal array"
	
	^self
		on: anArray
		size: size
		sortBlock: nil! !

!Heap class methodsFor: 'instance creation' stamp: 'ul 5/27/2016 20:02'!
on: anArray size: size sortBlock: aBlockOrNil
	"Create a new heap using the first size elements of anArray as the internal array and sorted by aBlockOrNil"
	
	anArray isArray ifFalse: [ self error: 'Array expected.' ].
	anArray size < size ifTrue: [ self error: 'size must not be larger than anArray size' ].
	^super new
		setCollection: anArray tally: size;
		sortBlock: aBlockOrNil;
		yourself! !

!Heap class methodsFor: 'instance creation' stamp: 'ar 9/10/1999 14:13'!
sortBlock: aBlock
	"Create a new heap sorted by the given block"
	^self new sortBlock: aBlock! !

!Heap class methodsFor: 'instance creation' stamp: 'ul 5/26/2016 20:00'!
withAll: aCollection
	"Create a new heap with all the elements from aCollection"
	
	^self withAll: aCollection sortBlock: nil! !

!Heap class methodsFor: 'instance creation' stamp: 'ul 5/26/2016 20:01'!
withAll: aCollection sortBlock: sortBlock
	"Create a new heap with all the elements from aCollection, sorted by sortBlock"
	
	| array |
	array := aCollection asArray.
	array == aCollection ifTrue: [ array := array copy ].
	^self on: array size: array size sortBlock: sortBlock! !

!Heap methodsFor: 'adding' stamp: 'ul 6/2/2016 11:27'!
add: anObject
	"Include newObject as one of the receiver's elements. Answer newObject."
	
	tally = array size ifTrue:[self grow].
	array at: (tally := tally + 1) put: anObject.
	indexUpdateBlock ifNotNil: [ indexUpdateBlock value: anObject value: tally ].
	self upHeap: tally.
	^anObject! !

!Heap methodsFor: 'private' stamp: 'ar 7/1/1999 04:19'!
array
	^array! !

!Heap methodsFor: 'accessing' stamp: 'ul 5/26/2016 00:20'!
at: index
	"Return the element at the given position within the receiver"
	
	index > tally ifTrue: [ ^self errorSubscriptBounds: index ].
	^array at: index! !

!Heap methodsFor: 'accessing' stamp: 'ar 7/1/1999 04:14'!
at: index put: newObject
	"Heaps are accessed with #add: not #at:put:"
	^self shouldNotImplement! !

!Heap methodsFor: 'accessing' stamp: 'ul 6/2/2016 14:38'!
capacity
	"Answer the current capacity of the receiver."

	^array size! !

!Heap methodsFor: 'enumerating' stamp: 'ul 5/27/2016 19:55'!
collect: aBlock

	^(array first: tally) replace: aBlock! !

!Heap methodsFor: 'growing' stamp: 'ul 6/2/2016 14:37'!
compact
	"Remove any empty slots in the receiver."

	self growTo: self size.! !

!Heap methodsFor: 'copying' stamp: 'nice 8/21/2010 15:53'!
copyEmpty
	"Answer a copy of the receiver without any of the receiver's elements."

	^self class sortBlock: sortBlock! !

!Heap methodsFor: 'private-heap' stamp: 'ul 6/2/2016 14:18'!
downHeapSingle: anIndex
	"This version is optimized for the case when only one element in the receiver can be at a wrong position. It avoids one comparison at each node when travelling down the heap and checks the heap upwards after the element is at a bottom position. Since the probability for being at the bottom of the heap is much larger than for being somewhere in the middle this version should be faster."

	| childIndex index value |
	index := anIndex.
	value := array at: anIndex.
	[ (childIndex := 2 * index) < tally ] whileTrue:[
		"Select the child with the larger value. We know there are two children."
		(sortBlock
			ifNil: [ (array at: childIndex + 1) <= (array at: childIndex) ]
			ifNotNil: [ sortBlock value: (array at: childIndex + 1) value: (array at: childIndex) ])
			ifTrue: [ childIndex := childIndex + 1 ].
		array at: index put: (array at: childIndex).
		indexUpdateBlock ifNotNil: [ indexUpdateBlock value: (array at: index) value: index ].
		"and repeat at the next level"
		index := childIndex ].
	childIndex = tally ifTrue: [ "Child with no sibling"
		array at: index put: (array at: childIndex).
		indexUpdateBlock ifNotNil: [ indexUpdateBlock value: (array at: index) value: index ].
		index := childIndex ].
	array at: index put: value.
	indexUpdateBlock ifNotNil: [ indexUpdateBlock value: value value: index ].	
	self upHeap: index! !

!Heap methodsFor: 'private-heap' stamp: 'ul 6/2/2016 14:18'!
downHeap: anIndex
	"Check the heap downwards for correctness starting at anIndex.
	 Everything above (i.e. left of) anIndex is ok."

	| childIndex childValue index value |
	index := anIndex.
	value := array at: anIndex.
	[ (childIndex := 2 * index) >= tally or: [
		"Select the child with the larger value. We know there are two children."
		childValue := array at: childIndex.
		(sortBlock
			ifNil: [ (array at: childIndex + 1) <= childValue ]
			ifNotNil: [ sortBlock value: (array at: childIndex + 1) value: childValue ])
			ifTrue: [ 
				childValue := array at: (childIndex := childIndex + 1) ].
		"Check if the value at index is at the right position."
		sortBlock
			ifNil: [ value <= childValue ]
			ifNotNil: [ sortBlock value: value value: childValue ] ] ]
		whileFalse: [
			"Move value downwards the tree."
			array at: index put: childValue.
			indexUpdateBlock ifNotNil: [ indexUpdateBlock value: childValue value: index ].
			"Contine from childIndex"
			index := childIndex ].
	childIndex = tally ifTrue: [ "Special case: there's only one child."
		"Check if the value at index is at the right position."
		childValue := array at: childIndex.
		(sortBlock
			ifNil: [ value <= childValue ]
			ifNotNil: [ sortBlock value: value value: childValue ]) 
			ifFalse: [
				"Move value downwards the tree."
				array at: index put: childValue.
				indexUpdateBlock ifNotNil: [ indexUpdateBlock value: childValue value: index ].
				"Contine from childIndex"
				index := childIndex ] ].
	array at: index put: value.
	indexUpdateBlock ifNotNil: [ indexUpdateBlock value: value value: index ]! !

!Heap methodsFor: 'enumerating' stamp: 'ar 9/10/1999 13:05'!
do: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument."
	1 to: tally do:[:i| aBlock value: (array at: i)]! !

!Heap methodsFor: 'accessing' stamp: 'md 1/19/2006 09:56'!
first
	"Return the first element in the receiver"
	^array at: 1! !

!Heap methodsFor: 'growing' stamp: 'ar 7/1/1999 04:17'!
grow
	"Become larger."
	self growTo: self size + self growSize.! !

!Heap methodsFor: 'growing' stamp: 'ar 7/1/1999 04:18'!
growSize
	"Return the size by which the receiver should grow if there are no empty slots left."
	^array size max: 5! !

!Heap methodsFor: 'growing' stamp: 'ul 6/2/2016 15:11'!
growTo: newSize
	"Grow to the requested size."

	| newArray |
	newArray := Array new: (newSize max: tally).
	newArray replaceFrom: 1 to: tally with: array startingAt: 1.
	array := newArray! !

!Heap methodsFor: 'accessing' stamp: 'ul 5/27/2016 19:37'!
indexUpdateBlock: aBlockOrNil
	"aBlockOrNil is either nil or a two argument block. The first argument is the object whose index has changed in the heap, the second is the new index. The block will be evaluated whenever an element is moved in the heap's internal array. If you don't plan to remove elements by index (see #removeAt:), then you should not set this block."

	indexUpdateBlock := aBlockOrNil.

! !

!Heap methodsFor: 'testing' stamp: 'ar 9/10/1999 13:03'!
isEmpty
	"Answer whether the receiver contains any elements."
	^tally = 0! !

!Heap methodsFor: 'testing' stamp: 'rhi 8/14/2003 08:51'!
isHeap

	^ true! !

!Heap methodsFor: 'testing' stamp: 'ul 6/2/2016 14:18'!
isValidHeap

	"Check the size first."
	(tally between: 0 and: array size) ifFalse: [ ^false ].
	"Check the sort order between parent and child nodes."
	1 to: (tally bitShift: -1) do: [ :index |
		| childIndex |
		childIndex := 2 * index.
		sortBlock
			ifNil: [ (array at: index) <= (array at: childIndex) ifFalse: [ ^false ] ]
			ifNotNil: [ (sortBlock value: (array at: index) value: (array at: childIndex)) ifFalse: [ ^false ] ].
		childIndex < tally ifTrue: [
			childIndex := childIndex + 1.
			sortBlock
				ifNil: [ (array at: index) <= (array at: childIndex) ifFalse: [ ^false ] ]
				ifNotNil: [ (sortBlock value: (array at: index) value: (array at: childIndex)) ifFalse: [ ^false ] ] ] ].
	"Check for elements left in array after tally."
	tally + 1 to: array size do: [ :index |
		(array at: index) ifNotNil: [ ^false ] ].
	^true! !

!Heap methodsFor: 'copying' stamp: 'nice 10/5/2009 08:47'!
postCopy
	super postCopy.
	array := array copy! !

!Heap methodsFor: 'private' stamp: 'ul 6/2/2016 15:12'!
privateRemoveAt: index
	"Remove the element at the given index and make sure the sorting order is okay. The value of index must not be larger than tally."

	| removed |
	removed := array at: index.
	index = tally ifTrue: [
		array at: index put: nil.
		tally := tally - 1.
		^removed ].
	array 
		at: index put: (array at: tally);
		at: tally put: nil.
	tally := tally - 1.
	2 * index <= tally "The node at index has at least one child."
		ifTrue: [ self downHeapSingle: index ]
		ifFalse: [ self upHeap: index ].
	^removed! !

!Heap methodsFor: 'private' stamp: 'ul 6/2/2016 15:13'!
privateReverseSort
	"Arrange to have the array sorted in reverse order.
	WARNING: this method breaks the heap invariants. It's up to the sender to restore them afterwards."
	
	| oldTally |
	self deprecated: 'Use #sort if you want to sort.'.
	oldTally := tally.
	[tally > 1] whileTrue:
		[array swap: 1 with: tally.
		tally := tally - 1.
		self downHeapSingle: 1].
	tally := oldTally! !

!Heap methodsFor: 'removing' stamp: 'klub 9/14/2009 19:10'!
removeAll

	array atAllPut: nil.
	tally := 0! !

!Heap methodsFor: 'removing' stamp: 'ul 5/27/2016 19:50'!
removeAt: index
	"Remove the element at the given index and make sure the sorting order is okay."

	index > tally ifTrue: [ self errorSubscriptBounds: index ].
	^self privateRemoveAt: index! !

!Heap methodsFor: 'removing' stamp: 'ar 9/10/1999 13:05'!
removeFirst
	"Remove the first element from the receiver"
	^self removeAt: 1! !

!Heap methodsFor: 'removing' stamp: 'ar 9/10/1999 13:04'!
remove: oldObject ifAbsent: aBlock
	"Remove oldObject as one of the receiver's elements. If several of the 
	elements are equal to oldObject, only one is removed. If no element is 
	equal to oldObject, answer the result of evaluating anExceptionBlock. 
	Otherwise, answer the argument, oldObject."
	1 to: tally do:[:i| 
		(array at: i) = oldObject ifTrue:[^self privateRemoveAt: i]].
	^aBlock value! !

!Heap methodsFor: 'enumerating' stamp: 'ul 5/27/2016 19:57'!
select: aBlock 
	"Evaluate aBlock with each of my elements as the argument. Collect into
	a new collection like the receiver, only those elements for which aBlock
	evaluates to true."

	| newCollection |
	newCollection := self copyEmpty.
	1 to: tally do: [ :index |
		| element |
		(aBlock value: (element := array at: index)) ifTrue: [
			newCollection add: element ] ].
	^ newCollection! !

!Heap methodsFor: 'private' stamp: 'ar 7/1/1999 04:35'!
setCollection: aCollection
	array := aCollection.
	tally := 0.! !

!Heap methodsFor: 'private' stamp: 'ar 9/10/1999 13:18'!
setCollection: aCollection tally: newTally
	array := aCollection.
	tally := newTally.! !

!Heap methodsFor: 'accessing' stamp: 'sma 5/12/2000 11:37'!
size
	"Answer how many elements the receiver contains."

	^ tally! !

!Heap methodsFor: 'sorting' stamp: 'ul 6/2/2016 14:21'!
sort
	"Fully sort the heap. This method preserves the heap invariants and can thus be sent safely"

	| start end element originalIndexUpdateBlock |
	end := tally.
	"Temporarly remove indexUpdateBlock to speed up sorting."
	originalIndexUpdateBlock := indexUpdateBlock.
	indexUpdateBlock := nil.
	[ tally > 1 ] whileTrue: [
		element := array at: tally.
		array 
			at: tally put: (array at: 1);
			at: 1 put: element.
		tally := tally - 1.
		self downHeapSingle: 1 ].
	tally := end.
	start := 1.
	originalIndexUpdateBlock ifNil: [
		"The was no indexUpdateBlock; just reverse the elements"
		[ start < end ] whileTrue: [
			element := array at: start.
			array
				at: start put: (array at: end);
				at: end put: element.
			start := start + 1.
			end := end - 1 ].
		^self ].
	"Restore indexUpdateBlock, reverse the elements and update the indices."
	indexUpdateBlock := originalIndexUpdateBlock.
	start := 1.
	[ start < end ] whileTrue: [
		| endValue |
		element := array at: start.
		endValue := array at: end.
		array
			at: start put: endValue;
			at: end put: element.
		indexUpdateBlock
			value: endValue value: start;
			value: element value: end.
		start := start + 1.
		end := end - 1 ].
	start = end ifTrue: [ indexUpdateBlock value: (array at: start) value: start ]! !

!Heap methodsFor: 'accessing' stamp: 'ar 7/1/1999 04:21'!
sortBlock
	^sortBlock! !

!Heap methodsFor: 'accessing' stamp: 'ul 6/2/2016 15:28'!
sortBlock: aBlock

	| oldIndexUpdateBlock |
	sortBlock := aBlock.
	"Restore the heap invariant."
	tally <= 1 ifTrue: [ ^self ].
	oldIndexUpdateBlock := indexUpdateBlock.
	indexUpdateBlock := nil.
	(tally bitShift: -1) to: 1 by: -1 do: [ :index | self downHeap: index ].
	indexUpdateBlock := oldIndexUpdateBlock ifNil: [ ^self ].
	1 to: tally do: [ :index |
		indexUpdateBlock value: (array at: index) value: index ]
	
! !

!Heap methodsFor: 'private-heap' stamp: 'ul 5/26/2016 00:38'!
upHeap: anIndex
	"Check the heap upwards for correctness starting at anIndex.
	 Everything below anIndex is ok."

	| index parentValue parentIndex value |
	anIndex = 1 ifTrue: [ ^self ].
	value := array at: (index := anIndex).
	[ index > 1 and: [
		parentValue := array at: (parentIndex := index bitShift: -1).
		sortBlock
			ifNil: [ value <= parentValue ]
			ifNotNil: [ sortBlock value: value value: parentValue ] ] ]
		whileTrue: [
			array at: index put: parentValue.
			indexUpdateBlock ifNotNil: [ indexUpdateBlock value: parentValue value: index ].
			index := parentIndex ].
	array at: index put: value.
	indexUpdateBlock ifNotNil: [ indexUpdateBlock value: value value: index ]! !

!Heap methodsFor: 'comparing' stamp: 'ul 5/27/2016 19:39'!
= anObject
	"Heap are considered to be equal when they have the same sortBlock and the same elements. This method is expensive due to the sorted copies of the arrays. Try not to use it."

	self == anObject ifTrue: [ ^true ].
	anObject isHeap ifFalse: [ ^false ].
	anObject size = tally ifFalse: [ ^false ].
	anObject sortBlock = sortBlock ifFalse: [ ^false ].
	^((array first: tally) sort: sortBlock) = ((anObject array first: tally) sort: sortBlock)! !

Collection subclass: #Matrix
	instanceVariableNames: 'nrows ncols contents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Unordered'!

!Matrix commentStamp: '<historical>' prior: 0!
I represent a two-dimensional array, rather like Array2D.
There are three main differences between me and Array2D:
(1) Array2D inherits from ArrayedCollection, but isn't one.  A lot of things that should work
    do not work in consequence of this.
(2) Array2D uses "at: column at: row" index order, which means that nothing you write using
    it is likely to work either.  I use the almost universal "at: row at: column" order, so it is
    much easier to adapt code from other languages without going doolally.
(3) Array2D lets you specify the class of the underlying collection, I don't.

Structure:
  nrows : a non-negative integer saying how many rows there are.
  ncols : a non-negative integer saying how many columns there are.
  contents : an Array holding the elements in row-major order.  That is, for a 2x3 array
    the contents are (11 12 13 21 22 23).  Array2D uses column major order.

    You can specify the class of 'contents' when you create a new Array2D,
    but Matrix always gives you an Array.

    There is a reason for this.  In strongly typed languages like Haskell and Clean,
    'unboxed arrays' save you both space AND time.  But in Squeak, while
    WordArray and FloatArray and so on do save space, it costs time to use them.
    A LOT of time.  I've measured aFloatArray sum running nearly twice as slow as
    anArray sum.  The reason is that whenever you fetch an element from an Array,
    that's all that happens, but when you fetch an element from aFloatArray, a whole
    new Float gets allocated to hold the value.  This takes time and churns memory.
    So the paradox is that if you want fast numerical stuff, DON'T use unboxed arrays!!

    Another reason for always insisting on an Array is that letting it be something
    else would make things like #, and #,, rather more complicated.  Always using Array
    is the simplest thing that could possibly work, and it works rather well.

I was trying to patch Array2D to make more things work, but just couldn't get my head
around the subscript order.  That's why I made Matrix.

Element-wise matrix arithmetic works; you can freely mix matrices and numbers but
don't try to mix matrices and arrays (yet).
Matrix multiplication, using the symbol +* (derived from APL's +.x), works between
(Matrix or Array) +* (Matrix or Array).  Don't try to use a number as an argument of +*.
Matrix * Number and Number * Matrix work fine, so you don't need +* with numbers.

Still to come: oodles of stuff.  Gaussian elimination maybe, other stuff probably not.
!

!Matrix class methodsFor: 'instance creation' stamp: 'raok 10/23/2002 20:58'!
column: aCollection
	"Should this be called #fromColumn:?"

	^self rows: aCollection size columns: 1 contents: aCollection asArray shallowCopy! !

!Matrix class methodsFor: 'instance creation' stamp: 'raok 10/22/2002 00:09'!
diagonal: aCollection
	|r i|
	r := self zeros: aCollection size.
	i := 0.
	aCollection do: [:each | i := i+1. r at: i at: i put: each].
	^r! !

!Matrix class methodsFor: 'instance creation' stamp: 'raok 10/23/2002 20:59'!
identity: n
	|r|

	r := self zeros: n.
	1 to: n do: [:i | r at: i at: i put: 1].
	^r! !

!Matrix class methodsFor: 'instance creation' stamp: 'raok 10/22/2002 00:06'!
new: dim
	"Answer a dim*dim matrix.  Is this an abuse of #new:?  The argument is NOT a size."
	^self rows: dim columns: dim! !

!Matrix class methodsFor: 'instance creation' stamp: 'raok 11/25/2002 12:51'!
new: dim element: element
	"Answer a dim*dim matrix with all elements set to element.
	 Is this an abuse of #new:?  The argument is NOT a size."

	^self rows: dim columns: dim element: element! !

!Matrix class methodsFor: 'instance creation' stamp: 'raok 10/22/2002 19:54'!
new: dim tabulate: aBlock
	"Answer a dim*dim matrix where it at: i at: j is aBlock value: i value: j."
	^self rows: dim columns: dim tabulate: aBlock! !

!Matrix class methodsFor: 'instance creation' stamp: 'raok 11/28/2002 14:08'!
ones: n
	^self new: n element: 1
! !

!Matrix class methodsFor: 'instance creation' stamp: 'raok 10/22/2002 00:04'!
rows: rows columns: columns
	^self rows: rows columns: columns contents: (Array new: rows*columns)! !

!Matrix class methodsFor: 'private' stamp: 'raok 10/21/2002 23:06'!
rows: rows columns: columns contents: contents
	^self new rows: rows columns: columns contents: contents! !

!Matrix class methodsFor: 'instance creation' stamp: 'raok 11/28/2002 14:10'!
rows: rows columns: columns element: element
	^self rows: rows columns: columns
		contents: ((Array new: rows*columns) atAllPut: element; yourself)! !

!Matrix class methodsFor: 'instance creation' stamp: 'raok 10/22/2002 19:51'!
rows: rows columns: columns tabulate: aBlock
	"Answer a new Matrix of the given dimensions where
	 result at: i at: j     is   aBlock value: i value: j"
	|a i|

	a := Array new: rows*columns.
	i := 0.
	1 to: rows do: [:row |
		1 to: columns do: [:column |
			a at: (i := i+1) put: (aBlock value: row value: column)]].
	^self rows: rows columns: columns contents: a
! !

!Matrix class methodsFor: 'instance creation' stamp: 'raok 10/23/2002 20:59'!
row: aCollection
	"Should this be called #fromRow:?"

	^self rows: 1 columns: aCollection size contents: aCollection asArray shallowCopy! !

!Matrix class methodsFor: 'instance creation' stamp: 'raok 11/28/2002 14:09'!
zeros: n
	^self new: n element: 0! !

!Matrix methodsFor: 'adding' stamp: 'raok 10/21/2002 22:53'!
add: newObject
	self shouldNotImplement! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:37'!
anyOne
	^contents anyOne! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 22:57'!
asArray
	^contents shallowCopy! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 22:57'!
asBag
	^contents asBag! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 22:58'!
asByteArray
	^contents asByteArray! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 22:58'!
asCharacterSet
	^contents asCharacterSet! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 23:00'!
asFloatArray
	^contents asFloatArray! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 22:58'!
asIdentitySet
	^contents asIdentitySet! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 23:00'!
asIntegerArray
	^contents asIntegerArray! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 22:58'!
asOrderedCollection
	^contents asOrderedCollection! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 22:58'!
asSet
	^contents asSet! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 22:58'!
asSortedArray
	^contents asSortedArray! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 22:59'!
asSortedCollection
	^contents asSortedCollection! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 22:59'!
asSortedCollection: aBlock
	^contents asSortedCollection: aBlock! !

!Matrix methodsFor: 'converting' stamp: 'raok 10/21/2002 23:00'!
asWordArray
	^contents asWordArray! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:42'!
atAllPut: value
	contents atAllPut: value! !

!Matrix methodsFor: 'accessing rows/columns' stamp: 'raok 11/22/2002 12:41'!
atColumn: column
	|p|

	p := (self indexForRow: 1 andColumn: column)-ncols.
	^(1 to: nrows) collect: [:row | contents at: (p := p+ncols)]
! !

!Matrix methodsFor: 'accessing rows/columns' stamp: 'raok 11/28/2002 14:21'!
atColumn: column put: aCollection
	|p|

	aCollection size = nrows ifFalse: [self error: 'wrong column size'].
	p := (self indexForRow: 1 andColumn: column)-ncols.
	aCollection do: [:each | contents at: (p := p+ncols) put: each].
	^aCollection
! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:43'!
atRandom
	^contents atRandom
! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:43'!
atRandom: aGenerator
	^contents atRandom: aGenerator! !

!Matrix methodsFor: 'accessing submatrices' stamp: 'raok 11/25/2002 13:09'!
atRows: rs columns: cs
	"Answer a Matrix obtained by slicing the receiver.
	 rs and cs should be sequenceable collections of positive integers."

	^self class rows: rs size columns: cs size tabulate: [:r :c |
		self at: (rs at: r) at: (cs at: c)]! !

!Matrix methodsFor: 'accessing submatrices' stamp: 'raok 11/25/2002 12:30'!
atRows: r1 to: r2 columns: c1 to: c2
	"Answer a submatrix [r1..r2][c1..c2] of the receiver."
	|rd cd|

	rd := r1 - 1.
	cd := c1 - 1.
	^self class rows: r2-rd columns: c2-cd tabulate: [:r :c| self at: r+rd at: c+cd]
! !

!Matrix methodsFor: 'accessing submatrices' stamp: 'raok 11/25/2002 13:05'!
atRows: r1 to: r2 columns: c1 to: c2 ifInvalid: element
	"Answer a submatrix [r1..r2][c1..c2] of the receiver.
	 Portions of the result outside the bounds of the original matrix
	 are filled in with element."
	|rd cd|

	rd := r1 - 1.
	cd := c1 - 1.
	^self class rows: r2-rd columns: c2-cd tabulate: [:r :c| self at: r+rd at: c+cd ifInvalid: element]
! !

!Matrix methodsFor: 'accessing submatrices' stamp: 'raok 11/25/2002 12:32'!
atRows: r1 to: r2 columns: c1 to: c2 put: aMatrix
	"Set the [r1..r2][c1..c2] submatrix of the receiver
	 from the [1..r2-r1+1][1..c2-c1+1] submatrix of aMatrix.
	 As long as aMatrix responds to at:at: and accepts arguments in the range shown,
	 we don't care if it is bigger or even if it is a Matrix at all."
	|rd cd|

	rd := r1 - 1.
	cd := c1 - 1.
	r1 to: r2 do: [:r |
		c1 to: c2 do: [:c |
			self at: r at: c put: (aMatrix at: r-rd at: c-cd)]].
	^aMatrix
! !

!Matrix methodsFor: 'accessing rows/columns' stamp: 'raok 10/21/2002 23:32'!
atRow: row
	(row between: 1 and: nrows)
		ifFalse: [self error: '1st subscript out of range'].
	^contents copyFrom: (row-1)*ncols+1 to: row*ncols! !

!Matrix methodsFor: 'accessing rows/columns' stamp: 'raok 11/22/2002 12:42'!
atRow: row put: aCollection
	|p|

	aCollection size = ncols ifFalse: [self error: 'wrong row size'].
	p := (self indexForRow: row andColumn: 1)-1.
	aCollection do: [:each | contents at: (p := p+1) put: each].
	^aCollection! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:37'!
at: row at: column
	^contents at: (self indexForRow: row andColumn: column)! !

!Matrix methodsFor: 'accessing' stamp: 'raok 11/28/2002 14:14'!
at: r at: c ifInvalid: v
	"If r,c is a valid index for this matrix, answer the corresponding element.
	 Otherwise, answer v."

	(r between: 1 and: nrows) ifFalse: [^v].
	(c between: 1 and: ncols) ifFalse: [^v].
	^contents at: (r-1)*ncols + c
! !

!Matrix methodsFor: 'accessing' stamp: 'raok 11/22/2002 12:37'!
at: row at: column incrementBy: value
	"Array2D>>at:at:add: was the origin of this method, but in Smalltalk add:
	 generally suggests adding an element to a collection, not doing a sum.
	 This method, and SequenceableCollection>>at:incrementBy: that supports
	 it, have been renamed to reveal their intention more clearly."

	^contents at: (self indexForRow: row andColumn: column) incrementBy: value! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:40'!
at: row at: column put: value
	^contents at: (self indexForRow: row andColumn: column) put: value! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:41'!
collect: aBlock
	"Answer a new matrix with transformed elements; transformations should be independent."

	^self class rows: nrows columns: ncols contents: (contents collect: aBlock)! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:44'!
columnCount
	^ncols! !

!Matrix methodsFor: 'accessing rows/columns' stamp: 'raok 10/23/2002 20:41'!
diagonal
	"Answer (1 to: (nrows min: ncols)) collect: [:i | self at: i at: i]"
	|i|

	i := ncols negated.
	^(1 to: (nrows min: ncols)) collect: [:j | contents at: (i := i + ncols + 1)]! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:42'!
difference: aCollection
	"Union is in because the result is always a Set.
	 Difference and intersection are out because the result is like the receiver,
	 and with irregular seleection that cannot be."
	self shouldNotImplement! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:40'!
do: aBlock
	"Pass elements to aBlock one at a time in row-major order."
	contents do: aBlock! !

!Matrix methodsFor: 'comparing' stamp: 'raok 11/22/2002 13:14'!
hash
	"I'm really not sure what would be a good hash function here.
	 The essential thing is that it must be compatible with #=, and
	 this satisfies that requirement."

	^contents hash! !

!Matrix methodsFor: 'testing' stamp: 'raok 10/21/2002 23:24'!
identityIncludes: anObject
	^contents identityIncludes: anObject! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:48'!
identityIndexOf: anElement
	^self identityIndexOf: anElement ifAbsent: [0@0]
! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:48'!
identityIndexOf: anElement ifAbsent: anExceptionBlock
	^self rowAndColumnForIndex:
		 (contents identityIndexOf: anElement ifAbsent: [^anExceptionBlock value])
! !

!Matrix methodsFor: 'testing' stamp: 'raok 10/21/2002 23:24'!
includesAllOf: aCollection
	^contents includesAllOf: aCollection! !

!Matrix methodsFor: 'testing' stamp: 'raok 10/21/2002 23:24'!
includesAnyOf: aCollection
	^contents includesAnyOf: aCollection! !

!Matrix methodsFor: 'testing' stamp: 'raok 10/21/2002 23:23'!
includes: anObject
	^contents includes: anObject! !

!Matrix methodsFor: 'private' stamp: 'raok 10/21/2002 22:40'!
indexForRow: row andColumn: column
	(row between: 1 and: nrows)
		ifFalse: [self error: '1st subscript out of range'].
	(column between: 1 and: ncols)
		ifFalse: [self error: '2nd subscript out of range'].
	^(row-1) * ncols + column! !

!Matrix methodsFor: 'accessing' stamp: 'raok 11/22/2002 13:13'!
indexOf: anElement
	"If there are integers r, c such that (self at: r at: c) = anElement,
	 answer some such r@c, otherwise answer 0@0.  This kind of perverse
	 result is provided by analogy with SequenceableCollection>>indexOf:.
	 The order in which the receiver are searched is UNSPECIFIED except
	 that it is the same as the order used by #indexOf:ifAbsent: and #readStream."

	^self indexOf: anElement ifAbsent: [0@0]
! !

!Matrix methodsFor: 'accessing' stamp: 'raok 11/22/2002 13:10'!
indexOf: anElement ifAbsent: anExceptionBlock
	"If there are integers r, c such that (self at: r at: c) = anElement,
	 answer some such r@c, otherwise answer the result of anExceptionBlock."

	^self rowAndColumnForIndex:
		 (contents indexOf: anElement ifAbsent: [^anExceptionBlock value])
! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/23/2002 20:57'!
indicesCollect: aBlock
	|r i|

	r := Array new: nrows * ncols.
	i := 0.
	1 to: nrows do: [:row |
		1 to: ncols do: [:column |
			r at: (i := i+1) put: (aBlock value: row value: column)]].
	^self class rows: nrows columns: ncols contents: r! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:49'!
indicesDo: aBlock
	1 to: nrows do: [:row |
		1 to: ncols do: [:column |
			aBlock value: row value: column]].! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:51'!
indicesInject: start into: aBlock
	|current|

	current := start.
	1 to: nrows do: [:row |
		1 to: ncols do: [:column |
			current := aBlock value: current value: row value: column]].
	^current! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:42'!
intersection: aCollection
	"Union is in because the result is always a Set.
	 Difference and intersection are out because the result is like the receiver,
	 and with irregular seleection that cannot be."
	self shouldNotImplement! !

!Matrix methodsFor: 'testing' stamp: 'raok 11/22/2002 13:03'!
isSequenceable
	"LIE so that arithmetic on matrices will work.
	 What matters for arithmetic is not that there should be random indexing
	 but that the structure should be stable and independent of the values of
	 the elements.  #isSequenceable is simply the wrong question to ask."
	^true! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:25'!
occurrencesOf: anObject
	^contents occurrencesOf: anObject! !

!Matrix methodsFor: 'copying' stamp: 'nice 10/5/2009 09:09'!
postCopy
	super postCopy.
	contents := contents copy! !

!Matrix methodsFor: 'arithmetic' stamp: 'raok 11/28/2002 14:22'!
preMultiplyByArray: a
	"Answer a +* self where a is an Array."

	nrows = 1 ifFalse: [self error: 'dimensions do not conform'].
	^Matrix rows: a size columns: ncols tabulate: [:row :col |
		(a at: row) * (contents at: col)]
! !

!Matrix methodsFor: 'arithmetic' stamp: 'nice 12/27/2009 03:10'!
preMultiplyByMatrix: m
	"Answer m +* self where m is a Matrix."
	

	nrows = m columnCount ifFalse: [self error: 'dimensions do not conform'].
	^Matrix rows: m rowCount columns: ncols tabulate: [:row :col | | s |
		s := 0.
		1 to: nrows do: [:k | s := (m at: row at: k) * (self at: k at: col) + s].
		s]! !

!Matrix methodsFor: 'private' stamp: 'raok 11/22/2002 12:56'!
privateContents
	"Only used in #, #,, and #= so far.
	 It used to be called #contents, but that clashes with Collection>>contents."

	^contents! !

!Matrix methodsFor: 'converting' stamp: 'raok 11/22/2002 13:02'!
readStream
	"Answer a ReadStream that returns all the elements of the receiver
	 in some UNSPECIFIED order."

	^ReadStream on: contents! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:42'!
reject: aBlock
	self shouldNotImplement! !

!Matrix methodsFor: 'removing' stamp: 'klub 9/14/2009 16:34'!
removeAll

	self shouldNotImplement! !

!Matrix methodsFor: 'removing' stamp: 'raok 10/21/2002 22:54'!
remove: anObject ifAbsent: anExceptionBlock
	self shouldNotImplement! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:49'!
replaceAll: oldObject with: newObject
	contents replaceAll: oldObject with: newObject! !

!Matrix methodsFor: 'private' stamp: 'raok 10/21/2002 22:47'!
rowAndColumnForIndex: index
	|t|

	t := index - 1.
	^(t // ncols + 1)@(t \\ ncols + 1)! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:44'!
rowCount
	^nrows! !

!Matrix methodsFor: 'private' stamp: 'raok 10/21/2002 23:05'!
rows: rows columns: columns contents: anArray
	self assert: [rows isInteger and: [rows >= 0]].
	self assert: [columns isInteger and: [columns >= 0]].
	self assert: [rows * columns = anArray size].
	nrows := rows.
	ncols := columns.
	contents := anArray.
	^self! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:42'!
select: aBlock
	self shouldNotImplement! !

!Matrix methodsFor: 'copying' stamp: 'raok 10/21/2002 23:27'!
shuffled
	^self class rows: nrows columns: ncols contents: (contents shuffled)! !

!Matrix methodsFor: 'copying' stamp: 'raok 10/21/2002 23:27'!
shuffledBy: aRandom
	^self class rows: nrows columns: ncols contents: (contents shuffledBy: aRandom)! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:49'!
size
	^contents size! !

!Matrix methodsFor: 'printing' stamp: 'raok 10/21/2002 23:22'!
storeOn: aStream
	aStream nextPut: $(; nextPutAll: self class name;
		nextPutAll: ' rows: '; store: nrows;
		nextPutAll: ' columns: '; store: ncols;
		nextPutAll: ' contents: '; store: contents;
		nextPut: $)! !

!Matrix methodsFor: 'accessing rows/columns' stamp: 'raok 11/28/2002 14:21'!
swapColumn: anIndex withColumn: anotherIndex
	|a b|

	a := self indexForRow: 1 andColumn: anIndex.
	b := self indexForRow: 1 andColumn: anotherIndex.
	nrows timesRepeat: [
		contents swap: a with: b.
		a := a + ncols.
		b := b + ncols].
! !

!Matrix methodsFor: 'accessing rows/columns' stamp: 'raok 11/28/2002 14:22'!
swapRow: anIndex withRow: anotherIndex
	|a b|

	a := self indexForRow: anIndex andColumn: 1.
	b := self indexForRow: anotherIndex andColumn: 1.
	ncols timesRepeat: [
		contents swap: a with: b.
		a := a + 1.
		b := b + 1].
! !

!Matrix methodsFor: 'accessing' stamp: 'raok 10/21/2002 22:52'!
swap: r1 at: c1 with: r2 at: c2
	contents swap: (self indexForRow: r1 andColumn: c1)
			 with: (self indexForRow: r2 andColumn: c2)! !

!Matrix methodsFor: 'accessing rows/columns' stamp: 'raok 10/22/2002 00:13'!
transposed
	self assert: [nrows = ncols].
	^self indicesCollect: [:row :column | self at: column at: row]! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:52'!
withIndicesCollect: aBlock
	|i r|

	i := 0.
	r := contents shallowCopy.
	1 to: nrows do: [:row |
		1 to: ncols do: [:column |
			i := i+1.
			r at: i put: (aBlock value: (r at: i) value: row value: column)]].
	^self class rows: nrows columns: ncols contents: r
! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:52'!
withIndicesDo: aBlock
	|i|

	i := 0.
	1 to: nrows do: [:row |
		1 to: ncols do: [:column |
			aBlock value: (contents at: (i := i+1)) value: row value: column]].
! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:52'!
withIndicesInject: start into: aBlock
	|i current|

	i := 0.
	current := start.
	1 to: nrows do: [:row |
		1 to: ncols do: [:column |
			current := aBlock value: current value: (contents at: (i := i+1)) 
							  value: row value: column]].
	^current! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/22/2002 00:15'!
with: aCollection collect: aBlock
	"aCollection must support #at:at: and be at least as large as the receiver."

	^self withIndicesCollect: [:each :row :column |
		aBlock value: each value: (aCollection at: row at: column)]
! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:53'!
with: aCollection do: aBlock
	"aCollection must support #at:at: and be at least as large as the receiver."

	self withIndicesDo: [:each :row :column |
		aBlock value: each value: (aCollection at: row at: column)].
! !

!Matrix methodsFor: 'enumerating' stamp: 'raok 10/21/2002 23:55'!
with: aCollection inject: startingValue into: aBlock
	"aCollection must support #at:at: and be at least as large as the receiver."

	^self withIndicesInject: startingValue into: [:value :each :row :column |
		aBlock value: value value: each value: (aCollection at: row at: column)]! !

!Matrix methodsFor: 'arithmetic' stamp: 'raok 10/22/2002 20:01'!
+* aCollection
	"Premultiply aCollection by self.  aCollection should be an Array or Matrix.
	 The name of this method is APL's +.x squished into Smalltalk syntax."

	^aCollection preMultiplyByMatrix: self
! !

!Matrix methodsFor: 'copying' stamp: 'raok 11/22/2002 12:57'!
, aMatrix
	"Answer a new matrix having the same number of rows as the receiver and aMatrix,
	 its columns being the columns of the receiver followed by the columns of aMatrix."
	|newCont newCols anArray oldCols a b c|

	self assert: [nrows = aMatrix rowCount].
	newCont := Array new: self size + aMatrix size.
	anArray := aMatrix privateContents.
	oldCols := aMatrix columnCount.
	newCols := ncols + oldCols.
	a := b := c := 1.
	1 to: nrows do: [:r |
		newCont replaceFrom: a to: a+ncols-1 with: contents startingAt: b.
		newCont replaceFrom: a+ncols to: a+newCols-1 with: anArray startingAt: c.
		a := a + newCols.
		b := b + ncols.
		c := c + oldCols].
	^self class rows: nrows columns: newCols contents: newCont
		
! !

!Matrix methodsFor: 'copying' stamp: 'raok 11/22/2002 12:58'!
,, aMatrix
	"Answer a new matrix having the same number of columns as the receiver and aMatrix,
	 its rows being the rows of the receiver followed by the rows of aMatrix."

	self assert: [ncols = aMatrix columnCount].
	^self class rows: nrows + aMatrix rowCount columns: ncols
		contents: contents , aMatrix privateContents
! !

!Matrix methodsFor: 'comparing' stamp: 'raok 11/22/2002 12:58'!
= aMatrix
	^aMatrix class == self class and: [
	 aMatrix rowCount = nrows and: [
	 aMatrix columnCount = ncols and: [
	 aMatrix privateContents = contents]]]! !

Collection subclass: #SequenceableCollection
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Abstract'!

!SequenceableCollection commentStamp: '<historical>' prior: 0!
I am an abstract superclass for collections that have a well-defined order associated with their elements. Thus each element is externally-named by integers referred to as indices.!

SequenceableCollection subclass: #ArrayedCollection
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Abstract'!

!ArrayedCollection commentStamp: '<historical>' prior: 0!
I am an abstract collection of elements with a fixed range of integers (from 1 to n>=0) as external keys.!

ArrayedCollection variableSubclass: #Array
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Arrayed'!

!Array commentStamp: '<historical>' prior: 0!
I present an ArrayedCollection whose elements are objects.!

!Array class methodsFor: 'brace support' stamp: 'di 11/18/1999 22:53'!
braceStream: nElements
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	^ WriteStream basicNew braceArray: (self new: nElements)
! !

!Array class methodsFor: 'brace support' stamp: 'di 11/19/1999 08:16'!
braceWithNone
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	^ self new: 0! !

!Array class methodsFor: 'brace support' stamp: 'di 11/19/1999 08:16'!
braceWith: a
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	| array |
	array := self new: 1.
	array at: 1 put: a.
	^ array! !

!Array class methodsFor: 'brace support' stamp: 'di 11/19/1999 08:15'!
braceWith: a with: b 
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	| array |
	array := self new: 2.
	array at: 1 put: a.
	array at: 2 put: b.
	^ array! !

!Array class methodsFor: 'brace support' stamp: 'di 11/19/1999 08:17'!
braceWith: a with: b with: c 
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	| array |
	array := self new: 3.
	array at: 1 put: a.
	array at: 2 put: b.
	array at: 3 put: c.
	^ array! !

!Array class methodsFor: 'brace support' stamp: 'di 11/19/1999 08:17'!
braceWith: a with: b with: c with: d
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	| array |
	array := self new: 4.
	array at: 1 put: a.
	array at: 2 put: b.
	array at: 3 put: c.
	array at: 4 put: d.
	^ array! !

!Array class methodsFor: 'instance creation' stamp: 'cmm 2/11/2010 21:12'!
empty
	"A canonicalized empty Array instance."
	^ #()! !

!Array class methodsFor: 'instance creation' stamp: 'md 7/19/2004 12:34'!
new: sizeRequested 
	"Answer an instance of this class with the number of indexable
	variables specified by the argument, sizeRequested.
	
	This is a shortcut (direct call of primitive, no #initialize, for performance"

	<primitive: 71>  "This method runs primitively if successful"
	^ self basicNew: sizeRequested  "Exceptional conditions will be handled in basicNew:"
! !

!Array methodsFor: 'converting' stamp: 'sma 5/12/2000 17:32'!
asArray
	"Answer with the receiver itself."

	^ self! !

!Array methodsFor: 'accessing' stamp: 'ar 8/26/2001 22:02'!
atWrap: index 
	"Optimized to go through the primitive if possible"
	<primitive: 60>
	^ self at: index - 1 \\ self size + 1! !

!Array methodsFor: 'accessing' stamp: 'ar 8/26/2001 22:03'!
atWrap: index put: anObject
	"Optimized to go through the primitive if possible"
	<primitive: 61>
	^ self at: index - 1 \\ self size + 1 put: anObject! !

!Array methodsFor: 'filter streaming' stamp: 'MPW 1/1/1901 01:42'!
byteEncode:aStream
	aStream writeArray:self.
! !

!Array methodsFor: 'copying' stamp: 'ar 2/11/2001 01:55'!
copyWithDependent: newElement
	self size = 0 ifTrue:[^DependentsArray with: newElement].
	^self copyWith: newElement! !

!Array methodsFor: 'converting' stamp: 'eem 6/23/2016 16:22'!
elementsExchangeIdentityWith: otherArray
	"This primitive performs a bulk mutation, causing all pointers to the elements of the
	 receiver to be replaced by pointers to the corresponding elements of otherArray.
	 At the same time, all pointers to the elements of otherArray are replaced by
	 pointers to the corresponding elements of this array.  The identityHashes remain
	 with the pointers rather than with the objects so that objects in hashed structures
	 should still be properly indexed after the mutation."

	<primitive: 128 error: ec>
	ec == #'bad receiver' ifTrue:
		[^self error: 'receiver must be of class Array'].
	ec == #'bad argument' ifTrue:
		[^self error: (otherArray class == Array
						ifTrue: ['arg must be of class Array']
						ifFalse: ['receiver and argument must have the same size'])].
	ec == #'inappropriate operation' ifTrue:
		[^self error: 'can''t become immediates such as SmallIntegers or Characters'].
	ec == #'no modification' ifTrue:
		[^self error: 'can''t become immutable objects'].
	ec == #'object is pinned' ifTrue:
		[^self error: 'can''t become pinned objects'].
	ec == #'insufficient object memory' ifTrue:
		[| maxRequired |
		 "In Spur, two-way become may involve making each pair of objects into a forwarder into a copy of the other.
		 So if become fails with #'insufficient object memory', garbage collect, and if necessary, grow memory."
		 maxRequired := (self detectSum: [:obj | obj class byteSizeOfInstanceOfSize: obj basicSize])
						+ (otherArray detectSum: [:obj | obj class byteSizeOfInstanceOfSize: obj basicSize]).
		 (Smalltalk garbageCollectMost < maxRequired
		  and: [Smalltalk garbageCollect < maxRequired]) ifTrue:
			[Smalltalk growMemoryByAtLeast: maxRequired].
		 ^self elementsExchangeIdentityWith: otherArray].
	self primitiveFailed! !

!Array methodsFor: 'converting' stamp: 'eem 6/23/2016 16:24'!
elementsForwardIdentityTo: otherArray
	"This primitive performs a bulk mutation, causing all pointers to the elements of the
	 receiver to be replaced by pointers to the corresponding elements of otherArray.
	 The identityHashes remain with the pointers rather than with the objects so that
	 the objects in this array should still be properly indexed in any existing hashed
	 structures after the mutation."
	<primitive: 72 error: ec>
	ec == #'bad receiver' ifTrue:
		[^self error: 'receiver must be of class Array'].
	ec == #'bad argument' ifTrue:
		[^self error: (otherArray class == Array
						ifTrue: ['arg must be of class Array']
						ifFalse: ['receiver and argument must have the same size'])].
	ec == #'inappropriate operation' ifTrue:
		[^self error: 'can''t become immediates such as SmallIntegers or Characters'].
	ec == #'no modification' ifTrue:
		[^self error: 'can''t become immutable objects'].
	ec == #'object is pinned' ifTrue:
		[^self error: 'can''t become pinned objects'].
	ec == #'insufficient object memory' ifTrue:
		[self error: 'The virtual machine is out-of-date.  Please upgrade.'].
	self primitiveFailed! !

!Array methodsFor: 'converting' stamp: 'eem 7/7/2016 11:08'!
elementsForwardIdentityTo: otherArray copyHash: copyHash
	"This primitive performs a bulk mutation, causing all pointers to the elements of the
	 receiver to be replaced by pointers to the corresponding elements of otherArray.
	 If copyHash is true, the identityHashes remain with the pointers rather than with the
	 objects so that the objects in the receiver should still be properly indexed in any
	 existing hashed structures after the mutation.  If copyHash is false, then the hashes
	 of the objects in otherArray remain unchanged.  If you know what you're doing this
	 may indeed be what you want."
	<primitive: 249 error: ec>
	ec == #'bad receiver' ifTrue:
		[^self error: 'receiver must be of class Array'].
	ec == #'bad argument' ifTrue:
		[^self error: (otherArray class == Array
						ifTrue: ['arg must be of class Array']
						ifFalse: ['receiver and argument must have the same size'])].
	ec == #'inappropriate operation' ifTrue:
		[^self error: 'can''t become immediates such as SmallIntegers or Characters'].
	ec == #'no modification' ifTrue:
		[^self error: 'can''t become immutable objects'].
	ec == #'object is pinned' ifTrue:
		[^self error: 'can''t become pinned objects'].
	self primitiveFailed! !

!Array methodsFor: 'converting' stamp: 'nice 12/27/2009 03:10'!
evalStrings
	   "Allows you to construct literal arrays.
    #(true false nil '5@6' 'Set new' '''text string''') evalStrings
    gives an array with true, false, nil, a Point, a Set, and a String
    instead of just a bunch of Symbols"
    

    ^ self collect: [:each | | it |
        it := each.
        each == #true ifTrue: [it := true].
		      each == #false ifTrue: [it := false].
        each == #nil ifTrue: [it := nil].
        (each isString and:[each isSymbol not]) ifTrue: [
			it := Compiler evaluate: each].
        each class == Array ifTrue: [it := it evalStrings].
        it]! !

!Array methodsFor: 'private' stamp: 'md 3/1/2006 21:09'!
hasLiteralSuchThat: testBlock
	"Answer true if testBlock returns true for any literal in this array, even if imbedded in 	further Arrays or CompiledMethods.  This method is only intended for private use by 	CompiledMethod 	hasLiteralSuchThat:"
	| lit |
	1 to: self size do: [:index |
		(testBlock value: (lit := self at: index)) ifTrue: [^ true].
		(lit hasLiteralSuchThat: testBlock) ifTrue: [^ true]].
	^ false! !

!Array methodsFor: 'private' stamp: 'nice 7/20/2011 09:03'!
hasLiteral: literal
	"Answer true if literal is identical to any literal in this array, even 
	if imbedded in further array structure. This method is only intended 
	for private use by CompiledMethod hasLiteralSymbol:"

	| lit |
	1 to: self size do: 
		[:index | 
		((lit := self at: index) literalEqual: literal) ifTrue: [^true].
		(Array == lit class and: [lit hasLiteral: literal]) ifTrue: [^true]].
	^false! !

!Array methodsFor: 'testing' stamp: 'md 7/30/2005 21:19'!
isArray
	^true! !

!Array methodsFor: 'testing' stamp: 'ul 2/3/2014 14:31'!
isLiteral
	
	^self class == Array and: [
		self isLiteralIfContainedBy: IdentitySet new ]! !

!Array methodsFor: 'testing' stamp: 'ul 2/3/2014 14:33'!
isLiteralIfContainedBy: parents
	" Answer whether the receiver has a literal text form recognized by the compiler. Precondition: the receiver is an instance of Array. "

	(parents includes: self) ifTrue: [ ^false ].
	parents add: self.
	1 to: self size do: [ :index |
		| element |
		element := self at: index.
		(element class == Array
			ifTrue: [ element isLiteralIfContainedBy: parents ]
			ifFalse: [ element isLiteral ]) ifFalse: [ ^false ] ].
	parents remove: self.
	^true! !

!Array methodsFor: 'comparing' stamp: 'ajh 2/2/2002 15:03'!
literalEqual: other

	self class == other class ifFalse: [^ false].
	self size = other size ifFalse: [^ false].
	self with: other do: [:e1 :e2 |
		(e1 literalEqual: e2) ifFalse: [^ false]].
	^ true! !

!Array methodsFor: 'arithmetic' stamp: 'raok 10/22/2002 20:10'!
preMultiplyByArray: a
	"Answer a+*self where a is an Array.  Arrays are always understood as column vectors,
	 so an n element Array is an n*1 Array.  This multiplication is legal iff self size = 1."

	self size = 1 ifFalse: [self error: 'dimensions do not conform'].
	^a * self first! !

!Array methodsFor: 'arithmetic' stamp: 'nice 12/27/2009 02:49'!
preMultiplyByMatrix: m
	"Answer m+*self where m is a Matrix."
	m columnCount = self size ifFalse: [self error: 'dimensions do not conform'].
	^(1 to: m rowCount) collect: [:row |
		| s |
		s := 0.
		1 to: self size do: [:k | s := (m at: row at: k) * (self at: k) + s].
		s]! !

!Array methodsFor: 'self evaluating' stamp: 'nice 10/30/2009 23:50'!
printAsBraceFormOn: aStream

	aStream nextPut: ${.
	self do: [:el | aStream print: el] separatedBy: [ aStream nextPutAll: ' . '].
	aStream nextPut: $}! !

!Array methodsFor: 'self evaluating' stamp: 'sd 7/31/2005 21:44'!
printAsLiteralFormOn: aStream
	aStream nextPut: $#.
	self printElementsOn: aStream
! !

!Array methodsFor: 'printing' stamp: 'ul 11/23/2010 13:28'!
printOn: aStream
	self shouldBePrintedAsLiteral ifTrue: [^self printAsLiteralFormOn: aStream].
	self class = Array ifTrue: [^self printAsBraceFormOn: aStream].
	^super printOn: aStream! !

!Array methodsFor: 'private' stamp: ''!
replaceFrom: start to: stop with: replacement startingAt: repStart 
	"Primitive. This destructively replaces elements from start to stop in the receiver starting at index, repStart, in the collection, replacement. Answer the receiver. Range checks are performed in the primitive only. Optional. See Object documentation whatIsAPrimitive."
	<primitive: 105>
	super replaceFrom: start to: stop with: replacement startingAt: repStart! !

!Array methodsFor: 'testing' stamp: 'eem 11/2/2011 11:02'!
shouldBePrintedAsLiteral

	^self class == Array
	  and: [self shouldBePrintedAsLiteralVisiting: (IdentitySet new: 8)]! !

!Array methodsFor: 'testing' stamp: 'eem 11/2/2011 11:01'!
shouldBePrintedAsLiteralVisiting: aSet
	self class == Array ifFalse:
		[^false].
	(aSet includes: self) ifTrue:
		[^false].
	aSet add: self.
	^self allSatisfy: [:each | each shouldBePrintedAsLiteralVisiting: aSet]! !

!Array methodsFor: 'sorting' stamp: 'ul 3/22/2010 12:16'!
sorted: aSortBlockOrNil
	"Return a new sequenceable collection which contains the same elements as self but its elements are sorted by aSortBlockOrNil. The block should take two arguments and return true if the first element should preceed the second one. If aSortBlock is nil then <= is used for comparison."
	
	^self copy sort: aSortBlockOrNil! !

!Array methodsFor: 'printing' stamp: 'ul 11/23/2010 13:28'!
storeOn: aStream 
	"Use the literal form if possible."

	self shouldBePrintedAsLiteral
		ifTrue: 
			[aStream nextPut: $#; nextPut: $(.
			self do: 
				[:element | 
				element storeOn: aStream.
				aStream space].
			aStream nextPut: $)]
		ifFalse: [super storeOn: aStream]! !

!Array methodsFor: 'arithmetic' stamp: 'raok 10/22/2002 20:09'!
+* aCollection
	"Premultiply aCollection by self.  aCollection should be an Array or Matrix.
	 The name of this method is APL's +.x squished into Smalltalk syntax."

	^aCollection preMultiplyByArray: self
! !

Array weakSubclass: #WeakArray
	instanceVariableNames: ''
	classVariableNames: 'FinalizationDependents FinalizationLock FinalizationProcess FinalizationSemaphore'
	poolDictionaries: ''
	category: 'Collections-Weak'!

!WeakArray commentStamp: '<historical>' prior: 0!
WeakArray is an array which holds only weakly on its elements. This means whenever an object is only referenced by instances of WeakArray it will be garbage collected.!

!WeakArray class methodsFor: 'accessing' stamp: 'ul 8/27/2010 23:38'!
addWeakDependent: anObject

	FinalizationLock
		critical: [
			| emptySlotIndex |
			emptySlotIndex := FinalizationDependents 
				identityIndexOf: nil
				ifAbsent: [ 
					| newIndex |
					newIndex := FinalizationDependents size + 1.
					"Grow linearly"
					FinalizationDependents := FinalizationDependents grownBy: 10.
					newIndex ].
			FinalizationDependents at: emptySlotIndex put: anObject ]
		ifError: [ :msg :rcvr | rcvr error: msg ]! !

!WeakArray class methodsFor: 'class initialization' stamp: 'nk 6/21/2004 10:22'!
initialize
	"WeakArray initialize"

	"Do we need to initialize specialObjectsArray?"
	Smalltalk specialObjectsArray size < 42 
		ifTrue:[Smalltalk recreateSpecialObjectsArray].

	Smalltalk addToStartUpList: self.
	self restartFinalizationProcess.! !

!WeakArray class methodsFor: 'accessing' stamp: 'ul 8/27/2010 23:28'!
isFinalizationSupported
	"This method is only here for backwards compatibility, all closure VMs support finalization"
	
	^true! !

!WeakArray class methodsFor: 'accessing' stamp: 'ul 8/27/2010 23:25'!
removeWeakDependent: anObject

	FinalizationLock critical:[
		1 to: FinalizationDependents size do:[:i|
			((FinalizationDependents at: i) == anObject) ifTrue:[
				FinalizationDependents at: i put: nil.
			].
		].
	] ifError:[:msg :rcvr| rcvr error: msg].! !

!WeakArray class methodsFor: 'private' stamp: 'ul 8/27/2010 23:26'!
restartFinalizationProcess
	"kill any old process, just in case"
	FinalizationProcess
		ifNotNil: [FinalizationProcess terminate.
			FinalizationProcess := nil].

	FinalizationSemaphore := Smalltalk specialObjectsArray at: 42.
	FinalizationDependents ifNil: [FinalizationDependents := WeakArray new: 10].
	FinalizationLock := Semaphore forMutualExclusion.
	FinalizationProcess := [self finalizationProcess]
		forkAt: Processor userInterruptPriority! !

!WeakArray class methodsFor: 'accessing' stamp: 'nk 10/28/2000 20:26'!
runningFinalizationProcess
	"Answer the FinalizationProcess I am running, if any"
	^FinalizationProcess! !

!WeakArray class methodsFor: 'system startup' stamp: 'nk 6/21/2004 09:22'!
startUp: resuming
	resuming ifFalse: [ ^self ].
	self restartFinalizationProcess.! !

!WeakArray methodsFor: 'copying' stamp: 'bf 4/7/2015 14:01'!
copyWithout: oldElement 
	"Re-implemented to not return a strong copy."
	^ (super copyWithout: oldElement) as: self class! !

!WeakArray methodsFor: 'copying' stamp: 'bf 4/7/2015 14:01'!
copyWith: newElement 
	"Re-implemented to not return a strong copy."
	^ (super copyWith: newElement) as: self class! !

!WeakArray methodsFor: 'private' stamp: 'cmm 2/5/2015 15:03'!
species
	"More useful to have strongly-referenced results of #select: and #collect:."
	^ Array! !

!ArrayedCollection class methodsFor: 'instance creation' stamp: ''!
new
	"Answer a new instance of me, with size = 0."

	^self new: 0! !

!ArrayedCollection class methodsFor: 'instance creation' stamp: 'ul 3/15/2010 04:52'!
newFromStream: s
	"Only meant for my subclasses that are raw bits and word-like.  For quick unpack form the disk."
	
	| len |
	(self isPointers or: [ self isWords not ]) ifTrue: [ ^self ].
	s next = 16r80 ifTrue: [
		"A compressed format.  Could copy what BitMap does, or use a 
		special sound compression format.  Callers normally compress their own way."
		^self error: 'not implemented' ].
	s skip: -1.
	len := s nextInt32.
	^s nextWordsInto: (self basicNew: len)! !

!ArrayedCollection class methodsFor: 'instance creation' stamp: ''!
newFrom: aCollection 
	"Answer an instance of me containing the same elements as aCollection."
	| newArray |
	newArray := self new: aCollection size.
	1 to: aCollection size do: [:i | newArray at: i put: (aCollection at: i)].
	^ newArray

"	Array newFrom: {1. 2. 3}
	{1. 2. 3} as: Array
	{1. 2. 3} as: ByteArray
	{$c. $h. $r} as: String
	{$c. $h. $r} as: Text
"! !

!ArrayedCollection class methodsFor: 'instance creation' stamp: ''!
new: size withAll: value 
	"Answer an instance of me, with number of elements equal to size, each 
	of which refers to the argument, value."

	^(self new: size) atAllPut: value! !

!ArrayedCollection class methodsFor: 'instance creation' stamp: 'sma 5/12/2000 17:37'!
withAll: aCollection
	"Create a new collection containing all the elements from aCollection."

	^ (self new: aCollection size) replaceFrom: 1 to: aCollection size with: aCollection! !

!ArrayedCollection class methodsFor: 'instance creation' stamp: ''!
with: anObject 
	"Answer a new instance of me, containing only anObject."

	| newCollection |
	newCollection := self new: 1.
	newCollection at: 1 put: anObject.
	^newCollection! !

!ArrayedCollection class methodsFor: 'instance creation' stamp: ''!
with: firstObject with: secondObject 
	"Answer a new instance of me, containing firstObject and secondObject."

	| newCollection |
	newCollection := self new: 2.
	newCollection at: 1 put: firstObject.
	newCollection at: 2 put: secondObject.
	^newCollection! !

!ArrayedCollection class methodsFor: 'instance creation' stamp: ''!
with: firstObject with: secondObject with: thirdObject 
	"Answer a new instance of me, containing only the three arguments as
	elements."

	| newCollection |
	newCollection := self new: 3.
	newCollection at: 1 put: firstObject.
	newCollection at: 2 put: secondObject.
	newCollection at: 3 put: thirdObject.
	^newCollection! !

!ArrayedCollection class methodsFor: 'instance creation' stamp: ''!
with: firstObject with: secondObject with: thirdObject with: fourthObject 
	"Answer a new instance of me, containing only the three arguments as
	elements."

	| newCollection |
	newCollection := self new: 4.
	newCollection at: 1 put: firstObject.
	newCollection at: 2 put: secondObject.
	newCollection at: 3 put: thirdObject.
	newCollection at: 4 put: fourthObject.
	^newCollection! !

!ArrayedCollection class methodsFor: 'instance creation' stamp: ''!
with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject
	"Answer a new instance of me, containing only the five arguments as
	elements."

	| newCollection |
	newCollection := self new: 5.
	newCollection at: 1 put: firstObject.
	newCollection at: 2 put: secondObject.
	newCollection at: 3 put: thirdObject.
	newCollection at: 4 put: fourthObject.
	newCollection at: 5 put: fifthObject.
	^newCollection! !

!ArrayedCollection class methodsFor: 'instance creation' stamp: 'sw 10/24/1998 22:22'!
with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject with: sixthObject
	"Answer a new instance of me, containing only the 6 arguments as elements."

	| newCollection |
	newCollection := self new: 6.
	newCollection at: 1 put: firstObject.
	newCollection at: 2 put: secondObject.
	newCollection at: 3 put: thirdObject.
	newCollection at: 4 put: fourthObject.
	newCollection at: 5 put: fifthObject.
	newCollection at: 6 put: sixthObject.
	^ newCollection! !

!ArrayedCollection methodsFor: 'adding' stamp: 'sma 5/12/2000 14:09'!
add: newObject
	self shouldNotImplement! !

!ArrayedCollection methodsFor: 'converting' stamp: 'sma 5/12/2000 18:18'!
asSortedArray
	self isSorted ifTrue: [^ self asArray].
	^ super asSortedArray! !

!ArrayedCollection methodsFor: 'objects from disk' stamp: 'nk 3/17/2004 15:22'!
byteSize
	^self basicSize * self bytesPerBasicElement
! !

!ArrayedCollection methodsFor: 'objects from disk' stamp: 'nk 3/17/2004 16:28'!
bytesPerBasicElement
	"Answer the number of bytes that each of my basic elements requires.
	In other words:
		self basicSize * self bytesPerBasicElement
	should equal the space required on disk by my variable sized representation."
	^self class isBytes ifTrue: [ 1 ] ifFalse: [ 4 ]! !

!ArrayedCollection methodsFor: 'objects from disk' stamp: 'nk 3/17/2004 18:51'!
bytesPerElement
	^self class isBytes ifTrue: [ 1 ] ifFalse: [ 4 ].
! !

!ArrayedCollection methodsFor: 'private' stamp: ''!
defaultElement

	^nil! !

!ArrayedCollection methodsFor: 'private' stamp: 'ul 11/21/2009 01:15'!
fillFrom: aCollection with: aBlock
	"Evaluate aBlock with each of aCollections's elements as the argument.  
	Collect the resulting values into self. Answer self."

	| index |
	index := 0.
	aCollection do: [ :each |
		self at: (index := index + 1) put: (aBlock value: each) ]! !

!ArrayedCollection methodsFor: 'filter streaming' stamp: 'sma 5/12/2000 14:20'!
flattenOnStream: aStream 
	aStream writeArrayedCollection: self! !

!ArrayedCollection methodsFor: 'enumerating' stamp: 'nice 2/8/2010 09:19'!
inject: anObject into: aBlock
	"Optimized for speed. See super."
	
	| result |
	result := anObject.
	1 to: self size do: [:i | result := aBlock value: result value: (self at: i)].
	^result! !

!ArrayedCollection methodsFor: 'sorting' stamp: 'ul 8/8/2013 14:17'!
isSorted
	"Return true if the receiver is sorted by #<=."

	^self isSortedBetween: 1 and: self size! !

!ArrayedCollection methodsFor: 'sorting' stamp: 'ul 8/8/2013 14:41'!
isSortedBetween: startIndex and: endIndex
	"Return true if the receiver is sorted by #<= between startIndex and endIndex."

	| previousElement |
	endIndex < startIndex ifTrue: [ ^true ].
	previousElement := self at: startIndex.
	startIndex + 1 to: endIndex do: [ :index |
		| element |
		element := self at: index.
		previousElement <= element ifFalse: [ ^false ].
		previousElement := element ].
	^true! !

!ArrayedCollection methodsFor: 'sorting' stamp: 'ul 8/8/2013 14:18'!
isSortedBy: aSortBlockOrNil
	"Return true if the receiver is sorted by aSortBlockOrNil. Use #<= for comparison if aSortBlockOrNil is nil."

	^self isSortedBy: aSortBlockOrNil between: 1 and: self size! !

!ArrayedCollection methodsFor: 'sorting' stamp: 'ul 8/8/2013 14:41'!
isSortedBy: aSortBlockOrNil between: startIndex and: endIndex
	"Return true if the receiver is sorted by aSortBlockOrNil between startIndex and endIndex. Use #<= for comparison if aSortBlockOrNil is nil."

	| previousElement |
	aSortBlockOrNil ifNil: [ ^self isSortedBetween: startIndex and: endIndex ].
	endIndex < startIndex ifTrue: [ ^true ].
	previousElement := self at: startIndex.
	startIndex + 1 to: endIndex do: [ :index |
		| element |
		element := self at: index.
		(aSortBlockOrNil value: previousElement value: element) ifFalse: [ ^false ].
		previousElement := element ].
	^true! !

!ArrayedCollection methodsFor: 'sorting' stamp: 'ul 11/2/2009 03:26'!
mergeFirst: first middle: middle last: last into: dst by: aBlock
	"Private. Merge the sorted ranges [first..middle] and [middle+1..last] 
	of the receiver into the range [first..last] of dst."

	| i1 i2 val1 val2 out |
	i1 := first.
	i2 := middle + 1.
	val1 := self at: i1.
	val2 := self at: i2.
	out := first - 1.  "will be pre-incremented"

	"select 'lower' half of the elements based on comparator"
	[ (i1 <= middle) and: [ i2 <= last ] ] whileTrue: 	[
		(aBlock 
			ifNil: [ val1 <= val2 ]
			ifNotNil: [ aBlock value: val1 value: val2 ])
				ifTrue: [
					dst at: (out := out + 1) put: val1.
					val1 := self at: (i1 := i1 + 1)]
				ifFalse: [
					dst at: (out := out + 1) put: val2.
					(i2 := i2 + 1) <= last ifTrue: [
						val2 := self at: i2 ] ] ].

	"copy the remaining elements"
	i1 <= middle
		ifTrue: [dst replaceFrom: out + 1 to: last with: self startingAt: i1]
		ifFalse: [dst replaceFrom: out + 1 to: last with: self startingAt: i2]! !

!ArrayedCollection methodsFor: 'sorting' stamp: 'sma 5/12/2000 14:25'!
mergeSortFrom: startIndex to: stopIndex by: aBlock
	"Sort the given range of indices using the mergesort algorithm.
	Mergesort is a worst-case O(N log N) sorting algorithm that usually
	does only half as many comparisons as heapsort or quicksort."

	"Details: recursively split the range to be sorted into two halves,
	mergesort each half, then merge the two halves together. An extra 
	copy of the data is used as temporary storage and successive merge 
	phases copy data back and forth between the receiver and this copy.
	The recursion is set up so that the final merge is performed into the
	receiver, resulting in the receiver being completely sorted."

	self size <= 1 ifTrue: [^ self].  "nothing to do"
	startIndex = stopIndex ifTrue: [^ self].
	self assert: [startIndex >= 1 and: [startIndex < stopIndex]]. "bad start index"
	self assert: [stopIndex <= self size]. "bad stop index"
	self
		mergeSortFrom: startIndex
		to: stopIndex 
		src: self clone 
		dst: self 
		by: aBlock! !

!ArrayedCollection methodsFor: 'sorting' stamp: 'sma 5/12/2000 14:26'!
mergeSortFrom: first to: last src: src dst: dst by: aBlock
	"Private. Split the range to be sorted in half, sort each half, and 
	merge the two half-ranges into dst."

	| middle |
	first = last ifTrue: [^ self].
	middle := (first + last) // 2.
	self mergeSortFrom: first to: middle src: dst dst: src by: aBlock.
	self mergeSortFrom: middle + 1 to: last src: dst dst: src by: aBlock.
	src mergeFirst: first middle: middle last: last into: dst by: aBlock! !

!ArrayedCollection methodsFor: 'enumerating' stamp: 'ul 3/8/2016 17:18'!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject. Optimized version."

	| tally |
	tally := 0.
	1 to: self size do: [ :index |
		(self at: index) = anObject ifTrue: [ tally := tally + 1 ] ].
	^tally! !

!ArrayedCollection methodsFor: 'sorting' stamp: 'ul 5/27/2016 19:16'!
quickSort
	"Sort elements of self to be nondescending according to #<= using an in-place quicksort with simple median-of-three partitioning with guaranteed O(log(n)) space usage."

	self quickSortFrom: 1 to: self size by: nil! !

!ArrayedCollection methodsFor: 'sorting' stamp: 'ul 11/13/2011 05:13'!
quickSortFrom: from to: to by: sortBlock
	"Sort elements i through j of self to be nondescending according to sortBlock using an in-place quicksort with simple median-of-three partitioning with guaranteed O(log(n)) space usage."

	| dij k l temp i j di dj n ij |
	i := from.
	j := to.
	[
		
		"The prefix d means the data at that index."
		"Sort di,dj."
		di := self at: i.
		dj := self at: j.
		(sortBlock ifNil: [ di <= dj ] ifNotNil: [ sortBlock value: di value: dj ]) ifFalse: [
			self at: i put: dj; at: j put: di. temp := dj. dj := di. di := temp "swap di with dj" ].
		(n := j + 1 - i) <= 2 ifTrue: [ ^self ].
		"More than two elements."
		dij := self at: (ij := i + j // 2). "ij is the midpoint of i and j. Sort di,dij,dj. Make dij be their median."
		(sortBlock ifNil: [ di <= dij ] ifNotNil: [ sortBlock value: di value: dij ])
			ifTrue: [
				(sortBlock ifNil: [ dij <= dj ] ifNotNil: [ sortBlock value: dij value: dj ]) ifFalse: [
					 "swap dij with dj, we don't need the value of the variable dj anymore"
					self at: j put: dij; at: ij put: dj. dij := dj ] ]
			ifFalse: [
				 "swap di with dij, we don't need the value of the variable di anymore"
				self at: i put: dij; at: ij put: di. dij := di ].
		n = 3 ifTrue: [ ^self ].
		"More than three elements."
		"Find k and l such that i<k<l<j and dk,dij,dl are in reverse order. Swap k and l. Repeat this procedure until k and l pass each other."
		k := i.
		l := j.
		[
			[ k <= (l := l - 1) and: [ 
				sortBlock ifNil: [ dij <= (self at: l) ] ifNotNil: [ sortBlock value: dij value: (self at: l) ] ] ] whileTrue.  "i.e. while dl succeeds dij"
			[ (k := k + 1) <= l and: [
				sortBlock ifNil: [ (self at: k) <= dij ] ifNotNil: [ sortBlock value: (self at: k) value: dij ] ] ] whileTrue.  "i.e. while dij succeeds dk"
			k <= l ] whileTrue: [ temp := self at: k. self at: k put: (self at: l); at: l put: temp. ].
		"Now l<k (either 1 or 2 less), and di through dl are all less than or equal to dk through dj. Sort the larger segment in this method and call another quicksort for the smaller segment. This ensures O(log(n)) space usage."
		i < l 
			ifFalse: [
				k < j
					ifFalse: [ ^self ]
					ifTrue: [ i := k ] ]
			ifTrue: [
				k < j
					ifFalse: [ j := l ]
					ifTrue: [
						l - i <  (j - k)
							ifTrue: [ 
								self quickSortFrom: i to: l by: sortBlock.
								i := k ]
							ifFalse: [
								self quickSortFrom: k to: j by: sortBlock.
								j := l ] ] ] ] repeat! !

!ArrayedCollection methodsFor: 'sorting' stamp: 'ul 5/27/2016 19:16'!
quickSort: sortBlock
	"Sort elements of self to be nondescending according to sortBlock using an in-place quicksort with simple median-of-three partitioning with guaranteed O(log(n)) space usage."

	self quickSortFrom: 1 to: self size by: sortBlock! !

!ArrayedCollection methodsFor: 'removing' stamp: 'klub 9/14/2009 16:27'!
removeAll

	self shouldNotImplement! !

!ArrayedCollection methodsFor: 'objects from disk' stamp: 'fbs 7/25/2013 07:07'!
restoreEndianness
	"This word object was just read in from a stream.  It was stored in Big Endian (Mac) format.  Reverse the byte order if the current machine is Little Endian.
	We only intend this for non-pointer arrays.  Do nothing if I contain pointers."

	self class isPointers | self class isWords not ifTrue: [^self].
	Smalltalk  isLittleEndian 
		ifTrue: 
			[Bitmap 
				swapBytesIn: self
				from: 1
				to: self basicSize]! !

!ArrayedCollection methodsFor: 'accessing' stamp: 'sma 5/12/2000 11:36'!
size
	"Answer how many elements the receiver contains."

	<primitive: 62>
	^ self basicSize! !

!ArrayedCollection methodsFor: 'sorting' stamp: 'ul 11/2/2009 04:56'!
sort
	"Sort this array into ascending order using the '<=' operator."

	self sort: nil! !

!ArrayedCollection methodsFor: 'sorting' stamp: 'ul 11/2/2009 05:18'!
sort: aSortBlock 
	"Sort this array using aSortBlock. The block should take two arguments
	and return true if the first element should preceed the second one.
	If aSortBlock is nil then <= is used for comparison."

	self
		mergeSortFrom: 1
		to: self size
		by: aSortBlock! !

!ArrayedCollection methodsFor: 'private' stamp: ''!
storeElementsFrom: firstIndex to: lastIndex on: aStream

	| noneYet defaultElement arrayElement |
	noneYet := true.
	defaultElement := self defaultElement.
	firstIndex to: lastIndex do: 
		[:index | 
		arrayElement := self at: index.
		arrayElement = defaultElement
			ifFalse: 
				[noneYet
					ifTrue: [noneYet := false]
					ifFalse: [aStream nextPut: $;].
				aStream nextPutAll: ' at: '.
				aStream store: index.
				aStream nextPutAll: ' put: '.
				aStream store: arrayElement]].
	^noneYet! !

!ArrayedCollection methodsFor: 'printing' stamp: ''!
storeOn: aStream

	aStream nextPutAll: '(('.
	aStream nextPutAll: self class name.
	aStream nextPutAll: ' new: '.
	aStream store: self size.
	aStream nextPut: $).
	(self storeElementsFrom: 1 to: self size on: aStream)
		ifFalse: [aStream nextPutAll: '; yourself'].
	aStream nextPut: $)! !

!ArrayedCollection methodsFor: 'objects from disk' stamp: 'tk 3/7/2001 17:36'!
swapHalves
		"A normal switch in endianness (byte order in words) reverses the order of 4 bytes.  That is not correct for SoundBuffers, which use 2-bytes units.  If a normal switch has be done, this method corrects it further by swapping the two halves of the long word.
	This method is only used for 16-bit quanities in SoundBuffer, ShortIntegerArray, etc."

	| hack blt |
	"The implementation is a hack, but fast for large ranges"
	hack := Form new hackBits: self.
	blt := (BitBlt toForm: hack) sourceForm: hack.
	blt combinationRule: Form reverse.  "XOR"
	blt sourceY: 0; destY: 0; height: self size; width: 2.
	blt sourceX: 0; destX: 2; copyBits.  "Exchange bytes 0&1 with 2&3"
	blt sourceX: 2; destX: 0; copyBits.
	blt sourceX: 0; destX: 2; copyBits.! !

!ArrayedCollection methodsFor: 'objects from disk' stamp: 'ul 3/15/2010 02:31'!
writeOnGZIPByteStream: aStream 
	"We only intend this for non-pointer arrays.  Do nothing if I contain pointers."

	(self class isPointers or: [ self class isWords not ]) ifTrue: [ ^self ].
	aStream nextPutAllWordArray: self! !

!ArrayedCollection methodsFor: 'objects from disk' stamp: 'ul 3/15/2010 02:30'!
writeOn: aStream 
	"Store the array of bits onto the argument, aStream.  (leading byte ~= 16r80) identifies this as raw bits (uncompressed).  Always store in Big Endian (Mac) byte order.  Do the writing at BitBlt speeds. We only intend this for non-pointer arrays.  Do nothing if I contain pointers."

	(self class isPointers or: [ self class isWords not ]) ifTrue: [ ^self ].
	aStream nextInt32Put: self basicSize.
	aStream nextWordsPutAll: self.! !

ArrayedCollection variableByteSubclass: #ByteArray
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Arrayed'!

!ByteArray commentStamp: '<historical>' prior: 0!
I represent an ArrayedCollection whose elements are integers between 0 and 255.
!

!ByteArray class methodsFor: 'byte based hash' stamp: 'dtl 10/24/2010 12:40'!
hashBytes: aByteArray startingWith: speciesHash
	"Answer the hash of a byte-indexed collection,
	using speciesHash as the initial value.
	See SmallInteger>>hashMultiply.

	The primitive should be renamed at a
	suitable point in the future"

	| byteArraySize hash low |
	<primitive: 'primitiveStringHash' module: 'MiscPrimitivePlugin'>
	<var: #aHash declareC: 'int speciesHash'>
	<var: #aByteArray declareC: 'unsigned char *aByteArray'>

	byteArraySize := aByteArray size.
	hash := speciesHash bitAnd: 16rFFFFFFF.
	1 to: byteArraySize do: [:pos |
		hash := hash + (aByteArray basicAt: pos).
		"Begin hashMultiply"
		low := hash bitAnd: 16383.
		hash := (16r260D * low + ((16r260D * (hash bitShift: -14) + (16r0065 * low) bitAnd: 16383) * 16384)) bitAnd: 16r0FFFFFFF.
	].
	^ hash! !

!ByteArray class methodsFor: 'instance creation' stamp: 'ar 2/23/2010 07:08'!
readHexFrom: aString
	"Create a byte array from a hexadecimal representation"
	^(self new: aString size // 2) readHexFrom: aString readStream! !

!ByteArray methodsFor: 'converting' stamp: 'sma 5/12/2000 17:35'!
asByteArray
	^ self! !

!ByteArray methodsFor: 'private' stamp: 'ar 1/28/2000 17:45'!
asByteArrayPointer
	"Return a ByteArray describing a pointer to the contents of the receiver."
	^self shouldNotImplement! !

!ByteArray methodsFor: 'converting' stamp: 'topa 7/9/2016 23:57'!
ascii85Encoded
	"Encode the receiver as Ascii85"
	"'Hello World' asByteArray ascii85Encoded"

	^ (Ascii85Converter encode: self readStream) contents! !

!ByteArray methodsFor: 'converting' stamp: ''!
asString
	"Convert to a String with Characters for each byte.
	Fast code uses primitive that avoids character conversion"

	^ (String new: self size) replaceFrom: 1 to: self size with: self! !

!ByteArray methodsFor: 'accessing' stamp: 'ar 4/12/2005 17:35'!
asWideString

	^ WideString fromByteArray: self.
! !

!ByteArray methodsFor: 'accessing' stamp: 'sma 4/22/2000 17:47'!
atAllPut: value
	"Fill the receiver with the given value"

	<primitive: 145>
	super atAllPut: value! !

!ByteArray methodsFor: 'accessing' stamp: 'tfel 4/9/2015 09:08'!
at: index put: value 
	<primitive: 61> "try primitiveAtPut, convert value to integer if that fails and try again" 
	^ self byteAt: index put: value asInteger
! !

!ByteArray methodsFor: 'converting' stamp: 'jcg 9/1/2009 00:30'!
base64Encoded
	"Encode the receiver as base64"
	"'Hello World' asByteArray base64Encoded"
	^(Base64MimeConverter mimeEncode: self readStream) contents! !

!ByteArray methodsFor: 'accessing' stamp: 'tfel 4/9/2015 08:54'!
byteAt: index
	<primitive: 60>
	^ super at: index! !

!ByteArray methodsFor: 'accessing' stamp: 'tfel 4/9/2015 08:54'!
byteAt: index put: value
	<primitive: 61>
	^ super at: index put: value! !

!ByteArray methodsFor: 'accessing' stamp: 'ar 3/3/2001 16:17'!
byteSize
	^self size! !

!ByteArray methodsFor: 'accessing' stamp: 'tk 3/13/2000 14:46'!
bytesPerElement
	"Number of bytes in each item.  This multiplied by (self size)*8 gives the number of bits stored."
	^ 1! !

!ByteArray methodsFor: 'private' stamp: ''!
defaultElement

	^0! !

!ByteArray methodsFor: 'platform independent access' stamp: 'jmb 12/3/2004 14:54'!
doubleAt: index bigEndian: bool 
	"Return a 64 bit float starting from the given byte index"
	| w1 w2 dbl |
	w1 := self unsignedLongAt: index bigEndian: bool.
	w2 := self unsignedLongAt: index + 4 bigEndian: bool.
	dbl := Float new: 2. 
	bool
		ifTrue: [dbl basicAt: 1 put: w1.
			dbl basicAt: 2 put: w2]
		ifFalse: [dbl basicAt: 1 put: w2.
			dbl basicAt: 2 put: w1].
	^ dbl! !

!ByteArray methodsFor: 'platform independent access' stamp: 'jmb 12/3/2004 14:54'!
doubleAt: index put: value bigEndian: bool 
	"Store a 64 bit float starting from the given byte index"
	| w1 w2 |
	bool
		ifTrue: [w1 := value basicAt: 1.
			w2 := value basicAt: 2]
		ifFalse: [w1 := value basicAt: 2.
			w2 := value basicAt: 1]. 
	self unsignedLongAt: index put: w1 bigEndian: bool.
	self unsignedLongAt: index + 4 put: w2 bigEndian: bool.
	^ value! !

!ByteArray methodsFor: 'comparing' stamp: 'SqR 8/13/2002 10:52'!
hash
	"#hash is implemented, because #= is implemented"

	^self class
		hashBytes: self
		startingWith: self species hash! !

!ByteArray methodsFor: 'converting' stamp: 'ar 2/20/2010 12:44'!
hex
	"Answer a hexa decimal representation of the receiver"
	| string v index map |
	map := '0123456789abcdef'.
	string := String new: self size * 2. "hex"
	index := 0.
	1 to: self size do:[:i| 
		v := self at: i.
		string at: (index := index + 1) put: (map at: (v bitShift: -4) + 1). 
		string at: (index := index + 1) put: (map at: (v bitAnd: 15) + 1).
	].
	^string! !

!ByteArray methodsFor: 'accessing' stamp: 'ul 8/19/2015 22:31'!
indexOf: anInteger

	^self indexOf: anInteger startingAt: 1! !

!ByteArray methodsFor: 'accessing' stamp: 'ul 8/19/2015 22:29'!
indexOf: anInteger startingAt: start

	anInteger isInteger ifFalse: [ ^0 ].
	0 <= anInteger ifFalse: [ ^0 ].
	anInteger <= 255 ifFalse: [ ^0 ].
	^ByteString indexOfAscii: anInteger inString: self startingAt: start! !

!ByteArray methodsFor: 'accessing' stamp: 'ul 8/19/2015 22:32'!
indexOf: anInteger startingAt: start ifAbsent: aBlock

	| index |
	(index := self indexOf: anInteger startingAt: start) = 0 ifTrue: [ ^aBlock value ].
	^index! !

!ByteArray methodsFor: 'testing' stamp: 'eem 6/4/2010 17:55'!
isLiteral
	"so that #(1 #[1 2 3] 5) prints itself"
	^self class == ByteArray! !

!ByteArray methodsFor: 'platform independent access' stamp: 'ul 9/27/2015 23:01'!
long64At: index bigEndian: bigEndian
	"Return a 64-bit signed integer quantity starting from the given byte index."

	| value |
	value := self unsignedLong64At: index bigEndian: bigEndian.
	value digitLength < 8 ifTrue: [ ^value ].
	(value digitAt: 8) < 16r80 ifTrue: [ ^value ].
	^value - 16r10000000000000000! !

!ByteArray methodsFor: 'platform independent access' stamp: 'ul 9/27/2015 22:57'!
long64At: index put: value bigEndian: bigEndian
	"Store a 64-bit signed integer quantity starting from the given byte index."
	
	^self
		unsignedLong64At: index
		put: (value negative
			ifFalse: [ value ]
			ifTrue: [ value + 16r10000000000000000 ])
		bigEndian: bigEndian! !

!ByteArray methodsFor: 'platform independent access' stamp: 'ul 9/9/2015 20:28'!
longAt: index bigEndian: bigEndian
	"Return a 32-bit integer quantity starting from the given byte index. Use #normalize where necessary to ensure compatibility with non-30-bit SmallIntegers."
	
	| byte result |
	bigEndian ifFalse: [
		(byte := self at: index + 3) <= 16r7F ifTrue: [ "Is the result non-negative?"
			byte <= 16r3F ifTrue: [
				^(((byte bitShift: 8) + (self at: index + 2) bitShift: 8) + (self at: index + 1) bitShift: 8) + (self at: index) ].
			^(LargePositiveInteger new: 4)
				replaceFrom: 1
					to: 4
					with: self
					startingAt: index;
				normalize ].
		"Negative"
		byte >= 16rC0 ifTrue: [
			^-1 - (((((byte bitShift: 8) + (self at: index + 2) bitShift: 8) + (self at: index + 1) bitXor: 16rFFFFFF) bitShift: 8) + ((self at: index) bitXor: 16rFF)) ].
		(result := LargeNegativeInteger new: 4)
			digitAt: 4 put: ((self at: index + 3) bitXor: 16rFF);
			digitAt: 3 put: ((self at: index + 2) bitXor: 16rFF);
			digitAt: 2 put: ((self at: index + 1) bitXor: 16rFF).
		(byte := ((self at: index) bitXor: 16rFF) + 1) <= 16rFF ifTrue: [
			^result
				digitAt: 1 put: byte;
				normalize ].
		^result
			digitAt: 1 put: 16rFF;
			- 1 "It's tempting to do the subtraction in a loop to avoid the LargeInteger creation, but it's actually slower than this." ].
	(byte := self at: index) <= 16r7F ifTrue: [ "Is the result non-negative?"
		byte <= 16r3F ifTrue: [
			^(((byte bitShift: 8) + (self at: index + 1) bitShift: 8) + (self at: index + 2) bitShift: 8) + (self at: index + 3) ].
		^(LargePositiveInteger new: 4)
			digitAt: 1 put: (self at: index + 3);
			digitAt: 2 put: (self at: index + 2);
			digitAt: 3 put: (self at: index + 1);
			digitAt: 4 put: byte;
			normalize ].
	"Negative"
	16rC0 <= byte ifTrue: [
		^-1 - (((((byte bitShift: 8) + (self at: index + 1) bitShift: 8) + (self at: index + 2) bitXor: 16rFFFFFF) bitShift: 8) + ((self at: index + 3) bitXor: 16rFF)) ].
	(result := LargeNegativeInteger new: 4)
		digitAt: 4 put: (byte bitXor: 16rFF);
		digitAt: 3 put: ((self at: index + 1) bitXor: 16rFF);
		digitAt: 2 put: ((self at: index + 2) bitXor: 16rFF).
	(byte := ((self at: index + 3) bitXor: 16rFF) + 1) <= 16rFF ifTrue: [
		^result
			digitAt: 1 put: byte;
			normalize ].
	^result 
		digitAt: 1 put: 16rFF;
		- 1 "It's tempting to do the subtraction in a loop to avoid the LargeInteger creation, but it's actually slower than this."! !

!ByteArray methodsFor: 'platform independent access' stamp: 'ul 9/6/2015 17:13'!
longAt: index put: value bigEndian: bigEndian
	"Store a 32-bit signed integer quantity starting from the given byte index"
	
	| v v2 |
	value isLarge ifTrue: [
		bigEndian ifFalse: [
			value positive ifTrue: [
				self 
					replaceFrom: index
					to: index + 3
					with: value
					startingAt: 1.
				^value ].
			v := 0.
			[ v <= 3 and: [ (v2 := ((value digitAt: v + 1) bitXor: 16rFF) + 1) = 16r100 ] ] whileTrue: [
				self at: index + v put: 0.
				v := v + 1 ].
			self at: index + v put: v2.
			v := v + 1.
			[ v <= 3 ] whileTrue: [
				self at: index + v put: ((value digitAt: (v := v + 1)) bitXor: 16rFF) ].
			^value ].
		value positive ifTrue: [
			self
				at: index put: (value digitAt: 4);
				at: index + 1 put: (value digitAt: 3);
				at: index + 2 put: (value digitAt: 2);
				at: index + 3 put: (value digitAt: 1).
			^value ].
		v := 3.
		[ 0 <= v and: [ (v2 := ((value digitAt: 4 - v) bitXor: 16rFF) + 1) = 16r100 ] ] whileTrue: [
			self at: index + v put: 0.
			v := v - 1 ].
		self at: index + v put: v2.
		[ 0 <= (v := v - 1) ] whileTrue: [
			self at: index + v put: ((value digitAt: 4 - v) bitXor: 16rFF) ].
		^value ].
	v := value bitShift: -24.
	0 <= (v := (v bitAnd: 16r7F) - (v bitAnd: 16r80)) ifFalse: [
		v := v + 16r100 ].
	bigEndian ifFalse: [
		self 
			at: index put: (value bitAnd: 16rFF);
			at: index + 1 put: ((value bitShift: -8) bitAnd: 16rFF);
			at: index + 2 put: ((value bitShift: -16) bitAnd: 16rFF);
			at: index + 3 put: v.
		^value ].
	self
		at: index put: v;
		at: index + 1 put: ((value bitShift: -16) bitAnd: 16rFF);
		at: index + 2 put: ((value bitShift: -8) bitAnd: 16rFF);
		at: index + 3 put: (value bitAnd: 16rFF).
	^value! !

!ByteArray methodsFor: 'enumerating' stamp: 'ul 3/8/2016 17:23'!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject. Optimized version."

	| tally |
	anObject isInteger ifFalse: [ ^0 ].
	anObject negative ifTrue: [ ^0 ].
	anObject > 255 ifTrue: [ ^0 ].
	tally := 0.
	1 to: self size do: [ :index |
		(self at: index) = anObject ifTrue: [ tally := tally + 1 ] ].
	^tally! !

!ByteArray methodsFor: 'printing' stamp: 'eem 10/31/2015 16:42:41'!
printAsLiteralByteArrayOn: aStream

	aStream nextPutAll: '#['.
	self
		do: [ :each | each printOn: aStream ]
		separatedBy: [ aStream nextPut: $ ].
	aStream nextPut: $]! !

!ByteArray methodsFor: 'printing' stamp: 'eem 10/31/2015 16:45:10'!
printOn: aStream

	self shouldBePrintedAsLiteral ifFalse:
		[super printOn: aStream.
		 aStream space].
	self printAsLiteralByteArrayOn: aStream! !

!ByteArray methodsFor: 'streaming' stamp: 'kph 9/27/2007 21:53'!
putOn: aStream

	aStream nextPutAll: self
! !

!ByteArray methodsFor: 'initialize' stamp: 'nice 2/26/2010 00:04'!
readHexFrom: aStream
	"Initialize the receiver from a hexadecimal string representation"
	| map v ch value |
	map := '0123456789abcdefABCDEF'.
	1 to: self size do:[:i|
		ch := aStream next.
		v := (map indexOf: ch) - 1.
		((v between: 0 and: 15) or: [((v:= v - 6) between: 0 and: 15)]) ifFalse:[^self error: 'Hex digit expected'].
		value := v bitShift: 4.
		ch := aStream next.
		v := (map indexOf: ch) - 1.
		((v between: 0 and: 15) or: [((v:= v - 6) between: 0 and: 15)]) ifFalse:[^self error: 'Hex digit expected'].
		value := value + v.
		self at: i put: value.
	].
! !

!ByteArray methodsFor: 'private' stamp: ''!
replaceFrom: start to: stop with: replacement startingAt: repStart 
	"Primitive. This destructively replaces elements from start to stop in the receiver starting at index, repStart, in the collection, replacement. Answer the receiver. Range checks are performed in the primitive only. Optional. See Object documentation whatIsAPrimitive."
	<primitive: 105>
	super replaceFrom: start to: stop with: replacement startingAt: repStart! !

!ByteArray methodsFor: 'platform independent access' stamp: 'ul 9/6/2015 23:16'!
shortAt: index bigEndian: bigEndian
	"Return a 16-bit signed integer quantity starting from the given byte index"

	| result |
	result := bigEndian
		ifFalse: [ ((self at: index + 1) bitShift: 8) + (self at: index) ]
		ifTrue: [ ((self at: index) bitShift: 8) + (self at: index + 1) ].
	result < 16r8000 ifTrue: [ ^result ].
	^result - 16r10000! !

!ByteArray methodsFor: 'platform independent access' stamp: 'ul 9/6/2015 23:16'!
shortAt: index put: value bigEndian: bigEndian
	"Store a 16-bit signed integer quantity starting from the given byte index"
	
	| unsignedValue |
	(unsignedValue := value) < 0 ifTrue: [
		unsignedValue := unsignedValue + 16r10000 ].
	bigEndian ifFalse: [
		self 
			at: index + 1 put: (unsignedValue bitShift: -8);
			at: index put: (unsignedValue bitAnd: 16rFF).
		^value ].
	self
		at: index put: (unsignedValue bitShift: -8);
		at: index + 1 put: (unsignedValue bitAnd: 16rFF).
	^value! !

!ByteArray methodsFor: 'platform independent access' stamp: 'ul 9/27/2015 22:49'!
signedByteAt: index
	"Return an 8-bit signed integer quantity from the given byte index."
	
	| byte |
	(byte := self at: index) <= 16r7F ifTrue: [ ^byte ].
	^byte - 16r100! !

!ByteArray methodsFor: 'platform independent access' stamp: 'ul 9/27/2015 22:55'!
signedByteAt: index put: anInteger
	"Store an 8-bit signed integer quantity at the given byte index."
	
	anInteger >= 0 ifTrue: [ ^self at: index put: anInteger ].
	self at: index put: anInteger + 16r100.
	^anInteger! !

!ByteArray methodsFor: 'printing' stamp: 'eem 10/31/2015 16:42:54'!
storeAsLiteralByteArrayOn: aStream
	aStream nextPutAll: '#['.
	self
		do: [ :each | each storeOn: aStream ]
		separatedBy: [ aStream nextPut: $ ].
	aStream nextPut: $]! !

!ByteArray methodsFor: 'printing' stamp: 'eem 10/31/2015 16:45:51'!
storeOn: aStream
	self shouldBePrintedAsLiteral
		ifTrue: [self storeAsLiteralByteArrayOn: aStream]
		ifFalse: [super storeOn: aStream]! !

!ByteArray methodsFor: 'platform independent access' stamp: 'eem 2/22/2016 09:36'!
unsignedLong64At: index bigEndian: bigEndian
	"Return a 64-bit unsigned integer quantity starting from the given byte index. Use #normalize where necessary to ensure compatibility with non-30-bit SmallIntegers."

	| byte |
	SmallInteger maxVal >  1073741823 ifTrue:
		[bigEndian
			ifTrue: "64-bit SmallIntegers have a 3 bit tag and a sign bit, so the most positive value has 16rF as its top byte."
				[(byte := self at: index) <= 16rF ifTrue:
					[^((((((((byte bitShift: 8) + (self at: index + 1) bitShift: 8) + (self at: index + 2) bitShift: 8) + (self at: index + 3)) bitShift: 8)
						+ (self at: index + 4) bitShift: 8) + (self at: index + 5) bitShift: 8) + (self at: index + 6) bitShift: 8) + (self at: index + 7)]]
			ifFalse:
				[(byte := self at: index + 7) <= 16rF ifTrue:
					[^((((((((byte bitShift: 8) + (self at: index + 6) bitShift: 8) + (self at: index + 5) bitShift: 8) + (self at: index + 4)) bitShift: 8)
						+ (self at: index + 3) bitShift: 8) + (self at: index + 2) bitShift: 8) + (self at: index + 1) bitShift: 8) + (self at: index)]]].
	bigEndian ifFalse: [
		(byte := self at: index + 7) = 0 ifFalse: [
			^(LargePositiveInteger new: 8)
				replaceFrom: 1 to: 8 with: self startingAt: index;
				normalize ].
		(byte := self at: index + 6) = 0 ifFalse: [
			^(LargePositiveInteger new: 7)
				replaceFrom: 1 to: 7 with: self startingAt: index;
				normalize ].
		(byte := self at: index + 5) = 0 ifFalse: [
			^(LargePositiveInteger new: 6)
				replaceFrom: 1 to: 6 with: self startingAt: index;
				normalize ].
		(byte := self at: index + 4) = 0 ifFalse: [
			^(LargePositiveInteger new: 5)
				replaceFrom: 1 to: 5 with: self startingAt: index;
				normalize ].
		(byte := self at: index + 3) <= 16r3F ifFalse: [
			^(LargePositiveInteger new: 4)
				replaceFrom: 1 to: 4 with: self startingAt: index;
				normalize ].
		^(((byte bitShift: 8) + (self at: index + 2) bitShift: 8) + (self at: index + 1) bitShift: 8) + (self at: index) ].
	(byte := self at: index) = 0 ifFalse: [
		^(LargePositiveInteger new: 8)
			digitAt: 1 put: (self at: index + 7);
			digitAt: 2 put: (self at: index + 6);
			digitAt: 3 put: (self at: index + 5);
			digitAt: 4 put: (self at: index + 4);
			digitAt: 5 put: (self at: index + 3);
			digitAt: 6 put: (self at: index + 2);
			digitAt: 7 put: (self at: index + 1);
			digitAt: 8 put: byte;
			normalize ].
	(byte := self at: index + 1) = 0 ifFalse: [	
		^(LargePositiveInteger new: 7)
			digitAt: 1 put: (self at: index + 7);
			digitAt: 2 put: (self at: index + 6);
			digitAt: 3 put: (self at: index + 5);
			digitAt: 4 put: (self at: index + 4);
			digitAt: 5 put: (self at: index + 3);
			digitAt: 6 put: (self at: index + 2);
			digitAt: 7 put: byte;
			normalize ].
	(byte := self at: index + 2) = 0 ifFalse: [	
		^(LargePositiveInteger new: 6)
			digitAt: 1 put: (self at: index + 7);
			digitAt: 2 put: (self at: index + 6);
			digitAt: 3 put: (self at: index + 5);
			digitAt: 4 put: (self at: index + 4);
			digitAt: 5 put: (self at: index + 3);
			digitAt: 6 put: byte;
			normalize ].
	(byte := self at: index + 3) = 0 ifFalse: [	
		^(LargePositiveInteger new: 5)
			digitAt: 1 put: (self at: index + 7);
			digitAt: 2 put: (self at: index + 6);
			digitAt: 3 put: (self at: index + 5);
			digitAt: 4 put: (self at: index + 4);
			digitAt: 5 put: byte;
			normalize ].		
	(byte := self at: index + 4) <= 16r3F ifFalse: [
		^(LargePositiveInteger new: 4)
			digitAt: 1 put: (self at: index + 7);
			digitAt: 2 put: (self at: index + 6);
			digitAt: 3 put: (self at: index + 5);
			digitAt: 4 put: byte;
			normalize ].
	^(((byte bitShift: 8) + (self at: index + 5) bitShift: 8) + (self at: index + 6) bitShift: 8) + (self at: index + 7)! !

!ByteArray methodsFor: 'platform independent access' stamp: 'ul 9/9/2015 20:18'!
unsignedLong64At: index put: value bigEndian: bigEndian
	"Store a 64-bit unsigned integer quantity starting from the given byte index"
	
	| i j |
	value isLarge ifTrue: [
		i := value digitLength.
		bigEndian ifFalse: [
			self
				replaceFrom: index
					to: index + i - 1
					with: value 
					startingAt: 1;
				replaceFrom: index + i
					to: index + 7
					with: #[0 0 0 0 0 0 0 0]
					startingAt: 1.
			^value ].
		j := index + 8.
		i <= 7 ifTrue: [
			self
				replaceFrom: index
				to: j - i - 1
				with: #[0 0 0 0 0 0 0 0]
				startingAt: 1 ].
		[ 1 <= i ] whileTrue: [
			self at: j - i put: (value digitAt: i).
			i := i - 1 ].
		^value ].
	bigEndian ifFalse: [
		j := index - 1.
		i := value.
		[ 1 <= i ] whileTrue: [
			self at: (j := j + 1) put: (i bitAnd: 16rFF).
			i := i bitShift: -8 ].
		self replaceFrom: j + 1
			to: index + 7
			with: #[0 0 0 0 0 0 0 0]
			startingAt: 1.
		^value ].
	j := index + 8.
	i := value.
	[ 1 <= i ] whileTrue: [
		self at: (j := j - 1) put: (i bitAnd: 16rFF).
		i := i bitShift: -8 ].
	self replaceFrom: index
		to: j - 1
		with: #[0 0 0 0 0 0 0 0]
		startingAt: 1.
	^value! !

!ByteArray methodsFor: 'platform independent access' stamp: 'eem 2/22/2016 09:20'!
unsignedLongAt: index bigEndian: bigEndian
	"Return a 32-bit unsigned integer quantity starting from the given byte index. Use #normalize where necessary to ensure compatibility with non-30-bit SmallIntegers."
	| byte |
	bigEndian ifTrue:
		[((byte := self at: index) <= 16r3F
		 or: [SmallInteger maxVal >  1073741823]) ifTrue:
			[^(((byte bitShift: 8) + (self at: index + 1) bitShift: 8) + (self at: index + 2) bitShift: 8) + (self at: index + 3)].
		^(LargePositiveInteger new: 4)
			digitAt: 1 put: (self at: index + 3);
			digitAt: 2 put: (self at: index + 2);
			digitAt: 3 put: (self at: index + 1);
			digitAt: 4 put: byte;
			normalize].
	((byte := self at: index + 3) <= 16r3F
	 or: [SmallInteger maxVal >  1073741823]) ifTrue:
		[^(((byte bitShift: 8) + (self at: index + 2) bitShift: 8) + (self at: index + 1) bitShift: 8) + (self at: index)].
	^(LargePositiveInteger new: 4)
		replaceFrom: 1 to: 4 with: self startingAt: index;
		normalize! !

!ByteArray methodsFor: 'platform independent access' stamp: 'ul 9/3/2015 01:33'!
unsignedLongAt: index put: value bigEndian: bigEndian
	"Store a 32-bit unsigned integer quantity starting from the given byte index"
	
	value isLarge
		ifTrue: [
			bigEndian ifFalse: [
				self
					replaceFrom: index
					to: index + 3
					with: value
					startingAt: 1.
				^value ].
			self
				at: index put: (value digitAt: 4);
				at: index + 1 put: (value digitAt: 3);
				at: index + 2 put: (value digitAt: 2);
				at: index +3 put: (value digitAt: 1) ]
		ifFalse: [
			bigEndian ifFalse: [
				self 
					at: index put: (value bitAnd: 16rFF);
					at: index + 1 put: ((value bitShift: -8) bitAnd: 16rFF);
					at: index + 2 put: ((value bitShift: -16) bitAnd: 16rFF);
					at: index + 3 put: (value bitShift: -24).
				^value ].
			self 
				at: index put: (value bitShift: -24);
				at: index + 1 put: ((value bitShift: -16) bitAnd: 16rFF);
				at: index + 2 put: ((value bitShift: -8) bitAnd: 16rFF);
				at: index + 3 put: (value bitAnd: 16rFF) ].
	^value! !

!ByteArray methodsFor: 'platform independent access' stamp: 'ul 9/3/2015 01:33'!
unsignedShortAt: index bigEndian: bigEndian
	"Return a 16-bit unsigned integer quantity starting from the given byte index"

	bigEndian ifFalse: [ ^((self at: index + 1) bitShift: 8) + (self at: index) ].
	^((self at: index) bitShift: 8) + (self at: index + 1)
	! !

!ByteArray methodsFor: 'platform independent access' stamp: 'ul 9/3/2015 15:29'!
unsignedShortAt: index put: value bigEndian: bigEndian
	"Store a 16-bit unsigned integer quantity starting from the given byte index"
	
	bigEndian ifFalse: [
		self 
			at: index + 1 put: (value bitShift: -8);
			at: index put: (value bitAnd: 16rFF).
		^value ].
	self
		at: index put: (value bitShift: -8);
		at: index+1 put: (value bitAnd: 16rFF).
	^value! !

ArrayedCollection variableWordSubclass: #ColorArray
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Arrayed'!

!ColorArray methodsFor: 'converting' stamp: 'ar 3/3/2001 20:06'!
asColorArray
	^self! !

!ColorArray methodsFor: 'accessing' stamp: 'ar 3/3/2001 20:03'!
at: index
	^(super at: index) asColorOfDepth: 32! !

!ColorArray methodsFor: 'accessing' stamp: 'ar 3/3/2001 20:04'!
at: index put: aColor
	^super at: index put: (aColor pixelWordForDepth: 32).! !

!ColorArray methodsFor: 'converting' stamp: 'RAA 3/8/2001 06:24'!
bytesPerElement

	^4! !

ArrayedCollection variableWordSubclass: #FloatArray
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Arrayed'!

!FloatArray commentStamp: '<historical>' prior: 0!
FloatArrays store 32bit IEEE floating point numbers.!

!FloatArray methodsFor: 'arithmetic' stamp: 'nice 11/24/2007 00:10'!
adaptToNumber: rcvr andSend: selector
	"If I am involved in arithmetic with a Number. If possible,
	convert it to a float and perform the (more efficient) primitive operation."
	selector == #+ ifTrue:[^self + rcvr].
	selector == #* ifTrue:[^self * rcvr].
	selector == #- ifTrue:[^self negated += rcvr].
	selector == #/ ifTrue:[
		"DO NOT USE TRIVIAL CODE
			^self reciprocal * rcvr
		BECAUSE OF GRADUAL UNDERFLOW
		self should: (1.0e-39 / (FloatArray with: 1.0e-39)) first < 2."
			^(self class new: self size withAll: rcvr) / self
		].
	^super adaptToNumber: rcvr andSend: selector! !

!FloatArray methodsFor: 'converting' stamp: 'ar 9/14/1998 23:46'!
asFloatArray
	^self! !

!FloatArray methodsFor: 'accessing' stamp: 'ar 2/2/2001 15:47'!
at: index
	<primitive: 'primitiveAt' module: 'FloatArrayPlugin'>
	^Float fromIEEE32Bit: (self basicAt: index)! !

!FloatArray methodsFor: 'accessing' stamp: 'ar 2/2/2001 15:47'!
at: index put: value
	<primitive: 'primitiveAtPut' module: 'FloatArrayPlugin'>
	value isFloat 
		ifTrue:[self basicAt: index put: value asIEEE32BitWord]
		ifFalse:[self at: index put: value asFloat].
	^value! !

!FloatArray methodsFor: 'accessing' stamp: 'ar 11/2/1998 12:19'!
defaultElement
	"Return the default element of the receiver"
	^0.0! !

!FloatArray methodsFor: 'arithmetic' stamp: 'ul 2/2/2011 23:42'!
dot: aFloatVector
	"Primitive. Return the dot product of the receiver and the argument.
	Fail if the argument is not of the same size as the receiver."

	| result |
	<primitive: 'primitiveDotProduct' module: 'FloatArrayPlugin'>
	self size = aFloatVector size ifFalse:[^self error:'Must be equal size'].
	result := 0.0.
	1 to: self size do:[:i|
		result := result + ((self at: i) * (aFloatVector at: i)).
	].
	^result! !

!FloatArray methodsFor: 'comparing' stamp: 'ar 5/3/2001 13:02'!
hash
	| result |
	<primitive:'primitiveHashArray' module: 'FloatArrayPlugin'>
	result := 0.
	1 to: self size do:[:i| result := result + (self basicAt: i) ].
	^result bitAnd: 16r1FFFFFFF! !

!FloatArray methodsFor: 'accessing' stamp: 'laza 3/24/2000 13:08'!
length
	"Return the length of the receiver"
	^self squaredLength sqrt! !

!FloatArray methodsFor: 'arithmetic' stamp: 'ar 9/7/2001 23:04'!
negated
	^self clone *= -1! !

!FloatArray methodsFor: 'arithmetic' stamp: 'ar 12/14/2010 08:20'!
normalize
	"Unsafely normalize the receiver in-place (become a unit vector).
 	 Div-by-Zero raised if len 0."
	<primitive: 'primitiveNormalize' module: 'FloatArrayPlugin'>
	self /= self length.! !

!FloatArray methodsFor: 'primitives-plugin' stamp: 'ar 2/2/2001 15:47'!
primAddArray: floatArray

	<primitive: 'primitiveAddFloatArray' module: 'FloatArrayPlugin'>
	1 to: self size do:[:i| self at: i put: (self at: i) + (floatArray at: i)].! !

!FloatArray methodsFor: 'primitives-plugin' stamp: 'ar 2/2/2001 15:47'!
primAddScalar: scalarValue

	<primitive: 'primitiveAddScalar' module: 'FloatArrayPlugin'>
	1 to: self size do:[:i| self at: i put: (self at: i) + scalarValue].! !

!FloatArray methodsFor: 'primitives-plugin' stamp: 'ar 2/2/2001 15:47'!
primDivArray: floatArray

	<primitive: 'primitiveDivFloatArray' module: 'FloatArrayPlugin'>
	1 to: self size do:[:i| self at: i put: (self at: i) / (floatArray at: i)].! !

!FloatArray methodsFor: 'primitives-plugin' stamp: 'ar 2/2/2001 15:47'!
primDivScalar: scalarValue

	<primitive: 'primitiveDivScalar' module: 'FloatArrayPlugin'>
	1 to: self size do:[:i| self at: i put: (self at: i) / scalarValue].! !

!FloatArray methodsFor: 'primitives-plugin' stamp: 'ar 2/2/2001 15:47'!
primMulArray: floatArray

	<primitive: 'primitiveMulFloatArray' module: 'FloatArrayPlugin'>
	1 to: self size do:[:i| self at: i put: (self at: i) * (floatArray at: i)].! !

!FloatArray methodsFor: 'primitives-plugin' stamp: 'ar 2/2/2001 15:47'!
primMulScalar: scalarValue

	<primitive: 'primitiveMulScalar' module: 'FloatArrayPlugin'>
	1 to: self size do:[:i| self at: i put: (self at: i) * scalarValue].! !

!FloatArray methodsFor: 'primitives-plugin' stamp: 'ar 2/2/2001 15:47'!
primSubArray: floatArray

	<primitive: 'primitiveSubFloatArray' module: 'FloatArrayPlugin'>
	1 to: self size do:[:i| self at: i put: (self at: i) - (floatArray at: i)].! !

!FloatArray methodsFor: 'primitives-plugin' stamp: 'ar 2/2/2001 15:47'!
primSubScalar: scalarValue

	<primitive: 'primitiveSubScalar' module: 'FloatArrayPlugin'>
	1 to: self size do:[:i| self at: i put: (self at: i) - scalarValue].! !

!FloatArray methodsFor: 'private' stamp: 'ar 10/9/1998 11:27'!
replaceFrom: start to: stop with: replacement startingAt: repStart 
	"Primitive. This destructively replaces elements from start to stop in the receiver starting at index, repStart, in the collection, replacement. Answer the receiver. Range checks are performed in the primitive only. Optional. See Object documentation whatIsAPrimitive."
	<primitive: 105>
	super replaceFrom: start to: stop with: replacement startingAt: repStart! !

!FloatArray methodsFor: 'accessing' stamp: 'laza 3/24/2000 13:08'!
squaredLength
	"Return the squared length of the receiver"
	^self dot: self! !

!FloatArray methodsFor: 'primitives-plugin' stamp: 'jcg 6/12/2003 17:54'!
sum

	<primitive: 'primitiveSum' module: 'FloatArrayPlugin'>
	^ super sum! !

!FloatArray methodsFor: 'arithmetic' stamp: 'ar 9/14/1998 22:33'!
* anObject
	^self clone *= anObject! !

!FloatArray methodsFor: 'arithmetic' stamp: 'ar 9/14/1998 22:49'!
*= anObject
	^anObject isNumber
		ifTrue:[self primMulScalar: anObject asFloat]
		ifFalse:[self primMulArray: anObject]! !

!FloatArray methodsFor: 'arithmetic' stamp: 'ar 9/14/1998 22:33'!
+ anObject
	^self clone += anObject! !

!FloatArray methodsFor: 'arithmetic' stamp: 'ar 9/14/1998 22:48'!
+= anObject
	^anObject isNumber
		ifTrue:[self primAddScalar: anObject asFloat]
		ifFalse:[self primAddArray: anObject]! !

!FloatArray methodsFor: 'arithmetic' stamp: 'ar 9/14/1998 22:33'!
- anObject
	^self clone -= anObject! !

!FloatArray methodsFor: 'arithmetic' stamp: 'ar 9/14/1998 22:49'!
-= anObject
	^anObject isNumber
		ifTrue:[self primSubScalar: anObject asFloat]
		ifFalse:[self primSubArray: anObject]! !

!FloatArray methodsFor: 'arithmetic' stamp: 'ar 9/14/1998 22:34'!
/ anObject
	^self clone /= anObject! !

!FloatArray methodsFor: 'arithmetic' stamp: 'ar 10/7/1998 19:58'!
/= anObject
	^anObject isNumber
		ifTrue:[self primDivScalar: anObject asFloat]
		ifFalse:[self primDivArray: anObject]! !

!FloatArray methodsFor: 'comparing' stamp: 'ar 2/2/2001 15:47'!
= aFloatArray 
	| length |
	<primitive: 'primitiveEqual' module: 'FloatArrayPlugin'>
	aFloatArray class = self class ifFalse: [^ false].
	length := self size.
	length = aFloatArray size ifFalse: [^ false].
	1 to: self size do: [:i | (self at: i)
			= (aFloatArray at: i) ifFalse: [^ false]].
	^ true! !

!FloatArray methodsFor: 'arithmetic' stamp: 'yo 9/14/2004 17:12'!
\\= other

	other isNumber ifTrue: [
		1 to: self size do: [:i |
			self at: i put: (self at: i) \\ other
		].
		^ self.
	].
	1 to: (self size min: other size) do: [:i |
		self at: i put: (self at: i) \\ (other at: i).
	].

! !

ArrayedCollection variableWordSubclass: #IntegerArray
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Arrayed'!

!IntegerArray commentStamp: '<historical>' prior: 0!
IntegerArrays store 32bit signed Integer values.
Negative values are stored as 2's complement.!

!IntegerArray methodsFor: 'converting' stamp: 'ar 10/10/1998 16:18'!
asIntegerArray
	^self! !

!IntegerArray methodsFor: 'accessing' stamp: 'dtl 3/8/2009 15:09'!
atAllPut: anInteger
	| word |
	anInteger < 0
		ifTrue:[anInteger < -16r80000000 ifTrue: [self error: anInteger asString , ' out of range'].
				"word := 16r100000000 + anInteger"
				word := (anInteger + 1) negated bitInvert32]
		ifFalse:[anInteger > 16r7FFFFFFF ifTrue: [self error: anInteger asString , ' out of range'].
				word := anInteger].
	self primFill: word.! !

!IntegerArray methodsFor: 'accessing' stamp: 'ar 3/3/2001 22:40'!
at: index
	| word |
	<primitive: 165>
	word := self basicAt: index.
	word < 16r3FFFFFFF ifTrue:[^word]. "Avoid LargeInteger computations"
	^word >= 16r80000000	"Negative?!!"
		ifTrue:["word - 16r100000000"
				(word bitInvert32 + 1) negated]
		ifFalse:[word]! !

!IntegerArray methodsFor: 'accessing' stamp: 'dtl 3/8/2009 11:52'!
at: index put: anInteger
	| word |
	<primitive: 166>
	anInteger < 0
		ifTrue:[anInteger < -16r80000000 ifTrue: [self error: anInteger asString , ' out of range'].
				"word := 16r100000000 + anInteger"
				word := (anInteger + 1) negated bitInvert32]
		ifFalse:[anInteger > 16r7FFFFFFF ifTrue: [self error: anInteger asString , ' out of range'].
				word := anInteger].
	self  basicAt: index put: word.
	^anInteger! !

!IntegerArray methodsFor: 'accessing' stamp: 'ar 11/2/1998 12:19'!
defaultElement
	"Return the default element of the receiver"
	^0! !

!IntegerArray methodsFor: 'private' stamp: 'ar 3/3/2001 23:34'!
primFill: aPositiveInteger
	"Fill the receiver, an indexable bytes or words object, with the given positive integer. The range of possible fill values is [0..255] for byte arrays and [0..(2^32 - 1)] for word arrays."

	<primitive: 145>
	self errorImproperStore.! !

ArrayedCollection subclass: #RunArray
	instanceVariableNames: 'runs values lastIndex lastRun lastOffset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Arrayed'!

!RunArray commentStamp: '<historical>' prior: 0!
My instances provide space-efficient storage of data which tends to be constant over long runs of the possible indices. Essentially repeated values are stored singly and then associated with a "run" length that denotes the number of consecutive occurrences of the value.

My two important variables are
	runs	An array of how many elements are in each run
	values	An array of what the value is over those elements

The variables lastIndex, lastRun and lastOffset cache the last access
so that streaming through RunArrays is not an N-squared process.

Many complexities of access can be bypassed by using the method
	RunArray withStartStopAndValueDo:!

!RunArray class methodsFor: 'instance creation' stamp: 'ar 10/16/2001 19:03'!
new

	^self runs: Array new values: Array new! !

!RunArray class methodsFor: 'instance creation' stamp: ''!
newFrom: aCollection 
	"Answer an instance of me containing the same elements as aCollection."

	| newCollection |
	newCollection := self new.
	aCollection do: [:x | newCollection addLast: x].
	^newCollection

"	RunArray newFrom: {1. 2. 2. 3}
	{1. $a. $a. 3} as: RunArray
	({1. $a. $a. 3} as: RunArray) values
"! !

!RunArray class methodsFor: 'instance creation' stamp: 'dc 5/24/2007 10:53'!
new: aSize
	^ self new: aSize withAll: nil! !

!RunArray class methodsFor: 'instance creation' stamp: 'ar 10/16/2001 19:04'!
new: size withAll: value 
	"Answer a new instance of me, whose every element is equal to the
	argument, value."

	size = 0 ifTrue: [^self new].
	^self runs: (Array with: size) values: (Array with: value)! !

!RunArray class methodsFor: 'instance creation' stamp: ''!
readFrom: aStream
	"Answer an instance of me as described on the stream, aStream."

	| size runs values |
	size := aStream nextWord.
	runs := Array new: size.
	values := Array new: size.
	1 to: size do:
		[:x |
		runs at: x put: aStream nextWord.
		values at: x put: aStream nextWord].
	^ self runs: runs values: values! !

!RunArray class methodsFor: 'instance creation' stamp: ''!
runs: newRuns values: newValues 
	"Answer an instance of me with runs and values specified by the 
	arguments."

	| instance |
	instance := self basicNew.
	instance setRuns: newRuns setValues: newValues.
	^instance! !

!RunArray class methodsFor: 'instance creation' stamp: 'dtl 1/16/2014 20:01'!
scanFrom: strm
	"Read the style section of a fileOut or sources file.  nextChunk has already been done.  We need to return a RunArray of TextAttributes of various kinds.  These are written by the implementors of writeScanOn:"
	| runs values attrList char |
	(strm peekFor: $( ) ifFalse: [^ nil].
	runs := OrderedCollection new.
	[strm skipSeparators.
	 strm peekFor: $)] whileFalse: 
		[runs add: (Number readFrom: strm)].
	values := OrderedCollection new.	"Value array"
	attrList := OrderedCollection new.	"Attributes list"
	[(char := strm peek) == nil] whileFalse: [
		(char isSeparator or: [ char = $!! ])
			ifTrue: [ "n.b. Skip $!! to meet expectations of RunArrayTest>>testScanFromTrailer.
					The example string used in that test does not seem to match the implemention
					of the fileOut serialization, but the test may be right and the implementation
					wrong. In any case, relax the parsing here to meet the test expectations, and to
					be more consistent with the original version of this method that assumed any
					unexpected charater to be a separator. -dtl Jan 2014"
				strm next "space, cr do nothing"]
			ifFalse: [char == $,
					ifTrue: [strm next.
						values add: attrList asArray.
						attrList := OrderedCollection new]
					ifFalse: [attrList add:  (TextAttribute newFrom: strm)]
				]
		].
	values add: attrList asArray.
	^ self runs: runs asArray values: (values copyFrom: 1 to: runs size) asArray
"
RunArray scanFrom: (ReadStream on: '(14 50 312)f1,f1b,f1LInteger +;i')
"! !

!RunArray methodsFor: 'adding' stamp: 'ar 10/16/2001 18:47'!
addFirst: value
	"Add value as the first element of the receiver."
	lastIndex := nil.  "flush access cache"
	(runs size=0 or: [values first ~= value])
	  ifTrue:
		[runs := {1}, runs.
		values := {value}, values]
	  ifFalse:
		[runs at: 1 put: runs first+1]! !

!RunArray methodsFor: 'adding' stamp: 'nice 3/20/2010 07:58'!
addLast: value
	"Add value as the last element of the receiver."
	lastIndex := nil.  "flush access cache"
	(runs size=0 or: [values last ~= value])
	  ifTrue:
		[runs := runs copyWith: 1.
		values := values copyWith: value]
	  ifFalse:
		[runs at: runs size put: runs last+1].
	^value! !

!RunArray methodsFor: 'adding' stamp: 'ar 10/16/2001 18:47'!
addLast: value  times: times
	"Add value as the last element of the receiver, the given number of times"
	times = 0 ifTrue: [ ^self ].
	lastIndex := nil.  "flush access cache"
	(runs size=0 or: [values last ~= value])
	  ifTrue:
		[runs := runs copyWith: times.
		values := values copyWith: value]
	  ifFalse:
		[runs at: runs size put: runs last+times]! !

!RunArray methodsFor: 'accessing' stamp: ''!
at: index

	self at: index setRunOffsetAndValue: [:run :offset :value | ^value]! !

!RunArray methodsFor: 'accessing' stamp: 'nice 2/14/2007 21:59'!
at: index put: aValue 
	"Set an element of the RunArray"
	| runIndex offsetInRun lastValue runLength runReplacement valueReplacement iStart iStop |
	index isInteger
		ifFalse: [self errorNonIntegerIndex].
	(index >= 1
			and: [index <= self size])
		ifFalse: [self errorSubscriptBounds: index].
	self
		at: index
		setRunOffsetAndValue: [:run :offset :value | 
			runIndex := run.
			offsetInRun := offset.
			lastValue := value].
	aValue = lastValue
		ifTrue: [^ aValue].
	runLength := runs at: runIndex.
	runReplacement := Array
				with: offsetInRun
				with: 1
				with: runLength - offsetInRun - 1.
	valueReplacement := Array
				with: lastValue
				with: aValue
				with: lastValue.
	iStart := offsetInRun = 0
				ifTrue: [2]
				ifFalse: [1].
	iStop := offsetInRun = (runLength - 1)
				ifTrue: [2]
				ifFalse: [3].
	self
		setRuns: (runs copyReplaceFrom: runIndex to: runIndex with: (runReplacement copyFrom: iStart to: iStop))
		setValues: (values copyReplaceFrom: runIndex to: runIndex with: (valueReplacement copyFrom: iStart to: iStop)).
	self coalesce.
	^ aValue! !

!RunArray methodsFor: 'private' stamp: ''!
at: index setRunOffsetAndValue: aBlock 
	"Supply all run information to aBlock."
	"Tolerates index=0 and index=size+1 for copyReplace: "
	| run limit offset |
	limit := runs size.
	(lastIndex == nil or: [index < lastIndex])
		ifTrue:  "cache not loaded, or beyond index - start over"
			[run := 1.
			offset := index-1]
		ifFalse:  "cache loaded and before index - start at cache"
			[run := lastRun.
			offset := lastOffset + (index-lastIndex)].
	[run <= limit and: [offset >= (runs at: run)]]
		whileTrue: 
			[offset := offset - (runs at: run).
			run := run + 1].
	lastIndex := index.  "Load cache for next access"
	lastRun := run.
	lastOffset := offset.
	run > limit
		ifTrue: 
			["adjustment for size+1"
			run := run - 1.
			offset := offset + (runs at: run)].
	^aBlock
		value: run	"an index into runs and values"
		value: offset	"zero-based offset from beginning of this run"
		value: (values at: run)	"value for this run"! !

!RunArray methodsFor: 'adding' stamp: 'tk 1/28/98 09:28'!
coalesce
	"Try to combine adjacent runs"
	| ind |
	ind := 2.
	[ind > values size] whileFalse: [
		(values at: ind-1) = (values at: ind) 
			ifFalse: [ind := ind + 1]
			ifTrue: ["two are the same, combine them"
				values := values copyReplaceFrom: ind to: ind with: #().
				runs at: ind-1 put: (runs at: ind-1) + (runs at: ind).
				runs := runs copyReplaceFrom: ind to: ind with: #().
				"self error: 'needed to combine runs' "]].
			! !

!RunArray methodsFor: 'copying' stamp: 'ls 10/10/1999 13:15'!
copyFrom: start to: stop
	| newRuns run1 run2 offset1 offset2 | 
	stop < start ifTrue: [^RunArray new].
	self at: start setRunOffsetAndValue: [:r :o :value1 | run1 := r. offset1
:= o.  value1].
	self at: stop setRunOffsetAndValue: [:r :o :value2 | run2 := r. offset2
:= o. value2].
	run1 = run2
		ifTrue: 
			[newRuns := Array with: offset2 - offset1 + 1]
		ifFalse: 
			[newRuns := runs copyFrom: run1 to: run2.
			newRuns at: 1 put: (newRuns at: 1) - offset1.
			newRuns at: newRuns size put: offset2 + 1].
	^RunArray runs: newRuns values: (values copyFrom: run1 to: run2)! !

!RunArray methodsFor: 'copying' stamp: ''!
copyReplaceFrom: start to: stop with: replacement

	^(self copyFrom: 1 to: start - 1)
		, replacement 
		, (self copyFrom: stop + 1 to: self size)! !

!RunArray methodsFor: 'accessing' stamp: 'ar 10/16/2001 18:56'!
first
	^values at: 1! !

!RunArray methodsFor: 'accessing' stamp: 'ar 10/16/2001 18:56'!
last
	^values at: values size! !

!RunArray methodsFor: 'private' stamp: 'ul 8/2/2011 20:09'!
mapValues: mapBlock
	"NOTE: only meaningful to an entire set of runs"
	
	values := values collect: mapBlock! !

!RunArray methodsFor: 'copying' stamp: 'nice 10/5/2009 08:44'!
postCopy
	super postCopy.
	runs := runs copy.
	values := values copy! !

!RunArray methodsFor: 'printing' stamp: 'sma 6/1/2000 09:47'!
printOn: aStream
	self printNameOn: aStream.
	aStream
		nextPutAll: ' runs: ';
		print: runs;
		nextPutAll: ' values: ';
		print: values! !

!RunArray methodsFor: 'adding' stamp: 'BG 6/12/2003 11:07'!
rangeOf: attr startingAt: startPos
	"Answer an interval that gives the range of attr at index position  startPos. An empty interval with start value startPos is returned when the attribute attr is not present at position startPos.  self size > 0 is assumed, it is the responsibility of the caller to test for emptiness of self.
Note that an attribute may span several adjancent runs. "

	self at: startPos 
		setRunOffsetAndValue: 
            [:run :offset :value | 
               ^(value includes: attr)
                  ifFalse: [startPos to: startPos - 1]
                  ifTrue:
                    [ | firstRelevantPosition lastRelevantPosition idxOfCandidateRun |
                     lastRelevantPosition := startPos - offset + (runs at: run) - 1.
                     firstRelevantPosition := startPos - offset.
                     idxOfCandidateRun := run + 1.
                     [idxOfCandidateRun <= runs size 
                             and: [(values at: idxOfCandidateRun) includes: attr]]
                        whileTrue:
                          [lastRelevantPosition := lastRelevantPosition + (runs at: idxOfCandidateRun).
                           idxOfCandidateRun := idxOfCandidateRun + 1]. 
                     idxOfCandidateRun := run - 1.
                     [idxOfCandidateRun >= 1 
                             and: [(values at: idxOfCandidateRun) includes: attr]]
                        whileTrue:
                          [firstRelevantPosition := firstRelevantPosition - (runs at: idxOfCandidateRun).
                           idxOfCandidateRun := idxOfCandidateRun - 1]. 
 
                    firstRelevantPosition to: lastRelevantPosition]
		  ]! !

!RunArray methodsFor: 'adding' stamp: 'ar 10/16/2001 18:48'!
repeatLastIfEmpty: defaultBlock
	"add the last value back again.  If we are empty, add (defaultBlock value)"
	lastIndex := nil.  "flush access cache"
	(runs size=0)
	  ifTrue:[
		 runs := runs copyWith: 1.
		values := values copyWith: defaultBlock value]
	  ifFalse:
		[runs at: runs size put: runs last+1]! !

!RunArray methodsFor: 'adding' stamp: 'ar 10/16/2001 18:48'!
repeatLast: times  ifEmpty: defaultBlock
	"add the last value back again, the given number of times.  If we are empty, add (defaultBlock value)"
	times = 0 ifTrue: [^self ].
	lastIndex := nil.  "flush access cache"
	(runs size=0)
	  ifTrue:
		[runs := runs copyWith: times.
		values := values copyWith: defaultBlock value]
	  ifFalse:
		[runs at: runs size put: runs last+times] ! !

!RunArray methodsFor: 'converting' stamp: 'BG 6/8/2003 15:17'!
reversed

  ^self class runs: runs reversed values: values reversed! !

!RunArray methodsFor: 'accessing' stamp: ''!
runLengthAt: index 
	"Answer the length remaining in run beginning at index."

	self at: index 
		setRunOffsetAndValue: [:run :offset :value | ^(runs at: run) - offset]! !

!RunArray methodsFor: 'private' stamp: ''!
runs

	^runs! !

!RunArray methodsFor: 'enumerating' stamp: 'ar 12/27/1999 13:43'!
runsAndValuesDo: aBlock
	"Evaluate aBlock with run lengths and values from the receiver"
	^runs with: values do: aBlock.! !

!RunArray methodsFor: 'enumerating' stamp: 'nice 12/26/2009 20:55'!
runsFrom: start to: stop do: aBlock
	"Evaluate aBlock with all existing runs in the range from start to stop"
	start > stop ifTrue:[^self].
	self at: start setRunOffsetAndValue:[:firstRun :offset :firstValue|
		| run value index |
		run := firstRun.
		value := firstValue.
		index := start + (runs at: run) - offset.
		[aBlock value: value.
		index <= stop] whileTrue:[
			run := run + 1.
			value := values at: run.
			index := index + (runs at: run)]].
! !

!RunArray methodsFor: 'private' stamp: 'ar 10/16/2001 18:47'!
setRuns: newRuns setValues: newValues
	lastIndex := nil.  "flush access cache"
	runs := newRuns asArray.
	values := newValues asArray.! !

!RunArray methodsFor: 'accessing' stamp: ''!
size
	| size |
	size := 0.
	1 to: runs size do: [:i | size := size + (runs at: i)].
	^size! !

!RunArray methodsFor: 'printing' stamp: ''!
storeOn: aStream

	aStream nextPut: $(.
	aStream nextPutAll: self class name.
	aStream nextPutAll: ' runs: '.
	runs storeOn: aStream.
	aStream nextPutAll: ' values: '.
	values storeOn: aStream.
	aStream nextPut: $)! !

!RunArray methodsFor: 'private' stamp: ''!
values
	"Answer the values in the receiver."

	^values! !

!RunArray methodsFor: 'accessing' stamp: 'nice 12/27/2009 04:54'!
withStartStopAndValueDo: aBlock
	| start |
	start := 1.
	runs with: values do:
		[:len : val | | stop |
		stop := start + len - 1.
		aBlock value: start value: stop value: val.
		start := stop + 1]
		! !

!RunArray methodsFor: 'printing' stamp: ''!
writeOn: aStream

	aStream nextWordPut: runs size.
	1 to: runs size do:
		[:x |
		aStream nextWordPut: (runs at: x).
		aStream nextWordPut: (values at: x)]! !

!RunArray methodsFor: 'printing' stamp: 'tk 12/16/97 09:18'!
writeScanOn: strm
	"Write out the format used for text runs in source files. (14 50 312)f1,f1b,f1LInteger +;i"

	strm nextPut: $(.
	runs do: [:rr | rr printOn: strm.  strm space].
	strm skip: -1; nextPut: $).
	values do: [:vv |
		vv do: [:att | att writeScanOn: strm].
		strm nextPut: $,].
	strm skip: -1.  "trailing comma"! !

!RunArray methodsFor: 'copying' stamp: 'ar 10/16/2001 18:57'!
, aRunArray 
	"Answer a new RunArray that is a concatenation of the receiver and
	aRunArray."

	| new newRuns |
	(aRunArray isMemberOf: RunArray)
		ifFalse: 
			[new := self copy.
			"attempt to be sociable"
			aRunArray do: [:each | new addLast: each].
			^new].
	runs size = 0 ifTrue: [^aRunArray copy].
	aRunArray runs size = 0 ifTrue: [^self copy].
	(values at: values size) ~= (aRunArray values at: 1)
		ifTrue: [^RunArray
					runs: runs , aRunArray runs
					values: values , aRunArray values].
	newRuns := runs
					copyReplaceFrom: runs size
					to: runs size
					with: aRunArray runs.
	newRuns at: runs size put: (runs at: runs size) + (aRunArray runs at: 1).
	^RunArray
		runs: newRuns
		values: 
			(values
				copyReplaceFrom: values size
				to: values size
				with: aRunArray values)! !

!RunArray methodsFor: 'accessing' stamp: 'di 1/15/1999 00:04'!
= otherArray 
	"Test if all my elements are equal to those of otherArray"

	(otherArray isMemberOf: RunArray) ifFalse: [^ self hasEqualElements: otherArray].

	"Faster test between two RunArrays"
 	^ (runs hasEqualElements: otherArray runs)
		and: [values hasEqualElements: otherArray values]! !

ArrayedCollection variableSubclass: #SparseLargeTable
	instanceVariableNames: 'base size chunkSize defaultValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Arrayed'!

!SparseLargeTable commentStamp: '<historical>' prior: 0!
Derivated from Stephan Pair's LargeArray, but to hold a sparse table, in which most of the entries are the same default value, it uses some tricks.!

SparseLargeTable variableSubclass: #SparseLargeArray
	instanceVariableNames: 'arrayClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Arrayed'!

!SparseLargeArray commentStamp: '<historical>' prior: 0!
A version of SparseLargeTable that does not populate its bins until a value other than the default is stored.!

!SparseLargeArray methodsFor: 'private' stamp: 'eem 11/5/2008 10:36'!
analyzeSpaceSaving

	| elems tablesTotal nonNilTables lastPage lastChunkSize |
	elems := 0.
	tablesTotal := self basicSize.
	nonNilTables := 0.
	lastPage := self basicAt: self basicSize.
	(lastChunkSize := size \\ chunkSize) = 0 ifTrue:
		[lastChunkSize := chunkSize].
	1 to: self basicSize do:
		[:i | | page |
		(page := self basicAt: i) ifNotNil:
			[nonNilTables := nonNilTables + 1.
			 1 to: (page == lastPage ifTrue: [lastChunkSize] ifFalse: [chunkSize]) do:
				[:j|
				(page at: j) ~= defaultValue ifTrue:
					[elems := elems + 1]]]].

	^String streamContents:
		[:strm |
		strm nextPutAll: 'total: '; print: size.
		strm nextPutAll: ' elements: '; print: elems.
		strm nextPutAll: ' tables: '; print: tablesTotal.
		strm nextPutAll: ' non-nil: '; print: nonNilTables]! !

!SparseLargeArray methodsFor: 'accessing' stamp: 'eem 11/1/2008 14:00'!
atAllPut: anObject 
	"Put anObject at every one of the receiver's indices."

	1 to: self basicSize do:
		[:i|
		self basicAt: i put: nil].
	defaultValue := anObject! !

!SparseLargeArray methodsFor: 'initialization' stamp: 'eem 11/1/2008 12:16'!
initChunkSize: aChunkSize size: aSize arrayClass: aClass base: b defaultValue: d
	chunkSize := aChunkSize.
	size := aSize.
	base := b.
	defaultValue := d.
	arrayClass := aClass
! !

!SparseLargeArray methodsFor: 'accessing' stamp: 'eem 11/1/2008 12:53'!
noCheckAt: index
	^(self basicAt: index - base // chunkSize + 1)
		ifNil: [defaultValue]
		ifNotNil: [:chunk| chunk at: index - base \\ chunkSize + 1]
! !

!SparseLargeArray methodsFor: 'accessing' stamp: 'eem 11/2/2008 11:57'!
noCheckAt: index put: value
	| chunkIndex chunk lastChunkSize |
	chunkIndex := index - base // chunkSize + 1.
	(chunk := self basicAt: chunkIndex) ifNil:
		[value = defaultValue ifTrue:
			[^value].
		chunk := arrayClass
					new: ((chunkIndex == self basicSize
						   and: [(lastChunkSize := size \\ chunkSize) > 0])
							ifTrue: [lastChunkSize]
							ifFalse: [chunkSize])
					withAll: defaultValue.
		self basicAt: chunkIndex put: chunk].
	^chunk at: index - base \\ chunkSize + 1 put: value! !

!SparseLargeTable class methodsFor: 'accessing' stamp: 'yo 12/1/2003 15:54'!
defaultChunkSize

	^100! !

!SparseLargeTable class methodsFor: 'accessing' stamp: 'yo 12/1/2003 15:54'!
defaultChunkSizeForFiles

	^8000! !

!SparseLargeTable class methodsFor: 'instance creation' stamp: 'yo 12/1/2003 16:06'!
new: size

	^self new: size chunkSize: self defaultChunkSize
! !

!SparseLargeTable class methodsFor: 'instance creation' stamp: 'yo 12/1/2003 16:07'!
new: size chunkSize: chunkSize

	^self new: size chunkSize: chunkSize arrayClass: Array
! !

!SparseLargeTable class methodsFor: 'instance creation' stamp: 'yo 12/1/2003 16:08'!
new: size chunkSize: chunkSize arrayClass: aClass

	^self new: size chunkSize: chunkSize arrayClass: Array base: 1.
! !

!SparseLargeTable class methodsFor: 'instance creation' stamp: 'yo 12/1/2003 16:37'!
new: size chunkSize: chunkSize arrayClass: aClass base: b

	^self new: size chunkSize: chunkSize arrayClass: Array base: 1 defaultValue: nil.
! !

!SparseLargeTable class methodsFor: 'instance creation' stamp: 'yo 12/1/2003 16:37'!
new: size chunkSize: chunkSize arrayClass: aClass base: b defaultValue: d

	| basicSize |
	(basicSize := ((size - 1) // chunkSize) + 1) = 0
		ifTrue: [basicSize := 1].
	^(self basicNew: basicSize)
		initChunkSize: chunkSize size: size arrayClass: aClass base: b defaultValue: d;
		yourself
! !

!SparseLargeTable methodsFor: 'private' stamp: 'yo 12/1/2003 18:58'!
allDefaultValueSubtableAt: index

	| t |
	t := self basicAt: index.
	t ifNil: [^ true].
	t do: [:e |
		e ~= defaultValue ifTrue: [^ false].
	].
	^ true.
! !

!SparseLargeTable methodsFor: 'private' stamp: 'yo 12/1/2003 17:10'!
analyzeSpaceSaving

	| total elems tablesTotal nonNilTables |
	total := size - base + 1.
	elems := 0.
	base to: size do: [:i | (self at: i) ~= defaultValue ifTrue: [elems := elems + 1]].
	tablesTotal := self basicSize.
	nonNilTables := 0.
	1 to: self basicSize do: [:i | (self basicAt: i) ifNotNil: [nonNilTables := nonNilTables + 1]].

	^ String streamContents: [:strm |
		strm nextPutAll: 'total: '.
		strm nextPutAll: total printString.
		strm nextPutAll: ' elements: '.
		strm nextPutAll: elems printString.
		strm nextPutAll: ' tables: '.
		strm nextPutAll: tablesTotal printString.
		strm nextPutAll: ' non-nil: '.
		strm nextPutAll: nonNilTables printString.
	].

! !

!SparseLargeTable methodsFor: 'accessing' stamp: 'yo 12/1/2003 15:58'!
arrayClass

	^(self basicAt: 1) class
! !

!SparseLargeTable methodsFor: 'accessing' stamp: 'yo 12/1/2003 15:58'!
at: index

	self pvtCheckIndex: index.
	^self noCheckAt: index.
! !

!SparseLargeTable methodsFor: 'accessing' stamp: 'yo 12/1/2003 15:58'!
at: index put: value
	
	self pvtCheckIndex: index.
	^self noCheckAt: index put: value
! !

!SparseLargeTable methodsFor: 'accessing' stamp: 'yo 12/1/2003 17:56'!
base

	^ base.
! !

!SparseLargeTable methodsFor: 'accessing' stamp: 'yo 12/1/2003 15:58'!
chunkSize

	^chunkSize
! !

!SparseLargeTable methodsFor: 'private' stamp: 'nk 8/31/2004 08:34'!
copyEmpty
	"Answer a copy of the receiver that contains no elements."
	^self speciesNew: 0
! !

!SparseLargeTable methodsFor: 'private' stamp: 'yo 12/1/2003 22:34'!
findLastNonNilSubTable

	(self basicAt: self basicSize) ifNotNil: [^ self basicSize].

	self basicSize - 1 to: 1 by: -1 do: [:lastIndex |
		(self basicAt: lastIndex) ifNotNil: [^ lastIndex].
	].
	^ 0.
! !

!SparseLargeTable methodsFor: 'initialization' stamp: 'yo 12/1/2003 16:58'!
initChunkSize: aChunkSize size: aSize arrayClass: aClass base: b defaultValue: d

	| lastChunkSize |
	chunkSize := aChunkSize.
	size := aSize.
	base := b.
	defaultValue := d.
	1 to: (self basicSize - 1) do: [ :in | self basicAt: in put: (aClass new: chunkSize withAll: defaultValue) ].
	lastChunkSize := size \\ chunkSize.
	lastChunkSize = 0 ifTrue: [lastChunkSize := chunkSize].
	size = 0 
		ifTrue: [self basicAt: 1 put: (aClass new: 0)]
		ifFalse: [self basicAt: self basicSize put: (aClass new: lastChunkSize withAll: defaultValue)].
! !

!SparseLargeTable methodsFor: 'accessing' stamp: 'tak 12/21/2004 16:59'!
noCheckAt: index
	| chunkIndex t |

	chunkIndex := index - base // chunkSize + 1.
	(chunkIndex > self basicSize or: [chunkIndex < 1]) ifTrue: [^ defaultValue].
	t := self basicAt: chunkIndex.
	t ifNil: [^ defaultValue].
	^ t at: (index - base + 1 - (chunkIndex - 1 * chunkSize))
! !

!SparseLargeTable methodsFor: 'accessing' stamp: 'yo 12/1/2003 19:18'!
noCheckAt: index put: value
	| chunkIndex t |

	chunkIndex := index - base // chunkSize + 1.
	chunkIndex > self basicSize ifTrue: [^ value].
	t :=  self basicAt: chunkIndex.
	t ifNil: [^ value].
	^ t at: (index - base + 1 - (chunkIndex - 1 * chunkSize)) put: value
! !

!SparseLargeTable methodsFor: 'copying' stamp: 'nice 10/5/2009 08:46'!
postCopy
	super postCopy.
	1 to: self basicSize do: [:i | self basicAt: i put: (self basicAt: i) copy]! !

!SparseLargeTable methodsFor: 'printing' stamp: 'yo 12/1/2003 17:06'!
printElementsOn: aStream
	| element |
	aStream nextPut: $(.
	base to: size do: [:index | element := self at: index. aStream print: element; space].
	self isEmpty ifFalse: [aStream skip: -1].
	aStream nextPut: $)
! !

!SparseLargeTable methodsFor: 'printing' stamp: 'yo 12/1/2003 15:58'!
printOn: aStream

	(#(String) includes: self arrayClass name) 
		ifTrue: [^self storeOn: aStream].
	^super printOn: aStream
! !

!SparseLargeTable methodsFor: 'private' stamp: 'yo 12/1/2003 19:19'!
privateSize: s

	size := s.
! !

!SparseLargeTable methodsFor: 'private' stamp: 'yo 12/1/2003 17:34'!
pvtCheckIndex: index 

	index isInteger ifFalse: [self errorNonIntegerIndex].
	index < 1 ifTrue: [self errorSubscriptBounds: index].
	index > size ifTrue: [self errorSubscriptBounds: index].
! !

!SparseLargeTable methodsFor: 'private' stamp: 'yo 12/1/2003 15:59'!
similarInstance

	^self class
		new: self size 
		chunkSize: self chunkSize 
		arrayClass: self arrayClass
! !

!SparseLargeTable methodsFor: 'private' stamp: 'yo 12/1/2003 15:59'!
similarInstance: newSize

	^self class
		new: newSize 
		chunkSize: self chunkSize 
		arrayClass: self arrayClass
! !

!SparseLargeTable methodsFor: 'private' stamp: 'yo 12/1/2003 15:59'!
similarSpeciesInstance

	^self similarInstance
! !

!SparseLargeTable methodsFor: 'private' stamp: 'yo 12/1/2003 16:00'!
similarSpeciesInstance: newSize

	^self similarInstance: newSize
! !

!SparseLargeTable methodsFor: 'accessing' stamp: 'yo 12/1/2003 15:58'!
size

	^size
! !

!SparseLargeTable methodsFor: 'private' stamp: 'yo 12/1/2003 16:00'!
speciesNew

	^self species
		new: self size 
		chunkSize: self chunkSize 
		arrayClass: self arrayClass
! !

!SparseLargeTable methodsFor: 'private' stamp: 'yo 12/1/2003 16:00'!
speciesNew: newSize

	^self species
		new: newSize 
		chunkSize: self chunkSize 
		arrayClass: self arrayClass
! !

!SparseLargeTable methodsFor: 'printing' stamp: 'yo 12/1/2003 15:59'!
storeOn: aStream

	| x |
	(#(String) includes: self arrayClass name) ifTrue: 
		[aStream nextPut: $'.
		1 to: self size do:
			[:i |
			aStream nextPut: (x := self at: i).
			x == $' ifTrue: [aStream nextPut: x]].
		aStream nextPutAll: ''' asLargeArrayChunkSize: '.
		aStream nextPutAll: self chunkSize asString.
		^self].
	^super storeOn: aStream
! !

!SparseLargeTable methodsFor: 'accessing' stamp: 'kwl 6/30/2006 03:02'!
zapDefaultOnlyEntries

	| lastIndex newInst |
	1 to: self basicSize do: [:i |
		(self allDefaultValueSubtableAt: i) ifTrue: [self basicAt: i put: nil].
	].

	lastIndex := self findLastNonNilSubTable.
	lastIndex = 0 ifTrue: [^ self].
	
	newInst := self class new: lastIndex*chunkSize chunkSize: chunkSize arrayClass: (self basicAt: lastIndex) class base: base defaultValue: defaultValue.
	newInst privateSize: self size.
	base to: newInst size do: [:i | newInst at: i put: (self at: i)].
	1 to: newInst basicSize do: [:i |
		(newInst allDefaultValueSubtableAt: i) ifTrue: [newInst basicAt: i put: nil].
	].

	" this is not allowed in production: self becomeForward: newInst. "
	^ newInst.
! !

ArrayedCollection subclass: #String
	instanceVariableNames: ''
	classVariableNames: 'AsciiOrder CaseInsensitiveOrder CaseSensitiveOrder CrLfExchangeTable CSMacroCharacters FormatCharacterSet HtmlEntities LowercasingTable Tokenish UppercasingTable'
	poolDictionaries: ''
	category: 'Collections-Strings'!

!String commentStamp: '<historical>' prior: 0!
A String is an indexed collection of Characters. Class String provides the abstract super class for ByteString (that represents an array of 8-bit Characters) and WideString (that represents an array of  32-bit characters).  In the similar manner of LargeInteger and SmallInteger, those subclasses are chosen accordingly for a string; namely as long as the system can figure out so, the String is used to represent the given string.

Strings support a vast array of useful methods, which can best be learned by browsing and trying out examples as you find them in the code.

Here are a few useful methods to look at...
	String match:
	String contractTo:

String also inherits many useful methods from its hierarchy, such as
	SequenceableCollection ,
	SequenceableCollection copyReplaceAll:with:
!

String variableByteSubclass: #ByteString
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Strings'!

!ByteString commentStamp: '<historical>' prior: 0!
This class represents the array of 8 bit wide characters.
!

!ByteString class methodsFor: 'primitives' stamp: 'dtl 10/24/2010 12:44'!
compare: string1 with: string2 collated: order
	"Return 1, 2 or 3, if string1 is <, =, or > string2, with the collating order of characters given by the order array."

	| len1 len2 c1 c2 |
	<primitive: 'primitiveCompareString' module: 'MiscPrimitivePlugin'>
	<var: #string1 declareC: 'unsigned char *string1'>
	<var: #string2 declareC: 'unsigned char *string2'>
	<var: #order declareC: 'unsigned char *order'>

	len1 := string1 size.
	len2 := string2 size.
	1 to: (len1 min: len2) do:
		[:i |
		c1 := order at: (string1 basicAt: i) + 1.
		c2 := order at: (string2 basicAt: i) + 1.
		c1 = c2 ifFalse: 
			[c1 < c2 ifTrue: [^ 1] ifFalse: [^ 3]]].
	len1 = len2 ifTrue: [^ 2].
	len1 < len2 ifTrue: [^ 1] ifFalse: [^ 3].
! !

!ByteString class methodsFor: 'primitives' stamp: 'ul 4/25/2016 01:35'!
findFirstInString: aString  inSet: inclusionMap  startingAt: start
	| i stringSize |
	<primitive: 'primitiveFindFirstInString' module: 'MiscPrimitivePlugin'>
	<var: #aString declareC: 'unsigned char *aString'>
	<var: #inclusionMap  declareC: 'char *inclusionMap'>

	inclusionMap size ~= 256 ifTrue: [ ^0 ].

	i := start.
	stringSize := aString size.
	[ i <= stringSize and: [ (inclusionMap at: (aString basicAt: i) + 1) = 0 ] ] whileTrue: [ 
		i := i + 1 ].

	i > stringSize ifTrue: [ ^0 ].
	^i! !

!ByteString class methodsFor: 'primitives' stamp: 'ul 4/25/2016 01:35'!
indexOfAscii: anInteger inString: aString startingAt: start

	| stringSize |
	<primitive: 'primitiveIndexOfAsciiInString' module: 'MiscPrimitivePlugin'>
	<var: #aCharacter declareC: 'int anInteger'>
	<var: #aString declareC: 'unsigned char *aString'>

	stringSize := aString size.
	start to: stringSize do: [:pos |
		(aString basicAt: pos) = anInteger ifTrue: [^ pos]].

	^ 0
! !

!ByteString class methodsFor: 'primitives' stamp: 'ul 4/25/2016 01:35'!
stringHash: aString initialHash: speciesHash

	| stringSize hash low |
	<primitive: 'primitiveStringHash' module: 'MiscPrimitivePlugin'>

	<var: #aHash declareC: 'int speciesHash'>
	<var: #aString declareC: 'unsigned char *aString'>

	stringSize := aString size.
	hash := speciesHash bitAnd: 16rFFFFFFF.
	1 to: stringSize do: [:pos |
		hash := hash + (aString basicAt: pos).
		"Begin hashMultiply"
		low := hash bitAnd: 16383.
		hash := (16r260D * low + ((16r260D * (hash bitShift: -14) + (16r0065 * low) bitAnd: 16383) * 16384)) bitAnd: 16r0FFFFFFF.
	].
	^ hash! !

!ByteString class methodsFor: 'primitives' stamp: 'ul 4/25/2016 01:35'!
translate: aString from: start  to: stop  table: table
	"translate the characters in the string by the given table, in place"
	<primitive: 'primitiveTranslateStringWithTable' module: 'MiscPrimitivePlugin'>
	<var: #table  declareC: 'unsigned char *table'>
	<var: #aString  declareC: 'unsigned char *aString'>

	start to: stop do: [ :i |
		aString at: i put: (table at: (aString basicAt: i) + 1) ]! !

!ByteString methodsFor: 'converting' stamp: 'ar 4/10/2005 17:20'!
asByteArray
	| ba sz |
	sz := self byteSize.
	ba := ByteArray new: sz.
	ba replaceFrom: 1 to: sz with: self startingAt: 1.
	^ba! !

!ByteString methodsFor: 'converting' stamp: 'ul 10/10/2015 03:25'!
asIntegerSigned: signed
	"Return the first decimal integer I can find or nil."

	| index integerValue result size negative |
	(size := self size) <= 16 ifFalse: [ ^super asIntegerSigned: signed ].
	"Find the first character between $0 and $9."
	index := 0.
	[ (index := index + 1) <= size and: [ 
		(integerValue := (self at: index) asInteger) <= 47 "$0 asInteger - 1" or: [ 
			58 "$9 asInteger + 1" <= integerValue ] ] ] whileTrue.
	index <= size ifFalse: [ ^nil ].
	negative := signed and: [ 2 <= index and: [ (self at: index - 1) == $- ] ].
	"Parse the number."
	result := integerValue - 48 "$0 asInteger".
	[ (index := index + 1) <= size
		and: [ (integerValue := (self at: index) asInteger) <= 57 "$9 asInteger"
		and: [ 48 "$0 asInteger" <= integerValue ] ] ]  whileTrue: [
		result := result * 10 + integerValue - 48 ].
	negative ifTrue: [ ^result negated ].
	^result! !

!ByteString methodsFor: 'converting' stamp: 'yo 8/28/2002 16:52'!
asOctetString

	^ self.
! !

!ByteString methodsFor: 'accessing' stamp: 'yo 8/26/2002 20:33'!
at: index 
	"Primitive. Answer the Character stored in the field of the receiver
	indexed by the argument. Fail if the index argument is not an Integer or
	is out of bounds. Essential. See Object documentation whatIsAPrimitive."

	<primitive: 63>
	^ Character value: (super at: index)! !

!ByteString methodsFor: 'accessing' stamp: 'ar 4/12/2005 17:33'!
at: index put: aCharacter
	"Primitive. Store the Character in the field of the receiver indicated by
	the index. Fail if the index is not an Integer or is out of bounds, or if
	the argument is not a Character. Essential. See Object documentation
	whatIsAPrimitive."

	<primitive: 64>
	aCharacter isCharacter 
		ifFalse:[^self errorImproperStore].
	aCharacter isOctetCharacter ifFalse:[
		"Convert to WideString"
		self becomeForward: (WideString from: self).
		^self at: index put: aCharacter.
	].
	index isInteger
		ifTrue: [self errorSubscriptBounds: index]
		ifFalse: [self errorNonIntegerIndex]! !

!ByteString methodsFor: 'accessing' stamp: 'ar 12/27/1999 13:44'!
byteAt: index
	<primitive: 60>
	^(self at: index) asciiValue! !

!ByteString methodsFor: 'accessing' stamp: 'ar 12/27/1999 13:44'!
byteAt: index put: value
	<primitive: 61>
	self at: index put: value asCharacter.
	^value! !

!ByteString methodsFor: 'accessing' stamp: 'ar 3/3/2001 16:17'!
byteSize
	^self size! !

!ByteString methodsFor: 'testing' stamp: 'cwp 10/24/2012 14:43'!
canBeToken
	"Optimized version for the common case."
	
	^ (self findSubstring: '~' in: self startingAt: 1 matchTable: Tokenish) = 0 
		! !

!ByteString methodsFor: 'converting' stamp: 'yo 11/11/2002 12:20'!
convertFromCompoundText

	| readStream writeStream converter |
	readStream := self readStream.
	writeStream := String new writeStream.
	converter := CompoundTextConverter new.
	converter ifNil: [^ self].
	[readStream atEnd] whileFalse: [
		writeStream nextPut: (converter nextFromStream: readStream)].
	^ writeStream contents
! !

!ByteString methodsFor: 'converting' stamp: 'mir 7/20/2004 15:50'!
convertFromSystemString

	| readStream writeStream converter |
	readStream := self readStream.
	writeStream := String new writeStream.
	converter := LanguageEnvironment defaultSystemConverter.
	converter ifNil: [^ self].
	[readStream atEnd] whileFalse: [
		writeStream nextPut: (converter nextFromStream: readStream)].
	^ writeStream contents
! !

!ByteString methodsFor: 'comparing' stamp: 'ul 4/25/2016 01:35'!
findSubstring: key in: body startingAt: start matchTable: matchTable
	"Answer the index in the string body at which the substring key first occurs, at or beyond start.  The match is determined using matchTable, which can be used to effect, eg, case-insensitive matches.  If no match is found, zero will be returned.

	The algorithm below is not optimum -- it is intended to be translated to C which will go so fast that it wont matter."
	| index |
	<primitive: 'primitiveFindSubstring' module: 'MiscPrimitivePlugin'>
	<var: #key declareC: 'unsigned char *key'>
	<var: #body declareC: 'unsigned char *body'>
	<var: #matchTable declareC: 'unsigned char *matchTable'>

	key size = 0 ifTrue: [^ 0].
	(start max: 1) to: body size - key size + 1 do:
		[:startIndex |
		index := 1.
			[(matchTable at: (body basicAt: startIndex+index-1) + 1)
				= (matchTable at: (key basicAt: index) + 1)]
				whileTrue:
				[index = key size ifTrue: [^ startIndex].
				index := index+1]].
	^ 0
"
' ' findSubstring: 'abc' in: 'abcdefabcd' startingAt: 1 matchTable: CaseSensitiveOrder 1
' ' findSubstring: 'abc' in: 'abcdefabcd' startingAt: 2 matchTable: CaseSensitiveOrder 7
' ' findSubstring: 'abc' in: 'abcdefabcd' startingAt: 8 matchTable: CaseSensitiveOrder 0
' ' findSubstring: 'abc' in: 'abcdefABcd' startingAt: 2 matchTable: CaseSensitiveOrder 0
' ' findSubstring: 'abc' in: 'abcdefABcd' startingAt: 2 matchTable: CaseInsensitiveOrder 7
"! !

!ByteString methodsFor: 'accessing' stamp: 'nice 12/9/2009 15:39'!
indexOfAnyOf: aCollection startingAt: start
	"Use double dispatching for speed"
	^aCollection findFirstInByteString: self startingAt: start! !

!ByteString methodsFor: 'accessing' stamp: 'nice 12/9/2009 15:40'!
indexOfAnyOf: aCollection startingAt: start ifAbsent: aBlock
	"Use double dispatching for speed"
	| index |
	^(index := aCollection findFirstInByteString: self startingAt: start) = 0
		ifTrue: [aBlock value]
		ifFalse: [index]! !

!ByteString methodsFor: 'testing' stamp: 'ar 4/10/2005 18:04'!
isByteString
	"Answer whether the receiver is a ByteString"
	^true! !

!ByteString methodsFor: 'testing' stamp: 'ar 4/10/2005 17:28'!
isOctetString
	"Answer whether the receiver can be represented as a byte string. 
	This is different from asking whether the receiver *is* a ByteString 
	(i.e., #isByteString)"
	^ true.
! !

!ByteString methodsFor: 'enumerating' stamp: 'ul 3/9/2016 21:51'!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject. Optimized version."

	| tally |
	anObject isCharacter ifFalse: [ ^0 ].
	anObject asInteger > 255 ifTrue: [ ^0 ].
	tally := 0.
	1 to: self size do: [ :index |
		(self at: index) == anObject ifTrue: [ tally := tally + 1 ] ].
	^tally! !

!ByteString methodsFor: 'accessing' stamp: 'ar 4/12/2005 17:33'!
replaceFrom: start to: stop with: replacement startingAt: repStart 
	"Primitive. This destructively replaces elements from start to stop in the receiver starting at index, repStart, in the collection, replacement. Answer the receiver. Range checks are performed in the primitive only. Optional. See Object documentation whatIsAPrimitive."
	<primitive: 105>
	replacement class == WideString ifTrue: [
		self becomeForward: (WideString from: self).
	]. 

	super replaceFrom: start to: stop with: replacement startingAt: repStart.
! !

!ByteString methodsFor: 'converting' stamp: 'ul 3/26/2010 23:30'!
squeakToUtf8
	"Convert the given string (self) to UTF-8 from Squeak's internal representation."

	^UTF8TextConverter encodeByteString: self! !

!ByteString methodsFor: 'converting' stamp: 'ul 4/1/2015 19:43'!
substrings
	"Answer an array of the substrings that compose the receiver."
	| result end beginning |
	result := WriteStream on: (Array new: 10).
	end := 0.
	"find one substring each time through this loop"
	[ "find the beginning of the next substring"
	beginning := self indexOfAnyOf: CharacterSet nonSeparators 
							startingAt: end+1 ifAbsent: [ nil ].
	beginning ~~ nil ] whileTrue: [
		"find the end"
		end := self indexOfAnyOf: CharacterSet separators 
					startingAt: beginning ifAbsent: [ self size + 1 ].
		end := end - 1.
		result nextPut: (self copyFrom: beginning to: end).
	].
	^result contents! !

!ByteString methodsFor: 'converting' stamp: 'ul 3/26/2010 23:29'!
utf8ToSqueak
	"Convert the given string (self) from UTF-8 to Squeak's internal representation."
	
	^UTF8TextConverter decodeByteString: self
! !

!String class methodsFor: 'primitives' stamp: 'ul 9/12/2010 04:08'!
compare: string1 with: string2 collated: order
	"Return 1, 2 or 3, if string1 is <, =, or > string2, with the collating order of characters given by the order array."

	| len1 len2 c1 c2 |
	order == nil ifTrue: [
		len1 := string1 size.
		len2 := string2 size.
		1 to: (len1 min: len2) do:[:i |
			c1 := string1 basicAt: i.
			c2 := string2 basicAt: i.
			c1 = c2 ifFalse: [c1 < c2 ifTrue: [^ 1] ifFalse: [^ 3]].
		].
		len1 = len2 ifTrue: [^ 2].
		len1 < len2 ifTrue: [^ 1] ifFalse: [^ 3].
	].
	len1 := string1 size.
	len2 := string2 size.
	1 to: (len1 min: len2) do:[:i |
		c1 := string1 basicAt: i.
		c2 := string2 basicAt: i.
		c1 < 256 ifTrue: [c1 := order at: c1 + 1].
		c2 < 256 ifTrue: [c2 := order at: c2 + 1].
		c1 = c2 ifFalse:[c1 < c2 ifTrue: [^ 1] ifFalse: [^ 3]].
	].
	len1 = len2 ifTrue: [^ 2].
	len1 < len2 ifTrue: [^ 1] ifFalse: [^ 3].
! !

!String class methodsFor: 'instance creation' stamp: 'jm 5/6/1998 18:39'!
cr
	"Answer a string containing a single carriage return character."

	^ self with: Character cr
! !

!String class methodsFor: 'accessing' stamp: 'ul 11/23/2010 08:17'!
crLfExchangeTable

	^CrLfExchangeTable! !

!String class methodsFor: 'instance creation' stamp: 'ls 9/10/1998 22:29'!
crlf
	"Answer a string containing a carriage return and a linefeed."

	^ self with: Character cr with: Character lf
! !

!String class methodsFor: 'instance creation' stamp: 'yo 11/3/2004 19:24'!
crlfcrlf
	^self crlf , self crlf.
! !

!String class methodsFor: 'instance creation' stamp: 'cmm 2/11/2010 21:11'!
empty
	"A canonicalized empty String instance."
	^ ''! !

!String class methodsFor: 'examples' stamp: 'yo 11/3/2004 19:24'!
example
	"To see the string displayed at the cursor point, execute this expression
	and select a point by pressing a mouse button."

	'this is some text' displayOn: Display at: Sensor waitButton! !

!String class methodsFor: 'formatting' stamp: 'md 6/5/2005 07:49'!
expandMacro: macroType argument: argument withExpansions: expansions 
	macroType = $s ifTrue: [^expansions at: argument].
	macroType = $p ifTrue: [^(expansions at: argument) printString].
	macroType = $n ifTrue: [^String cr].
	macroType = $t ifTrue: [^String tab].
	self error: 'unknown expansion type'! !

!String class methodsFor: 'primitives' stamp: 'ul 9/12/2010 04:07'!
findFirstInString: aString inSet: inclusionMap startingAt: start
	"Trivial, non-primitive version"
	
	| i stringSize ascii more |
	inclusionMap size ~= 256 ifTrue: [^ 0].
	stringSize := aString size.
	more := true.
	i := start - 1.
	[more and: [(i := i + 1) <= stringSize]] whileTrue: [
		ascii := aString basicAt: i.
		more := ascii < 256 ifTrue: [(inclusionMap at: ascii + 1) = 0] ifFalse: [true].
	].

	i > stringSize ifTrue: [^ 0].
	^ i! !

!String class methodsFor: 'instance creation' stamp: 'yo 10/23/2002 23:43'!
fromByteArray: aByteArray

	^ aByteArray asString
! !

!String class methodsFor: 'instance creation' stamp: 'yo 11/3/2004 19:24'!
fromPacked: aLong
	"Convert from a longinteger to a String of length 4."

	| s |
	s := self new: 4.
	s at: 1 put: (aLong digitAt: 4) asCharacter.
	s at: 2 put: (aLong digitAt: 3) asCharacter.
	s at: 3 put: (aLong digitAt: 2) asCharacter.
	s at: 4 put: (aLong digitAt: 1) asCharacter.
	^s

"String fromPacked: 'TEXT' asPacked"
! !

!String class methodsFor: 'instance creation' stamp: 'di 1/14/98 10:26'!
fromString: aString 
	"Answer an instance of me that is a copy of the argument, aString."
	
	^ aString copyFrom: 1 to: aString size! !

!String class methodsFor: 'accessing' stamp: 'mt 5/3/2015 20:27'!
htmlEntities
	
	^ HtmlEntities! !

!String class methodsFor: 'primitives' stamp: 'ul 9/12/2010 04:06'!
indexOfAscii: anInteger inString: aString startingAt: start
	"Trivial, non-primitive version"

	start to: aString size do: [ :index |
		(aString basicAt: index) = anInteger ifTrue: [ ^index ] ].
	^0
! !

!String class methodsFor: 'initialization' stamp: 'ul 4/2/2015 23:22'!
initialize   "self initialize"

	| order |
	AsciiOrder := (0 to: 255) as: ByteArray.

	CaseInsensitiveOrder := AsciiOrder copy.
	($a to: $z) do:
		[:c | CaseInsensitiveOrder at: c asciiValue + 1
				put: (CaseInsensitiveOrder at: c asUppercase asciiValue +1)].

	"Case-sensitive compare sorts space, digits, letters, all the rest..."
	CaseSensitiveOrder := ByteArray new: 256 withAll: 255.
	order := -1.
	' 0123456789' do:  "0..10"
		[:c | CaseSensitiveOrder at: c asciiValue + 1 put: (order := order+1)].
	($a to: $z) do:     "11-64"
		[:c | CaseSensitiveOrder at: c asUppercase asciiValue + 1 put: (order := order+1).
		CaseSensitiveOrder at: c asciiValue + 1 put: (order := order+1)].
	1 to: CaseSensitiveOrder size do:
		[:i | (CaseSensitiveOrder at: i) = 255 ifTrue:
			[CaseSensitiveOrder at: i put: (order := order+1)]].
	order = 255 ifFalse: [self error: 'order problem'].

	"a table for translating to lower case"
	LowercasingTable := String withAll: (Character allByteCharacters collect: [:c | c asLowercase]).

	"a table for translating to upper case"
	UppercasingTable := String withAll: (Character allByteCharacters collect: [:c | c asUppercase]).

	"a table for testing tokenish (for fast numArgs)"
	Tokenish := String withAll: (Character allByteCharacters collect:
									[:c | c tokenish ifTrue: [c] ifFalse: [$~]]).
 
	"% and < for #expandMacros*"
	CSMacroCharacters := CharacterSet newFrom: '%<'.

	"{\ used by #format:"
	FormatCharacterSet := CharacterSet newFrom: '{\'.
	
	"a table for exchanging cr with lf and vica versa"
	CrLfExchangeTable := Character allByteCharacters collect: [ :each |
		each
			caseOf: {
				[ Character cr ] -> [ Character lf ].
				[ Character lf ] -> [ Character cr ] }
			otherwise: [ each ] ]! !

!String class methodsFor: 'initialization' stamp: 'yo 8/11/2003 21:11'!
initializeHtmlEntities
	"self initializeHtmlEntities"

	HtmlEntities := (Dictionary new: 128)
		at: 'amp'	put: $&;
		at: 'lt'		put: $<;
		at: 'gt'		put: $>;
		at: 'quot'	put: $";
		at: 'euro'	put: Character euro;
		yourself.
	#('nbsp' 'iexcl' 'cent' 'pound' 'curren' 'yen' 'brvbar' 'sect' 'uml' 'copy' 'ordf' 'laquo' 'not' 'shy' 'reg' 'hibar' 'deg' 'plusmn' 'sup2' 'sup3' 'acute' 'micro' 'para' 'middot' 'cedil' 'sup1' 'ordm' 'raquo' 'frac14' 'frac12' 'frac34' 'iquest' 'Agrave' 'Aacute' 'Acirc' 'Atilde' 'Auml' 'Aring' 'AElig' 'Ccedil' 'Egrave' 'Eacute' 'Ecirc' 'Euml' 'Igrave' 'Iacute' 'Icirc' 'Iuml' 'ETH' 'Ntilde' 'Ograve' 'Oacute' 'Ocirc' 'Otilde' 'Ouml' 'times' 'Oslash' 'Ugrave' 'Uacute' 'Ucirc' 'Uuml' 'Yacute' 'THORN' 'szlig' 'agrave' 'aacute' 'acirc' 'atilde' 'auml' 'aring' 'aelig' 'ccedil' 'egrave' 'eacute' 'ecirc' 'euml' 'igrave' 'iacute' 'icirc' 'iuml' 'eth' 'ntilde' 'ograve' 'oacute' 'ocirc' 'otilde' 'ouml' 'divide' 'oslash' 'ugrave' 'uacute' 'ucirc' 'uuml' 'yacute' 'thorn' 'yuml' ) withIndexDo: [:each :index | HtmlEntities at: each put: (index + 159) asCharacter]! !

!String class methodsFor: 'instance creation' stamp: 'yo 11/3/2004 19:24'!
lf
	"Answer a string containing a single carriage return character."

	^ self with: Character lf! !

!String class methodsFor: 'instance creation' stamp: 'ar 4/10/2005 23:26'!
new: sizeRequested 
	"Answer an instance of this class with the number of indexable
	variables specified by the argument, sizeRequested."
	self == String 
		ifTrue:[^ByteString new: sizeRequested]
		ifFalse:[^self basicNew: sizeRequested].! !

!String class methodsFor: 'instance creation' stamp: 'yo 8/28/2002 13:27'!
readFrom: inStream
	"Answer an instance of me that is determined by reading the stream, 
	inStream. Embedded double quotes become the quote Character."

	| outStream char done |
	outStream := WriteStream on: (self new: 16).
	"go to first quote"
	inStream skipTo: $'.
	done := false.
	[done or: [inStream atEnd]]
		whileFalse: 
			[char := inStream next.
			char = $'
				ifTrue: 
					[char := inStream next.
					char = $'
						ifTrue: [outStream nextPut: char]
						ifFalse: [done := true]]
				ifFalse: [outStream nextPut: char]].
	^outStream contents! !

!String class methodsFor: 'instance creation' stamp: 'dc 2/12/2007 10:47'!
space
	"Answer a string containing a single space character."

	^ self with: Character space
! !

!String class methodsFor: 'primitives' stamp: 'ul 9/12/2010 04:02'!
stringHash: aString initialHash: speciesHash

	| stringSize hash low |
	stringSize := aString size.
	hash := speciesHash bitAnd: 16r0FFFFFFF.
	1 to: stringSize do: [ :pos |
		hash := hash + (aString basicAt: pos).
		"Begin hashMultiply"
		low := hash bitAnd: 16383.
		hash := (16r260D * low + ((16r260D * (hash // 16384) + (16r0065 * low) bitAnd: 16383) * 16384)) bitAnd: 16r0FFFFFFF ].
	^hash.


! !

!String class methodsFor: 'instance creation' stamp: 'sw 6/15/1999 22:59'!
tab
	"Answer a string containing a single tab character."

	^ self with: Character tab
! !

!String class methodsFor: 'primitives' stamp: 'ul 9/12/2010 04:09'!
translate: aString from: start  to: stop  table: table
	"Trivial, non-primitive version"

	start to: stop do: [ :i |
		| char |
		(char := aString basicAt: i) < 256 ifTrue: [
			aString at: i put: (table at: char+1) ] ].
! !

!String class methodsFor: 'instance creation' stamp: 'yo 8/28/2002 13:29'!
value: anInteger

	^ self with: (Character value: anInteger).
! !

!String class methodsFor: 'instance creation' stamp: 'ar 4/12/2005 17:34'!
with: aCharacter
	| newCollection |
	aCharacter asInteger < 256
		ifTrue:[newCollection := ByteString new: 1]
		ifFalse:[newCollection := WideString new: 1].
	newCollection at: 1 put: aCharacter.
	^newCollection! !

!String methodsFor: 'converting' stamp: 'di 11/6/1998 13:47'!
adaptToCollection: rcvr andSend: selector
	"If I am involved in arithmetic with a collection, convert me to a number."

	^ rcvr perform: selector with: self asNumber! !

!String methodsFor: 'converting' stamp: 'di 11/9/1998 12:17'!
adaptToNumber: rcvr andSend: selector
	"If I am involved in arithmetic with a number, convert me to a number."

	^ rcvr perform: selector with: self asNumber! !

!String methodsFor: 'converting' stamp: 'di 11/6/1998 13:49'!
adaptToPoint: rcvr andSend: selector
	"If I am involved in arithmetic with a point, convert me to a number."

	^ rcvr perform: selector with: self asNumber! !

!String methodsFor: 'converting' stamp: 'di 11/6/1998 13:49'!
adaptToString: rcvr andSend: selector
	"If I am involved in arithmetic with a string, convert us both to
	numbers, and return the printString of the result."

	^ (rcvr asNumber perform: selector with: self asNumber) printString! !

!String methodsFor: 'comparing' stamp: 'yo 11/3/2004 19:24'!
alike: aString 
	"Answer some indication of how alike the receiver is to the argument,  0 is no match, twice aString size is best score.  Case is ignored."

	| i j k minSize bonus |
	minSize := (j := self size) min: (k := aString size).
	bonus := (j - k) abs < 2 ifTrue: [ 1 ] ifFalse: [ 0 ].
	i := 1.
	[(i <= minSize) and: [((super at: i) bitAnd: 16rDF)  = ((aString at: i) asciiValue bitAnd: 16rDF)]]
		whileTrue: [ i := i + 1 ].
	[(j > 0) and: [(k > 0) and:
		[((super at: j) bitAnd: 16rDF) = ((aString at: k) asciiValue bitAnd: 16rDF)]]]
			whileTrue: [ j := j - 1.  k := k - 1. ].
	^ i - 1 + self size - j + bonus. ! !

!String methodsFor: 'converting' stamp: 'cmm 8/6/2013 18:11'!
asAscii
	^ self select: [ : each | each isAscii ]! !

!String methodsFor: 'converting' stamp: 'ar 4/10/2005 17:18'!
asByteArray
	"Convert to a ByteArray with the ascii values of the string."
	| b |
	b := ByteArray new: self byteSize.
	1 to: self size * 4 do: [:i |
		b at: i put: (self byteAt: i).
	].
	^ b.
! !

!String methodsFor: 'converting' stamp: 'ar 4/10/2005 16:22'!
asByteString
	"Convert the receiver into a ByteString"
	^self asOctetString! !

!String methodsFor: 'converting' stamp: 'fbs 2/12/2013 09:28'!
asCamelCase
	"Convert to CamelCase. Can be convenient 
	in conjunction with #asLegalSelector 
	'A man, a plan, a canal, panama' asCamelCase.
	'A man, a plan, a canal, panama' asCamelCase asLegalSelector.
	'Here 123should % be 6 the name6 of the method' asCamelCase.
	'Here 123should % be 6 the name6 of the method' asCamelCase asLegalSelector."
	
	^ self class streamContents: [:stream | 
		self substrings do: [:sub |
			stream nextPutAll: sub capitalized]]! !

!String methodsFor: 'converting' stamp: 'ar 4/10/2005 17:03'!
asCharacter
	"Answer the receiver's first character, or '*' if none.  Idiosyncratic, provisional."

	^ self size > 0 ifTrue: [self first] ifFalse:[$·]! !

!String methodsFor: 'converting' stamp: 'topa 7/10/2016 00:00'!
ascii85Decoded
	"Decode the receiver from Ascii85"
	"'<~87cURD]i,""Ebo7~>' ascii85Decoded"

	^ self ascii85DecodedAs: self class
! !

!String methodsFor: 'converting' stamp: 'topa 7/10/2016 00:00'!
ascii85DecodedAs: aClass
	"Decode the receiver from Ascii85"
	"'<~87cURD]i,""Ebo7~>' ascii85DecodedAs: String"

	^ Ascii85Converter decode: self as: aClass! !

!String methodsFor: 'converting' stamp: 'topa 7/9/2016 23:57'!
ascii85Encoded
	"Encode the receiver as Ascii85"
	"'Hello World' ascii85Encoded"

	^ (Ascii85Converter encode: self readStream) contents
! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
asDate
	"Many allowed forms, see Date>>#readFrom:"

	^ Date fromString: self! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
asDateAndTime

	"Convert from UTC format" 	^ DateAndTime fromString: self! !

!String methodsFor: 'converting' stamp: 'ar 8/27/2009 20:41'!
asDecomposedUnicode
	"Convert the receiver into a decomposed Unicode representation.
	Optimized for the common case that no decomposition needs to take place."
	| lastIndex nextIndex out decomposed |
	lastIndex := 1.
	nextIndex := 0.
	[(nextIndex := nextIndex+1) <= self size] whileTrue:[
		decomposed := Unicode decompose: (self at: nextIndex).
		decomposed ifNotNil:[
			lastIndex = 1 ifTrue:[out := WriteStream on: (String new: self size)].
			out nextPutAll: (self copyFrom: lastIndex to: nextIndex-1).
			out nextPutAll: decomposed.
			lastIndex := nextIndex+1.
		].
	].
	^out ifNil:[self] ifNotNil:[
		out nextPutAll: (self copyFrom: lastIndex to: self size).
		out contents]! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
asDisplayText
	"Answer a DisplayText whose text string is the receiver."

	^DisplayText text: self asText! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
asDuration
	"convert from [nnnd]hh:mm:ss[.nanos] format. [] implies optional elements"

	^ Duration fromString: self
! !

!String methodsFor: 'user interface' stamp: 'md 7/28/2005 20:42'!
asExplorerString

	^ self printString! !

!String methodsFor: 'converting' stamp: 'ar 4/12/2005 13:55'!
asFileName
	"Answer a String made up from the receiver that is an acceptable file 
	name."

	| string checkedString |
	string := FileDirectory checkName: self fixErrors: true.
	checkedString := (FilePath pathName: string) asVmPathName.
	^ (FilePath pathName: checkedString isEncoded: true) asSqueakPathName.
! !

!String methodsFor: 'converting' stamp: 'yo 8/27/2002 14:38'!
asFourCode

	| result |
	self size = 4 ifFalse: [^self error: 'must be exactly four characters'].
	result := self inject: 0 into: [:val :each | 256 * val + each asciiValue].
	(result bitAnd: 16r80000000) = 0 
		ifFalse: [self error: 'cannot resolve fourcode'].
	(result bitAnd: 16r40000000) = 0 ifFalse: [^result - 16r80000000].
	^ result
! !

!String methodsFor: 'converting' stamp: 'yo 8/26/2002 23:06'!
asHex
	| stream |
	stream := WriteStream on: (String new: self size * 4).
	self do: [ :ch | stream nextPutAll: ch hex ].
	^stream contents! !

!String methodsFor: 'converting' stamp: 'sma 3/11/2000 17:25'!
asHtml
	"Do the basic character conversion for HTML.  Leave all original return 
	and tabs in place, so can conver back by simply removing bracked 
	things. 4/4/96 tk"
	| temp |
	temp := self copyReplaceAll: '&' with: '&amp;'.
	HtmlEntities keysAndValuesDo:
		[:entity :char |
		char = $& ifFalse:
			[temp := temp copyReplaceAll: char asString with: '&' , entity , ';']].
	temp := temp copyReplaceAll: '	' with: '	<IMG SRC="tab.gif" ALT="    ">'.
	temp := temp copyReplaceAll: '
' with: '
<BR>'.
	^ temp

"
	'A<&>B' asHtml
"! !

!String methodsFor: 'converting' stamp: 'ul 11/15/2010 10:07'!
asIdentifier: shouldBeCapitalized
	"Return a legal identifier, with first character in upper case if shouldBeCapitalized is true, else lower case.  This will always return a legal identifier, even for an empty string"

	| aString firstChar firstLetterPosition |
	aString := self select: [:el | el isAlphaNumeric].
	firstLetterPosition := aString findFirst: [:ch | ch isLetter].
	aString := firstLetterPosition = 0
		ifFalse:
			[aString copyFrom: firstLetterPosition to: aString size]
		ifTrue:
			['a', aString].
	firstChar := shouldBeCapitalized ifTrue: [aString first asUppercase] ifFalse: [aString first asLowercase].

	^ firstChar asString, (aString copyFrom: 2 to: aString size)
"
'234Fred987' asIdentifier: false
'235Fred987' asIdentifier: true
'' asIdentifier: true
'()87234' asIdentifier: false
'())z>=PPve889  U >' asIdentifier: false

"! !

!String methodsFor: 'converting' stamp: 'ul 9/28/2015 20:32'!
asInteger 
	
	^self asIntegerSigned: true
! !

!String methodsFor: 'converting' stamp: 'ul 10/10/2015 03:13'!
asIntegerSigned: signed
	"Return the first decimal integer I can find or nil."

	| index character size result negative |
	index := 0.
	size := self size.
	"Find the first character between $0 and $9."
	[ (index := index + 1) > size or: [ (self at: index) isDigit ] ] whileFalse.
	index > size ifTrue: [ ^nil ].
	negative := signed and: [ index > 1 and: [ (self at: index - 1) == $- ] ].
	"Parse the number."
	size - index > 15 ifTrue: [
		negative ifTrue: [ index := index - 1 ].
		^Integer readFrom: (
			ReadStream
				on: self
				from: index
				to: size) ].
	result := (self at: index) digitValue.
	[ (index := index + 1) <= size
		and: [ (character := self at: index) isDigit ] ]  whileTrue: [
		result := result * 10 + character digitValue ].
	negative ifTrue: [ ^result negated ].
	^result! !

!String methodsFor: 'converting' stamp: 'di 2/2/98 12:59'!
askIfAddStyle: priorMethod req: requestor
	^ self   "we are a string with no text style"! !

!String methodsFor: 'converting' stamp: 'nice 3/12/2010 23:18'!
asLegalSelector
	| toUse |
	toUse := self select: [:char | char isAlphaNumeric].
	(toUse size = 0 or: [toUse first isLetter not])
		ifTrue: [toUse := 'v', toUse].
	^ toUse withFirstCharacterDownshifted! !

!String methodsFor: 'converting' stamp: 'sma 6/12/2000 11:40'!
asLowercase
	"Answer a String made up from the receiver whose characters are all 
	lowercase."

	^ self copy asString translateToLowercase! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
asNumber 
	"Answer the Number created by interpreting the receiver as the string 
	representation of a number."

	^Number readFromString: self! !

!String methodsFor: 'converting' stamp: 'ar 4/10/2005 20:55'!
asOctetString
	"Convert the receiver into an octet string"
	| string |
	string := String new: self size.
	1 to: self size do: [:i | string at: i put: (self at: i)].
	^string! !

!String methodsFor: 'converting' stamp: 'nice 12/26/2009 01:02'!
asPacked
	"Convert to a longinteger that describes the string"

	^ self inject: 0 into: [ :pack :next | pack * 256 + next asInteger ].! !

!String methodsFor: 'converting' stamp: 'dew 9/13/2001 01:17'!
asPluralBasedOn: aNumberOrCollection
	"Append an 's' to this string based on whether aNumberOrCollection is 1 or of size 1."

	^ (aNumberOrCollection = 1 or:
		[aNumberOrCollection isCollection and: [aNumberOrCollection size = 1]])
			ifTrue: [self]
			ifFalse: [self, 's']
! !

!String methodsFor: 'converting' stamp: 'ar 8/27/2009 20:41'!
asPrecomposedUnicode
	"Convert the receiver into a precomposed Unicode representation.
	Optimized for the common case that no composition needs to take place."
	| lastIndex nextIndex composed out |
	lastIndex := 1.
	nextIndex := 0.
	[(nextIndex := nextIndex+1) < self size] whileTrue:[
		composed := Unicode compose: (self at: nextIndex) with: (self at: nextIndex+1).
		composed ifNotNil:[
			lastIndex = 1 ifTrue:[out := WriteStream on: (String new: self size)].
			out nextPutAll: (self copyFrom: lastIndex to: nextIndex-1).
			out nextPut: composed.
			nextIndex := nextIndex+1.
			lastIndex := nextIndex+1.
		].
	].
	^out ifNil:[self] ifNotNil:[
		out nextPutAll: (self copyFrom: lastIndex to: self size).
		out contents]! !

!String methodsFor: 'converting' stamp: 'ul 10/9/2015 22:45'!
asSignedInteger
	"Return the first signed integer I can find or nil."
	
	^self asIntegerSigned: true! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
asSmalltalkComment
	"return this string, munged so that it can be treated as a comment in Smalltalk code.  Quote marks are added to the beginning and end of the string, and whenever a solitary quote mark appears within the string, it is doubled"

	^String streamContents:  [ :str |
		| quoteCount first |

		str nextPut: $".
	
		quoteCount := 0.
		first := true.
		self do: [ :char |
			char = $"
				ifTrue: [
					first ifFalse: [
						str nextPut: char.
						quoteCount := quoteCount + 1 ] ]
				ifFalse: [
					quoteCount odd ifTrue: [
						"add a quote to even the number of quotes in a row"
						str nextPut: $" ].
					quoteCount := 0.
					str nextPut: char ].
			first := false ]. 

		quoteCount odd ifTrue: [
			"check at the end"
			str nextPut: $". ].

		str nextPut: $".
	].
	! !

!String methodsFor: 'converting' stamp: 'yo 12/19/2003 21:16'!
asSqueakPathName

	^ self.
! !

!String methodsFor: 'converting' stamp: 'jm 5/14/1998 10:40'!
asString
	"Answer this string."

	^ self
! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
asStringOrText
	"Answer this string."

	^ self
! !

!String methodsFor: 'converting' stamp: 'ar 4/10/2005 19:24'!
asSymbol
	"Answer the unique Symbol whose characters are the characters of the 
	string."
	^Symbol intern: self! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
asText
	"Answer a Text whose string is the receiver."

	^Text fromString: self! !

!String methodsFor: 'converting' stamp: 'mt 5/3/2015 17:45'!
asTextFromHtml
	"Answer a Text by interpreting the receiver as HTML."

	^ (HtmlReadWriter on: self readStream) nextText! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
asTime
	"Many allowed forms, see Time>>readFrom:"

	^ Time fromString: self.! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
asTimeStamp
	"Convert from obsolete TimeStamp format"

	^ TimeStamp fromString: self! !

!String methodsFor: 'converting' stamp: 'bf 11/26/2012 18:38'!
asUnHtml
	"Strip out all Html stuff (commands in angle brackets <>) and convert
the characters &<> back to their real value.  Leave actual cr and tab as
they were in text."
	| in out char rest |
	in := ReadStream on: self.
	out := WriteStream on: (String new: self size).
	[in atEnd] whileFalse:
		[in peek = $<
			ifTrue: [in unCommand] 	"Absorb <...><...>"
			ifFalse: [(char := in next) = $&
						ifTrue: [rest := in upTo: $;.
								out nextPut: (HtmlEntities
									at: rest
									ifAbsent: [
										(rest beginsWith: '#')
											ifTrue: [Character value: rest allButFirst asInteger]
											ifFalse: [Character space]])]
						ifFalse: [out nextPut: char]].
		].
	^ out contents! !

!String methodsFor: 'converting' stamp: 'ul 10/9/2015 22:45'!
asUnsignedInteger 
	"Returns the first unsigned integer I can find or nil."

	^self asIntegerSigned: false! !

!String methodsFor: 'converting' stamp: 'SqR 6/13/2000 20:10'!
asUppercase
	"Answer a String made up from the receiver whose characters are all 
	uppercase."

	^self copy asString translateToUppercase! !

!String methodsFor: 'converting' stamp: 'yo 2/24/2005 18:33'!
asVmPathName

	^ (FilePath pathName: self) asVmPathName.
! !

!String methodsFor: 'converting' stamp: 'ar 4/12/2005 17:36'!
asWideString 
	self isWideString
		ifTrue:[^self]
		ifFalse:[^WideString from: self]! !

!String methodsFor: 'converting' stamp: 'ar 7/22/2009 22:00'!
base64Decoded
	"Decode the receiver from base 64"
	"'SGVsbG8gV29ybGQ=' base64Decoded"
	^(Base64MimeConverter mimeDecode: self as: self class)! !

!String methodsFor: 'converting' stamp: 'ul 10/30/2012 02:09'!
base64Encoded
	"Encode the receiver as base64"
	"'Hello World' base64Encoded"

	^(Base64MimeConverter
		mimeEncode: (ReadStream on: self)
		multiLine: false) contents! !

!String methodsFor: 'testing' stamp: 'ul 4/3/2016 00:38'!
beginsWith: sequence
	"Answer true if the receiver starts with the argument collection. The comparison is case-sensitive. Overridden for better performance."

	| index sequenceSize |
	sequence isString ifFalse: [ ^super beginsWith: sequence ].
	((sequenceSize := sequence size) = 0 or: [ self size < sequenceSize ]) ifTrue: [ ^false ].
	index := 0.
	[ (index := index + 1) <= sequenceSize ] whileTrue: [
		(sequence at: index) == (self at: index) ifFalse: [ ^false ] ].
	^true! !

!String methodsFor: 'accessing' stamp: 'ar 4/12/2005 16:30'!
byteAt: index
	^self subclassResponsibility! !

!String methodsFor: 'accessing' stamp: 'ar 4/12/2005 16:30'!
byteAt: index put: value
	^self subclassResponsibility! !

!String methodsFor: 'filter streaming' stamp: 'yo 8/26/2002 22:31'!
byteEncode:aStream

	^aStream writeString: self.
! !

!String methodsFor: 'accessing' stamp: 'ar 4/12/2005 16:30'!
byteSize
	^self subclassResponsibility! !

!String methodsFor: 'testing' stamp: 'cwp 10/24/2012 14:43'!
canBeToken
	"Extracted from #numArgs to allow specialization by subclasses"
	
	^ self allSatisfy: [:c | c tokenish]! !

!String methodsFor: 'converting' stamp: 'ls 9/10/1998 08:48'!
capitalized
	"Return a copy with the first letter capitalized"
	| cap |
	self isEmpty ifTrue: [ ^self copy ].
	cap := self copy.
	cap at: 1 put: (cap at: 1) asUppercase.
	^ cap! !

!String methodsFor: 'comparing' stamp: 'ar 4/10/2005 16:39'!
caseInsensitiveLessOrEqual: aString 
	"Answer whether the receiver sorts before or equal to aString.
	The collation order is case insensitive."
	^(self compare: aString caseSensitive: false) <= 2! !

!String methodsFor: 'comparing' stamp: 'ar 4/10/2005 16:39'!
caseSensitiveLessOrEqual: aString 
	"Answer whether the receiver sorts before or equal to aString.
	The collation order is case sensitive."
	^(self compare: aString caseSensitive: true) <= 2! !

!String methodsFor: 'comparing' stamp: 'yo 8/27/2002 14:15'!
charactersExactlyMatching: aString
	"Do a character-by-character comparison between the receiver and aString.  Return the index of the final character that matched exactly."

	| count |
	count := self size min: aString size.
	1 to: count do: [:i | 
		(self at: i) = (aString at: i) ifFalse: [
			^ i - 1]].
	^ count! !

!String methodsFor: 'enumerating' stamp: 'nice 11/1/2011 22:18'!
combinations: kk atATimeDo: aBlock
	"Gather the combinations into a String rather than an Array"
	
	" 'abcde' combinations: 3 atATimeDo: [:each | Transcript cr; show: each printString]"

	| aCollection |
	aCollection := String new: kk.
	self combinationsAt: 1 in: aCollection after: 0 do: aBlock! !

!String methodsFor: '*collections' stamp: 'cmm 9/7/2013 12:09'!
compareSafely: aString 
	^ aString isString
		ifTrue: [ self < aString ]
		ifFalse: [ super compareSafely: aString ]! !

!String methodsFor: 'comparing' stamp: 'ar 4/10/2005 16:38'!
compare: aString 
	"Answer a comparison code telling how the receiver sorts relative to aString:
		1 - before
		2 - equal
		3 - after.
	The collation sequence is ascii with case differences ignored.
	To get the effect of a <= b, but ignoring case, use (a compare: b) <= 2."
	^self compare: aString caseSensitive: false! !

!String methodsFor: 'comparing' stamp: 'ar 4/10/2005 16:42'!
compare: aString caseSensitive: aBool
	"Answer a comparison code telling how the receiver sorts relative to aString:
		1 - before
		2 - equal
		3 - after.
	"
	| map |
	map := aBool ifTrue:[CaseSensitiveOrder] ifFalse:[CaseInsensitiveOrder].
	^self compare: self with: aString collated: map! !

!String methodsFor: 'comparing' stamp: 'yo 12/15/2005 14:28'!
compare: string1 with: string2 collated: order

	(string1 isByteString and: [string2 isByteString]) ifTrue: [
		^ ByteString compare: string1 with: string2 collated: order
	].
     "Primitive does not fail properly right now"
      ^ String compare: string1 with: string2 collated: order

"
self assert: 'abc' = 'abc' asWideString.
self assert: 'abc' asWideString = 'abc'.
self assert: ((ByteArray with: 97 with: 0 with: 0 with: 0) asString ~= 'a000' asWideString).
self assert: ('a000' asWideString ~= (ByteArray with: 97 with: 0 with: 0 with: 0) asString).

self assert: ('abc' sameAs: 'aBc' asWideString).
self assert: ('aBc' asWideString sameAs: 'abc').
self assert: ((ByteArray with: 97 with: 0 with: 0 with: 0) asString sameAs: 'Abcd' asWideString) not.
self assert: ('a000' asWideString sameAs: (ByteArray with: 97 with: 0 with: 0 with: 0) asString) not.

"! !

!String methodsFor: 'converting' stamp: 'nice 5/31/2013 00:05'!
compressWithTable: tokens
	"Return a string with all substrings that occur in tokens replaced
	by a character with ascii code = 127 + token index.
	This will work best if tokens are sorted by size.
	Assumes this string contains no characters > 127, or that they
	are intentionally there and will not interfere with this process."
	| str null finalSize result ri c |
	null := Character null.
	str := self copyFrom: 1 to: self size.  "Working string will get altered"
	finalSize := str size.
	tokens doWithIndex:
		[:token :tIndex |
		| start ts |
		start := 1.
		[(start := str findString: token startingAt: start) > 0]
			whileTrue:
			[ts := token size.
			((start + ts) <= str size
				and: [(str at: start + ts) = $  and: [tIndex*2 <= 128]])
				ifTrue: [ts := token size + 1.  "include training blank"
						str at: start put: (Character value: tIndex*2 + 127)]
				ifFalse: [str at: start put: (Character value: tIndex + 127)].
			str at: start put: (Character value: tIndex + 127).
			1 to: ts-1 do: [:i | str at: start+i put: null].
			finalSize := finalSize - (ts - 1).
			start := start + ts]].
	result := String new: finalSize.
	ri := 0.
	1 to: str size do:
		[:i | (c := str at: i) = null ifFalse: [result at: (ri := ri+1) put: c]].
	^ result! !

!String methodsFor: 'converting' stamp: 'di 1/16/98 16:33'!
contractTo: smallSize
	"return myself or a copy shortened by ellipsis to smallSize"
	| leftSize |
	self size <= smallSize
		ifTrue: [^ self].  "short enough"
	smallSize < 5
		ifTrue: [^ self copyFrom: 1 to: smallSize].    "First N characters"
	leftSize := smallSize-2//2.
	^ self copyReplaceFrom: leftSize+1		"First N/2 ... last N/2"
		to: self size - (smallSize - leftSize - 3)
		with: '...'
"
	'A clear but rather long-winded summary' contractTo: 18
"! !

!String methodsFor: 'converting' stamp: 'KR 1/30/2006 21:47'!
convertFromEncoding: encodingName
	^self convertFromWithConverter: (TextConverter newForEncoding: encodingName)! !

!String methodsFor: 'converting' stamp: 'KR 1/30/2006 21:47'!
convertFromSuperSwikiServerString
	^self convertFromEncoding: 'shift_jis'! !

!String methodsFor: 'converting' stamp: 'ul 3/13/2010 19:43'!
convertFromWithConverter: converter

	^(converter ifNil: [ ^self]) decodeString: self! !

!String methodsFor: 'converting' stamp: 'KR 1/30/2006 21:49'!
convertToEncoding: encodingName
	^self convertToWithConverter: (TextConverter newForEncoding: encodingName).! !

!String methodsFor: 'converting' stamp: 'KR 1/30/2006 21:50'!
convertToSuperSwikiServerString
	^self convertToEncoding: 'shift_jis'! !

!String methodsFor: 'converting' stamp: 'nice 11/22/2009 21:06'!
convertToSystemString
	^self convertToWithConverter: LanguageEnvironment defaultSystemConverter! !

!String methodsFor: 'converting' stamp: 'ul 3/13/2010 19:43'!
convertToWithConverter: converter 

	^(converter ifNil: [ ^self]) encodeString: self! !

!String methodsFor: 'copying' stamp: 'yo 11/3/2004 19:24'!
copyReplaceTokens: oldSubstring with: newSubstring 
	"Replace all occurrences of oldSubstring that are surrounded
	by non-alphanumeric characters"
	^ self copyReplaceAll: oldSubstring with: newSubstring asTokens: true
	"'File asFile Files File''s File' copyReplaceTokens: 'File' with: 'Snick'"! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
correctAgainstDictionary: wordDict continuedFrom: oldCollection
	"Like correctAgainst:continuedFrom:.  Use when you want to correct against a dictionary."

	^ wordDict isNil
		ifTrue: [ self correctAgainstEnumerator: nil
					continuedFrom: oldCollection ]
		ifFalse: [ self correctAgainstEnumerator: [ :action | wordDict keysDo: action ]
					continuedFrom: oldCollection ]! !

!String methodsFor: 'private' stamp: 'nice 12/27/2009 03:10'!
correctAgainstEnumerator: wordBlock continuedFrom: oldCollection
	"The guts of correction, instead of a wordList, there is a block that should take another block and enumerate over some list with it."

	| choices results maxChoices scoreMin |
	scoreMin := self size // 2 min: 3.
	maxChoices := 10.
	oldCollection isNil
		ifTrue: [ choices := SortedCollection sortBlock: [ :x :y | x value > y value ] ]
		ifFalse: [ choices := oldCollection ].
	wordBlock isNil
		ifTrue:
			[ results := OrderedCollection new.
			1 to: (maxChoices min: choices size) do: [ :i | results add: (choices at: i) key ] ]
		ifFalse:
			[ wordBlock value: [ :word | | score |
				(score := self alike: word) >= scoreMin ifTrue:
					[ choices add: (Association key: word value: score).
						(choices size >= maxChoices) ifTrue: [ scoreMin := (choices at: maxChoices) value] ] ].
			results := choices ].
	^ results! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
correctAgainst: wordList
	"Correct the receiver: assume it is a misspelled word and return the (maximum of five) nearest words in the wordList.  Depends on the scoring scheme of alike:"
	| results |
	results := self correctAgainst: wordList continuedFrom: nil.
	results := self correctAgainst: nil continuedFrom: results.
	^ results! !

!String methodsFor: 'converting' stamp: 'ar 5/1/1999 05:01'!
correctAgainst: wordList continuedFrom: oldCollection
	"Like correctAgainst:.  Use when you want to correct against several lists, give nil as the first oldCollection, and nil as the last wordList."

	^ wordList isNil
		ifTrue: [ self correctAgainstEnumerator: nil
					continuedFrom: oldCollection ]
		ifFalse: [ self correctAgainstEnumerator: [ :action | wordList do: action without: nil]
					continuedFrom: oldCollection ]! !

!String methodsFor: 'comparing' stamp: 'ar 4/10/2005 17:27'!
crc16
	"Compute a 16 bit cyclic redundancy check."

	| crc |
	crc := 0.
	1 to: self byteSize do: [:i |
		crc := (crc bitShift: -8) bitXor: (
		 #(	16r0000	16rC0C1	16rC181	16r0140	16rC301	16r03C0	16r0280	16rC241
			16rC601	16r06C0	16r0780	16rC741	16r0500	16rC5C1	16rC481	16r0440
			16rCC01	16r0CC0	16r0D80	16rCD41	16r0F00	16rCFC1	16rCE81	16r0E40
			16r0A00	16rCAC1	16rCB81	16r0B40	16rC901	16r09C0	16r0880	16rC841
			16rD801	16r18C0	16r1980	16rD941	16r1B00	16rDBC1	16rDA81	16r1A40
			16r1E00	16rDEC1	16rDF81	16r1F40	16rDD01	16r1DC0	16r1C80	16rDC41
			16r1400	16rD4C1	16rD581	16r1540	16rD701	16r17C0	16r1680	16rD641
			16rD201	16r12C0	16r1380	16rD341	16r1100	16rD1C1	16rD081	16r1040
			16rF001	16r30C0	16r3180	16rF141	16r3300	16rF3C1	16rF281	16r3240
			16r3600	16rF6C1	16rF781	16r3740	16rF501	16r35C0	16r3480	16rF441
			16r3C00	16rFCC1	16rFD81	16r3D40	16rFF01	16r3FC0	16r3E80	16rFE41
			16rFA01	16r3AC0	16r3B80	16rFB41	16r3900	16rF9C1	16rF881	16r3840
			16r2800	16rE8C1	16rE981	16r2940	16rEB01	16r2BC0	16r2A80	16rEA41
			16rEE01	16r2EC0	16r2F80	16rEF41	16r2D00	16rEDC1	16rEC81	16r2C40
			16rE401	16r24C0	16r2580	16rE541	16r2700	16rE7C1	16rE681	16r2640
			16r2200	16rE2C1	16rE381	16r2340	16rE101	16r21C0	16r2080	16rE041
			16rA001	16r60C0	16r6180	16rA141	16r6300	16rA3C1	16rA281	16r6240
			16r6600	16rA6C1	16rA781	16r6740	16rA501	16r65C0	16r6480	16rA441
			16r6C00	16rACC1	16rAD81	16r6D40	16rAF01	16r6FC0	16r6E80	16rAE41
			16rAA01	16r6AC0	16r6B80	16rAB41	16r6900	16rA9C1	16rA881	16r6840
			16r7800	16rB8C1	16rB981	16r7940	16rBB01	16r7BC0	16r7A80	16rBA41
			16rBE01	16r7EC0	16r7F80	16rBF41	16r7D00	16rBDC1	16rBC81	16r7C40
			16rB401	16r74C0	16r7580	16rB541	16r7700	16rB7C1	16rB681	16r7640
			16r7200	16rB2C1	16rB381	16r7340	16rB101	16r71C0	16r7080	16rB041
			16r5000	16r90C1	16r9181	16r5140	16r9301	16r53C0	16r5280	16r9241
			16r9601	16r56C0	16r5780	16r9741	16r5500	16r95C1	16r9481	16r5440
			16r9C01	16r5CC0	16r5D80	16r9D41	16r5F00	16r9FC1	16r9E81	16r5E40
			16r5A00	16r9AC1	16r9B81	16r5B40	16r9901	16r59C0	16r5880	16r9841
			16r8801	16r48C0	16r4980	16r8941	16r4B00	16r8BC1	16r8A81	16r4A40
			16r4E00	16r8EC1	16r8F81	16r4F40	16r8D01	16r4DC0	16r4C80	16r8C41
			16r4400	16r84C1	16r8581	16r4540	16r8701	16r47C0	16r4680	16r8641
			16r8201	16r42C0	16r4380	16r8341	16r4100	16r81C1	16r8081	16r4040)
			 at: ((crc bitXor: (self byteAt: i)) bitAnd: 16rFF) + 1) ].
	^crc! !

!String methodsFor: 'internet' stamp: 'yo 12/28/2003 01:17'!
decodeMimeHeader
	"See RFC 2047, MIME Part Three: Message Header Extension for Non-ASCII  
	Text. Text containing non-ASCII characters is encoded by the sequence  
	=?character-set?encoding?encoded-text?=  
	Encoding is Q (quoted printable) or B (Base64), handled by  
	Base64MimeConverter / RFC2047MimeConverter.

	Thanks to Yokokawa-san, it works in m17n package.  Try the following:

	'=?ISO-2022-JP?B?U1dJS0lQT1AvGyRCPUJDKyVpJXMlQRsoQi8=?= =?ISO-2022-JP?B?GyRCJVElRiUjJSobKEIoUGF0aW8p?=' decodeMimeHeader.
"
	| input output temp charset decoder encodedStream encoding pos |
	input := ReadStream on: self.
	output := WriteStream on: String new.
	[output
		nextPutAll: (input upTo: $=).
	"ASCII Text"
	input atEnd]
		whileFalse: [(temp := input next) = $?
				ifTrue: [charset := input upTo: $?.
					encoding := (input upTo: $?) asUppercase.
					temp := input upTo: $?.
					input next.
					"Skip final ="
					(charset isNil or: [charset size = 0]) ifTrue: [charset := 'LATIN-1'].
					encodedStream := MultiByteBinaryOrTextStream on: String new encoding: charset.
					decoder := encoding = 'B'
								ifTrue: [Base64MimeConverter new]
								ifFalse: [RFC2047MimeConverter new].
					decoder
						mimeStream: (ReadStream on: temp);
						 dataStream: encodedStream;
						 mimeDecode.
					output nextPutAll: encodedStream reset contents.
					pos := input position.
					input skipSeparators.
					"Delete spaces if followed by ="
					input peek = $=
						ifFalse: [input position: pos]]
				ifFalse: [output nextPut: $=;
						 nextPut: temp]].
	^ output contents! !

!String methodsFor: 'internet' stamp: 'yo 11/3/2004 19:24'!
decodeQuotedPrintable
	"Assume receiver is in MIME 'quoted-printable' encoding, and decode it."
  
	^QuotedPrintableMimeConverter mimeDecode: self as: self class! !

!String methodsFor: 'copying' stamp: 'yo 11/3/2004 19:24'!
deepCopy
	"DeepCopy would otherwise mean make a copy of the character;  since 
	characters are unique, just return a shallowCopy."

	^self shallowCopy! !

!String methodsFor: 'displaying' stamp: 'di 5/28/1998 13:25'!
displayAt: aPoint 
	"Display the receiver as a DisplayText at aPoint on the display screen."

	self displayOn: Display at: aPoint! !

!String methodsFor: 'displaying' stamp: 'yo 11/3/2004 19:24'!
displayOn: aDisplayMedium
	"Display the receiver on the given DisplayMedium.  5/16/96 sw"

	self displayOn: aDisplayMedium at: 0 @ 0! !

!String methodsFor: 'displaying' stamp: 'yo 11/3/2004 19:24'!
displayOn: aDisplayMedium at: aPoint 
	"Show a representation of the receiver as a DisplayText at location aPoint on aDisplayMedium, using black-colored text."

	self displayOn: aDisplayMedium at: aPoint textColor: Color black! !

!String methodsFor: 'displaying' stamp: 'yo 11/3/2004 19:24'!
displayOn: aDisplayMedium at: aPoint textColor: aColor
	"Show a representation of the receiver as a DisplayText at location aPoint on aDisplayMedium, rendering the text in the designated color"

	(self asDisplayText foregroundColor: (aColor ifNil: [Color black]) backgroundColor: Color white)
		displayOn: aDisplayMedium at: aPoint! !

!String methodsFor: 'accessing' stamp: 'yo 11/3/2004 19:24'!
do: aBlock toFieldNumber: aNumber
	"Considering the receiver as a holder of tab-delimited fields, evaluate aBlock on behalf of a field in this string"

	| start end index |
	start := 1.
	index := 1.
	[start <= self size] whileTrue: 
		[end := self indexOf: Character tab startingAt: start ifAbsent: [self size + 1].
		end := end - 1.
		aNumber = index ifTrue:
			[aBlock value: (self copyFrom: start  to: end).
			^ self].
		index := index + 1.
		start := end + 2]

"
1 to: 6 do:
	[:aNumber |
		'fred	charlie	elmo		wimpy	friml' do:
			[:aField | Transcript cr; show: aField] toFieldNumber: aNumber]
"! !

!String methodsFor: 'printing' stamp: 'yo 8/26/2002 22:57'!
encodeDoublingQuoteOn: aStream 
	"Print inside string quotes, doubling inbedded quotes."
	| x |
	aStream print: $'.
	1 to: self size do:
		[:i |
		aStream print: (x := self at: i).
		x = $' ifTrue: [aStream print: x]].
	aStream print: $'! !

!String methodsFor: 'converting' stamp: 'yo 7/29/2005 16:04'!
encodeForHTTP
	"change dangerous characters to their %XX form, for use in HTTP transactions"

	^ self encodeForHTTPWithTextEncoding: 'utf-8' conditionBlock: [:c | c isSafeForHTTP].
! !

!String methodsFor: 'converting' stamp: 'yo 7/29/2005 16:04'!
encodeForHTTPWithTextEncoding: encodingName

	^ self encodeForHTTPWithTextEncoding: encodingName conditionBlock: [:c | c isSafeForHTTP].
! !

!String methodsFor: 'converting' stamp: 'nice 12/27/2009 03:10'!
encodeForHTTPWithTextEncoding: encodingName conditionBlock: conditionBlock
	"change dangerous characters to their %XX form, for use in HTTP transactions"

	| httpSafeStream encodedStream |
	httpSafeStream := WriteStream on: (String new).
	encodedStream := MultiByteBinaryOrTextStream on: (String new: 6).
	encodedStream converter: (TextConverter newForEncoding: encodingName).
	self do: [:c | | cont |
		(conditionBlock value: c)
			ifTrue: [httpSafeStream nextPut: (Character value: c charCode)]
			ifFalse: [
				encodedStream text; resetToStart.
				encodedStream nextPut: c.
				encodedStream position: 0.
				encodedStream binary.
				cont := encodedStream contents.
				cont do: [:byte |
					httpSafeStream nextPut: $%.
					httpSafeStream nextPut: (byte // 16) asHexDigit.
					httpSafeStream nextPut: (byte \\ 16) asHexDigit.
				].
			].
	].
	^ httpSafeStream contents.
! !

!String methodsFor: 'testing' stamp: 'sw 10/20/2004 17:51'!
endsWithAColon 
	"Answer whether the final character of the receiver is a colon"

	^ self size > 0 and: [self last == $:]

"
#fred: endsWithAColon
'fred' endsWithAColon
"! !

!String methodsFor: 'testing' stamp: 'ar 11/19/1998 21:48'!
endsWithAnyOf: aCollection
	aCollection do:[:suffix|
		(self endsWith: suffix) ifTrue:[^true].
	].
	^false! !

!String methodsFor: 'testing' stamp: 'yo 11/3/2004 19:24'!
endsWithDigit
	"Answer whether the receiver's final character represents a digit.  3/11/96 sw"

	^ self size > 0 and: [self last isDigit]! !

!String methodsFor: 'testing' stamp: 'ul 4/3/2016 00:41'!
endsWith: sequence
	"Answer true if the receiver ends with the argument collection. The comparison is case-sensitive."
	
	| index sequenceSize offset |
	sequence isString ifFalse: [ ^ super endsWith: sequence ].
	((sequenceSize := sequence size) = 0 or: [ (offset := self size - sequenceSize) < 0 ]) ifTrue: [ ^false ].
	index := 0.
	[ (index := index + 1) <= sequenceSize ] whileTrue: [
		(sequence at: index) == (self at: index + offset) ifFalse: [ ^false ] ].
	^true! !

!String methodsFor: 'formatting' stamp: 'md 5/26/2005 13:34'!
expandMacros
	^self expandMacrosWithArguments: #()! !

!String methodsFor: 'formatting' stamp: 'ul 4/2/2015 22:58'!
expandMacrosWithArguments: anArray 
	
	^self class new: self size streamContents: [ :output |
		| lastIndex nextIndex |
		lastIndex := 1.
		[ (nextIndex := self indexOfAnyOf: CSMacroCharacters startingAt: lastIndex) = 0 ] whileFalse: [
			nextIndex = lastIndex ifFalse: [
				output next: nextIndex - lastIndex putAll: self startingAt: lastIndex ].
			(self at: nextIndex) == $% 
				ifTrue: [ output nextPut: (self at: (nextIndex := nextIndex + 1)) ]
				ifFalse: [ 
					| nextCharacter argumentIndex |
					nextCharacter := (self at: (nextIndex := nextIndex + 1)) asUppercase.
					nextCharacter == $N ifTrue: [ output cr ].
					nextCharacter == $T ifTrue: [ output tab ].
					(nextCharacter between: $0 and: $9) ifTrue: [
						argumentIndex := nextCharacter digitValue.
						[ (nextIndex := nextIndex + 1) <= self size and: [
							(nextCharacter := self at: nextIndex) between: $0 and: $9 ] ] whileTrue: [
								argumentIndex := argumentIndex * 10 + nextCharacter digitValue ].
						nextCharacter := nextCharacter asUppercase ].
					nextCharacter == $P ifTrue: [ output print: (anArray at: argumentIndex) ].
					nextCharacter == $S ifTrue: [ output nextPutAll: (anArray at: argumentIndex) ].
					nextCharacter == $? ifTrue: [ 
						| trueEnd falseEnd |
						trueEnd := self indexOf: $: startingAt: nextIndex + 1.
						falseEnd := self indexOf: $> startingAt: trueEnd + 1.
						(anArray at: argumentIndex)
							ifTrue: [ output next: trueEnd - nextIndex - 1 putAll: self startingAt: nextIndex + 1 ]
							ifFalse: [ output next: falseEnd - trueEnd - 1 putAll: self startingAt: trueEnd + 1 ].
						nextIndex := falseEnd - 1 ].
					(self at: (nextIndex := nextIndex + 1)) == $> ifFalse: [
						self error: '> expected' ] ].
			lastIndex := nextIndex + 1 ].
		lastIndex <= self size ifTrue: [
			output next: self size - lastIndex + 1 putAll: self startingAt: lastIndex ] ]! !

!String methodsFor: 'formatting' stamp: 'md 5/26/2005 13:34'!
expandMacrosWith: anObject 
	^self expandMacrosWithArguments: (Array with: anObject)! !

!String methodsFor: 'formatting' stamp: 'md 5/26/2005 13:34'!
expandMacrosWith: anObject with: anotherObject 
	^self 
		expandMacrosWithArguments: (Array with: anObject with: anotherObject)! !

!String methodsFor: 'formatting' stamp: 'md 5/26/2005 13:34'!
expandMacrosWith: anObject with: anotherObject with: thirdObject 
	^self expandMacrosWithArguments: (Array 
				with: anObject
				with: anotherObject
				with: thirdObject)! !

!String methodsFor: 'formatting' stamp: 'md 5/26/2005 13:34'!
expandMacrosWith: anObject with: anotherObject with: thirdObject with: fourthObject 
	^self expandMacrosWithArguments: (Array 
				with: anObject
				with: anotherObject
				with: thirdObject
				with: fourthObject)! !

!String methodsFor: 'accessing' stamp: 'nice 2/26/2010 00:53'!
findAnySubStr: delimiters startingAt: start 
	"Answer the index of the character within the receiver, starting at start, that begins a substring matching one of the delimiters.  delimiters is an Array of Strings (Characters are permitted also).  If the receiver does not contain any of the delimiters, answer size + 1."

	^delimiters inject: 1 + self size into: [:min :delim |
		"delim may be a char, a string of length 1, or a substring"
		| ind |
		ind := delim isCharacter 
			ifTrue: [self indexOf: delim
						startingAt: start ifAbsent: [min]]
			ifFalse: [self indexOfSubCollection: delim 
						startingAt: start ifAbsent: [min]].
		min min: ind]! !

!String methodsFor: 'accessing' stamp: 'dew 11/15/2000 15:02'!
findBetweenSubStrs: delimiters
	"Answer the collection of String tokens that result from parsing self.  Tokens are separated by 'delimiters', which can be a collection of Strings, or a collection of Characters.  Several delimiters in a row are considered as just one separation."

	| tokens keyStart keyStop |
	tokens := OrderedCollection new.
	keyStop := 1.
	[keyStop <= self size] whileTrue:
		[keyStart := self skipAnySubStr: delimiters startingAt: keyStop.
		keyStop := self findAnySubStr: delimiters startingAt: keyStart.
		keyStart < keyStop
			ifTrue: [tokens add: (self copyFrom: keyStart to: (keyStop - 1))]].
	^tokens! !

!String methodsFor: 'accessing' stamp: 'yo 11/3/2004 19:24'!
findCloseParenthesisFor: startIndex
	"assume (self at: startIndex) is $(.  Find the matching $), allowing parentheses to nest."
	" '(1+(2-3))-3.14159' findCloseParenthesisFor: 1 "
	" '(1+(2-3))-3.14159' findCloseParenthesisFor: 4 "
	| pos nestLevel |
	pos := startIndex+1.
	nestLevel := 1.
	[ pos <= self size ] whileTrue: [
		(self at: pos) = $( ifTrue: [ nestLevel := nestLevel + 1 ].
		(self at: pos) = $) ifTrue: [ nestLevel := nestLevel - 1 ].
		nestLevel = 0 ifTrue: [ ^pos ].
		pos := pos + 1.
	].
	^self size + 1! !

!String methodsFor: 'accessing' stamp: 'yo 11/3/2004 19:24'!
findDelimiters: delimiters startingAt: start 
	"Answer the index of the character within the receiver, starting at start, that matches one of the delimiters. If the receiver does not contain any of the delimiters, answer size + 1."

	start to: self size do: [:i |
		delimiters do: [:delim | delim = (self at: i) ifTrue: [^ i]]].
	^ self size + 1! !

!String methodsFor: 'accessing' stamp: 'nice 10/5/2009 03:15'!
findLastOccurrenceOfString: subString startingAt: start 
	"Answer the index of the last occurrence of subString within the receiver, starting at start. If 
	the receiver does not contain subString, answer 0.  Case-sensitive match used."

	| last now |
	last := self findString: subString startingAt: start.
	last = 0 ifTrue: [^ 0].
	[last > 0] whileTrue:
		[now := last.
		last := self findString: subString startingAt: last + 1].

	^ now
! !

!String methodsFor: 'converting' stamp: 'bf 3/2/2013 12:36'!
findSelector
	"Dan's code for hunting down selectors with keyword parts; while this doesn't give a true parse, in most cases it does what we want, in where it doesn't, we're none the worse for it."
	| sel possibleParens |
	sel := self withBlanksTrimmed.
	(sel includes: $:) ifTrue:
		[sel := sel copyReplaceAll: ':' with: ': '.	"for the style (aa max:bb) with no space"
		sel := sel copyReplaceAll: '[:' with: '[ :'.    "for the style ([:a) with no space"  
		possibleParens := sel findTokens: Character separators.
		sel := self class streamContents:
			[:s | | level |
			level := 0.
			possibleParens do:
				[:token |
				(level = 0 and: [token endsWith: ':'])
					ifTrue: [s nextPutAll: token]
					ifFalse: [level := level
							+ (token occurrencesOf: $() - (token occurrencesOf: $))
							+ (token occurrencesOf: $[) - (token occurrencesOf: $])
							+ (token occurrencesOf: ${) - (token occurrencesOf: $})]]]].
	sel isEmpty ifTrue: [^ nil].
	sel isOctetString ifTrue: [sel := sel asOctetString].
	Symbol hasInterned: sel ifTrue:
		[:aSymbol | ^ aSymbol].
	^ nil! !

!String methodsFor: 'accessing' stamp: 'ar 4/11/1999 22:15'!
findString: subString
	"Answer the index of subString within the receiver, starting at start. If 
	the receiver does not contain subString, answer 0."
	^self findString: subString startingAt: 1.! !

!String methodsFor: 'accessing' stamp: 'nice 7/27/2007 23:34'!
findString: subString startingAt: start 
	"Answer the index of subString within the receiver, starting at start. If 
	the receiver does not contain subString, answer 0."

	^self findString: subString startingAt: start caseSensitive: true! !

!String methodsFor: 'accessing' stamp: 'nice 7/27/2007 23:33'!
findString: key startingAt: start caseSensitive: caseSensitive 
	"Answer the index in this String at which the substring key first occurs,
	at or beyond start. The match can be case-sensitive or not. If no match
	is found, zero will be returned."
	
	"IMPLEMENTATION NOTE: do not use CaseSensitiveOrder because it is broken for WideString
	This is a temporary work around until Wide CaseSensitiveOrder search is fixed
	Code should revert to:
	caseSensitive
		ifTrue: [^ self findSubstring: key in: self startingAt: start matchTable: CaseSensitiveOrder]
		ifFalse: [^ self findSubstring: key in: self startingAt: start matchTable: CaseInsensitiveOrder]"
		
	^caseSensitive
		ifTrue: [
			(self class isBytes and: [key class isBytes])
				ifTrue: [self
						findSubstring: key
						in: self
						startingAt: start
						matchTable: CaseSensitiveOrder]
				ifFalse: [WideString new
						findSubstring: key
						in: self
						startingAt: start
						matchTable: nil]]
		ifFalse: [
			(self class isBytes and: [key class isBytes])
				ifTrue: [self
						findSubstring: key
						in: self
						startingAt: start
						matchTable: CaseInsensitiveOrder]
				ifFalse: [WideString new
						findSubstring: key
						in: self
						startingAt: start
						matchTable: CaseInsensitiveOrder]]! !

!String methodsFor: 'accessing' stamp: 'ul 4/3/2016 02:21'!
findSubstring: key in: body startingAt: start matchTable: matchTable
	"Answer the index in the string body at which the substring key first occurs, at or beyond start.  The match is determined using matchTable, which can be used to effect, eg, case-insensitive matches.  If no match is found, zero will be returned."

	| index c1 c2 keySize matchTableSize |
	(keySize := key size) = 0 ifTrue: [ ^0 ].
	matchTable ifNil: [
		start to: body size - key size + 1 do: [ :startIndex |
			index := 0.
			[ (body at: startIndex + index) == (key at: (index := index + 1)) ] whileTrue: [
				index = keySize ifTrue: [ ^startIndex ] ] ].
		^0 ].
	matchTableSize := matchTable size.
	start to: body size - key size + 1 do: [ :startIndex |
		index := 0.
		[
			(c1 := (body at: startIndex + index) asInteger + 1) <= matchTableSize ifTrue: [
				c1 := matchTable at: c1 ].
			(c2 := (key at: (index := index + 1)) asInteger + 1) <= matchTableSize ifTrue: [
				c2 := matchTable at: c2 ].
			c1 = c2 ]
			whileTrue: [
				index = keySize ifTrue: [ ^startIndex ] ] ].
	^0! !

!String methodsFor: 'accessing' stamp: 'ar 4/10/2005 17:13'!
findTokens: delimiters
	"Answer the collection of tokens that result from parsing self.  Return strings between the delimiters.  Any character in the Collection delimiters marks a border.  Several delimiters in a row are considered as just one separation.  Also, allow delimiters to be a single character."

	| tokens keyStart keyStop separators |

	tokens := OrderedCollection new.
	separators := delimiters isCharacter 
		ifTrue: [Array with: delimiters]
		ifFalse: [delimiters].
	keyStop := 1.
	[keyStop <= self size] whileTrue:
		[keyStart := self skipDelimiters: separators startingAt: keyStop.
		keyStop := self findDelimiters: separators startingAt: keyStart.
		keyStart < keyStop
			ifTrue: [tokens add: (self copyFrom: keyStart to: (keyStop - 1))]].
	^tokens! !

!String methodsFor: 'accessing' stamp: 'dtl 8/14/2004 11:27'!
findTokens: delimiters escapedBy: quoteDelimiters 
	"Answer a collection of Strings separated by the delimiters, where  
	delimiters is a Character or collection of characters. Two delimiters in a  
	row produce an empty string (compare this to #findTokens, which  
	treats sequential delimiters as one).  
	 
	The characters in quoteDelimiters are treated as quote characters, such  
	that any delimiter within a pair of matching quoteDelimiter characters  
	is treated literally, rather than as a delimiter.  
	 
	The quoteDelimiter characters may be escaped within a quoted string.  
	Two sequential quote characters within a quoted string are treated as  
	a single character.  
	 
	This method is useful for parsing comma separated variable strings for  
	spreadsheet import and export."

	| tokens rs activeEscapeCharacter ts char token delimiterChars quoteChars |
	delimiterChars := (delimiters isNil
				ifTrue: ['']
				ifFalse: [delimiters]) asString.
	quoteChars := (quoteDelimiters isNil
				ifTrue: ['']
				ifFalse: [quoteDelimiters]) asString.
	tokens := OrderedCollection new.
	rs := ReadStream on: self.
	activeEscapeCharacter := nil.
	ts := WriteStream on: ''.
	[rs atEnd]
		whileFalse: [char := rs next.
			activeEscapeCharacter isNil
				ifTrue: [(quoteChars includes: char)
						ifTrue: [activeEscapeCharacter := char]
						ifFalse: [(delimiterChars includes: char)
								ifTrue: [token := ts contents.
									tokens add: token.
									ts := WriteStream on: '']
								ifFalse: [ts nextPut: char]]]
				ifFalse: [char == activeEscapeCharacter
						ifTrue: [rs peek == activeEscapeCharacter
								ifTrue: [ts nextPut: rs next]
								ifFalse: [activeEscapeCharacter := nil]]
						ifFalse: [ts nextPut: char]]].
	token := ts contents.
	(tokens isEmpty and: [token isEmpty])
		ifFalse: [tokens add: token].
	^ tokens! !

!String methodsFor: 'accessing' stamp: 'ul 8/14/2015 18:04'!
findTokens: delimiters includes: subString
	"Divide self into pieces using delimiters.  Return the piece that includes subString anywhere in it.  Is case sensitive (say asLowercase to everything beforehand to make insensitive)."

^ (self findTokens: delimiters) 
	detect: [:str | (str includesSubstring: subString)] 
	ifNone: [nil]! !

!String methodsFor: 'accessing' stamp: 'tk 6/19/97 10:05'!
findTokens: delimiters keep: keepers
	"Answer the collection of tokens that result from parsing self.  The tokens are seperated by delimiters, any of a string of characters.  If a delimiter is also in keepers, make a token for it.  (Very useful for carriage return.  A sole return ends a line, but is also saved as a token so you can see where the line breaks were.)"

	| tokens keyStart keyStop |
	tokens := OrderedCollection new.
	keyStop := 1.
	[keyStop <= self size] whileTrue:
		[keyStart := self skipDelimiters: delimiters startingAt: keyStop.
		keyStop to: keyStart-1 do: [:ii | 
			(keepers includes: (self at: ii)) ifTrue: [
				tokens add: (self copyFrom: ii to: ii)]].	"Make this keeper be a token"
		keyStop := self findDelimiters: delimiters startingAt: keyStart.
		keyStart < keyStop
			ifTrue: [tokens add: (self copyFrom: keyStart to: (keyStop - 1))]].
	^tokens! !

!String methodsFor: 'accessing' stamp: 'nice 7/27/2007 23:13'!
findWordStart: key startingAt: start
	| ind |
	"HyperCard style searching.  Answer the index in self of the substring key, when that key is preceeded by a separator character.  Must occur at or beyond start.  The match is case-insensitive.  If no match is found, zero will be returned."

	ind := start.
	[ind := self findString: key startingAt: ind caseSensitive: false.
	ind = 0 ifTrue: [^ 0].	"not found"
	ind = 1 ifTrue: [^ 1].	"First char is the start of a word"
	(self at: ind-1) isSeparator] whileFalse: [ind := ind + 1].
	^ ind	"is a word start"! !

!String methodsFor: 'formatting' stamp: 'ul 4/3/2015 00:27'!
format: aCollection 
	"format the receiver with aCollection  
	 
	simplest example:  
	'foo {1} bar' format: {Date today}.
	 
	complete example:  
	'\{ \} \\ foo {1} bar {2}' format: {12. 'string'}. 
	"
	^self class new: self size * 11 // 10 "+10%" streamContents: [ :output |
		| lastIndex nextIndex |
		lastIndex := 1.
		[ (nextIndex := self indexOfAnyOf: FormatCharacterSet startingAt: lastIndex) = 0 ] whileFalse: [
			nextIndex = lastIndex ifFalse: [
				output next: nextIndex - lastIndex putAll: self startingAt: lastIndex ].
			(self at: nextIndex) == $\
				ifTrue: [ output nextPut: (self at: (nextIndex := nextIndex + 1)) ]
				ifFalse: [ "${"
					"Parse the index - a positive integer in base 10."
					| digitValue collectionIndex |
					collectionIndex := 0.
					[ (digitValue := self basicAt: (nextIndex := nextIndex + 1)) between: 48 "$0 asciiValue" and: 57 "$9 asciiValue" ] whileTrue: [
						collectionIndex := collectionIndex * 10 + digitValue - 48. "$0 asciiValue" ].
					digitValue =  125 "$} asciiValue" ifFalse: [ self error: '$} expected' ].
					output nextPutAll: (aCollection at: collectionIndex) asString ].
			lastIndex := nextIndex + 1 ].
		lastIndex <= self size ifTrue: [
			output next: self size - lastIndex + 1 putAll: self startingAt: lastIndex ] ]! !

!String methodsFor: 'comparing' stamp: 'md 2/16/2006 17:49'!
hash
	"#hash is implemented, because #= is implemented"
	"ar 4/10/2005: I had to change this to use ByteString hash as initial 
	hash in order to avoid having to rehash everything and yet compute
	the same hash for ByteString and WideString.
	md 16/10/2006: use identityHash as initialHash, as behavior hash will 
    use String hash (name) to have a better hash soon"
	^ self class stringHash: self initialHash: ByteString identityHash! !

!String methodsFor: 'comparing' stamp: 'ul 11/2/2009 03:01'!
hashWithInitialHash: initialHash
	
	^ self class stringHash: self initialHash: initialHash! !

!String methodsFor: 'comparing' stamp: 'yo 11/3/2004 19:24'!
howManyMatch: string 
	"Count the number of characters that match up in self and aString."
	| count shorterLength |
	
	count  :=  0 .
	shorterLength  :=  ((self size ) min: (string size ) ) .
	(1 to: shorterLength  do: [:index |
		 (((self at: index ) = (string at: index )  ) ifTrue: [count  :=  (count + 1 ) .
			]   ).
		]   ).
	^  count 
	
	! !

!String methodsFor: 'testing' stamp: 'ul 8/14/2015 17:57'!
includesSubstring: aString

	^(self findString: aString startingAt: 1) > 0! !

!String methodsFor: 'testing' stamp: 'di 11/15/1998 16:53'!
includesSubstring: aString caseSensitive: caseSensitive
	
	^ (self findString: aString startingAt: 1 caseSensitive: caseSensitive) > 0! !

!String methodsFor: 'testing' stamp: 'ar 4/10/2005 16:49'!
includesUnifiedCharacter
	^false! !

!String methodsFor: 'paragraph support' stamp: 'ul 4/1/2015 19:44'!
indentationIfBlank: aBlock
	"Answer the number of leading tabs in the receiver.  If there are
	 no visible characters, pass the number of tabs to aBlock and return its value."

	| leadingTabs nonTab nonTabIndex nonSepIndex lineEndIndex |
	nonTab := (CharacterSet with: Character tab) complement.
	nonTabIndex := self indexOfAnyOf: nonTab startingAt: 1.
	nonTabIndex = 0 ifTrue: [
		"Only made of tabs or empty"
		^aBlock value: self size].
	leadingTabs := nonTabIndex - 1.
	nonSepIndex := self indexOfAnyOf: CharacterSet nonSeparators startingAt: 1.
	nonSepIndex = 0 ifTrue: [
		"Only made of separators"
		^aBlock value: leadingTabs].
	lineEndIndex := self indexOfAnyOf: CharacterSet crlf startingAt: 1.
	(lineEndIndex between: 1 and: nonSepIndex) ifTrue: [
		"Only made of separators up to a line end"
		^aBlock value: leadingTabs].
	^leadingTabs! !

!String methodsFor: 'accessing' stamp: 'yo 11/3/2004 19:24'!
indexOfSubCollection: sub 
	#Collectn.
	"Added 2000/04/08 For ANSI <sequenceReadableCollection> protocol."
	^ self
		indexOfSubCollection: sub
		startingAt: 1
		ifAbsent: [0]! !

!String methodsFor: 'accessing' stamp: 'nice 7/27/2007 22:58'!
indexOfSubCollection: sub startingAt: start ifAbsent: exceptionBlock
	| index |
	index := self findString: sub startingAt: start.
	index = 0 ifTrue: [^ exceptionBlock value].
	^ index! !

!String methodsFor: 'accessing' stamp: 'ul 8/20/2015 21:06'!
indexOf: aCharacter

	^self indexOf: aCharacter startingAt: 1
! !

!String methodsFor: 'accessing' stamp: 'ar 4/12/2005 16:31'!
indexOf: aCharacter startingAt: start

	(aCharacter isCharacter) ifFalse: [^ 0].
	^ self class indexOfAscii: aCharacter asciiValue inString: self startingAt: start! !

!String methodsFor: 'accessing' stamp: 'ul 8/20/2015 21:05'!
indexOf: aCharacter  startingAt: start  ifAbsent: aBlock

	| index |
	(index := self indexOf: aCharacter startingAt: start) = 0 ifTrue: [ ^aBlock value ].
	^index! !

!String methodsFor: 'converting' stamp: 'ul 11/15/2010 10:07'!
initialIntegerOrNil
	"Answer the integer represented by the leading digits of the receiver, or nil if the receiver does not begin with a digit"
	| firstNonDigit |
	(self size = 0 or: [self first isDigit not]) ifTrue: [^ nil].
	firstNonDigit := (self findFirst: [:m | m isDigit not]).
	firstNonDigit = 0 ifTrue: [firstNonDigit := self size + 1].
	^ (self copyFrom: 1  to: (firstNonDigit - 1)) asNumber
"
'234Whoopie' initialIntegerOrNil
'wimpy' initialIntegerOrNil
'234' initialIntegerOrNil
'2N' initialIntegerOrNil
'2' initialIntegerOrNil
'  89Ten ' initialIntegerOrNil
'78 92' initialIntegerOrNil
"
! !

!String methodsFor: 'testing' stamp: 'djp 10/26/1999 21:25'!
isAllDigits
	"whether the receiver is composed entirely of digits"
	self do: [:c | c isDigit ifFalse: [^ false]].
	^ true! !

!String methodsFor: 'testing' stamp: 'ls 7/4/1998 19:17'!
isAllSeparators
	"whether the receiver is composed entirely of separators"
	self do: [ :c | c isSeparator ifFalse: [ ^false ] ].
	^true! !

!String methodsFor: 'testing' stamp: 'ul 2/11/2011 18:16'!
isAsciiString

	^self allSatisfy: [ :each | each asciiValue <= 127 ]! !

!String methodsFor: 'testing' stamp: 'ar 4/10/2005 16:23'!
isByteString
	"Answer whether the receiver is a ByteString"
	^false! !

!String methodsFor: 'testing' stamp: 'di 4/25/2000 12:32'!
isLiteral
	"Answer whether the receiver is a valid Smalltalk literal."

	^ true! !

!String methodsFor: 'testing' stamp: 'ar 4/10/2005 23:25'!
isOctetString
	"Answer whether the receiver can be represented as a byte string. 
	This is different from asking whether the receiver *is* a ByteString 
	(i.e., #isByteString)"
	1 to: self size do: [:pos |
		(self at: pos) asInteger >= 256 ifTrue: [^ false].
	].
	^ true.
! !

!String methodsFor: 'internet' stamp: 'ar 4/9/2005 22:16'!
isoToSqueak
	^self "no longer needed"! !

!String methodsFor: 'internet' stamp: 'nice 12/26/2009 22:13'!
isoToUtf8
	"Convert ISO 8559-1 to UTF-8"
	| s |
	s := WriteStream on: (String new: self size).
	self do: [:c |
		| v |
		v := c asciiValue.
		(v > 128)
			ifFalse: [s nextPut: c]
			ifTrue: [
				s nextPut: (192+(v >> 6)) asCharacter.
				s nextPut: (128+(v bitAnd: 63)) asCharacter]].
	^s contents. 
! !

!String methodsFor: 'testing' stamp: 'sma 6/15/2000 15:48'!
isString
	^ true! !

!String methodsFor: 'testing' stamp: 'ar 4/12/2005 19:52'!
isWideString
	"Answer whether the receiver is a WideString"
	^false! !

!String methodsFor: 'converting' stamp: 'nice 12/26/2009 22:14'!
keywords
	"Answer an array of the keywords that compose the receiver."
	| keywords |
	keywords := Array streamContents:
		[:kwds |
		| kwd |
		kwd := WriteStream on: (String new: 16).
		1 to: self size do:
			[:i |
			| char |
			kwd nextPut: (char := self at: i).
			char = $: ifTrue: 
					[kwds nextPut: kwd contents.
					kwd reset]].
		(kwd position = 0) ifFalse: [kwds nextPut: kwd contents]].
	(keywords size >= 1 and: [(keywords at: 1) = ':']) ifTrue:
		["Has an initial keyword, as in #:if:then:else:"
		keywords := keywords allButFirst].
	(keywords size >= 2 and: [(keywords at: keywords size - 1) = ':']) ifTrue:
		["Has a final keyword, as in #nextPut::andCR"
		keywords := keywords copyReplaceFrom: keywords size - 1
								to: keywords size with: {':' , keywords last}].
	^ keywords! !

!String methodsFor: 'accessing' stamp: 'sw 11/5/1998 17:41'!
lastSpacePosition
	"Answer the character position of the final space or other separator character in the receiver, and 0 if none"
	self size to: 1 by: -1 do:
		[:i | ((self at: i) isSeparator) ifTrue: [^ i]].
	^ 0

"
'fred the bear' lastSpacePosition
'ziggie' lastSpacePosition
'elvis ' lastSpacePosition
'wimpy  ' lastSpacePosition
'' lastSpacePosition
"! !

!String methodsFor: 'accessing' stamp: 'yo 12/17/2002 16:56'!
leadingCharRunLengthAt: index

	| leadingChar |
	leadingChar := (self at: index) leadingChar.
	index to: self size do: [:i |
		(self at: i) leadingChar ~= leadingChar ifTrue: [^ i - index].
	].
	^ self size - index + 1.
! !

!String methodsFor: 'accessing' stamp: 'nice 11/16/2009 12:32'!
lineCorrespondingToIndex: anIndex
	"Answer a string containing the line at the given character position."

	self lineIndicesDo: [:start :endWithoutDelimiters :end |
		anIndex <= end ifTrue: [^self copyFrom: start to: endWithoutDelimiters]].
	^''! !

!String methodsFor: 'accessing' stamp: 'nice 11/16/2009 13:23'!
lineCount
	"Answer the number of lines represented by the receiver, where every line delimiter CR, LF or CRLF pair adds one line."

	| lineCount |
	lineCount := 0.
	self lineIndicesDo: [:start :endWithoutDelimiters :end |
		lineCount := lineCount + 1].
	^lineCount! !

!String methodsFor: 'accessing' stamp: 'nice 1/19/2010 00:29'!
lineIndicesDo: aBlock
	"execute aBlock with 3 arguments for each line:
	- start index of line
	- end index of line without line delimiter
	- end index of line including line delimiter(s) CR, LF or CRLF"
	
	| cr lf start sz nextLF nextCR |
	start := 1.
	sz := self size.
	cr := Character cr.
	nextCR := self indexOf: cr startingAt: 1.
	lf := Character lf.
	nextLF := self indexOf: lf startingAt: 1.
	[ start <= sz ] whileTrue: [
		(nextLF = 0 and: [ nextCR = 0 ])
			ifTrue: [ "No more CR, nor LF, the string is over"
					aBlock value: start value: sz value: sz.
					^self ].
		(nextCR = 0 or: [ 0 < nextLF and: [ nextLF < nextCR ] ])
			ifTrue: [ "Found a LF"
					aBlock value: start value: nextLF - 1 value: nextLF.
					start := 1 + nextLF.
					nextLF := self indexOf: lf startingAt: start ]
			ifFalse: [ 1 + nextCR = nextLF
				ifTrue: [ "Found a CR-LF pair"
					aBlock value: start value: nextCR - 1 value: nextLF.
					start := 1 + nextLF.
					nextCR := self indexOf: cr startingAt: start.
					nextLF := self indexOf: lf startingAt: start ]
				ifFalse: [ "Found a CR"
					aBlock value: start value: nextCR - 1 value: nextCR.
					start := 1 + nextCR.
					nextCR := self indexOf: cr startingAt: start ]]]! !

!String methodsFor: 'accessing' stamp: 'nice 11/16/2009 13:23'!
lineNumber: anIndex
	"Answer a string containing the characters in the given line number."

	| lineCount |
	lineCount := 0.
	self lineIndicesDo: [:start :endWithoutDelimiters :end |
		(lineCount := lineCount + 1) = anIndex ifTrue: [^self copyFrom: start to: endWithoutDelimiters]].
	^nil! !

!String methodsFor: 'accessing' stamp: 'nice 6/11/2010 21:44'!
lines
	"Answer an array of lines composing this receiver without the line ending delimiters."

	^Array new: (self size // 60 max: 16) streamContents: [:lines |
		self linesDo: [:aLine | lines nextPut: aLine]]! !

!String methodsFor: 'accessing' stamp: 'nice 11/16/2009 12:26'!
linesDo: aBlock
	"Execute aBlock with each line in this string. The terminating line delimiters CR, LF or CRLF pairs are not included in what is passed to aBlock"

	self lineIndicesDo: [:start :endWithoutDelimiters :end |
		aBlock value: (self copyFrom: start to: endWithoutDelimiters)]! !

!String methodsFor: 'internet' stamp: 'ar 4/10/2005 15:58'!
macToSqueak
	"Convert the receiver from MacRoman to Squeak encoding"
	^ self collect: [:each | each macToSqueak]! !

!String methodsFor: 'comparing' stamp: 'ul 1/15/2011 11:48'!
match: text
	"Answer whether text matches the pattern in this string.
	Matching ignores upper/lower case differences.
	Where this string contains #, text may contain any character.
	Where this string contains *, text may contain any sequence of characters."

	^ self startingAt: 1 match: text startingAt: 1
"
	'*'			match: 'zort' true
	'*baz'		match: 'mobaz' true
	'*baz'		match: 'mobazo' false
	'*baz*'		match: 'mobazo' true
	'*baz*'		match: 'mozo' false
	'foo*'		match: 'foozo' true
	'foo*'		match: 'bozo' false
	'foo*baz'	match: 'foo23baz' true
	'foo*baz'	match: 'foobaz' true
	'foo*baz'	match: 'foo23bazo' false
	'foo'		match: 'Foo' true
	'foo*baz*zort' match: 'foobazort' false
	'foo*baz*zort' match: 'foobazzort' true
	'*foo#zort'	match: 'afoo3zortthenfoo3zort' true
	'*foo*zort'	match: 'afoodezortorfoo3zort' true
"! !

!String methodsFor: 'accessing' stamp: 'ul 3/15/2013 22:00'!
numArgs 
	"Answer either the number of arguments that the receiver would take if considered a selector.  Answer -1 if it couldn't be a selector. It is intended mostly for the assistance of spelling correction."

	| numColons index size |
	(size := self size) = 0 ifTrue: [ ^-1 ].
	index := 1.
	(self at: index) isSpecial ifTrue: [
		2 to: size do: [ :i | (self at: i) isSpecial ifFalse: [ ^-1 ] ].
		^1 ].
	self canBeToken ifFalse: [ ^-1 ].
	"Fast colon count"
	numColons := 0.
	[ 
		(self at: index) isLetter ifFalse: [ ^-1 ].
		(index := (self indexOf: $: startingAt: index) + 1) > 1 ifFalse: [ 
			numColons = 0 ifTrue: [ ^0 ].
			^-1 ].
		numColons := numColons + 1.
		index <= size ] whileTrue.
	^numColons! !

!String methodsFor: 'converting' stamp: 'sw 5/30/2000 14:58'!
numericSuffix
	^ self stemAndNumericSuffix last

"
'abc98' numericSuffix
'98abc' numericSuffix
"! !

!String methodsFor: 'enumerating' stamp: 'ul 3/8/2016 17:20'!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject. Optimized version."

	| tally |
	anObject isCharacter ifFalse: [ ^0 ].
	tally := 0.
	1 to: self size do: [ :index |
		(self at: index) == anObject ifTrue: [ tally := tally + 1 ] ].
	^tally! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
onlyLetters
	"answer the receiver with only letters"
	^ self select:[:each | each isLetter]! !

!String methodsFor: 'copying' stamp: 'ul 3/19/2016 22:22'!
padded: leftOrRight to: length with: aCharacter

	| result paddingLength |
	(paddingLength := length - self size) <= 0 ifTrue: [ ^self ].
	result := self species new: length.
	leftOrRight = #left ifTrue: [
		^result
			from: 1
				to: paddingLength
				put: aCharacter;
			replaceFrom: paddingLength + 1
				to: length
				with: self
				startingAt: 1;
			yourself ].
	leftOrRight = #right ifTrue: [
		^result
			replaceFrom: 1
				to: self size
				with: self
				startingAt: 1;
			from: self size + 1
				to: length
				put: aCharacter;
			yourself ]! !

!String methodsFor: 'printing' stamp: 'sd 7/8/2006 18:06'!
printOn: aStream 
	"Print inside string quotes, doubling inbedded quotes."
 
	self storeOn: aStream! !

!String methodsFor: 'filter streaming' stamp: 'yo 8/26/2002 22:31'!
putOn:aStream

	^aStream nextPutAll: self.
! !

!String methodsFor: 'private' stamp: 'yo 8/26/2002 22:53'!
replaceFrom: start to: stop with: replacement startingAt: repStart 
	"Primitive. This destructively replaces elements from start to stop in the receiver starting at index, repStart, in the collection, replacement. Answer the receiver. Range checks are performed in the primitive only. Optional. See Object documentation whatIsAPrimitive."
	<primitive: 105>
	super replaceFrom: start to: stop with: replacement startingAt: repStart! !

!String methodsFor: 'converting' stamp: 'sma 5/20/2000 17:17'!
romanNumber
	| value v1 v2 |
	value := v1 := v2 := 0.
	self reverseDo:
		[:each |
		v1 := #(1 5 10 50 100 500 1000) at: ('IVXLCDM' indexOf: each).
		v1 >= v2
			ifTrue: [value := value + v1]
			ifFalse: [value := value - v1].
		v2 := v1].
	^ value! !

!String methodsFor: 'comparing' stamp: 'ar 4/10/2005 17:35'!
sameAs: aString 
	"Answer whether the receiver sorts equal to aString. The 
	collation sequence is ascii with case differences ignored."
	^(self compare: aString caseSensitive: false) = 2! !

!String methodsFor: 'converting' stamp: 'ul 11/15/2010 10:07'!
sansPeriodSuffix
	"Return a copy of the receiver up to, but not including, the first period.  If the receiver's *first* character is a period, then just return the entire receiver. "

	| likely |
	likely := self copyUpTo: $..
	^ likely size = 0
		ifTrue:	[self]
		ifFalse:	[likely]! !

!String methodsFor: 'accessing' stamp: 'yo 8/28/2002 14:28'!
skipAnySubStr: delimiters startingAt: start 
	"Answer the index of the last character within the receiver, starting at start, that does NOT match one of the delimiters. delimiters is a Array of substrings (Characters also allowed).  If the receiver is all delimiters, answer size + 1."

	| any this ind ii |
	ii := start-1.
	[(ii := ii + 1) <= self size] whileTrue: [ "look for char that does not match"
		any := false.
		delimiters do: [:delim |
			delim isCharacter 
				ifTrue: [(self at: ii) == delim ifTrue: [any := true]]
				ifFalse: ["a substring"
					delim size > (self size - ii + 1) ifFalse: "Here's where the one-off error was."
						[ind := 0.
						this := true.
						delim do: [:dd | 
							dd == (self at: ii+ind) ifFalse: [this := false].
							ind := ind + 1].
						this ifTrue: [ii := ii + delim size - 1.  any := true]]
							ifTrue: [any := false] "if the delim is too big, it can't match"]].
		any ifFalse: [^ ii]].
	^ self size + 1! !

!String methodsFor: 'accessing' stamp: 'ul 12/7/2009 06:41'!
skipDelimiters: delimiters startingAt: start 
	"Answer the index of the character within the receiver, starting at start, that does NOT match one of the delimiters. If the receiver does not contain any of the delimiters, answer size + 1.  Assumes the delimiters to be a non-empty string."

	start to: self size do: [:i |
		(delimiters includes: (self at: i)) ifFalse: [ ^i ] ].
	^self size + 1! !

!String methodsFor: 'converting' stamp: 'yo 8/27/2002 11:13'!
splitInteger
	"Answer an array that is a splitting of self into a string and an integer.
	'43Sam' ==> #(43 'Sam').  'Try90' ==> #('Try' 90)
	BUT NOTE: 'Sam' ==> #('Sam' 0), and '90' ==> #('' 90)  ie, (<string> <integer>)."

	| pos |
	(pos := self findFirst: [:d | d isDigit not]) = 0 ifTrue: [^ Array with: '' with: self asNumber].
	self first isDigit ifTrue: [
		^ Array with: (self copyFrom: 1 to: pos - 1) asNumber 
				with: (self copyFrom: pos to: self size)].
	(pos := self findFirst: [:d | d isDigit]) = 0 ifTrue: [^ Array with: self with: 0].
	^ Array with: (self copyFrom: 1 to: pos - 1)
			with: (self copyFrom: pos to: self size) asNumber! !

!String methodsFor: 'internet' stamp: 'ar 4/9/2005 22:16'!
squeakToIso
	^self "no longer needed"! !

!String methodsFor: 'internet' stamp: 'ar 4/10/2005 15:55'!
squeakToMac
	"Convert the receiver from Squeak to MacRoman encoding"
	^ self collect: [:each | each squeakToMac]! !

!String methodsFor: 'converting' stamp: 'ul 3/26/2010 23:51'!
squeakToUtf8
	"Convert the receiver into a UTF8-encoded string"
	
	^self convertToWithConverter: UTF8TextConverter new.! !

!String methodsFor: 'comparing' stamp: 'ul 1/15/2011 13:54'!
startingAt: keyStart match: text startingAt: textStart
	"Answer whether text matches the pattern in this string.
	Matching ignores upper/lower case differences.
	Where this string contains #, text may contain any character.
	Where this string contains *, text may contain any sequence of characters."
	
	| anyMatch matchStart matchEnd i matchStr j ii jj |
	i := keyStart.
	j := textStart.
	
	"Process consecutive *s and #s at the beginning."
	anyMatch := false.
	[ i <= self size and: [
		(self at: i)
			caseOf: {
				[ $* ] -> [ 
					anyMatch := true.
					i := i + 1.
					true ].
				[ $# ] -> [
					i := i + 1.
					j := j + 1.
					true ] }
			otherwise: [ false ] ] ] whileTrue.
	i > self size ifTrue: [
		^j - 1 = text size or: [ "We reached the end by matching the character with a #."
			anyMatch and: [ j <= text size ] "Or there was a * before the end." ] ].
	matchStart := i.

	"Now determine the match string"
	matchEnd := self size.
	(ii := self indexOf: $* startingAt: matchStart) > 0 ifTrue: [ matchEnd := ii-1 ].
	(ii := self indexOf: $# startingAt: matchStart) > 0 ifTrue: [ matchEnd := matchEnd min: ii-1 ].
	matchStr := self copyFrom: matchStart to: matchEnd.

	"Now look for the match string"
	[jj := text findString: matchStr startingAt: j caseSensitive: false.
	anyMatch ifTrue: [jj > 0] ifFalse: [jj = j]]
		whileTrue:
		["Found matchStr at jj.  See if the rest matches..."
		(self startingAt: matchEnd+1 match: text startingAt: jj + matchStr size) ifTrue:
			[^ true "the rest matches -- success"].
		"The rest did not match."
		anyMatch ifFalse: [^ false].
		"Preceded by * -- try for a later match"
		j := j+1].
	^ false "Failed to find the match string"! !

!String methodsFor: 'testing' stamp: 'sw 4/23/1998 10:53'!
startsWithDigit
	"Answer whether the receiver's first character represents a digit"

	^ self size > 0 and: [self first isDigit]! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
stemAndNumericSuffix
	"Parse the receiver into a string-valued stem and a numeric-valued suffix.  6/7/96 sw"

	| stem suffix position |

	stem := self.
	suffix := 0.
	position := 1.
	[stem endsWithDigit and: [stem size > 1]] whileTrue:
		[suffix :=  stem last digitValue * position + suffix.
		position := position * 10.
		stem := stem copyFrom: 1 to: stem size - 1].
	^ Array with: stem with: suffix

"'Fred2305' stemAndNumericSuffix"! !

!String methodsFor: 'printing' stamp: 'yo 8/26/2002 22:58'!
storeOn: aStream 
	"Print inside string quotes, doubling inbedded quotes."
	| x |
	aStream nextPut: $'.
	1 to: self size do:
		[:i |
		aStream nextPut: (x := self at: i).
		x = $' ifTrue: [aStream nextPut: x]].
	aStream nextPut: $'! !

!String methodsFor: 'converting' stamp: 'md 5/26/2005 13:35'!
string
	^self! !

!String methodsFor: 'private' stamp: 'yo 8/28/2002 15:22'!
stringhash

	^ self hash.
! !

!String methodsFor: 'printing' stamp: 'yo 11/3/2004 19:24'!
stringRepresentation
	"Answer a string that represents the receiver.  For most objects this is simply its printString, but for strings themselves, it's themselves, to avoid the superfluous extra pair of quotes.  6/12/96 sw"

	^ self ! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
subStrings
	"Answer an array of the substrings that compose the receiver."
	#Collectn.
	"Added 2000/04/08 For ANSI <readableString> protocol."
	^ self substrings! !

!String methodsFor: 'converting' stamp: 'eem 2/3/2015 12:04'!
subStrings: separators 
	"Answer an array containing the substrings in the receiver separated 
	by the elements of separators."
	| char result sourceStream subString |
	#Collectn.
	"Changed 2000/04/08 For ANSI <readableString> protocol."
	(separators isString or:[separators allSatisfy: [:element | element isCharacter]]) ifFalse:
		[^ self error: 'separators must be Characters.'].
	sourceStream := ReadStream on: self.
	result := OrderedCollection new.
	subString := String new.
	[sourceStream atEnd]
		whileFalse: 
			[char := sourceStream next.
			(separators includes: char)
				ifTrue: [subString notEmpty
						ifTrue: 
							[result add: subString copy.
							subString := String new]]
				ifFalse: [subString := subString , (String with: char)]].
	subString notEmpty ifTrue: [result add: subString copy].
	^ result asArray! !

!String methodsFor: 'converting' stamp: 'ar 7/21/2007 21:50'!
substrings
	"Answer an array of the substrings that compose the receiver."
	| result end beginning |
	result := WriteStream on: (Array new: 10).
	end := 0.
	"find one substring each time through this loop"
	[ "find the beginning of the next substring"
	beginning := end+1.
	[beginning <= self size and:[(self at: beginning) isSeparator]]
		whileTrue:[beginning := beginning + 1].
	beginning <= self size] whileTrue: [
		"find the end"
		end := beginning.
		[end <= self size and:[(self at: end) isSeparator not]]
			whileTrue:[end := end + 1].
		end := end - 1.
		result nextPut: (self copyFrom: beginning to: end).
	].
	^result contents! !

!String methodsFor: 'converting' stamp: 'sw 11/25/96'!
surroundedBySingleQuotes
	"Answer the receiver with leading and trailing quotes.  "

	^ $' asString, self, $' asString! !

!String methodsFor: 'accessing' stamp: 'yo 11/3/2004 19:24'!
tabDelimitedFieldsDo: aBlock
	"Considering the receiver as a holder of tab-delimited fields, evaluate execute aBlock with each field in this string.  The separatilng tabs are not included in what is passed to aBlock"

	| start end |
	"No senders but was useful enough in earlier work that it's retained for the moment."
	start := 1.
	[start <= self size] whileTrue: 
		[end := self indexOf: Character tab startingAt: start ifAbsent: [self size + 1].
		end := end - 1.
		aBlock value: (self copyFrom: start  to: end).
		start := end + 2]

"
'fred	charlie	elmo		2' tabDelimitedFieldsDo: [:aField | Transcript cr; show: aField]
"! !

!String methodsFor: 'converting' stamp: 'yo 8/28/2002 15:14'!
translateFrom: start  to: stop  table: table
	"translate the characters in the string by the given table, in place"
	self class translate: self from: start to: stop table: table! !

!String methodsFor: 'converting' stamp: 'ls 8/18/1998 07:44'!
translateToLowercase
	"Translate all characters to lowercase, in place"

	self translateWith: LowercasingTable! !

!String methodsFor: 'converting' stamp: 'nice 11/1/2013 23:34'!
translateToUppercase
	"Translate all characters to uppercase, in place"

	self translateWith: UppercasingTable! !

!String methodsFor: 'converting' stamp: 'yo 8/28/2002 15:13'!
translateWith: table
	"translate the characters in the string by the given table, in place"
	^ self translateFrom: 1 to: self size table: table! !

!String methodsFor: 'converting' stamp: 'yo 11/3/2004 19:24'!
truncateTo: smallSize
	"return myself or a copy shortened to smallSize.  1/18/96 sw"

	^ self size <= smallSize
		ifTrue:
			[self]
		ifFalse:
			[self copyFrom: 1 to: smallSize]! !

!String methodsFor: 'converting' stamp: 'sw 9/2/1998 17:09'!
truncateWithElipsisTo: maxLength
	"Return myself or a copy suitably shortened but with elipsis added"

	^ self size <= maxLength
		ifTrue:
			[self]
		ifFalse:
			[(self copyFrom: 1 to: (maxLength - 3)), '...']


	"'truncateWithElipsisTo:' truncateWithElipsisTo: 20"! !

!String methodsFor: 'converting' stamp: 'ul 8/9/2011 13:31'!
unescapePercents
	"decode %xx form.  This is the opposite of #encodeForHTTP. Assume UTF-8 encoding by default."
	
	| unescaped |
	unescaped := self unescapePercentsRaw.
	^[ unescaped utf8ToSqueak ]
		on: Error
		do: [ unescaped ]! !

!String methodsFor: 'converting' stamp: 'ul 8/19/2012 01:45'!
unescapePercentsRaw
	"Decode myself if I'm percent-encoded, also replace + with space. Return self if the encoding is not valid."
	
	^String new: self size streamContents: [ :stream |
		| value1 value2 specialChars startIndex endIndex |
		specialChars := '+%' asCharacterSet.
		startIndex := 1.
		[ (endIndex := self indexOfAnyOf: specialChars startingAt: startIndex) > 0 ] whileTrue: [
			stream 	next: endIndex - startIndex putAll: self startingAt: startIndex.
			(self at: endIndex) == $%
				ifTrue: [
					endIndex + 2 <= self size ifFalse: [ ^self ].
					value1 := (self at: endIndex + 1) asUppercase digitValue.
					(value1 < 0 or: [ value1 > 15 ]) ifTrue: [ ^self ].
					value2 := (self at: endIndex + 2) asUppercase digitValue.
					(value2 < 0 or: [ value2 > 15 ]) ifTrue: [ ^self ].
					stream nextPut: (Character value: value1 * 16 + value2).
					startIndex := endIndex + 3 ]
				ifFalse: [ "$+"
					stream nextPut: Character space.
					startIndex := endIndex + 1 ] ].
		startIndex <= self size ifTrue: [
			stream next: self size + 1 - startIndex putAll: self startingAt: startIndex ] ]! !

!String methodsFor: 'converting' stamp: 'ul 8/9/2011 13:29'!
unescapePercentsWithTextConverter: aTextConverter
	"decode string including %XX form"

	| unescaped |
	unescaped := self unescapePercentsRaw.
	^[ unescaped convertFromWithConverter: aTextConverter ]
		on: Error
		do: ["the contents may be squeak-encoded"
			unescaped ]! !

!String methodsFor: 'converting' stamp: 'ul 8/9/2011 13:30'!
unescapePercentsWithTextEncoding: encodingName 
	"decode string including %XX form"
	
	| converter |
	converter := (TextConverter newForEncoding: encodingName)
		ifNil: [ TextConverter newForEncoding: nil ].
	^self unescapePercentsWithTextConverter: converter! !

!String methodsFor: 'converting' stamp: 'yo 8/27/2002 11:20'!
unparenthetically
	"If the receiver starts with (..( and ends with matching )..), strip them"

	| curr |
	curr := self.
	[((curr first = $() and: [curr last = $)])] whileTrue:
		[curr := curr copyFrom: 2 to: (curr size - 1)].

	^ curr

"

'((fred the bear))' unparenthetically

"
		! !

!String methodsFor: 'internet' stamp: 'ul 11/15/2010 10:07'!
utf8ToIso
	"Only UTF-8 characters that maps to 8-bit ISO-8559-1 values are converted. Others raises an error"
	| s i c v c2 v2 |
	s := WriteStream on: (String new: self size).
	
	i := 1.
	[i <= self size] whileTrue: [
		c := self at: i. i:=i+1.
		v := c asciiValue.
		(v > 128)
			ifFalse: [ s nextPut: c ]
			ifTrue: [((v bitAnd: 252) = 192)
				ifFalse: [self error: 'illegal UTF-8 ISO character']
				ifTrue: [
					(i > self size) ifTrue: [ self error: 'illegal end-of-string, expected 2nd byte of UTF-8'].
					c2 := self at: i. i:=i+1.
					v2 := c2 asciiValue.
					((v2 bitAnd: 192) = 128) ifFalse: [self error: 'illegal 2nd UTF-8 char']. 
					s nextPut: ((v2 bitAnd: 63) bitOr: ((v << 6) bitAnd: 192)) asCharacter]]].
	^s contents. 
! !

!String methodsFor: 'converting' stamp: 'ul 3/26/2010 23:51'!
utf8ToSqueak
	"Convert the receiver from a UTF8-encoded string"
	
	^self convertFromWithConverter: UTF8TextConverter new.! !

!String methodsFor: 'converting' stamp: 'nice 12/26/2009 22:15'!
withBlanksCondensed
	"Return a copy of the receiver with leading/trailing blanks removed
	 and consecutive white spaces condensed."

	^String streamContents: [:stream |
		| trimmed lastBlank |
		trimmed := self withBlanksTrimmed.
		lastBlank := false.
		trimmed do: [:c | (c isSeparator and: [lastBlank]) ifFalse: [stream nextPut: c].
			lastBlank := c isSeparator]].

	" ' abc  d   ' withBlanksCondensed"
! !

!String methodsFor: 'converting' stamp: 'ul 4/1/2015 19:44'!
withBlanksTrimmed
	"Return a copy of the receiver from which leading and trailing blanks have been trimmed."

	| first last |
	first := self indexOfAnyOf: CharacterSet nonSeparators startingAt: 1 ifAbsent: [0].
	first = 0 ifTrue: [ ^'' ].  "no non-separator character"
	last := self lastIndexOfAnyOf: CharacterSet nonSeparators startingAt: self size ifAbsent: [self size].
	(first = 1 and: [ last = self size ]) ifTrue: [ ^self copy ].
	^self
		copyFrom: first
		to: last
! !

!String methodsFor: 'formatting' stamp: 'yo 11/3/2004 19:24'!
withCRs
	"Return a copy of the receiver in which backslash (\) characters have been replaced with carriage returns."

	^ self collect: [ :c | c = $\ ifTrue: [ Character cr ] ifFalse: [ c ]].! !

!String methodsFor: 'converting' stamp: 'md 9/19/2004 15:19'!
withFirstCharacterDownshifted
	"Return a copy with the first letter downShifted"
	
	| answer |
	
	self ifEmpty: [^ self copy].
	answer := self copy.
	answer at: 1 put: (answer at: 1) asLowercase.
	^ answer. ! !

!String methodsFor: 'internet' stamp: 'ul 11/22/2010 13:20'!
withInternetLineEndings
	"change line endings from CR's and LF's to CRLF's.  This is probably in prepration for sending a string over the Internet"
	
	^self withLineEndings: String crlf! !

!String methodsFor: 'internet' stamp: 'ul 11/22/2010 13:19'!
withLineEndings: lineEndingString

	| stream |
	stream := nil.
	self lineIndicesDo: [ :start :endWithoutDelimiters :end |
		(stream isNil and: [ endWithoutDelimiters ~= end ]) ifTrue: [
			(self copyFrom: endWithoutDelimiters + 1 to: end) = lineEndingString ifFalse: [
				stream := WriteStream with: self copy.
				stream position: start - 1 ] ].
		stream ifNotNil: [
			stream next: endWithoutDelimiters - start + 1 putAll: self startingAt: start.
			endWithoutDelimiters = end ifFalse: [
				stream nextPutAll: lineEndingString ] ] ].
	^stream
		ifNil: [ self ]
		ifNotNil: [ 
			stream position = self size
				ifTrue: [ stream originalContents ]
				ifFalse: [ stream contents ] ]! !

!String methodsFor: 'converting' stamp: 'ul 4/1/2015 19:44'!
withNoLineLongerThan: aNumber
	"Answer a string with the same content as receiver, but rewrapped so that no line has more characters than the given number"
	aNumber isNumber not | (aNumber < 1) ifTrue: [self error: 'too narrow'].
	^self class
		new: self size * (aNumber + 1) // aNumber "provision for supplementary line breaks"
		streamContents: [ :stream |
			self lineIndicesDo: [ :start :endWithoutDelimiters :end |
				| pastEnd lineStart |
				pastEnd := endWithoutDelimiters + 1.
				"eliminate spaces at beginning of line"
				lineStart := (self indexOfAnyOf: CharacterSet nonSeparators startingAt: start ifAbsent: [pastEnd]) min: pastEnd.
				[| lineStop lineEnd spacePosition |
				lineEnd := lineStop  := lineStart + aNumber min: pastEnd..
				spacePosition := lineStart.
				[spacePosition < lineStop] whileTrue: [
					spacePosition := self indexOfAnyOf: CharacterSet separators startingAt: spacePosition + 1 ifAbsent: [pastEnd].
					spacePosition <= lineStop ifTrue: [lineEnd := spacePosition].
				].
				"split before space or before lineStop if no space"
				stream nextPutAll: (self copyFrom: lineStart to: lineEnd - 1).
				"eliminate spaces at beginning of next line"
				lineStart := self indexOfAnyOf: CharacterSet nonSeparators startingAt: lineEnd ifAbsent: [pastEnd].
				lineStart <= endWithoutDelimiters ]
					whileTrue: [stream cr].
				stream nextPutAll: (self copyFrom: pastEnd to: end) ] ]! !

!String methodsFor: 'converting' stamp: 'jmv 8/6/2009 09:17'!
withoutJustTrailingDigits
	"Answer the portion of the receiver that precedes any trailing series of digits.  If the receiver consists entirely of digits and blanks, return an empty string"
	| firstDigit |
	firstDigit := (self findFirst: [:m | m isDigit]).
	^ firstDigit > 0
		ifTrue:
			[(self copyFrom: 1 to: firstDigit-1) withoutTrailingBlanks]
		ifFalse:
			[self]

"
'Wh oopi e234' withoutJustTrailingDigits
'Wh oopi e 234' withoutJustTrailingDigits
"
! !

!String methodsFor: 'converting' stamp: 'md 10/5/2005 11:01'!
withoutLeadingBlanks
	
	"Return a copy of the receiver from which leading blanks have been
trimmed."

	
	| first |
	
	first := self findFirst: [:c | c isSeparator not ].

	first = 0 ifTrue: [^ ''].  
	
	"no non-separator character"
	
	^ self copyFrom: first to: self size

	
		
	" '    abc  d' withoutLeadingBlanks"
! !

!String methodsFor: 'converting' stamp: 'yo 8/27/2002 14:06'!
withoutLeadingDigits
	"Answer the portion of the receiver that follows any leading series of digits and blanks.  If the receiver consists entirely of digits and blanks, return an empty string"
	| firstNonDigit |
	firstNonDigit := (self findFirst: [:m | m isDigit not and: [m ~= $ ]]).
	^ firstNonDigit > 0
		ifTrue:
			[self copyFrom: firstNonDigit  to: self size]
		ifFalse:
			['']

"
'234Whoopie' withoutLeadingDigits
' 4321 BlastOff!!' withoutLeadingDigits
'wimpy' withoutLeadingDigits
'  89Ten ' withoutLeadingDigits
'78 92' withoutLeadingDigits
"
! !

!String methodsFor: 'converting' stamp: 'ul 4/22/2015 07:52'!
withoutLineEndings

	^self withLineEndings: ' '! !

!String methodsFor: 'internet' stamp: 'ul 11/7/2010 06:46'!
withoutQuoting
	"remove the initial and final quote marks, if present"
	"'''h''' withoutQuoting"
	| quote |
	self size < 2 ifTrue: [ ^self ].
	quote := self first.
	(quote = self last and: [ quote = $' or: [ quote = $" ] ])
		ifTrue: [ ^self copyFrom: 2 to: self size - 1 ]
		ifFalse: [ ^self ].! !

!String methodsFor: 'converting' stamp: 'bf 11/24/1998 19:58'!
withoutTrailingBlanks
	"Return a copy of the receiver from which trailing blanks have been trimmed."

	| last |
	last := self findLast: [:c | c isSeparator not].
	last = 0 ifTrue: [^ ''].  "no non-separator character"
	^ self copyFrom: 1 to: last

	" ' abc  d   ' withoutTrailingBlanks"
! !

!String methodsFor: 'converting' stamp: 'yo 8/27/2002 14:06'!
withoutTrailingDigits
	"Answer the portion of the receiver that precedes any trailing series of digits and blanks.  If the receiver consists entirely of digits and blanks, return an empty string"
	| firstDigit |
	firstDigit := (self findFirst: [:m | m isDigit or: [m = $ ]]).
	^ firstDigit > 0
		ifTrue:
			[self copyFrom: 1 to: firstDigit-1]
		ifFalse:
			[self]

"
'Whoopie234' withoutTrailingDigits
' 4321 BlastOff!!' withoutLeadingDigits
'wimpy' withoutLeadingDigits
'  89Ten ' withoutLeadingDigits
'78 92' withoutLeadingDigits
"
! !

!String methodsFor: 'converting' stamp: 'ul 4/1/2015 19:42'!
withSeparatorsCompacted
	"replace each sequences of whitespace by a single space character"
	"' test ' withSeparatorsCompacted = ' test '"
	"' test test' withSeparatorsCompacted = ' test test'"
	"'test test		' withSeparatorsCompacted = 'test test '"

	| out in next isSeparator |
	self isEmpty ifTrue: [^ self].

	out := WriteStream on: (String new: self size).
	in := self readStream.
	isSeparator := [:char | char asciiValue < 256
				and: [CharacterSet separators includes: char]].
	[in atEnd] whileFalse: [
		next := in next.
		(isSeparator value: next)
			ifTrue: [
				out nextPut: $ .
				[in atEnd or:
					[next := in next.
					(isSeparator value: next)
						ifTrue: [false]
						ifFalse: [out nextPut: next. true]]] whileFalse]
			ifFalse: [out nextPut: next]].
	^ out contents! !

!String methodsFor: 'internet' stamp: 'bf 7/4/2012 16:50'!
withSqueakLineEndings
	"Assume the string is textual, and that CR, LF, and CRLF are all valid line endings.
	Replace each occurence with a single CR."

	(self includes: Character lf) ifFalse: [ ^self ].
	(self includes: Character cr) ifFalse: [
		^self copy translateWith: String crLfExchangeTable ].
	^self withLineEndings: String cr! !

!String methodsFor: 'internet' stamp: 'bf 7/4/2012 16:50'!
withUnixLineEndings
	"Assume the string is textual, and that CR, LF, and CRLF are all valid line endings.
	Replace each occurence with a single LF."

	(self includes: Character cr) ifFalse: [ ^self ].
	(self includes: Character lf) ifFalse: [
		^self copy translateWith: String crLfExchangeTable ].
	^self withLineEndings: String lf! !

!String methodsFor: 'encoding' stamp: 'ar 4/10/2005 17:18'!
writeLeadingCharRunsOn: stream

	| runLength runValues runStart leadingChar |
	self isEmpty ifTrue: [^ self].

	runLength := OrderedCollection new.
	runValues := OrderedCollection new.
	runStart := 1.
	leadingChar := (self at: runStart) leadingChar.
	2 to: self size do: [:index |
		(self at: index) leadingChar = leadingChar ifFalse: [
			runValues add: leadingChar.
			runLength add: (index - runStart).
			leadingChar := (self at: index) leadingChar.
			runStart := index.
		].
	].
	runValues add: (self last) leadingChar.
	runLength add: self size + 1 -  runStart.

	stream nextPut: $(.
	runLength do: [:rr | rr printOn: stream. stream space].
	stream skip: -1; nextPut: $).
	runValues do: [:vv | vv printOn: stream. stream nextPut: $,].
	stream skip: -1.
! !

!String methodsFor: 'arithmetic' stamp: 'di 11/6/1998 14:03'!
* arg

	^ arg adaptToString: self andSend: #*! !

!String methodsFor: 'arithmetic' stamp: 'di 11/6/1998 14:03'!
+ arg

	^ arg adaptToString: self andSend: #+! !

!String methodsFor: 'converting' stamp: 'ar 12/27/2009 14:56'!
, anObject
	"Concatenate the argument to the receiver.
		Transcript cr; show: 'The value is: ', 3.
	"
	^ self copyReplaceFrom: self size + 1
		  to: self size
		  with: anObject asString! !

!String methodsFor: 'arithmetic' stamp: 'di 11/6/1998 14:03'!
- arg

	^ arg adaptToString: self andSend: #-! !

!String methodsFor: 'arithmetic' stamp: 'di 11/6/1998 14:03'!
/ arg

	^ arg adaptToString: self andSend: #/! !

!String methodsFor: 'arithmetic' stamp: 'di 11/6/1998 14:03'!
// arg

	^ arg adaptToString: self andSend: #//! !

!String methodsFor: 'comparing' stamp: 'ar 4/10/2005 16:47'!
< aString 
	"Answer whether the receiver sorts before aString.
	The collation order is simple ascii (with case differences)."
	^(self compare: aString caseSensitive: true) = 1! !

!String methodsFor: 'comparing' stamp: 'ar 4/10/2005 16:47'!
<= aString 
	"Answer whether the receiver sorts before or equal to aString.
	The collation order is simple ascii (with case differences)."
	^(self compare: aString caseSensitive: true) <= 2! !

!String methodsFor: 'sorting' stamp: 'ul 5/24/2015 04:14'!
<=> aCharacterArray
	"Return a collation order of -1, 0, or 1, indicating whether I should be collated before the receiver, am equal, or after.
	See also:  http://en.wikipedia.org/wiki/Spaceship_operator"

	aCharacterArray isString ifTrue: [ ^(self compare: aCharacterArray) - 2 ].
	self = aCharacterArray 	ifTrue: [ ^0 ].
	self < aCharacterArray 	ifTrue: [ ^-1 ].
	^1! !

!String methodsFor: 'comparing' stamp: 'ul 5/24/2015 04:18'!
= aString 
	"Answer whether the receiver sorts equally as aString.
	The collation order is simple ascii (with case differences)."
	
	self == aString ifTrue: [ ^true ].
	aString isString ifFalse: [ ^false ].
	self size = aString size ifFalse: [ ^false ].
	^ (self compare: self with: aString collated: AsciiOrder) = 2! !

!String methodsFor: 'comparing' stamp: 'ar 4/10/2005 16:48'!
> aString 
	"Answer whether the receiver sorts after aString.
	The collation order is simple ascii (with case differences)."
	^(self compare: aString caseSensitive: true) = 3! !

!String methodsFor: 'comparing' stamp: 'ar 4/10/2005 16:48'!
>= aString 
	"Answer whether the receiver sorts after or equal to aString.
	The collation order is simple ascii (with case differences)."
	^(self compare: aString caseSensitive: true) >= 2! !

!String methodsFor: 'arithmetic' stamp: 'di 11/6/1998 14:03'!
\\ arg

	^ arg adaptToString: self andSend: #\\! !

String subclass: #Symbol
	instanceVariableNames: ''
	classVariableNames: 'NewSymbols OneCharacterSymbols SymbolTable'
	poolDictionaries: ''
	category: 'Collections-Strings'!

!Symbol commentStamp: '<historical>' prior: 0!
I represent Strings that are created uniquely. Thus, someString asSymbol == someString asSymbol.!

Symbol variableByteSubclass: #ByteSymbol
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Strings'!

!ByteSymbol commentStamp: '<historical>' prior: 0!
This class represents the symbols containing 8bit characters.!

!ByteSymbol class methodsFor: 'primitives' stamp: 'ar 8/18/2005 13:46'!
findFirstInString: aString inSet: inclusionMap startingAt: start
	^ByteString findFirstInString: aString  inSet: inclusionMap startingAt: start! !

!ByteSymbol class methodsFor: 'primitives' stamp: 'ar 8/18/2005 13:46'!
indexOfAscii: anInteger inString: aString startingAt: start
	^ByteString indexOfAscii: anInteger inString: aString startingAt: start! !

!ByteSymbol class methodsFor: 'primitives' stamp: 'ar 8/18/2005 13:46'!
stringHash: aString initialHash: speciesHash
	^ByteString stringHash: aString initialHash: speciesHash! !

!ByteSymbol class methodsFor: 'primitives' stamp: 'ar 8/18/2005 13:47'!
translate: aString from: start  to: stop  table: table
	^ByteString translate: aString from: start  to: stop  table: table! !

!ByteSymbol methodsFor: 'converting' stamp: 'ar 4/10/2005 22:12'!
asByteArray
	| ba sz |
	sz := self byteSize.
	ba := ByteArray new: sz.
	ba replaceFrom: 1 to: sz with: self startingAt: 1.
	^ba! !

!ByteSymbol methodsFor: 'converting' stamp: 'ar 4/10/2005 22:12'!
asOctetString
	^ self! !

!ByteSymbol methodsFor: 'accessing' stamp: 'ar 4/10/2005 22:10'!
at: index 
	"Primitive. Answer the Character stored in the field of the receiver
	indexed by the argument. Fail if the index argument is not an Integer or
	is out of bounds. Essential. See Object documentation whatIsAPrimitive."

	<primitive: 63>
	^ Character value: (super at: index)! !

!ByteSymbol methodsFor: 'accessing' stamp: 'ar 4/10/2005 22:10'!
at: anInteger put: anObject 
	"You cannot modify the receiver."
	self errorNoModification! !

!ByteSymbol methodsFor: 'accessing' stamp: 'ar 4/10/2005 22:10'!
byteAt: index
	<primitive: 60>
	^(self at: index) asciiValue! !

!ByteSymbol methodsFor: 'accessing' stamp: 'ar 4/10/2005 22:10'!
byteAt: anInteger put: anObject 
	"You cannot modify the receiver."
	self errorNoModification! !

!ByteSymbol methodsFor: 'accessing' stamp: 'ar 4/10/2005 22:11'!
byteSize
	^self size! !

!ByteSymbol methodsFor: 'testing' stamp: 'ul 10/25/2012 01:32'!
canBeToken
	"Optimized version for the common case."
	
	| index |
	index := 0.
	[ (index := self findSubstring: '~' in: self startingAt: index + 1 matchTable: Tokenish) = 0 ]
		whileFalse: [
			(self at: index) == $_ ifFalse: [ ^false ] ].
	^true
		! !

!ByteSymbol methodsFor: 'comparing' stamp: 'ar 4/10/2005 22:14'!
findSubstring: key in: body startingAt: start matchTable: matchTable
	"Answer the index in the string body at which the substring key first occurs, at or beyond start.  The match is determined using matchTable, which can be used to effect, eg, case-insensitive matches.  If no match is found, zero will be returned."
	<primitive: 'primitiveFindSubstring' module: 'MiscPrimitivePlugin'>
	^super findSubstring: key in: body startingAt: start matchTable: matchTable! !

!ByteSymbol methodsFor: 'accessing' stamp: 'nice 12/9/2009 15:40'!
indexOfAnyOf: aCollection startingAt: start
	"Use double dispatching for speed"
	^aCollection findFirstInByteString: self startingAt: start! !

!ByteSymbol methodsFor: 'accessing' stamp: 'nice 12/9/2009 15:40'!
indexOfAnyOf: aCollection startingAt: start ifAbsent: aBlock
	"Use double dispatching for speed"
	| index |
	^(index := aCollection findFirstInByteString: self startingAt: start) = 0
		ifTrue: [aBlock value]
		ifFalse: [index]! !

!ByteSymbol methodsFor: 'testing' stamp: 'ar 4/10/2005 22:14'!
isByteString
	"Answer whether the receiver is a ByteString"
	^true! !

!ByteSymbol methodsFor: 'testing' stamp: 'ar 4/10/2005 22:14'!
isOctetString
	"Answer whether the receiver can be represented as a byte string. 
	This is different from asking whether the receiver *is* a ByteString 
	(i.e., #isByteString)"
	^ true.
! !

!ByteSymbol methodsFor: 'enumerating' stamp: 'ul 3/9/2016 21:51'!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject. Optimized version."

	| tally |
	anObject isCharacter ifFalse: [ ^0 ].
	anObject asInteger > 255 ifTrue: [ ^0 ].
	tally := 0.
	1 to: self size do: [ :index |
		(self at: index) == anObject ifTrue: [ tally := tally + 1 ] ].
	^tally! !

!ByteSymbol methodsFor: 'private' stamp: 'ar 4/11/2005 00:08'!
pvtAt: index put: aCharacter
	"Primitive. Store the Character in the field of the receiver indicated by
	the index. Fail if the index is not an Integer or is out of bounds, or if
	the argument is not a Character. Essential. See Object documentation
	whatIsAPrimitive."

	<primitive: 64>
	aCharacter isCharacter 
		ifFalse:[^self errorImproperStore].
	index isInteger
		ifTrue: [self errorSubscriptBounds: index]
		ifFalse: [self errorNonIntegerIndex]! !

!ByteSymbol methodsFor: 'accessing' stamp: 'ar 4/12/2005 17:51'!
species
	"Answer the preferred class for reconstructing the receiver."
	^ByteString
! !

!ByteSymbol methodsFor: 'private' stamp: 'ar 4/10/2005 23:02'!
string: aString
	1 to: aString size do: [:j | self pvtAt: j put: (aString at: j)].
	^self! !

!Symbol class methodsFor: 'access' stamp: 'ul 10/27/2014 00:23'!
allSymbols
	"Answer all interned symbols"
	
	| originalNewSymbols originalSymbolTable |
	originalNewSymbols := NewSymbols.
	originalSymbolTable := SymbolTable.
	^Array
		new: originalNewSymbols slowSize + originalSymbolTable slowSize
		streamContents:[ :stream |
			stream
				nextPutAll: originalNewSymbols;
				nextPutAll: originalSymbolTable ]
! !

!Symbol class methodsFor: 'class initialization' stamp: 'ul 10/27/2014 00:22'!
allSymbolTablesDo: aBlock
	
	| originalNewSymbols originalSymbolTable |
	originalNewSymbols := NewSymbols.
	originalSymbolTable := SymbolTable.
	originalNewSymbols do: aBlock.
	originalSymbolTable do: aBlock.! !

!Symbol class methodsFor: 'class initialization' stamp: 'ul 10/27/2014 00:22'!
allSymbolTablesDo: aBlock after: aSymbol

	| originalNewSymbols originalSymbolTable |
	originalNewSymbols := NewSymbols.
	originalSymbolTable := SymbolTable.
	(originalNewSymbols includes: aSymbol) 
		ifTrue: [
			originalNewSymbols do: aBlock after: aSymbol.
			originalSymbolTable do: aBlock after: aSymbol ]
		ifFalse: [
			originalSymbolTable do: aBlock after: aSymbol.
			originalNewSymbols do: aBlock after: aSymbol ]
	! !

!Symbol class methodsFor: 'class initialization' stamp: 'ul 10/27/2014 00:37'!
cleanUp
	"Flush caches"

	self condenseNewSymbols! !

!Symbol class methodsFor: 'class initialization' stamp: 'ul 10/27/2014 00:37'!
compactSymbolTable
	"Reduce the size of the symbol table so that it holds all existing symbols with 25% free space."

	self deprecated: 'Use #condenseNewSymbols instead'.
	self condenseNewSymbols! !

!Symbol class methodsFor: 'private' stamp: 'ul 10/27/2014 00:34'!
condenseNewSymbols
	"Move all symbols from NewSymbols to SymbolTable, and compact SymbolTable."

	| originalNewSymbols originalSymbolTable newNewSymbols newSymbolTable |
	originalNewSymbols := NewSymbols.
	originalSymbolTable := SymbolTable.
	newNewSymbols := WeakSet new.
	newSymbolTable := originalSymbolTable copy
		addAll: originalNewSymbols;
		compact;
		yourself.
	originalNewSymbols == NewSymbols ifFalse: [
		"Some other process has modified the symbols. Try again."
		^self condenseNewSymbols ].
	NewSymbols := newNewSymbols.
	SymbolTable := newSymbolTable! !

!Symbol class methodsFor: 'instance creation' stamp: ''!
findInterned:aString

	self hasInterned:aString ifTrue:[:symbol| ^symbol].
	^nil.! !

!Symbol class methodsFor: 'private' stamp: 'ar 4/10/2005 22:43'!
hasInterned: aString ifTrue: symBlock 
	"Answer with false if aString hasnt been interned (into a Symbol),  
	otherwise supply the symbol to symBlock and return true."

	| symbol |
	^ (symbol := self lookup: aString)
		ifNil: [false]
		ifNotNil: [symBlock value: symbol.
			true]! !

!Symbol class methodsFor: 'class initialization' stamp: 'nice 9/8/2013 22:19'!
initialize

	"Symbol initialize"

	Symbol rehash.
	Smalltalk addToShutDownList: self.
! !

!Symbol class methodsFor: 'instance creation' stamp: 'nice 9/8/2013 22:19'!
internCharacter: aCharacter
	^self intern: aCharacter asString! !

!Symbol class methodsFor: 'instance creation' stamp: 'ul 10/27/2014 00:45'!
intern: aStringOrSymbol 

	| originalNewSymbols |
	originalNewSymbols := NewSymbols.
	^(self lookup: aStringOrSymbol) ifNil:[
		| aClass aSymbol newNewSymbols |
		aStringOrSymbol isSymbol ifTrue:[
			aSymbol := aStringOrSymbol.
		] ifFalse:[
			aClass := aStringOrSymbol isOctetString ifTrue:[ByteSymbol] ifFalse:[WideSymbol].
			aSymbol := aClass new: aStringOrSymbol size.
			aSymbol string: aStringOrSymbol.
		].
		newNewSymbols := originalNewSymbols copyWith: aSymbol.
		originalNewSymbols == NewSymbols
			ifTrue: [
				NewSymbols := newNewSymbols.
				newNewSymbols size > 1000 ifTrue: [ self condenseNewSymbols ].
				aSymbol ]
			ifFalse: [
				"Some other process has modified the symbols. Try again."
				self intern: aStringOrSymbol ] ]! !

!Symbol class methodsFor: 'instance creation' stamp: 'ul 10/27/2014 00:24'!
lookup: aStringOrSymbol

	| originalNewSymbols originalSymbolTable |
	originalNewSymbols := NewSymbols.
	originalSymbolTable := SymbolTable.
	^(originalNewSymbols like: aStringOrSymbol) ifNil: [
		originalSymbolTable like: aStringOrSymbol ]! !

!Symbol class methodsFor: 'instance creation' stamp: ''!
newFrom: aCollection 
	"Answer an instance of me containing the same elements as aCollection."

	^ (aCollection as: String) asSymbol

"	Symbol newFrom: {$P. $e. $n}
	{$P. $e. $n} as: Symbol
"! !

!Symbol class methodsFor: 'private' stamp: 'nice 12/27/2009 04:53'!
possibleSelectorsFor: misspelled 
	"Answer an ordered collection of possible corrections
	for the misspelled selector in order of likelyhood"

	| numArgs candidates lookupString best binary short long first |
	lookupString := misspelled asLowercase. "correct uppercase selectors to lowercase"
	numArgs := lookupString numArgs.
	(numArgs < 0 or: [lookupString size < 2]) ifTrue: [^ OrderedCollection new: 0].
	first := lookupString first.
	short := lookupString size - (lookupString size // 4 max: 3) max: 2.
	long := lookupString size + (lookupString size // 4 max: 3).

	"First assemble candidates for detailed scoring"
	candidates := OrderedCollection new.
	self allSymbolTablesDo: [:s | | ss |
		(((ss := s size) >= short	"not too short"
			and: [ss <= long			"not too long"
					or: [(s at: 1) = first]])	"well, any length OK if starts w/same letter"
			and: [s numArgs = numArgs])	"and numArgs is the same"
			ifTrue: [candidates add: s]].

	"Then further prune these by correctAgainst:"
	best := lookupString correctAgainst: candidates.
	((misspelled last ~~ $:) and: [misspelled size > 1]) ifTrue: [
		binary := misspelled, ':'.		"try for missing colon"
		Symbol hasInterned: binary ifTrue: [:him | best addFirst: him]].
	^ best! !

!Symbol class methodsFor: 'instance creation' stamp: 'di 10/11/1999 00:02'!
readFrom: strm  "Symbol readFromString: '#abc'"

	strm peek = $# ifFalse: [self error: 'Symbols must be introduced by #'].
	^ (Scanner new scan: strm) advance  "Just do what the code scanner does"! !

!Symbol class methodsFor: 'private' stamp: 'ul 10/27/2014 00:35'!
rehash
	"Rebuild the hash table, reclaiming unreferenced Symbols. This method will intern all symbols. You're probably looking for #condenseNewSymbols instead."

	| originalNewSymbols originalSymbolTable newNewSymbols newSymbolTable |
	originalNewSymbols := NewSymbols.
	originalSymbolTable := SymbolTable.
	newNewSymbols := WeakSet new.
	newSymbolTable := WeakSet withAll: self allSubInstances.
	originalNewSymbols == NewSymbols ifFalse: [
		"Some other process has modified the symbols. Try again."
		^self rehash ].
	NewSymbols := newNewSymbols.
	SymbolTable := newSymbolTable! !

!Symbol class methodsFor: 'access' stamp: 'ul 9/4/2013 20:41'!
selectorsContaining: aString
	"Answer a list of selectors that contain aString within them. Case-insensitive. Does return symbols that begin with a capital letter."

	| size selectorList |
	selectorList := OrderedCollection new.
	(size := aString size) = 0 ifTrue: [ ^selectorList ].
	self allSymbolTablesDo: [ :each |
		(each size >= size
			and: [ (each includesSubstring: aString caseSensitive: false) 
			and: [ each numArgs ~= -1 ] ])
				ifTrue: [ selectorList add: each ] ].
	^selectorList

	"Symbol selectorsContaining: 'scon'"! !

!Symbol class methodsFor: 'access' stamp: 'ul 9/4/2013 20:39'!
selectorsMatching: aStringPattern
	"Answer a list of selectors that match aStringPattern within them. Case-insensitive. Does return symbols that begin with a capital letter."

	| selectorList |
	selectorList := OrderedCollection new.
	aStringPattern isEmpty ifTrue: [ ^selectorList ].
	self allSymbolTablesDo: [ :each | 
		((aStringPattern match: each) and: [ each numArgs ~= -1 ])
			ifTrue: [selectorList add: each ] ].
	^selectorList

	"Symbol selectorsMatching: 'parse:*'"! !

!Symbol class methodsFor: 'private' stamp: 'ul 10/27/2014 00:34'!
shutDown: aboutToQuit

	self condenseNewSymbols! !

!Symbol class methodsFor: 'access' stamp: 'tween 9/13/2004 10:09'!
thatStartsCaseSensitive: leadingCharacters skipping: skipSym
	"Same as thatStarts:skipping: but caseSensitive"
	| size firstMatch key |

	size := leadingCharacters size.
	size = 0 ifTrue: [^skipSym ifNil: [#''] ifNotNil: [nil]].
	firstMatch := leadingCharacters at: 1.
	size > 1 ifTrue: [key := leadingCharacters copyFrom: 2 to: size].
	self allSymbolTablesDo: [:each |
			each size >= size ifTrue:
				[
					((each at: 1) == firstMatch and:
						[key == nil or:
							[(each findString: key startingAt: 2 caseSensitive: true) = 2]])
								ifTrue: [^each]
				]
		] after: skipSym.

	^nil
! !

!Symbol class methodsFor: 'access' stamp: 'RAA 5/29/2001 14:35'!
thatStarts: leadingCharacters skipping: skipSym
	"Answer a selector symbol that starts with leadingCharacters.
	Symbols beginning with a lower-case letter handled directly here.
	Ignore case after first char.
	If skipSym is not nil, it is a previous answer; start searching after it.
	If no symbols are found, answer nil.
	Used by Alt-q (Command-q) routines"

	| size firstMatch key |

	size := leadingCharacters size.
	size = 0 ifTrue: [^skipSym ifNil: [#''] ifNotNil: [nil]].

	firstMatch := leadingCharacters at: 1.
	size > 1 ifTrue: [key := leadingCharacters copyFrom: 2 to: size].

	self allSymbolTablesDo: [:each |
			each size >= size ifTrue:
				[
					((each at: 1) == firstMatch and:
						[key == nil or:
							[(each findString: key startingAt: 2 caseSensitive: false) = 2]])
								ifTrue: [^each]
				]
		] after: skipSym.

	^nil

"Symbol thatStarts: 'sf' skipping: nil"
"Symbol thatStarts: 'sf' skipping: #sfpGetFile:with:with:with:with:with:with:with:with:"
"Symbol thatStarts: 'candidate' skipping: nil"
! !

!Symbol methodsFor: 'sorting' stamp: 'nice 3/12/2014 23:17'!
ascending
	"Return a SortFunction around the receiver, where the receiver will be used as a unary message to send to both a and b during sorting, and then the result of said send will be collated in ascending order using the <=> method."
	"Example: #('abc'  'de' 'fghi') sorted: #size asscending"

	^SortFunction ascend sendMessage: self! !

!Symbol methodsFor: 'user interface' stamp: 'sma 11/12/2000 11:46'!
asExplorerString
	^ self printString! !

!Symbol methodsFor: 'converting' stamp: 'cmm 7/18/2013 10:29'!
asMutator
	"Return a setter message from a getter message. For example, #name asMutator returns #name:"
	^ self last = $:
		ifTrue: [ self ]
		ifFalse: [ (self copyWith: $:) asSymbol ]! !

!Symbol methodsFor: 'sorting' stamp: 'nice 3/12/2014 23:17'!
asSortFunction

	^self ascending! !

!Symbol methodsFor: 'converting' stamp: 'ar 4/10/2005 22:42'!
asString 
	"Refer to the comment in String|asString."
	| newString |
	newString := self species new: self size.
	newString replaceFrom: 1 to: newString size with: self startingAt: 1.
	^newString! !

!Symbol methodsFor: 'converting' stamp: ''!
asSymbol 
	"Refer to the comment in String|asSymbol."! !

!Symbol methodsFor: 'accessing' stamp: ''!
at: anInteger put: anObject 
	"You cannot modify the receiver."

	self errorNoModification! !

!Symbol methodsFor: 'filter streaming' stamp: 'mpw 1/1/1901 00:20'!
byteEncode:aStream
	^aStream writeSymbol:self.
! !

!Symbol methodsFor: 'testing' stamp: 'cwp 10/24/2012 14:42'!
canBeToken
	"Since definition of #tokenish depends on a preference, we want to make sure
	that underscores are always considered tokenish. This is so that selectors created
	when the preference was turned on don't suddenly become invalid when the
	preference is turned off."
	
	^ self allSatisfy: [:c | c = $_ or: [c tokenish]]! !

!Symbol methodsFor: 'converting' stamp: 'sw 1/28/98 18:18'!
capitalized
	^ self asString capitalized asSymbol! !

!Symbol methodsFor: 'copying' stamp: 'tk 6/26/1998 11:35'!
clone
	"Answer with the receiver, because Symbols are unique."! !

!Symbol methodsFor: 'copying' stamp: ''!
copy
	"Answer with the receiver, because Symbols are unique."! !

!Symbol methodsFor: 'evaluating' stamp: 'ul 11/2/2010 18:23'!
cull: anObject 

	^anObject perform: self! !

!Symbol methodsFor: 'sorting' stamp: 'nice 3/12/2014 23:17'!
descending
	"Return a SortFunction around the receiver, where the receiver will be used as a unary message to send to both a and b during sorting, and then the result of said send will be collated in descending order using the <=> method."
	"Example: #('abc'  'de' 'fghi') sorted: #size descending"

	^SortFunction descend sendMessage: self! !

!Symbol methodsFor: 'private' stamp: ''!
errorNoModification

	self error: 'symbols can not be modified.'! !

!Symbol methodsFor: 'system primitives' stamp: 'di 1/2/1999 17:00'!
flushCache
	"Tell the interpreter to remove all entries with this symbol as a selector from its method lookup cache, if it has one.  This primitive must be called whenever a method is defined or removed.
	NOTE:  Only one of the two selective flush methods needs to be used.
	Squeak 2.3 and later uses 116 (See CompiledMethod flushCache)."

	<primitive: 119>
! !

!Symbol methodsFor: 'testing' stamp: 'md 1/20/2006 16:16'!
includesKey: sym
	^self == sym.! !

!Symbol methodsFor: 'testing' stamp: 'fbs 8/26/2012 20:32'!
isBinary
	^ self isInfix.! !

!Symbol methodsFor: 'testing' stamp: 'md 8/27/2005 16:33'!
isDoIt

	^ (self == #DoIt) or: [self == #DoItIn:].! !

!Symbol methodsFor: 'testing' stamp: 'ul 11/15/2010 10:07'!
isInfix
	"Answer whether the receiver is an infix message selector."

	^ self precedence = 2! !

!Symbol methodsFor: 'testing' stamp: 'ul 11/15/2010 10:07'!
isKeyword
	"Answer whether the receiver is a message keyword."

	^ self precedence = 3! !

!Symbol methodsFor: 'printing' stamp: 'sw 8/19/1999 11:30'!
isOrientedFill
	"Needs to be implemented here because symbols can occupy 'color' slots of morphs."

	^ false! !

!Symbol methodsFor: 'testing' stamp: 'sma 2/5/2000 12:13'!
isPvtSelector
	"Answer whether the receiver is a private message selector, that is,
	begins with 'pvt' followed by an uppercase letter, e.g. pvtStringhash."

	^ (self beginsWith: 'pvt') and: [self size >= 4 and: [(self at: 4) isUppercase]]! !

!Symbol methodsFor: 'testing' stamp: 'md 4/30/2003 15:31'!
isSymbol
	^ true ! !

!Symbol methodsFor: 'testing' stamp: 'ul 11/15/2010 10:07'!
isUnary
	"Answer whether the receiver is an unary message selector."

	^ self precedence = 1! !

!Symbol methodsFor: 'system primitives' stamp: 'ul 5/1/2015 13:28'!
numArgs: n
	"Answer a string that can be used as a selector with n arguments.
	 TODO: need to be extended to support shrinking and for selectors like #+ " 

	| numArgs offset |.
	(numArgs := self numArgs) >= n ifTrue: [ ^self ].
	numArgs = 0
		ifTrue: [ offset := 1 ]
		ifFalse: [ offset := 0 ].
	^(String new: n - numArgs + offset * 5 + offset + self size streamContents: [ :stream |
		stream nextPutAll: self.
		numArgs = 0 ifTrue: [ stream nextPut: $:. ].
		numArgs + offset + 1 to: n do: [ :i | stream nextPutAll: 'with:' ] ]) asSymbol! !

!Symbol methodsFor: 'accessing' stamp: 'cwp 8/14/2012 10:28'!
precedence
	"Answer the receiver's precedence, assuming it is a valid Smalltalk
	message selector or 0 otherwise.  The numbers are 1 for unary,
	2 for binary and 3 for keyword selectors."

	self size = 0 ifTrue: [ ^ 0 ].
	self first canBeIdentifierInitial ifFalse: [ ^ 2 ].
	self last = $: ifTrue: [ ^ 3 ].
	^ 1! !

!Symbol methodsFor: 'accessing' stamp: ''!
replaceFrom: start to: stop with: replacement startingAt: repStart

	self errorNoModification! !

!Symbol methodsFor: 'copying' stamp: ''!
shallowCopy
	"Answer with the receiver, because Symbols are unique."! !

!Symbol methodsFor: 'printing' stamp: 'di 4/25/2000 12:32'!
storeOn: aStream 

	aStream nextPut: $#.
	(Scanner isLiteralSymbol: self)
		ifTrue: [aStream nextPutAll: self]
		ifFalse: [super storeOn: aStream]! !

!Symbol methodsFor: 'private' stamp: ''!
string: aString

	1 to: aString size do: [:j | super at: j put: (aString at: j)].
	^self  ! !

!Symbol methodsFor: 'evaluating' stamp: 'md 3/24/2006 12:09'!
value: anObject 
	^anObject perform: self.! !

!Symbol methodsFor: 'evaluating' stamp: 'ul 11/4/2009 09:51'!
value: anObject value: anotherObject

	^anObject perform: self with: anotherObject! !

!Symbol methodsFor: 'copying' stamp: 'cmm 8/13/2013 15:02'!
veryDeepCopy
	"Overridden for performance."
	^ self! !

!Symbol methodsFor: 'copying' stamp: 'tk 8/19/1998 16:05'!
veryDeepCopyWith: deepCopier
	"Return self.  I am immutable in the Morphic world.  Do not record me."! !

!Symbol methodsFor: 'converting' stamp: 'md 8/10/2004 10:54'!
withFirstCharacterDownshifted
	"Answer an object like the receiver but with first character downshifted if necesary"

	^self asString withFirstCharacterDownshifted asSymbol.! !

!Symbol methodsFor: 'comparing' stamp: 'ar 4/10/2005 23:45'!
= aSymbol
	"Compare the receiver and aSymbol." 
	self == aSymbol ifTrue: [^ true].
	self class == aSymbol class ifTrue: [^ false].
	"Use String comparison otherwise"
	^ super = aSymbol! !

Symbol variableWordSubclass: #WideSymbol
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Strings'!

!WideSymbol commentStamp: '<historical>' prior: 0!
This class represents the symbols containing 32bit characters.!

!WideSymbol class methodsFor: 'class initialization' stamp: 'ar 4/10/2005 23:55'!
initialize
	Smalltalk removeFromShutDownList: self. "@@@ Remove this later @@@"! !

!WideSymbol methodsFor: 'accessing' stamp: 'ar 4/10/2005 23:38'!
at: index 
	"Answer the Character stored in the field of the receiver indexed by the argument."
	^ Character value: (self wordAt: index).
! !

!WideSymbol methodsFor: 'accessing' stamp: 'yo 8/30/2002 14:15'!
at: anInteger put: anObject 
	"You cannot modify the receiver."

	self errorNoModification
! !

!WideSymbol methodsFor: 'accessing' stamp: 'ar 4/10/2005 23:38'!
byteAt: index

	| d r |
	d := (index + 3) // 4.
	r := (index - 1) \\ 4 + 1.
	^ (self wordAt: d) digitAt: ((4 - r) + 1).
! !

!WideSymbol methodsFor: 'accessing' stamp: 'ar 4/10/2005 23:38'!
byteAt: index put: aByte
	self errorNoModification.! !

!WideSymbol methodsFor: 'accessing' stamp: 'ar 4/10/2005 23:38'!
byteSize

	^ self size * 4.
! !

!WideSymbol methodsFor: 'private' stamp: 'ar 4/12/2005 14:12'!
fixUponLoad: aProject seg: anImageSegment
	"We are in an old project that is being loaded from disk. 
	Fix up conventions that have changed."
	| ms |
	"Yoshiki did not put MultiSymbols into outPointers in older 
images!!
	When all old images are gone, remove this method."
	ms := Symbol intern: self asString.
	self == ms ifFalse: [
		"For a project from older m17n image, this is necessary."
		self becomeForward: ms.
		aProject projectParameters at: #MultiSymbolInWrongPlace put: true
	].

	"MultiString>>capitalized was not implemented 
correctly. 
	Fix eventual accessors and mutators here."
	((self beginsWith: 'get')
		and:[(self at: 4) asInteger < 256
		and:[(self at: 4) isLowercase]]) ifTrue:[
			ms := self asString.
			ms at: 4 put: (ms at: 4) asUppercase.
			ms := ms asSymbol.
			self becomeForward: ms.
			aProject projectParameters at: #MultiSymbolInWrongPlace put: true.
		].
	((self beginsWith: 'set')
		and:[(self at: 4) asInteger < 256
		and:[(self at: 4) isLowercase
		and:[self last = $:
		and:[(self occurrencesOf: $:) = 1]]]]) ifTrue:[
			ms := self asString.
			ms at: 4 put: (ms at: 4) asUppercase.
			ms := ms asSymbol.
			self becomeForward: ms.
			aProject projectParameters at: #MultiSymbolInWrongPlace put: true.
		].
	^ super fixUponLoad: aProject seg: anImageSegment	"me, 
not the label"
! !

!WideSymbol methodsFor: 'testing' stamp: 'ar 4/12/2005 19:52'!
isWideString
	"Answer whether the receiver is a WideString"
	^true! !

!WideSymbol methodsFor: 'private' stamp: 'yo 7/29/2005 21:53'!
mutateJISX0208StringToUnicode

	| c |
	1 to: self size do: [:i |
		c := self at: i.
		(c leadingChar = JISX0208 leadingChar or: [
			c leadingChar = (JISX0208 leadingChar bitShift: 2)]) ifTrue: [
			self basicAt: i put: (Character leadingChar: JapaneseEnvironment leadingChar code: (c asUnicode)) asciiValue.
		]
	].
! !

!WideSymbol methodsFor: 'private' stamp: 'ar 4/11/2005 00:09'!
pvtAt: index put: aCharacter
	"Primitive. Store the Character in the field of the receiver indicated by
	the index. Fail if the index is not an Integer or is out of bounds, or if
	the argument is not a Character. Essential. See Object documentation
	whatIsAPrimitive."

	<primitive: 61>
	index isInteger
		ifTrue: [self errorSubscriptBounds: index]
		ifFalse: [self errorNonIntegerIndex]! !

!WideSymbol methodsFor: 'accessing' stamp: 'ar 4/12/2005 17:34'!
species
	"Answer the preferred class for reconstructing the receiver."
	^WideString
! !

!WideSymbol methodsFor: 'private' stamp: 'ar 4/10/2005 23:58'!
string: aString
	1 to: aString size do: [:j | self pvtAt: j put: (aString at: j) asInteger].
	^self! !

!WideSymbol methodsFor: 'accessing' stamp: 'ar 4/10/2005 23:38'!
wordAt: index
	<primitive: 60>
	^ (self basicAt: index).
! !

!WideSymbol methodsFor: 'accessing' stamp: 'ar 4/10/2005 23:39'!
wordAt: index put: anInteger
	self errorNoModification.! !

String variableWordSubclass: #WideString
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Strings'!

!WideString commentStamp: 'yo 10/19/2004 22:34' prior: 0!
This class represents the array of 32 bit wide characters.
!

!WideString class methodsFor: 'enumeration' stamp: 'ul 7/25/2011 02:00'!
allMultiStringMethods  

	^CurrentReadOnlySourceFiles cacheDuring: [
		self systemNavigation allMethodsSelect: [ :method |
			method getSource asString isOctetString not ] ]! !

!WideString class methodsFor: 'enumeration' stamp: 'ul 7/25/2011 02:00'!
allNonAsciiMethods

	^CurrentReadOnlySourceFiles cacheDuring: [
		self systemNavigation allMethodsSelect: [ :method |
			method getSource asString isAsciiString not ] ]! !

!WideString class methodsFor: 'instance creation' stamp: 'ar 4/10/2005 19:38'!
fromByteArray: aByteArray 

	| inst |
	aByteArray size \\ 4 = 0 ifFalse: [^ ByteString fromByteArray: aByteArray ].
	inst := self new: aByteArray size // 4.
	4 to: aByteArray size by: 4 do: [:i |
		inst basicAt: i // 4
			put: ((aByteArray at: i - 3) << 24) + 
				((aByteArray at: i - 2) << 16) +
				 ((aByteArray at: i - 1) << 8) +
				(aByteArray at: i)
	].

	^ inst
! !

!WideString class methodsFor: 'instance creation' stamp: 'yo 8/30/2002 17:00'!
fromISO2022JPString: string 

	| tempFileName stream contents |
	tempFileName := Time millisecondClockValue printString , '.txt'.
	FileDirectory default deleteFileNamed: tempFileName ifAbsent: [].
	stream := StandardFileStream fileNamed: tempFileName.
	[stream nextPutAll: string]
		ensure: [stream close].
	stream := FileStream fileNamed: tempFileName.
	contents := stream contentsOfEntireFile.
	FileDirectory default deleteFileNamed: tempFileName ifAbsent: [].
	^ contents
! !

!WideString class methodsFor: 'instance creation' stamp: 'ar 4/12/2005 19:58'!
fromPacked: aLong
	"Convert from a longinteger to a String of length 4."

	| s val |
	s := self new: 1.
	val := (((aLong digitAt: 4) << 24) bitOr:((aLong digitAt: 3) << 16))
				bitOr: (((aLong digitAt: 2) << 8) bitOr: (aLong digitAt: 1)).
	s basicAt: 1 put: val.
	^ s.

"WideString fromPacked: 'TEXT' asPacked"
! !

!WideString class methodsFor: 'instance creation' stamp: 'yo 8/28/2002 13:39'!
fromString: aString 
	"Answer an instance of me that is a copy of the argument, aString."

	| inst |
	(aString isMemberOf: self) ifTrue: [
		^ aString copy.
	].
	inst := self new: aString size.
	1 to: aString size do: [:pos |
		inst basicAt: pos put: (aString basicAt: pos).
	].
	^ inst.
! !

!WideString class methodsFor: 'instance creation' stamp: 'ar 4/12/2005 20:00'!
from: aString 

	| newString |
	(aString isMemberOf: self)
		ifTrue: [^ aString copy].
	newString := self new: aString size.
	1 to: aString size do: [:index | newString basicAt: index put: (aString basicAt: index)].
	^ newString
! !

!WideString methodsFor: 'converting' stamp: 'yo 8/28/2002 14:46'!
asFourCode

	| result |
	self size = 1 ifFalse: [^self error: 'must be exactly four octets'].
	result := self basicAt: 1.
	(result bitAnd: 16r80000000) = 0 
		ifFalse: [self error: 'cannot resolve fourcode'].
	(result bitAnd: 16r40000000) = 0 ifFalse: [^result - 16r80000000].
	^ result
! !

!WideString methodsFor: 'converting' stamp: 'nice 11/1/2013 23:36'!
asLowercase
	^self collect: [:e | e asLowercase]! !

!WideString methodsFor: 'converting' stamp: 'nice 12/26/2009 01:03'!
asPacked
	^self inject: 0 into: [:pack :next | pack * 16r100000000 + next asInteger].! !

!WideString methodsFor: 'converting' stamp: 'nice 11/1/2013 23:35'!
asUppercase
	^self collect: [:e | e asUppercase]! !

!WideString methodsFor: 'accessing' stamp: 'eem 2/6/2014 14:38'!
at: index
	"Answer the Character stored in the field of the receiver indexed by the
	 argument.  Primitive.  Fail if the index argument is not an Integer or is out
	 of bounds.  Essential.  See Object documentation whatIsAPrimitive."

	<primitive: 63>
	^index isInteger
		ifTrue:
			[self errorSubscriptBounds: index]
		ifFalse:
			[index isNumber
				ifTrue: [self at: index asInteger]
				ifFalse: [self errorNonIntegerIndex]]! !

!WideString methodsFor: 'accessing' stamp: 'eem 2/6/2014 14:40'!
at: index put: aCharacter
	"Store the Character into the field of the receiver indicated by the index.
	 Primitive.  Fail if the index is not an Integer or is out of bounds, or if the
	 argument is not a Character.  Essential.  See Object documentation whatIsAPrimitive."

	<primitive: 64>
	^aCharacter isCharacter
		ifTrue:
			[index isInteger
				ifTrue: [self errorSubscriptBounds: index]
				ifFalse: [self errorNonIntegerIndex]]
		ifFalse:
			[self errorImproperStore]! !

!WideString methodsFor: 'accessing' stamp: 'yo 10/31/2002 22:29'!
byteAt: index

	| d r |
	d := (index + 3) // 4.
	r := (index - 1) \\ 4 + 1.
	^ (self wordAt: d) digitAt: ((4 - r) + 1).
! !

!WideString methodsFor: 'accessing' stamp: 'yo 11/3/2002 13:19'!
byteAt: index put: aByte

	| d r w |
	d := (index + 3) // 4.
	r := (index - 1) \\ 4 + 1.
	w := (self wordAt: d) bitAnd: ((16rFF<<((4 - r)*8)) bitInvert32).
	w := w + (aByte<<((4 - r)*8)).
	self basicAt: d put: w.
	^ aByte.
! !

!WideString methodsFor: 'accessing' stamp: 'yo 8/27/2002 14:22'!
byteSize

	^ self size * 4.
! !

!WideString methodsFor: 'converting' stamp: 'yo 3/14/2005 11:41'!
copyFrom: start to: stop

	| n |
	n := super copyFrom: start to: stop.
	n isOctetString ifTrue: [^ n asOctetString].
	^ n.
! !

!WideString methodsFor: 'testing' stamp: 'yo 7/29/2003 14:10'!
includesUnifiedCharacter

	^ self isUnicodeStringWithCJK
! !

!WideString methodsFor: 'testing' stamp: 'ar 4/12/2005 14:10'!
isUnicodeStringWithCJK

	self do: [:c |
		(c isTraditionalDomestic not and: [Unicode isUnifiedKanji: c charCode]) ifTrue: [
			^ true
		].
	].

	^ false.
! !

!WideString methodsFor: 'testing' stamp: 'ar 4/12/2005 19:52'!
isWideString
	"Answer whether the receiver is a WideString"
	^true! !

!WideString methodsFor: 'private' stamp: 'ar 4/9/2005 22:31'!
mutateJISX0208StringToUnicode

	| c |
	1 to: self size do: [:i |
		c := self at: i.
		(c leadingChar = JISX0208 leadingChar or: [
			c leadingChar = (JISX0208 leadingChar bitShift: 2)]) ifTrue: [
			self basicAt: i put: (Character leadingChar: JapaneseEnvironment leadingChar code: (c asUnicode)) asciiValue.
		]
	].
! !

!WideString methodsFor: 'accessing' stamp: 'ul 12/14/2009 07:58'!
replaceFrom: start to: stop with: replacement startingAt: repStart 

	<primitive: 105>
	replacement class == ByteString ifTrue: [
		^self replaceFrom: start to: stop with: replacement asWideString startingAt: repStart ]. 
	^super replaceFrom: start to: stop with: replacement startingAt: repStart.
! !

!WideString methodsFor: 'accessing' stamp: 'ar 4/10/2005 17:14'!
wordAt: index
	<primitive: 60>
	^ (self basicAt: index).
! !

!WideString methodsFor: 'accessing' stamp: 'ar 4/10/2005 17:14'!
wordAt: index put: anInteger
	<primitive: 61>
	self basicAt: index put: anInteger.
! !

ArrayedCollection variableWordSubclass: #WordArray
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Arrayed'!

!WordArray commentStamp: '<historical>' prior: 0!
WordArrays store 32-bit unsigned Integer values.
!

WordArray variableWordSubclass: #WordArrayForSegment
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Arrayed'!

!WordArrayForSegment methodsFor: 'as yet unclassified' stamp: 'tk 1/24/2000 23:22'!
restoreEndianness
	"This word object was just read in from a stream.  Do not correct the Endianness because the load primitive will reverse bytes as needed."

	"^ self"
! !

!WordArrayForSegment methodsFor: 'as yet unclassified' stamp: 'tk 1/24/2000 23:22'!
writeOn: aByteStream
	"Write quickly and disregard the endianness of the words.  Store the array of bits onto the argument, aStream.  (leading byte ~= 16r80) identifies this as raw bits (uncompressed)."

	aByteStream nextInt32Put: self size.	"4 bytes"
	aByteStream nextPutAll: self
! !

!WordArray class methodsFor: 'as yet unclassified' stamp: 'nice 12/27/2009 03:11'!
bobsTest
	| wa answer |
"
WordArray bobsTest
"
	answer := OrderedCollection new.
	wa := WordArray with: 16r01020304 with: 16r05060708.
	{false. true} do: [ :pad | | rawData s1 s2 wa2 |
		0 to: 3 do: [ :skip |
			s1 := RWBinaryOrTextStream on: ByteArray new.

			s1 next: skip put: 0.		"start at varying positions"
			wa writeOn: s1.
			pad ifTrue: [s1 next: 4-skip put: 0].	"force length to be multiple of 4"

			rawData := s1 contents.
			s2 := RWBinaryOrTextStream with: rawData.
			s2 reset.
			s2 skip: skip.			"get to beginning of object"
			wa2 := WordArray newFromStream: s2.
			answer add: {
				rawData size. 
				skip. 
				wa2 = wa. 
				wa2 asArray collect: [ :each | each radix: 16]
			}
		].
	].
	^answer explore! !

!WordArray methodsFor: 'converting' stamp: 'ar 9/14/1998 23:46'!
asWordArray
	^self! !

!WordArray methodsFor: 'accessing' stamp: 'sma 4/22/2000 17:47'!
atAllPut: value
	"Fill the receiver with the given value"

	<primitive: 145>
	super atAllPut: value! !

!WordArray methodsFor: 'accessing' stamp: 'ar 3/3/2001 16:18'!
byteSize
	^self size * 4! !

!WordArray methodsFor: 'accessing' stamp: 'tk 3/13/2000 14:46'!
bytesPerElement
	"Number of bytes in each item.  This multiplied by (self size)*8 gives the number of bits stored."
	^ 4! !

!WordArray methodsFor: 'accessing' stamp: 'ar 11/2/1998 12:19'!
defaultElement
	"Return the default element of the receiver"
	^0! !

!WordArray methodsFor: 'private' stamp: 'ar 2/15/1999 00:51'!
replaceFrom: start to: stop with: replacement startingAt: repStart 
	<primitive: 105>
	^super replaceFrom: start to: stop with: replacement startingAt: repStart ! !

SequenceableCollection subclass: #Interval
	instanceVariableNames: 'start stop step'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Sequenceable'!

!Interval commentStamp: '<historical>' prior: 0!
I represent a finite arithmetic progression.!

!Interval class methodsFor: 'instance creation' stamp: 'eem 7/19/2014 05:41'!
from: startInteger to: stopInteger 
	"Answer an instance of me, starting at startNumber, ending at 
	stopNumber, and with an interval increment of 1."

	^self basicNew
		setFrom: startInteger
		to: stopInteger
		by: 1! !

!Interval class methodsFor: 'instance creation' stamp: 'eem 7/19/2014 05:41'!
from: startInteger to: stopInteger by: stepInteger 
	"Answer an instance of me, starting at startNumber, ending at 
	stopNumber, and with an interval increment of stepNumber."

	^self basicNew
		setFrom: startInteger
		to: stopInteger
		by: stepInteger! !

!Interval class methodsFor: 'instance creation' stamp: 'nice 3/27/2008 00:17'!
newFrom: aCollection 
	"Answer an instance of me containing the same elements as aCollection."

    | newInterval n |

    (n := aCollection size) <= 1 ifTrue: [
		n = 0 ifTrue: [^self from: 1 to: 0].
		^self from: aCollection first to: aCollection last].
    	newInterval := self from: aCollection first to: aCollection last
	by: (aCollection last - aCollection first) // (n - 1).
	aCollection ~= newInterval
		ifTrue: [
			"Give a second chance, because progression might be arithmetic, but = answer false"
			(newInterval hasEqualElements: aCollection) ifFalse: [
				self error: 'The argument is not an arithmetic progression']].
	^newInterval

"	Interval newFrom: {1. 2. 3}
	{33. 5. -23} as: Interval
	{33. 5. -22} as: Interval    (an error)
	(-4 to: -12 by: -1) as: Interval
	#(2 4 6) asByteArray as: Interval.
"! !

!Interval methodsFor: 'adding' stamp: 'ul 4/12/2015 23:53'!
addAllFirstTo: anOrderedCollection
	"Add all of my elements to the beginning of anOrderedCollection"

	self reverseDo: [ :each | anOrderedCollection addFirst: each ]! !

!Interval methodsFor: 'adding' stamp: ''!
add: newObject 
	"Adding to an Interval is not allowed."

	self shouldNotImplement! !

!Interval methodsFor: 'accessing' stamp: 'nice 2/9/2010 20:43'!
anyOne
	"This message will fail for an empty Interval, super would not."
	^self at: 1! !

!Interval methodsFor: 'accessing' stamp: ''!
at: anInteger 
	"Answer the anInteger'th element."

	(anInteger >= 1 and: [anInteger <= self size])
		ifTrue: [^start + (step * (anInteger - 1))]
		ifFalse: [self errorSubscriptBounds: anInteger]! !

!Interval methodsFor: 'accessing' stamp: ''!
at: anInteger put: anObject 
	"Storing into an Interval is not allowed."

	self error: 'you can not store into an interval'! !

!Interval methodsFor: 'enumerating' stamp: ''!
collect: aBlock
	| nextValue result |
	result := self species new: self size.
	nextValue := start.
	1 to: result size do:
		[:i |
		result at: i put: (aBlock value: nextValue).
		nextValue := nextValue + step].
	^ result! !

!Interval methodsFor: 'enumerating' stamp: 'nice 12/22/2009 16:11'!
do: aBlock 
	"Evaluate aBlock for each value of the interval.
	Implementation note: instead of repeatedly incrementing the value
	    aValue := aValue + step.
	until stop is reached,
	We prefer to recompute value from start
	    aValue := start + (index * step).
	This is better for floating points accuracy, while not degrading Integer and Fraction speed too much.
	Moreover, this is consistent with methods #at: and #size"
	
	| aValue index size |
	index := 0.
	size := self size.
	[index < size]
		whileTrue: [aValue := start + (index * step).
			index := index + 1.
			aBlock value: aValue]! !

!Interval methodsFor: 'accessing' stamp: 'stp 8/19/2000 23:52'!
extent 
	"Answer the max - min of the receiver interval."
	"(10 to: 50) extent"

	^stop - start! !

!Interval methodsFor: 'accessing' stamp: ''!
first 
	"Refer to the comment in SequenceableCollection|first."

	^start! !

!Interval methodsFor: 'comparing' stamp: ''!
hash
	"Hash is reimplemented because = is implemented."

	^(((start hash bitShift: 2)
		bitOr: stop hash)
		bitShift: 1)
		bitOr: self size! !

!Interval methodsFor: 'accessing' stamp: ''!
increment
	"Answer the receiver's interval increment."

	^step! !

!Interval methodsFor: 'accessing' stamp: 'nice 12/22/2009 15:50'!
indexOf: anElement startingAt: startIndex ifAbsent: exceptionBlock 
	"startIndex is an positive integer, the collection index where the search is started."
	"during the computation of val , floats are only used when the receiver	contains floats"

	| index val |
	(self rangeIncludes: anElement)
		ifFalse: [^ exceptionBlock value].
	val := anElement - self first / self increment.
	val isFloat
		ifTrue: [(val - val rounded) abs * 100000000 < 1
				ifTrue: [index := val rounded + 1]
				ifFalse: [^ exceptionBlock value]]
		ifFalse: [val isInteger
				ifTrue: [index := val + 1]
				ifFalse: [^ exceptionBlock value]].
	"finally, the value of startIndex comes into play:"
	^ (index between: startIndex and: self size)
		ifTrue: [index]
		ifFalse: [exceptionBlock value]! !

!Interval methodsFor: 'testing' stamp: 'rhi 8/12/2003 09:52'!
isInterval

	^ true! !

!Interval methodsFor: 'accessing' stamp: ''!
last 
	"Refer to the comment in SequenceableCollection|last."

	^stop - (stop - start \\ step)! !

!Interval methodsFor: 'enumerating' stamp: 'dtl 5/31/2003 16:45'!
permutationsDo: aBlock
	"Repeatly value aBlock with a single copy of the receiver. Reorder the copy
	so that aBlock is presented all (self size factorial) possible permutations."
	"(1 to: 4) permutationsDo: [:each | Transcript cr; show: each printString]"

	self asArray permutationsDo: aBlock
! !

!Interval methodsFor: 'printing' stamp: 'sma 6/1/2000 09:50'!
printOn: aStream
	aStream nextPut: $(;
	 print: start;
	 nextPutAll: ' to: ';
	 print: stop.
	step ~= 1 ifTrue: [aStream nextPutAll: ' by: '; print: step].
	aStream nextPut: $)! !

!Interval methodsFor: 'accessing' stamp: 'di 12/6/1999 11:00'!
rangeIncludes: aNumber
	"Return true if the number lies in the interval between start and stop."

	step >= 0
		ifTrue: [^ aNumber between: start and: stop]
		ifFalse: [^ aNumber between: stop and: start]
! !

!Interval methodsFor: 'removing' stamp: ''!
remove: newObject 
	"Removing from an Interval is not allowed."

	self error: 'elements cannot be removed from an Interval'! !

!Interval methodsFor: 'enumerating' stamp: 'nice 4/30/2007 18:28'!
reverseDo: aBlock 
	"Evaluate aBlock for each element of my interval, in reverse order.
	Implementation notes: see do: for an explanation on loop detail"

	| aValue index |
	index := self size.
	[index > 0]
		whileTrue: [
			index := index - 1.
			aValue := start + (index * step).
			aBlock value: aValue]! !

!Interval methodsFor: 'private' stamp: ''!
setFrom: startInteger to: stopInteger by: stepInteger

	start := startInteger.
	stop := stopInteger.
	step := stepInteger! !

!Interval methodsFor: 'accessing' stamp: 'sma 5/12/2000 11:38'!
size
	"Answer how many elements the receiver contains."

	step < 0
		ifTrue: [start < stop
				ifTrue: [^ 0]
				ifFalse: [^ stop - start // step + 1]]
		ifFalse: [stop < start
				ifTrue: [^ 0]
				ifFalse: [^ stop - start // step + 1]]! !

!Interval methodsFor: 'private' stamp: ''!
species

	^Array! !

!Interval methodsFor: 'printing' stamp: 'nice 5/12/2007 00:55'!
storeOn: aStream 
	aStream nextPut: $(;
	 store: start;
	 nextPutAll: ' to: ';
	 store: stop.
	step ~= 1 ifTrue: [aStream nextPutAll: ' by: '; store: step].
	aStream nextPut: $)! !

!Interval methodsFor: 'accessing' stamp: 'ul 11/19/2014 13:43'!
sum
	"Optimized version. Use the sum(n * i - k, i=a..b) = -1/2 * (a - b - 1) * (n * (a + b) - 2 * k) equation with a = 1, n = step, b = self size and k = step - start."

	| b |
	b := self size.
	^b * ((b - 1) * step + (start * 2)) / 2! !

!Interval methodsFor: 'arithmetic' stamp: 'ajh 3/13/2003 15:45'!
+ number

	^ start + number to: stop + number by: step! !

!Interval methodsFor: 'arithmetic' stamp: 'ajh 3/13/2003 15:46'!
- number

	^ start - number to: stop - number by: step! !

!Interval methodsFor: 'comparing' stamp: 'rhi 8/14/2003 10:08'!
= anObject

	^ self == anObject
		ifTrue: [true]
		ifFalse: [anObject isInterval
			ifTrue: [start = anObject first
				and: [step = anObject increment
					and: [self last = anObject last]]]
			ifFalse: [super = anObject]]! !

SequenceableCollection subclass: #LinkedList
	instanceVariableNames: 'firstLink lastLink'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Sequenceable'!

!LinkedList commentStamp: '<historical>' prior: 0!
I represent a collection of links, which are containers for other objects. Using the message sequence addFirst:/removeLast causes the receiver to behave as a stack; using addLast:/removeFirst causes the receiver to behave as a queue.!

!LinkedList methodsFor: 'adding' stamp: ''!
addFirst: aLink 
	"Add aLink to the beginning of the receiver's list. Answer aLink."

	self isEmpty ifTrue: [lastLink := aLink].
	aLink nextLink: firstLink.
	firstLink := aLink.
	^aLink! !

!LinkedList methodsFor: 'adding' stamp: ''!
addLast: aLink 
	"Add aLink to the end of the receiver's list. Answer aLink."

	self isEmpty
		ifTrue: [firstLink := aLink]
		ifFalse: [lastLink nextLink: aLink].
	lastLink := aLink.
	^aLink! !

!LinkedList methodsFor: 'adding' stamp: ''!
add: aLink 
	"Add aLink to the end of the receiver's list. Answer aLink."

	^self addLast: aLink! !

!LinkedList methodsFor: 'adding' stamp: 'nice 7/28/2008 22:37'!
add: link after: otherLink

	"Add otherLink  after link in the list. Answer aLink."

	| savedLink |
	lastLink == otherLink ifTrue: [^ self addLast: link].
	savedLink := otherLink nextLink.
	otherLink nextLink: link.
	link nextLink:  savedLink.
	^link.! !

!LinkedList methodsFor: 'adding' stamp: 'ajh 8/22/2002 14:17'!
add: link before: otherLink

	| aLink |
	firstLink == otherLink ifTrue: [^ self addFirst: link].
	aLink := firstLink.
	[aLink == nil] whileFalse: [
		aLink nextLink == otherLink ifTrue: [
			link nextLink: aLink nextLink.
			aLink nextLink: link.
			^ link
		].
		 aLink := aLink nextLink.
	].
	^ self errorNotFound: otherLink! !

!LinkedList methodsFor: 'accessing' stamp: 'ajh 8/6/2002 15:46'!
at: index

	| i |
	i := 0.
	self do: [:link |
		(i := i + 1) = index ifTrue: [^ link]].
	^ self errorSubscriptBounds: index! !

!LinkedList methodsFor: 'accessing' stamp: 'nice 10/29/2011 22:41'!
at: index ifAbsent: exceptionBlock
	"Optimized to scan the list once, super would twice."
	| i |
	index < 1 ifTrue: [^exceptionBlock value].
	i := 0.
	self do: [:link |
		(i := i + 1) = index ifTrue: [^ link]].
	^exceptionBlock value! !

!LinkedList methodsFor: 'enumerating' stamp: ''!
do: aBlock

	| aLink |
	aLink := firstLink.
	[aLink == nil] whileFalse:
		[aBlock value: aLink.
		 aLink := aLink nextLink]! !

!LinkedList methodsFor: 'accessing' stamp: ''!
first
	"Answer the first link. Create an error notification if the receiver is 
	empty."

	self emptyCheck.
	^firstLink! !

!LinkedList methodsFor: 'testing' stamp: ''!
isEmpty

	^firstLink == nil! !

!LinkedList methodsFor: 'accessing' stamp: ''!
last
	"Answer the last link. Create an error notification if the receiver is 
	empty."

	self emptyCheck.
	^lastLink! !

!LinkedList methodsFor: 'copying' stamp: 'nice 10/5/2009 08:49'!
postCopy
	| aLink |
	super postCopy.
	firstLink isNil ifFalse: [
		aLink := firstLink := firstLink copy.
		[aLink nextLink isNil] whileFalse: [aLink nextLink: (aLink := aLink nextLink copy)].
		lastLink := aLink].! !

!LinkedList methodsFor: 'removing' stamp: 'nice 1/10/2009 00:23'!
removeAll
	"Implementation note: this has to be fast"

	firstLink := lastLink := nil! !

!LinkedList methodsFor: 'removing' stamp: ''!
removeFirst
	"Remove the first element and answer it. If the receiver is empty, create 
	an error notification."

	| oldLink |
	self emptyCheck.
	oldLink := firstLink.
	firstLink == lastLink
		ifTrue: [firstLink := nil. lastLink := nil]
		ifFalse: [firstLink := oldLink nextLink].
	oldLink nextLink: nil.
	^oldLink! !

!LinkedList methodsFor: 'removing' stamp: ''!
removeLast
	"Remove the receiver's last element and answer it. If the receiver is 
	empty, create an error notification."

	| oldLink aLink |
	self emptyCheck.
	oldLink := lastLink.
	firstLink == lastLink
		ifTrue: [firstLink := nil. lastLink := nil]
		ifFalse: [aLink := firstLink.
				[aLink nextLink == oldLink] whileFalse:
					[aLink := aLink nextLink].
				 aLink nextLink: nil.
				 lastLink := aLink].
	oldLink nextLink: nil.
	^oldLink! !

!LinkedList methodsFor: 'removing' stamp: ''!
remove: aLink ifAbsent: aBlock  
	"Remove aLink from the receiver. If it is not there, answer the result of
	evaluating aBlock."

	| tempLink |
	aLink == firstLink
		ifTrue: [firstLink := aLink nextLink.
				aLink == lastLink
					ifTrue: [lastLink := nil]]
		ifFalse: [tempLink := firstLink.
				[tempLink == nil ifTrue: [^aBlock value].
				 tempLink nextLink == aLink]
					whileFalse: [tempLink := tempLink nextLink].
				tempLink nextLink: aLink nextLink.
				aLink == lastLink
					ifTrue: [lastLink := tempLink]].
	aLink nextLink: nil.
	^aLink! !

!LinkedList methodsFor: 'accessing' stamp: 'eem 10/2/2011 19:07'!
size
	"Answer how many elements the receiver contains."

	| tally |
	tally := 0.
	self do: [:each | tally := tally + 1].
	^ tally! !

!LinkedList methodsFor: 'enumerating' stamp: 'ajh 8/6/2002 16:39'!
species

	^ Array! !

SequenceableCollection subclass: #OrderedCollection
	instanceVariableNames: 'array firstIndex lastIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Sequenceable'!

!OrderedCollection commentStamp: '<historical>' prior: 0!
I represent a collection of objects ordered by the collector.!

OrderedCollection subclass: #FloatCollection
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Sequenceable'!

!FloatCollection commentStamp: 'cmm 1/28/2013 19:49' prior: 0!
FloatCollctions store 32bit IEEE floating point numbers.!

!FloatCollection class methodsFor: 'overriding' stamp: 'cmm 1/28/2013 19:50'!
arrayType
	^ FloatArray! !

!FloatCollection methodsFor: 'adding' stamp: 'cmm 1/28/2013 19:55'!
addLast: aFloat
	aFloat isNumber ifFalse: [ self error: 'This collection can only store Floats.' ].
	^ super addLast: aFloat! !

!FloatCollection methodsFor: 'adding' stamp: 'ul 1/18/2015 23:45'!
asFloatArray
	"Optimized version"

	^array copyFrom: firstIndex to: lastIndex! !

!OrderedCollection class methodsFor: 'private' stamp: 'cmm 10/25/2010 22:26'!
arrayType
	^ Array! !

!OrderedCollection class methodsFor: 'instance creation' stamp: 'sma 5/12/2000 17:41'!
new
	^ self new: 10! !

!OrderedCollection class methodsFor: 'instance creation' stamp: 'ul 9/10/2011 15:11'!
newFrom: aCollection 
	"Answer an instance of me containing the same elements as aCollection."

	^(self new: aCollection size)
		addAll: aCollection;
		yourself

"	OrderedCollection newFrom: {1. 2. 3}
	{1. 2. 3} as: OrderedCollection
	{4. 2. 7} as: SortedCollection
"! !

!OrderedCollection class methodsFor: 'instance creation' stamp: 'cmm 10/25/2010 22:27'!
new: anInteger 
	^ self basicNew setCollection: (self arrayType new: anInteger)! !

!OrderedCollection class methodsFor: 'instance creation' stamp: 'cmm 10/25/2010 22:28'!
new: anInteger withAll: anObject
	^ self basicNew setContents: (self arrayType new: anInteger withAll: anObject)! !

!OrderedCollection class methodsFor: 'instance creation' stamp: 'apb 10/15/2000 22:02'!
ofSize: n
	"Create a new collection of size n with nil as its elements.
	This method exists because OrderedCollection new: n creates an
	empty collection,  not one of size n."
	| collection |
	collection := self new: n.
	collection setContents: (collection collector).
	^ collection
! !

!OrderedCollection methodsFor: 'adding' stamp: 'sw 3/1/2001 11:03'!
addAllFirstUnlessAlreadyPresent: anOrderedCollection 
	"Add each element of anOrderedCollection at the beginning of the receiver, preserving the order, but do not add any items that are already in the receiver.  Answer anOrderedCollection."

	anOrderedCollection reverseDo:
		[:each | (self includes: each) ifFalse: [self addFirst: each]].
	^ anOrderedCollection! !

!OrderedCollection methodsFor: 'adding' stamp: 'ul 4/12/2015 23:56'!
addAllFirst: aCollection 
	"Add all elements of aCollection to the beginning of the me. Answer aCollection. Use double dispatch to add elements in reverse order if aCollection implements #reverseDo:."

	^aCollection addAllFirstTo: self! !

!OrderedCollection methodsFor: 'adding' stamp: 'ul 4/12/2015 23:55'!
addAllLast: aCollection 
	"Add each element of aCollection at the end of me. Answer aCollection."

	^aCollection do: [ :each | self addLast: each ]! !

!OrderedCollection methodsFor: 'adding' stamp: 'sma 5/12/2000 11:26'!
addAll: aCollection 
	"Add each element of aCollection at my end. Answer	aCollection."

	^ self addAllLast: aCollection! !

!OrderedCollection methodsFor: 'adding' stamp: 'ul 2/24/2011 08:46'!
addFirst: newObject 
	"Add newObject to the beginning of the receiver. Answer newObject."

	firstIndex = 1 ifTrue: [ self makeRoomAtFirst ].
	^array at: (firstIndex := firstIndex - 1) put: newObject! !

!OrderedCollection methodsFor: 'adding' stamp: 'ul 2/24/2011 08:46'!
addLast: newObject 
	"Add newObject to the end of the receiver. Answer newObject."

	array size = lastIndex ifTrue: [ self makeRoomAtLast ].
	^array at: (lastIndex := lastIndex + 1) put: newObject! !

!OrderedCollection methodsFor: 'adding' stamp: ''!
add: newObject

	^self addLast: newObject! !

!OrderedCollection methodsFor: 'adding' stamp: 'ar 7/15/2008 23:03'!
add: newObject afterIndex: index 
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just after index. Answer newObject."
	(index between: 0 and: self size) ifFalse:[^self errorSubscriptBounds: index].
	self insert: newObject before: firstIndex + index.
	^ newObject! !

!OrderedCollection methodsFor: 'adding' stamp: ''!
add: newObject after: oldObject 
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just succeeding oldObject. Answer newObject."
	
	| index |
	index := self find: oldObject.
	self insert: newObject before: index + 1.
	^newObject! !

!OrderedCollection methodsFor: 'adding' stamp: 'ar 7/15/2008 23:05'!
add: newObject beforeIndex: index 
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just before index. Answer newObject."
	(index between: 1 and: self size+1) ifFalse:[^self errorSubscriptBounds: index].
	self insert: newObject before: firstIndex + index - 1.
	^ newObject! !

!OrderedCollection methodsFor: 'adding' stamp: ''!
add: newObject before: oldObject 
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just preceding oldObject. Answer newObject."
	
	| index |
	index := self find: oldObject.
	self insert: newObject before: index.
	^newObject! !

!OrderedCollection methodsFor: 'converting' stamp: 'ul 12/31/2014 13:50'!
asArray
	"Overriden for speed"

	| result |
	result := Array new: self size.
	result
		replaceFrom: 1
		to: result size
		with: array
		startingAt: firstIndex.
	^result! !

!OrderedCollection methodsFor: 'accessing' stamp: 'ul 4/11/2015 23:10'!
at: anInteger 
	"Answer my element at index anInteger. at: is used by a knowledgeable
	client to access an existing element"

	| index |
	1 <= anInteger ifFalse: [ self errorNoSuchElement ].
	(index := anInteger + firstIndex - 1) <= lastIndex ifFalse: [ self errorNoSuchElement ].
	^array at: index! !

!OrderedCollection methodsFor: 'accessing' stamp: 'ul 4/12/2015 04:51'!
at: index ifAbsentPut: block
	"Return value at index, however, if value does not exist (nil or out of bounds) then add block's value at index (growing self if necessary)"

	[ index <= self size ] whileFalse: [ self add: nil ].
	^(self at: index) ifNil: [ self at: index put: block value ]! !

!OrderedCollection methodsFor: 'accessing' stamp: 'ul 4/11/2015 23:10'!
at: anInteger put: anObject 
	"Put anObject at element index anInteger. at:put: cannot be used to
	append, front or back, to an ordered collection; it is used by a
	knowledgeable client to replace an element."

	| index |
	1 <= anInteger ifFalse: [ self errorNoSuchElement ].
	(index := anInteger + firstIndex - 1) <= lastIndex ifFalse: [ self errorNoSuchElement ].
	^array at: index put: anObject! !

!OrderedCollection methodsFor: 'accessing' stamp: 'sma 5/12/2000 11:42'!
capacity
	"Answer the current capacity of the receiver."

	^ array size! !

!OrderedCollection methodsFor: 'private' stamp: ''!
collector  "Private"
	^ array! !

!OrderedCollection methodsFor: 'enumerating' stamp: 'sma 2/5/2000 15:22'!
collect: aBlock 
	"Evaluate aBlock with each of my elements as the argument. Collect the 
	resulting values into a collection that is like me. Answer the new 
	collection. Override superclass in order to use addLast:, not at:put:."

	| newCollection |
	newCollection := self species new: self size.
	firstIndex to: lastIndex do:
		[:index |
		newCollection addLast: (aBlock value: (array at: index))].
	^ newCollection! !

!OrderedCollection methodsFor: 'enumerating' stamp: 'bf 5/18/2000 17:34'!
collect: aBlock from: fromIndex to: toIndex
	"Override superclass in order to use addLast:, not at:put:."
	| result |
	(fromIndex < 1 or:[toIndex + firstIndex - 1 > lastIndex])
		ifTrue: [^self errorNoSuchElement].
	result := self species new: toIndex - fromIndex + 1.
	firstIndex + fromIndex - 1 to: firstIndex + toIndex - 1 do:
		[:index | result addLast: (aBlock value: (array at: index))].
	^ result
! !

!OrderedCollection methodsFor: 'copying' stamp: ''!
copyEmpty
	"Answer a copy of the receiver that contains no elements."

	^self species new! !

!OrderedCollection methodsFor: 'copying' stamp: 'nice 10/5/2009 10:18'!
copyFrom: startIndex to: endIndex 
	"Answer a copy of the receiver that contains elements from position
	startIndex to endIndex."

	^self shallowCopy postCopyFrom: startIndex to: endIndex! !

!OrderedCollection methodsFor: 'copying' stamp: ''!
copyReplaceFrom: start to: stop with: replacementCollection 
	"Answer a copy of the receiver with replacementCollection's elements in
	place of the receiver's start'th to stop'th elements. This does not expect
	a 1-1 map from replacementCollection to the start to stop elements, so it
	will do an insert or append."

	| newOrderedCollection delta startIndex stopIndex |
	"if start is less than 1, ignore stop and assume this is inserting at the front. 
	if start greater than self size, ignore stop and assume this is appending. 
	otherwise, it is replacing part of me and start and stop have to be within my 
	bounds. "
	delta := 0.
	startIndex := start.
	stopIndex := stop.
	start < 1
		ifTrue: [startIndex := stopIndex := 0]
		ifFalse: [startIndex > self size
				ifTrue: [startIndex := stopIndex := self size + 1]
				ifFalse: 
					[(stopIndex < (startIndex - 1) or: [stopIndex > self size])
						ifTrue: [self errorOutOfBounds].
					delta := stopIndex - startIndex + 1]].
	newOrderedCollection := 
		self species new: self size + replacementCollection size - delta.
	1 to: startIndex - 1 do: [:index | newOrderedCollection add: (self at: index)].
	1 to: replacementCollection size do: 
		[:index | newOrderedCollection add: (replacementCollection at: index)].
	stopIndex + 1 to: self size do: [:index | newOrderedCollection add: (self at: index)].
	^newOrderedCollection! !

!OrderedCollection methodsFor: 'copying' stamp: ''!
copyWith: newElement 
	"Answer a copy of the receiver that is 1 bigger than the receiver and 
	includes the argument, newElement, at the end."

	| newCollection |
	newCollection := self copy.
	newCollection add: newElement.
	^newCollection! !

!OrderedCollection methodsFor: 'enumerating' stamp: ''!
do: aBlock 
	"Override the superclass for performance reasons."
	| index |
	index := firstIndex.
	[index <= lastIndex]
		whileTrue: 
			[aBlock value: (array at: index).
			index := index + 1]! !

!OrderedCollection methodsFor: 'private' stamp: 'ul 4/12/2015 23:32'!
errorNoSuchElement

	self error: (String streamContents: [ :stream | 
		stream nextPutAll: 'attempt to index a non-existent element in '.
		self printNameOn: stream ])! !

!OrderedCollection methodsFor: 'private' stamp: 'ul 4/12/2015 23:30'!
errorNotEnoughElements

	self error: (String streamContents: [ :stream | 
		stream nextPutAll: 'attempt to remove more elements than possible from '.
		self printNameOn: stream ])! !

!OrderedCollection methodsFor: 'private' stamp: 'BG 1/9/2004 12:26'!
find: oldObject
  "  This method answers an index in the range firstIndex .. lastIndex, which is meant for internal use only.
     Never use this method in your code, the methods for public use are:
        #indexOf:
        #indexOf:ifAbsent: "

	| index |
	index := firstIndex.
	[index <= lastIndex]
		whileTrue:
			[(array at: index) = oldObject ifTrue: [^ index].
			index := index + 1].
	self errorNotFound: oldObject! !

!OrderedCollection methodsFor: 'private' stamp: 'cmm 10/25/2010 22:27'!
growAtFirst
	"Add new empty slots to the front of array, while keeping the empty slots at the end."

	| newArray newFirstIndex newLastIndex |
	newArray := self class arrayType new: (array size * 2 max: 1).
	newFirstIndex := newArray size - array size + firstIndex.
	newLastIndex := newFirstIndex + lastIndex - firstIndex.
	newArray 
		replaceFrom: newFirstIndex
		to: newLastIndex
		with: array
		startingAt: firstIndex.
	array := newArray.
	firstIndex := newFirstIndex.
	lastIndex := newLastIndex! !

!OrderedCollection methodsFor: 'private' stamp: 'cmm 10/25/2010 22:27'!
growAtLast
	"Add new empty slots to the end of array, while keeping the empty slots at the front."

	| newArray |
	newArray := self class arrayType new: (array size * 2 max: 1).
	newArray 
		replaceFrom: firstIndex
		to: lastIndex
		with: array
		startingAt: firstIndex.
	array := newArray! !

!OrderedCollection methodsFor: 'accessing' stamp: 'ul 4/12/2015 04:41'!
indexOf: anElement startingAt: start ifAbsent: exceptionBlock
	"Optimized version."

	firstIndex + start - 1 to: lastIndex do: [ :index |
		(array at: index) = anElement ifTrue: [ ^index - firstIndex + 1 ] ].
	^exceptionBlock value! !

!OrderedCollection methodsFor: 'private' stamp: 'BG 1/9/2004 12:29'!
insert: anObject before: spot

  "  spot is an index in the range firstIndex .. lastIndex, such an index is not known from outside the collection. 
     Never use this method in your code, it is meant for private use by OrderedCollection only.
     The methods for use are:
        #add:before:   to insert an object before another object
        #add:beforeIndex:   to insert an object before a given position. "
	| "index" delta spotIndex|
	spotIndex := spot.
	delta := spotIndex - firstIndex.
	firstIndex = 1
		ifTrue: 
			[self makeRoomAtFirst.
			spotIndex := firstIndex + delta].
	firstIndex := firstIndex - 1.
	array
		replaceFrom: firstIndex
		to: spotIndex - 2
		with: array
		startingAt: firstIndex + 1.
	array at: spotIndex - 1 put: anObject.
"	index := firstIndex := firstIndex - 1.
	[index < (spotIndex - 1)]
		whileTrue: 
			[array at: index put: (array at: index + 1).
			index := index + 1].
	array at: index put: anObject."
	^ anObject! !

!OrderedCollection methodsFor: 'sorting' stamp: 'ul 8/8/2013 14:24'!
isSorted
	"Return true if the receiver is sorted by #<=."
	
	^array
		isSortedBetween: firstIndex
		and: lastIndex! !

!OrderedCollection methodsFor: 'sorting' stamp: 'ul 8/8/2013 14:22'!
isSortedBetween: startIndex and: endIndex
	"Return true if the receiver is sorted by #<= between startIndex and endIndex."
	
	^array isSortedBetween: startIndex + firstIndex - 1 and: endIndex + firstIndex - 1! !

!OrderedCollection methodsFor: 'sorting' stamp: 'ul 8/8/2013 14:25'!
isSortedBy: aSortBlockOrNil
	"Return true if the receiver is sorted by aSortBlockOrNil. Use #<= for comparison if aSortBlockOrNil is nil."
	
	^array
		isSortedBy: aSortBlockOrNil
		between: firstIndex
		and: lastIndex! !

!OrderedCollection methodsFor: 'sorting' stamp: 'ul 4/12/2015 23:06'!
isSortedBy: aSortBlockOrNil between: startIndex and: endIndex
	"Return true if the receiver is sorted by aSortBlockOrNil between startIndex and endIndex. Use #<= for comparison if aSortBlockOrNil is nil."
	
	^array
		isSortedBy: aSortBlockOrNil
		between: startIndex + firstIndex - 1
		and: endIndex + firstIndex - 1! !

!OrderedCollection methodsFor: 'private' stamp: 'ul 5/10/2010 13:42'!
makeRoomAtFirst
	"Make some empty slots at the front of the array. If we have more than 50% free space, then just move the elements, so that the first 50% of the slots are free, otherwise add new free slots to the front by growing. Precondition: firstIndex = 1"
	
	| tally newFirstIndex newLastIndex |
	tally := self size.
	tally * 2 >= array size ifTrue: [ ^self growAtFirst ].
	tally = 0 ifTrue: [ ^self resetTo: array size + 1 ].
	newFirstIndex := array size // 2 + 1.
	newLastIndex := newFirstIndex - firstIndex + lastIndex.
	0 to: tally - 1 do: [ :offset |
		array at: newLastIndex - offset put: (array at: lastIndex - offset) ].
	array from: firstIndex to: newFirstIndex - 1 put: nil.
	firstIndex := newFirstIndex.
	lastIndex := newLastIndex! !

!OrderedCollection methodsFor: 'private' stamp: 'ul 5/10/2010 13:42'!
makeRoomAtLast
	"Make some empty slots at the end of the array. If we have more than 50% free space, then just move the elements, so that the last 50% of the slots are free, otherwise add new free slots to the end by growing. Precondition: lastIndex = array size"
	
	| tally newFirstIndex newLastIndex |
	tally := self size.
	tally * 2 >= lastIndex ifTrue: [ ^self growAtLast ].
	tally = 0 ifTrue: [ ^self resetTo: 1 ].
	newLastIndex := lastIndex // 2.
	newFirstIndex := newLastIndex - lastIndex + firstIndex.
	array 
		replaceFrom: newFirstIndex
		to: newLastIndex
		with: array
		startingAt: firstIndex.
	array from: newLastIndex + 1 to: lastIndex put: nil.
	firstIndex := newFirstIndex.
	lastIndex := newLastIndex! !

!OrderedCollection methodsFor: 'enumerating' stamp: 'ul 3/9/2016 21:57'!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject. Optimized version."

	| tally |
	tally := 0.
	firstIndex to: lastIndex do: [ :index |
		(array at: index) = anObject ifTrue: [ tally := tally + 1 ] ].
	^tally! !

!OrderedCollection methodsFor: 'copying' stamp: 'nice 10/5/2009 08:50'!
postCopy
	array := array copy! !

!OrderedCollection methodsFor: 'copying' stamp: 'nice 5/28/2008 21:02'!
postCopyFrom: startIndex to: endIndex 
	"finish copying the array in a certain range."

	endIndex < startIndex ifFalse: [
		"Because actual size of the array may be greater than used size,
		postCopyFrom:to: may fail to fail and answer an incorrect result
		if this sanity check were not applied"
		(startIndex between: 1 and: self size) ifFalse: [^self error: 'startIndex is out of bounds'].
		(endIndex between: 1 and: self size) ifFalse: [^self error: 'endIndex is out of bounds']].
	
	"Add a protection that lacks in Array>>postcopy"
	array := array copyFrom: startIndex + firstIndex - 1 to: (endIndex max: startIndex - 1) + firstIndex - 1.
	firstIndex := 1.
	lastIndex := array size! !

!OrderedCollection methodsFor: 'removing' stamp: 'cmm 10/25/2010 22:26'!
removeAll
	"remove all the elements from this collection.
	Keep same amount of storage"
	
	self setCollection: (self class arrayType new: array size)! !

!OrderedCollection methodsFor: 'removing' stamp: 'ul 4/12/2015 23:15'!
removeAllSuchThat: aBlock 
	"Remove each element of the receiver for which aBlock evaluates to true.
	The method in Collection is O(N^2), this is O(N)."

	| n |
	n := firstIndex.
	firstIndex to: lastIndex do: [ :index |
		| element |
		(aBlock value: (element := array at: index)) ifFalse: [
			array at: n put: element.
			n := n + 1 ] ].
	array from: n to: lastIndex put: nil.
	lastIndex := n - 1! !

!OrderedCollection methodsFor: 'removing' stamp: 'ar 5/22/2000 12:19'!
removeAt: index
	| removed |
	removed := self at: index.
	self removeIndex: index + firstIndex - 1.
	^removed! !

!OrderedCollection methodsFor: 'removing' stamp: ''!
removeFirst
	"Remove the first element of the receiver and answer it. If the receiver is 
	empty, create an error notification."
	| firstObject |
	self emptyCheck.
	firstObject := array at: firstIndex.
	array at: firstIndex put: nil.
	firstIndex := firstIndex + 1.
	^ firstObject! !

!OrderedCollection methodsFor: 'removing' stamp: 'ul 4/12/2015 23:33'!
removeFirst: n 
	"Remove the first n objects into an array."

	| lastIndexToRemove result |
	n < 1 ifTrue: [ self errorNoSuchElement ].
	lastIndex < (lastIndexToRemove := firstIndex + n - 1) ifTrue: [ self errorNotEnoughElements ].
	result := array copyFrom: firstIndex to: lastIndexToRemove.
	array from: firstIndex to: lastIndexToRemove put: nil.
	firstIndex := lastIndexToRemove + 1.
	^result! !

!OrderedCollection methodsFor: 'private' stamp: 'BG 1/9/2004 12:28'!
removeIndex: removedIndex
  "  removedIndex is an index in the range firstIndex .. lastIndex, such an index is not known from outside the collection.
    Never use this method in your code, it is meant for private use by OrderedCollection only.
     The method for public use is:
        #removeAt: "

	array 
		replaceFrom: removedIndex 
		to: lastIndex - 1 
		with: array 
		startingAt: removedIndex+1.
	array at: lastIndex put: nil.
	lastIndex := lastIndex - 1.! !

!OrderedCollection methodsFor: 'removing' stamp: ''!
removeLast
	"Remove the last element of the receiver and answer it. If the receiver is 
	empty, create an error notification."
	| lastObject |
	self emptyCheck.
	lastObject := array at: lastIndex.
	array at: lastIndex put: nil.
	lastIndex := lastIndex - 1.
	^ lastObject! !

!OrderedCollection methodsFor: 'removing' stamp: 'ul 4/12/2015 23:37'!
removeLast: n 
	"Remove the last n objects into an array with last in last position."

	| firstIndexToRemove result |
	n < 1 ifTrue: [ self errorNoSuchElement ].
	(firstIndexToRemove := lastIndex - n + 1) < firstIndex ifTrue: [ self errorNotEnoughElements ].
	result := array copyFrom: firstIndexToRemove to: lastIndex.
	array from: firstIndexToRemove to: lastIndex put: nil.
	lastIndex := firstIndexToRemove - 1.
	^result! !

!OrderedCollection methodsFor: 'removing' stamp: ''!
remove: oldObject ifAbsent: absentBlock

	| index |
	index := firstIndex.
	[index <= lastIndex]
		whileTrue: 
			[oldObject = (array at: index)
				ifTrue: 
					[self removeIndex: index.
					^ oldObject]
				ifFalse: [index := index + 1]].
	^ absentBlock value! !

!OrderedCollection methodsFor: 'removing' stamp: 'ul 2/24/2011 14:33'!
reset
	"Quickly remove all elements. The objects will be still referenced, but will not be 	accessible."
	
	self resetTo: 1! !

!OrderedCollection methodsFor: 'removing' stamp: 'ul 2/24/2011 14:38'!
resetTo: index
	"Quickly remove all elements. The objects will be still referenced, but will not be 	accessible. Also make sure that the first object will be inserted at index. Choosing the 	right value has had great impact on performance, but it's neglible with the current 	implementation, so it's better to use #reset instead in most cases."
	
	firstIndex := index.
	lastIndex := firstIndex - 1! !

!OrderedCollection methodsFor: 'enumerating' stamp: ''!
reverseDo: aBlock 
	"Override the superclass for performance reasons."
	| index |
	index := lastIndex.
	[index >= firstIndex]
		whileTrue: 
			[aBlock value: (array at: index).
			index := index - 1]! !

!OrderedCollection methodsFor: 'enumerating' stamp: 'sma 2/5/2000 15:13'!
select: aBlock 
	"Evaluate aBlock with each of my elements as the argument. Collect into
	a new collection like the receiver, only those elements for which aBlock
	evaluates to true."

	| newCollection element |
	newCollection := self copyEmpty.
	firstIndex to: lastIndex do:
		[:index |
		(aBlock value: (element := array at: index))
			ifTrue: [newCollection addLast: element]].
	^ newCollection! !

!OrderedCollection methodsFor: 'private' stamp: 'di 11/14/97 12:54'!
setCollection: anArray
	array := anArray.
	self reset! !

!OrderedCollection methodsFor: 'private' stamp: 'apb 10/15/2000 18:10'!
setContents: anArray
	array := anArray.
	firstIndex := 1.
	lastIndex := array size.! !

!OrderedCollection methodsFor: 'accessing' stamp: 'sma 5/12/2000 11:39'!
size
	"Answer how many elements the receiver contains."

	^ lastIndex - firstIndex + 1! !

!OrderedCollection methodsFor: 'sorting' stamp: 'ul 11/2/2009 04:56'!
sort
	"Sort this array into ascending order using the '<=' operator."

	self sort: nil! !

!OrderedCollection methodsFor: 'sorting' stamp: 'ul 3/22/2010 12:16'!
sorted: aSortBlockOrNil
	"Return a new sequenceable collection which contains the same elements as self but its elements are sorted by aSortBlockOrNil. The block should take two arguments and return true if the first element should preceed the second one. If aSortBlock is nil then <= is used for comparison."
	
	^self copy sort: aSortBlockOrNil! !

!OrderedCollection methodsFor: 'sorting' stamp: 'ul 12/31/2014 12:25'!
sort: aSortBlock 
	"Sort this collection using aSortBlock. The block should take two arguments and return true if the first element should preceed the second one. If aSortBlock is nil then <= is used for comparison."

	self size <= 1 ifTrue: [ ^self ].
	array
		mergeSortFrom: firstIndex
		to: lastIndex
		by: aSortBlock! !

!OrderedCollection methodsFor: 'enumerating' stamp: 'bf 5/16/2000 16:30'!
withIndexCollect: elementAndIndexBlock 
	"Just like with:collect: except that the iteration index supplies the second argument to the block. Override superclass in order to use addLast:, not at:put:."

	| newCollection |
	newCollection := self species new: self size.
	firstIndex to: lastIndex do:
		[:index |
		newCollection addLast: (elementAndIndexBlock
			value: (array at: index)
			value: index - firstIndex + 1)].
	^ newCollection! !

!OrderedCollection methodsFor: 'enumerating' stamp: 'di 8/31/1999 13:13'!
with: otherCollection collect: twoArgBlock 
	"Collect and return the result of evaluating twoArgBlock with 
	corresponding elements from this collection and otherCollection."
	| result |
	otherCollection size = self size ifFalse: [self error: 'otherCollection must be the same size'].
	result := self species new: self size.
	1 to: self size do:
		[:index | result addLast: (twoArgBlock value: (self at: index)
									value: (otherCollection at: index))].
	^ result! !

OrderedCollection subclass: #SortedCollection
	instanceVariableNames: 'sortBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Sequenceable'!

!SortedCollection commentStamp: 'dtl 9/6/2009 16:02' prior: 0!
I represent a collection of objects ordered by some property of the objects themselves. The ordering is specified in a BlockContext. The default sorting function is a <= comparison on elements.!

!SortedCollection class methodsFor: 'instance creation' stamp: ''!
sortBlock: aBlock 
	"Answer an instance of me such that its elements are sorted according to 
	the criterion specified in aBlock."

	^(super new: 10) sortBlock: aBlock! !

!SortedCollection methodsFor: 'adding' stamp: 'sma 4/28/2000 18:35'!
addAll: aCollection
	aCollection size > (self size // 3)
		ifTrue:
			[aCollection do: [:each | self addLast: each].
			self reSort]
		ifFalse: [aCollection do: [:each | self add: each]].
	^ aCollection! !

!SortedCollection methodsFor: 'adding' stamp: 'go 4/26/2000 17:26'!
addFirst: newObject
	self shouldNotImplement! !

!SortedCollection methodsFor: 'adding' stamp: 'go 4/27/2000 13:19'!
add: newObject
	^ super insert: newObject before: (self indexForInserting: newObject)! !

!SortedCollection methodsFor: 'accessing' stamp: 'sma 4/28/2000 17:47'!
at: anInteger put: anObject
	self shouldNotImplement! !

!SortedCollection methodsFor: 'enumerating' stamp: 'sma 2/5/2000 15:22'!
collect: aBlock 
	"Evaluate aBlock with each of my elements as the argument. Collect the 
	resulting values into an OrderedCollection. Answer the new collection. 
	Override the superclass in order to produce an OrderedCollection instead
	of a SortedCollection."

	| newCollection | 
	newCollection := OrderedCollection new: self size.
	self do: [:each | newCollection addLast: (aBlock value: each)].
	^ newCollection! !

!SortedCollection methodsFor: 'adding' stamp: 'MPH 10/23/2000 13:31'!
copyEmpty
	"Answer a copy of the receiver without any of the receiver's elements."

	^self species sortBlock: sortBlock! !

!SortedCollection methodsFor: 'private' stamp: 'stp 04/23/1999 05:36'!
indexForInserting: newObject

	| index low high |
	low := firstIndex.
	high := lastIndex.
	sortBlock isNil
		ifTrue: [[index := high + low // 2.  low > high]
			whileFalse: 
				[((array at: index) <= newObject)
					ifTrue: [low := index + 1]
					ifFalse: [high := index - 1]]]
		ifFalse: [[index := high + low // 2.  low > high]
			whileFalse: 
				[(sortBlock value: (array at: index) value: newObject)
					ifTrue: [low := index + 1]
					ifFalse: [high := index - 1]]].
	^low! !

!SortedCollection methodsFor: 'private' stamp: 'go 4/26/2000 17:17'!
insert: anObject before: spot
	self shouldNotImplement! !

!SortedCollection methodsFor: 'accessing' stamp: 'tk 3/28/1999 22:55'!
median
	"Return the middle element, or as close as we can get."

	^ self at: self size + 1 // 2! !

!SortedCollection methodsFor: 'private' stamp: 'ul 6/8/2011 23:18'!
reSort

	firstIndex < lastIndex ifTrue: [ 
		array quickSortFrom: firstIndex to: lastIndex by: sortBlock ]! !

!SortedCollection methodsFor: 'converting' stamp: 'nice 2/26/2008 22:28'!
reversed
	"Answer a collection that Sort elements in reverse order"
	
	^self shallowCopy reverseInPlace! !

!SortedCollection methodsFor: 'converting' stamp: 'ul 3/15/2011 15:03'!
reverseInPlace
	"Change this colleciton into its reversed.
	Do not make a copy like reversed do, but change self in place."
	
	| newFirstIndex |
	newFirstIndex := 1 + array size - lastIndex.
	lastIndex := 1 + array size - firstIndex.
	firstIndex := newFirstIndex.
	array := array reversed.
	sortBlock := sortBlock
		ifNil: [ [ :a :b | b <= a ] ]
		ifNotNil: [ [ :a :b | sortBlock value: b value: a ] ]! !

!SortedCollection methodsFor: 'private' stamp: 'hg 12/17/2001 19:30'!
should: a precede: b

	^sortBlock ifNil: [a <= b] ifNotNil: [sortBlock value: a value: b]
! !

!SortedCollection methodsFor: 'accessing' stamp: ''!
sortBlock
	"Answer the blockContext which is the criterion for sorting elements of 
	the receiver."

	^sortBlock! !

!SortedCollection methodsFor: 'accessing' stamp: 'ul 6/2/2016 20:13'!
sortBlock: aBlock 
	"Make the argument, aBlock, be the criterion for ordering elements of the 
	receiver."

	sortBlock := aBlock.
	"sortBlocks with side effects may not work right"
	self size > 1 ifTrue: [self reSort]! !

!SortedCollection methodsFor: 'topological sort' stamp: 'hg 1/2/2002 13:34'!
sortTopologically
	"Plenty of room for increased efficiency in this one."

	| remaining result pick |
	remaining := self asOrderedCollection.
	result := OrderedCollection new.
	[remaining isEmpty] whileFalse: [
		pick := remaining select: [:item |
			remaining allSatisfy: [:anotherItem |
				item == anotherItem or: [self should: item precede: anotherItem]]].
		pick isEmpty ifTrue: [self error: 'bad topological ordering'].
		result addAll: pick.
		remaining removeAll: pick].
	^self copySameFrom: result! !

!SortedCollection methodsFor: 'sorting' stamp: 'ul 12/23/2009 12:12'!
sort: aSortBlock 
	"Sort this collection using aSortBlock. The block should take two arguments
	and return true if the first element should preceed the second one.
	If aSortBlock is nil then <= is used for comparison."

	super sort: aSortBlock.
	sortBlock := aSortBlock! !

!SortedCollection methodsFor: 'comparing' stamp: ''!
= aSortedCollection
	"Answer true if my and aSortedCollection's species are the same,
	and if our blocks are the same, and if our elements are the same."

	self species = aSortedCollection species ifFalse: [^ false].
	sortBlock = aSortedCollection sortBlock
		ifTrue: [^ super = aSortedCollection]
		ifFalse: [^ false]! !

OrderedCollection subclass: #WeakOrderedCollection
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Weak'!

!WeakOrderedCollection class methodsFor: 'as yet unclassified' stamp: 'cmm 10/25/2010 22:27'!
arrayType
	^ WeakArray! !

!SequenceableCollection class methodsFor: 'stream creation' stamp: 'ul 1/20/2014 17:05'!
new: newSize streamContents: blockWithArg

	| stream originalContents |
	stream := WriteStream on: (self new: newSize).
	blockWithArg value: stream.
	originalContents := stream originalContents.
	originalContents size =  stream position
		ifTrue: [ ^originalContents ]
		ifFalse: [ ^stream contents ]! !

!SequenceableCollection class methodsFor: 'stream creation' stamp: 'ul 9/12/2009 19:33'!
streamContents: blockWithArg
	
	^self new: 100 streamContents: blockWithArg! !

!SequenceableCollection class methodsFor: 'stream creation' stamp: 'di 6/20/97 09:07'!
streamContents: blockWithArg limitedTo: sizeLimit
	| stream |
	stream := LimitedWriteStream on: (self new: (100 min: sizeLimit)).
	stream setLimit: sizeLimit limitBlock: [^ stream contents].
	blockWithArg value: stream.
	^ stream contents
"
String streamContents: [:s | 1000 timesRepeat: [s nextPutAll: 'Junk']] limitedTo: 25
 'JunkJunkJunkJunkJunkJunkJ'
"! !

!SequenceableCollection methodsFor: 'adding' stamp: 'ul 4/12/2015 23:53'!
addAllFirstTo: anOrderedCollection
	"Add all of my elements to the beginning of anOrderedCollection"

	self reverseDo: [ :each | anOrderedCollection addFirst: each ]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 5/12/2000 18:00'!
after: target
	"Answer the element after target.  Raise an error if target is not
	in the receiver, or if there are no elements after it."

	^ self after: target ifAbsent: [self errorNotFound: target]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ul 11/15/2010 10:06'!
after: target ifAbsent: exceptionBlock
	"Answer the element after target.  Answer the result of evaluation
	the exceptionBlock if target is not in the receiver, or if there are 
	no elements after it."

	| index |
	index := self indexOf: target.
	^ (index = 0 or: [index = self size])
		ifTrue: [exceptionBlock value]
		ifFalse: [self at: index + 1]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 6/1/2000 15:34'!
allButFirst
	"Answer a copy of the receiver containing all but the first
	element. Raise an error if there are not enough elements."

	^ self allButFirst: 1! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ajh 8/6/2002 15:03'!
allButFirstDo: block

	2 to: self size do:
		[:index | block value: (self at: index)]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 6/1/2000 15:35'!
allButFirst: n
	"Answer a copy of the receiver containing all but the first n
	elements. Raise an error if there are not enough elements."

	^ self copyFrom: n + 1 to: self size! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 6/1/2000 15:35'!
allButLast
	"Answer a copy of the receiver containing all but the last
	element. Raise an error if there are not enough elements."

	^ self allButLast: 1! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ajh 8/6/2002 15:01'!
allButLastDo: block

	1 to: self size - 1 do:
		[:index | block value: (self at: index)]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 6/1/2000 15:35'!
allButLast: n
	"Answer a copy of the receiver containing all but the last n
	elements. Raise an error if there are not enough elements."

	^ self copyFrom: 1 to: self size - n! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 5/12/2000 11:33'!
anyOne
	^ self first! !

!SequenceableCollection methodsFor: 'converting' stamp: 'sma 5/12/2000 17:32'!
asArray
	"Answer an Array whose elements are the elements of the receiver."

	^ Array withAll: self! !

!SequenceableCollection methodsFor: 'converting' stamp: 'sma 5/12/2000 17:36'!
asByteArray
	"Answer a ByteArray whose elements are the elements of the receiver."

	^ ByteArray withAll: self! !

!SequenceableCollection methodsFor: 'converting' stamp: 'ar 3/3/2001 20:06'!
asColorArray
	^ColorArray withAll: self! !

!SequenceableCollection methodsFor: 'private' stamp: 'tk 7/30/97 12:42'!
asDigitsAt: anInteger in: aCollection do: aBlock
	"(0 to: 1) asDigitsToPower: 4 do: [:each | Transcript cr; show: each printString]"

	self do: 
		[:each | 
		aCollection at: anInteger put: each.
		anInteger = aCollection size 
			ifTrue: [aBlock value: aCollection]
			ifFalse: [self asDigitsAt: anInteger + 1 in: aCollection do: aBlock]].! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'tk 7/30/97 12:41'!
asDigitsToPower: anInteger do: aBlock
	"Repeatedly value aBlock with a single Array.  Adjust the collection
	so that aBlock is presented all (self size raisedTo: anInteger) possible 
	combinations of the receiver's elements taken as digits of an anInteger long number."
	"(0 to: 1) asDigitsToPower: 4 do: [:each | Transcript cr; show: each printString]"

	| aCollection |
	aCollection := Array new: anInteger.
	self asDigitsAt: 1 in: aCollection do: aBlock! !

!SequenceableCollection methodsFor: 'converting' stamp: 'ar 9/14/1998 23:47'!
asFloatArray
	"Answer a FloatArray whose elements are the elements of the receiver, in 
	the same order."

	| floatArray |
	floatArray := FloatArray new: self size.
	1 to: self size do:[:i| floatArray at: i put: (self at: i) asFloat ].
	^floatArray! !

!SequenceableCollection methodsFor: 'converting' stamp: 'ar 10/10/1998 16:19'!
asIntegerArray
	"Answer an IntegerArray whose elements are the elements of the receiver, in 
	the same order."

	| intArray |
	intArray := IntegerArray new: self size.
	1 to: self size do:[:i| intArray at: i put: (self at: i)].
	^intArray! !

!SequenceableCollection methodsFor: 'converting' stamp: 'ar 4/10/2005 18:02'!
asStringWithCr
	"Convert to a string with returns between items.  Elements are
usually strings.
	 Useful for labels for PopUpMenus."
	| labelStream |
	labelStream := WriteStream on: (String new: 200).
	self do: [:each |
		each isString
			ifTrue: [labelStream nextPutAll: each; cr]
			ifFalse: [each printOn: labelStream. labelStream cr]].
	self size > 0 ifTrue: [labelStream skip: -1].
	^ labelStream contents! !

!SequenceableCollection methodsFor: 'converting' stamp: 'ar 10/10/1998 16:20'!
asWordArray
	"Answer a WordArray whose elements are the elements of the receiver, in 
	the same order."

	| wordArray |
	wordArray := WordArray new: self size.
	1 to: self size do:[:i| wordArray at: i put: (self at: i)].
	^wordArray! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 4/22/2000 17:45'!
atAllPut: anObject 
	"Put anObject at every one of the receiver's indices."

	| size |
	(size := self size) > 26 "first method faster from 27 accesses and on"
		ifTrue: [self from: 1 to: size put: anObject]
		ifFalse: [1 to: size do: [:index | self at: index put: anObject]]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'apb 11/4/2000 22:51'!
atAll: indexArray
	"Answer a new collection like the receiver which contains all elements
	of the receiver at the indices of indexArray."
	"#('one' 'two' 'three' 'four') atAll: #(3 2 4)"

	| newCollection |
	newCollection := self species ofSize: indexArray size.
	1 to: indexArray size do:
		[:index |
		newCollection at: index put: (self at: (indexArray at: index))].
	^ newCollection! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 5/12/2000 12:18'!
atAll: indexArray putAll: valueArray
	"Store the elements of valueArray into the slots
	of this collection selected by indexArray."

	indexArray with: valueArray do: [:index :value | self at: index put: value].
	^ valueArray! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 5/12/2000 12:17'!
atAll: aCollection put: anObject 
	"Put anObject at every index specified by the elements of aCollection."

	aCollection do: [:index | self at: index put: anObject].
	^ anObject! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'nk 6/12/2004 17:06'!
atLast: indexFromEnd
	"Return element at indexFromEnd from the last position.
	 atLast: 1, returns the last element"

	^ self atLast: indexFromEnd ifAbsent: [self error: 'index out of range']! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ajh 6/27/2002 17:52'!
atLast: indexFromEnd ifAbsent: block
	"Return element at indexFromEnd from the last position.
	 atLast: 1 ifAbsent: [] returns the last element"

	^ self at: self size + 1 - indexFromEnd ifAbsent: block! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ajh 6/27/2002 18:10'!
atLast: indexFromEnd put: obj
	"Set the element at indexFromEnd from the last position.
	 atLast: 1 put: obj, sets the last element"

	^ self at: self size + 1 - indexFromEnd put: obj! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'di 11/6/1998 14:32'!
atPin: index 
	"Return the index'th element of me if possible.
	Return the first or last element if index is out of bounds."

	index < 1 ifTrue: [^ self first].
	index > self size ifTrue: [^ self last].
	^ self at: index! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 5/12/2000 12:50'!
atRandom: aGenerator
	"Answer a random element of the receiver.  Uses aGenerator which
	should be kept by the user in a variable and used every time. Use
	this instead of #atRandom for better uniformity of random numbers 
	because only you use the generator.  Causes an error if self has no 
	elements."

	^ self at: (aGenerator nextInt: self size)! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 5/12/2000 13:51'!
atWrap: index 
	"Answer the index'th element of the receiver.  If index is out of bounds,
	let it wrap around from the end to the beginning until it is in bounds."

	^ self at: index - 1 \\ self size + 1! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 5/12/2000 13:52'!
atWrap: index put: value
	"Store value into the index'th element of the receiver.  If index is out
	of bounds, let it wrap around from the end to the beginning until it 
	is in bounds. Answer value."

	^ self at: index  - 1 \\ self size + 1 put: value! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ul 9/17/2009 06:00'!
at: index ifAbsent: exceptionBlock 
	"Answer the element at my position index. If I do not contain an element 
	at index, answer the result of evaluating the argument, exceptionBlock."

	(index <= self size  and: [ 1 <= index ]) ifTrue: [ ^self at: index ].
	^exceptionBlock value! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'raok 11/22/2002 12:34'!
at: index incrementBy: value
	^self at: index put: (self at: index) + value! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 5/12/2000 18:04'!
before: target
	"Answer the receiver's element immediately before target. Raise an
	error if target is not an element of the receiver, or if there are no 
	elements before it (i.e. it is the first element)."

	^ self before: target ifAbsent: [self errorNotFound: target]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ul 11/15/2010 10:06'!
before: target ifAbsent: exceptionBlock
	"Answer the receiver's element immediately before target. Answer
	the result of evaluating the exceptionBlock if target is not an element
	of the receiver, or if there are no elements before it."

	| index |
	index := self indexOf: target.
	^ (index = 0 or: [index = 1])
		ifTrue: [exceptionBlock value]
		ifFalse: [self at: index - 1]! !

!SequenceableCollection methodsFor: 'testing' stamp: 'ar 2/1/2010 19:06'!
beginsWithAnyOf: aCollection
	"Return true if the receiver starts with any of the elements in aCollection."
	^aCollection anySatisfy:[:prefix| self beginsWith: prefix].! !

!SequenceableCollection methodsFor: 'testing' stamp: 'ul 4/1/2011 04:03'!
beginsWith: sequence
	"Answer true if the receiver starts with the argument collection."
	
	| sequenceSize |
	((sequenceSize := sequence size) = 0 or: [ self size < sequence size ]) ifTrue: [ ^false ].
	1 to: sequenceSize do: [ :index |
		(sequence at: index) = (self at: index) ifFalse: [ ^false ] ].
	^true! !

!SequenceableCollection methodsFor: 'private' stamp: 'sma 5/12/2000 13:57'!
checkedAt: index
	index > self size ifTrue: [self error: 'not enough elements'].
	^ self at: index! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'di 7/13/97 09:44'!
collectWithIndex: elementAndIndexBlock
	"Use the new version with consistent naming"
	^ self withIndexCollect: elementAndIndexBlock! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 11:46'!
collect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into a collection like the receiver. Answer  
	the new collection."

	| newCollection |
	newCollection := self species new: self size.
	1 to: self size do:
		[:index |
		newCollection at: index put: (aBlock value: (self at: index))].
	^ newCollection! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ul 3/15/2010 03:50'!
collect: aBlock from: firstIndex to: lastIndex
	"Refer to the comment in Collection|collect:."

	| size result |
	size := lastIndex - firstIndex + 1.
	result := self species new: size.
	1 to: size do: [ :index |
		result at: index put: (aBlock value: (self at: index + firstIndex - 1)) ].
	^result! !

!SequenceableCollection methodsFor: 'private' stamp: 'tk 7/30/97 12:42'!
combinationsAt: jj in: aCollection after: nn do: aBlock
	"Choose k of N items and put in aCollection.  jj-1 already chosen.  Indexes of items are in numerical order, to avoid the same combo being used twice.  In this slot, we are allowed to use items in self indexed by nn+1 to self size.  nn is the index used for position jj-1."
	"(1 to: 6) combinationsSize: 3 do: [:each | Transcript cr; show: each printString]"

nn+1 to: self size do: [:index | 
		aCollection at: jj put: (self at: index).
		jj = aCollection size 
			ifTrue: [aBlock value: aCollection]
			ifFalse: [self combinationsAt: jj + 1 in: aCollection after: index do: aBlock]].! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'nice 11/1/2011 22:19'!
combinations: kk atATimeDo: aBlock
	"Take the items in the receiver, kk at a time, and evaluate the block for each combination.  Hand in an array of elements of self as the block argument.  Each combination only occurs once, and order of the elements does not matter.  There are (self size take: kk) combinations."
	" (1 to: 5) combinations: 3 atATimeDo: [:each | Transcript cr; show: each printString]"

	| aCollection |
	aCollection := Array new: kk.
	self combinationsAt: 1 in: aCollection after: 0 do: aBlock! !

!SequenceableCollection methodsFor: 'converting' stamp: 'mt 1/14/2015 11:34'!
concatenation
	"Flattens the collection by one level into an Array. Avoids using #gather: and #streamContents: for performance reasons."

	| result index |
	result := Array new: (self inject: 0 into: [:sum :each | sum + each size]).
	index := 0.
	self do: [:each | each do: [:item | result at: (index := index+1) put: item]].
	^ result! !

!SequenceableCollection methodsFor: 'copying' stamp: 'sma 6/1/2000 16:05'!
copyAfterLast: anElement
	"Answer a copy of the receiver from after the last occurence
	of anElement up to the end. If no such element exists, answer 
	an empty copy."

	^ self allButFirst: (self lastIndexOf: anElement ifAbsent: [^ self copyEmpty])! !

!SequenceableCollection methodsFor: 'copying' stamp: 'sma 6/1/2000 16:05'!
copyAfter: anElement
	"Answer a copy of the receiver from after the first occurence
	of anElement up to the end. If no such element exists, answer 
	an empty copy."

	^ self allButFirst: (self indexOf: anElement ifAbsent: [^ self copyEmpty])! !

!SequenceableCollection methodsFor: 'copying' stamp: 'sma 6/1/2000 16:07'!
copyEmpty
	^ self species new: 0! !

!SequenceableCollection methodsFor: 'copying' stamp: ''!
copyFrom: start to: stop 
	"Answer a copy of a subset of the receiver, starting from element at 
	index start until element at index stop."

	| newSize |
	newSize := stop - start + 1.
	^(self species new: newSize)
		replaceFrom: 1
		to: newSize
		with: self
		startingAt: start! !

!SequenceableCollection methodsFor: 'copying' stamp: 'sma 6/1/2000 16:06'!
copyLast: num
	"Deprecated. Use #last:"

	^ self last: num! !

!SequenceableCollection methodsFor: 'copying' stamp: ''!
copyReplaceAll: oldSubstring with: newSubstring 
	"Default is not to do token matching.
	See also String copyReplaceTokens:with:"
	^ self copyReplaceAll: oldSubstring with: newSubstring asTokens: false
	"'How now brown cow?' copyReplaceAll: 'ow' with: 'ello'"
	"'File asFile Files File''s File' copyReplaceTokens: 'File' with: 'Pile'"! !

!SequenceableCollection methodsFor: 'private' stamp: 'yo 9/2/2002 18:22'!
copyReplaceAll: oldSubstring with: newSubstring asTokens: ifTokens
	"Answer a copy of the receiver in which all occurrences of
	oldSubstring have been replaced by newSubstring.
	ifTokens (valid for Strings only) specifies that the characters
	surrounding the recplacement must not be alphanumeric.
		Bruce Simth,  must be incremented by 1 and not 
	newSubstring if ifTokens is true.  See example below. "

	| aString startSearch currentIndex endIndex |
	(ifTokens and: [(self isString) not])
		ifTrue: [(self isKindOf: Text) ifFalse: [
			self error: 'Token replacement only valid for Strings']].
	aString := self.
	startSearch := 1.
	[(currentIndex := aString indexOfSubCollection: oldSubstring startingAt: startSearch)
			 > 0]
		whileTrue: 
		[endIndex := currentIndex + oldSubstring size - 1.
		(ifTokens not
			or: [(currentIndex = 1
					or: [(aString at: currentIndex-1) isAlphaNumeric not])
				and: [endIndex = aString size
					or: [(aString at: endIndex+1) isAlphaNumeric not]]])
			ifTrue: [aString := aString
					copyReplaceFrom: currentIndex
					to: endIndex
					with: newSubstring.
				startSearch := currentIndex + newSubstring size]
			ifFalse: [
				ifTokens 
					ifTrue: [startSearch := currentIndex + 1]
					ifFalse: [startSearch := currentIndex + newSubstring size]]].
	^ aString

"Test case:
	'test te string' copyReplaceAll: 'te' with: 'longone' asTokens: true   "
! !

!SequenceableCollection methodsFor: 'copying' stamp: 'ar 10/16/2001 19:03'!
copyReplaceFrom: start to: stop with: replacementCollection 
	"Answer a copy of the receiver satisfying the following conditions: If 
	stop is less than start, then this is an insertion; stop should be exactly 
	start-1, start = 1 means insert before the first character, start = size+1 
	means append after last character. Otherwise, this is a replacement; start 
	and stop have to be within the receiver's bounds."

	| newSequenceableCollection newSize endReplacement |
	newSize := self size - (stop - start + 1) + replacementCollection size.
	endReplacement := start - 1 + replacementCollection size.
	newSequenceableCollection := self species new: newSize.
	start > 1 ifTrue:[
		newSequenceableCollection
			replaceFrom: 1
			to: start - 1
			with: self
			startingAt: 1].
	start <= endReplacement ifTrue:[
		newSequenceableCollection
			replaceFrom: start
			to: endReplacement
			with: replacementCollection
			startingAt: 1].
	endReplacement < newSize ifTrue:[
		newSequenceableCollection
			replaceFrom: endReplacement + 1
			to: newSize
			with: self
			startingAt: stop + 1].
	^newSequenceableCollection! !

!SequenceableCollection methodsFor: 'copying' stamp: 'ul 4/3/2015 20:50'!
copyUpThrough: anElement 
	"Answer all elements up to and including anObject. If there
	is no such object, answer a copy of the receiver."

	| index |
	index := self indexOf: anElement.
	index = 0 ifTrue: [ ^self copy ].
	^self first: index! !

!SequenceableCollection methodsFor: 'copying' stamp: 'ul 4/3/2015 20:50'!
copyUpToLast: anElement
	"Answer a copy of the receiver from index 1 to the last occurrence of 
	anElement, not including anElement."

	| index |
	index := self lastIndexOf: anElement.
	index = 0 ifTrue: [ ^self copy ].
	^self first: index - 1! !

!SequenceableCollection methodsFor: 'copying' stamp: 'ul 4/3/2015 20:51'!
copyUpTo: anElement 
	"Answer all elements up to but not including anObject. If there
	is no such object, answer a copy of the receiver."

	| index |
	index := self indexOf: anElement.
	index = 0 ifTrue: [ ^self copy ].
	^self first: index - 1! !

!SequenceableCollection methodsFor: 'copying' stamp: 'ajh 9/27/2002 12:09'!
copyWithFirst: newElement 
	"Answer a copy of the receiver that is 1 bigger than the receiver with newElement as the first element."

	| newIC |
	newIC := self species ofSize: self size + 1.
	newIC 
		replaceFrom: 2
		to: self size + 1
		with: self
		startingAt: 1.
	newIC at: 1 put: newElement.
	^ newIC! !

!SequenceableCollection methodsFor: 'copying' stamp: 'sma 6/1/2000 15:38'!
copyWithoutFirst
	"Deprecatd. Return a copy of the receiver which doesn't include
	the first element."

	^ self allButFirst! !

!SequenceableCollection methodsFor: 'copying' stamp: 'rhi 12/6/2001 14:04'!
copyWithoutIndex: index
	"Return a copy containing all elements except the index-th."

	| copy |
	copy := self species ofSize: self size - 1.
	copy replaceFrom: 1 to: index-1 with: self startingAt: 1.
	copy replaceFrom: index to: copy size with: self startingAt: index+1.
	^ copy! !

!SequenceableCollection methodsFor: 'copying' stamp: ''!
copyWith: newElement 
	"Answer a copy of the receiver that is 1 bigger than the receiver and has 
	newElement at the last element."

	| newIC |
	newIC := self species new: self size + 1.
	newIC 
		replaceFrom: 1
		to: self size
		with: self
		startingAt: 1.
	newIC at: newIC size put: newElement.
	^newIC! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'di 7/13/97 09:43'!
doWithIndex: elementAndIndexBlock
	"Use the new version with consistent naming"
	^ self withIndexDo: elementAndIndexBlock! !

!SequenceableCollection methodsFor: 'enumerating' stamp: ''!
do: aBlock 
	"Refer to the comment in Collection|do:."
	1 to: self size do:
		[:index | aBlock value: (self at: index)]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 11:56'!
do: elementBlock separatedBy: separatorBlock
	"Evaluate the elementBlock for all elements in the receiver,
	and evaluate the separatorBlock between."

	1 to: self size do:
		[:index |
		index = 1 ifFalse: [separatorBlock value].
		elementBlock value: (self at: index)]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ar 5/1/1999 05:01'!
do: aBlock without: anItem
	"Enumerate all elements in the receiver.
	Execute aBlock for those elements that are not equal to the given item"
	"Refer to the comment in Collection|do:."
	1 to: self size do:
		[:index | anItem = (self at: index) ifFalse:[aBlock value: (self at: index)]]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'JMM 10/30/2005 09:21'!
eighth
	"Answer the eighth element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 8! !

!SequenceableCollection methodsFor: 'testing' stamp: 'ar 2/1/2010 19:10'!
endsWithAnyOf: aCollection
	"Return true if the receiver ends with any of the elements in aCollection."
	^aCollection anySatisfy:[:suffix| self endsWith: suffix].! !

!SequenceableCollection methodsFor: 'testing' stamp: 'ul 4/1/2011 04:42'!
endsWith: sequence
	"Answer true if the receiver ends with the argument collection."
	
	| sequenceSize offset |
	((sequenceSize := sequence size) = 0 or: [ (offset := self size - sequence size) < 0 ]) ifTrue: [ ^false ].
	1 to: sequenceSize do: [ :index |
		(sequence at: index) = (self at: index + offset) ifFalse: [ ^false ] ].
	^true! !

!SequenceableCollection methodsFor: 'private' stamp: 'sma 5/12/2000 18:06'!
errorFirstObject: anObject
	self error: 'specified object is first object'! !

!SequenceableCollection methodsFor: 'private' stamp: 'sma 5/12/2000 18:03'!
errorLastObject: anObject
	self error: 'specified object is last object'! !

!SequenceableCollection methodsFor: 'private' stamp: ''!
errorOutOfBounds

	self error: 'indices are out of bounds'! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'JMM 10/30/2005 09:21'!
fifth
	"Answer the fifth element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 5! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'jmv 5/10/2010 14:03'!
findBinaryIndex: aBlock
	"Search for an element in the receiver using binary search.
	The argument aBlock is a one-element block returning
		0 	- if the element is the one searched for
		<0	- if the search should continue in the first half
		>0	- if the search should continue in the second half
	If no matching element is found, raise an error.
	Examples:
		#(1 3 5 7 11 15 23) findBinaryIndex: [ :arg | 11 - arg ]
	"
	^self findBinaryIndex: aBlock do: [ :found | found ] ifNone: [ self errorNotFound: aBlock]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ul 11/17/2010 12:15'!
findBinaryIndex: aBlock do: actionBlock ifNone: exceptionBlock
	"Search for an element in the receiver using binary search.
	The argument aBlock is a one-element block returning
		0 	- if the element is the one searched for
		<0	- if the search should continue in the first half
		>0	- if the search should continue in the second half
	If found, evaluate actionBlock with the index as argument
	If no matching element is found, evaluate exceptionBlock,
	with the indexes of the 'bounding' elements as optional
	arguments. 	Warning: Might give invalid indexes, see
	examples below.
	Examples:
		#(1 3 5 7 11 15 23)
			findBinaryIndex: [ :arg | 11 - arg ]
			do: [ :found | found ]
			ifNone: [ :a :b | ('between: ', {a. b} printString)]
		#(1 3 5 7 11 15 23)
			findBinaryIndex: [ :arg | 12 - arg ]
			do: [ :found | found ]
			ifNone: [ :a :b | ('between: ', {a. b} printString) ]
		#(1 3 5 7 11 15 23) d
			findBinaryIndex: [ :arg | 0.5 - arg ]
			do: [ :found | found ]
			ifNone: [ :a :b | ('between: ', {a. b} printString) ]
		#(1 3 5 7 11 15 23)
			findBinaryIndex: [ :arg | 25 - arg ]
			do: [ :found | found ]
			ifNone: [ :a :b | ('between: ',{a. b} printString) ]
	"
	| index low high |
	low := 1.
	high := self size.
	[
		index := high + low // 2.
		low > high ] whileFalse: [
			| test |
			test := aBlock value: (self at: index).
			test = 0 
				ifTrue: [ ^actionBlock value: index ]
				ifFalse: [ test > 0
					ifTrue: [ low := index + 1 ]
					ifFalse: [ high := index - 1 ] ] ].
	^exceptionBlock cull: high cull: low! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ul 11/17/2010 12:13'!
findBinaryIndex: aBlock ifNone: exceptionBlock
	"Search for an element in the receiver using binary search.
	The argument aBlock is a one-element block returning
		0 	- if the element is the one searched for
		<0	- if the search should continue in the first half
		>0	- if the search should continue in the second half
	If no matching element is found, evaluate exceptionBlock,
	with the indexes of the 'bounding' elements as optional
	arguments.	Warning: Might give invalid indexes."

	^self findBinaryIndex: aBlock do: [ :found | found ] ifNone: exceptionBlock! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'jmv 5/10/2010 14:05'!
findBinary: aBlock
	"Search for an element in the receiver using binary search.
	The argument aBlock is a one-element block returning
		0 	- if the element is the one searched for
		<0	- if the search should continue in the first half
		>0	- if the search should continue in the second half
	If no matching element is found, raise an error.
	Examples:
		#(1 3 5 7 11 15 23) findBinary: [ :arg | 11 - arg ]
	"
	^self findBinary: aBlock do: [ :found | found ] ifNone: [ self errorNotFound: aBlock ]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ul 11/17/2010 12:15'!
findBinary: aBlock do: actionBlock ifNone: exceptionBlock
	"Search for an element in the receiver using binary search.
	The argument aBlock is a one-element block returning
		0 	- if the element is the one searched for
		<0	- if the search should continue in the first half
		>0	- if the search should continue in the second half
	If found, evaluate actionBlock with the found element as argument
	If no matching element is found, evaluate exceptionBlock,
	with the 'bounding' elements (or nil) as optional arguments.
	Examples:
		#(1 3 5 7 11 15 23)
			findBinary: [ :arg | 11 - arg ]
			do: [ :found | found ]
			ifNone: [ :a :b | ('between: ', {a. b} printString) ]
		#(1 3 5 7 11 15 23)
			findBinary: [ :arg | 12 - arg ]
			do: [ :found | found ]
			ifNone: [ :a :b | ('between: ', {a. b} printString) ]
		#(1 3 5 7 11 15 23)
			findBinary: [ :arg | 0.5 - arg ]
			do: [ :found | found ]
			ifNone: [ :a :b | ('between: ', {a. b} printString) ]
		#(1 3 5 7 11 15 23)
			findBinary: [ :arg | 25 - arg ]
			do: [ :found | found ]
			ifNone: [ :a :b | ('between: ',{a. b} printString) ]
	"
	^self
		findBinaryIndex: aBlock
		do: [ :foundIndex | actionBlock value: (self at: foundIndex) ]
		ifNone: [ :prevIndex :nextIndex |
			exceptionBlock
				cull: (prevIndex > 0 ifTrue: [ self at: prevIndex ])
				cull: (nextIndex <= self size ifTrue: [ self at: nextIndex ]) ]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ul 11/17/2010 12:13'!
findBinary: aBlock ifNone: exceptionBlock
	"Search for an element in the receiver using binary search.
	The argument aBlock is a one-element block returning
		0 	- if the element is the one searched for
		<0	- if the search should continue in the first half
		>0	- if the search should continue in the second half
	If no matching element is found, evaluate exceptionBlock,
	with the 'bounding' elements (or nil) as optional arguments."
	
	^self findBinary: aBlock do: [ :found | found ] ifNone: exceptionBlock! !

!SequenceableCollection methodsFor: 'enumerating' stamp: ''!
findFirst: aBlock
	"Return the index of my first element for which aBlock evaluates as true."

	| index |
	index := 0.
	[(index := index + 1) <= self size] whileTrue:
		[(aBlock value: (self at: index)) ifTrue: [^index]].
	^ 0! !

!SequenceableCollection methodsFor: 'enumerating' stamp: ''!
findLast: aBlock
	"Return the index of my last element for which aBlock evaluates as true."

	| index |
	index := self size + 1.
	[(index := index - 1) >= 1] whileTrue:
		[(aBlock value: (self at: index)) ifTrue: [^index]].
	^ 0! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'eem 5/19/2010 09:57'!
findLast: aBlock startingAt: i
	"Return the index of my last element with index >= i for which aBlock evaluates as true."

	| index |
	index := self size + 1.
	[(index := index - 1) >= i] whileTrue:
		[(aBlock value: (self at: index)) ifTrue: [^index]].
	^ 0! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ul 9/10/2011 15:21'!
findNearbyBinaryIndex: aBlock
	"Search for an element in the receiver using binary search.
	The argument aBlock is a one-element block returning
		0 	- if the element is the one searched for
		<0	- if the search should continue in the first half
		>0	- if the search should continue in the second half
	If no matching element is found, answer the closest index we could find,
	answering 0 if the element should preceed all items in the collection,
	and size + 1 if the element should follow all items in the collection."
	
	^self
		findBinaryIndex: aBlock
		ifNone: [ :lower :upper |
			lower = self size
				ifTrue: [ upper ]
				ifFalse: [ lower ] ]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'md 1/19/2006 09:57'!
first
	"Answer the first element of the receiver"

	^ self at: 1! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 6/1/2000 15:31'!
first: n
	"Answer the first n elements of the receiver.
	Raise an error if there are not enough elements."

	^ self copyFrom: 1 to: n! !

!SequenceableCollection methodsFor: 'converting' stamp: 'mt 4/12/2016 10:05'!
flatten
	"Similar to #concatenation but removes all nesting except for strings.
Example: {3 .4 .{2 .4 .{'hi'} .'ho'}} flatten = {3 .4 .2 .4 .'hi' .'ho'}"

	^ Array streamContents: [:stream |
		self do: [:each |
			((each isCollection and: [each isString not]) or: [each isStream])
				ifFalse: [stream nextPut: each]
				ifTrue: [stream nextPutAll: each flatten]]]! !

!SequenceableCollection methodsFor: 'converting' stamp: 'topa 1/16/2015 08:25'!
flattened
	"An alias for #flatten
	This message's name is in line with messages like #sorted or #reversed
	while #flatten's is in line with #reverse (as per ANSI, see comment there)"

	^ self flatten! !

!SequenceableCollection methodsFor: 'copying' stamp: 'nice 2/26/2009 11:48'!
forceTo: length paddingStartWith: elem 
	"Force the length of the collection to length, padding  
	the beginning of the result if necessary with elem.  
	Note that this makes a copy."
	| newCollection padLen |
	newCollection := self species ofSize: length.
	padLen := length - self size max: 0.
	newCollection
		from: 1
		to: padLen
		put: elem.
	newCollection
		replaceFrom: padLen + 1
		to: ((padLen + self size) min: length)
		with: self
		startingAt:  1.
	^ newCollection! !

!SequenceableCollection methodsFor: 'copying' stamp: 'nice 2/24/2009 11:33'!
forceTo: length paddingWith: elem
	"Force the length of the collection to length, padding
	if necessary with elem.  Note that this makes a copy."

	| newCollection |
	newCollection := self species new: length withAll: elem.
	newCollection replaceFrom: 1 to: (self size min: length) with: self startingAt: 1.
	^ newCollection! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'JMM 10/30/2005 09:21'!
fourth
	"Answer the fourth element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 4! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 18:11'!
from: start to: stop do: aBlock
	"Evaluate aBlock for all elements between start and stop (inclusive)."

	start to: stop do: [:index | aBlock value: (self at: index)]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'SqR 10/30/2000 22:06'!
from: startIndex to: endIndex put: anObject
	"Put anObject in all indexes between startIndex 
	and endIndex. Very fast. Faster than to:do: for
	more than 26 positions. Answer anObject"

	| written toWrite thisWrite |

	startIndex > endIndex ifTrue: [^self].
	self at: startIndex put: anObject.
	written := 1.
	toWrite := endIndex - startIndex + 1.
	[written < toWrite] whileTrue:
		[
			thisWrite := written min: toWrite - written.
			self 
				replaceFrom: startIndex + written
				to: startIndex + written + thisWrite - 1
				with: self startingAt: startIndex.
			written := written + thisWrite
		].
	^anObject! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ul 12/2/2014 18:27'!
groupsDo: aBlock 
	"Evaluate aBlock with my elements taken n at a time, where n is the number of arguments of aBlock. Ignore any leftovers at the end."
	
	| index argumentArray numArgs endIndex |
	numArgs := aBlock numArgs.
	numArgs
		caseOf: { 
			[ 0 ] -> [ ^self error: 'At least one block argument expected.' ].
			[ 1 ] -> [ ^self do: aBlock ].
			[ 2 ] -> [ ^self pairsDo: aBlock ] }
		otherwise: [].
	argumentArray := Array new: numArgs.
	index := 1.
	endIndex := self size - numArgs + 1.
	[ index <= endIndex ] whileTrue: [
		argumentArray
			replaceFrom: 1
			to: numArgs
			with: self
			startingAt: index.
		aBlock valueWithArguments: argumentArray.
		index := index + numArgs ].! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'nice 12/26/2009 21:21'!
groupsOf: n atATimeCollect: aBlock 
	"Evaluate aBlock with my elements taken n at a time. Ignore any 
	leftovers at the end. 
	Allows use of a flattened  
	array for things that naturally group into groups of n. 
	If aBlock has a single argument, pass it an array of n items, 
	otherwise, pass the items as separate arguments. 
	See also pairsDo:"
	| passArray |
	passArray := aBlock numArgs = 1.
	^(n
		to: self size
		by: n)
		collect: [:index | 
			| args |
			args := (self copyFrom: index - n + 1 to: index) asArray.
			passArray
				ifTrue: [aBlock value: args]
				ifFalse: [aBlock valueWithArguments: args]]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'nice 12/26/2009 21:21'!
groupsOf: n atATimeDo: aBlock 
	"Evaluate aBlock with my elements taken n at a time. Ignore any leftovers at the end.
	Allows use of a flattened 
	array for things that naturally group into groups of n.
	If aBlock has a single argument, pass it an array of n items,
	otherwise, pass the items as separate arguments.
	See also pairsDo:"
	| passArray |
	passArray := (aBlock numArgs = 1).
	n
		to: self size
		by: n
		do: [:index | 
			| args |
			args := (self copyFrom: index - n + 1 to: index) asArray.
			passArray ifTrue: [ aBlock value: args ]
				ifFalse: [ aBlock valueWithArguments: args ]].! !

!SequenceableCollection methodsFor: 'copying' stamp: 'cmm 1/3/2016 15:32'!
grownBy: length 
	"Answer a copy of receiver collection with size grown by length"
	^ (self class ofSize: self size + length)
		replaceFrom: 1 to: self size with: self startingAt: 1 ;
		yourself! !

!SequenceableCollection methodsFor: 'comparing' stamp: 'eem 2/3/2015 10:27'!
hasEqualElements: otherCollection
	"Answer whether the receiver's size is the same as otherCollection's
	size, and each of the receiver's elements equal the corresponding 
	element of otherCollection.
	This should probably replace the current definition of #= ."

	| size |
	otherCollection isSequenceable ifFalse: [^ false].
	(size := self size) = otherCollection size ifFalse: [^ false].
	1 to: size do:
		[:index |
		(self at: index) = (otherCollection at: index) ifFalse: [^ false]].
	^ true! !

!SequenceableCollection methodsFor: 'comparing' stamp: 'SqR 8/3/2000 13:39'!
hash
	| hash |

	hash := self species hash.
	1 to: self size do: [:i | hash := (hash + (self at: i) hash) hashMultiply].
	^hash! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ar 8/14/1998 21:20'!
identityIndexOf: anElement 
	"Answer the index of anElement within the receiver. If the receiver does 
	not contain anElement, answer 0."

	^self identityIndexOf: anElement ifAbsent: [0]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ar 8/14/1998 21:21'!
identityIndexOf: anElement ifAbsent: exceptionBlock
	"Answer the index of anElement within the receiver. If the receiver does 
	not contain anElement, answer the result of evaluating the argument, 
	exceptionBlock."
	1 to: self size do:
		[:i | (self at: i) == anElement ifTrue: [^ i]].
	^ exceptionBlock value! !

!SequenceableCollection methodsFor: 'testing' stamp: 'sma 5/12/2000 14:08'!
includes: anObject
	"Answer whether anObject is one of the receiver's elements."

	^ (self indexOf: anObject) ~= 0! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'nice 12/9/2009 14:55'!
indexOfAnyOf: aCollection
	"Answer the index of the first occurence of any element included in aCollection within the receiver.
	If the receiver does not contain anElement, answer zero, which is an invalid index."

	^self indexOfAnyOf: aCollection startingAt: 1 ifAbsent: [0]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'nice 12/9/2009 14:55'!
indexOfAnyOf: aCollection ifAbsent: exceptionBlock
	"Answer the index of the first occurence of any element included in aCollection within the receiver.
	If the receiver does not contain anElement, answer the result of evaluating the argument, exceptionBlock."

	^self indexOfAnyOf: aCollection startingAt: 1 ifAbsent: exceptionBlock! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'nice 12/9/2009 14:54'!
indexOfAnyOf: aCollection startingAt: start
	"Answer the index of the first occurence of any element included in aCollection after start within the receiver.
	If the receiver does not contain anElement, answer zero, which is an invalid index."

	^self indexOfAnyOf: aCollection startingAt: start ifAbsent: [0]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'nice 12/9/2009 14:57'!
indexOfAnyOf: aCollection startingAt: start ifAbsent: exceptionBlock
	"Answer the index of the first occurence of any element included in aCollection after start within the receiver.
	If the receiver does not contain anElement, answer the result of evaluating the argument, exceptionBlock.
	Note: it is user responsibility to provide aCollection that behaves relatevily fast when asked for includes: (like a Set)"

	start to: self size do:
		[:index |
		(aCollection includes: (self at: index)) ifTrue: [^ index]].
	^ exceptionBlock value! !

!SequenceableCollection methodsFor: 'accessing' stamp: ''!
indexOfSubCollection: aSubCollection startingAt: anIndex 
	"Answer the index of the receiver's first element, such that that element 
	equals the first element of aSubCollection, and the next elements equal 
	the rest of the elements of aSubCollection. Begin the search at element 
	anIndex of the receiver. If no such match is found, answer 0."

	^self
		indexOfSubCollection: aSubCollection
		startingAt: anIndex
		ifAbsent: [0]! !

!SequenceableCollection methodsFor: 'accessing' stamp: ''!
indexOfSubCollection: sub startingAt: start ifAbsent: exceptionBlock
	"Answer the index of the receiver's first element, such that that element 
	equals the first element of sub, and the next elements equal 
	the rest of the elements of sub. Begin the search at element 
	start of the receiver. If no such match is found, answer the result of 
	evaluating argument, exceptionBlock."
	| first index |
	sub isEmpty ifTrue: [^ exceptionBlock value].
	first := sub first.
	start to: self size - sub size + 1 do:
		[:startIndex |
		(self at: startIndex) = first ifTrue:
			[index := 1.
			[(self at: startIndex+index-1) = (sub at: index)]
				whileTrue:
				[index = sub size ifTrue: [^startIndex].
				index := index+1]]].
	^ exceptionBlock value! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 6/1/2000 15:46'!
indexOf: anElement
	"Answer the index of the first occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer 0."

	^ self indexOf: anElement ifAbsent: [0]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 6/1/2000 15:47'!
indexOf: anElement ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the 
	result of evaluating the argument, exceptionBlock."

	^ self indexOf: anElement startingAt: 1 ifAbsent: exceptionBlock! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ul 9/12/2010 04:34'!
indexOf: anElement startingAt: start
	"Answer the index of the first occurence of anElement after start
	within the receiver. If the receiver does not contain anElement, 
	answer 0."

	^self indexOf: anElement startingAt: start ifAbsent: 0! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 6/1/2000 15:47'!
indexOf: anElement startingAt: start ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement after start
	within the receiver. If the receiver does not contain anElement, 
	answer the 	result of evaluating the argument, exceptionBlock."

	start to: self size do:
		[:index |
		(self at: index) = anElement ifTrue: [^ index]].
	^ exceptionBlock value! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ar 3/3/2001 22:43'!
integerAt: index
	"Return the integer at the given index"
	^self at: index! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ar 3/3/2001 22:43'!
integerAt: index put: value
	"Return the integer at the given index"
	^self at: index put: value! !

!SequenceableCollection methodsFor: 'testing' stamp: 'di 11/6/1998 09:35'!
isSequenceable
	^ true! !

!SequenceableCollection methodsFor: 'converting' stamp: 'mt 1/14/2015 12:45'!
join
	"Example: #(H e l l o W o r l d) join = 'HelloWorld'.  "

	^ self joinSeparatedBy: ''! !

!SequenceableCollection methodsFor: 'converting' stamp: 'mt 1/14/2015 12:41'!
joinSeparatedBy: aSeparator
	"Returns a string, which is a concatenation of each element's string representation separated by another string."

	^ String streamContents: [:stream |
		self
			do: [:ea | stream nextPutAll: ea asString]
			separatedBy: [stream nextPutAll: aSeparator asString]]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'sma 5/12/2000 18:13'!
keysAndValuesDo: aBlock 
	"Enumerate the receiver with all the keys (aka indices) and values."

	1 to: self size do: [:index | aBlock value: index value: (self at: index)]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'md 1/19/2006 09:56'!
last
	"Answer the last element of the receiver"

	^ self at: self size! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'nice 11/16/2009 10:31'!
lastIndexOfAnyOf: aCollection startingAt: lastIndex ifAbsent: exceptionBlock
	"Answer the index of the last occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the
	result of evaluating the argument, exceptionBlock."

	lastIndex to: 1 by: -1 do:
		[:index |
		(aCollection includes: (self at: index)) ifTrue: [^ index]].
	^ exceptionBlock value! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ar 12/16/2001 01:06'!
lastIndexOf: anElement
	"Answer the index of the last occurence of anElement within the 
	receiver. If the receiver does not contain anElement, answer 0."

	^ self lastIndexOf: anElement startingAt: self size ifAbsent: [0]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ar 12/16/2001 01:06'!
lastIndexOf: anElement ifAbsent: exceptionBlock
	"Answer the index of the last occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the
	result of evaluating the argument, exceptionBlock."
	^self lastIndexOf: anElement startingAt: self size ifAbsent: exceptionBlock! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ar 12/16/2001 01:05'!
lastIndexOf: anElement startingAt: lastIndex ifAbsent: exceptionBlock
	"Answer the index of the last occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the
	result of evaluating the argument, exceptionBlock."

	lastIndex to: 1 by: -1 do:
		[:index |
		(self at: index) = anElement ifTrue: [^ index]].
	^ exceptionBlock value! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'sma 6/1/2000 15:30'!
last: n
	"Answer the last n elements of the receiver.  
	Raise an error if there are not enough elements."

	| size |
	size := self size.
	^ self copyFrom: size - n + 1 to: size! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'md 1/19/2006 09:56'!
middle
	"Answer the middle element of the receiver."
	^ self at: self size // 2 + 1! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'nk 8/23/2003 10:42'!
nextToLast
	^self at: self size - 1! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'JMM 10/30/2005 09:21'!
ninth
	"Answer the ninth element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 9! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ul 3/1/2011 08:14'!
overlappingPairsCollect: aBlock 
	"Answer the result of evaluating aBlock with all of the overlapping pairs of my elements."
	| retval |
	retval := self species ofSize: self size - 1.
	1 to: self size - 1
		do: [:i | retval at: i put: (aBlock value: (self at: i) value: (self at: i + 1)) ].
	^retval! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'nk 7/17/2003 17:55'!
overlappingPairsDo: aBlock 
	"Emit overlapping pairs of my elements into aBlock"

	1 to: self size - 1
		do: [:i | aBlock value: (self at: i) value: (self at: i + 1)]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'nk 7/17/2003 17:54'!
overlappingPairsWithIndexDo: aBlock 
	"Emit overlapping pairs of my elements into aBlock, along with an index."

	1 to: self size - 1
		do: [:i | aBlock value: (self at: i) value: (self at: i + 1) value: i ]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'nk 3/30/2002 12:45'!
paddedWith: otherCollection do: twoArgBlock 
	"Evaluate twoArgBlock with corresponding elements from this collection and otherCollection.
	Missing elements from either will be passed as nil."
	1 to: (self size max: otherCollection size) do:
		[:index | twoArgBlock value: (self at: index ifAbsent: [])
				value: (otherCollection at: index ifAbsent: [])]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'di 11/12/1998 15:01'!
pairsCollect: aBlock 
	"Evaluate aBlock with my elements taken two at a time, and return an Array with the results"

	^ (1 to: self size // 2) collect:
		[:index | aBlock value: (self at: 2 * index - 1) value: (self at: 2 * index)]
"
#(1 'fred' 2 'charlie' 3 'elmer') pairsCollect:
	[:a :b | b, ' is number ', a printString]
"! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'di 11/12/1998 15:01'!
pairsDo: aBlock 
	"Evaluate aBlock with my elements taken two at a time.  If there's an odd number of items, ignore the last one.  Allows use of a flattened array for things that naturally group into pairs.  See also pairsCollect:"

	1 to: self size // 2 do:
		[:index | aBlock value: (self at: 2 * index - 1) value: (self at: 2 * index)]
"
#(1 'fred' 2 'charlie' 3 'elmer') pairsDo:
	[:a :b | Transcript cr; show: b, ' is number ', a printString]
"! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ward 7/28/97 09:41'!
permutationsDo: aBlock
	"Repeatly value aBlock with a single copy of the receiver. Reorder the copy
	so that aBlock is presented all (self size factorial) possible permutations."
	"(1 to: 4) permutationsDo: [:each | Transcript cr; show: each printString]"

	self shallowCopy permutationsStartingAt: 1 do: aBlock! !

!SequenceableCollection methodsFor: 'private' stamp: 'ward 7/28/97 09:38'!
permutationsStartingAt: anInteger do: aBlock
	"#(1 2 3 4) permutationsDo: [:each | Transcript cr; show: each printString]"

	anInteger > self size ifTrue: [^self].
	anInteger = self size ifTrue: [^aBlock value: self].
	anInteger to: self size do:
		[:i | self swap: anInteger with: i.
		self permutationsStartingAt: anInteger + 1 do: aBlock.
		self swap: anInteger with: i]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'eem 4/28/2012 17:00'!
piecesCutWhere: binaryBlock
	"Answer substrings of the receiver derived from cutting the receiver
	 at points where binaryBlock answers true for adjacent elements."

	| pieces |
	pieces := OrderedCollection new.
	self piecesCutWhere: binaryBlock
		do: [:piece|
			pieces add: piece].
	^pieces

	"'Now is the time for all good people to come to the aid of the cause of world peace.  It is just fine, even desirable, to love your country, if that means wanting it to play a beneficial role in the course of world events and be the best possible example of a good society.  But if it means wanting dominion over the rest of the world, it is not love but defensiveness or self-glorification, and will lead only to oblivion.'
		piecesCutWhere: [:a :b| a = $. and: [b isSeparator]]"! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'eem 4/28/2012 16:55'!
piecesCutWhere: binaryBlock do: pieceBlock
	"Evaluate pieceBlock with substrings of the receiver derived from cutting the
	 receiver at points where binaryBlock answers true for adjacent elements."

	| size lastCut this next |
	(size := self size) <= 1 ifTrue:
		[size = 1 ifTrue:
			[pieceBlock value: self].
		 ^self].
	lastCut := 1.
	this := self at: 1.
	2 to: size do:
		[:i|
		next := self at: i.
		(binaryBlock value: this value: next) ifTrue:
			[pieceBlock value: (self copyFrom: lastCut to: i - 1).
			lastCut := i].
		this := next].
	pieceBlock value: (self copyFrom: lastCut to: size)! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ul 8/4/2011 02:33'!
polynomialEval: thisX
	"Treat myself as the coeficients of a polynomial in X.  Evaluate it with thisX.  First element is the constant and last is the coeficient for the highest power."
	"  #(1 2 3) polynomialEval: 2   "   "is 3*X^2 + 2*X + 1 with X = 2"

	| index sum |
	sum := self at: (index := self size).
	[ (index := index - 1) >= 1 ] whileTrue: [
		sum := sum * thisX + (self at: index) ].
	^sum! !

!SequenceableCollection methodsFor: 'filter streaming' stamp: 'kph 9/27/2007 21:18'!
putOn: aStream

	self do: [ :each | each putOn: aStream ]! !

!SequenceableCollection methodsFor: 'converting' stamp: 'sma 5/12/2000 12:51'!
readStream
	^ ReadStream on: self! !

!SequenceableCollection methodsFor: 'removing' stamp: ''!
remove: oldObject ifAbsent: anExceptionBlock 
	"SequencableCollections cannot implement removing."

	self shouldNotImplement! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'ar 1/20/98 16:22'!
replaceAll: oldObject with: newObject 
	"Replace all occurences of oldObject with newObject"
	| index |
	index := self
				indexOf: oldObject
				startingAt: 1
				ifAbsent: [0].
	[index = 0]
		whileFalse: 
			[self at: index put: newObject.
			index := self
						indexOf: oldObject
						startingAt: index + 1
						ifAbsent: [0]]! !

!SequenceableCollection methodsFor: 'accessing' stamp: ''!
replaceFrom: start to: stop with: replacement 
	"This destructively replaces elements from start to stop in the receiver. 
	Answer the receiver itself. Use copyReplaceFrom:to:with: for 
	insertion/deletion which may alter the size of the result."

	replacement size = (stop - start + 1)
		ifFalse: [self error: 'Size of replacement doesnt match'].
	^self replaceFrom: start to: stop with: replacement startingAt: 1! !

!SequenceableCollection methodsFor: 'accessing' stamp: ''!
replaceFrom: start to: stop with: replacement startingAt: repStart 
	"This destructively replaces elements from start to stop in the receiver 
	starting at index, repStart, in the sequenceable collection, 
	replacementCollection. Answer the receiver. No range checks are 
	performed."

	| index repOff |
	repOff := repStart - start.
	index := start - 1.
	[(index := index + 1) <= stop]
		whileTrue: [self at: index put: (replacement at: repOff + index)]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ul 11/11/2009 19:38'!
replace: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into self."

	1 to: self size do: [ :index |
		self at: index put: (aBlock value: (self at: index)) ]! !

!SequenceableCollection methodsFor: 'converting' stamp: 'ul 3/15/2011 14:47'!
reverse
	"The ANSI standard (5.7.8.26) requires this method to return a copy of the receiver. If you want to reverse the collection in place, then use #reverseInPlace."
	
	^self reversed! !

!SequenceableCollection methodsFor: 'enumerating' stamp: ''!
reverseDo: aBlock
	"Evaluate aBlock with each of the receiver's elements as the argument, 
	starting with the last element and taking each in sequence up to the 
	first. For SequenceableCollections, this is the reverse of the enumeration 
	for do:."

	self size to: 1 by: -1 do: [:index | aBlock value: (self at: index)]! !

!SequenceableCollection methodsFor: 'converting' stamp: 'ul 3/15/2011 14:42'!
reversed
	"Answer a copy of the receiver with element order reversed."
	"Example: 'frog' reversed"

	| n result src |
	n := self size.
	result := self species ofSize: n.
	src := n + 1.
	1 to: n do: [:i | result at: i put: (self at: (src := src - 1))].
	^ result
! !

!SequenceableCollection methodsFor: 'converting' stamp: 'ul 4/11/2011 03:07'!
reverseInPlace
	"Reverse this collection in place."
	
	| start end |
	start := 1.
	end := self size.
	[ start < end ] whileTrue: [
		| temp |
		temp := self at: start.
		self
			at: start put: (self at: end);
			at: end put: temp.
		start := start + 1.
		end := end - 1 ]
! !

!SequenceableCollection methodsFor: 'enumerating' stamp: ''!
reverseWith: aSequenceableCollection do: aBlock 
	"Evaluate aBlock with each of the receiver's elements, in reverse order, 
	along with the  
	corresponding element, also in reverse order, from 
	aSequencableCollection. "

	self size ~= aSequenceableCollection size ifTrue: [^ self errorNoMatch].
	self size
		to: 1
		by: -1
		do: [:index | aBlock value: (self at: index)
				value: (aSequenceableCollection at: index)]! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'JMM 10/30/2005 09:21'!
second
	"Answer the second element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 2! !

!SequenceableCollection methodsFor: 'enumerating' stamp: ''!
select: aBlock 
	"Refer to the comment in Collection|select:."
	| aStream |
	aStream := WriteStream on: (self species new: self size).
	1 to: self size do: 
		[:index |
		(aBlock value: (self at: index))
			ifTrue: [aStream nextPut: (self at: index)]].
	^ aStream contents! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'JMM 10/30/2005 09:21'!
seventh
	"Answer the seventh element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 7! !

!SequenceableCollection methodsFor: 'shuffling' stamp: 'ul 12/17/2014 14:53'!
shuffle

	^self shuffleBy: ThreadSafeRandom value! !

!SequenceableCollection methodsFor: 'shuffling' stamp: 'ul 3/26/2010 16:48'!
shuffleBy: aRandom
	"Durstenfeld's version of the Fisher-Yates shuffle"

	self size to: 2 by: -1 do: [ :i | 
		self swap: i with: (aRandom nextInt: i) ]! !

!SequenceableCollection methodsFor: 'shuffling' stamp: 'ul 4/3/2015 20:52'!
shuffled

	^self shuffledBy: ThreadSafeRandom value

"Examples:
	($A to: $Z) shuffled
"! !

!SequenceableCollection methodsFor: 'shuffling' stamp: 'ul 3/28/2010 01:03'!
shuffledBy: aRandom
	"Durstenfeld's version of the Fisher-Yates shuffle"

	^self copy shuffleBy: aRandom! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'JMM 10/30/2005 09:22'!
sixth
	"Answer the sixth element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 6! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'nice 10/2/2011 19:10'!
size
	^self subclassResponsibility! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ar 2/27/2010 22:29'!
splitBy: aCollection
	"Answer the receiver, split by aCollection.

	This method works similarly to findTokens: but
		a) takes a collection argument (i.e., 'hello<p>world<p>' splitBy: '<p>')
		b) is 'strict' in its splitting, for example:
				'a///b' findTokens: '/' ==> #('a' 'b')
				'a///b' splitBy: '/' ==> #('a' '' '' 'b')
	"

	^Array streamContents:[:stream|
		self splitBy: aCollection do:[:each| stream nextPut: each].
	].
! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'ar 2/27/2010 22:27'!
splitBy: aCollection do: aBlock
	"Split the receiver by aCollection. Evaluate aBlock with the parts.

	This method works similarly to findTokens: but
		a) takes a collection argument (i.e., 'hello<p>world<p>' splitBy: '<p>')
		b) is 'strict' in its splitting, for example:
				'a///b' findTokens: '/' ==> #('a' 'b')
				'a///b' splitBy: '/' ==> #('a' '' '' 'b')
	"

	| lastIndex nextIndex |
	lastIndex := 1.
	[nextIndex := self indexOfSubCollection: aCollection startingAt: lastIndex.
	nextIndex = 0] whileFalse:[
		aBlock value: (self copyFrom: lastIndex to: nextIndex-1).
		lastIndex := nextIndex+ aCollection size.
	].
	aBlock value: (self copyFrom: lastIndex to: self size).
! !

!SequenceableCollection methodsFor: 'accessing' stamp: ''!
swap: oneIndex with: anotherIndex 
	"Move the element at oneIndex to anotherIndex, and vice-versa."

	| element |
	element := self at: oneIndex.
	self at: oneIndex put: (self at: anotherIndex).
	self at: anotherIndex put: element! !

!SequenceableCollection methodsFor: 'accessing' stamp: 'JMM 10/30/2005 09:22'!
third
	"Answer the third element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 3! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'huma 12/11/2004 20:43'!
upTo: anObject
	"Deprecated. Use copyUpTo:"
	self deprecated: 'Use #copyUpTo:'.
	^ self copyUpTo: anObject! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'di 5/17/1998 13:34'!
withIndexCollect: elementAndIndexBlock 
	"Just like with:collect: except that the iteration index supplies the second argument to the block."
	| result |
	result := self species new: self size.
	1 to: self size do:
		[:index | result at: index put:
		(elementAndIndexBlock
			value: (self at: index)
			value: index)].
	^ result! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'di 7/13/97 09:35'!
withIndexDo: elementAndIndexBlock 
	"Just like with:do: except that the iteration index supplies the second argument to the block."
	1 to: self size do:
		[:index |
		elementAndIndexBlock
			value: (self at: index)
			value: index]! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'di 8/31/1999 13:13'!
with: otherCollection collect: twoArgBlock 
	"Collect and return the result of evaluating twoArgBlock with corresponding elements from this collection and otherCollection."
	| result |
	otherCollection size = self size ifFalse: [self error: 'otherCollection must be the same size'].
	result := self species new: self size.
	1 to: self size do:
		[:index | result at: index put:
		(twoArgBlock
			value: (self at: index)
			value: (otherCollection at: index))].
	^ result! !

!SequenceableCollection methodsFor: 'enumerating' stamp: 'di 8/3/1999 15:26'!
with: otherCollection do: twoArgBlock 
	"Evaluate twoArgBlock with corresponding elements from this collection and otherCollection."
	otherCollection size = self size ifFalse: [self error: 'otherCollection must be the same size'].
	1 to: self size do:
		[:index |
		twoArgBlock value: (self at: index)
				value: (otherCollection at: index)]! !

!SequenceableCollection methodsFor: 'converting' stamp: 'sma 5/12/2000 12:52'!
writeStream
	^ WriteStream on: self! !

!SequenceableCollection methodsFor: 'copying' stamp: 'cmm 4/7/2012 15:30'!
, otherCollection 
	"Concatenate two Strings or Collections."
	^ self copyReplaceFrom: self size + 1
		  to: self size
		  with: otherCollection asCollection
"
#(2 4 6 8) , #(who do we appreciate)
((2989 storeStringBase: 16) copyFrom: 4 to: 6) , ' boy!!'
"! !

!SequenceableCollection methodsFor: 'comparing' stamp: 'sma 5/12/2000 14:04'!
= otherCollection 
	"Answer true if the receiver is equivalent to the otherCollection.
	First test for identity, then rule out different species and sizes of
	collections. As a last resort, examine each element of the receiver
	and the otherCollection."

	self == otherCollection ifTrue: [^ true].
	self species == otherCollection species ifFalse: [^ false].
	^ self hasEqualElements: otherCollection! !

!SequenceableCollection methodsFor: 'converting' stamp: 'TAG 11/6/1998 15:55'!
@ aCollection 
	^ self with: aCollection collect: [:a :b | a @ b]! !

Collection subclass: #WideCharacterSet
	instanceVariableNames: 'map byteArrayMap bitsetCapacity highBitsShift lowBitsMask'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Support'!

!WideCharacterSet commentStamp: 'nice 12/10/2009 19:17' prior: 0!
WideCharacterSet is used to store a Set of WideCharacter with fast access and inclusion test.

Implementation should be efficient in memory if sets are sufficently sparse.

Wide Characters are at most 32bits.
We split them into 16 highBits and 16 lowBits.

map is a dictionary key: 16 highBits value: map of 16 lowBits.

Maps of lowBits  are stored as arrays of bits in a ByteArray.
If a bit is set to 1, this indicate that corresponding character is present.
8192 bytes are necessary in each lowmap.
Empty lowmap are removed from the map Dictionary.

A byteArrayMap is maintained in parallel with map for fast handling of ByteString.
(byteArrayMap at: i+1) = 0 means that character of asciiValue i is absent, = 1 means present.!

!WideCharacterSet class methodsFor: 'instance creation' stamp: 'nice 11/15/2007 22:38'!
newFrom: aCollection
	| newCollection |
	newCollection := self new.
	newCollection addAll: aCollection.
	^newCollection! !

!WideCharacterSet methodsFor: 'collection ops' stamp: 'ul 8/22/2015 12:32'!
add: aCharacter

	| value highBits lowBits |
	(value := aCharacter asInteger) < 256 ifTrue: [
		byteArrayMap at: value + 1 put: 1 ].
	highBits := value bitShift: highBitsShift.
	lowBits := value bitAnd: lowBitsMask.
	(map at: highBits ifAbsentPut: [ Bitset new: bitsetCapacity ])
		setBitAt: lowBits.
	^aCharacter! !

!WideCharacterSet methodsFor: 'comparing' stamp: 'ul 8/20/2015 01:51'!
byteArrayMap
	"return a ByteArray mapping each ascii value to a 1 if that ascii value is in the set, and a 0 if it isn't.
	Intended for use by primitives only. (and comparison)
	This version will answer a subset with only byte characters"
	
	^byteArrayMap! !

!WideCharacterSet methodsFor: 'converting' stamp: 'nice 11/20/2007 00:19'!
complement
	"return a character set containing precisely the characters the receiver does not"
	
	^CharacterSetComplement of: self copy! !

!WideCharacterSet methodsFor: 'collection ops' stamp: 'ul 8/22/2015 12:33'!
do: aBlock
 
	map keysAndValuesDo: [ :index :bitset |
		| highBits |
		highBits := index * bitsetCapacity.
		bitset do: [ :lowBits |
			aBlock value: (Character value: highBits + lowBits) ] ]! !

!WideCharacterSet methodsFor: 'collection ops' stamp: 'ul 8/22/2015 12:37'!
findFirstInByteString: aByteString startingAt: startIndex
	"Double dispatching: since we know this is a ByteString, we can use a superfast primitive using a ByteArray map with 0 slots for byte characters not included and 1 for byte characters included in the receiver."

	^ByteString
		findFirstInString: aByteString
		inSet: byteArrayMap
		startingAt: startIndex! !

!WideCharacterSet methodsFor: 'comparing' stamp: 'ul 8/22/2015 12:49'!
hash
	"Answer a hash code aimed at storing and retrieving the receiver in a Set or Dictionary.
	Two equal objects should have equal hash.
	Note: as the receiver can be equal to an ordinary CharacterSet,
	the hash code must reflect this"
	
	self hasWideCharacters ifTrue: [ ^map hash ].
	^byteArrayMap hash! !

!WideCharacterSet methodsFor: 'testing' stamp: 'nice 5/9/2006 23:25'!
hasWideCharacters
	"Answer true if i contain any wide character"
	
	self do: [:e | e asciiValue >= 256 ifTrue: [^true]].
	^false! !

!WideCharacterSet methodsFor: 'collection ops' stamp: 'ul 3/10/2016 14:08'!
includes: anObject 

	| value |
	anObject isCharacter ifFalse: [ ^false ].
	(value := anObject asInteger) < 256 ifTrue: [
		^(byteArrayMap at: value + 1) ~= 0 ].
	^((map at: (value bitShift: highBitsShift) ifAbsent: nil) ifNil: [ ^false ])
		includes: (value bitAnd: lowBitsMask)! !

!WideCharacterSet methodsFor: 'initialize-release' stamp: 'ul 8/19/2015 22:18'!
initialize

	map := PluggableDictionary integerDictionary.
	byteArrayMap := ByteArray new: 256.
	self initializeWithLowBits: 8! !

!WideCharacterSet methodsFor: 'initialize-release' stamp: 'ul 8/22/2015 12:33'!
initializeWithLowBits: lowBits

	bitsetCapacity := 1 bitShift: lowBits.
	highBitsShift := 0 - lowBits.
	lowBitsMask := bitsetCapacity - 1.
	! !

!WideCharacterSet methodsFor: 'enumerating' stamp: 'ul 3/9/2016 21:54'!
occurrencesOf: anObject
	"Answer how many of the receiver's elements are equal to anObject. Optimized version."

	(self includes: anObject) ifTrue: [ ^1 ].
	^0! !

!WideCharacterSet methodsFor: 'copying' stamp: 'nice 10/5/2009 08:53'!
postCopy
	super postCopy.
	map := map collect: [:each | each copy]! !

!WideCharacterSet methodsFor: 'collection ops' stamp: 'ul 8/22/2015 12:45'!
removeAll

	map isEmpty ifTrue: [ ^self ].
	map removeAll.
	byteArrayMap atAllPut: 0! !

!WideCharacterSet methodsFor: 'collection ops' stamp: 'ul 8/22/2015 12:39'!
remove: aCharacter
	"Don't signal an error when aCharacter is not present."

	^self remove: aCharacter ifAbsent: aCharacter! !

!WideCharacterSet methodsFor: 'collection ops' stamp: 'ul 8/22/2015 12:46'!
remove: aCharacter ifAbsent: aBlock

	| value highBits lowBits bitset |
	(value := aCharacter asInteger) < 256 ifTrue: [
		(byteArrayMap at: value + 1) = 0 ifTrue: [ ^aBlock value ].
		byteArrayMap at: value + 1 put: 0 ].
	highBits := value bitShift: highBitsShift.
	lowBits := value bitAnd: lowBitsMask.
	bitset := (map at: highBits ifAbsent: nil) ifNil: [ ^aBlock value ].
	((bitset clearBitAt: lowBits) and: [ bitset size = 0 ]) ifTrue: [
		map removeKey: highBits ].
	^aCharacter! !

!WideCharacterSet methodsFor: 'collection ops' stamp: 'ul 8/22/2015 12:45'!
size

	^map detectSum: [ :each | each size ]! !

!WideCharacterSet methodsFor: 'comparing' stamp: 'nice 5/9/2006 23:29'!
species
	^self hasWideCharacters
		ifTrue: [WideCharacterSet]
		ifFalse: [CharacterSet]! !

!WideCharacterSet methodsFor: 'comparing' stamp: 'nice 5/9/2006 23:14'!
wideCharacterMap
	^map! !

!WideCharacterSet methodsFor: 'comparing' stamp: 'nice 5/9/2006 23:29'!
= anObject
	^self species == anObject species and: [
		self wideCharacterMap = anObject wideCharacterMap ]! !

Object subclass: #Link
	instanceVariableNames: 'nextLink'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Support'!

!Link commentStamp: '<historical>' prior: 0!
An instance of me is a simple record of a pointer to another Link. I am an abstract class; my concrete subclasses, for example, Process, can be stored in a LinkedList structure.!

!Link class methodsFor: 'instance creation' stamp: 'apb 10/3/2000 15:55'!
nextLink: aLink 
	"Answer an instance of me referring to the argument, aLink."

	^self new nextLink: aLink; yourself! !

!Link methodsFor: 'accessing' stamp: ''!
nextLink
	"Answer the link to which the receiver points."

	^nextLink! !

!Link methodsFor: 'accessing' stamp: ''!
nextLink: aLink 
	"Store the argument, aLink, as the link to which the receiver refers. 
	Answer aLink."

	^nextLink := aLink! !

Link subclass: #StackLink
	instanceVariableNames: 'element'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Stack'!

!StackLink commentStamp: '<historical>' prior: 0!
I implement an element of a stack. I'm a container for any type of object, saved into the 'element' variable. My superclass Link allows me to be part of a LinkedList.!

!StackLink class methodsFor: 'instance creation' stamp: 'dc 7/25/2005 10:15'!
with: anObject 
	^ self new element: anObject! !

!StackLink methodsFor: 'accessing' stamp: 'dc 7/24/2005 15:34'!
element
	^element! !

!StackLink methodsFor: 'accessing' stamp: 'dc 7/25/2005 10:16'!
element: anObject 
	"Any kind of Object."
	element := anObject! !

!StackLink methodsFor: 'printing' stamp: 'dc 7/25/2005 10:15'!
printOn: aStream 
	aStream nextPutAll: self class printString;
		 nextPutAll: ' with: ';
		 nextPutAll: self element printString! !

Object subclass: #LRUCache
	instanceVariableNames: 'map head calls hits size factory'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Cache'!

!LRUCache commentStamp: 'ul 4/12/2015 02:55' prior: 0!
I'm a cache of values, given a key I return a Value from the cache or from the factory.

I use a Dictionary to find the corresponding value for the given key. I also store the key-value pairs in a circular doubly-linked list with a head element. The list is implemented by an LRUCacheHeadNode - stored in the head instance variable - and an LRUCacheNode for each key-value pair. The nodes in the list are ordered by access time. The first node (next of head) is the most recently accessed, the last one (previous of head) is the least recently accessed.
If the number of stored key-value pairs is equal to size, and a new pair needs to be stored, then I remove the least recently used pair before adding the new pair.!

!LRUCache class methodsFor: 'instance creation' stamp: 'dgd 3/26/2003 22:29'!
size: aNumber factory: aBlock 
	"answer an instance of the receiver"
	^ self new initializeSize: aNumber factory: aBlock! !

!LRUCache class methodsFor: 'testing' stamp: 'dgd 3/26/2003 22:22'!
test
	" 
	LRUCache test 
	"
	| c |
	c := LRUCache
				size: 5
				factory: [:key | key * 2].
	c at: 1.
	c at: 2.
	c at: 3.
	c at: 4.
	c at: 1.
	c at: 5.
	c at: 6.
	c at: 7.
	c at: 8.
	c at: 1.
	^ c! !

!LRUCache class methodsFor: 'testing' stamp: 'dgd 3/26/2003 22:22'!
test2
	" 
	LRUCache test2.  
	Time millisecondsToRun:[LRUCache test2]. 
	MessageTally spyOn:[LRUCache test2].  
	"
	| c |
	c := LRUCache
				size: 600
				factory: [:key | key * 2].
	1
		to: 6000
		do: [:each | c at: each].
	^ c! !

!LRUCache methodsFor: 'accessing' stamp: 'ul 4/13/2015 05:36'!
at: aKey 
	"answer the object for aKey, if not present in the cache creates it"

	calls := calls + 1.
	^map
		at: aKey
		ifPresent: [ :node |
			hits := hits + 1.
			head next == node ifFalse: [
				node
					unlink;
					linkAfter: head ].
			node value ]
		ifAbsent: [
			| node |
			map size = size
				ifTrue: [ 
					node := head previous.
					node unlink.
					map removeKey: node key. ]
				ifFalse: [ node := LRUCacheNode new ].
			node linkAfter: head.
			map at: aKey put: node.
			node 
				key: aKey;
				value: (factory value: aKey);
				value ]! !

!LRUCache methodsFor: 'initialization' stamp: 'ul 4/12/2015 02:46'!
initializeSize: anInteger factory: aBlock 
	"initialize the receiver's size and factory"
	
	anInteger strictlyPositive ifFalse: [ self error: 'Size must be at least 1' ].
	size := anInteger.
	head := LRUCacheHeadNode new.
	map := PluggableDictionary integerDictionary.
	factory := aBlock.
	calls := 0.
	hits := 0! !

!LRUCache methodsFor: 'printing' stamp: 'laza 3/9/2010 21:33'!
printOn: aStream 
	"Append to the argument, aStream, a sequence of characters 
	that identifies the receiver."
	aStream nextPutAll: self class name;
		 nextPutAll: ' size:';
		 nextPutAll: size asString;
		 nextPutAll: ', calls:';
		 nextPutAll: calls asString;
		 nextPutAll: ', hits:';
		 nextPutAll: hits asString;
		 nextPutAll: ', ratio:';
		 nextPutAll: ((hits isNumber and: [calls isNumber and: [calls ~= 0]])
			ifTrue: [hits / calls]
			ifFalse: [0]) asFloat asString! !

!LRUCache methodsFor: 'initialization' stamp: 'ul 4/12/2015 00:25'!
reset

	self initializeSize: size factory: factory! !

Object subclass: #LRUCacheHeadNode
	instanceVariableNames: 'next previous'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Cache'!

!LRUCacheHeadNode methodsFor: 'initialize-release' stamp: 'ul 4/12/2015 02:30'!
initialize

	previous := next := self! !

!LRUCacheHeadNode methodsFor: 'accessing' stamp: 'ul 4/12/2015 00:20'!
next

	^next! !

!LRUCacheHeadNode methodsFor: 'accessing' stamp: 'ul 4/12/2015 00:20'!
next: anObject

	next := anObject! !

!LRUCacheHeadNode methodsFor: 'accessing' stamp: 'ul 4/12/2015 00:20'!
previous

	^previous! !

!LRUCacheHeadNode methodsFor: 'accessing' stamp: 'ul 4/12/2015 00:20'!
previous: anObject

	previous := anObject! !

LRUCacheHeadNode subclass: #LRUCacheNode
	instanceVariableNames: 'key value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Cache'!

!LRUCacheNode methodsFor: 'accessing' stamp: 'ul 4/12/2015 00:17'!
key

	^key! !

!LRUCacheNode methodsFor: 'accessing' stamp: 'ul 4/12/2015 00:17'!
key: anObject

	key := anObject! !

!LRUCacheNode methodsFor: 'list operations' stamp: 'ul 4/12/2015 00:22'!
linkAfter: anLRUCacheHeadNode

	next := anLRUCacheHeadNode next.
	previous := anLRUCacheHeadNode.
	next previous: self.
	previous next: self! !

!LRUCacheNode methodsFor: 'accessing' stamp: 'ul 4/12/2015 00:18'!
printOn: stream

	super printOn: stream.
	stream
		nextPut: $(;
		print: key;
		nextPutAll: ', ';
		print: value;
		nextPut: $)! !

!LRUCacheNode methodsFor: 'list operations' stamp: 'ul 4/12/2015 00:19'!
unlink

	next previous: previous.
	previous next: next.
	next := previous := nil! !

!LRUCacheNode methodsFor: 'accessing' stamp: 'ul 4/12/2015 00:17'!
value

	^value! !

!LRUCacheNode methodsFor: 'accessing' stamp: 'ul 4/12/2015 00:17'!
value: anObject

	value := anObject! !

Object subclass: #MimeConverter
	instanceVariableNames: 'dataStream mimeStream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

MimeConverter subclass: #Base64MimeConverter
	instanceVariableNames: 'data multiLine'
	classVariableNames: 'FromCharTable ToCharTable'
	poolDictionaries: ''
	category: 'Collections-Streams'!

!Base64MimeConverter commentStamp: '<historical>' prior: 0!
This class encodes and decodes data in Base64 format.  This is MIME encoding.  We translate a whole stream at once, taking a Stream as input and giving one as output.  Returns a whole stream for the caller to use.
           0 A            17 R            34 i            51 z
           1 B            18 S            35 j            52 0
           2 C            19 T            36 k            53 1
           3 D            20 U            37 l            54 2
           4 E            21 V            38 m            55 3
           5 F            22 W            39 n            56 4
           6 G            23 X            40 o            57 5
           7 H            24 Y            41 p            58 6
           8 I            25 Z            42 q            59 7
           9 J            26 a            43 r            60 8
          10 K            27 b            44 s            61 9
          11 L            28 c            45 t            62 +
          12 M            29 d            46 u            63 /
          13 N            30 e            47 v
          14 O            31 f            48 w         (pad) =
          15 P            32 g            49 x
          16 Q            33 h            50 y
Outbound: bytes are broken into 6 bit chunks, and the 0-63 value is converted to a character.  3 data bytes go into 4 characters.
Inbound: Characters are translated in to 0-63 values and shifted into 8 bit bytes.

(See: N. Borenstein, Bellcore, N. Freed, Innosoft, Network Working Group, Request for Comments: RFC 1521, September 1993, MIME (Multipurpose Internet Mail Extensions) Part One: Mechanisms for Specifying and Describing the Format of Internet Message Bodies. Sec 6.2)

By Ted Kaehler, based on Tim Olson's Base64Filter.!

!Base64MimeConverter class methodsFor: 'as yet unclassified' stamp: 'tk 2/19/2000 15:53'!
decodeInteger: mimeString
	| bytes sum |
	"Decode the MIME string into an integer of any length"

	bytes := (Base64MimeConverter mimeDecodeToBytes: 
				(ReadStream on: mimeString)) contents.
	sum := 0.
	bytes reverseDo: [:by | sum := sum * 256 + by].
	^ sum! !

!Base64MimeConverter class methodsFor: 'as yet unclassified' stamp: 'nice 7/26/2014 23:00'!
encodeInteger: int
	| strm |
	"Encode an integer of any length and return the MIME string"

	strm := WriteStream on: (ByteArray new: int digitLength).
	1 to: int digitLength do: [:ii | strm nextPut: (int digitAt: ii)].
	^ ((self mimeEncode: strm readStream) contents) copyUpTo: $=	"remove padding"! !

!Base64MimeConverter class methodsFor: 'as yet unclassified' stamp: 'nice 7/30/2013 01:59'!
initialize

	FromCharTable := Array new: 256.	"nils"
	ToCharTable := ($A to: $Z) , ($a to: $z) , ($0 to: $9) , '+/'.
	ToCharTable keysAndValuesDo: [:ind :char |
		FromCharTable at: char asciiValue + 1 put: ind - 1].! !

!Base64MimeConverter class methodsFor: 'as yet unclassified' stamp: 'nice 7/26/2014 23:03'!
mimeDecodeToBytes: aStream 
	"Return a ReadStream of the original ByteArray.  aStream has only 65 innocuous character values.  aStream is not binary.  (See class comment). 4 bytes in aStream goes to 3 bytes in output."

	| me |
	aStream position: 0.
	me := self new mimeStream: aStream.
	me dataStream: (WriteStream on: (ByteArray new: aStream size * 3 // 4)).
	me mimeDecodeToByteArray.
	^ me dataStream readStream! !

!Base64MimeConverter class methodsFor: 'as yet unclassified' stamp: 'nice 7/26/2014 23:01'!
mimeDecodeToChars: aStream 
	"Return a ReadWriteStream of the original String.  aStream has only 65 innocuous character values.  It is not binary.  (See class comment). 4 bytes in aStream goes to 3 bytes in output."

	| me |
	aStream position: 0.
	me := self new mimeStream: aStream.
	me dataStream: (WriteStream on: (String new: aStream size * 3 // 4)).
	me mimeDecode.
	^ me dataStream readStream! !

!Base64MimeConverter class methodsFor: 'as yet unclassified' stamp: 'ar 3/9/2010 22:17'!
mimeEncodeContinue: aStream
	"Return a ReadWriteStream of characters.  The data of aStream is encoded as 65 innocuous characters.  (See class comment). 3 bytes in aStream goes to 4 bytes in output."
	^self mimeEncode: aStream multiLine: true atStart: false! !

!Base64MimeConverter class methodsFor: 'as yet unclassified' stamp: 'ar 3/9/2010 22:17'!
mimeEncode: aStream
	"Return a ReadWriteStream of characters.  The data of aStream is encoded as 65 innocuous characters.  (See class comment). 3 bytes in aStream goes to 4 bytes in output."
	^self mimeEncode: aStream multiLine: true atStart: true! !

!Base64MimeConverter class methodsFor: 'as yet unclassified' stamp: 'ar 3/9/2010 22:17'!
mimeEncode: aStream multiLine: aBool
	"Return a ReadWriteStream of characters.  The data of aStream is encoded as 65 innocuous characters.  (See class comment). 3 bytes in aStream goes to 4 bytes in output."

	^self mimeEncode: aStream multiLine: aBool atStart: true! !

!Base64MimeConverter class methodsFor: 'as yet unclassified' stamp: 'nice 7/26/2014 22:59'!
mimeEncode: aStream multiLine: aBool atStart: resetInput
	"Return a ReadStream of characters.  The data of aStream is encoded as 65 innocuous characters.  (See class comment). 3 bytes in aStream goes to 4 bytes in output."

	| me |
	resetInput ifTrue:[aStream position: 0].
	me := self new dataStream: aStream.
	me multiLine: aBool.
	me mimeStream: (WriteStream on: (String new: aStream size + 20 * 4 // 3)).
	me mimeEncode.
	^ me mimeStream readStream! !

!Base64MimeConverter methodsFor: 'conversion' stamp: 'tk 12/9/97 13:34'!
mimeDecode
	"Convert a stream in base 64 with only a-z,A-Z,0-9,+,/ to a full byte stream of characters.  Reutrn a whole stream for the user to read."

	| nibA nibB nibC nibD |
	[mimeStream atEnd] whileFalse: [
		(nibA := self nextValue) ifNil: [^ dataStream].
		(nibB := self nextValue) ifNil: [^ dataStream].
		dataStream nextPut: ((nibA bitShift: 2) + (nibB bitShift: -4)) asCharacter.
		nibB := nibB bitAnd: 16rF.
		(nibC := self nextValue) ifNil: [^ dataStream].
		dataStream nextPut: ((nibB bitShift: 4) + (nibC bitShift: -2)) asCharacter.
		nibC := nibC bitAnd: 16r3.
		(nibD := self nextValue) ifNil: [^ dataStream].
		dataStream nextPut: ((nibC bitShift: 6) + nibD) asCharacter.
		].
	^ dataStream! !

!Base64MimeConverter methodsFor: 'conversion' stamp: 'tk 12/9/97 13:39'!
mimeDecodeToByteArray
	"Convert a stream in base 64 with only a-z,A-Z,0-9,+,/ to a full ByteArray of 0-255 values.  Reutrn a whole stream for the user to read."

	| nibA nibB nibC nibD |
	[mimeStream atEnd] whileFalse: [
		(nibA := self nextValue) ifNil: [^ dataStream].
		(nibB := self nextValue) ifNil: [^ dataStream].
		dataStream nextPut: ((nibA bitShift: 2) + (nibB bitShift: -4)).
		nibB := nibB bitAnd: 16rF.
		(nibC := self nextValue) ifNil: [^ dataStream].
		dataStream nextPut: ((nibB bitShift: 4) + (nibC bitShift: -2)).
		nibC := nibC bitAnd: 16r3.
		(nibD := self nextValue) ifNil: [^ dataStream].
		dataStream nextPut: ((nibC bitShift: 6) + nibD).
		].
	^ dataStream! !

!Base64MimeConverter methodsFor: 'conversion' stamp: 'ar 4/15/2008 17:58'!
mimeEncode
	"Convert from data to 6 bit characters."

	| phase1 phase2 raw nib lineLength |
	phase1 := phase2 := false.
	lineLength := 0.
	[dataStream atEnd] whileFalse: [
		(multiLine and:[lineLength >= 70]) ifTrue: [ mimeStream cr.  lineLength := 0. ].
		data := raw := dataStream next asInteger.
		nib := (data bitAnd: 16rFC) bitShift: -2.
		mimeStream nextPut: (ToCharTable at: nib+1).
		(raw := dataStream next) ifNil: [raw := 0. phase1 := true].
		data := ((data bitAnd: 3) bitShift: 8) + raw asInteger.
		nib := (data bitAnd: 16r3F0) bitShift: -4.
		mimeStream nextPut: (ToCharTable at: nib+1).
		(raw := dataStream next) ifNil: [raw := 0. phase2 := true].
		data := ((data bitAnd: 16rF) bitShift: 8) + (raw asInteger).
		nib := (data bitAnd: 16rFC0) bitShift: -6.
		mimeStream nextPut: (ToCharTable at: nib+1).
		nib := (data bitAnd: 16r3F).
		mimeStream nextPut: (ToCharTable at: nib+1).

		lineLength := lineLength + 4.].
	phase1 ifTrue: [mimeStream skip: -2; nextPut: $=; nextPut: $=.
			^ mimeStream].
	phase2 ifTrue: [mimeStream skip: -1; nextPut: $=.
			^ mimeStream].

! !

!Base64MimeConverter methodsFor: 'accessing' stamp: 'ar 4/15/2008 17:58'!
multiLine
	"Determines whether we allow multi-line encodings (the default) or force everything into a single line (for use with URLs etc. where the continuation marker and the line break cause problems)"
	^multiLine! !

!Base64MimeConverter methodsFor: 'accessing' stamp: 'ar 4/15/2008 17:58'!
multiLine: aBool
	"Determines whether we allow multi-line encodings (the default) or force everything into a single line (for use with URLs etc. where the continuation marker and the line break cause problems)"
	multiLine := aBool! !

!Base64MimeConverter methodsFor: 'conversion' stamp: 'ul 6/17/2011 12:36'!
nextValue
	"The next six bits of data char from the mimeStream, or nil.  Skip all other chars"
	| raw num |
	[raw := mimeStream next.
	raw ifNil: [^ nil].	"end of stream"
	raw == $= ifTrue: [^ nil].
	num := FromCharTable at: raw asciiValue + 1.
	num ifNotNil: [^ num].
	"else ignore space, return, tab, ..."
	] repeat! !

!MimeConverter class methodsFor: 'convenience' stamp: 'bf 3/10/2000 14:47'!
forEncoding: encodingString
	"Answer a converter class for the given encoding or nil if unknown"
	encodingString ifNil: [^nil].
	^ encodingString asLowercase caseOf: 
		{ ['base64'] -> [Base64MimeConverter].
		  ['quoted-printable'] -> [QuotedPrintableMimeConverter]}
		otherwise: [].
! !

!MimeConverter class methodsFor: 'convenience' stamp: 'bf 3/10/2000 14:43'!
mimeDecode: aStringOrStream as: contentsClass
	^ contentsClass streamContents: [:out |
		self mimeDecode: aStringOrStream to: out]! !

!MimeConverter class methodsFor: 'convenience' stamp: 'bf 3/10/2000 14:40'!
mimeDecode: aStringOrStream to: outStream
	self new
		mimeStream: (aStringOrStream isStream
			ifTrue: [aStringOrStream]
			ifFalse: [ReadStream on: aStringOrStream]);
		dataStream: outStream;
		mimeDecode! !

!MimeConverter class methodsFor: 'convenience' stamp: 'bf 3/10/2000 14:40'!
mimeEncode: aCollectionOrStream
	^ String streamContents: [:out |
		self mimeEncode: aCollectionOrStream to: out]! !

!MimeConverter class methodsFor: 'convenience' stamp: 'bf 3/10/2000 14:40'!
mimeEncode: aCollectionOrStream to: outStream
	self new
		dataStream: (aCollectionOrStream isStream
			ifTrue: [aCollectionOrStream]
			ifFalse: [ReadStream on: aCollectionOrStream]);
		mimeStream: outStream;
		mimeEncode! !

!MimeConverter methodsFor: 'accessing' stamp: 'tk 12/9/97 13:55'!
dataStream
	^dataStream! !

!MimeConverter methodsFor: 'accessing' stamp: 'tk 12/9/97 13:51'!
dataStream: anObject
	dataStream := anObject! !

!MimeConverter methodsFor: 'conversion' stamp: 'bf 11/12/1998 13:30'!
mimeDecode
	"Do conversion reading from mimeStream writing to dataStream"

	self subclassResponsibility! !

!MimeConverter methodsFor: 'conversion' stamp: 'bf 11/12/1998 13:31'!
mimeEncode
	"Do conversion reading from dataStream writing to mimeStream"

	self subclassResponsibility! !

!MimeConverter methodsFor: 'accessing' stamp: 'tk 12/9/97 13:53'!
mimeStream
	^mimeStream! !

!MimeConverter methodsFor: 'accessing' stamp: 'tk 12/9/97 13:51'!
mimeStream: anObject
	mimeStream := anObject! !

MimeConverter subclass: #QuotedPrintableMimeConverter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!QuotedPrintableMimeConverter commentStamp: '<historical>' prior: 0!
I do quoted printable MIME decoding as specified in RFC 2045 "MIME Part One: Format of Internet Message Bodies".

Short version of RFC2045, Sect. 6.7:

	(1) Any octet, except a CR or LF that is part of a CRLF line break of the canonical (standard) form of the data being encoded, may be represented by an "=" followed by a two digit hexadecimal representation of the octet's value. [...]

	(2) Octets with decimal values of 33 through 60 inclusive, and 62 through 126, inclusive, MAY be represented as the US-ASCII characters which correspond to those octets [...].

	(3) Octets with values of 9 and 32 MAY be represented as US-ASCII TAB (HT) and SPACE characters,
 respectively, but MUST NOT be so represented at the end of an encoded line.  [...]

	(4) A line break in a text body, represented as a CRLF sequence in the text canonical form, must be represented by a (RFC 822) line break, which is also a CRLF sequence, in the Quoted-Printable encoding.  [...]

	(5) The Quoted-Printable encoding REQUIRES that encoded lines be no more than 76 characters long.  If longer lines are to be encoded with the Quoted-Printable encoding, "soft" line breaks
 must be used.  An equal sign as the last character on a encoded line indicates such a non-significant ("soft") line break in the encoded text.


--bf 11/27/1998 16:50!

!QuotedPrintableMimeConverter methodsFor: 'conversion' stamp: 'bf 11/24/1998 20:33'!
mimeDecode
	"Do conversion reading from mimeStream writing to dataStream"

	| line s c1 v1 c2 v2 |
	[(line := mimeStream nextLine) isNil] whileFalse: [
		line := line withoutTrailingBlanks.
		line size = 0
			ifTrue: [dataStream cr]
			ifFalse: [
				s := ReadStream on: line.
				[dataStream nextPutAll: (s upTo: $=).
				s atEnd] whileFalse: [
					c1 := s next. v1 := c1 digitValue.
					((v1 between: 0 and: 15) and: [s atEnd not])
						ifFalse: [dataStream nextPut: $=; nextPut: c1]
						ifTrue: [c2 := s next. v2 := c2 digitValue.
							(v2 between: 0 and: 15)
								ifFalse: [dataStream nextPut: $=; nextPut: c1; nextPut: c2]
								ifTrue: [dataStream nextPut: (Character value: v1 * 16 + v2)]]].
				line last = $= ifFalse: [dataStream cr]]].
	^ dataStream! !

QuotedPrintableMimeConverter subclass: #RFC2047MimeConverter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!RFC2047MimeConverter commentStamp: '<historical>' prior: 0!
I do quoted printable MIME decoding as specified in RFC 2047 ""MIME Part Three: Message Header Extensions for Non-ASCII Text". See String>>decodeMimeHeader!

!RFC2047MimeConverter methodsFor: 'private-encoding' stamp: 'bf 3/11/2000 23:16'!
encodeChar: aChar to: aStream

	aChar = Character space
		ifTrue: [^ aStream nextPut: $_].
	((aChar asciiValue between: 32 and: 127) and: [('?=_' includes: aChar) not])
		ifTrue: [^ aStream nextPut: aChar].
	aStream nextPut: $=;
		nextPut: (Character digitValue: aChar asciiValue // 16);
		nextPut: (Character digitValue: aChar asciiValue \\ 16)
! !

!RFC2047MimeConverter methodsFor: 'private-encoding' stamp: 'bf 3/11/2000 23:13'!
encodeWord: aString

	(aString allSatisfy: [:c | c asciiValue < 128])
		ifTrue: [^ aString].
	^ String streamContents: [:stream |
		stream nextPutAll: '=?iso-8859-1?Q?'.
		aString do: [:c | self encodeChar: c to: stream].
		stream nextPutAll: '?=']! !

!RFC2047MimeConverter methodsFor: 'private-encoding' stamp: 'bf 3/12/2000 14:36'!
isStructuredField: aString

	| fName |
	fName := aString copyUpTo: $:.
	('Resent' sameAs: (fName copyUpTo: $-))
		ifTrue: [fName := fName copyFrom: 8 to: fName size].
	^#('Sender' 'From' 'Reply-To' 'To' 'cc' 'bcc') anySatisfy: [:each | fName sameAs: each]! !

!RFC2047MimeConverter methodsFor: 'conversion' stamp: 'bf 11/27/1998 17:04'!
mimeDecode
	"Do conversion reading from mimeStream writing to dataStream. See String>>decodeMimeHeader"

	| c |
	[mimeStream atEnd] whileFalse: [
		c := mimeStream next.
		c = $=
			ifTrue: [c := Character value: mimeStream next digitValue * 16
				+ mimeStream next digitValue]
			ifFalse: [c = $_ ifTrue: [c := $ ]].
		dataStream nextPut: c].
	^ dataStream! !

!RFC2047MimeConverter methodsFor: 'conversion' stamp: 'bf 3/10/2000 16:06'!
mimeEncode
	"Do conversion reading from dataStream writing to mimeStream. Break long lines and escape non-7bit chars."

	| word pos wasGood isGood max |
	true ifTrue: [mimeStream nextPutAll: dataStream upToEnd].
	pos := 0.
	max := 72.
	wasGood := true.
	[dataStream atEnd] whileFalse: [
		word := self readWord.
		isGood := word allSatisfy: [:c | c asciiValue < 128].
		wasGood & isGood ifTrue: [
			pos + word size < max
				ifTrue: [dataStream nextPutAll: word.
					pos := pos + word size]
				ifFalse: []
		]
	].
	^ mimeStream! !

!RFC2047MimeConverter methodsFor: 'private-encoding' stamp: 'bf 3/11/2000 22:30'!
readWord

	| strm |
	strm := WriteStream on: (String new: 20)
	dataStream skipSeparators.
	[dataStream atEnd] whileFalse: 
		[ | c |
		c := dataStream next.
		strm nextPut: c.
		c isSeparator ifTrue: [^ strm contents]].
	^ strm contents! !

!Object methodsFor: '*collections' stamp: 'cmm 9/7/2013 12:05'!
compareSafely: anObject
	^ self class = anObject class
		ifTrue: [ self printString < anObject printString ]
		ifFalse: [ self class name < anObject class name ]! !

Object subclass: #SetElement
	instanceVariableNames: 'enclosedElement'
	classVariableNames: 'NilElement'
	poolDictionaries: ''
	category: 'Collections-Support'!

!SetElement commentStamp: 'nice 8/26/2010 22:21' prior: 0!
A SetElement is a special wrapper used to handle addition of some special elements into Set.
This is necessary mainly for storing an UndefinedObject in a Set, since nil is used in Set algorithm to designate free slots in internal storage.

Instance Variables
	enclosedElement:		<Object>

enclosedElement
	- the real element we wish to put into the set
!

!SetElement class methodsFor: 'class initialization' stamp: 'Igor.Stasenko 11/13/2009 05:55'!
initialize
	NilElement := self with: nil.
! !

!SetElement class methodsFor: 'accessing' stamp: 'Igor.Stasenko 11/13/2009 05:56'!
withNil
	^ NilElement
! !

!SetElement class methodsFor: 'instance creation' stamp: 'Igor.Stasenko 11/13/2009 05:53'!
with: anObject
	^ self new enclosedSetElement: anObject! !

!SetElement methodsFor: 'converting' stamp: 'Igor.Stasenko 11/13/2009 06:29'!
asSetElement
	"A receiver has to be included into a set, as a distinct object.
	We need to wrap receiver in own turn, otherwise #enclosedSetElement will return wrong object for that set"
	^ SetElement with: self! !

!SetElement methodsFor: 'accessing' stamp: 'Igor.Stasenko 11/13/2009 05:49'!
enclosedSetElement
	^ enclosedElement! !

!SetElement methodsFor: 'accessing' stamp: 'Igor.Stasenko 11/13/2009 05:49'!
enclosedSetElement: anObject
	enclosedElement := anObject! !

!SetElement methodsFor: 'comparing' stamp: 'Igor.Stasenko 11/13/2009 05:52'!
hash
	^ enclosedElement hash
! !

!SetElement methodsFor: 'comparing' stamp: 'Igor.Stasenko 11/13/2009 05:51'!
= anObject
	^ anObject class = self class and: [ enclosedElement = anObject enclosedSetElement ]
! !

Object subclass: #SharedQueue
	instanceVariableNames: 'contentsArray readPosition writePosition accessProtect readSynch'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Sequenceable'!

!SharedQueue commentStamp: '<historical>' prior: 0!
I provide synchronized communication of arbitrary objects between Processes. An object is sent by sending the message nextPut: and received by sending the message next. If no object has been sent when a next message is sent, the Process requesting the object will be suspended until one is sent.!

!SharedQueue class methodsFor: 'instance creation' stamp: ''!
new
	"Answer a new instance of SharedQueue that has 10 elements."

	^self new: 10! !

!SharedQueue class methodsFor: 'instance creation' stamp: 'nice 4/4/2006 22:10'!
new: anInteger 
	^super new initialize: anInteger! !

!SharedQueue methodsFor: 'accessing' stamp: 'bf 2/11/2006 15:17'!
flush
	"Throw out all pending contents"
	accessProtect critical: [
		"nil out flushed slots --bf 02/11/2006"
		contentsArray from: readPosition to: writePosition-1 put: nil.
		readPosition := 1.
		writePosition := 1.
		"Reset the read synchronization semaphore"
		readSynch initSignals].! !

!SharedQueue methodsFor: 'accessing' stamp: 'ul 12/8/2010 04:45'!
flushAllSuchThat: aBlock
	"Remove from the queue all objects that satisfy aBlock."
	
	accessProtect critical: [
		| newReadPos |
		newReadPos := writePosition.
		writePosition - 1 to: readPosition by: -1 do: [ :i |
			| value |
			value := contentsArray at: i.
			contentsArray at: i put: nil.
			((aBlock value: value) and: [ (readSynch waitIfLocked: [ nil ]) notNil ]) ifFalse: [
				newReadPos := newReadPos - 1.
				contentsArray at: newReadPos put: value ] ].
		readPosition := newReadPos ]! !

!SharedQueue methodsFor: 'private' stamp: 'nice 4/4/2006 22:09'!
initialize: size

	contentsArray := Array new: size.
	readPosition := 1.
	writePosition := 1.
	accessProtect := Semaphore forMutualExclusion.
	readSynch := Semaphore new! !

!SharedQueue methodsFor: 'testing' stamp: ''!
isEmpty
	"Answer whether any objects have been sent through the receiver and 
	not yet received by anyone."

	^readPosition = writePosition! !

!SharedQueue methodsFor: 'private' stamp: 'ul 10/19/2010 04:53'!
makeRoomAtEnd

	| contentsSize newContentsArray |
	contentsSize := writePosition - readPosition.
	newContentsArray := contentsSize * 2 > contentsArray size
		ifTrue: [ contentsArray class new: contentsArray size * 2 ]
		ifFalse: [
			(contentsArray size > 10 and: [ contentsSize * 4 <= contentsArray size ])
				ifTrue: [ contentsArray class new: (contentsSize * 2 max: 10) ]
				ifFalse: [ contentsArray ] ].
	newContentsArray
		replaceFrom: 1
		to: contentsSize
		with: contentsArray
		startingAt: readPosition.
	contentsArray == newContentsArray 
		ifFalse: [ contentsArray := newContentsArray ]
		ifTrue: [ contentsArray from: contentsSize + 1 to: contentsArray size put: nil ].
	readPosition := 1.
	writePosition := contentsSize + 1! !

!SharedQueue methodsFor: 'accessing' stamp: 'nice 12/26/2009 21:25'!
next
	"Answer the object that was sent through the receiver first and has not 
	yet been received by anyone. If no object has been sent, suspend the 
	requesting process until one is."

	readSynch wait.
	^accessProtect
		critical: [
			| value |
			readPosition = writePosition
					ifTrue: 
						[self error: 'Error in SharedQueue synchronization'.
						 value := nil]
					ifFalse: 
						[value := contentsArray at: readPosition.
						 contentsArray at: readPosition put: nil.
						 readPosition := readPosition + 1].
			value].! !

!SharedQueue methodsFor: 'accessing' stamp: 'ul 12/8/2010 04:31'!
nextOrNil
	"Answer the object that was sent through the receiver first and has not 
	yet been received by anyone. If no object has been sent, answer <nil>."

	readSynch waitIfLocked: [ ^nil ].
	^accessProtect
		critical: [
			| value |
			readPosition = writePosition
					ifTrue: 
						[self error: 'Error in SharedQueue synchronization'.
						 value := nil]
					ifFalse: 
						[value := contentsArray at: readPosition.
						 contentsArray at: readPosition put: nil.
						 readPosition := readPosition + 1].
			value ]! !

!SharedQueue methodsFor: 'accessing' stamp: 'ul 12/8/2010 04:44'!
nextOrNilSuchThat: aBlock
	"Answer the next object that satisfies aBlock, skipping any intermediate objects.
	If no object has been sent, answer <nil> and leave me intact.
	NOTA BENE:  aBlock MUST NOT contain a non-local return (^)."

	^accessProtect critical: [
		| value readPos |
		value := nil.
		readPos := readPosition.
		[ readPos < writePosition and: [ value isNil ] ] whileTrue: [
			value := contentsArray at: readPos.
			readPos := readPos + 1.
			(aBlock value: value)
				ifFalse: [ value := nil ]
				ifTrue: [
					readSynch waitIfLocked: [ ^nil ]. "We found the value, but someone else booked it."
					readPosition to: readPos - 1 do: [ :j | contentsArray at: j put: nil ].
					readPosition := readPos ] ].
		value ].
"===
q := SharedQueue new.
1 to: 10 do: [ :i | q nextPut: i].
c := OrderedCollection new.
[
	v := q nextOrNilSuchThat: [ :e | e odd].
	v notNil
] whileTrue: [
	c add: {v. q size}
].
{c. q} explore
==="! !

!SharedQueue methodsFor: 'accessing' stamp: ''!
nextPut: value 
	"Send value through the receiver. If a Process has been suspended 
	waiting to receive a value through the receiver, allow it to proceed."

	accessProtect
		critical: [writePosition > contentsArray size
						ifTrue: [self makeRoomAtEnd].
				 contentsArray at: writePosition put: value.
				 writePosition := writePosition + 1].
	readSynch signal.
	^value! !

!SharedQueue methodsFor: 'accessing' stamp: 'ul 12/8/2010 04:37'!
peek
	"Answer the object that was sent through the receiver first and has not 
	yet been received by anyone but do not remove it from the receiver. If 
	no object has been sent, return nil"

	^readSynch
		critical: [
			accessProtect critical: [
				readPosition >= writePosition ifFalse: [
					contentsArray at: readPosition ] ] ]
		ifLocked: [ nil ]! !

!SharedQueue methodsFor: 'copying' stamp: 'nice 10/5/2009 09:01'!
postCopy
	super postCopy.
	contentsArray := contentsArray copy.
	accessProtect := Semaphore forMutualExclusion.
	readSynch := Semaphore new! !

!SharedQueue methodsFor: 'private' stamp: 'ar 10/4/2006 12:43'!
printOn: aStream
	super printOn: aStream.
	"Print a guesstimate of the size of the queue without aquiring the lock properly"
	aStream nextPut: $(.
	aStream print: writePosition - readPosition.
	aStream nextPut: $).! !

!SharedQueue methodsFor: 'accessing' stamp: ''!
size
	"Answer the number of objects that have been sent through the
	receiver and not yet received by anyone."

	^writePosition - readPosition! !

Object subclass: #SortFunction
	instanceVariableNames: 'direction collator'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Support'!

!SortFunction commentStamp: 'nice 3/13/2014 22:24' prior: 0!
I am intended to be used in place of two arg sort blocks.

Usage

In the following example, an ascending SortFunction is created based on the result of the #first message send to each object.
#(#(1 2) #(2 3) #(0 0)) sorted: #first ascending.

To sort by the #last element, but descending, the following would be used:
#(#(1 2) #(2 3) #(0 0)) sorted: #last descending.

One can use blocks as well. The following sorts in descending order, the sub elements based on the sum of their values.
| sumBlock |
sumBlock := [:sequence | sequence inject: 0 into: [:sum :each | sum + each]].
#(#(1 2) #(2 3) #(0 0)) sorted: sumBlock descending.

One can even use 2 arg blocks, for those cases where the function isn't expressible with objects that respond to < and =. The only catch, is that such a function has to return not true and false, but instead a collation order, values of -1 (for before), 0 (the same) or 1 (to follow). For example:

| oddBlock |
oddBlock :=
		[:a :b |
		a odd = b odd ifTrue: [0] ifFalse: [a odd ifTrue: [-1] ifFalse: [1]]].
#(1 5 1 3 2 7 9 4 6) asSortedCollection: oddBlock descending

Instance Variables
	collator	<BlockClosure>	This is the collation function that must return a -1, 0, or 1. It is usually composed by an initialization method such as sendMessage: or monadicBlock:, but may be set directly.
	direction	<SmallInteger>	1 for ascending, -1 for descending

!

SortFunction subclass: #ChainedSortFunction
	instanceVariableNames: 'next'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Support'!

!ChainedSortFunction commentStamp: 'nice 3/13/2014 22:25' prior: 0!
I add to my parent the idea of a "next" function to use when two objects are equal by my particular collator.

Usage

SortFunctions can be chained together in primary, secondary, tertiary, etc order using the comma method. Consider a sequence of customer objects, where each customer object responds to the messages firstName, lastName, and age. If we want to sort them lastName first, then firstName, and finally oldest first, we would use an expression like:

customers sort: #lastName ascending, #firstName ascending, #age descending

As noted in my super's comment, unary symbols or single arg blocks can be used. One can omit the the ascending methods on arguments (not the receiver), it will default blocks or symbols to be ascending if none is specified. In other words, the above expression could be simplified slightly as

customers sort: #lastName ascending, #firstName, #age descending

(note the missing ascending on the #firstName argument)

Instance Variables
	next	<SortFunction>	the next SortFunction to evaluate in the event my evaluator results in equal values.

!

!ChainedSortFunction methodsFor: 'initialize-release' stamp: 'nice 3/12/2014 23:06'!
next: anObject

	next := anObject! !

!ChainedSortFunction methodsFor: 'evaluating' stamp: 'nice 3/12/2014 23:07'!
value: anObject value: bObject
	"Refinement of the parent behavior. If the result of my collator is 0, then pass on to the next variable to work it out."

	| result |
	result := (collator value: anObject value: bObject) * direction.
	^result isZero
		ifTrue: [next value: anObject value: bObject]
		ifFalse: [result < 0]! !

!ChainedSortFunction methodsFor: 'converting' stamp: 'nice 3/12/2014 23:07'!
, aSortFunction

	self next: next , aSortFunction! !

!SortFunction class methodsFor: 'instance creation' stamp: 'nice 3/12/2014 23:05'!
ascend

	^self new ascend! !

!SortFunction class methodsFor: 'instance creation' stamp: 'nice 3/12/2014 23:05'!
descend

	^self new descend! !

!SortFunction methodsFor: 'initailize-release' stamp: 'nice 3/12/2014 23:08'!
ascend

	direction := 1! !

!SortFunction methodsFor: 'converting' stamp: 'nice 3/12/2014 23:10'!
asSortFunction

	^self! !

!SortFunction methodsFor: 'initailize-release' stamp: 'nice 3/12/2014 23:09'!
collator: a2ArgBlock
	"a2ArgBlock MUST return the collation order, -1, 0, or 1"

	collator := a2ArgBlock! !

!SortFunction methodsFor: 'initailize-release' stamp: 'nice 3/12/2014 23:08'!
descend

	direction := -1! !

!SortFunction methodsFor: 'initailize-release' stamp: 'nice 3/12/2014 23:09'!
direction: anInteger

	direction := anInteger! !

!SortFunction methodsFor: 'testing' stamp: 'nice 3/12/2014 23:09'!
isAscending

	^direction = 1! !

!SortFunction methodsFor: 'testing' stamp: 'nice 3/12/2014 23:09'!
isDescending

	^direction = -1! !

!SortFunction methodsFor: 'initailize-release' stamp: 'nice 3/12/2014 23:09'!
monadicBlock: aSingleArgBlock
	"Initialze the receiver's collation block to compare the results of evaluating aSingleArgBlock with each argument, and then collate the return values with the <=> method."

	collator := [:a :b | (aSingleArgBlock value: a) <=> (aSingleArgBlock value: b)]! !

!SortFunction methodsFor: 'initailize-release' stamp: 'nice 3/12/2014 23:09'!
sendMessage: aUnarySymbol
	"Initialze the receiver's collation block to compare the results of sending aUnarySymbol to each argument, and then collate them with the <=> method."

	collator := [:a :b | (a perform: aUnarySymbol) <=> (b perform: aUnarySymbol)]! !

!SortFunction methodsFor: 'converting' stamp: 'nice 3/12/2014 23:11'!
toggleDirection
	"Invert my current direction, if I'm currently ascending, this will cause me to be descending now, and vice-versa."

	direction := direction * -1! !

!SortFunction methodsFor: 'evaluating' stamp: 'nice 3/12/2014 23:10'!
value: anObject value: bObject
	"Masquerade as a two argument block, used by many of the sorting APIs, by returning whether anObject should be placed before bObject or not."

	| result |
	result := (collator value: anObject value: bObject) * direction.
	^result <= 0! !

!SortFunction methodsFor: 'converting' stamp: 'nice 3/12/2014 23:10'!
, aSortFunction
	"Return a new SortFunction which is the concatenation of aSortFunction to me, I will be the primary sort, but if I compare equal, I will defer to the argument."

	^(ChainedSortFunction new)
		collator: collator;
		direction: direction;
		next: aSortFunction asSortFunction! !

Object subclass: #Stack
	instanceVariableNames: 'linkedList'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Stack'!

!Stack commentStamp: 'dc 7/24/2005 15:41' prior: 0!
I implement a simple Stack. #push: adds a new object of any kind on top of the stack. #pop returns the first element and remove it from the stack. #top answer the first element of the stack without removing it.!

!Stack methodsFor: 'private' stamp: 'dc 7/25/2005 10:05'!
errorEmptyStack
	self error: 'this stack is empty'! !

!Stack methodsFor: 'initialize-release' stamp: 'dc 7/25/2005 11:39'!
initialize
	super initialize.
	linkedList := LinkedList new! !

!Stack methodsFor: 'testing' stamp: 'dc 7/25/2005 10:05'!
isEmpty
	^ self linkedList isEmpty! !

!Stack methodsFor: 'private' stamp: 'dc 7/25/2005 10:20'!
linkedList
	"The stack is implemented with a LinkedList. Do NOT call this function, it  
	is for private use !!"
	^ linkedList! !

!Stack methodsFor: 'private' stamp: 'dc 7/25/2005 10:05'!
notEmptyCheck
	"Ensure the stack is not empty."
	self isEmpty
		ifTrue: [self errorEmptyStack]! !

!Stack methodsFor: 'removing' stamp: 'dc 7/24/2005 16:16'!
pop
	"Returns the first element and remove it from the stack."

	self notEmptyCheck.
	^self linkedList removeFirst element! !

!Stack methodsFor: 'copying' stamp: 'nice 10/5/2009 08:52'!
postCopy
	super postCopy.
	linkedList := linkedList copy! !

!Stack methodsFor: 'adding' stamp: 'dc 7/25/2005 10:22'!
push: anObject 
	"Adds a new object of any kind on top of the stack."
	self linkedList
		addFirst: (StackLink with: anObject).
	^ anObject.! !

!Stack methodsFor: 'accessing' stamp: 'dc 7/25/2005 10:04'!
size
	"How many objects in me ?"
	^ self linkedList size! !

!Stack methodsFor: 'accessing' stamp: 'sd 3/25/2006 15:05'!
top
	"Answer the first element of the stack without removing it."
	self notEmptyCheck.
	^ self linkedList first element! !

Object subclass: #Stream
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!Stream commentStamp: '<historical>' prior: 0!
I am an abstract class that represents an accessor for a sequence of objects. This sequence is referred to as my "contents".!

Stream subclass: #AttributedTextStream
	instanceVariableNames: 'characters attributeRuns attributeValues currentAttributes currentRun'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!AttributedTextStream commentStamp: '<historical>' prior: 0!
a stream on Text's which keeps track of the last attribute put; new characters are added with those attributes.

instance vars:

	characters - a WriteStream of the characters in the stream
	attributeRuns - a RunArray with the attributes for the stream
	currentAttributes - the attributes to be used for new text
	attributesChanged - whether the attributes have changed since the last addition!

!AttributedTextStream class methodsFor: 'instance creation' stamp: 'gk 2/9/2004 18:50'!
new
	"For this class we override Stream class>>new since this
	class actually is created using #new, even though it is a Stream."
	
	^self basicNew initialize! !

!AttributedTextStream methodsFor: 'accessing' stamp: 'nice 12/9/2009 08:26'!
contents
	| ans |
	currentRun > 0 ifTrue:[
		attributeValues nextPut: currentAttributes.
		attributeRuns nextPut: currentRun.
		currentRun := 0].
	ans := Text string: characters contents  runs: 
		(RunArray runs: attributeRuns contents values: attributeValues contents).
	^ans! !

!AttributedTextStream methodsFor: 'character writing' stamp: 'ul 11/2/2010 21:51'!
cr
	"Append a carriage return character to the receiver."

	self nextPut: Character cr! !

!AttributedTextStream methodsFor: 'character writing' stamp: 'ul 11/2/2010 21:51'!
crlf
	"Append a carriage return and a line feed to the receiver."

	self nextPut: Character cr; nextPut: Character lf! !

!AttributedTextStream methodsFor: 'accessing' stamp: 'ls 6/27/1998 15:09'!
currentAttributes
	"return the current attributes"
	^currentAttributes! !

!AttributedTextStream methodsFor: 'accessing' stamp: 'ar 10/16/2001 22:57'!
currentAttributes: newAttributes
	"set the current attributes"
	(currentRun > 0 and:[currentAttributes ~= newAttributes]) ifTrue:[
		attributeRuns nextPut: currentRun.
		attributeValues nextPut: currentAttributes.
		currentRun := 0.
	].
	currentAttributes := newAttributes.
! !

!AttributedTextStream methodsFor: 'initialize-release' stamp: 'ul 11/2/2010 21:57'!
initialize

	characters := String new writeStream.
	currentAttributes := #().
	currentRun := 0.
	attributeValues := (Array new: 50) writeStream.
	attributeRuns := (Array new: 50) writeStream! !

!AttributedTextStream methodsFor: 'character writing' stamp: 'ul 11/2/2010 21:45'!
lf
	"Append a line feed character to the receiver."

	self nextPut: Character lf! !

!AttributedTextStream methodsFor: 'accessing' stamp: 'nice 3/19/2010 19:12'!
nextPutAll: aString
	"add an entire string with the same attributes"
	currentRun := currentRun + aString size.
	^characters nextPutAll: aString.! !

!AttributedTextStream methodsFor: 'accessing' stamp: 'nice 3/19/2010 19:03'!
nextPut: aChar
	currentRun := currentRun + 1.
	^characters nextPut: aChar! !

!AttributedTextStream methodsFor: 'accessing' stamp: 'ls 9/10/1998 03:36'!
size
	"number of characters in the stream so far"
	^characters size! !

!AttributedTextStream methodsFor: 'accessing' stamp: 'ul 11/2/2010 06:42'!
withAttributes: attributes do: aBlock

	| previousAttributes |
	previousAttributes := currentAttributes.
	[
		self currentAttributes: attributes.
		aBlock value ]
			ensure: [ self currentAttributes: previousAttributes ]! !

!AttributedTextStream methodsFor: 'accessing' stamp: 'ul 11/2/2010 21:57'!
withAttribute: attribute do: aBlock

	^self withAttributes: { attribute } do: aBlock! !

Stream subclass: #Generator
	instanceVariableNames: 'block next continue home'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!Generator commentStamp: 'ar 2/10/2010 20:51' prior: 0!
A Generator transforms callback interfaces into stream interfaces. 

When a producer algorithm provide results as callbacks (blocks) and a consumer algorithm expects streamable input, a Generator transforms one into the other, for example:

	| generator |
	generator := Generator on: [:g| Integer primesUpTo: 100 do:[:prime| g yield: prime]].
	[generator atEnd] whileFalse:[Transcript show: generator next].

Instance Variables
	block:		<BlockClosure> The block associated with the generator.
	continue:	<MethodContext>	The continuation to return to.
	home:		<MethodContext>	The home (root) context of the activated block
	next:		<Object>		The next object to return from the Generator.
!

!Generator class methodsFor: 'instance-creation' stamp: 'lr 1/8/2009 15:54'!
on: aBlock
	^ self basicNew initializeOn: aBlock! !

!Generator methodsFor: 'testing' stamp: 'ar 2/10/2010 21:00'!
atEnd
	"Answer whether the receiver can access any more objects."

	^ continue isNil! !

!Generator methodsFor: 'accessing' stamp: 'ar 2/10/2010 20:54'!
close
	"Close the receiving generator and unwind its ensure-blocks."

	continue ifNotNil:[continue unwindTo: home].
	continue := block := next := nil! !

!Generator methodsFor: 'accessing' stamp: 'lr 4/26/2009 11:50'!
contents
	"Answer the contents of this generator. Do not call this method on infinite generators."

	| stream |
	stream := (Array new: 10) writeStream.
	[ self atEnd ]
		whileFalse: [ stream nextPut: self next ].
	^ stream contents! !

!Generator methodsFor: 'private' stamp: 'ar 2/10/2010 20:46'!
fork
	| result |
	home := thisContext.
	block reentrant value: self.
	thisContext swapSender: continue.
	result := next.
	continue := next := home := nil.
	^ result! !

!Generator methodsFor: 'initialization' stamp: 'lr 1/8/2009 16:18'!
initializeOn: aBlock
	block := aBlock.
	self reset! !

!Generator methodsFor: 'accessing' stamp: 'ar 2/10/2010 20:45'!
next
	"Generate and answer the next object in the receiver."

	^ self atEnd ifFalse: [
		home swapSender: thisContext sender.
		continue := thisContext swapSender: continue
	]! !

!Generator methodsFor: 'accessing' stamp: 'lr 2/10/2010 09:16'!
nextPut: anObject
	"Add anObject into the generator. A synonym to #yield: and value:."

	| previous |
	previous := next.
	next := anObject.
	continue := thisContext swapSender: continue.
	^ previous! !

!Generator methodsFor: 'accessing' stamp: 'lr 2/10/2010 09:16'!
peek
	"Answer the upcoming object of the receiver."

	^ next! !

!Generator methodsFor: 'printing' stamp: 'lr 1/8/2009 16:21'!
printOn: aStream
	aStream nextPutAll: self class name; nextPutAll: ' on: '; print: block! !

!Generator methodsFor: 'public' stamp: 'ar 2/10/2010 21:04'!
reset
	"Reset the generator, i.e., start it over"
	continue ifNotNil:[continue unwindTo: home].
	next := nil.
	continue := thisContext.
	[ self fork ] value! !

!Generator methodsFor: 'accessing' stamp: 'lr 2/10/2010 09:16'!
size
	"A generator does not know its size."

	^ self shouldNotImplement! !

!Generator methodsFor: 'public' stamp: 'ar 2/10/2010 21:01'!
value: anObject 
	"Allows passing generators as arguments to methods expecting blocks.
	A synonym for #yield: / #nextPut:."
	^ self nextPut: anObject! !

!Generator methodsFor: 'public' stamp: 'ar 2/10/2010 20:52'!
yield: anObject 
	"Yield the next value to the consumer of the generator.
	A synonym for #nextPut:"
	^ self nextPut: anObject! !

Stream subclass: #NullStream
	instanceVariableNames: 'binary position'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!NullStream commentStamp: 'ar 2/25/2010 14:49' prior: 0!
NullStream is a stream generating and consuming an infinite number of elements. It can be used as an equivalent of /dev/null or for performance benchmarks.!

!NullStream class methodsFor: 'instance creation' stamp: 'ar 2/25/2010 04:20'!
new
	"Creates a new instance"

	^self basicNew initialize! !

!NullStream methodsFor: 'accessing' stamp: 'ar 2/25/2010 04:18'!
ascii
	"Switches the stream to ascii mode"

	binary := false.! !

!NullStream methodsFor: 'testing' stamp: 'ar 2/25/2010 04:19'!
atEnd
	"Answer whether the receiver can access any more objects."

	^false! !

!NullStream methodsFor: 'accessing' stamp: 'ar 2/25/2010 04:18'!
binary
	"Switches the stream to binary mode"

	binary := true! !

!NullStream methodsFor: 'accessing' stamp: 'ar 2/25/2010 04:18'!
collectionSpecies
	"The type of collection returned by the stream"

	^binary ifTrue:[ByteArray] ifFalse:[ByteString]! !

!NullStream methodsFor: 'accessing' stamp: 'ar 2/25/2010 03:35'!
contents
	"Answer all of the contents of the receiver."

	self shouldNotImplement! !

!NullStream methodsFor: 'writing' stamp: 'nice 2/2/2014 17:43'!
cr! !

!NullStream methodsFor: 'writing' stamp: 'nice 2/2/2014 17:43'!
crtab: n! !

!NullStream methodsFor: 'accessing' stamp: 'nice 5/31/2013 00:01'!
element
	"The element returned by the stream"

	^binary ifTrue:[0] ifFalse:[Character null]! !

!NullStream methodsFor: 'writing' stamp: 'nice 12/29/2012 23:30'!
ensureCr! !

!NullStream methodsFor: 'initialize' stamp: 'ar 2/25/2010 14:45'!
initialize
	"Initialize the receiver"

	binary := false.
	position := 0.! !

!NullStream methodsFor: 'testing' stamp: 'ar 2/25/2010 04:19'!
isBinary
	"Return true if the receiver is a binary byte stream"

	^binary! !

!NullStream methodsFor: 'testing' stamp: 'ar 2/25/2010 04:19'!
isEmpty
	"Answer whether the receiver's contents has no elements."

	^false
! !

!NullStream methodsFor: 'reading' stamp: 'ar 2/25/2010 14:48'!
next
	"Answer the next object accessible by the receiver."

	position := position +1.
	^self element! !

!NullStream methodsFor: 'reading' stamp: 'ar 2/25/2010 04:19'!
nextInto: aCollection
	"Read the next elements of the receiver into aCollection.
	Return aCollection or a partial copy if less than aCollection
	size elements have been read."

	^self next: aCollection size into: aCollection startingAt: 1.! !

!NullStream methodsFor: 'writing' stamp: 'ar 2/25/2010 14:48'!
nextPutAll: aCollection 
	"Append the elements of aCollection to the sequence of objects accessible 
	by the receiver. Answer aCollection."

	position := position + aCollection size.
	^aCollection! !

!NullStream methodsFor: 'writing' stamp: 'ar 2/25/2010 14:48'!
nextPut: anObject 
	"Insert the argument, anObject, as the next object accessible by the 
	receiver. Answer anObject."

	position := position +1.
	^anObject! !

!NullStream methodsFor: 'reading' stamp: 'ar 2/25/2010 14:48'!
next: anInteger 
	"Answer the next anInteger elements of my collection. Must override 
	because default uses self contents species, which might involve a large 
	collection."

	position := position +anInteger.
	^self collectionSpecies new: anInteger! !

!NullStream methodsFor: 'reading' stamp: 'ar 2/25/2010 14:48'!
next: n into: aCollection startingAt: startIndex
	"Read n objects into the given collection. 
	Return aCollection or a partial copy if less than
	n elements have been read."

	position := position +n.
	^aCollection! !

!NullStream methodsFor: 'writing' stamp: 'ar 2/25/2010 04:19'!
next: anInteger putAll: aCollection
	"Store the next anInteger elements from the given collection."

	^self next: anInteger putAll: aCollection startingAt: 1! !

!NullStream methodsFor: 'writing' stamp: 'ar 2/25/2010 14:48'!
next: anInteger putAll: aCollection startingAt: startIndex
	"Store the next anInteger elements from the given collection."

	position := position + anInteger.
	^aCollection! !

!NullStream methodsFor: 'writing' stamp: 'nice 2/2/2014 17:43'!
nl! !

!NullStream methodsFor: 'writing' stamp: 'ar 2/25/2010 04:20'!
peek
	"Answer what would be returned if the message next were sent to the 
	receiver. If the receiver is at the end, answer nil."

	^self element! !

!NullStream methodsFor: 'positioning' stamp: 'ar 2/25/2010 14:47'!
position
	"Answer the current position of accessing the sequence of objects."

	^position! !

!NullStream methodsFor: 'positioning' stamp: 'ar 2/25/2010 14:47'!
position: anInteger 
	"Set the current position for accessing the objects to be anInteger, as long 
	as anInteger is within the bounds of the receiver's contents. If it is not, 
	create an error notification."

	(anInteger >= 0)
		ifTrue: [position := anInteger]
		ifFalse: [self positionError]! !

!NullStream methodsFor: 'reading' stamp: 'nice 3/16/2010 23:04'!
readInto: aCollection startingAt: startIndex count: n
	"Read n objects into the given collection. 
	Return number of elements that have been read."

	position := position + n.
	^n! !

!NullStream methodsFor: 'positioning' stamp: 'ar 2/25/2010 14:47'!
reset
	"Set the receiver's position to the beginning of the sequence of objects."

	position := 0! !

!NullStream methodsFor: 'positioning' stamp: 'ar 2/25/2010 14:47'!
skip: anInteger 
	"Set the receiver's position to be the current position+anInteger. A 
	subclass might choose to be more helpful and select the minimum of the 
	receiver's size and position+anInteger, or the maximum of 1 and 
	position+anInteger for the repositioning."

	self position: position + anInteger! !

!NullStream methodsFor: 'writing' stamp: 'nice 2/2/2014 17:43'!
space! !

!NullStream methodsFor: 'writing' stamp: 'nice 2/2/2014 17:43'!
tab! !

!NullStream methodsFor: 'writing' stamp: 'nice 2/2/2014 17:44'!
tab: n! !

Stream subclass: #PositionableStream
	instanceVariableNames: 'collection position readLimit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!PositionableStream commentStamp: '<historical>' prior: 0!
I represent an accessor for a sequence of objects (a collection) that are externally named by indices so that the point of access can be repositioned. I am abstract in that I do not implement the messages next and nextPut: which are inherited from my superclass Stream.!

!PositionableStream class methodsFor: 'instance creation' stamp: ''!
on: aCollection 
	"Answer an instance of me, streaming over the elements of aCollection."

	^self basicNew on: aCollection! !

!PositionableStream class methodsFor: 'instance creation' stamp: ''!
on: aCollection from: firstIndex to: lastIndex 
	"Answer an instance of me, streaming over the elements of aCollection 
	starting with the element at firstIndex and ending with the one at 
	lastIndex."

	^self basicNew on: (aCollection copyFrom: firstIndex to: lastIndex)! !

!PositionableStream methodsFor: 'converting' stamp: 'tk 2/7/2000 11:08'!
asBinaryOrTextStream
	"Convert to a stream that can switch between bytes and characters"

	^ (RWBinaryOrTextStream with: self contentsOfEntireFile) reset! !

!PositionableStream methodsFor: 'testing' stamp: ''!
atEnd
	"Primitive. Answer whether the receiver can access any more objects.
	Optional. See Object documentation whatIsAPrimitive."

	<primitive: 67>
	^position >= readLimit! !

!PositionableStream methodsFor: 'accessing' stamp: 'dc 8/1/2007 16:21'!
back
	"Go back one element and return it."
	self position = 0 ifTrue: [self errorCantGoBack].
	self skip: -1.
	^ self peek! !

!PositionableStream methodsFor: 'positioning' stamp: 'tk 3/22/2002 19:33'!
backUpTo: subCollection
	"Back up the position to he subCollection.  Position must be somewhere within the stream initially.  Leave it just after it.  Return true if succeeded.  No wildcards, and case does matter."
"Example:
	| strm | strm := ReadStream on: 'zabc abdc'.
	strm setToEnd; backUpTo: 'abc'; position 
"

	| pattern startMatch |
	pattern := ReadStream on: subCollection reversed.
	startMatch := nil.
	[pattern atEnd] whileFalse: 
		[self position = 0 ifTrue: [^ false].
		self skip: -1.
		(self next) = (pattern next) 
			ifTrue: [pattern position = 1 ifTrue: [startMatch := self position]]
			ifFalse: [pattern position: 0.
					startMatch ifNotNil: [
						self position: startMatch-1.
						startMatch := nil]].
		self skip: -1].
	self position: startMatch.
	^ true

! !

!PositionableStream methodsFor: 'accessing - multibyte support' stamp: 'ul 4/11/2011 04:19'!
basicBack

	^self back! !

!PositionableStream methodsFor: 'private basic' stamp: 'eem 3/30/2016 13:36'!
basicUpTo: anObject 
	"Answer a subcollection from the current access position to the 
	occurrence (if any, but not inclusive) of anObject in the receiver. If 
	anObject is not in the collection, answer the entire rest of the receiver."
	| newStream element |
	newStream := WriteStream on: (self collectionSpecies new: 100).
	[self atEnd or: [(element := self next) = anObject]]
		whileFalse: [newStream nextPut: element].
	^newStream contents! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 10/5/2001 12:09'!
boolean
	"Answer the next boolean value from this (binary) stream."

	^ self next ~= 0
! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 10/5/2001 12:11'!
boolean: aBoolean
	"Store the given boolean value on this (binary) stream."

	self nextPut: (aBoolean ifTrue: [1] ifFalse: [0]).
! !

!PositionableStream methodsFor: 'private' stamp: 'nice 11/22/2009 18:03'!
collectionSpecies
	"Answer the species of collection into which the receiver can stream"
	
	^collection species! !

!PositionableStream methodsFor: 'accessing' stamp: ''!
contents
	"Answer with a copy of my collection from 1 to readLimit."

	^collection copyFrom: 1 to: readLimit! !

!PositionableStream methodsFor: 'accessing' stamp: 'sw 3/10/98 13:55'!
contentsOfEntireFile
	"For non-file streams"
	^ self contents! !

!PositionableStream methodsFor: 'accessing' stamp: 'wiz 5/10/2008 20:19'!
errorCantGoBack
	self error: ' CantGoBack '! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 7/16/2001 14:43'!
int16
	"Answer the next signed, 16-bit integer from this (binary) stream."

	| n |
	n := self next.
	n := (n bitShift: 8) + (self next).
	n >= 16r8000 ifTrue: [n := n - 16r10000].
	^ n
! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 7/16/2001 14:44'!
int16: anInteger
	"Store the given signed, 16-bit integer on this (binary) stream."

	| n |
	(anInteger < -16r8000) | (anInteger >= 16r8000)
		ifTrue: [self error: 'outside 16-bit integer range'].

	anInteger < 0
		ifTrue: [n := 16r10000 + anInteger]
		ifFalse: [n := anInteger].
	self nextPut: (n digitAt: 2).
	self nextPut: (n digitAt: 1).
! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 7/16/2001 15:15'!
int32
	"Answer the next signed, 32-bit integer from this (binary) stream."
	"Details: As a fast check for negative number, check the high bit of the first digit"

	| n firstDigit |
	n := firstDigit := self next.
	n := (n bitShift: 8) + self next.
	n := (n bitShift: 8) + self next.
	n := (n bitShift: 8) + self next.
	firstDigit >= 128 ifTrue: [n := -16r100000000 + n].  "decode negative 32-bit integer"
	^ n
! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 7/16/2001 14:46'!
int32: anInteger
	"Store the given signed, 32-bit integer on this (binary) stream."

	| n |
	(anInteger < -16r80000000) | (anInteger >= 16r80000000)
		ifTrue: [self error: 'outside 32-bit integer range'].

	anInteger < 0
		ifTrue: [n := 16r100000000 + anInteger]
		ifFalse: [n := anInteger].
	self nextPut: (n digitAt: 4).
	self nextPut: (n digitAt: 3).
	self nextPut: (n digitAt: 2).
	self nextPut: (n digitAt: 1).
! !

!PositionableStream methodsFor: 'testing' stamp: 'ar 1/2/2000 17:24'!
isBinary
	"Return true if the receiver is a binary byte stream"
	^collection class == ByteArray! !

!PositionableStream methodsFor: 'testing' stamp: 'dc 2/27/2007 16:16'!
isEmpty
	"Answer whether the receiver's contents has no elements."

	"Returns true if both the set of past and future sequence values of
the receiver are empty. Otherwise returns false"

	^ self atEnd and: [position = 0]! !

!PositionableStream methodsFor: 'accessing' stamp: 'bf 3/4/2013 14:38'!
last
	"Return the element that was read last"

	^ collection at: position! !

!PositionableStream methodsFor: 'positioning' stamp: 'hmm 10/22/1999 21:18'!
match: subCollection
	"Set the access position of the receiver to be past the next occurrence of the subCollection. Answer whether subCollection is found.  No wildcards, and case does matter."

	| pattern startMatch |
	pattern := ReadStream on: subCollection.
	startMatch := nil.
	[pattern atEnd] whileFalse: 
		[self atEnd ifTrue: [^ false].
		(self next) = (pattern next) 
			ifTrue: [pattern position = 1 ifTrue: [startMatch := self position]]
			ifFalse: [pattern position: 0.
					startMatch ifNotNil: [
						self position: startMatch.
						startMatch := nil]]].
	^ true

! !

!PositionableStream methodsFor: 'accessing' stamp: 'tk 3/27/98 08:44'!
nextDelimited: terminator
	"Answer the contents of the receiver, up to the next terminator character. Doubled terminators indicate an embedded terminator character.  For example: 'this '' was a quote'. Start postioned before the initial terminator."

	| out ch |
	out := WriteStream on: (String new: 1000).
	self atEnd ifTrue: [^ ''].
	self next == terminator ifFalse: [self skip: -1].	"absorb initial terminator"
	[(ch := self next) == nil] whileFalse: [
		(ch == terminator) ifTrue: [
			self peek == terminator ifTrue: [
				self next.  "skip doubled terminator"
			] ifFalse: [
				^ out contents  "terminator is not doubled; we're done!!"
			].
		].
		out nextPut: ch.
	].
	^ out contents! !

!PositionableStream methodsFor: 'nonhomogeneous accessing' stamp: 'sw 3/10/98 13:55'!
nextInt32
	"Read a 32-bit signed integer from the next 4 bytes"
	| s |
	s := 0.
	1 to: 4 do: [:i | s := (s bitShift: 8) + self next].
	(s bitAnd: 16r80000000) = 0
		ifTrue: [^ s]
		ifFalse: [^ -1 - s bitInvert32]! !

!PositionableStream methodsFor: 'nonhomogeneous accessing' stamp: 'sw 3/10/98 13:55'!
nextInt32Put: int32
	"Write a signed integer to the next 4 bytes"
	| pos |
	pos := int32 < 0
		ifTrue: [(0-int32) bitInvert32 + 1]
		ifFalse: [int32].
	1 to: 4 do: [:i | self nextPut: (pos digitAt: 5-i)].
	^ int32! !

!PositionableStream methodsFor: 'accessing' stamp: 'nice 12/7/2009 08:40'!
nextLine
	"Answer next line (may be empty) without line end delimiters, or nil if at end.
	Let the stream positioned after the line delimiter(s).
	Handle a zoo of line delimiters CR, LF, or CR-LF pair"

	self atEnd ifTrue: [^nil].
	^self upToAnyOf: CharacterSet crlf do: [:char | char = Character cr ifTrue: [self peekFor: Character lf]]! !

!PositionableStream methodsFor: 'nonhomogeneous accessing' stamp: 'jm 4/9/98 21:36'!
nextLittleEndianNumber: n 
	"Answer the next n bytes as a positive Integer or LargePositiveInteger, where the bytes are ordered from least significant to most significant."

	| bytes s |
	bytes := self next: n.
	s := 0.
	n to: 1 by: -1 do: [:i | s := (s bitShift: 8) bitOr: (bytes at: i)].
	^ s
! !

!PositionableStream methodsFor: 'nonhomogeneous accessing' stamp: 'di 12/6/1999 10:13'!
nextLittleEndianNumber: n put: value
	"Answer the next n bytes as a positive Integer or LargePositiveInteger, where the bytes are ordered from least significant to most significant."
	| bytes |
	bytes := ByteArray new: n.
	1 to: n do: [: i | bytes at: i put: (value digitAt: i)].
	self nextPutAll: bytes! !

!PositionableStream methodsFor: 'nonhomogeneous accessing' stamp: ''!
nextNumber: n 
	"Answer the next n bytes as a positive Integer or LargePositiveInteger."
	| s |
	s := 0.
	1 to: n do: 
		[:i | s := (s bitShift: 8) bitOr: self next asInteger].
	^ s normalize! !

!PositionableStream methodsFor: 'nonhomogeneous accessing' stamp: 'sw 3/10/98 13:55'!
nextNumber: n put: v 
	"Append to the receiver the argument, v, which is a positive 
	SmallInteger or a LargePositiveInteger, as the next n bytes.
	Possibly pad with leading zeros."

	1 to: n do: [:i | self nextPut: (v digitAt: n+1-i)].
	^ v
! !

!PositionableStream methodsFor: 'nonhomogeneous accessing' stamp: 'nice 7/18/2014 01:04'!
nextString
	"Read a string from the receiver. The first byte is the length of the string, unless it is greater than 192, in which case the first four bytes encode the length. This is restricted to Latin1 encoded String."

	| length aByteArray |
	length := self next.		"first byte."
	length >= 192 ifTrue: [length := length - 192.
		1 to: 3 do: [:ii | length := length * 256 + self next]].
	aByteArray := ByteArray new: length.

	self nextInto: aByteArray.
	^aByteArray asString.
! !

!PositionableStream methodsFor: 'nonhomogeneous accessing' stamp: 'tk 6/8/1998 21:01'!
nextStringOld
	"Read a string from the receiver. The first byte is the length of the 
	string, unless it is greater than 192, in which case the first *two* bytes 
	encode the length.  Max size 16K. "

	| aString length |
	length := self next.		"first byte."
	length >= 192 ifTrue: [length := (length - 192) * 256 + self next].
	aString := String new: length.
	1 to: length do: [:ii | aString at: ii put: self next asCharacter].
	^aString! !

!PositionableStream methodsFor: 'nonhomogeneous accessing' stamp: 'yo 4/16/2001 17:56'!
nextStringPut: s 
	"Append the string, s, to the receiver.  Only used by DataStream.  Max size of 64*256*256*256."

	| length |
	(length := s size) < 192
		ifTrue: [self nextPut: length]
		ifFalse: 
			[self nextPut: (length digitAt: 4)+192.
			self nextPut: (length digitAt: 3).
			self nextPut: (length digitAt: 2).
			self nextPut: (length digitAt: 1)].
	self nextPutAll: s asByteArray.
	^s! !

!PositionableStream methodsFor: 'nonhomogeneous accessing' stamp: 'sw 3/10/98 13:55'!
nextWord
	"Answer the next two bytes from the receiver as an Integer."

	| high low |
	high := self next.
		high==nil ifTrue: [^false].
	low := self next.
		low==nil ifTrue: [^false].
	^(high asInteger bitShift: 8) + low asInteger! !

!PositionableStream methodsFor: 'nonhomogeneous accessing' stamp: 'sw 3/10/98 13:55'!
nextWordPut: aWord 
	"Append to the receiver an Integer as the next two bytes."

	self nextPut: ((aWord bitShift: -8) bitAnd: 255).
	self nextPut: (aWord bitAnd: 255).
	^aWord! !

!PositionableStream methodsFor: 'accessing' stamp: 'tpr 3/15/2013 12:37'!
nextWordsInto: aBitmap 
	"Fill the word based buffer from my collection. 
	Stored on stream as Big Endian. Optimized for speed. 
	Read in BigEndian, then restoreEndianness."
	| blt pos source byteSize |
	collection class isBytes
		ifFalse: [^ self next: aBitmap size into: aBitmap startingAt: 1].

	byteSize := aBitmap byteSize.
	"is the test on collection basicSize \\ 4 necessary?"
	((self position bitAnd: 3) = 0 and: [ (collection basicSize bitAnd: 3) = 0])
		ifTrue: [source := collection.
			pos := self position.
			self skip: byteSize]
		ifFalse: ["forced to copy it into a buffer"
			source := self next: byteSize.
			pos := 0].

	"Now use BitBlt to copy the bytes to the bitmap."
	blt := (BitBlt
				toForm: (Form new hackBits: aBitmap))
				sourceForm: (Form new hackBits: source).
	blt combinationRule: Form over. "store"
	blt sourceX: 0;
		 sourceY: pos // 4;
		 height: byteSize // 4;
		 width: 4.
	blt destX: 0;
		 destY: 0.
	blt copyBits.

	"And do whatever the bitmap needs to do to convert from big-endian order."
	aBitmap restoreEndianness.

	^ aBitmap 	"May be WordArray, ColorArray, etc"
! !

!PositionableStream methodsFor: 'accessing' stamp: 'nice 11/22/2009 18:10'!
next: anInteger 
	"Answer the next anInteger elements of my collection. Must override 
	because default uses self contents species, which might involve a large 
	collection."

	| newArray |
	newArray := self collectionSpecies new: anInteger.
	1 to: anInteger do: [:index | newArray at: index put: self next].
	^newArray! !

!PositionableStream methodsFor: 'accessing' stamp: 'ar 1/2/2000 15:32'!
next: anInteger putAll: aCollection
	"Store the next anInteger elements from the given collection."
	^self next: anInteger putAll: aCollection startingAt: 1! !

!PositionableStream methodsFor: 'accessing' stamp: 'ul 3/19/2016 13:25'!
next: anInteger putAll: aCollection startingAt: startIndex
	"Store the next anInteger elements from the given collection."
	
	startIndex to: startIndex + anInteger - 1 do: [ :index |
		self nextPut: (aCollection at: index) ].
	^aCollection! !

!PositionableStream methodsFor: 'private' stamp: ''!
on: aCollection

	collection := aCollection.
	readLimit := aCollection size.
	position := 0.
	self reset! !

!PositionableStream methodsFor: 'accessing' stamp: ''!
originalContents
	"Answer the receiver's actual contents collection, NOT a copy.  1/29/96 sw"

	^ collection! !

!PositionableStream methodsFor: 'positioning' stamp: 'di 5/25/1998 15:16'!
padToNextLongPut: char 
	"Make position be on long word boundary, writing the padding 
	character, char, if necessary."
	[self position \\ 4 = 0]
		whileFalse: [self nextPut: char]! !

!PositionableStream methodsFor: 'positioning' stamp: 'di 2/15/98 14:41'!
padTo: nBytes put: aCharacter 
	"Pad using the argument, aCharacter, to the next boundary of nBytes characters."
	| rem |
	rem := nBytes - (self position \\ nBytes).
	rem = nBytes ifTrue: [^ 0].
	self next: rem put: aCharacter.! !

!PositionableStream methodsFor: 'accessing' stamp: ''!
peek
	"Answer what would be returned if the message next were sent to the 
	receiver. If the receiver is at the end, answer nil."

	| nextObject |
	self atEnd ifTrue: [^nil].
	nextObject := self next.
	position := position - 1.
	^nextObject! !

!PositionableStream methodsFor: 'accessing' stamp: 'topa 10/8/2015 20:31'!
peekBack
	"Return the element at the previous position, without changing position.  Use indirect messages in case self is a StandardFileStream."

	| element |
	self position = 0 ifTrue: [self errorCantGoBack].
	self position = 1 ifTrue: [self position: 0.  ^ nil].
	self skip: -2.
	element := self next.
	self skip: 1.
	^ element! !

!PositionableStream methodsFor: 'accessing' stamp: 'ul 3/29/2016 07:23'!
peekFor: anObject 
	"Answer false and do not move over the next element if it is not equal to 
	the argument, anObject, or if the receiver is at the end. Answer true 
	and increment the position for accessing elements, if the next element is 
	equal to anObject."

	self atEnd ifTrue: [^false].
	"gobble it if found"
	self next = anObject ifTrue: [ ^true ].
	position := position - 1.
	^false! !

!PositionableStream methodsFor: 'positioning' stamp: ''!
position
	"Answer the current position of accessing the sequence of objects."

	^position! !

!PositionableStream methodsFor: 'private' stamp: ''!
positionError
	"Since I am not necessarily writable, it is up to my subclasses to override 
	position: if expanding the collection is preferrable to giving this error."

	self error: 'Attempt to set the position of a PositionableStream out of bounds'! !

!PositionableStream methodsFor: 'positioning' stamp: 'mir 6/29/2004 17:35'!
positionOfSubCollection: subCollection
	"Return a position such that that element at the new position equals the first element of sub, and the next elements equal the rest of the elements of sub. Begin the search at the current position.
	If no such match is found, answer 0."

	^self positionOfSubCollection: subCollection ifAbsent: [0]! !

!PositionableStream methodsFor: 'positioning' stamp: 'ul 2/22/2011 16:26'!
positionOfSubCollection: subCollection ifAbsent: exceptionBlock
	"Return a position such that that element at the new position equals the first element of sub, and the next elements equal the rest of the elements of sub. Begin the search at the current position.
	If no such match is found, answer the result of evaluating argument, exceptionBlock."

	| pattern startPosition currentPosition |
	pattern := subCollection readStream.
	startPosition := self position.
	[ pattern atEnd or: [ self atEnd ] ] whileFalse: [
		self next = pattern next ifFalse: [
			self position: self position - pattern position + 1.
			pattern reset ] ].
	currentPosition := self position.
	self position: startPosition.
	pattern atEnd ifTrue: [ ^currentPosition + 1 - subCollection size ].
	^exceptionBlock value! !

!PositionableStream methodsFor: 'positioning' stamp: 'nice 3/10/2008 22:29'!
position: anInteger 
	"Set the current position for accessing the objects to be anInteger, as long 
	as anInteger is within the bounds of the receiver's contents. If it is not, 
	create an error notification."

	(anInteger >= 0 and: [anInteger <= readLimit])
		ifTrue: [position := anInteger]
		ifFalse: [self positionError]! !

!PositionableStream methodsFor: 'positioning' stamp: 'mir 5/14/2003 18:45'!
pushBack: aString
	"Compatibility with SocketStreams"
	self skip: aString size negated! !

!PositionableStream methodsFor: 'positioning' stamp: ''!
reset
	"Set the receiver's position to the beginning of the sequence of objects."

	position := 0! !

!PositionableStream methodsFor: 'positioning' stamp: 'sw 3/10/98 13:55'!
resetContents
	"Set the position and limits to 0."

	position := 0.
	readLimit := 0! !

!PositionableStream methodsFor: 'private' stamp: ''!
setFrom: newStart to: newStop

	position := newStart - 1.
	readLimit := newStop! !

!PositionableStream methodsFor: 'positioning' stamp: ''!
setToEnd
	"Set the position of the receiver to the end of the sequence of objects."

	position := readLimit! !

!PositionableStream methodsFor: 'positioning' stamp: 'di 6/13/97 12:00'!
skipSeparators
	[self atEnd]
		whileFalse:
		[self next isSeparator ifFalse: [^ self position: self position-1]]! !

!PositionableStream methodsFor: 'positioning' stamp: ''!
skipTo: anObject 
	"Set the access position of the receiver to be past the next occurrence of 
	anObject. Answer whether anObject is found."

	[self atEnd]
		whileFalse: [self next = anObject ifTrue: [^true]].
	^false! !

!PositionableStream methodsFor: 'positioning' stamp: ''!
skip: anInteger 
	"Set the receiver's position to be the current position+anInteger. A 
	subclass might choose to be more helpful and select the minimum of the 
	receiver's size and position+anInteger, or the maximum of 1 and 
	position+anInteger for the repositioning."

	self position: position + anInteger! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 9/5/2001 07:35'!
string
	"Answer the next string from this (binary) stream."

	| size |
	size := self uint16.
	^ (self next: size) asString
! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 9/5/2001 12:09'!
string: aString
	"Store the given string on this (binary) stream. The string must contain 65535 or fewer characters."

	aString size > 16rFFFF ifTrue: [self error: 'string too long for this format'].
	self uint16: aString size.
	self nextPutAll: aString asByteArray.
! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 8/20/2001 07:53'!
uint16
	"Answer the next unsigned, 16-bit integer from this (binary) stream."

	| n |
	n := self next.
	n := (n bitShift: 8) + (self next).
	^ n
! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 8/20/2001 07:53'!
uint16: anInteger
	"Store the given unsigned, 16-bit integer on this (binary) stream."

	(anInteger < 0) | (anInteger >= 16r10000)
		ifTrue: [self error: 'outside unsigned 16-bit integer range'].

	self nextPut: (anInteger digitAt: 2).
	self nextPut: (anInteger digitAt: 1).
! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 8/20/2001 08:07'!
uint24
	"Answer the next unsigned, 24-bit integer from this (binary) stream."

	| n |
	n := self next.
	n := (n bitShift: 8) + self next.
	n := (n bitShift: 8) + self next.
	^ n
! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 8/20/2001 08:07'!
uint24: anInteger
	"Store the given unsigned, 24-bit integer on this (binary) stream."

	(anInteger < 0) | (anInteger >= 16r1000000)
		ifTrue: [self error: 'outside unsigned 24-bit integer range'].

	self nextPut: (anInteger digitAt: 3).
	self nextPut: (anInteger digitAt: 2).
	self nextPut: (anInteger digitAt: 1).
! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 8/20/2001 07:53'!
uint32
	"Answer the next unsigned, 32-bit integer from this (binary) stream."

	| n |
	n := self next.
	n := (n bitShift: 8) + self next.
	n := (n bitShift: 8) + self next.
	n := (n bitShift: 8) + self next.
	^ n
! !

!PositionableStream methodsFor: 'data get/put' stamp: 'jm 8/20/2001 07:52'!
uint32: anInteger
	"Store the given unsigned, 32-bit integer on this (binary) stream."

	(anInteger < 0) | (anInteger >= 16r100000000)
		ifTrue: [self error: 'outside unsigned 32-bit integer range'].

	self nextPut: (anInteger digitAt: 4).
	self nextPut: (anInteger digitAt: 3).
	self nextPut: (anInteger digitAt: 2).
	self nextPut: (anInteger digitAt: 1).
! !

!PositionableStream methodsFor: 'positioning' stamp: 'tak 8/5/2005 10:34'!
untilEndWithFork: aBlock displayingProgress: aString 
	| sem done result |
	sem := Semaphore new.
	done := false.
	[[result := aBlock value]
		ensure: [done := true.
			sem signal]] fork.
	self
		untilEnd: [done
				ifTrue: [^ result].
			(Delay forSeconds: 0.2) wait]
		displayingProgress: aString.
	sem wait.
	^ result! !

!PositionableStream methodsFor: 'accessing' stamp: 'tk 7/18/1999 17:10'!
upToAll: aCollection
	"Answer a subcollection from the current access position to the occurrence (if any, but not inclusive) of aCollection. If aCollection is not in the stream, answer the entire rest of the stream."

	| startPos endMatch result |
	startPos := self position.
	(self match: aCollection) 
		ifTrue: [endMatch := self position.
			self position: startPos.
			result := self next: endMatch - startPos - aCollection size.
			self position: endMatch.
			^ result]
		ifFalse: [self position: startPos.
			^ self upToEnd]! !

!PositionableStream methodsFor: 'accessing' stamp: 'nice 12/7/2009 08:38'!
upToAnyOf: aCollection 
	"Answer a subcollection from the current access position to the 
	occurrence (if any, but not inclusive) of any object in the collection. If 
	no matching object is found, answer the entire rest of the receiver."
	^self upToAnyOf: aCollection do: [:matchingObject | ]! !

!PositionableStream methodsFor: 'accessing' stamp: 'nice 12/7/2009 08:30'!
upToAnyOf: subcollection do: aBlock
	"Answer a subcollection from the current access position to the occurrence (if any, but not inclusive) of any object in the collection.
	Evaluate aBlock with this occurence as argument.
	If no matching object is found, don't evaluate aBlock and answer the entire rest of the receiver."
	
	^self collectionSpecies new: 1000 streamContents: [ :stream |
		| ch |
		[ self atEnd or: [ (subcollection includes: (ch := self next)) and: [aBlock value: ch. true] ] ] 
			whileFalse: [ stream nextPut: ch ] ]! !

!PositionableStream methodsFor: 'accessing' stamp: 'nice 11/22/2009 18:09'!
upToEnd
	"Answer a subcollection from the current access position through the last element of the receiver."

	| newStream |
	newStream := WriteStream on: (self collectionSpecies new: 100).
	[self atEnd] whileFalse: [ newStream nextPut: self next ].
	^ newStream contents! !

!PositionableStream methodsFor: 'accessing' stamp: 'nice 11/22/2009 18:09'!
upTo: anObject 
	"Answer a subcollection from the current access position to the 
	occurrence (if any, but not inclusive) of anObject in the receiver. If 
	anObject is not in the collection, answer the entire rest of the receiver."
	| newStream element |
	newStream := WriteStream on: (self collectionSpecies new: 100).
	[self atEnd or: [(element := self next) = anObject]]
		whileFalse: [newStream nextPut: element].
	^newStream contents! !

PositionableStream subclass: #ReadStream
	instanceVariableNames: 'initialPositionOrNil'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!ReadStream commentStamp: '<historical>' prior: 0!
I represent an accessor for a sequence of objects that can only read objects from the sequence.!

!ReadStream class methodsFor: 'instance creation' stamp: ''!
on: aCollection from: firstIndex to: lastIndex 
	"Answer with a new instance streaming over a copy of aCollection from
	firstIndex to lastIndex."

	^self basicNew
		on: aCollection
		from: firstIndex
		to: lastIndex! !

!ReadStream methodsFor: 'accessing' stamp: ''!
ascii! !

!ReadStream methodsFor: 'accessing' stamp: ''!
binary! !

!ReadStream methodsFor: 'accessing' stamp: 'eem 5/2/2014 09:23'!
contents
	"Answer with a copy of my collection from the start to readLimit."

	^collection copyFrom: (initialPositionOrNil ifNil: [1]) to: readLimit! !

!ReadStream methodsFor: 'file stream compatibility' stamp: 'nk 12/13/2002 12:00'!
localName
	^'ReadStream'! !

!ReadStream methodsFor: 'accessing' stamp: ''!
next
	"Primitive. Answer the next object in the Stream represented by the
	receiver. Fail if the collection of this stream is not an Array or a String.
	Fail if the stream is positioned at its end, or if the position is out of
	bounds in the collection. Optional. See Object documentation
	whatIsAPrimitive."

	<primitive: 65>
	position >= readLimit
		ifTrue: [^nil]
		ifFalse: [^collection at: (position := position + 1)]! !

!ReadStream methodsFor: 'accessing' stamp: 'ul 4/2/2015 00:00'!
nextFloat
	"Read a floating point value from the receiver. This method is highly optimized for cases
	where many floating point values need to be read subsequently. And if this needs to go
	even faster, look at the inner loops fetching the characters - moving those into a plugin
	would speed things up even more."
	| buffer count sign index cc value digit fraction exp startIndex anyDigit digitNeeded |
	buffer := collection.
	count := readLimit.
	index := position+1.

	"Skip separators"
	index := ByteString findFirstInString: buffer inSet: CharacterSet nonSeparators byteArrayMap startingAt: index.
	index = 0 ifTrue:[self setToEnd. ^nil].

	"check for sign"
	digitNeeded := false.
	sign := 1. cc := buffer byteAt: index.
	cc = 45 "$- asciiValue"
		ifTrue:[sign := -1. index := index+1. digitNeeded := true]
		ifFalse:[cc =  43 "$+ asciiValue" ifTrue:[index := index+1. digitNeeded := true]].

	"Read integer part"
	startIndex := index.
	value := 0.
	[index <= count and:[
		digit := (buffer byteAt: index) - 48. "$0 asciiValue"
		digit >= 0 and:[digit <= 9]]] whileTrue:[
			value := value * 10 + digit.
			index := index + 1.
	].
	anyDigit := index > startIndex.
	index > count ifTrue:[
		(digitNeeded and:[anyDigit not]) ifTrue:[^self error: 'At least one digit expected'].
		self setToEnd. ^value asFloat * sign].

	(buffer byteAt: index) = 46 "$. asciiValue" ifTrue:["<integer>.<fraction>"
		index := index+1.
		startIndex := index.
		"NOTE: fraction and exp below can overflow into LargeInteger range. If they do, then things slow down horribly due to the relatively slow LargeInt -> Float conversion. This can be avoided by changing fraction and exp to use floats to begin with (0.0 and 1.0 respectively), however, this will give different results to Float>>readFrom: and it is not clear if that is acceptable here."
		fraction := 0. exp := 1.
		[index <= count and:[
			digit := (buffer byteAt: index) - 48. "$0 asciiValue"
			digit >= 0 and:[digit <= 9]]] whileTrue:[
				fraction := fraction * 10 + digit.
				exp := exp * 10.
				index := index + 1.
		].
		value := value + (fraction asFloat / exp asFloat).
		anyDigit := anyDigit or:[index > startIndex].
	].
	value := value asFloat * sign.

	"At this point we require at least one digit to avoid allowing:
		- . ('0.0' without leading digits)
		- e32 ('0e32' without leading digits) 
		- .e32 ('0.0e32' without leading digits)
	but these are currently allowed:
		- .5 (0.5)
		- 1. ('1.0')
		- 1e32 ('1.0e32')
		- 1.e32 ('1.0e32')
		- .5e32 ('0.5e32')
	"
	anyDigit ifFalse:["Check for NaN/Infinity first"
		(count - index >= 2 and:[(buffer copyFrom: index to: index+2) = 'NaN'])
			ifTrue:[position := index+2. ^Float nan * sign].
		(count - index >= 7 and:[(buffer copyFrom: index to: index+7) = 'Infinity'])
			ifTrue:[position := index+7. ^Float infinity * sign].
		^self error: 'At least one digit expected'
	].

	index > count ifTrue:[self setToEnd. ^value asFloat].

	(buffer byteAt: index) = 101 "$e asciiValue" ifTrue:["<number>e[+|-]<exponent>"
		index := index+1. "skip e"
		sign := 1. cc := buffer byteAt: index.
		cc = 45 "$- asciiValue"
			ifTrue:[sign := -1. index := index+1]
			ifFalse:[cc = 43 "$+ asciiValue" ifTrue:[index := index+1]].
		startIndex := index.
		exp := 0. anyDigit := false.
		[index <= count and:[
			digit := (buffer byteAt: index) - 48. "$0 asciiValue"
			digit >= 0 and:[digit <= 9]]] whileTrue:[
				exp := exp * 10 + digit.
				index := index + 1.
		].
		index> startIndex ifFalse:[^self error: 'Exponent expected'].
		value := value * (10.0 raisedToInteger: exp * sign).
	].

	position := index-1.
	^value! !

!ReadStream methodsFor: 'accessing' stamp: ''!
nextPut: anObject

	self shouldNotImplement! !

!ReadStream methodsFor: 'accessing' stamp: 'ls 8/16/1998 00:46'!
next: anInteger 
	"Answer the next anInteger elements of my collection.  overriden for efficiency"

	| ans endPosition |

	endPosition := position + anInteger  min:  readLimit.
	ans := collection copyFrom: position+1 to: endPosition.
	position := endPosition.
	^ans
! !

!ReadStream methodsFor: 'accessing' stamp: 'ar 12/23/1999 15:01'!
next: n into: aCollection startingAt: startIndex
	"Read n objects into the given collection. 
	Return aCollection or a partial copy if less than
	n elements have been read."
	| max |
	max := (readLimit - position) min: n.
	aCollection 
		replaceFrom: startIndex 
		to: startIndex+max-1
		with: collection
		startingAt: position+1.
	position := position + max.
	max = n
		ifTrue:[^aCollection]
		ifFalse:[^aCollection copyFrom: 1 to: startIndex+max-1]! !

!ReadStream methodsFor: 'private' stamp: 'eem 5/2/2014 09:22'!
on: aCollection from: firstIndex to: lastIndex

	| len |
	collection := aCollection.
	readLimit :=  lastIndex > (len := collection size)
						ifTrue: [len]
						ifFalse: [lastIndex].
	position := firstIndex <= 1
				ifTrue: [0]
				ifFalse: [firstIndex - 1].
	initialPositionOrNil := position + 1! !

!ReadStream methodsFor: 'file stream compatibility' stamp: 'nk 12/13/2002 12:01'!
openReadOnly! !

!ReadStream methodsFor: 'accessing' stamp: 'nice 3/16/2010 23:08'!
readInto: aCollection startingAt: startIndex count: n
	"Read n objects into the given collection. 
	Return number of elements that have been read."
	| max |
	max := (readLimit - position) min: n.
	aCollection 
		replaceFrom: startIndex 
		to: startIndex+max-1
		with: collection
		startingAt: position+1.
	position := position + max.
	^max! !

!ReadStream methodsFor: 'file stream compatibility' stamp: 'nk 12/13/2002 12:00'!
readOnly! !

!ReadStream methodsFor: 'accessing' stamp: 'ajh 9/5/2002 22:11'!
readStream
	"polymorphic with SequenceableCollection.  Return self"

	^ self! !

!ReadStream methodsFor: 'accessing' stamp: 'ar 11/2/1998 12:20'!
size
	"Compatibility with other streams (e.g., FileStream)"
	^readLimit! !

!ReadStream methodsFor: 'accessing' stamp: 'ul 4/25/2016 02:53'!
skipTo: anObject
	"fast version using indexOf:"

	(position := collection indexOf: anObject startingAt: position + 1) = 0 ifTrue: [
		"not present. consume all characters"
		position := readLimit.
		^false ].
	^true! !

!ReadStream methodsFor: 'accessing' stamp: 'nice 12/8/2009 14:22'!
upToAnyOf: aCollection do: aBlock
	"Overriden for speed"
	| end result |
	end := collection indexOfAnyOf: aCollection startingAt: 1 + position ifAbsent: [0].
	(end = 0 or: [end > readLimit]) ifTrue: [^self upToEnd].
	result := collection copyFrom: 1 + position to: -1 + end.
	position := end.
	aBlock value: (collection at: end).
	^result! !

!ReadStream methodsFor: 'accessing' stamp: 'nice 12/8/2009 15:24'!
upToEnd
	| start |
	start := 1 + position.
	position := readLimit.
	^collection copyFrom: start to: position! !

!ReadStream methodsFor: 'accessing' stamp: 'ul 4/25/2016 02:06'!
upTo: anObject
	"fast version using indexOf:"
	| start end |

	start := position+1.
	end := collection indexOf: anObject startingAt: start.

	"not present--return rest of the collection"	
	(end = 0 or: [end > readLimit]) ifTrue: [ ^self upToEnd ].

	"skip to the end and return the data passed over"
	position := end.
	^collection copyFrom: start to: (end-1)! !

PositionableStream subclass: #WriteStream
	instanceVariableNames: 'writeLimit initialPositionOrNil'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!WriteStream commentStamp: '<historical>' prior: 0!
I represent an accessor for a sequence of objects that can only store objects in the sequence.!

WriteStream subclass: #LimitedWriteStream
	instanceVariableNames: 'limit limitBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!LimitedWriteStream commentStamp: '<historical>' prior: 0!
A LimitedWriteStream is a specialized WriteStream that has a maximum size of the collection it streams over. When this limit is reached a special limitBlock is executed. This can for example be used to "bail out" of lengthy streaming operations before they have finished.  For a simple example take a look at the universal Object printString.

The message SequenceableCollection class streamContents:limitedTo: creates a LimitedWriteStream. In this case it prevents very large (or possibly recursive) object structures to "overdo" their textual representation. !

!LimitedWriteStream methodsFor: 'as yet unclassified' stamp: 'nice 3/19/2010 19:14'!
nextPutAll: aCollection

	| newEnd |
	collection class == aCollection class ifFalse:
		[^ super nextPutAll: aCollection ].

	newEnd := position + aCollection size.
	newEnd > limit ifTrue: [
		super nextPutAll: (aCollection copyFrom: 1 to: (limit - position max: 0)).
		limitBlock value.
		^aCollection
	].
	newEnd > writeLimit ifTrue: [
		self growTo: newEnd + 10
	].

	collection replaceFrom: position+1 to: newEnd  with: aCollection startingAt: 1.
	position := newEnd.
	^aCollection! !

!LimitedWriteStream methodsFor: 'accessing' stamp: 'nice 3/19/2010 19:11'!
nextPut: anObject 
	"Ensure that the limit is not exceeded"

	position >= limit
		ifTrue:
			[ limitBlock value.
			^anObject ].
    ^super nextPut: anObject
! !

!LimitedWriteStream methodsFor: 'as yet unclassified' stamp: 'di 10/28/2001 12:49'!
pastEndPut: anObject
	collection size >= limit ifTrue: [limitBlock value].  "Exceptional return"
	^ super pastEndPut: anObject! !

!LimitedWriteStream methodsFor: 'as yet unclassified' stamp: 'di 6/20/97 09:07'!
setLimit: sizeLimit limitBlock: aBlock
	"Limit the numer of elements this stream will write..."
	limit := sizeLimit.
	"Execute this (typically ^ contents) when that limit is exceded"
	limitBlock := aBlock! !

WriteStream subclass: #ReadWriteStream
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!ReadWriteStream commentStamp: '<historical>' prior: 0!
I represent an accessor for a sequence of objects. My instances can both read and store objects.!

!ReadWriteStream methodsFor: 'file status' stamp: ''!
close
	"Presumably sets the status of the receiver to be closed. This message does 
	nothing at this level, but is included for FileStream compatibility."

	^self! !

!ReadWriteStream methodsFor: 'file status' stamp: ''!
closed
	"If you have close (for FileStream compatibility), you must respond to closed.  The result in nonsense here.  TK 29 May 96"

	^ false! !

!ReadWriteStream methodsFor: 'accessing' stamp: 'eem 5/2/2014 09:23'!
contents
	"Answer with a copy of my collection from the start to readLimit."

	readLimit := readLimit max: position.
	^collection copyFrom: (initialPositionOrNil ifNil: [1]) to: readLimit! !

!ReadWriteStream methodsFor: 'testing' stamp: 'tk 12/2/2001 17:13'!
hash

	self class == ReadWriteStream ifFalse: [^ super hash].
	^ (self position + readLimit + 53) hash! !

!ReadWriteStream methodsFor: 'accessing' stamp: ''!
name
	^ 'a stream'   "for fileIn compatibility"! !

!ReadWriteStream methodsFor: 'accessing' stamp: ''!
next
	"Primitive. Return the next object in the Stream represented by the
	receiver. Fail if the collection of this stream is not an Array or a String.
	Fail if the stream is positioned at its end, or if the position is out of
	bounds in the collection. Optional. See Object documentation
	whatIsAPrimitive."

	<primitive: 65>
	"treat me as a FIFO"
	position >= readLimit
		ifTrue: [^nil]
		ifFalse: [^collection at: (position := position + 1)]! !

!ReadWriteStream methodsFor: 'accessing' stamp: 'ar 8/5/2003 02:23'!
next: anInteger 
	"Answer the next anInteger elements of my collection.  overriden for efficiency"

	| ans endPosition |
	readLimit := readLimit max: position.

	endPosition := position + anInteger  min:  readLimit.
	ans := collection copyFrom: position+1 to: endPosition.
	position := endPosition.
	^ans
! !

!ReadWriteStream methodsFor: 'converting' stamp: 'ajh 9/14/2002 20:37'!
readStream
	"polymorphic with SequenceableCollection.  Return self"

	^ self! !

!ReadWriteStream methodsFor: 'testing' stamp: 'tk 11/29/2001 12:47'!
= other

	(self class == ReadWriteStream and: [other class == ReadWriteStream]) ifFalse: [
		^ super = other].	"does an identity test.  Don't read contents of FileStream"
	^ self position = other position and: [self contents = other contents]! !

ReadWriteStream subclass: #RWBinaryOrTextStream
	instanceVariableNames: 'isBinary'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!RWBinaryOrTextStream commentStamp: '<historical>' prior: 0!
A simulation of a FileStream, but living totally in memory.  Hold the contents of a file or web page from the network.  Can then fileIn like a normal FileStream.

Need to be able to switch between binary and text, as a FileStream does, without recopying the whole collection.  Convert to binary upon input and output.  Always keep as text internally.!

!RWBinaryOrTextStream methodsFor: 'converting' stamp: 'tk 2/4/2000 09:15'!
asBinaryOrTextStream

	^ self! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'tk 6/25/97 13:22'!
ascii
	isBinary := false! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'tk 6/20/97 19:46'!
binary
	isBinary := true! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'tk 6/21/97 12:49'!
contents
	"Answer with a copy of my collection from 1 to readLimit."

	| newArray |
	isBinary ifFalse: [^ super contents].	"String"
	readLimit := readLimit max: position.
	newArray := ByteArray new: readLimit.
	^ newArray replaceFrom: 1
		to: readLimit
		with: collection
		startingAt: 1.! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'jm 11/4/97 08:25'!
contentsOfEntireFile
	"For compatibility with file streams."

	^ self contents! !

!RWBinaryOrTextStream methodsFor: 'testing' stamp: 'tk 6/20/97 19:47'!
isBinary
	^ isBinary! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'tk 12/13/97 13:07'!
next

	| byte |
	^ isBinary 
			ifTrue: [byte := super next.
				 byte ifNil: [nil] ifNotNil: [byte asciiValue]]
			ifFalse: [super next].
! !

!RWBinaryOrTextStream methodsFor: 'writing' stamp: 'ar 9/1/2009 00:22'!
nextPutAll: aCollection
	"Optimized for ByteArrays"
	aCollection class == ByteArray 
		ifTrue:[^super nextPutAll: aCollection asString].
	^super nextPutAll: aCollection! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'nice 3/19/2010 19:06'!
nextPut: charOrByte

	^super nextPut: charOrByte asCharacter! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'ar 4/10/2005 19:26'!
next: anInteger 
	"Answer the next anInteger elements of my collection. Must override to get class right."

	| newArray |
	newArray := (isBinary ifTrue: [ByteArray] ifFalse: [ByteString]) new: anInteger.
	^ self nextInto: newArray! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'ls 3/27/2000 22:24'!
next: n into: aCollection startingAt: startIndex
	"Read n objects into the given collection. 
	Return aCollection or a partial copy if less than n elements have been read."
	"Overriden for efficiency"
	| max |
	max := (readLimit - position) min: n.
	aCollection 
		replaceFrom: startIndex 
		to: startIndex+max-1
		with: collection
		startingAt: position+1.
	position := position + max.
	max = n
		ifTrue:[^aCollection]
		ifFalse:[^aCollection copyFrom: 1 to: startIndex+max-1]! !

!RWBinaryOrTextStream methodsFor: 'writing' stamp: 'ar 9/1/2009 00:23'!
next: anInteger putAll: aCollection startingAt: startIndex
	"Optimized for ByteArrays"
	aCollection class == ByteArray 
		ifTrue:[^super next: anInteger putAll: aCollection asString startingAt: startIndex].
	^super next: anInteger putAll: aCollection startingAt: startIndex! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'nice 7/28/2014 22:40'!
peekLast
	"Return that item just put at the end of the stream"

	^ position > 0 
		ifTrue: [self isBinary
			ifTrue: [(collection at: position) asInteger]
			ifFalse: [(collection at: position) asCharacter]]
		ifFalse: [nil]! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'nice 3/16/2010 23:06'!
readInto: aCollection startingAt: startIndex count: n
	"Read n objects into the given collection. 
	Return number of elements that have been read."
	"Overriden for efficiency"
	| max |
	max := (readLimit - position) min: n.
	aCollection 
		replaceFrom: startIndex 
		to: startIndex+max-1
		with: collection
		startingAt: position+1.
	position := position + max.
	^max! !

!RWBinaryOrTextStream methodsFor: 'positioning' stamp: 'tk 6/21/97 13:04'!
reset
	"Set the receiver's position to the beginning of the sequence of objects."

	super reset.
	isBinary ifNil: [isBinary := false].
	collection class == ByteArray ifTrue: ["Store as String and convert as needed."
		collection := collection asString.
		isBinary := true].
! !

!RWBinaryOrTextStream methodsFor: 'as yet unclassified' stamp: 'tk 10/1/1998 11:54'!
setFileTypeToObject
	"do nothing.  We don't have a file type"! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'tk 6/20/97 19:47'!
text
	isBinary := false! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'ar 4/10/2005 19:27'!
upToEnd
	"Must override to get class right."
	| newArray |
	newArray := (isBinary ifTrue: [ByteArray] ifFalse: [ByteString]) new: self size - self position.
	^ self nextInto: newArray! !

!RWBinaryOrTextStream methodsFor: 'accessing' stamp: 'ul 8/9/2011 13:24'!
upTo: anObject
	"fast version using indexOf:"

	| start end |
	start := position+1.
	isBinary
		ifTrue: [ anObject isInteger ifFalse: [ ^self upToEnd ] ]
		ifFalse: [ anObject isCharacter ifFalse: [ ^self upToEnd ] ].
	end := collection indexOf: anObject asCharacter startingAt: start ifAbsent: [ 0 ].
	"not present--return rest of the collection"	
	(end = 0 or: [end > readLimit]) ifTrue: [ ^self upToEnd ].
	"skip to the end and return the data passed over"
	position := end.
	^((isBinary ifTrue: [ ByteArray ] ifFalse: [ String ]) new: end - start)
		replaceFrom: 1
		to: end - start
		with: collection
		startingAt: start! !

WriteStream subclass: #TextStream
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Streams'!

!TextStream methodsFor: 'as yet unclassified' stamp: ''!
applyAttribute: att beginningAt: startPos
	collection addAttribute: att from: startPos to: self position! !

!TextStream methodsFor: 'as yet unclassified' stamp: 'nice 3/19/2010 19:15'!
nextPutAll: aCollection 
	"Optimized access to get around Text at:Put: overhead"
	| n |
	n := aCollection size.
	position + n > writeLimit
		ifTrue:
			[self growTo: position + n + 10].
	collection 
		replaceFrom: position+1
		to: position + n
		with: aCollection
		startingAt: 1.
	position := position + n.
	^aCollection! !

!TextStream methodsFor: 'as yet unclassified' stamp: 'djp 11/6/1999 20:30'!
withAttributes: attributes do: streamBlock 
	| pos1 val |
	pos1 := self position.
	val := streamBlock value.
	attributes do: [:attribute |
		collection
			addAttribute: attribute
			from: pos1 + 1
			to: self position].
	^ val! !

!TextStream methodsFor: 'as yet unclassified' stamp: ''!
withAttribute: att do: strmBlock
	| pos1 val |
	pos1 := self position.
	val := strmBlock value.
	collection addAttribute: att from: pos1+1 to: self position.
	^ val! !

WriteStream subclass: #TranscriptStream
	instanceVariableNames: 'lastChar'
	classVariableNames: 'AccessSema ForceUpdate RedirectToStdOut'
	poolDictionaries: ''
	category: 'Collections-Streams'!

!TranscriptStream commentStamp: 'fbs 12/30/2013 09:53' prior: 0!
This class is a much simpler implementation of Transcript protocol that supports multiple views and very simple conversion to morphic.  Because it inherits from Stream, it is automatically compatible with code that is designed to write to streams.!

!TranscriptStream class methodsFor: 'preferences' stamp: 'mt 5/1/2015 19:48'!
forceUpdate

	<preference: 'Force transcript updates to screen'
		categoryList: #(printing morphic debug)
		description: 'When enabled, transcript updates will immediately shown in the screen no matter how busy the UI process is.'
		type: #Boolean>
	^ ForceUpdate ifNil: [true]! !

!TranscriptStream class methodsFor: 'preferences' stamp: 'mt 5/1/2015 19:49'!
forceUpdate: aBoolean

	ForceUpdate := aBoolean.! !

!TranscriptStream class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 12:05'!
initialize

	self registerInFlapsRegistry.	! !

!TranscriptStream class methodsFor: 'instance creation' stamp: 'di 5/8/1998 13:51'!
new
	^ self on: (String new: 1000)
"
INSTALLING:
TextCollector allInstances do:
	[:t | t breakDependents.
	t become: TranscriptStream new].

TESTING: (Execute this text in a workspace)
Do this first...
	tt := TranscriptStream new.
	tt openLabel: 'Transcript test 1'.
Then this will open a second view -- ooooh...
	tt openLabel: 'Transcript test 2'.
And finally make them do something...
	tt clear.
	[Sensor anyButtonPressed] whileFalse:
		[1 to: 20 do: [:i | tt print: (2 raisedTo: i-1); cr; endEntry]].
"! !

!TranscriptStream class methodsFor: 'instance creation' stamp: 'di 5/8/1998 12:44'!
newTranscript: aTextCollector 
	"Store aTextCollector as the value of the system global Transcript."
	Smalltalk at: #Transcript put: aTextCollector! !

!TranscriptStream class methodsFor: 'preferences' stamp: 'tpr 8/12/2016 17:10'!
redirectToStdOut
	<preference: 'Redirect transcript to stdout'
		categoryList: #(printing morphic debug)
		description: 'When enabled, anything sent to the transcript will be redirected to the stdout stream and (hopefully) the OS terminal.'
		type: #Boolean>
	^ RedirectToStdOut ifNil: [false]! !

!TranscriptStream class methodsFor: 'preferences' stamp: 'mt 5/1/2015 20:05'!
redirectToStdOut: aBoolean

	RedirectToStdOut := aBoolean.! !

!TranscriptStream class methodsFor: 'class initialization' stamp: 'KR 8/12/2007 11:01'!
registerInFlapsRegistry
	"Register the receiver in the system's flaps registry"
	self environment
		at: #Flaps
		ifPresent: [:cl | cl registerQuad: {#TranscriptStream. #openMorphicTranscript.	'Transcript' translatedNoop.			'A Transcript is a window usable for logging and debugging; browse references to #Transcript for examples of how to write to it.' translatedNoop}
						forFlapNamed: 'Tools']
! !

!TranscriptStream class methodsFor: 'preferences' stamp: 'mt 7/9/2016 11:04'!
themeProperties
	
	^ Model themeProperties! !

!TranscriptStream class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 12:41'!
unload
	"Unload the receiver from global registries"

	self environment at: #Flaps ifPresent: [:cl |
	cl unregisterQuadsWithReceiver: self] ! !

!TranscriptStream methodsFor: 'model protocol' stamp: 'mt 7/9/2016 11:02'!
applyUserInterfaceTheme

	self dependents do: [:ea |
		ea isSystemWindow ifTrue: [
			ea refreshWindowColor]].! !

!TranscriptStream methodsFor: 'stream extensions' stamp: 'sma 3/15/2000 21:28'!
bs
	self position > 0 ifTrue: [^ self skip: -1].
	self changed: #bs! !

!TranscriptStream methodsFor: 'access' stamp: 'di 3/16/1999 21:38'!
characterLimit
	"Tell the views how much to retain on screen"
	^ 20000! !

!TranscriptStream methodsFor: 'stream extensions' stamp: 'di 5/8/1998 12:35'!
clear
	"Clear all characters and redisplay the view"
	self changed: #clearText.
	self reset! !

!TranscriptStream methodsFor: 'initialization' stamp: 'dtl 2/10/2010 17:15'!
closeAllViews
	"Transcript closeAllViews"

	self changed: #close
! !

!TranscriptStream methodsFor: 'accessing' stamp: 'eem 4/11/2012 09:35'!
contents
	"Override to update lastChar."
	position > 0 ifTrue:
		[lastChar := collection at: position].
	^super contents! !

!TranscriptStream methodsFor: 'private' stamp: 'dtl 2/10/2010 17:32'!
countOpenTranscripts
	"Transcript countOpenTranscripts"

	^ (self dependents select: [:e | e isTextView]) size
! !

!TranscriptStream methodsFor: 'stream extensions' stamp: 'mt 5/1/2015 19:50'!
endEntry
	"Display all the characters since the last endEntry, and reset the stream"
	self semaphore critical:[
		self class forceUpdate
			ifTrue: [self changed: #appendEntry]
			ifFalse: [self changed: #appendEntryLater].
		self reset.
	].! !

!TranscriptStream methodsFor: 'stream extensions' stamp: 'sma 4/22/2000 16:58'!
flush
	self endEntry
! !

!TranscriptStream methodsFor: 'stream extensions' stamp: 'di 5/8/1998 12:35'!
pastEndPut: anObject
	"If the stream reaches its limit, just output the contents and reset."
	self endEntry.
	^ self nextPut: anObject! !

!TranscriptStream methodsFor: 'character writing' stamp: 'eem 7/28/2014 09:49'!
peekLast
	"Return that item just put at the end of the stream"

	^ position > 0 
		ifTrue: [collection at: position]
		ifFalse: [lastChar]! !

!TranscriptStream methodsFor: 'model protocol' stamp: 'di 5/29/1998 17:13'!
perform: selector orSendTo: otherTarget
	"Selector was just chosen from a menu by a user.  If can respond, then
perform it on myself. If not, send it to otherTarget, presumably the
editPane from which the menu was invoked."

	(self respondsTo: selector)
		ifTrue: [^ self perform: selector]
		ifFalse: [^ otherTarget perform: selector]! !

!TranscriptStream methodsFor: 'model protocol' stamp: 'di 5/3/1999 22:49'!
release

	self dependents do:
		[:view | (view isMorph and: [view isInWorld not])
					ifTrue: [self removeDependent: view]]! !

!TranscriptStream methodsFor: 'positioning' stamp: 'eem 1/15/2016 16:21'!
reset
	"Override to set lastChar"
	position > 0 ifTrue:
		[lastChar := collection at: position].
	^super reset! !

!TranscriptStream methodsFor: 'private' stamp: 'mir 1/11/2000 11:41'!
semaphore
	^AccessSema ifNil:[AccessSema := Semaphore forMutualExclusion]! !

!TranscriptStream methodsFor: 'stream extensions' stamp: 'mt 5/1/2015 20:01'!
showln: anObject
	"TextCollector compatibility. Ensure a new line before inserting a message."
	
	[
		self target
			cr;
			nextPutAll: anObject asString.
		self endEntry.
	] on: FileWriteError do: [self class redirectToStdOut: false].! !

!TranscriptStream methodsFor: 'stream extensions' stamp: 'mt 5/1/2015 20:01'!
show: anObject
	"TextCollector compatibility"
	
	[
		self target nextPutAll: anObject asString.
		self endEntry
	] on: FileWriteError do: [self class redirectToStdOut: false].! !

!TranscriptStream methodsFor: 'model protocol' stamp: 'sw 3/2/2001 10:18'!
step
	"Objects that may be models of SystemWindows need to respond to this, albeit vacuously"! !

!TranscriptStream methodsFor: 'stream extensions' stamp: 'mt 5/1/2015 19:55'!
target

	^ self class redirectToStdOut
		ifTrue: [FileStream stdout]
		ifFalse: [self]! !

!TranscriptStream methodsFor: 'model protocol' stamp: 'mt 5/13/2016 21:35'!
windowActiveOnFirstClick

	^ Model windowActiveOnFirstClick! !

!TranscriptStream methodsFor: 'model protocol' stamp: 'mt 7/10/2016 12:20'!
windowColorToUse

	^ Color colorFrom: (Model useColorfulWindows
		ifTrue: [self userInterfaceTheme customWindowColor ifNil: [Color r: 0.9 g: 0.75 b: 0.45]]
		ifFalse: [self userInterfaceTheme uniformWindowColor ifNil: [Color veryVeryLightGray]])! !

!WriteStream class methodsFor: 'instance creation' stamp: ''!
on: aCollection from: firstIndex to: lastIndex 
	"Answer an instance of me on a copy of the argument, aCollection, 
	determined by the indices firstIndex and lastIndex. Position the instance 
	at the beginning of the collection."

	^self basicNew
		on: aCollection
		from: firstIndex
		to: lastIndex! !

!WriteStream class methodsFor: 'instance creation' stamp: ''!
with: aCollection 
	"Answer an instance of me on the argument, aCollection, positioned to 
	store objects at the end of aCollection."

	^self basicNew with: aCollection! !

!WriteStream class methodsFor: 'instance creation' stamp: ''!
with: aCollection from: firstIndex to: lastIndex 
	"Answer an instance of me on the subcollection of the argument, 
	aCollection, determined by the indices firstIndex and lastIndex. Position 
	the instance to store at the end of the subcollection."

	^self basicNew with: (aCollection copyFrom: firstIndex to: lastIndex)! !

!WriteStream methodsFor: 'private' stamp: 'di 11/18/1999 22:55'!
braceArray
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	^ collection! !

!WriteStream methodsFor: 'private' stamp: 'di 11/18/1999 22:50'!
braceArray: anArray
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	collection := anArray.
	position := 0.
	readLimit := 0.
	writeLimit := anArray size.! !

!WriteStream methodsFor: 'accessing' stamp: 'eem 5/2/2014 09:22'!
contents
	"Answer with a copy of my collection from the start to the current position."
	readLimit := readLimit max: position.
	^collection copyFrom: (initialPositionOrNil ifNil: [1]) to: position! !

!WriteStream methodsFor: 'character writing' stamp: 'ul 11/2/2010 21:50'!
cr
	"Append a carriage return character to the receiver."

	self nextPut: Character cr! !

!WriteStream methodsFor: 'character writing' stamp: 'ul 11/2/2010 21:50'!
crlf
	"Append a carriage return and a line feed to the receiver."

	self nextPut: Character cr; nextPut: Character lf! !

!WriteStream methodsFor: 'character writing' stamp: ''!
crtab
	"Append a return character, followed by a single tab character, to the 
	receiver."

	self nextPut: Character cr.
	self nextPut: Character tab! !

!WriteStream methodsFor: 'character writing' stamp: ''!
crtab: anInteger 
	"Append a return character, followed by anInteger tab characters, to the 
	receiver."

	self nextPut: Character cr.
	anInteger timesRepeat: [self nextPut: Character tab]! !

!WriteStream methodsFor: 'character writing' stamp: 'dc 2/11/2007 14:18'!
ensureASpace
	"Append a space character to the receiver IFF there is not one on the end."

	self ensureEndsWith: Character space! !

!WriteStream methodsFor: 'character writing' stamp: 'eem 10/21/2010 10:22'!
ensureCr
	"Append a cr character to the receiver IFF there is not one on the end."

	self ensureEndsWith: Character cr! !

!WriteStream methodsFor: 'accessing' stamp: 'eem 7/28/2014 09:50'!
ensureEndsWith: anObject
	"Append anObject to the receiver IFF there is not one on the end."

	self peekLast = anObject ifFalse:
		[self nextPut: anObject]! !

!WriteStream methodsFor: 'character writing' stamp: 'tk 9/23/2001 01:16'!
ensureNoSpace
	"If there is not one on the end, remove it."

	(position > 0 and: [(collection at: position) = Character space]) 
		ifTrue: [self skip: -1].! !

!WriteStream methodsFor: 'accessing' stamp: 'mir 8/10/1999 11:56'!
flush! !

!WriteStream methodsFor: 'private' stamp: 'BG 5/24/2003 22:49'!
growTo: anInteger

   " anInteger is the required minimal new size of the collection "
	| oldSize grownCollection newSize |
	oldSize := collection size.
     newSize := anInteger + (oldSize // 4 max: 20).
	grownCollection := collection class new: newSize.
	collection := grownCollection replaceFrom: 1 to: oldSize with: collection startingAt: 1.
	writeLimit := collection size.
! !

!WriteStream methodsFor: 'character writing' stamp: 'ul 12/2/2009 05:47'!
lf
	"Append a line feed character to the receiver."

	self nextPut: Character lf! !

!WriteStream methodsFor: 'accessing' stamp: ''!
next

	self shouldNotImplement! !

!WriteStream methodsFor: 'fileIn/Out' stamp: 'yo 8/28/2002 16:13'!
nextChunkPutWithStyle: aStringOrText
	"Append the argument, aText, to the receiver, doubling embedded terminators.  Put out one chunk for the string and one for the style runs.  Mark the style with ]style[."

	aStringOrText isString ifTrue: [^ self nextChunkPut: aStringOrText].
	aStringOrText runs coalesce.
	aStringOrText unembellished ifTrue: [^ self nextChunkPut: aStringOrText asString].

	self nextChunkPut: aStringOrText asString.
	self cr; nextPutAll: ']style['.
	self nextChunkPut: 
		(String streamContents: [:strm | 
			aStringOrText runs writeScanOn: strm]).
! !

!WriteStream methodsFor: 'fileIn/Out' stamp: 'yo 8/13/2003 12:18'!
nextChunkPut: aString
	"Append the argument, aString, to the receiver, doubling embedded terminators."

	| i remainder terminator |
	terminator := $!!.
	remainder := aString.
	[(i := remainder indexOf: terminator) = 0] whileFalse:
		[self nextPutAll: (remainder copyFrom: 1 to: i).
		self nextPut: terminator.  "double imbedded terminators"
		remainder := remainder copyFrom: i+1 to: remainder size].
	self nextPutAll: remainder.
	aString includesUnifiedCharacter ifTrue: [
		self nextPut: terminator.
		self nextPutAll: ']lang['.
		aString writeLeadingCharRunsOn: self.
	].
	self nextPut: terminator.
! !

!WriteStream methodsFor: 'accessing' stamp: 'ul 3/19/2016 13:32'!
nextPutAll: aCollection

	| newEnd |
	(collection class == aCollection class
		or: [ collection isString 
			and: [ aCollection isString
			and: [ collection class format = aCollection class format ] ] ]) "Let Strings with the same field size as collection take the quick route too."
		ifFalse: [ ^ super nextPutAll: aCollection ].

	newEnd := position + aCollection size.
	newEnd > writeLimit ifTrue:
		[self growTo: newEnd + 10].

	collection replaceFrom: position+1 to: newEnd  with: aCollection startingAt: 1.
	position := newEnd.
	^aCollection! !

!WriteStream methodsFor: 'accessing' stamp: 'ar 4/12/2005 17:34'!
nextPut: anObject 
	"Primitive. Insert the argument at the next position in the Stream
	represented by the receiver. Fail if the collection of this stream is not an
	Array or a String. Fail if the stream is positioned at its end, or if the
	position is out of bounds in the collection. Fail if the argument is not
	of the right type for the collection. Optional. See Object documentation
	whatIsAPrimitive."

	<primitive: 66>
	((collection class == ByteString) and: [
		anObject isCharacter and:[anObject isOctetCharacter not]]) ifTrue: [
			collection _ (WideString from: collection).
			^self nextPut: anObject.
	].
	position >= writeLimit
		ifTrue: [^ self pastEndPut: anObject]
		ifFalse: 
			[position _ position + 1.
			^collection at: position put: anObject]! !

!WriteStream methodsFor: 'accessing' stamp: 'ul 3/19/2016 13:33'!
next: anInteger putAll: aCollection startingAt: startIndex
	"Store the next anInteger elements from the given collection."

	| newEnd |
	(collection class == aCollection class
		or: [ collection isString 
			and: [ aCollection isString
			and: [ collection class format = aCollection class format ] ] ]) "Let Strings with the same field size as collection take the quick route too."
		ifFalse: [ ^super next: anInteger putAll: aCollection startingAt: startIndex ].

	newEnd := position + anInteger.
	newEnd > writeLimit ifTrue:
		[self growTo: newEnd + 10].

	collection replaceFrom: position+1 to: newEnd  with: aCollection startingAt: startIndex.
	position := newEnd.

	^aCollection! !

!WriteStream methodsFor: 'private' stamp: ''!
on: aCollection

	super on: aCollection.
	readLimit := 0.
	writeLimit := aCollection size! !

!WriteStream methodsFor: 'private' stamp: 'eem 5/2/2014 09:21'!
on: aCollection from: firstIndex to: lastIndex

	| len |
	collection := aCollection.
	readLimit := 
		writeLimit := lastIndex > (len := collection size)
						ifTrue: [len]
						ifFalse: [lastIndex].
	position := firstIndex <= 1
				ifTrue: [0]
				ifFalse: [firstIndex - 1].
	initialPositionOrNil := position + 1! !

!WriteStream methodsFor: 'private' stamp: 'nice 2/26/2009 11:26'!
pastEndPut: anObject
	"Grow the collection by doubling the size, but keeping the growth between 20 and 1000000.
	Then put <anObject> at the current write position."

	collection := collection grownBy: ((collection size max: 20) min: 1000000).
	writeLimit := collection size.
	collection at: (position := position + 1) put: anObject.
	^ anObject! !

!WriteStream methodsFor: 'character writing' stamp: 'tk 10/19/2001 11:12'!
peekLast
	"Return that item just put at the end of the stream"

	^ position > 0 
		ifTrue: [collection at: position]
		ifFalse: [nil]! !

!WriteStream methodsFor: 'positioning' stamp: ''!
position: anInteger 
	"Refer to the comment in PositionableStream|position:."

	readLimit := readLimit max: position.
	super position: anInteger! !

!WriteStream methodsFor: 'converting' stamp: 'nice 7/24/2014 22:27'!
readStream
	"Answer a readStream on my contents truncated to current position.
	Beware, the readStream shares the contents, so it will be modified if I'm written backward."
	readLimit := readLimit max: position.
	^ReadStream on: collection from: (initialPositionOrNil ifNil: [1]) to: position! !

!WriteStream methodsFor: 'positioning' stamp: ''!
reset 
	"Refer to the comment in PositionableStream|reset."

	readLimit := readLimit max: position.
	position := 0! !

!WriteStream methodsFor: 'positioning' stamp: 'ar 11/12/1998 21:27'!
resetToStart
	readLimit := position := 0.! !

!WriteStream methodsFor: 'positioning' stamp: 'ajh 5/25/2001 20:19'!
setToEnd 
	"Refer to the comment in PositionableStream|setToEnd."

	readLimit := readLimit max: position.
	super setToEnd.! !

!WriteStream methodsFor: 'accessing' stamp: ''!
size

	^readLimit := readLimit max: position! !

!WriteStream methodsFor: 'character writing' stamp: ''!
space
	"Append a space character to the receiver."

	self nextPut: Character space! !

!WriteStream methodsFor: 'character writing' stamp: 'JF 7/31/2003 13:01'!
space: anInteger 
	"Append anInteger space characters to the receiver."

	anInteger timesRepeat: [self space]! !

!WriteStream methodsFor: 'printing' stamp: ''!
store: anObject 
	"Have anObject print on the receiver for purposes of rereading."

	anObject storeOn: self! !

!WriteStream methodsFor: 'character writing' stamp: ''!
tab
	"Append a tab character to the receiver."

	self nextPut: Character tab! !

!WriteStream methodsFor: 'character writing' stamp: 'JF 7/31/2003 13:00'!
tab: anInteger 
	"Append anInteger tab characters to the receiver."

	anInteger timesRepeat: [self tab]! !

!WriteStream methodsFor: 'fileIn/Out' stamp: 'fbs 7/25/2013 07:21'!
timeStamp
	"Append the current time to the receiver as a String."
	self nextChunkPut:	"double string quotes and !!s"
		(String streamContents: [:s | Smalltalk timeStamp: s]) printString.
	self cr! !

!WriteStream methodsFor: 'private' stamp: 'djp 11/6/1999 23:15'!
withAttributes: attributes do: strmBlock 
	"No-op here is overriden in TextStream for font emphasis"
	^ strmBlock value! !

!WriteStream methodsFor: 'private' stamp: ''!
withAttribute: att do: strmBlock 
	"No-op here is overriden in TextStream for font emphasis"
	^ strmBlock value! !

!WriteStream methodsFor: 'private' stamp: ''!
with: aCollection

	super on: aCollection.
	position := readLimit := writeLimit := aCollection size! !

!WriteStream methodsFor: 'printing' stamp: 'kph 11/2/2008 01:16'!
<< aCollection
	"we want a readable version of nextPutAll however it may be difficult to fully recreate nextPutAll:
	for all the different types of stream. Rather then simply send to nextPutAll:
	we handle the String (or ByteArray) argument
	as fast as possible - the rest we delegate to putOn: This means that we handle single characters and bytes 
	whereas nextPutAll: is only for sequencable collections.
	.
	Note this may not work in every case that nextPutAll: does subject to extensive testing, 
	but it should work in the important cases"

	| newEnd |
	collection class == aCollection class ifFalse:
		[ aCollection putOn: self. ^ self ].

	newEnd := position + aCollection size.
	newEnd > writeLimit ifTrue:
		[self growTo: newEnd + 10].

	collection replaceFrom: position+1 to: newEnd  with: aCollection startingAt: 1.
	position := newEnd.

! !

Stream subclass: #SharedQueue2
	instanceVariableNames: 'monitor items'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Sequenceable'!

!SharedQueue2 commentStamp: 'ls 6/25/2005 13:48' prior: 0!
An implementation of a shared queue based on class Monitor.  Clients may may place items on the queue using nextPut: or remove them using methods like next or nextOrNil.  Items are removed in first-in first-out (FIFO) order.  It is safe for multiple threads to access the same shared queue, which is why this is a "shared" queue.

[monitor] is used to synchronize access from multiple threads.

[items] is an ordered collection holding the items that are in the queue.  New items are added  at the end, and old items are removed from the beginning.

All methods must hold the monitor while they run.
!

!SharedQueue2 class methodsFor: 'instance creation' stamp: 'ls 6/25/2005 13:53'!
new
	^self basicNew initialize! !

!SharedQueue2 methodsFor: 'initializing' stamp: 'ls 6/25/2005 13:48'!
initialize
	monitor := Monitor new.
	items := OrderedCollection new.
! !

!SharedQueue2 methodsFor: 'size' stamp: 'ls 6/25/2005 14:03'!
isEmpty
	^monitor critical: [ items isEmpty ]! !

!SharedQueue2 methodsFor: 'accessing' stamp: 'ul 10/15/2010 08:24'!
next

	^monitor critical: [
		monitor waitWhile: [ items isEmpty ].
		items removeFirst ]
! !

!SharedQueue2 methodsFor: 'accessing' stamp: 'ls 6/25/2005 14:00'!
nextOrNil
	^monitor critical: [
		items isEmpty ifTrue: [ nil ] ifFalse: [ items removeFirst ] ]! !

!SharedQueue2 methodsFor: 'accessing' stamp: 'nice 12/26/2009 21:34'!
nextOrNilSuchThat: aBlock
	"Answer the next object that satisfies aBlock, skipping any intermediate objects.
	If no such object has been queued, answer <nil> and leave me intact."

	^monitor critical: [
		| index |
		index := items findFirst: aBlock.
		index = 0 ifTrue: [
			nil ]
		ifFalse: [
			items removeAt: index ] ].
! !

!SharedQueue2 methodsFor: 'accessing' stamp: 'ul 10/15/2010 08:23'!
nextPut: item

	monitor critical: [
		items addLast: item.
		monitor signal.  ].
	^item! !

!SharedQueue2 methodsFor: 'accessing' stamp: 'ls 6/25/2005 14:00'!
peek
	"Answer the object that was sent through the receiver first and has not 
	yet been received by anyone but do not remove it from the receiver. If 
	no object has been sent, return nil"
	^monitor critical: [
		items isEmpty ifTrue: [ nil ] ifFalse: [ items first ] ]
! !

!SharedQueue2 methodsFor: 'copying' stamp: 'nice 10/5/2009 19:07'!
postCopy
	super postCopy.
	monitor critical:
		[items := items copy.
		monitor := Monitor new]! !

!SharedQueue2 methodsFor: 'printing' stamp: 'ls 6/25/2005 13:56'!
printOn: aStream
	monitor critical: [
		aStream 
			nextPutAll: self class name;
			nextPutAll: ' with ';
			print: items size;
		 	nextPutAll: ' items' ].! !

!SharedQueue2 methodsFor: 'accessing' stamp: 'ul 10/15/2010 08:23'!
removeAll
	monitor critical: [
		items removeAll ].! !

!SharedQueue2 methodsFor: 'accessing' stamp: 'ls 6/25/2005 14:01'!
removeAllSuchThat: aBlock
	"Remove from the queue all objects that satisfy aBlock."
	monitor critical: [
		items removeAllSuchThat: aBlock ]! !

!SharedQueue2 methodsFor: 'size' stamp: 'ls 6/25/2005 14:04'!
size
	^monitor critical: [ items size ]! !

!Stream class methodsFor: 'instance creation' stamp: ''!
new

	self error: 'Streams are created with on: and with:'! !

!Stream methodsFor: 'testing' stamp: ''!
atEnd
	"Answer whether the receiver can access any more objects."

	self subclassResponsibility! !

!Stream methodsFor: 'accessing - multibyte support' stamp: 'yo 8/30/2002 17:13'!
basicNext

	^ self next.
! !

!Stream methodsFor: 'accessing - multibyte support' stamp: 'yo 8/30/2002 17:13'!
basicNextPutAll: aCollection 

	^ self nextPutAll: aCollection.
! !

!Stream methodsFor: 'accessing - multibyte support' stamp: 'yo 8/30/2002 17:13'!
basicNextPut: anObject 

	^ self nextPut: anObject! !

!Stream methodsFor: 'accessing - multibyte support' stamp: 'ul 4/11/2011 02:58'!
basicNext: anInteger 

	^self next: anInteger
! !

!Stream methodsFor: 'accessing - multibyte support' stamp: 'ul 1/31/2011 01:37'!
basicNext: anInteger putAll: aCollection startingAt: startIndex

	^self next: anInteger putAll: aCollection startingAt: startIndex
! !

!Stream methodsFor: 'accessing' stamp: 'nk 2/24/2001 17:31'!
binary! !

!Stream methodsFor: 'file open/close' stamp: 'mir 8/10/1999 12:04'!
close! !

!Stream methodsFor: 'testing' stamp: 'ab 8/28/2003 18:30'!
closed
	^ false! !

!Stream methodsFor: 'enumerating' stamp: 'mt 4/12/2016 10:00'!
collect: block

	^ Generator on: [:g | 
		[self atEnd] whileFalse: [
			g yield: (self next
				ifNil: [nil]
				ifNotNil: [:object | block value: object])]]! !

!Stream methodsFor: 'accessing' stamp: ''!
contents
	"Answer all of the contents of the receiver."

	self subclassResponsibility! !

!Stream methodsFor: 'enumerating' stamp: ''!
do: aBlock 
	"Evaluate aBlock for each of the objects accessible by receiver."

	[self atEnd]
		whileFalse: [aBlock value: self next]! !

!Stream methodsFor: 'converting' stamp: 'mt 4/12/2016 10:04'!
flatten

	^ Generator on: [:g | 
		[self atEnd] whileFalse: [
			self next in: [:object |
				((object isCollection and: [object isString not]) or: [object isStream])
					ifFalse: [g yield: object]
					ifTrue: [object flatten do: [:each | g yield: each]]]]]! !

!Stream methodsFor: 'converting' stamp: 'mt 4/12/2016 09:43'!
flattened
	"An alias for #flatten
	This message's name is in line with messages like #sorted or #reversed
	while #flatten's is in line with #reverse (as per ANSI, see comment there)"

	^ self flatten! !

!Stream methodsFor: 'accessing' stamp: 'sma 4/22/2000 17:07'!
flush
	"Do nothing by default"! !

!Stream methodsFor: 'enumerating' stamp: 'mt 4/12/2016 10:00'!
gather: block

	^ Generator on: [:g | 
		[self atEnd] whileFalse: [
			self next
				ifNil: [g yield: nil]
				ifNotNil: [:object |
					(block value: object) do: [:ea |
						g yield: ea]]]]! !

!Stream methodsFor: 'testing' stamp: 'ar 12/23/1999 15:43'!
isStream
	"Return true if the receiver responds to the stream protocol"
	^true! !

!Stream methodsFor: 'testing' stamp: 'mir 11/10/2003 18:22'!
isTypeHTTP
	^false! !

!Stream methodsFor: 'accessing' stamp: 'nk 4/29/2004 10:38'!
localName
	^'a stream'! !

!Stream methodsFor: 'accessing' stamp: ''!
next
	"Answer the next object accessible by the receiver."

	self subclassResponsibility! !

!Stream methodsFor: 'accessing' stamp: 'ar 12/23/1999 14:53'!
nextInto: aCollection
	"Read the next elements of the receiver into aCollection.
	Return aCollection or a partial copy if less than aCollection
	size elements have been read."
	^self next: aCollection size into: aCollection startingAt: 1.! !

!Stream methodsFor: 'accessing' stamp: 'ar 12/23/1999 14:59'!
nextInto: aCollection startingAt: startIndex
	"Read the next elements of the receiver into aCollection.
	Return aCollection or a partial copy if less than aCollection
	size elements have been read."
	^self next: (aCollection size - startIndex+1) into: aCollection startingAt: startIndex.! !

!Stream methodsFor: 'accessing' stamp: ''!
nextMatchAll: aColl
    "Answer true if next N objects are the ones in aColl,
     else false.  Advance stream of true, leave as was if false."
    | save |
    save := self position.
    aColl do: [:each |
       (self next) = each ifFalse: [
            self position: save.
            ^ false]
        ].
    ^ true! !

!Stream methodsFor: 'accessing' stamp: ''!
nextMatchFor: anObject 
	"Gobble the next object and answer whether it is equal to the argument, 
	anObject."

	^anObject = self next! !

!Stream methodsFor: 'accessing' stamp: ''!
nextPutAll: aCollection 
	"Append the elements of aCollection to the sequence of objects accessible 
	by the receiver. Answer aCollection."

	aCollection do: [:v | self nextPut: v].
	^aCollection! !

!Stream methodsFor: 'accessing' stamp: ''!
nextPut: anObject 
	"Insert the argument, anObject, as the next object accessible by the 
	receiver. Answer anObject."

	self subclassResponsibility! !

!Stream methodsFor: 'accessing' stamp: 'mt 4/12/2016 09:33'!
nextSatisfy: aBlock

	self do: [:each | (aBlock value: each) ifTrue: [^ each]].
	Error signal: 'No object could satisfy the block'.! !

!Stream methodsFor: 'testing' stamp: 'ar 5/17/2001 19:07'!
nextWordsPutAll: aCollection
	"Write the argument a word-like object in big endian format on the receiver.
	May be used to write other than plain word-like objects (such as ColorArray)."
	aCollection class isPointers | aCollection class isWords not 
		ifTrue: [^self error: aCollection class name,' is not word-like'].
	1 to: aCollection basicSize do:[:i|
		self nextNumber: 4 put: (aCollection basicAt: i).
	].
	^aCollection! !

!Stream methodsFor: 'accessing' stamp: ''!
next: anInteger 
	"Answer the next anInteger number of objects accessible by the receiver."

	| aCollection |
	aCollection := OrderedCollection new.
	anInteger timesRepeat: [aCollection addLast: self next].
	^aCollection! !

!Stream methodsFor: 'accessing' stamp: 'ar 12/23/1999 14:58'!
next: n into: aCollection
	"Read n objects into the given collection.
	Return aCollection or a partial copy if less than
	n elements have been read."
	^self next: n into: aCollection startingAt: 1! !

!Stream methodsFor: 'accessing' stamp: 'nice 3/16/2010 23:12'!
next: n into: aCollection startingAt: startIndex
	"Read n objects into the given collection. 
	Return aCollection or a partial copy if less than n elements have been read."
	
	| count |
	count := self readInto: aCollection startingAt: startIndex count: n.
	count = n
		ifTrue:[ ^aCollection ]
		ifFalse:[ ^aCollection copyFrom: 1 to: startIndex + count - 1 ]! !

!Stream methodsFor: 'accessing' stamp: ''!
next: anInteger put: anObject 
	"Make anObject be the next anInteger number of objects accessible by the 
	receiver. Answer anObject."

	anInteger timesRepeat: [self nextPut: anObject].
	^anObject! !

!Stream methodsFor: 'accessing' stamp: 'nk 4/29/2004 10:40'!
openReadOnly
	^self! !

!Stream methodsFor: 'printing' stamp: 'djp 7/21/1998 17:13'!
printHtml: anObject
	anObject printHtmlOn: self! !

!Stream methodsFor: 'printing' stamp: 'bf 4/9/2013 13:34'!
printOn: stream

	super printOn: stream.
"
	stream space.
	self contents printOn: stream.
	
	NOTE: some subclasses actually read from the stream to return its contents.
	To not modify the state we must *not* send contents here. 
"! !

!Stream methodsFor: 'printing' stamp: 'sma 6/1/2000 09:56'!
print: anObject
	"Have anObject print itself on the receiver."

	anObject printOn: self! !

!Stream methodsFor: 'printing' stamp: 'bf 5/16/2012 18:38'!
print: anObject  maxDecimalPlaces: placesDesired
	"Have anObject print itself on the receiver, with at most the given number of decimal places desired."

	anObject printOn: self maxDecimalPlaces: placesDesired! !

!Stream methodsFor: 'accessing' stamp: 'nice 3/16/2010 23:05'!
readInto: aCollection startingAt: startIndex count: n
	"Read n objects into the given collection. 
	Return number of elements that have been read."
	| obj |
	0 to: n - 1 do: [:i |
		(obj := self next) == nil ifTrue: [^i].
		aCollection at: startIndex + i put: obj].
	^n! !

!Stream methodsFor: 'accessing' stamp: 'nk 4/29/2004 10:41'!
readOnly
	^self! !

!Stream methodsFor: 'enumerating' stamp: 'mt 4/12/2016 09:39'!
reject: aBlock

	^ self select: [:element | (aBlock value: element) == false]! !

!Stream methodsFor: 'enumerating' stamp: 'mt 4/12/2016 10:07'!
select: block

	^ Generator on: [:g |
		[self atEnd] whileFalse: [
			self next
				ifNil: [g yield: nil]
				ifNotNil: [:object |
					(block value: object)
						ifTrue: [g yield: object]]]]! !

!Stream methodsFor: 'enumerating' stamp: 'mt 4/11/2016 17:49'!
select: block thenCollect: anotherBlock

	^ (self select: block) collect: anotherBlock! !

!Stream methodsFor: 'as yet unclassified' stamp: 'RAA 9/11/2000 19:12'!
sleep

	"an FTP-based stream might close the connection here"! !

!Stream methodsFor: 'accessing' stamp: 'mt 4/12/2016 10:19'!
upToEnd
	"Answer the remaining elements in the stream."
	
	| elements |
	elements := OrderedCollection new.
	[self atEnd] whileFalse: [ 
		elements add: self next].
	^ elements! !

!Stream methodsFor: 'filter streaming' stamp: 'MPW 1/1/1901 00:48'!
write:encodedObject
	^encodedObject putOn:self.
! !

!Stream methodsFor: 'readability' stamp: 'kph 9/27/2007 21:53'!
<< items

 	items putOn: self.
	
	^ self! !

Object subclass: #TextAttribute
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!TextAttribute commentStamp: 'tk 7/22/2002 18:33' prior: 0!
Tells a piece of text to be a certain way.

Select text, press Command-6, choose a attribute.  If selected text is of the form 
	Hi There<Smalltalk beep>
the part in angle brackets is saved for action, and the Hi There appears in the paragraph.  If selection has no angle brackets, use the whole thing as both the text and the action.

TextDoIt  --  eval as a Smalltalk expression (the part in angle brackets)

TextLink -- Show a method, class comment, class hierarchy, or class defintion.
	<Point extent:>, <Point Comment>, <Point Hierarchy>, or <Point Defintion> are what you type.

TextURL -- Show the web page. <www.disney.com>

These attributes of text need to be stored on the disk in a regular file-out.  It is done in this form: 	Hi There   
	in the text, and a Run containing   dSmalltalk beep;;
	Click here to see the extent:   
	in the text, and a Run containing   method LPoint extent:;
See RunArray class scanFrom: where decoding is done.
!

TextAttribute subclass: #TextAction
	instanceVariableNames: ''
	classVariableNames: 'Purple'
	poolDictionaries: ''
	category: 'Collections-Text'!

TextAction subclass: #PluggableTextAttribute
	instanceVariableNames: 'evalBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!PluggableTextAttribute commentStamp: '<historical>' prior: 0!
An attribute which evaluates an arbitrary block when it is selected.!

!PluggableTextAttribute class methodsFor: 'instance creation' stamp: 'ls 6/21/2001 18:09'!
evalBlock: aBlock
	^super new evalBlock: aBlock! !

!PluggableTextAttribute methodsFor: 'event handling' stamp: 'ls 6/21/2001 18:13'!
actOnClickFor: anObject
	evalBlock ifNil: [ ^self ].
	evalBlock numArgs = 0 ifTrue: [ evalBlock value.  ^true ].
	evalBlock numArgs = 1 ifTrue: [ evalBlock value: anObject.  ^true ].
	self error: 'evalBlock should have 0 or 1 arguments'! !

!PluggableTextAttribute methodsFor: 'initialization' stamp: 'ls 6/21/2001 18:06'!
evalBlock: aBlock
	evalBlock := aBlock! !

!PluggableTextAttribute methodsFor: 'as yet unclassified' stamp: 'topa 7/8/2015 15:47'!
writeScanOn: aStream
	"Impossible for this kind of attribute"
	^ self shouldNotImplement
	! !

!TextAction class methodsFor: 'preferences' stamp: 'mt 8/11/2016 16:37'!
applyUserInterfaceTheme

	Purple := nil.! !

!TextAction class methodsFor: 'class initialization' stamp: 'di 1/14/98 09:30'!
initialize   "TextAction initialize"
	Purple := Color r: 0.4 g: 0 b: 1.0! !

!TextAction class methodsFor: 'preferences' stamp: 'mt 8/11/2016 16:33'!
themeProperties

	^ super themeProperties, {
		{ #color. 'Colors'. 'Color for clickable text links.' }
		} ! !

!TextAction methodsFor: 'as yet unclassified' stamp: 'nk 2/26/2005 09:49'!
analyze: aString
	"Analyze the selected text to find both the parameter to store and the text to emphesize (may be different from original selection).  Does not return self!!.  May be of the form:
3+4
<3+4>
Click Here<3+4>
<3+4>Click Here
"
	"Obtain the showing text and the instructions"
	| b1 b2 trim param show |
	b1 := aString indexOf: $<.
	b2 := aString indexOf: $>.
	(b1 < b2) & (b1 > 0) ifFalse: ["only one part"
		param := self validate: aString.
		param ifNil: [ ^{ nil. nil } ].
		^ Array with: param with: (param size = 0 ifTrue: [nil] ifFalse: [param])].
	"Two parts"
	trim := aString withBlanksTrimmed.
	(trim at: 1) == $< 
		ifTrue: [(trim last) == $>
			ifTrue: ["only instructions" 
				param := self validate: (aString copyFrom: b1+1 to: b2-1).
				show := param size = 0 ifTrue: [nil] ifFalse: [param]]
			ifFalse: ["at the front"
				param := self validate: (aString copyFrom: b1+1 to: b2-1).
				show := param size = 0 ifTrue: [nil] 
						ifFalse: [aString copyFrom: b2+1 to: aString size]]]
		ifFalse: [(trim last) == $>
			ifTrue: ["at the end"
				param := self validate: (aString copyFrom: b1+1 to: b2-1).
				show := param size = 0 ifTrue: [nil] 
						ifFalse: [aString copyFrom: 1 to: b1-1]]
			ifFalse: ["Illegal -- <> has text on both sides"
				show := nil]].
	^ Array with: param with: show
! !

!TextAction methodsFor: 'updating' stamp: 'mt 8/11/2016 16:34'!
applyUserInterfaceTheme

	"Ignore. Only class-side cache."! !

!TextAction methodsFor: 'updating' stamp: 'mt 8/16/2016 14:55'!
canApplyUserInterfaceTheme

	^ false! !

!TextAction methodsFor: 'testing' stamp: 'sw 11/9/1999 17:21'!
couldDeriveFromPrettyPrinting
	^ false! !

!TextAction methodsFor: 'testing' stamp: 'di 10/31/97 13:11'!
dominatedByCmd0
	"Cmd-0 should turn off active text"
	^ true! !

!TextAction methodsFor: 'as yet unclassified' stamp: 'mt 8/11/2016 16:36'!
emphasizeScanner: scanner
	"Set the emphasis for text display"
	Purple ifNil: [Purple := self userInterfaceTheme color ifNil: [Color r: 0.4 g: 0.0 b: 1]].
	scanner textColor: Purple.! !

!TextAction methodsFor: 'as yet unclassified' stamp: 'DSM 3/30/1999 13:15'!
info
	^ 'no hidden info'! !

!TextAction methodsFor: 'testing' stamp: ''!
mayActOnClick

	^ true! !

!TextAction methodsFor: 'as yet unclassified' stamp: 'tk 12/16/97 16:48'!
validate: aString
	"any format is OK with me"
	^ aString! !

TextAction subclass: #TextDoIt
	instanceVariableNames: 'evalString'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!TextDoIt class methodsFor: 'instance creation' stamp: 'tk 12/6/97 20:28'!
evalString: str
	^ self new evalString: str! !

!TextDoIt class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 11:56'!
scanCharacter
	"The character used to identify a subclass of TextAttribute for filein and fileout"
	^$d! !

!TextDoIt class methodsFor: 'fileIn/Out' stamp: 'tk 12/16/97 09:06'!
scanFrom: strm
	"read a doit in the funny format used by Text styles on files. d10 factorial;;  end with two semicolons"

	| pos end doit |
	pos := strm position.
	[strm skipTo: $;. strm peek == $;] whileFalse.
	end := strm position - 1.
	strm position: pos.
	doit := strm next: end-pos.
	strm skip: 2.  ";;"
	^ self evalString: doit! !

!TextDoIt methodsFor: 'as yet unclassified' stamp: 'nice 12/16/2013 15:47'!
actOnClickFor: anObject
	"Note: evalString gets evaluated IN THE CONTEXT OF anObject
	 -- meaning that self and all instVars are accessible"
	Compiler evaluate: evalString for: anObject.
	^ true ! !

!TextDoIt methodsFor: 'as yet unclassified' stamp: 'tk 12/16/97 16:46'!
analyze: aString

	| list |
	list := super analyze: aString.
	evalString := list at: 1.
	^ list at: 2! !

!TextDoIt methodsFor: 'accessing' stamp: 'dtl 1/1/2014 20:59'!
evalString
	^evalString! !

!TextDoIt methodsFor: 'accessing' stamp: 'tk 12/5/97 17:01'!
evalString: str
	evalString := str ! !

!TextDoIt methodsFor: 'comparing' stamp: 'dtl 1/1/2014 21:06'!
hash
	^evalString hash! !

!TextDoIt methodsFor: 'as yet unclassified' stamp: 'tk 12/30/97 10:33'!
info
	^ evalString! !

!TextDoIt methodsFor: 'fileIn/fileOut' stamp: 'dtl 1/2/2014 11:57'!
writeScanOn: strm

	strm nextPut: self class scanCharacter; nextPutAll: evalString; nextPutAll: ';;'! !

!TextDoIt methodsFor: 'comparing' stamp: 'dtl 1/1/2014 21:03'!
= textAttribute
	^ textAttribute class == self class
		and: [textAttribute evalString = evalString]! !

TextDoIt subclass: #TextPrintIt
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!TextPrintIt class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 11:58'!
scanCharacter
	"The character used to identify a subclass of TextAttribute for filein and fileout"
	^$P! !

!TextPrintIt methodsFor: 'mouse events' stamp: 'nice 12/16/2013 15:47'!
actOnClickFor: anObject in: aParagraph at: clickPoint editor: editor
	"Note: evalString gets evaluated IN THE CONTEXT OF anObject
	 -- meaning that self and all instVars are accessible"
	| result range index |
	result := Compiler evaluate: evalString for: anObject.
	result := ' ', result printString,' '.
	"figure out where the attribute ends in aParagraph"
	index := (aParagraph characterBlockAtPoint: clickPoint) stringIndex.
	range := aParagraph text rangeOf: self startingAt: index.
	editor selectFrom: range last+1 to: range last.
	editor zapSelectionWith: result.
	editor selectFrom: range last to: range last + result size.
	^ true ! !

TextAction subclass: #TextLink
	instanceVariableNames: 'classAndMethod'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!TextLink class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 12:13'!
scanCharacter
	"The character used to identify a subclass of TextAttribute for filein and fileout"
	^$L! !

!TextLink class methodsFor: 'fileIn/Out' stamp: 'tk 12/16/97 08:53'!
scanFrom: strm
	"read a link in the funny format used by Text styles on files. LPoint +;LPoint Comment;"

	^ self new classAndMethod: (strm upTo: $;)! !

!TextLink methodsFor: 'event handling' stamp: 'tk 5/1/2001 18:17'!
actOnClickFor: aMessageSet
	"Add to the end of the list.  'aClass selector', 'aClass Comment', 'aClass Definition', 'aClass Hierarchy' are the formats allowed."

	aMessageSet addItem: classAndMethod.
	^ true! !

!TextLink methodsFor: 'initialize-release' stamp: 'tk 12/16/97 16:49'!
analyze: aString

	| list |
	list := super analyze: aString.
	classAndMethod := list at: 1.
	^ list at: 2! !

!TextLink methodsFor: 'initialize-release' stamp: 'LC 10/8/2001 10:53'!
analyze: aString with: nonMethod
	"Initalize this attribute holder with a piece text the user typed into a paragraph.  Returns the text to emphesize (may be different from selection)  Does not return self!!.  nonMethod is what to show when clicked, i.e. the last part of specifier (Comment, Definition, or Hierarchy).  May be of the form:
Point
<Point>
Click Here<Point>
<Point>Click Here
"
	"Obtain the showing text and the instructions"
	| b1 b2 trim |
	b1 := aString indexOf: $<.
	b2 := aString indexOf: $>.
	(b1 < b2) & (b1 > 0) ifFalse: ["only one part"
		classAndMethod := self validate: aString, ' ', nonMethod.
		^ classAndMethod ifNotNil: [aString]].
	"Two parts"
	trim := aString withBlanksTrimmed.
	(trim at: 1) == $< 
		ifTrue: [(trim last) == $>
			ifTrue: ["only instructions" 
				classAndMethod := self validate: (aString copyFrom: b1+1 to: b2-1), ' ', nonMethod.
				^ classAndMethod ifNotNil: [classAndMethod]]
			ifFalse: ["at the front"
				classAndMethod := self validate: (aString copyFrom: b1+1 to: b2-1), ' ', nonMethod.
				^ classAndMethod ifNotNil: [aString copyFrom: b2+1 to: aString size]]]
		ifFalse: [(trim last) == $>
			ifTrue: ["at the end"
				classAndMethod := self validate: (aString copyFrom: b1+1 to: b2-1), ' ', nonMethod.
				^ classAndMethod ifNotNil: [aString copyFrom: 1 to: b1-1]]
			ifFalse: ["Illegal -- <> has text on both sides"
				^ nil]]
! !

!TextLink methodsFor: 'accessing' stamp: 'tk 12/5/97 17:09'!
classAndMethod: aString
	classAndMethod := aString! !

!TextLink methodsFor: 'comparing' stamp: 'dtl 1/1/2014 23:06'!
hash
	"#hash is re-implemented because #= is re-implemented"
	^classAndMethod hash! !

!TextLink methodsFor: 'accessing' stamp: 'tk 12/30/97 10:33'!
info
	^ classAndMethod! !

!TextLink methodsFor: 'initialize-release' stamp: 'nk 2/26/2005 09:50'!
validate: specString
	"Can this string be decoded to be Class space Method (or Comment, Definition, Hierarchy)? If so, return it in valid format, else nil" 

	| list first mid last |
	list := specString findTokens: ' 	.|'.
	list isEmpty ifTrue: [ ^nil ].
	last := list last.
	last first isUppercase ifTrue: [
		(#('Comment' 'Definition' 'Hierarchy') includes: last) ifFalse: [^ nil].
		"Check for 'Rectangle Comment Comment' and remove last one"
		(list at: list size - 1 ifAbsent: [^nil]) = last ifTrue: [list := list allButLast]].
	list size > 3 ifTrue: [^ nil].
	list size < 2 ifTrue: [^ nil].
	Symbol hasInterned: list first ifTrue: [:sym | first := sym].
	first ifNil: [^ nil].
	Smalltalk at: first ifAbsent: [^ nil].
	mid := list size = 3 
		ifTrue: [(list at: 2) = 'class' ifTrue: ['class '] ifFalse: [^ nil]]
		ifFalse: [''].
	"OK if method name is not interned -- may not be defined yet"
	^ first, ' ', mid, last! !

!TextLink methodsFor: 'fileIn/fileOut' stamp: 'dtl 1/2/2014 12:14'!
writeScanOn: strm

	strm nextPut: self class scanCharacter; nextPutAll: classAndMethod; nextPut: $;! !

!TextLink methodsFor: 'comparing' stamp: 'dtl 1/1/2014 23:07'!
= other 
	^ (other class == self class) 
		and: [other info = self info]! !

TextAction subclass: #TextURL
	instanceVariableNames: 'url'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

TextURL subclass: #TextSqkPageLink
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!TextSqkPageLink commentStamp: '<historical>' prior: 0!
I represent a link to either a SqueakPage in a BookMorph, or a regular url.  See TextMorphEditor changeEmphasis:.  
!

!TextSqkPageLink class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 12:15'!
scanCharacter
	"The character used to identify a subclass of TextAttribute for filein and fileout"
	^$q! !

!TextURL class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 12:15'!
scanCharacter
	"The character used to identify a subclass of TextAttribute for filein and fileout"
	^$R! !

!TextURL class methodsFor: 'fileIn/Out' stamp: 'tk 12/16/97 09:24'!
scanFrom: strm
	"read a link in the funny format used by Text styles on files. Rhttp://www.disney.com;"

	^ self new url: (strm upTo: $;)! !

!TextURL methodsFor: 'as yet unclassified' stamp: 'mt 8/12/2016 13:29'!
analyze: aString

	| list |
	list := super analyze: aString.

	(UIManager default request: 'URL to open' translated initialAnswer: (list at: 1))
		in: [:answer | answer ifEmpty: [url := list at: 1] ifNotEmpty: [url := answer]].
	
	^ list at: 2! !

!TextURL methodsFor: 'comparing' stamp: 'dtl 1/1/2014 23:53'!
hash
	"#hash is re-implemented because #= is re-implemented"
	^url hash! !

!TextURL methodsFor: 'accessing' stamp: 'tk 12/30/97 10:33'!
info
	^ url! !

!TextURL methodsFor: 'accessing' stamp: 'tk 12/16/97 08:55'!
url: aString
	url := aString! !

!TextURL methodsFor: 'fileIn/fileOut' stamp: 'dtl 1/2/2014 12:17'!
writeScanOn: strm

	strm nextPut: self class scanCharacter; nextPutAll: url; nextPut: $;! !

!TextURL methodsFor: 'comparing' stamp: 'dtl 1/1/2014 23:52'!
= other 
	^ (other class == self class) 
		and: [other info = self info]! !

!TextAttribute class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 12:06'!
classFor: scanCharacter
	"Answer the class that uses scanCharacter to identify itself in a text fileout" 
	^ self allSubclasses detect: [:cls | cls scanCharacters includes: scanCharacter]
! !

!TextAttribute class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 13:24'!
consumeIdentifierFrom: scanStream.
	"When scanning, a subclass may require access to the identifier character.
	Otherwise, consume it here."
	scanStream next! !

!TextAttribute class methodsFor: 'instance creation' stamp: 'dtl 1/2/2014 13:23'!
newFrom: scanStream
	"scanStream contains a class identifier character possibly followed by data"
	| cls |
	cls := self classFor: scanStream peek.
	cls consumeIdentifierFrom: scanStream.
	^cls scanFrom: scanStream! !

!TextAttribute class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 10:17'!
scanCharacter
	"The character used to identify a subclass of TextAttribute for filein and fileout"
	^nil! !

!TextAttribute class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 12:06'!
scanCharacters
	"All scan characters corresponding to the given class. Usually this is an array of one."
	^Array with: self scanCharacter! !

!TextAttribute methodsFor: 'mouse events' stamp: ''!
actOnClickFor: model
	"Subclasses may override to provide, eg, hot-spot actions"
	^ false! !

!TextAttribute methodsFor: 'mouse events' stamp: 'ar 9/22/2001 16:00'!
actOnClickFor: model in: aParagraph
	^self actOnClickFor: model! !

!TextAttribute methodsFor: 'mouse events' stamp: 'ar 9/22/2001 16:08'!
actOnClickFor: model in: aParagraph at: clickPoint
	^self actOnClickFor: model in: aParagraph! !

!TextAttribute methodsFor: 'mouse events' stamp: 'ar 9/22/2001 16:22'!
actOnClickFor: model in: aParagraph at: clickPoint editor: editor
	^self actOnClickFor: model in: aParagraph at: clickPoint! !

!TextAttribute methodsFor: 'as yet unclassified' stamp: 'ar 12/16/2001 23:18'!
anchoredMorph
	"If one hides here, return it"
	^nil! !

!TextAttribute methodsFor: 'html' stamp: 'dgd 9/2/2005 12:39'!
closeHtmlOn: aStream 
	"put on the given stream the tag to close the html  
	representation of the receiver"
	self subclassResponsibility! !

!TextAttribute methodsFor: 'testing' stamp: 'sw 11/9/1999 17:25'!
couldDeriveFromPrettyPrinting
	"Answer whether the receiver is a kind of attribute that could have been generated by doing polychrome pretty-printing of a method without functional text attributes."

	^ true! !

!TextAttribute methodsFor: 'testing' stamp: 'di 10/31/97 13:09'!
dominatedByCmd0
	"Subclasses may override if cmd-0 should turn them off"
	^ false! !

!TextAttribute methodsFor: 'as yet unclassified' stamp: ''!
dominates: another
	"Subclasses may override condense multiple attributes"
	^ false! !

!TextAttribute methodsFor: 'as yet unclassified' stamp: 'di 4/1/1999 15:16'!
emphasisCode
	"Subclasses may override to add bold, italic, etc"
	^ 0! !

!TextAttribute methodsFor: 'as yet unclassified' stamp: ''!
emphasizeScanner: scanner
	"Subclasses may override to set, eg, font, color, etc"! !

!TextAttribute methodsFor: 'as yet unclassified' stamp: 'di 11/9/97 17:46'!
forFontInStyle: aTextStyle do: aBlock
	"No action is the default.  Overridden by font specs"! !

!TextAttribute methodsFor: 'testing' stamp: 'ar 9/21/2000 14:16'!
isKern
	^false! !

!TextAttribute methodsFor: 'as yet unclassified' stamp: ''!
mayActOnClick
	"Subclasses may override to provide, eg, hot-spot actions"
	^ false! !

!TextAttribute methodsFor: 'as yet unclassified' stamp: 'di 11/10/97 14:05'!
mayBeExtended
	"A quality that may be overridden by subclasses, such as TextAnchors, that really only apply to a single character"
	^ true! !

!TextAttribute methodsFor: 'as yet unclassified' stamp: 'ar 5/5/2004 17:15'!
menu
	^nil! !

!TextAttribute methodsFor: 'as yet unclassified' stamp: ''!
oldEmphasisCode: default
	"Allows running thorugh possibly multiple attributes
	and getting the emphasis out of any that has an emphasis (font number)"
	^ default! !

!TextAttribute methodsFor: 'html' stamp: 'dgd 9/2/2005 12:38'!
openHtmlOn: aStream 
	"put on the given stream the tag to open the html  
	representation of the receiver"
	self subclassResponsibility! !

!TextAttribute methodsFor: 'as yet unclassified' stamp: 'di 10/31/97 11:03'!
reset
	"Allow subclasses to prepare themselves for merging attributes"! !

!TextAttribute methodsFor: 'as yet unclassified' stamp: ''!
set
	"Respond true to include this attribute (as opposed to, eg, a bold
	emphasizer that is clearing the property"
	^ true! !

TextAttribute subclass: #TextColor
	instanceVariableNames: 'color'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!TextColor commentStamp: '<historical>' prior: 0!
A TextColor encodes a text color change applicable over a given range of text.!

!TextColor class methodsFor: 'constants' stamp: ''!
black
	^ self new color: Color black! !

!TextColor class methodsFor: 'constants' stamp: ''!
blue
	^ self new color: Color blue! !

!TextColor class methodsFor: 'instance creation' stamp: ''!
color: aColor
	^ self new color: aColor! !

!TextColor class methodsFor: 'constants' stamp: ''!
cyan
	^ self new color: Color cyan! !

!TextColor class methodsFor: 'constants' stamp: 'ajh 9/10/2002 02:26'!
gray
	^ self new color: Color gray! !

!TextColor class methodsFor: 'constants' stamp: ''!
green
	^ self new color: Color green! !

!TextColor class methodsFor: 'constants' stamp: ''!
magenta
	^ self new color: Color magenta! !

!TextColor class methodsFor: 'constants' stamp: ''!
red
	^ self new color: Color red! !

!TextColor class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 11:55'!
scanCharacter
	"The character used to identify a subclass of TextAttribute for filein and fileout"
	^$c! !

!TextColor class methodsFor: 'fileIn/Out' stamp: 'sma 3/24/2000 10:49'!
scanFrom: strm
	"read a color in the funny format used by Text styles on files. c125000255 or cblue;"

	| r g b |
	strm peek isDigit
		ifTrue:
			[r := (strm next: 3) asNumber.
			g := (strm next: 3) asNumber.
			b := (strm next: 3) asNumber.
			^ self color: (Color r: r g: g b: b range: 255)].
	"A name of a color"
	^ self color: (Color perform: (strm upTo: $;) asSymbol)! !

!TextColor class methodsFor: 'constants' stamp: 'sma 3/24/2000 10:50'!
white 
	^ self new color: Color white! !

!TextColor class methodsFor: 'constants' stamp: ''!
yellow
	^ self new color: Color yellow! !

!TextColor methodsFor: 'html' stamp: 'dgd 9/2/2005 12:39'!
closeHtmlOn: aStream 
	"put on the given stream the tag to close the html  
	representation of the receiver"
	aStream nextPutAll: '</font>'! !

!TextColor methodsFor: 'accessing' stamp: ''!
color
	^ color! !

!TextColor methodsFor: 'accessing' stamp: ''!
color: aColor
	color := aColor! !

!TextColor methodsFor: 'testing' stamp: 'bf 12/6/2012 19:10'!
dominatedByCmd0
	"Cmd-0 should remove text color"
	^ true! !

!TextColor methodsFor: 'scanning' stamp: 'di 10/31/97 11:20'!
dominates: other
	^ other class == self class! !

!TextColor methodsFor: 'scanning' stamp: ''!
emphasizeScanner: scanner
	"Set the emphasis for text display"
	scanner textColor: color! !

!TextColor methodsFor: 'comparing' stamp: 'sma 3/24/2000 10:51'!
hash
	^ color hash! !

!TextColor methodsFor: 'html' stamp: 'dgd 9/2/2005 12:38'!
openHtmlOn: aStream 
	"put on the given stream the tag to open the html  
	representation of the receiver"
	aStream nextPutAll: '<font color="#' , color printHtmlString , '">'! !

!TextColor methodsFor: 'printing' stamp: 'sma 3/24/2000 10:51'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' code: '; print: color! !

!TextColor methodsFor: 'fileIn/fileOut' stamp: 'dtl 1/2/2014 11:55'!
writeScanOn: strm
	"Two formats.  c125000255 or cblue;"

	| nn |
	strm nextPut: self class scanCharacter.
	(nn := color name) ifNotNil: [
		(self class respondsTo: nn) ifTrue: [
			^ strm nextPutAll: nn; nextPut: $;]].
	(Array with: color red with: color green with: color blue) do: [:float |
		| str |
		str := '000', (float * 255) asInteger printString.
		strm nextPutAll: (str copyFrom: str size-2 to: str size)]! !

!TextColor methodsFor: 'comparing' stamp: 'di 10/31/97 11:19'!
= other 
	^ (other class == self class) 
		and: [other color = color]! !

TextAttribute subclass: #TextEmphasis
	instanceVariableNames: 'emphasisCode setMode'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!TextEmphasis commentStamp: '<historical>' prior: 0!
A TextEmphasis, encodes a characteristic applicable to all fonts.  The encoding is as follows:
	1	bold
	2	itallic
	4	underlined
	8	narrow
	16	struck out!

!TextEmphasis class methodsFor: 'instance creation' stamp: ''!
bold
	^ self new emphasisCode: 1! !

!TextEmphasis class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 13:25'!
consumeIdentifierFrom: scanStream.
	"Do not consume the identifier character. Leave it in the stream
	for use in identifying the type of emphasis."
! !

!TextEmphasis class methodsFor: 'instance creation' stamp: ''!
italic
	^ self new emphasisCode: 2! !

!TextEmphasis class methodsFor: 'instance creation' stamp: 'di 10/31/97 13:05'!
narrow
	^ TextKern kern: -1! !

!TextEmphasis class methodsFor: 'instance creation' stamp: ''!
normal
	^ self new emphasisCode: 0! !

!TextEmphasis class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 12:10'!
scanCharacters
	"All scan characters corresponding to this class. See writeScanOn:"
	^ #( $b $i $n $= $u )! !

!TextEmphasis class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 13:37'!
scanFrom: strm
	^strm next
		caseOf: {
			[ $b ] -> [ self bold ] .	
			[ $i ] -> [ self italic ] .	
			[ $u ] -> [ self underlined ] .	
			[ $= ] -> [ self struckOut ] .	
			[ $n ] -> [ self normal ]
		}
		otherwise: [self error: 'unrecognized identifier ']! !

!TextEmphasis class methodsFor: 'instance creation' stamp: ''!
struckOut
	^ self new emphasisCode: 16! !

!TextEmphasis class methodsFor: 'instance creation' stamp: ''!
underlined
	^ self new emphasisCode: 4! !

!TextEmphasis methodsFor: 'html' stamp: 'dgd 9/2/2005 12:39'!
closeHtmlOn: aStream 
	"put on the given stream the tag to close the html  
	representation of the receiver"
	emphasisCode = 1
		ifTrue: [aStream nextPutAll: '</b>'].
	emphasisCode = 2
		ifTrue: [aStream nextPutAll: '</i>'].
	emphasisCode = 4
		ifTrue: [aStream nextPutAll: '</u>']! !

!TextEmphasis methodsFor: 'testing' stamp: 'di 10/31/97 13:11'!
dominatedByCmd0
	"Cmd-0 should turn off emphasis"
	^ true! !

!TextEmphasis methodsFor: 'as yet unclassified' stamp: 'di 10/31/97 13:13'!
dominates: other
	(emphasisCode = 0 and: [other dominatedByCmd0]) ifTrue: [^ true].
	^ (other class == self class)
		and: [emphasisCode = other emphasisCode]! !

!TextEmphasis methodsFor: 'accessing' stamp: ''!
emphasisCode
	^ emphasisCode! !

!TextEmphasis methodsFor: 'initialize-release' stamp: ''!
emphasisCode: int
	emphasisCode := int.
	setMode := true! !

!TextEmphasis methodsFor: 'as yet unclassified' stamp: 'di 10/29/97 11:57'!
emphasizeScanner: scanner
	"Set the emphasist for text scanning"
	scanner addEmphasis: emphasisCode! !

!TextEmphasis methodsFor: 'comparing' stamp: 'ar 9/9/2003 22:03'!
hash
	"#hash is re-implemented because #= is re-implemented"
	^emphasisCode hash
! !

!TextEmphasis methodsFor: 'html' stamp: 'dgd 9/2/2005 12:38'!
openHtmlOn: aStream 
	"put on the given stream the tag to open the html  
	representation of the receiver"
	emphasisCode = 1
		ifTrue: [aStream nextPutAll: '<b>'].
	emphasisCode = 2
		ifTrue: [aStream nextPutAll: '<i>'].
	emphasisCode = 4
		ifTrue: [aStream nextPutAll: '<u>']! !

!TextEmphasis methodsFor: 'printing' stamp: ''!
printOn: strm
	super printOn: strm.
	strm nextPutAll: ' code: '; print: emphasisCode! !

!TextEmphasis methodsFor: 'accessing' stamp: ''!
set
	^ setMode and: [emphasisCode ~= 0]! !

!TextEmphasis methodsFor: 'initialize-release' stamp: ''!
turnOff
	setMode := false! !

!TextEmphasis methodsFor: 'fileIn/fileOut' stamp: 'tk 12/16/97 09:28'!
writeScanOn: strm

	emphasisCode = 1 ifTrue: [strm nextPut: $b].
	emphasisCode = 2 ifTrue: [strm nextPut: $i].
	emphasisCode = 0 ifTrue: [strm nextPut: $n].
	emphasisCode = 16 ifTrue: [strm nextPut: $=].
	emphasisCode = 4 ifTrue: [strm nextPut: $u].! !

!TextEmphasis methodsFor: 'comparing' stamp: 'di 10/31/97 11:15'!
= other 
	^ (other class == self class) 
		and: [other emphasisCode = emphasisCode]! !

TextAttribute subclass: #TextFontChange
	instanceVariableNames: 'fontNumber'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!TextFontChange commentStamp: '<historical>' prior: 0!
A TextFontChange encodes a font change applicable over a given range of text.  The font number is interpreted relative to the textStyle governing display of this text.!

!TextFontChange class methodsFor: 'instance creation' stamp: 'sw 12/6/1999 17:52'!
defaultFontChange
	"Answer a TextFontChange that represents the default font"

	^ self new fontNumber: TextStyle default defaultFontIndex! !

!TextFontChange class methodsFor: 'instance creation' stamp: ''!
font1
	^ self new fontNumber: 1! !

!TextFontChange class methodsFor: 'instance creation' stamp: ''!
font2
	^ self new fontNumber: 2! !

!TextFontChange class methodsFor: 'instance creation' stamp: ''!
font3
	^ self new fontNumber: 3! !

!TextFontChange class methodsFor: 'instance creation' stamp: ''!
font4
	^ self new fontNumber: 4! !

!TextFontChange class methodsFor: 'instance creation' stamp: ''!
fontNumber: n
	^ self new fontNumber: n! !

!TextFontChange class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 12:11'!
scanCharacter
	"The character used to identify a subclass of TextAttribute for filein and fileout"
	^$f! !

!TextFontChange class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 13:12'!
scanFrom: strm
	^self fontNumber: (Integer readFrom: strm ifFail: [0])! !

!TextFontChange methodsFor: 'html' stamp: 'dgd 9/2/2005 12:39'!
closeHtmlOn: aStream 
	"put on the given stream the tag to close the html  
	representation of the receiver"
	| font |
	font := TextStyle default fontAt: fontNumber.
	font closeHtmlOn: aStream! !

!TextFontChange methodsFor: 'as yet unclassified' stamp: 'nk 9/3/2004 15:48'!
dominates: other
	^ other isKindOf: TextFontChange! !

!TextFontChange methodsFor: 'as yet unclassified' stamp: ''!
emphasizeScanner: scanner
	"Set the font for text display"
	scanner setFont: fontNumber! !

!TextFontChange methodsFor: 'accessing' stamp: ''!
fontNumber
	^ fontNumber! !

!TextFontChange methodsFor: 'accessing' stamp: ''!
fontNumber: int
	fontNumber := int! !

!TextFontChange methodsFor: 'as yet unclassified' stamp: 'di 11/9/97 17:46'!
forFontInStyle: aTextStyle do: aBlock
	aBlock value: (aTextStyle fontAt: fontNumber)! !

!TextFontChange methodsFor: 'comparing' stamp: 'ar 9/9/2003 22:03'!
hash
	"#hash is re-implemented because #= is re-implemented"
	^fontNumber hash! !

!TextFontChange methodsFor: 'html' stamp: 'dgd 9/2/2005 12:38'!
openHtmlOn: aStream 
	"put on the given stream the tag to open the html  
	representation of the receiver"
	| font |
	font := TextStyle default fontAt: fontNumber.
	font openHtmlOn: aStream! !

!TextFontChange methodsFor: 'printing' stamp: ''!
printOn: strm
	super printOn: strm.
	strm nextPutAll: ' font: '; print: fontNumber! !

!TextFontChange methodsFor: 'fileIn/fileOut' stamp: 'dtl 1/2/2014 12:11'!
writeScanOn: strm

	strm nextPut: self class scanCharacter.
	fontNumber printOn: strm.! !

!TextFontChange methodsFor: 'comparing' stamp: 'di 10/31/97 11:15'!
= other 
	^ (other class == self class) 
		and: [other fontNumber = fontNumber]! !

TextFontChange subclass: #TextFontReference
	instanceVariableNames: 'font'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!TextFontReference commentStamp: '<historical>' prior: 0!
A TextFontReference encodes a font change applicable over a given range of text.  The font reference is absolute:  unlike a TextFontChange, it is independent of the textStyle governing display of this text.!

!TextFontReference class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 12:12'!
scanCharacter
	"The character used to identify a subclass of TextAttribute for filein and fileout"
	^$F! !

!TextFontReference class methodsFor: 'fileIn/Out' stamp: 'dtl 1/6/2014 17:45'!
scanFrom: strm
	^self toFont: 
		(StrikeFont familyName: (strm upTo: $#) size: (Number readFrom: strm ifFail: [0]))! !

!TextFontReference class methodsFor: 'as yet unclassified' stamp: ''!
toFont: aFont
	^ self new toFont: aFont! !

!TextFontReference methodsFor: 'testing' stamp: 'sw 11/9/1999 17:22'!
couldDeriveFromPrettyPrinting
	^ false! !

!TextFontReference methodsFor: 'initialize-release' stamp: ''!
emphasizeScanner: scanner
	"Set the actual font for text display"
	scanner setActualFont: font! !

!TextFontReference methodsFor: 'accessing' stamp: 'di 5/10/1999 23:47'!
font

	^ font! !

!TextFontReference methodsFor: 'as yet unclassified' stamp: 'di 11/9/97 17:47'!
forFontInStyle: aTextStyle do: aBlock
	aBlock value: font! !

!TextFontReference methodsFor: 'comparing' stamp: 'ar 9/9/2003 22:03'!
hash
	"#hash is re-implemented because #= is re-implemented"
	^font hash! !

!TextFontReference methodsFor: 'comparing' stamp: 'nk 9/3/2004 15:24'!
printOn: aStream
	aStream nextPutAll: 'a TextFontReference(';
		print: font;
		nextPut: $)! !

!TextFontReference methodsFor: 'initialize-release' stamp: ''!
toFont: aFont

	font := aFont! !

!TextFontReference methodsFor: 'fileIn/fileOut' stamp: 'dtl 1/2/2014 12:13'!
writeScanOn: strm

	strm nextPut: self class scanCharacter.
	strm nextPutAll: font familyName; nextPut: $#.
	font height printOn: strm.! !

!TextFontReference methodsFor: 'comparing' stamp: 'nk 9/3/2004 15:43'!
= other 
	^ (other class == self class) 
		and: [other font = font]! !

TextAttribute subclass: #TextIndent
	instanceVariableNames: 'amount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!TextIndent commentStamp: '<historical>' prior: 0!
create a hanging indent. !

!TextIndent class methodsFor: 'instance creation' stamp: 'ls 6/27/1998 15:55'!
amount: amount
	"create a TextIndent which will indent by the given amount.  Currently this is a number of tabs, but may change in the futur"
	^super new amount: amount! !

!TextIndent class methodsFor: 'example' stamp: 'ls 6/24/1998 18:06'!
example
	"TextIndent example"
	| text pg |

	"create an example text with some indentation"
	text := 'abcdao euoaeuo aeuo aeuoaeu o aeuoeauefgh bcd efghi'  asText.
	text addAttribute: (TextColor red)  from: 3 to: 8.
	text addAttribute: (TextIndent amount: 1) from: 1 to: 2.
	text addAttribute: (TextIndent amount: 2) from: 20 to: 35.

	"stick it in a paragraph and display it"
	pg := text asParagraph.
	pg compositionRectangle: (0@0 extent: 100@200).
	pg textStyle alignment: 2.
	pg displayAt: 0@0.
! !

!TextIndent class methodsFor: 'instance creation' stamp: 'ls 6/27/1998 15:54'!
tabs: numTabs
	"create an indentation by the given number of tabs"
	^self amount: numTabs! !

!TextIndent methodsFor: 'access' stamp: 'ls 6/22/1998 17:51'!
amount
	"number of tab spaces to indent by"
	^amount! !

!TextIndent methodsFor: 'access' stamp: 'ls 6/22/1998 17:51'!
amount: anInteger
	"change the number of tabs to indent by"
	amount := anInteger! !

!TextIndent methodsFor: 'condensing' stamp: 'ls 6/22/1998 19:27'!
dominates: anAttribute
	^(self class == anAttribute class)! !

!TextIndent methodsFor: 'setting indentation' stamp: 'ls 6/22/1998 18:56'!
emphasizeScanner: scanner
	scanner indentationLevel: amount! !

!TextIndent methodsFor: 'printing' stamp: 'ls 6/22/1998 18:03'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' amount: '.
	amount printOn: aStream! !

TextAttribute subclass: #TextKern
	instanceVariableNames: 'kern active'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!TextKern commentStamp: '<historical>' prior: 0!
A TextKern encodes a kerning change applicable over a given range of text.  Positive values of kern spread letters out, negative kern will cause them to overlap more.  Note that kerns other than 0 will display somewhat slower, as kerning is not yet supported in the text scanning primitive. !

!TextKern class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 13:44'!
consumeIdentifierFrom: scanStream.
	"Do not consume the identifier character. Leave it in the stream
	for use in identifying the kern value."
! !

!TextKern class methodsFor: 'instance creation' stamp: 'di 10/29/97 11:49'!
kern: kernValue
	^ self new kern: kernValue! !

!TextKern class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 12:31'!
scanCharacters
	"All scan characters corresponding to this class. See writeScanOn:"
	^ #( $- $+ )! !

!TextKern class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 13:51'!
scanFrom: strm
	| char k |
	char := strm next.
	char = $+
		ifTrue: [
			k := 1.
			[strm atEnd not and: [strm peek = char]]
				whileTrue: [strm next. k := k + 1].
			^ self kern: k].
	char = $-
		ifTrue: [
			k := -1.
			[strm atEnd not and: [strm peek = char]]
				whileTrue: [strm next. k := k - 1].
			^ self kern: k].
	self error: 'invalid identifier character'
			
	
! !

!TextKern methodsFor: 'testing' stamp: 'sw 11/9/1999 17:21'!
couldDeriveFromPrettyPrinting
	^ false! !

!TextKern methodsFor: 'testing' stamp: 'di 10/31/97 13:10'!
dominatedByCmd0
	"Cmd-0 should turn off kerning"
	^ true! !

!TextKern methodsFor: 'testing' stamp: 'di 10/31/97 11:10'!
dominates: other
	"NOTE: The use of active in this code is specific to its use in the method
		Text class addAttribute: att toArray: others"
	(active and: [other class == self class and: [other kern + kern = 0]])
		ifTrue: [active := false.  ^ true].  "can only dominate once"
	^ false! !

!TextKern methodsFor: 'kerning' stamp: 'di 10/29/97 11:50'!
emphasizeScanner: scanner
	"Augment (or diminish) the kerning offset for text display"
	scanner addKern: kern! !

!TextKern methodsFor: 'comparing' stamp: 'ar 9/9/2003 22:03'!
hash
	"#hash is re-implemented because #= is re-implemented"
	^kern hash! !

!TextKern methodsFor: 'testing' stamp: 'ar 9/21/2000 14:16'!
isKern
	^true! !

!TextKern methodsFor: 'accessing' stamp: 'di 10/31/97 11:12'!
kern
	^ kern! !

!TextKern methodsFor: 'initialize-release' stamp: 'tk 12/30/97 09:59'!
kern: kernValue
	kern := kernValue.
	self reset.! !

!TextKern methodsFor: 'initialize-release' stamp: 'di 10/31/97 11:04'!
reset
	active := true! !

!TextKern methodsFor: 'initialize-release' stamp: 'di 10/31/97 11:11'!
set
	^ active! !

!TextKern methodsFor: 'fileIn/fileOut' stamp: 'tk 9/21/1999 15:57'!
writeScanOn: strm

	kern > 0 ifTrue: [
		1 to: kern do: [:kk | strm nextPut: $+]].
	kern < 0 ifTrue: [
		1 to: 0-kern do: [:kk | strm nextPut: $-]].! !

!TextKern methodsFor: 'comparing' stamp: 'di 10/31/97 11:15'!
= other 
	^ (other class == self class) 
		and: [other kern = kern]! !

Object subclass: #TextReadWriter
	instanceVariableNames: 'stream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!TextReadWriter commentStamp: 'mt 5/7/2016 10:22' prior: 0!
Subclasses of me provide means to create instances of Text from various document formats such as HTML (Hyper Text Markup Language), DOCX (Microsoft Word Document), ODF (Open Document Format), and RTF (Rich Text Format). They should also be able to write such formats.!

TextReadWriter subclass: #HtmlReadWriter
	instanceVariableNames: 'count offset runStack runArray string'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Text'!

!HtmlReadWriter methodsFor: 'accessing' stamp: 'mt 5/3/2015 22:28'!
ignoredTags
	"Because we cannot process all of them."

	^ #(body script table tr td ul ol li form select option input)! !

!HtmlReadWriter methodsFor: 'testing' stamp: 'mt 5/3/2015 21:32'!
isTagIgnored: aTag

	| space t |
	space := aTag indexOf: Character space.
	t := space > 0
		ifTrue: [aTag copyFrom: 2 to: space - 1]
		ifFalse: [aTag copyFrom: 2 to: aTag size - 1].
	^ self ignoredTags includes: t! !

!HtmlReadWriter methodsFor: 'mapping' stamp: 'mt 8/11/2016 16:24'!
mapATag: aTag

	| result startIndex stopIndex attribute |
	result := OrderedCollection new.
	
	"<a href=""http://google.de"">"
	attribute := 'href'.
	startIndex := aTag findString: attribute.
	startIndex > 0 ifTrue: [
		startIndex := aTag findString: '"' startingAt: startIndex+attribute size.
		stopIndex := aTag findString: '"' startingAt: startIndex+1.
		result add: (TextURL new url: (aTag copyFrom: startIndex+1 to: stopIndex-1))].
	
	^ result! !

!HtmlReadWriter methodsFor: 'mapping' stamp: 'mt 8/2/2016 09:59'!
mapFontTag: aTag

	| result colorStartIndex colorStopIndex attribute |
	result := OrderedCollection new.
	
	"<font color=""#00FFCC""> or <font color=""green"">"
	attribute := 'color'.
	colorStartIndex := aTag findString: attribute.
	colorStartIndex > 0 ifTrue: [
		colorStartIndex := aTag findString: '"' startingAt: colorStartIndex+attribute size.		
		colorStopIndex := aTag findString: '"' startingAt: colorStartIndex+1.
		result add: (TextColor color:
			(Color fromString: (aTag copyFrom: colorStartIndex+1 to: colorStopIndex-1)))].

	^ result! !

!HtmlReadWriter methodsFor: 'mapping' stamp: 'mt 5/3/2015 22:05'!
mapTagToAttribute: aTag

	aTag = '<b>' ifTrue: [^ {TextEmphasis bold}].
	aTag = '<i>' ifTrue: [^ {TextEmphasis italic}].
	aTag = '<u>' ifTrue: [^ {TextEmphasis underlined}].
	"aTag = '<code>' ifTrue: [^ {TextFontReference toFont: Preferences standardCodeFont}]."
	(aTag beginsWith: '<font') ifTrue: [^ self mapFontTag: aTag].
	(aTag beginsWith: '<a') ifTrue: [^ self mapATag: aTag].

	"h1, h2, h3, ..."
	(aTag second = $h and: [aTag third isDigit])
		ifTrue: [^ {TextEmphasis bold}].

	^ {}! !

!HtmlReadWriter methodsFor: 'accessing' stamp: 'mt 5/3/2015 16:35'!
nextPutText: aText

	aText runs
		withStartStopAndValueDo: [:start :stop :attributes | 
			| att str | 
			att := aText attributesAt: start.
			str := aText string copyFrom: start to: stop.
			
			att do: [:each | self writeStartTagFor: each].
			self writeContent: str.
			att reverse do: [:each | self writeEndTagFor: each]]! !

!HtmlReadWriter methodsFor: 'accessing' stamp: 'mt 5/3/2015 22:00'!
nextText

	count := 0.
	offset := 0. "To ignore characters in the input string that are used by tags."
	
	runStack := Stack new.
	
	runArray := RunArray new.
	string := OrderedCollection new.
	
	"{text attributes. start index. end index. number of open tags}"
	runStack push: {OrderedCollection new. 1. nil. 0}.

	[stream atEnd] whileFalse: [self processNextTag].
	self processRunStackTop. "Add last run."

	string := String withAll: string.
	runArray coalesce.
	
	^ Text
		string: string
		runs: runArray! !

!HtmlReadWriter methodsFor: 'reading' stamp: 'mt 5/3/2015 22:37'!
processComment: aComment
! !

!HtmlReadWriter methodsFor: 'reading' stamp: 'mt 5/3/2015 21:26'!
processEmptyTag: aTag

	(aTag beginsWith: '<br') ifTrue: [
		string add: Character cr.
		count := count + 1.
		^ self].
	
	(self ignoredTags includes: (aTag copyFrom: 2 to: aTag size - 3))
		ifTrue: [^ self].
		
	"TODO..."! !

!HtmlReadWriter methodsFor: 'reading' stamp: 'mt 5/3/2015 22:24'!
processEndTag: aTag

	| index |
	index := count - offset.
	
	(self ignoredTags includes: (aTag copyFrom: 3 to: aTag size -1))
		ifTrue: [^ self].

	"De-Accumulate adjacent tags."
	runStack top at: 4 put: runStack top fourth - 1.
	runStack top fourth > 0
		ifTrue: [^ self "not yet"].
		
	self processRunStackTop.

	runStack pop.
	runStack top at: 2 put: index + 1.! !

!HtmlReadWriter methodsFor: 'reading' stamp: 'mt 5/3/2015 20:35'!
processHtmlEscape: aString

	(String htmlEntities at: (aString copyFrom: 2 to: aString size - 1) ifAbsent: [])
		ifNotNil: [:char | 
			string add: char.
			count := count + 1].! !

!HtmlReadWriter methodsFor: 'reading' stamp: 'mt 8/11/2016 16:18'!
processNextTag

	| tag htmlEscape lookForNewTag lookForHtmlEscape tagFound valid inComment inTagString |
	lookForNewTag := true.
	lookForHtmlEscape := false.
	tagFound := false.
	tag := OrderedCollection new.
	htmlEscape := OrderedCollection new.
	inComment := false.
	inTagString := false.
	
	[stream atEnd not and: [tagFound not]] whileTrue: [
		| character |
		character := stream next.
		valid := (#(10 13) includes: character asciiValue) not.
		count := count + 1.
	
		character = $< ifTrue: [lookForNewTag := false].
		character = $& ifTrue: [
			inComment ifFalse: [lookForHtmlEscape := true]].
		
		lookForNewTag
			ifTrue: [
				lookForHtmlEscape
					ifFalse: [valid ifTrue: [string add: character] ifFalse: [offset := offset + 1]]
					ifTrue: [valid ifTrue: [htmlEscape add: character]. offset := offset + 1]]
			ifFalse: [valid ifTrue: [tag add: character]. offset := offset + 1].

		"Toggle within tag string/text."
		(character = $" and: [lookForNewTag not])
			ifTrue: [inTagString := inTagString not].
		
		inComment := ((lookForNewTag not and: [tag size >= 4])
			and: [tag beginsWith: '<!!--'])
			and: [(tag endsWith: '-->') not].

		(((character = $> and: [inComment not]) and: [lookForNewTag not]) and: [inTagString not]) ifTrue: [
			lookForNewTag := true.
			(tag beginsWith: '<!!--')
				ifTrue: [self processComment: (String withAll: tag)]
				ifFalse: [tag second ~= $/
					ifTrue: [
						(tag atLast: 2) == $/
							ifTrue: [self processEmptyTag: (String withAll: tag)]
							ifFalse: [self processStartTag: (String withAll: tag)]]
					ifFalse: [self processEndTag: (String withAll: tag)]].			
			tagFound := true].

		(((character = $; and: [lookForNewTag])
			and: [htmlEscape notEmpty]) and: [htmlEscape first = $&]) ifTrue: [
				lookForHtmlEscape := false.
				self processHtmlEscape: (String withAll: htmlEscape).
				htmlEscape := OrderedCollection new]].
! !

!HtmlReadWriter methodsFor: 'reading' stamp: 'mt 5/3/2015 17:35'!
processRunStackTop
	"Write accumulated attributes to run array."
	
	| index start end attrs |
	index := count - offset.
	
	"Set end index."
	runStack top at: 3 put: index.
	"Write to run array."
	start := runStack top second.
	end := runStack top third.
	attrs := runStack top first.
	runArray
		addLast: attrs asArray
		times: end - start + 1.! !

!HtmlReadWriter methodsFor: 'reading' stamp: 'mt 5/3/2015 22:24'!
processStartTag: aTag

	| index |
	(self isTagIgnored: aTag) ifTrue: [^ self].

	index := count - offset.

	aTag = '<br>' ifTrue: [
		string add: Character cr.
		count := count + 1.
		^ self].
	(aTag beginsWith: '<img') ifTrue: [
		string addAll: '[image]'.
		count := count + 7.
		^ self].
	
	"Accumulate adjacent tags."
	(runStack size > 1 and: [runStack top second = (index + 1) "= adjacent start tags"])
		ifTrue: [
			runStack top at: 1 put: (runStack top first copy addAll: (self mapTagToAttribute: aTag); yourself).
			runStack top at: 4 put: (runStack top fourth + 1). "increase number of open tags"
			^self].
	
	self processRunStackTop.

	"Remove start/end info to reuse attributes later."
	runStack top at: 2 put: nil.
	runStack top at: 3 put: nil.
	"Copy attr list and add new attr."
	runStack push: ({runStack top first copy addAll: (self mapTagToAttribute: aTag); yourself. index + 1. nil. 1}).! !

!HtmlReadWriter methodsFor: 'writing' stamp: 'mt 5/3/2015 20:47'!
writeContent: aString

	aString do: [:char |
		(#(10 13) includes: char asciiValue)
			ifTrue: [stream nextPutAll: '<br>'; cr]
			ifFalse: [char = Character tab
				ifTrue: [stream nextPutAll: '&nbsp;&nbsp;&nbsp;&nbsp;']
				ifFalse: [(String htmlEntities keyAtValue: char ifAbsent: [])
					ifNil: [stream nextPut: char]
					ifNotNil: [:escapeSequence |
						stream
							nextPut: $&;
							nextPutAll: escapeSequence;
							nextPut: $;]]]].! !

!HtmlReadWriter methodsFor: 'writing' stamp: 'mt 5/3/2015 19:22'!
writeEndTagFor: aTextAttribute

	[aTextAttribute closeHtmlOn: stream]
		on: MessageNotUnderstood do: []! !

!HtmlReadWriter methodsFor: 'writing' stamp: 'mt 5/3/2015 19:22'!
writeStartTagFor: aTextAttribute

	[aTextAttribute openHtmlOn: stream]
		on: MessageNotUnderstood do: [].! !

!TextReadWriter class methodsFor: 'instance creation' stamp: 'mt 5/3/2015 17:46'!
on: stream

	^ self new on: stream! !

!TextReadWriter class methodsFor: 'instance creation' stamp: 'mt 8/11/2016 13:17'!
textFromFileNamed: fileName

	^ self textFromStream: (FileStream readOnlyFileNamed: fileName)! !

!TextReadWriter class methodsFor: 'instance creation' stamp: 'mt 8/11/2016 13:17'!
textFromStream: aBinaryStream
	
	| reader readerClass text |
	readerClass := self. "no auto-detection yet"
	aBinaryStream reset.
	reader := readerClass new on: aBinaryStream.
	Cursor read showWhile: [
		text := reader nextText.
		reader close].
	^ text
! !

!TextReadWriter methodsFor: 'initialize-release' stamp: 'mt 8/11/2016 13:17'!
close

	stream close.! !

!TextReadWriter methodsFor: 'accessing' stamp: 'mt 5/3/2015 16:17'!
nextPutText: aText
	"Encoding aText on stream."
	
	self subclassResponsibility.! !

!TextReadWriter methodsFor: 'accessing' stamp: 'mt 5/3/2015 16:16'!
nextText
	"Decoding a text object on stream and answer that text object."
	
	^ self subclassResponsibility.! !

!TextReadWriter methodsFor: 'initialize-release' stamp: 'mt 5/3/2015 17:34'!
on: aStream

	stream := aStream.! !

Magnitude immediateSubclass: #Character
	instanceVariableNames: ''
	classVariableNames: 'AlphaNumericMask CharacterTable ClassificationTable DigitBit DigitValues LetterMask LowercaseBit UppercaseBit'
	poolDictionaries: ''
	category: 'Collections-Strings'!

!Character commentStamp: 'eem 8/12/2014 14:53' prior: 0!
I represent a character by storing its associated Unicode as an unsigned 30-bit value.  Characters are created uniquely, so that all instances of a particular Unicode are identical.  My instances are encoded in tagged pointers in the VM, so called immediates, and therefore are pure immutable values.

	The code point is based on Unicode.  Since Unicode is 21-bit wide character set, we have several bits available for other information.  As the Unicode Standard  states, a Unicode code point doesn't carry the language information.  This is going to be a problem with the languages so called CJK (Chinese, Japanese, Korean.  Or often CJKV including Vietnamese).  Since the characters of those languages are unified and given the same code point, it is impossible to display a bare Unicode code point in an inspector or such tools.  To utilize the extra available bits, we use them for identifying the languages.  Since the old implementation uses the bits to identify the character encoding, the bits are sometimes called "encoding tag" or neutrally "leading char", but the bits rigidly denotes the concept of languages.

	The other languages can have the language tag if you like.  This will help to break the large default font (font set) into separately loadable chunk of fonts.  However, it is open to the each native speakers and writers to decide how to define the character equality, since the same Unicode code point may have different language tag thus simple #= comparison may return false.!

!Character class methodsFor: 'instance creation' stamp: 'ar 10/25/2010 19:00'!
allByteCharacters
	"Answer all the characters that can be encoded in a byte"
	^ (0 to: 255) collect: [:v | Character value: v] as: String
! !

!Character class methodsFor: 'instance creation' stamp: 'ar 4/9/2005 22:37'!
allCharacters
	"This name is obsolete since only the characters that will fit in a byte can be queried"
	^self allByteCharacters
	
! !

!Character class methodsFor: 'constants' stamp: 'rhi 9/8/2000 14:57'!
alphabet
	"($a to: $z) as: String"

	^ 'abcdefghijklmnopqrstuvwxyz' copy! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'NS 7/11/2000 09:20'!
arrowDown
	^ self value: 31! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'NS 7/11/2000 09:20'!
arrowLeft
	^ self value: 28! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'NS 7/11/2000 09:20'!
arrowRight
	^ self value: 29! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'NS 7/11/2000 09:20'!
arrowUp
	^ self value: 30! !

!Character class methodsFor: 'accessing untypeable characters' stamp: ''!
backspace
	"Answer the Character representing a backspace."

	^self value: 8! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'tpr 9/30/2013 16:03'!
characterForColumnBreak
"use a newPage to split columns in character scanners"
	^self newPage! !

!Character class methodsFor: 'constants' stamp: ''!
characterTable
	"Answer the class variable in which unique Characters are stored."

	^CharacterTable! !

!Character class methodsFor: 'instance creation' stamp: 'ar 9/1/2009 00:23'!
codePoint: integer 
	"Return a character whose encoding value is integer.
	For ansi compability."
	^self value: integer! !

!Character class methodsFor: 'private' stamp: 'lr 11/21/2005 17:24'!
constantNameFor: aCharacter
	^ self constantNames
		detect: [ :each | (self perform: each) = aCharacter ]
		ifNone: [ nil ].! !

!Character class methodsFor: 'private' stamp: 'nice 5/30/2013 23:48'!
constantNames
	^ #( backspace cr delete escape lf null newPage space tab ).! !

!Character class methodsFor: 'accessing untypeable characters' stamp: ''!
cr
	"Answer the Character representing a carriage return."

	^self value: 13! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'NS 7/11/2000 09:19'!
delete
	^ self value: 127! !

!Character class methodsFor: 'instance creation' stamp: 'eem 9/24/2013 23:43'!
digitValue: x 
	"Answer the Character whose digit value is x. For example,
	 answer $9 for x=9, $0 for x=0, $A for x=10, $Z for x=35."

	| n |
	n := x asInteger.
	^self value: (n < 10 ifTrue: [n + 48] ifFalse: [n + 55])! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'NS 7/11/2000 09:21'!
end
	^ self value: 4! !

!Character class methodsFor: 'accessing untypeable characters' stamp: ''!
enter
	"Answer the Character representing enter."

	^self value: 3! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'ls 9/2/1999 08:06'!
escape
	"Answer the ASCII ESC character"

	^self value: 27! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'nice 3/29/2010 18:14'!
euro
	"The Euro currency sign, that E with two dashes. The code point is a official unicode ISO/IEC-10646-1"

	^ Unicode value: 16r20AC! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'NS 7/11/2000 09:21'!
home
	^ self value: 1! !

!Character class methodsFor: 'class initialization' stamp: 'ul 8/14/2015 18:36'!
initialize
	"Character initialize"
	
	self
		initializeClassificationTable;
		initializeDigitValues! !

!Character class methodsFor: 'class initialization' stamp: 'ul 5/1/2015 19:17'!
initializeClassificationTable
	"Initialize the classification table.
	The classification table is a compact encoding of upper and lower cases and digits of characters with
		- bits 0-7: The lower case value of this character or 0, if its greater than 255.
		- bits 8-15: The upper case value of this character or 0, if its greater than 255.
		- bit 16: lowercase bit (isLowercase == true)
		- bit 17: uppercase bit (isUppercase == true)
		- bit 18: digit bit (isDigit == true)"
	" self initializeClassificationTable "

	| encodedCharSet newClassificationTable |
	"Base the table on the EncodedCharset of these characters' leadingChar - 0."
	encodedCharSet := EncodedCharSet charsetAt: 0.

	LowercaseBit := 1 bitShift: 16.
	UppercaseBit := 1 bitShift: 17.
	DigitBit := 1 bitShift: 18.

	"Initialize the letter mask (e.g., isLetter == true)"
	LetterMask := LowercaseBit bitOr: UppercaseBit.

	"Initialize the alphanumeric mask (e.g. isAlphaNumeric == true)"
	AlphaNumericMask := LetterMask bitOr: DigitBit.

	"Initialize the table based on encodedCharSet."
	newClassificationTable := Array new: 256.
	0 to: 255 do: [ :code |
		| isLowercase isUppercase isDigit lowercaseCode uppercaseCode value |
		isLowercase := encodedCharSet isLowercaseCode: code.
		isUppercase := encodedCharSet isUppercaseCode: code.
		isDigit := encodedCharSet isDigitCode: code.
		lowercaseCode := encodedCharSet toLowercaseCode: code.
		lowercaseCode > 255 ifTrue: [ lowercaseCode := 0 ].
		uppercaseCode := encodedCharSet toUppercaseCode: code.
		uppercaseCode > 255 ifTrue: [ uppercaseCode := 0 ].
		value := (uppercaseCode bitShift: 8) + lowercaseCode.
		isLowercase ifTrue: [ value := value bitOr: LowercaseBit ].
		isUppercase ifTrue: [ value := value bitOr: UppercaseBit ].
		isDigit ifTrue: [ value := value bitOr: DigitBit ].
		newClassificationTable at: code + 1 put: value ].
	ClassificationTable := newClassificationTable! !

!Character class methodsFor: 'class initialization' stamp: 'ul 8/14/2015 18:36'!
initializeDigitValues
	"Initialize the well known digit value of ascii characters.
	Note that the DigitValues table is 1-based while ascii values are 0-based, thus the offset +1."
	
	| newDigitValues |
	newDigitValues := Array new: 256 withAll: -1.
	"the digits"
	0 to: 9 do: [:i | newDigitValues at: 48 + i + 1 put: i].
	"the uppercase letters"
	10 to: 35 do: [:i | newDigitValues at: 55 + i + 1 put: i].
	"the lowercase letters"
	10 to: 35 do: [:i | newDigitValues at: 87 + i + 1 put: i].
	DigitValues := newDigitValues! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'NS 7/11/2000 09:19'!
insert
	^ self value: 5! !

!Character class methodsFor: 'instance creation' stamp: 'ul 4/26/2011 02:29'!
leadingChar: leadChar code: code

	code >= 16r400000 ifTrue: [
		self error: 'code is out of range'.
	].
	leadChar >= 256 ifTrue: [
		self error: 'lead is out of range'.
	].
	code < 256 ifTrue: [ ^self value: code ].
	^self value: (leadChar bitShift: 22) + code.! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'ls 9/8/1998 22:15'!
lf
	"Answer the Character representing a linefeed."

	^self value: 10! !

!Character class methodsFor: 'accessing untypeable characters' stamp: ''!
linefeed
	"Answer the Character representing a linefeed."

	^self value: 10! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'wiz 4/9/2006 20:30'!
nbsp
	"non-breakable space. Latin1 encoding common usage."

	^ Character value: 160! !

!Character class methodsFor: 'instance creation' stamp: ''!
new
	"Creating new characters is not allowed."

	self error: 'cannot create new characters'! !

!Character class methodsFor: 'accessing untypeable characters' stamp: ''!
newPage
	"Answer the Character representing a form feed."

	^self value: 12! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'CamilloBruni 10/20/2012 23:51'!
null
	^ self value: 0! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'NS 7/11/2000 09:20'!
pageDown
	^ self value: 12! !

!Character class methodsFor: 'accessing untypeable characters' stamp: 'NS 7/11/2000 09:21'!
pageUp
	^ self value: 11! !

!Character class methodsFor: 'instance creation' stamp: 'ar 10/25/2010 18:54'!
separators
	"Answer a collection of the standard ASCII separator characters."

	^ #(32 "space"
		13 "cr"
		9 "tab"
		10 "line feed"
		12 "form feed")
		collect: [:v | Character value: v] as: String! !

!Character class methodsFor: 'accessing untypeable characters' stamp: ''!
space
	"Answer the Character representing a space."

	^self value: 32! !

!Character class methodsFor: 'accessing untypeable characters' stamp: ''!
tab
	"Answer the Character representing a tab."

	^self value: 9! !

!Character class methodsFor: 'instance creation' stamp: 'eem 2/5/2014 12:49'!
value: anInteger
	"Answer the Character whose value is anInteger."
	<primitive: 170>
	^self primitiveFailed! !

!Character methodsFor: 'converting' stamp: ''!
asCharacter
	"Answer the receiver itself."

	^self! !

!Character methodsFor: 'accessing' stamp: 'eem 2/5/2014 12:57'!
asciiValue
	"Answer the receiver's character code.
	 This will be ascii for characters with value <= 127,
	 and Unicode for those with higher values."
	<primitive: 171>
	^self primitiveFailed! !

!Character methodsFor: 'converting' stamp: 'eem 2/5/2014 12:57'!
asInteger
	"Answer the receiver's character code."
	<primitive: 171>
	^self primitiveFailed! !

!Character methodsFor: 'converting' stamp: 'ul 8/12/2015 22:07'!
asLowercase
	"Answer the receiver's matching lowercase Character."
	
	| integerValue |
	(integerValue := self asInteger) > 255 ifFalse: [ 
		| result |
		(result := (ClassificationTable at: integerValue + 1) bitAnd: 16rFF) > 0
			ifTrue: [ ^self class value: result ] ].
	^self class value: (self encodedCharSet toLowercaseCode: integerValue)! !

!Character methodsFor: 'converting' stamp: 'sma 3/11/2000 17:21'!
asString
	^ String with: self! !

!Character methodsFor: 'converting' stamp: 'raa 5/26/2001 09:54'!
asSymbol 
	"Answer a Symbol consisting of the receiver as the only element."

	^Symbol internCharacter: self! !

!Character methodsFor: 'converting' stamp: 'tk 9/4/2000 12:05'!
asText
	^ self asString asText! !

!Character methodsFor: 'converting' stamp: 'ul 8/12/2015 22:10'!
asUnicode
	"Answer the unicode encoding of the receiver"
	
	| integerValue |
	(integerValue := self asInteger) <= 16r3FFFFF ifTrue: [ ^integerValue ].
	^self encodedCharSet charsetClass convertToUnicode: (integerValue bitAnd: 16r3FFFFF)
! !

!Character methodsFor: 'converting' stamp: 'ul 8/12/2015 22:08'!
asUppercase
	"Answer the receiver's matching uppercase Character."
	
	| integerValue |
	(integerValue := self asInteger) > 255 ifFalse: [ 
		| result |
		(result := ((ClassificationTable at: integerValue + 1) bitShift: -8) bitAnd: 16rFF) > 0
			ifTrue: [ ^self class value: result ] ].
	^self class value: (self encodedCharSet toUppercaseCode: integerValue)! !

!Character methodsFor: 'testing' stamp: 'ul 7/7/2014 21:53'!
canBeGlobalVarInitial

	^self encodedCharSet canBeGlobalVarInitial: self! !

!Character methodsFor: 'testing' stamp: 'cwp 8/14/2012 10:39'!
canBeIdentifierInitial
	^ self == $_ or: [self isLetter]! !

!Character methodsFor: 'testing' stamp: 'ul 7/7/2014 21:53'!
canBeNonGlobalVarInitial

	^self encodedCharSet canBeNonGlobalVarInitial: self
! !

!Character methodsFor: 'accessing' stamp: 'yo 12/29/2002 10:11'!
charCode

	^ (self asInteger bitAnd: 16r3FFFFF).
! !

!Character methodsFor: 'copying' stamp: 'eem 2/5/2014 12:52'!
clone
	"Answer the receiver, because Characters are unique."
	^self! !

!Character methodsFor: 'accessing' stamp: 'ul 10/8/2014 12:18'!
codePoint
	"Return the encoding value of the receiver."
	#Fundmntl.

	^self asInteger! !

!Character methodsFor: 'copying' stamp: 'eem 2/5/2014 12:52'!
copy
	"Answer the receiver, because Characters are unique."
	^self! !

!Character methodsFor: 'copying' stamp: 'eem 2/5/2014 12:52'!
deepCopy
	"Answer the receiver, because Characters are unique."
	^self! !

!Character methodsFor: 'accessing' stamp: 'ul 8/12/2015 22:08'!
digitValue
	"Answer 0-9 if the receiver is $0-$9, 10-35 if it is $A-$Z, and < 0 
	otherwise. This is used to parse literal numbers of radix 2-36."

	| integerValue |
	(integerValue := self asInteger) > 16rFF ifTrue: [^self encodedCharSet digitValueOf: self].
	^DigitValues at: integerValue + 1! !

!Character methodsFor: 'accessing' stamp: 'ul 5/1/2015 13:26'!
encodedCharSet
	
	self asInteger < 16r400000 ifTrue: [ ^Unicode ]. "Shortcut"
	^EncodedCharSet charsetAt: self leadingChar
! !

!Character methodsFor: 'comparing' stamp: 'eem 2/5/2014 12:55'!
hash
	"Hash is reimplemented because = is implemented.
	 Answer the receiver's character code."
	<primitive: 171>
	^self primitiveFailed! !

!Character methodsFor: 'printing' stamp: 'eem 5/21/2010 11:11'!
hex
	^self asInteger printStringBase: 16! !

!Character methodsFor: 'comparing' stamp: 'eem 2/5/2014 12:56'!
identityHash
	"Answer the receiver's character code."
	<primitive: 171>
	^self primitiveFailed! !

!Character methodsFor: 'testing' stamp: 'ul 8/12/2015 22:11'!
isAlphaNumeric
	"Answer whether the receiver is a letter or a digit."

	| integerValue |
	(integerValue := self asInteger) > 255 ifFalse: [ 
		^((ClassificationTable at: integerValue + 1) bitAnd: AlphaNumericMask) > 0 ].
	^self encodedCharSet isAlphaNumeric: self! !

!Character methodsFor: 'testing' stamp: 'ul 8/12/2015 22:14'!
isAscii

	^self asInteger < 128! !

!Character methodsFor: 'testing' stamp: 'yo 8/28/2002 13:42'!
isCharacter

	^ true.
! !

!Character methodsFor: 'testing' stamp: 'ul 8/12/2015 22:12'!
isDigit

	| integerValue |
	(integerValue := self asInteger) > 255 ifFalse: [
		^((ClassificationTable at: integerValue + 1) bitAnd: DigitBit) > 0 ].
	^self encodedCharSet isDigit: self.
! !

!Character methodsFor: 'testing' stamp: 'ul 8/12/2015 22:13'!
isLetter

	| integerValue |
	(integerValue := self asInteger) > 255 ifFalse: [
		^((ClassificationTable at: integerValue + 1) bitAnd: LetterMask) > 0 ].
	^self encodedCharSet isLetter: self! !

!Character methodsFor: 'testing' stamp: ''!
isLiteral

	^true! !

!Character methodsFor: 'testing' stamp: 'ul 8/12/2015 22:13'!
isLowercase

	| integerValue |
	(integerValue := self asInteger) > 255 ifFalse: [
		^((ClassificationTable at: integerValue + 1) bitAnd: LowercaseBit) > 0 ].
	^self encodedCharSet isLowercase: self.
! !

!Character methodsFor: 'testing' stamp: 'yo 8/27/2002 15:18'!
isOctetCharacter

	^ self asInteger < 256.
! !

!Character methodsFor: 'converting' stamp: 'ar 4/9/2005 22:15'!
isoToSqueak 
	^self "no longer needed"! !

!Character methodsFor: 'testing' stamp: 'yo 7/29/2005 15:21'!
isSafeForHTTP
	"whether a character is 'safe', or needs to be escaped when used, eg, in a URL"
	"[GG]  See http://www.faqs.org/rfcs/rfc1738.html. ~ is unsafe and has been removed"
	^ self charCode < 128
		and: [self isAlphaNumeric
				or: ['.-_' includes: (Character value: self charCode)]]! !

!Character methodsFor: 'testing' stamp: 'ul 8/14/2015 18:58'!
isSeparator
	"Answer whether the receiver is one of the separator characters--space, 
	cr, tab, line feed, or form feed."

	| integerValue |
	(integerValue := self asInteger) > 32 ifTrue: [ ^false ].
	integerValue
		caseOf: {
			[ 32 "space" ] -> [ ^true ].
			[ 9 "cr" ] -> [ ^true ].
			[ 13 "tab"] -> [ ^true ].
			[ 10 "line feed" ] -> [ ^true ] }
		otherwise: [ ^integerValue = 12 "form feed" ]! !

!Character methodsFor: 'testing' stamp: 'di 4/3/1999 00:38'!
isSpecial
	"Answer whether the receiver is one of the special characters"

	^'+-/\*~<>=@,%|&?!!' includes: self! !

!Character methodsFor: 'testing' stamp: 'ul 7/7/2014 21:53'!
isTraditionalDomestic
	"Yoshiki's note about #isUnicode says:
		[This method] is for the backward compatibility when we had domestic
		traditional encodings for CJK languages.  To support loading the
		projects in traditional domestic encodings (From Nihongo4), and load
		some changesets.  Once we decided to get rid of classes like JISX0208
		from the EncodedCharSet table, the need for isUnicode will not be
		necessary.
	I (Andreas) decided to change the name from isUnicode to #isTraditionalDomestic
	since I found isUnicode to be horribly confusing (how could the character *not*
	be Unicode after all?). But still, we should remove this method in due time."
	^ (self encodedCharSet isKindOf: LanguageEnvironment class) not! !

!Character methodsFor: 'testing' stamp: 'ul 8/12/2015 22:14'!
isUppercase

	| integerValue |
	(integerValue := self asInteger) > 255 ifFalse: [
		^((ClassificationTable at: integerValue + 1) bitAnd: UppercaseBit) > 0 ].
	^self encodedCharSet isUppercase: self.
! !

!Character methodsFor: 'testing' stamp: ''!
isVowel
	"Answer whether the receiver is one of the vowels, AEIOU, in upper or 
	lower case."

	^'AEIOU' includes: self asUppercase! !

!Character methodsFor: 'accessing' stamp: 'nice 5/1/2011 22:13'!
leadingChar
	"Answer the value of the 8 highest bits which is used to identify the language.
	This is mostly used for east asian languages CJKV as a workaround against unicode han-unification."
	^ self asInteger bitShift: -22! !

!Character methodsFor: 'converting' stamp: 'ul 10/8/2014 12:18'!
macToSqueak
	"Convert the receiver from MacRoman to Squeak encoding"
	| asciiValue |
	self asInteger < 128 ifTrue: [^ self].
	self asInteger > 255 ifTrue: [^ self].
	asciiValue := #[
		196 197 199 201 209 214 220 225 224 226 228 227 229 231 233 232	"80-8F"
		234 235 237 236 238 239 241 243 242 244 246 245 250 249 251 252	"90-9F"
		134 176 162 163 167 149 182 223 174 169 153 180 168 128 198 216	"A0-AF"
		129 177 138 141 165 181 142 143 144 154 157 170 186 158 230 248	"B0-BF"
		191 161 172 166 131 173 178 171 187 133 160 192 195 213 140 156	"C0-CF"
		150 151 147 148 145 146 247 179 255 159 185 164 139 155 188 189	"D0-DF"
		135 183 130 132 137 194 202 193 203 200 205 206 207 204 211 212	"E0-EF"
		190 210 218 219 217 208 136 152 175 215 221 222 184 240 253 254 ]	"F0-FF"
			at: self asInteger - 127.
	^ Character value: asciiValue.! !

!Character methodsFor: 'object fileIn' stamp: 'tk 2/16/2001 14:52'!
objectForDataStream: refStrm
	"I am being collected for inclusion in a segment.  Do not include Characters!!  Let them be in outPointers."

	refStrm insideASegment
		ifFalse: ["Normal use" ^ self]
		ifTrue: ["recording objects to go into an ImageSegment"			
			"remove it from references.  Do not trace."
			refStrm references removeKey: self ifAbsent: [].
			^ nil]
! !

!Character methodsFor: 'printing' stamp: 'ul 8/12/2015 22:18'!
printOn: aStream

	| integerValue |
	((integerValue := self asInteger) > 32 and: [ integerValue ~= 127 ]) ifTrue: [
		aStream nextPut: $$; nextPut: self.
		^self ].
	(self class constantNameFor: self)
		ifNotNil: [ :name | aStream nextPutAll: self class name; space; nextPutAll: name ]
		ifNil: [ aStream nextPutAll: self class name; nextPutAll: ' value: '; print: integerValue ]! !

!Character methodsFor: 'comparing' stamp: 'ul 8/22/2015 21:23'!
sameAs: aCharacter 
	"Answer whether the receiver is equal to aCharacter, ignoring case"
	
	self == aCharacter ifTrue: [ ^true ].
	^self asLowercase == aCharacter asLowercase! !

!Character methodsFor: 'copying' stamp: 'eem 2/5/2014 12:53'!
shallowCopy
	"Answer the receiver, because Characters are unique."
	^self! !

!Character methodsFor: 'testing' stamp: 'ul 8/16/2015 14:05'!
shouldBePrintedAsLiteral

	| integerValue |
	(integerValue := self asInteger) < 33 ifTrue: [ ^false ].
	255 < integerValue ifTrue: [ ^false ].
	^integerValue ~= 127! !

!Character methodsFor: 'testing' stamp: 'eem 11/2/2011 11:03'!
shouldBePrintedAsLiteralVisiting: aSet

	^self shouldBePrintedAsLiteral! !

!Character methodsFor: 'converting' stamp: 'ar 4/9/2005 22:16'!
squeakToIso
	^self "no longer needed"! !

!Character methodsFor: 'converting' stamp: 'nice 7/14/2011 22:34'!
squeakToMac
	"Convert the receiver from Squeak to MacRoman encoding."
	self asInteger < 128 ifTrue: [^ self].
	self asInteger > 255 ifTrue: [^ self].
	^ Character value: (#[
		173 176 226 196 227 201 160 224 246 228 178 220 206 179 182 183	"80-8F"
		184 212 213 210 211 165 208 209 247 170 185 221 207 186 189 217	"90-9F"
		202 193 162 163 219 180 195 164 172 169 187 199 194 197 168 248	"A0-AF"
		161 177 198 215 171 181 166 225 252 218 188 200 222 223 240 192 	"B0-BF"
		203 231 229 204 128 129 174 130 233 131 230 232 237 234 235 236 	"C0-CF"
		245 132 241 238 239 205 133 249 175 244 242 243 134 250 251 167	"D0-DF"
		136 135 137 139 138 140 190 141 143 142 144 145 147 146 148 149	"E0-EF"
		253 150 152 151 153 155 154 214 191 157 156 158 159 254 255 216	"F0-FF"
	] at: self asInteger - 127)
! !

!Character methodsFor: 'printing' stamp: 'ul 8/12/2015 22:18'!
storeBinaryOn: aStream
	"Store the receiver on a binary (file) stream"
	
	| integerValue |
	(integerValue := self asInteger) < 256 
		ifTrue: [ aStream basicNextPut: self ]
		ifFalse: [ aStream nextInt32Put: integerValue ]! !

!Character methodsFor: 'object fileIn' stamp: 'ul 8/12/2015 22:19'!
storeDataOn: aDataStream
	" Store characters in reference-like way, with value like instvar.
	This is compatible with various Squeak Memory Systems"

	aDataStream
		beginInstance: self class
			size: 1;
		nextPut: self asInteger! !

!Character methodsFor: 'printing' stamp: 'ul 8/12/2015 22:21'!
storeOn: aStream
	"Common character literals are preceded by '$', however special need to be encoded differently: for some this might be done by using one of the shortcut constructor methods for the rest we have to create them by ascii-value."

	self shouldBePrintedAsLiteral ifTrue: [ 
		aStream nextPut: $$; nextPut: self.
		^self ].
	(self class constantNameFor: self) ifNotNil: [ :name |
		aStream nextPutAll: self class name; space; nextPutAll: name.
		^self ].
	aStream 
		nextPut: $(; nextPutAll: self class name; 
		nextPutAll: ' value: '; print: self asInteger; nextPut: $)! !

!Character methodsFor: 'testing' stamp: 'ul 5/1/2015 13:12'!
tokenish
	"Answer whether the receiver is a valid token-character--letter, digit, or colon."

	self == $_ ifTrue: [ ^Scanner prefAllowUnderscoreSelectors ].
	^self == $: or: [ self isAlphaNumeric ]! !

!Character methodsFor: 'converting' stamp: 'ul 10/8/2014 12:19'!
to: other
	"Answer with a collection in ascii order -- $a to: $z"
	^ (self asInteger to: other asciiValue)
		collect:	[:ascii | Character value: ascii]
		as: String! !

!Character methodsFor: 'copying' stamp: 'eem 2/5/2014 12:53'!
veryDeepCopyWith: deepCopier
	"Answer the receiver, because Characters are unique."
	^self! !

!Character methodsFor: 'comparing' stamp: 'ul 8/12/2015 22:06'!
< aCharacter 
	"Answer true if the receiver's value < aCharacter's value."

	^self asInteger < aCharacter asInteger! !

!Character methodsFor: 'comparing' stamp: 'ul 8/12/2015 22:06'!
<= aCharacter 
	"Answer true if the receiver's value <= aCharacter's value."

	^self asInteger <= aCharacter asInteger! !

!Character methodsFor: 'comparing' stamp: 'eem 10/11/2013 12:55'!
= aCharacter 
	"Primitive. Answer if the receiver and the argument are the
	 same object (have the same object pointer). Optional. See
	 Object documentation whatIsAPrimitive."
	<primitive: 110>
	^self == aCharacter! !

!Character methodsFor: 'comparing' stamp: 'ul 8/12/2015 22:06'!
> aCharacter 
	"Answer true if the receiver's value > aCharacter's value."

	^self asInteger > aCharacter asInteger! !

!Character methodsFor: 'comparing' stamp: 'ul 8/12/2015 22:06'!
>= aCharacter 
	"Answer true if the receiver's value >= aCharacter's value."

	^self asInteger >= aCharacter asInteger! !

Magnitude subclass: #LookupKey
	instanceVariableNames: 'key'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Support'!

!LookupKey commentStamp: '<historical>' prior: 0!
I represent a key for looking up entries in a data structure. Subclasses of me, such as Association, typically represent dictionary entries.!

LookupKey subclass: #Association
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Support'!

!Association commentStamp: '<historical>' prior: 0!
I represent a pair of associated objects--a key and a value. My instances can serve as entries in a dictionary.!

!Association class methodsFor: 'instance creation' stamp: 'ul 12/14/2009 17:57'!
key: newKey value: newValue
	"Answer an instance of me with the arguments as the key and value of 
	the association."

	^self basicNew key: newKey value: newValue! !

!Association methodsFor: 'comparing' stamp: 'eem 11/29/2008 11:09'!
analogousCodeTo: anObject
	"For MethodProperties comparison."
	^anObject isVariableBinding
	  and: [key = anObject key
	  and: [value = anObject value]]! !

!Association methodsFor: 'filter streaming' stamp: 'MPW 1/1/1901 20:53'!
byteEncode: aStream
	aStream writeAssocation:self.! !

!Association methodsFor: 'comparing' stamp: 'ar 2/11/2010 23:50'!
hash
	"Hash is reimplemented because = is implemented.
	The super implementation is used to avoid hashing the value
	which can be extremely expensive."
	^super hash! !

!Association methodsFor: 'testing' stamp: 'ar 8/14/2001 23:06'!
isSpecialWriteBinding
	"Return true if this variable binding is write protected, e.g., should not be accessed primitively but rather by sending #value: messages"
	^false! !

!Association methodsFor: 'accessing' stamp: ''!
key: aKey value: anObject 
	"Store the arguments as the variables of the receiver."

	key := aKey.
	value := anObject! !

!Association methodsFor: 'testing' stamp: 'nice 7/20/2011 09:00'!
literalEqual: otherLiteral
	"Answer true if the receiver and otherLiteral represent the same literal.
	Variable bindings are literally equals only if identical.
	This is how variable sharing works, by preserving identity and changing only the value."
	^self == otherLiteral! !

!Association methodsFor: 'printing' stamp: 'fbs 5/2/2013 21:01'!
printOn: aStream
	| arrow |
	super printOn: aStream.
	"If the key is a binary selector and we don't use whitespace, we will stream (key, '->') asSymbol."
	arrow := (key isSymbol and: [key isBinary]) ifTrue: [' -> '] ifFalse: ['->'].
	aStream nextPutAll: arrow.
	value printOn: aStream! !

!Association methodsFor: 'printing' stamp: 'fbs 5/2/2013 21:03'!
storeOn: aStream
	| arrow |
	"Store in the format (key->value)"
	aStream nextPut: $(.
	key storeOn: aStream.
	"If the key is a binary selector and we don't use whitespace, we will stream (key, '->') asSymbol."
	arrow := (key isSymbol and: [key isBinary]) ifTrue: [' -> '] ifFalse: ['->'].
	aStream nextPutAll: arrow.
	value storeOn: aStream.
	aStream nextPut: $)! !

!Association methodsFor: 'accessing' stamp: ''!
value
	"Answer the value of the receiver."

	^value! !

!Association methodsFor: 'accessing' stamp: ''!
value: anObject 
	"Store the argument, anObject, as the value of the receiver."

	value := anObject! !

!Association methodsFor: 'comparing' stamp: 'md 1/27/2004 17:27'!
= anAssociation

	^ super = anAssociation and: [value = anAssociation value]! !

Association subclass: #WeakKeyAssociation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Support'!

!WeakKeyAssociation commentStamp: '<historical>' prior: 0!
I am an association holding only weakly on my key.!

!WeakKeyAssociation methodsFor: 'comparing' stamp: 'ar 3/21/98 15:46'!
hash
	"Hash is reimplemented because = is implemented."

	^self key hash! !

!WeakKeyAssociation methodsFor: 'accessing' stamp: 'ul 2/17/2010 13:54'!
key

	^key ifNotNil: [ key at: 1 ]! !

!WeakKeyAssociation methodsFor: 'accessing' stamp: 'ar 3/21/98 15:45'!
key: aKey
	key := WeakArray with: aKey! !

!WeakKeyAssociation methodsFor: 'accessing' stamp: 'raok 11/29/2002 14:49'!
key: aKey value: anObject
	key := WeakArray with: aKey.
	value := anObject.! !

!WeakKeyAssociation methodsFor: 'printing' stamp: 'ar 3/21/98 15:53'!
printOn: aStream
	self key printOn: aStream.
	aStream nextPutAll: '->'.
	self value printOn: aStream! !

!WeakKeyAssociation methodsFor: 'printing' stamp: 'ar 3/21/98 15:53'!
storeOn: aStream
	aStream 
		nextPut: $(;
		nextPutAll: self class name;
		nextPutAll:' key: '.
	self key storeOn: aStream.
	aStream nextPutAll: ' value: '.
	self value storeOn: aStream.
	aStream nextPut: $)! !

!WeakKeyAssociation methodsFor: 'comparing' stamp: 'ar 3/21/98 15:45'!
< aLookupKey 
	"Refer to the comment in Magnitude|<."

	^self key < aLookupKey key! !

!WeakKeyAssociation methodsFor: 'comparing' stamp: 'ar 3/21/98 15:46'!
= aLookupKey

	self species = aLookupKey species
		ifTrue: [^self key = aLookupKey key]
		ifFalse: [^false]! !

!LookupKey class methodsFor: 'instance creation' stamp: 'md 6/29/2005 16:34'!
key: aKey 
	"Answer an instance of me with the argument as the lookup up."

	^self basicNew key: aKey! !

!LookupKey methodsFor: 'bindings' stamp: 'cwp 7/19/2012 13:26'!
beBindingOfType: aClass announcing: aBool
	"Make the receiver a global binding of the given type"

	self class == aClass ifTrue:[^self].
	self becomeForward: (aClass key: self key value: self value).
	"NOTE: Now self == read-only (e.g., the new binding)"
	^self recompileBindingsAnnouncing: aBool! !

!LookupKey methodsFor: 'bindings' stamp: 'ar 8/16/2001 11:50'!
beReadOnlyBinding
	"Make the receiver (a global read-write binding) be a read-only binding"
	^self beReadOnlyBindingAnnouncing: true! !

!LookupKey methodsFor: 'bindings' stamp: 'ar 8/16/2001 11:50'!
beReadOnlyBindingAnnouncing: aBool
	"Make the receiver (a global read-write binding) be a read-only binding"
	^self beBindingOfType: ReadOnlyVariableBinding announcing: aBool! !

!LookupKey methodsFor: 'bindings' stamp: 'ar 8/16/2001 11:50'!
beReadWriteBinding
	"Make the receiver (a global read-only binding) be a read-write binding"
	^self beReadWriteBindingAnnouncing: true! !

!LookupKey methodsFor: 'bindings' stamp: 'ar 8/16/2001 11:51'!
beReadWriteBindingAnnouncing: aBool
	"Make the receiver (a global read-write binding) be a read-write binding"
	^self beBindingOfType: Association announcing: aBool! !

!LookupKey methodsFor: 'accessing' stamp: 'ajh 9/12/2002 12:04'!
canAssign

	^ true! !

!LookupKey methodsFor: 'comparing' stamp: ''!
hash
	"Hash is reimplemented because = is implemented."

	^key hash! !

!LookupKey methodsFor: 'testing' stamp: 'ar 8/14/2003 01:52'!
isSpecialReadBinding
	"Return true if this variable binding is read protected, e.g., should not be accessed primitively but rather by sending #value messages"
	^false! !

!LookupKey methodsFor: 'testing' stamp: 'ar 8/14/2001 22:39'!
isVariableBinding
	"Return true if I represent a literal variable binding"
	^true! !

!LookupKey methodsFor: 'accessing' stamp: ''!
key
	"Answer the lookup key of the receiver."

	^key! !

!LookupKey methodsFor: 'accessing' stamp: ''!
key: anObject 
	"Store the argument, anObject, as the lookup key of the receiver."

	key := anObject! !

!LookupKey methodsFor: 'accessing' stamp: 'ajh 3/24/2003 21:14'!
name

	^ self key isString
		ifTrue: [self key]
		ifFalse: [self key printString]! !

!LookupKey methodsFor: 'printing' stamp: ''!
printOn: aStream

	key printOn: aStream! !

!LookupKey methodsFor: 'bindings' stamp: 'fbs 6/4/2013 17:26'!
recompileBindingsAnnouncing: aBool 
	"Make the receiver (a global read-write binding) be a read-only binding"

	aBool 
		ifTrue: 
			[UIManager default informUserDuring: 
					[:bar | 
					(self systemNavigation allCallsOn: self) do: 
							[:mref | 
							bar value: 'Recompiling ' , mref asStringOrText.
							mref actualClass recompile: mref methodSymbol]]]
		ifFalse: 
			[(self systemNavigation allCallsOn: self) 
				do: [:mref | mref actualClass recompile: mref methodSymbol]]! !

!LookupKey methodsFor: 'filter streaming' stamp: 'MPW 1/1/1901 02:35'!
writeOnFilterStream: aStream

	aStream write:key.! !

!LookupKey methodsFor: 'comparing' stamp: ''!
< aLookupKey 
	"Refer to the comment in Magnitude|<."

	^key < aLookupKey key! !

!LookupKey methodsFor: 'comparing' stamp: ''!
= aLookupKey

	self species = aLookupKey species
		ifTrue: [^key = aLookupKey key]
		ifFalse: [^false]! !

LookupKey subclass: #ReadOnlyVariableBinding
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Support'!

!ReadOnlyVariableBinding class methodsFor: 'instance creation' stamp: 'ar 8/14/2001 23:11'!
key: key value: aValue
	^self new privateSetKey: key value: aValue! !

!ReadOnlyVariableBinding methodsFor: 'accessing' stamp: 'ajh 9/12/2002 12:06'!
canAssign

	^ false! !

!ReadOnlyVariableBinding methodsFor: 'testing' stamp: 'ar 8/14/2001 23:08'!
isSpecialWriteBinding
	"Return true if this variable binding is write protected, e.g., should not be accessed primitively but rather by sending #value: messages"
	^true! !

!ReadOnlyVariableBinding methodsFor: 'private' stamp: 'ar 8/14/2001 23:11'!
privateSetKey: aKey value: aValue
	key := aKey.
	value := aValue! !

!ReadOnlyVariableBinding methodsFor: 'accessing' stamp: 'ar 8/14/2001 23:09'!
value
	^value! !

!ReadOnlyVariableBinding methodsFor: 'accessing' stamp: 'ar 8/17/2001 18:03'!
value: aValue
	(AttemptToWriteReadOnlyGlobal signal: 'Cannot store into read-only bindings') == true ifTrue:[
		value := aValue.
	].! !

LookupKey weakSubclass: #WeakValueAssociation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Support'!

!WeakValueAssociation commentStamp: '<historical>' prior: 0!
I am a lookup key (acting like an association but) holding only weakly on my value.!

!WeakValueAssociation class methodsFor: 'instance creation' stamp: 'r++ 5/27/2000 18:07'!
key: anObject value: bObject
	^ self new key: anObject value: bObject! !

!WeakValueAssociation class methodsFor: 'as yet unclassified' stamp: 'r++ 5/27/2000 18:12'!
new
	^ self new: 1! !

!WeakValueAssociation methodsFor: 'accessing' stamp: 'r++ 5/27/2000 18:11'!
key: aKey value: anObject 
	"Store the arguments as the variables of the receiver."

	key := aKey.
	self value: anObject! !

!WeakValueAssociation methodsFor: 'accessing' stamp: 'r++ 5/27/2000 18:08'!
value
	^self at: 1! !

!WeakValueAssociation methodsFor: 'accessing' stamp: 'r++ 5/27/2000 18:08'!
value: anObject 
	"Store the argument, anObject, as the value of the receiver."

	self at: 1 put: anObject! !

!Number methodsFor: '*collections' stamp: 'cmm 9/7/2013 12:09'!
compareSafely: aNumber 
	^ aNumber isNumber
		ifTrue: [ self < aNumber ]
		ifFalse: [ super compareSafely: aNumber ]! !

Error subclass: #EndOfStream
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Exceptions'!

!EndOfStream commentStamp: '<historical>' prior: 0!
Signalled when ReadStream>>next encounters a premature end.!

!EndOfStream methodsFor: 'exceptionDescription' stamp: 'RAA 5/17/2000 03:10'!
defaultAction
	"Answer ReadStream>>next default reply."

	^ nil! !

!EndOfStream methodsFor: 'description' stamp: 'hh 5/17/2000 00:30'!
isResumable
	"EndOfStream is resumable, so ReadStream>>next can answer"

	^ true! !

Error subclass: #KeyNotFound
	instanceVariableNames: 'key'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Exceptions'!

!KeyNotFound class methodsFor: 'instance creation' stamp: 'ar 11/20/2007 14:49'!
key: anObject
	^self new key: anObject! !

!KeyNotFound methodsFor: 'accessing' stamp: 'nice 11/19/2013 23:21'!
description
	"Return a textual description of the exception."
	^self messageText! !

!KeyNotFound methodsFor: 'accessing' stamp: 'ar 11/20/2007 14:49'!
key
	"The key which wasn't found"
	^key! !

!KeyNotFound methodsFor: 'accessing' stamp: 'ar 11/20/2007 14:49'!
key: anObject
	"The key which wasn't found"
	key := anObject! !

!KeyNotFound methodsFor: 'accessing' stamp: 'ar 3/9/2010 22:02'!
messageText
	"Return a textual description of the exception."
	^messageText ifNil:['Key not found: ', key]! !

Error subclass: #NotFound
	instanceVariableNames: 'object'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Exceptions'!

!NotFound class methodsFor: 'instance creation' stamp: 'nice 11/19/2013 23:25'!
object: anObject
	^self new object: anObject! !

!NotFound methodsFor: 'accessing' stamp: 'nice 11/19/2013 23:23'!
messageText
	"Return a textual description of the exception."
	^messageText ifNil:['Object is not in the collection.']! !

!NotFound methodsFor: 'accessing' stamp: 'nice 11/19/2013 23:22'!
object
	^object! !

!NotFound methodsFor: 'accessing' stamp: 'nice 11/19/2013 23:22'!
object: anObject
	object := anObject! !

Object subclass: #LimitingLineStreamWrapper
	instanceVariableNames: 'stream line limitingBlock position'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Collections-Streams'!

!LimitingLineStreamWrapper commentStamp: '<historical>' prior: 0!
I'm a wrapper for a stream optimized for line-by-line access using #nextLine. My instances can be nested.

I read one line ahead. Reading terminates when the stream ends, or if the limitingBlock evaluated with the line answers true. To skip the delimiting line for further reading use #skipThisLine.

Character-based reading (#next) is permitted, too. Send #updatePosition when switching from line-based reading.

See examples at the class side.

--bf 2/19/1999 12:52!

!LimitingLineStreamWrapper class methodsFor: 'examples' stamp: 'bf 2/19/1999 11:48'!
example1
	"LimitingLineStreamWrapper example1"
	"Separate chunks of text delimited by a special string"
	| inStream msgStream messages |
	inStream := self exampleStream.
	msgStream := LimitingLineStreamWrapper on: inStream delimiter: 'From '.
	messages := OrderedCollection new.
	[inStream atEnd] whileFalse: [
		msgStream skipThisLine.
		messages add: msgStream upToEnd].
	^messages
			! !

!LimitingLineStreamWrapper class methodsFor: 'examples' stamp: 'bf 2/19/1999 12:46'!
example2
	"LimitingLineStreamWrapper example2"
	"Demo nesting wrappers - get header lines from some messages"
	| inStream msgStream headers headerStream |
	inStream := self exampleStream.
	msgStream := LimitingLineStreamWrapper on: inStream delimiter: 'From '.
	headers := OrderedCollection new.
	[inStream atEnd] whileFalse: [
		msgStream skipThisLine. "Skip From"
		headerStream := LimitingLineStreamWrapper on: msgStream delimiter: ''.
		headers add: headerStream linesUpToEnd.
		[msgStream nextLine isNil] whileFalse. "Skip Body"
	].
	^headers
			! !

!LimitingLineStreamWrapper class methodsFor: 'examples' stamp: 'bf 2/19/1999 12:44'!
exampleStream
	^ReadStream on:
'From me@somewhere
From: me
To: you
Subject: Test

Test

From you@elsewhere
From: you
To: me
Subject: Re: test

okay
'! !

!LimitingLineStreamWrapper class methodsFor: 'instance creation' stamp: 'bf 11/24/1998 14:31'!
on: aStream delimiter: aString

	^self new setStream: aStream delimiter: aString
! !

!LimitingLineStreamWrapper methodsFor: 'testing' stamp: 'bf 11/13/1998 16:55'!
atEnd

	^line isNil or: [limitingBlock value: line]! !

!LimitingLineStreamWrapper methodsFor: 'stream protocol' stamp: 'bf 11/13/1998 17:00'!
close
	^stream close! !

!LimitingLineStreamWrapper methodsFor: 'accessing' stamp: 'bf 11/24/1998 14:25'!
delimiter: aString
	"Set limitBlock to check for a delimiting string. Be unlimiting if nil"

	self limitingBlock: (aString caseOf: {
		[nil] -> [[:aLine | false]].
		[''] -> [[:aLine | aLine size = 0]]
	} otherwise: [[:aLine | aLine beginsWith: aString]])
! !

!LimitingLineStreamWrapper methodsFor: 'accessing' stamp: 'bf 11/13/1998 13:08'!
lastLineRead
	"Return line last read. At stream end, this is the boundary line or nil"

	^ line! !

!LimitingLineStreamWrapper methodsFor: 'accessing' stamp: 'nice 1/11/2010 20:56'!
limitingBlock: aBlock
	"The limitingBlock is evaluated with a line to check if this line terminates the stream"

	limitingBlock := aBlock.
	self updatePosition! !

!LimitingLineStreamWrapper methodsFor: 'accessing' stamp: 'bf 2/19/1999 11:45'!
linesUpToEnd

	| elements ln |
	elements := OrderedCollection new.
	[(ln := self nextLine) isNil] whileFalse: [ 
		elements add: ln].
	^elements! !

!LimitingLineStreamWrapper methodsFor: 'accessing' stamp: 'bf 11/24/1998 14:37'!
next
	"Provide character-based access"

	position isNil ifTrue: [^nil].
	position < line size ifTrue: [^line at: (position := position + 1)].
	line := stream nextLine.
	self updatePosition.
	^ Character cr! !

!LimitingLineStreamWrapper methodsFor: 'accessing' stamp: 'bf 11/24/1998 14:09'!
nextLine

	| thisLine |
	self atEnd ifTrue: [^nil].
	thisLine := line.
	line := stream nextLine.
	^thisLine
! !

!LimitingLineStreamWrapper methodsFor: 'accessing' stamp: 'bf 11/13/1998 13:04'!
peekLine

	self atEnd ifTrue: [^nil].
	^ line! !

!LimitingLineStreamWrapper methodsFor: 'printing' stamp: 'bf 11/24/1998 13:39'!
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: ' on '.
	stream printOn: aStream! !

!LimitingLineStreamWrapper methodsFor: 'private' stamp: 'bf 11/24/1998 14:30'!
setStream: aStream delimiter: aString

	stream := aStream.
	line := stream nextLine.
	self delimiter: aString.	"sets position"
! !

!LimitingLineStreamWrapper methodsFor: 'accessing' stamp: 'bf 11/24/1998 16:53'!
skipThisLine

	line := stream nextLine.
	self updatePosition.
! !

!LimitingLineStreamWrapper methodsFor: 'accessing' stamp: 'bf 11/24/1998 14:37'!
updatePosition
	"Call this before doing character-based access"

	position := self atEnd ifFalse: [0]! !

!LimitingLineStreamWrapper methodsFor: 'accessing' stamp: 'nice 12/26/2009 10:24'!
upToEnd
	^String streamContents: [:strm |
		| ln |
		[(ln := self nextLine) isNil] whileFalse: [ 
			strm nextPutAll: ln; cr]]! !

ArrayedCollection subclass: #Text
	instanceVariableNames: 'string runs'
	classVariableNames: 'IgnoreStyleIfOnlyBold'
	poolDictionaries: 'TextConstants'
	category: 'Collections-Text'!

!Text commentStamp: '<historical>' prior: 0!
I represent a character string that has been marked with abstract changes in character appearance. Actual display is performed in the presence of a TextStyle which indicates, for each abstract code, an actual font to be used.  A Text associates a set of TextAttributes with each character in its character string.  These attributes may be font numbers, emphases such as bold or italic, or hyperling actions.  Font numbers are interpreted relative to whatever textStyle appears, along with the text, in a Paragraph.  Since most characters have the same attributes as their neighbors, the attributes are stored in a RunArray for efficiency.  Each of my instances has
	string		a String
	runs		a RunArray!

!Text class methodsFor: 'private' stamp: 'di 10/31/97 11:22'!
addAttribute: att toArray: others 
	"Add a new text attribute to an existing set"
	"NOTE: The use of reset and set in this code is a specific
	hack for merging TextKerns."
	att reset.
	^ Array streamContents:
		[:strm | others do:
			[:other | (att dominates: other) ifFalse: [strm nextPut: other]].
		att set ifTrue: [strm nextPut: att]]! !

!Text class methodsFor: 'filler text' stamp: 'topa 6/1/2016 10:20'!
allDigits

	^ self fromString: '0123456789'! !

!Text class methodsFor: 'instance creation' stamp: 'ar 8/4/2009 23:54'!
fromString: aString 
	"Answer an instance of me whose characters are those of the argument, aString."

	^ self string: aString attributes:  #()! !

!Text class methodsFor: 'filler text' stamp: 'topa 6/1/2016 10:35'!
hamburgefonstiv

	^ self fromString: 'Hamburgefonstiv'! !

!Text class methodsFor: 'preferences' stamp: 'fbs 1/3/2014 16:50'!
ignoreStyleIfOnlyBold
		<preference: 'Ignore style if only bold' category: 'browsing' description: 'If true, then any method submission in which the only style change is for bolding will be treated as a method with no style specifications.' type: #Boolean>
		^ IgnoreStyleIfOnlyBold ifNil: [true].! !

!Text class methodsFor: 'preferences' stamp: 'fbs 1/3/2014 16:27'!
ignoreStyleIfOnlyBold: aBoolean
	IgnoreStyleIfOnlyBold := aBoolean.! !

!Text class methodsFor: 'class initialization' stamp: ''!
initialize	"Text initialize"
	"Initialize constants shared by classes associated with text display."

	TextConstants at: #CaretForm put:
				(Form extent: 16@5
					fromArray: #(2r001100e26 2r001100e26 2r011110e26 2r111111e26 2r110011e26)
					offset: -3@0).
	self initTextConstants! !

!Text class methodsFor: 'class initialization' stamp: 'nice 2/28/2010 19:18'!
initTextConstants 
	"Initialize constants shared by classes associated with text display, e.g., 
	Space, Tab, Cr, Bs, ESC."
		"1/24/96 sw: in exasperation and confusion, changed cmd-g mapping from 231 to 232 to see if I could gain any relief?!!"


	| letter varAndValue tempArray width |
	"CtrlA..CtrlZ, Ctrla..Ctrlz"
	letter := $A.
 	#(		212 230 228 196 194 226 241 243 214 229 200 217 246 
			245 216 202 210 239 211 240 197 198 209 215 242 231
	 		1 166 228 132 130 12 232 179 150 165 136 153 182 
			14 15 138 17 18 19 11 21 134 145 151 178 167 ) do:
		[:kbd |
		TextConstants at: ('Ctrl', letter asSymbol) asSymbol put: kbd asCharacter.
		letter := letter == $Z ifTrue: [$a] ifFalse: [(letter asciiValue + 1) asCharacter]].

	varAndValue := #(
		Space	32
		Tab		9
		CR		13
		Enter	3
		BS		8
		BS2		158
		ESC		160
		Clear 	173
	).

	varAndValue size odd ifTrue: [self error: 'unpaired text constant'].
	(2 to: varAndValue size by: 2) do:
		[:i | TextConstants at: (varAndValue at: i - 1) put: (varAndValue at: i) asCharacter].

	varAndValue := #(
		CtrlDigits 			(159 144 143 128 127 129 131 180 149 135)
		CtrlOpenBrackets	(201 7 218 249 219 15)
			"lparen gottn by ctrl-:= = 201; should be 213 but can't type that on Mac"

			"values for alignment"
		LeftFlush	0
		RightFlush	1
		Centered	2
		Justified	3

			"subscripts for a marginTabsArray tuple"
		LeftMarginTab	1
		RightMarginTab	2

			"font faces"
		Basal	0
		Bold	1
		Italic	2

			"in case font doesn't have a width for space character"
			"some plausible numbers-- are they the right ones?"
		DefaultSpace			4
		DefaultTab				24
		DefaultLineGrid			16
		DefaultBaseline			12
		DefaultFontFamilySize	3	"basal, bold, italic"
	).

	varAndValue size odd ifTrue: [self error: 'unpaired text constant'].
	(2 to: varAndValue size by: 2) do:
		[:i | TextConstants at: (varAndValue at: i - 1) put: (varAndValue at: i)].

	TextConstants at: #DefaultRule	put: Form over.
	TextConstants at: #DefaultMask	put: Color black.

	width := Display width max: 720.
	tempArray := Array new: width // DefaultTab.
	1 to: tempArray size do:
		[:i | tempArray at: i put: DefaultTab * i].
	TextConstants at: #DefaultTabsArray put: tempArray.
	tempArray := Array new: (width // DefaultTab) // 2.
	1 to: tempArray size do:
		[:i | tempArray at: i put: (Array with: (DefaultTab*i) with: (DefaultTab*i))].
	TextConstants at: #DefaultMarginTabsArray put: tempArray.

"Text initTextConstants "! !

!Text class methodsFor: 'filler text' stamp: 'topa 6/1/2016 10:20'!
loremIpsum

	^ self fromString: 'Lorem ipsum dolor sit amet, consectetur adipisicing elit, 
sed do eiusmod tempor incididunt ut labore et dolore 
magna aliqua. Ut enim ad minim veniam, quis nostrud 
exercitation ullamco laboris nisi ut aliquip ex ea commodo 
consequat. Duis aute irure dolor in reprehenderit in voluptate 
velit esse cillum dolore eu fugiat nulla pariatur. Excepteur 
sint occaecat cupidatat non proident, sunt in culpa qui 
officia deserunt mollit anim id est laborum.'! !

!Text class methodsFor: 'instance creation' stamp: ''!
new: stringSize

	^self fromString: (String new: stringSize)! !

!Text class methodsFor: 'instance creation' stamp: ''!
streamContents: blockWithArg 
	| stream |
	stream := TextStream on: (self new: 400).
	blockWithArg value: stream.
	^ stream contents! !

!Text class methodsFor: 'instance creation' stamp: ''!
string: aString attributes: atts
	"Answer an instance of me whose characters are those of aString.
	atts is an array of TextAttributes."

	^self string: aString runs: (RunArray new: aString size withAll: atts)! !

!Text class methodsFor: 'instance creation' stamp: ''!
string: aString attribute: att
	"Answer an instance of me whose characters are aString.
	att is a TextAttribute."

	^self string: aString attributes: (Array with: att)! !

!Text class methodsFor: 'instance creation' stamp: ''!
string: aString emphasis: emphasis
	"This is an old method that is mainly used by old applications"

	emphasis isNumber ifTrue:
		[self halt: 'Numeric emphasis is not supported in Squeak'.
		"But if you proceed, we will do our best to give you what you want..."
		^ self string: aString runs: (RunArray new: aString size withAll: 
			(Array with: (TextFontChange new fontNumber: emphasis)))].
	^ self string: aString attributes: emphasis! !

!Text class methodsFor: 'private' stamp: ''!
string: aString runs: anArray
 
	^self basicNew setString: aString setRuns: anArray! !

!Text class methodsFor: 'filler text' stamp: 'topa 6/1/2016 10:43'!
symbolSample

	^ self streamContents: [:stream | | lineLength character|
		lineLength := 0.
		33 to: 255 do:[ :index |
			character := stream nextPut: (Character value: index).
			lineLength := lineLength + 1.
			(('@Z`z' includes: character) or: [lineLength >= 30]) 
				ifTrue: [
					lineLength :=0.
					stream cr]]]! !

!Text class methodsFor: 'filler text' stamp: 'topa 6/1/2016 10:20'!
THEQUICKBROWNFOX

	^ self fromString: 'THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG.'! !

!Text class methodsFor: 'filler text' stamp: 'topa 6/1/2016 10:35'!
textSample

	^ self streamContents: [:stream |
		stream 
			nextPutAll: self hamburgefonstiv; cr;
			nextPutAll: self theQuickBrownFox; cr;
			nextPutAll: self THEQUICKBROWNFOX; cr; cr;
			nextPutAll: self allDigits; cr; cr;
			nextPutAll: self loremIpsum]! !

!Text class methodsFor: 'filler text' stamp: 'topa 6/1/2016 10:20'!
theQuickBrownFox

	^ self fromString: 'the quick brown fox jumps over the lazy dog'! !

!Text methodsFor: 'emphasis' stamp: ''!
addAttribute: att 
	^ self addAttribute: att from: 1 to: self size! !

!Text methodsFor: 'emphasis' stamp: ''!
addAttribute: att from: start to: stop 
	"Set the attribute for characters in the interval start to stop."
	runs :=  runs copyReplaceFrom: start to: stop
			with: ((runs copyFrom: start to: stop)
				mapValues:
				[:attributes | Text addAttribute: att toArray: attributes])
! !

!Text methodsFor: 'emphasis' stamp: 'ul 11/11/2009 16:28'!
alignmentAt: characterIndex ifAbsent: aBlock
	| attributes emph |
	self size = 0 ifTrue: [^aBlock value].
	emph := nil.
	attributes := runs at: characterIndex.
	attributes do:[:att | (att isKindOf: TextAlignment) ifTrue:[emph := att]].
	emph ifNil: [ ^aBlock value ].
	^emph alignment! !

!Text methodsFor: 'emphasis' stamp: ''!
allBold 
	"Force this whole text to be bold."
	string size = 0 ifTrue: [^self].
	self makeBoldFrom: 1 to: string size! !

!Text methodsFor: 'accessing' stamp: 'tk 9/4/2000 16:04'!
append: stringOrText

	self replaceFrom: string size + 1
				to: string size with: stringOrText! !

!Text methodsFor: 'converting' stamp: ''!
asDisplayText
	"Answer a DisplayText whose text is the receiver."

	^DisplayText text: self! !

!Text methodsFor: 'attributes' stamp: 'fbs 12/29/2013 17:09'!
askIfAddStyle: priorMethod req: requestor
	"Ask the user if we have a complex style (i.e. bold) for the first time"
	| tell answ old |
	(SystemBrowser browseWithPrettyPrint)
		ifTrue: [self couldDeriveFromPrettyPrinting ifTrue: [^ self asString]].
	self runs coalesce.
	self unembellished ifTrue: [^ self asString].
	priorMethod ifNotNil: [old := priorMethod getSourceFromFile].
	(old == nil or: [old unembellished])
		ifTrue:
			[tell := 'This method contains style for the first time (e.g. bold or colored text).
Do you really want to save the style info?'.
			answ := (UIManager default 
						chooseFrom: #('Save method with style' 'Save method simply')
						title: tell).
			answ = 2 ifTrue: [^ self asString]]! !

!Text methodsFor: 'converting' stamp: ''!
asNumber
	"Answer the number created by interpreting the receiver as the textual 
	representation of a number."

	^string asNumber! !

!Text methodsFor: 'converting' stamp: 'ar 4/12/2005 17:32'!
asOctetStringText

	string class == WideString ifTrue: [
		^ self class string: string asOctetString runs: self runs copy.
	].
	^self.
! !

!Text methodsFor: 'converting' stamp: ''!
asString
	"Answer a String representation of the textual receiver."

	^string! !

!Text methodsFor: 'converting' stamp: 'RAA 5/28/2001 06:19'!
asStringOrText	
	"Answer the receiver itself."

	^self! !

!Text methodsFor: 'converting' stamp: 'mt 5/3/2015 17:48'!
asStringToHtml
	"Inverse to String >> #asTextFromHtml"
	
	^ self printHtmlString! !

!Text methodsFor: 'converting' stamp: ''!
asText	
	"Answer the receiver itself."

	^self! !

!Text methodsFor: 'emphasis' stamp: 'sw 12/7/1999 12:30'!
attributesAt: characterIndex 
	"Answer the code for characters in the run beginning at characterIndex."
	"NB: no senders any more (supplanted by #attributesAt:forStyle: but retained for the moment in order not to break user code that may exist somewhere that still calls this"
	| attributes |
	self size = 0
		ifTrue: [^ Array with: (TextFontChange new fontNumber: 1)].  "null text tolerates access"
	attributes := runs at: characterIndex.
	^ attributes! !

!Text methodsFor: 'emphasis' stamp: 'ar 12/17/2001 01:17'!
attributesAt: characterIndex do: aBlock
	"Answer the code for characters in the run beginning at characterIndex."
	"NB: no senders any more (supplanted by #attributesAt:forStyle: but retained for the moment in order not to break user code that may exist somewhere that still calls this"
	self size = 0 ifTrue:[^self].
	(runs at: characterIndex) do: aBlock! !

!Text methodsFor: 'emphasis' stamp: 'sw 12/7/1999 11:32'!
attributesAt: characterIndex forStyle: aTextStyle
	"Answer the code for characters in the run beginning at characterIndex."
	| attributes |
	self size = 0
		ifTrue: [^ Array with: (TextFontChange new fontNumber: aTextStyle defaultFontIndex)].  "null text tolerates access"
	attributes := runs at: characterIndex.
	^ attributes! !

!Text methodsFor: 'accessing' stamp: ''!
at: index

	^string at: index! !

!Text methodsFor: 'accessing' stamp: ''!
at: index put: character

	^string at: index put: character! !

!Text methodsFor: 'emphasis' stamp: 'mt 11/12/2015 09:58'!
colorAt: characterIndex 

	^ self colorAt: characterIndex ifNone: [Color black]! !

!Text methodsFor: 'emphasis' stamp: 'mt 11/12/2015 10:13'!
colorAt: characterIndex ifNone: block

	self size = 0 ifTrue: [^ block value]. "null text tolerates access."

	^ (runs at: characterIndex)
		detect: [:attr | attr class == TextColor]
		ifFound: [:attr | attr color]
		ifNone: block! !

!Text methodsFor: 'copying' stamp: 'nice 10/19/2009 08:32'!
copyFrom: start to: stop 
	"Answer a copied subrange of the receiver."

	| realStart realStop |
	stop > self size
		ifTrue: [realStop := self size]		"handle selection at end of string"
		ifFalse: [realStop := stop].
	start < 1
		ifTrue: [realStart := 1]			"handle selection before start of string"
		ifFalse: [realStart := start].
	^self class 
		string: (string copyFrom: realStart to: realStop)
		runs: (runs copyFrom: realStart to: realStop)! !

!Text methodsFor: 'copying' stamp: 'BG 6/12/2003 13:11'!
copyReplaceFrom: start to: stop with: aTextOrString

	| txt |
	txt := aTextOrString asText.	"might be a string"
	^self class 
             string: (string copyReplaceFrom: start to: stop with: txt string)
             runs: (runs copyReplaceFrom: start to: stop with: txt runs)
! !

!Text methodsFor: 'copying' stamp: 'tk 1/7/98 10:58'!
copyReplaceTokens: oldSubstring with: newSubstring 
	"Replace all occurrences of oldSubstring that are surrounded
	by non-alphanumeric characters"
	^ self copyReplaceAll: oldSubstring with: newSubstring asTokens: true
	"'File asFile Files File''s File' copyReplaceTokens: 'File' with: 'Snick'"! !

!Text methodsFor: 'attributes' stamp: 'sw 11/16/1999 22:33'!
couldDeriveFromPrettyPrinting
	"Return true if the receiver has any TextAttributes that are functional rather than simply appearance-related"
	runs values do:
		[:emphArray | emphArray do:
			[:emph | emph couldDeriveFromPrettyPrinting ifFalse: [^ false]]].
	^ true! !

!Text methodsFor: 'copying' stamp: 'di 11/9/97 17:13'!
deepCopy

	^ self copy "Both string and runs are assumed to be read-only"! !

!Text methodsFor: 'accessing' stamp: 'gm 2/15/2003 14:59'!
embeddedMorphs
	"return the list of morphs embedded in me"

	| morphs |
	morphs := IdentitySet new.
	runs withStartStopAndValueDo: 
			[:start :stop :attribs | 
			attribs 
				do: [:attrib | attrib anchoredMorph ifNotNil: [morphs add: attrib anchoredMorph]]].
	^morphs select: [:m | m isMorph]! !

!Text methodsFor: 'accessing' stamp: 'gm 2/15/2003 14:59'!
embeddedMorphsFrom: start to: stop 
	"return the list of morphs embedded in me"

	| morphs |
	morphs := IdentitySet new.
	runs 
		runsFrom: start
		to: stop
		do: 
			[:attribs | 
			attribs 
				do: [:attr | attr anchoredMorph ifNotNil: [morphs add: attr anchoredMorph]]].
	^morphs select: [:m | m isMorph]! !

!Text methodsFor: 'emphasis' stamp: 'nice 1/10/2010 17:55'!
emphasisAt: characterIndex
	"Answer the fontfor characters in the run beginning at characterIndex."
	| attributes |
	self size = 0 ifTrue: [^ 0].	"null text tolerates access"
	attributes := runs at: characterIndex.
	^attributes inject: 0 into: 
		[:emph :att | emph bitOr: att emphasisCode].
	! !

!Text methodsFor: 'accessing' stamp: ''!
findString: aString startingAt: start 
	"Answer the index of subString within the receiver, starting at index 
	start. If the receiver does not contain subString, answer 0."

	^string findString: aString asString startingAt: start! !

!Text methodsFor: 'accessing' stamp: 'di 11/23/1998 11:53'!
findString: aString startingAt: start caseSensitive: caseSensitive
	"Answer the index of subString within the receiver, starting at index 
	start. If the receiver does not contain subString, answer 0."

	^string findString: aString asString startingAt: start caseSensitive: caseSensitive! !

!Text methodsFor: 'emphasis' stamp: 'di 11/10/97 13:36'!
find: attribute
	"Return the first interval over which this attribute applies"
	| begin end |
	begin := 0.
	runs withStartStopAndValueDo:
		[:start :stop :attributes |
		(attributes includes: attribute)
			ifTrue: [begin = 0 ifTrue: [begin := start].
					end := stop]
			ifFalse: [begin > 0 ifTrue: [^ begin to: end]]].
	begin > 0 ifTrue: [^ begin to: end].
	^ nil! !

!Text methodsFor: 'emphasis' stamp: 'sw 12/7/1999 10:58'!
fontAt: characterIndex withStyle: aTextStyle
	"Answer the fontfor characters in the run beginning at characterIndex."
	| attributes font |
	self size = 0 ifTrue: [^ aTextStyle defaultFont].	"null text tolerates access"
	attributes := runs at: characterIndex.
	font := aTextStyle defaultFont.  "default"
	attributes do: 
		[:att | att forFontInStyle: aTextStyle do: [:f | font := f]].
	^ font! !

!Text methodsFor: 'emphasis' stamp: ''!
fontNumberAt: characterIndex 
	"Answer the fontNumber for characters in the run beginning at characterIndex."
	| attributes fontNumber |
	self size = 0 ifTrue: [^1].	"null text tolerates access"
	attributes := runs at: characterIndex.
	fontNumber := 1.
	attributes do: [:att | (att isMemberOf: TextFontChange) ifTrue: [fontNumber := att fontNumber]].
	^ fontNumber
	! !

!Text methodsFor: 'testing' stamp: 'mt 8/13/2016 12:20'!
hasClickableAttribute

	^ self runs anySatisfy: [:attrs | attrs anySatisfy: [:attr |
			attr respondsTo: #mayActOnClick]]! !

!Text methodsFor: 'testing' stamp: 'mt 8/13/2016 12:18'!
hasColorAttribute

	^ self runs anySatisfy: [:attrs | attrs anySatisfy: [:attr |
			attr respondsTo: #color]]! !

!Text methodsFor: 'testing' stamp: 'mt 8/13/2016 12:19'!
hasFontAttribute

	^ self runs anySatisfy: [:attrs | attrs anySatisfy: [:attr |
			(attr respondsTo: #fontNumber) or: [attr respondsTo: #font]]]! !

!Text methodsFor: 'comparing' stamp: 'tk 10/17/2001 14:12'!
hash
	"#hash is implemented, because #= is implemented.  We are now equal to a string with the same characters.  Hash must reflect that."

	^ string hash! !

!Text methodsFor: 'comparing' stamp: 'topa 10/7/2015 23:07'!
hashWithInitialHash: initialHash 
	"Implemented to be polymorphic with String"
	^ self string hashWithInitialHash: initialHash
! !

!Text methodsFor: 'comparing' stamp: 'tk 9/6/2000 11:59'!
howManyMatch: aString

	^ self string howManyMatch: aString! !

!Text methodsFor: 'converting' stamp: 'ar 4/9/2005 22:16'!
isoToSqueak
	^self "no longer needed"! !

!Text methodsFor: 'testing' stamp: ''!
isText
	^ true! !

!Text methodsFor: 'accessing' stamp: 'tk 9/6/2000 12:33'!
lineCount

	^ string lineCount! !

!Text methodsFor: 'converting' stamp: 'ar 4/10/2005 15:58'!
macToSqueak
	"Convert the receiver from MacRoman to Squeak encoding"
	^ self class new setString: string macToSqueak setRuns: runs copy! !

!Text methodsFor: 'emphasis' stamp: ''!
makeBoldFrom: start to: stop

	^ self addAttribute: TextEmphasis bold from: start to: stop! !

!Text methodsFor: 'emphasis' stamp: 'nice 2/23/2011 21:50'!
makeSelectorBold
	"For formatting Smalltalk source code, set the emphasis of that portion of 
	the receiver's string that parses as a message selector to be bold."

	| parser i |
	string size = 0 ifTrue: [^ self].
	i := 0.
	[(string at: (i := i + 1)) isSeparator] whileTrue.
	(string at: i) = $[ ifTrue: [^ self].  "block, no selector"
	[(parser := Compiler newParser) parseSelector: string] on: Error do: [^ self].
	self makeBoldFrom: 1 to: (parser endOfLastToken min: string size)! !

!Text methodsFor: 'emphasis' stamp: 'nice 2/23/2011 21:50'!
makeSelectorBoldIn: aClass
	"For formatting Smalltalk source code, set the emphasis of that portion of 
	the receiver's string that parses as a message selector to be bold."

	| parser |
	string size = 0 ifTrue: [^self].
	(parser := aClass newParser) parseSelector: string.
	self makeBoldFrom: 1 to: (parser endOfLastToken min: string size)! !

!Text methodsFor: 'copying' stamp: 'nice 10/5/2009 08:53'!
postCopy
	super postCopy.
	string := string copy.
	runs := runs copy! !

!Text methodsFor: 'accessing' stamp: 'ar 12/27/2001 00:03'!
prepend: stringOrText

	self replaceFrom: 1 to: 0 with: stringOrText! !

!Text methodsFor: 'html' stamp: 'mt 5/3/2015 16:37'!
printHtmlOn: aStream 
	
	(HtmlReadWriter on: aStream)
		nextPutText: self.! !

!Text methodsFor: 'html' stamp: 'mt 5/3/2015 17:47'!
printHtmlString
	"answer a string whose characters are the html representation 
	of the receiver"
	
	^ String streamContents: [:stream |
		self printHtmlOn: stream]! !

!Text methodsFor: 'printing' stamp: 'sma 6/1/2000 09:49'!
printOn: aStream
	self printNameOn: aStream.
	aStream nextPutAll: ' for '; print: string! !

!Text methodsFor: 'accessing' stamp: 'BG 6/8/2003 16:18'!
rangeOf: attribute startingAt: index
"Answer an interval that gives the range of attribute at index position  index. An empty interval with start value index is returned when the attribute is not present at position index.  "
   ^string size = 0
      ifTrue: [index to: index - 1]
	 ifFalse: [runs rangeOf: attribute startingAt: index]! !

!Text methodsFor: 'converting' stamp: 'nice 12/27/2009 03:10'!
removeAttributesThat: removalBlock replaceAttributesThat: replaceBlock by: convertBlock
	"Enumerate all attributes in the receiver. Remove those passing removalBlock and replace those passing replaceBlock after converting it through convertBlock"
	| added removed |
	"Deliberately optimized for the no-op default."
	added := removed := nil.
	runs withStartStopAndValueDo: [ :start :stop :attribs | 
		attribs do: [ :attrib | | new |
			(removalBlock value: attrib) ifTrue:[
				removed ifNil:[removed := WriteStream on: #()].
				removed nextPut: {start. stop. attrib}.
			] ifFalse:[
				(replaceBlock value: attrib) ifTrue:[
					removed ifNil:[removed := WriteStream on: #()].
					removed nextPut: {start. stop. attrib}.
					new := convertBlock value: attrib.
					added ifNil:[added := WriteStream on: #()].
					added nextPut: {start. stop. new}.
				].
			].
		].
	].
	(added == nil and:[removed == nil]) ifTrue:[^self].
	"otherwise do the real work"
	removed ifNotNil:[removed contents do:[:spec|
		self removeAttribute: spec last from: spec first to: spec second]].
	added ifNotNil:[added contents do:[:spec|
		self addAttribute: spec last from: spec first to: spec second]].! !

!Text methodsFor: 'emphasis' stamp: ''!
removeAttribute: att from: start to: stop 
	"Remove the attribute over the interval start to stop."
	runs :=  runs copyReplaceFrom: start to: stop
			with: ((runs copyFrom: start to: stop)
				mapValues:
				[:attributes | attributes copyWithout: att])
! !

!Text methodsFor: 'accessing' stamp: 'tk 12/30/97 07:17'!
replaceFrom: start to: stop with: aText

	| txt |
	txt := aText asText.	"might be a string"
	string := string copyReplaceFrom: start to: stop with: txt string.
	runs := runs copyReplaceFrom: start to: stop with: txt runs! !

!Text methodsFor: 'converting' stamp: 'dvf 10/1/2003 02:58'!
replaceFrom: start to: stop with: replacement startingAt: repStart 
	"This destructively replaces elements from start to stop in the receiver starting at index, repStart, in replacementCollection. Do it to both the string and the runs."

	| rep newRepRuns |
	rep := replacement asText.	"might be a string"
	string replaceFrom: start to: stop with: rep string startingAt: repStart.
	newRepRuns := rep runs copyFrom: repStart to: repStart + stop - start.
	runs := runs copyReplaceFrom: start to: stop with: newRepRuns! !

!Text methodsFor: 'converting' stamp: 'nice 10/19/2009 08:32'!
reversed

	"Answer a copy of the receiver with element order reversed."

	^ self class string: string reversed runs: runs reversed.

  "  It is assumed that  self size = runs size  holds. "! !

!Text methodsFor: 'emphasis' stamp: 'ar 7/22/2009 21:52'!
runLengthFor: characterIndex 
	"Answer the count of characters remaining in run beginning with 
	characterIndex."
	self size = 0 ifTrue:[^0]. "null tolerates access"
	^runs runLengthAt: characterIndex! !

!Text methodsFor: 'private' stamp: ''!
runs

	^runs! !

!Text methodsFor: 'accessing' stamp: 'tween 9/13/2004 10:07'!
runs: anArray

	runs := anArray! !

!Text methodsFor: 'private' stamp: 'ar 4/9/2010 21:14'!
setString: aString setRunsChecking: aRunArray
	"Check runs and do the best you can to make them fit..."

	string := aString.
	"check the runs"
	aRunArray ifNil: [^ aString asText].
	(aRunArray isKindOf: RunArray) ifFalse: [^ aString asText].
	aRunArray runs size = aRunArray values size ifFalse: [^ aString asText].
	aRunArray size = aString size ifFalse: [^ aString asText].
	
	runs := aRunArray.! !

!Text methodsFor: 'private' stamp: ''!
setString: aString setRuns: anArray

	string := aString.
	runs := anArray! !

!Text methodsFor: 'accessing' stamp: ''!
size

	^string size! !

!Text methodsFor: 'converting' stamp: 'ar 4/9/2005 22:16'!
squeakToIso
	^self "no longer needed"! !

!Text methodsFor: 'converting' stamp: 'ar 4/10/2005 15:56'!
squeakToMac
	"Convert the receiver from Squeak to MacRoman encoding"
	^ self class new setString: string squeakToMac setRuns: runs copy! !

!Text methodsFor: 'printing' stamp: ''!
storeOn: aStream

	aStream nextPutAll: '(Text string: ';
		store: string;
		nextPutAll: ' runs: ';
		store: runs;
		nextPut: $)! !

!Text methodsFor: 'accessing' stamp: ''!
string
	"Answer the string representation of the receiver."

	^string! !

!Text methodsFor: 'testing' stamp: 'fbs 1/3/2014 16:27'!
unembellished 
	"Return true if the only emphases are the default font and bold"
	| font1 bold |
	font1 := TextFontChange defaultFontChange.
	bold := TextEmphasis bold.
	Text ignoreStyleIfOnlyBold ifFalse:
		["Ignore font1 only or font1-bold followed by font1-plain"
		^ (runs values = (Array with: (Array with: font1)))
		or: [runs values = (Array with: (Array with: font1 with: bold)
 								with: (Array with: font1))]].

	"If preference is set, then ignore any combo of font1 and bold"
	runs withStartStopAndValueDo:
		[:start :stop :emphArray |
		emphArray do:
			[:emph | (font1 = emph or: [bold = emph]) ifFalse: [^ false]]].
	^ true! !

!Text methodsFor: 'converting' stamp: 'ul 4/1/2015 19:47'!
withBlanksTrimmed
	"Return a copy of the receiver from which leading and trailing blanks have been trimmed."

	| first last |
	first := string indexOfAnyOf: CharacterSet nonSeparators startingAt: 1.
	first = 0 ifTrue: [ ^'' ].  "no non-separator character"
	last := string lastIndexOfAnyOf: CharacterSet nonSeparators startingAt: self size ifAbsent: [self size].
	(first = 1 and: [ last = self size ]) ifTrue: [ ^self copy ].
	^self
		copyFrom: first
		to: last
! !

!Text methodsFor: 'converting' stamp: 'nk 9/16/2003 16:46'!
withSqueakLineEndings
	"Answer a copy of myself in which all sequences of <CR><LF> or <LF> have been changed to <CR>"
	| newText |
	(string includes: Character lf) ifFalse: [ ^self copy ].
	newText := self copyReplaceAll: String crlf with: String cr asTokens: false.
	(newText asString includes: Character lf) ifFalse: [ ^newText ].
	^newText copyReplaceAll: String lf with: String cr asTokens: false.! !

!Text methodsFor: 'sorting' stamp: 'ul 5/24/2015 05:04'!
<=> aCharacterArray
	"Return a collation order of -1, 0, or 1, indicating whether I should be collated before the receiver, am equal, or after.
	See also:  http://en.wikipedia.org/wiki/Spaceship_operator"

	aCharacterArray isString ifTrue: [ ^string <=> aCharacterArray ].
	^string <=> aCharacterArray asString! !

!Text methodsFor: 'comparing' stamp: 'ul 5/24/2015 05:05'!
= other
	"Am I equal to the other Text or String?  
	***** Warning ***** Two Texts are considered equal if they have the same characters in them.  They might have completely different emphasis, fonts, sizes, text actions, or embedded morphs.  If you need to find out if one is a true copy of the other, you must do (text1 = text2 and: [text1 runs = text2 runs])."

	other isText ifTrue:	[ ^string = other string ].
	other isString ifTrue: [ ^string = other ].
	^false! !

TextAttribute subclass: #TextAlignment
	instanceVariableNames: 'alignment'
	classVariableNames: ''
	poolDictionaries: 'TextConstants'
	category: 'Collections-Text'!

!TextAlignment class methodsFor: 'as yet unclassified' stamp: 'KR 12/5/2005 00:53'!
alignmentSymbol: alignmentCode
	^#(leftFlush rightFlush centered justified) at: (alignmentCode + 1)! !

!TextAlignment class methodsFor: 'instance creation' stamp: 'nice 3/6/2008 22:33'!
centered
	^self new alignment: Centered! !

!TextAlignment class methodsFor: 'instance creation' stamp: 'nice 3/6/2008 22:33'!
justified
	^self new alignment: Justified! !

!TextAlignment class methodsFor: 'instance creation' stamp: 'nice 3/6/2008 22:33'!
leftFlush
	^self new alignment: LeftFlush! !

!TextAlignment class methodsFor: 'instance creation' stamp: 'nice 3/6/2008 22:33'!
rightFlush
	^self new alignment: RightFlush! !

!TextAlignment class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 10:23'!
scanCharacter
	"The character used to identify a subclass of TextAttribute for filein and fileout"
	^$a! !

!TextAlignment class methodsFor: 'fileIn/Out' stamp: 'dtl 1/2/2014 13:56'!
scanFrom: strm
	^self new alignment: (Integer readFrom: strm ifFail: [0])! !

!TextAlignment methodsFor: 'accessing' stamp: 'ar 12/15/2001 23:33'!
alignment
	^alignment! !

!TextAlignment methodsFor: 'accessing' stamp: 'ar 12/15/2001 23:33'!
alignment: aNumber
	alignment := aNumber.! !

!TextAlignment methodsFor: 'as yet unclassified' stamp: 'ar 12/16/2001 00:20'!
dominates: other
	"There can be only one..."
	^self class == other class! !

!TextAlignment methodsFor: 'as yet unclassified' stamp: 'ar 12/15/2001 23:34'!
emphasizeScanner: scanner
	"Set the emphasist for text scanning"
	scanner setAlignment: alignment.! !

!TextAlignment methodsFor: 'comparing' stamp: 'ar 9/9/2003 22:03'!
hash
	"#hash is re-implemented because #= is re-implemented"
	^ alignment hash! !

!TextAlignment methodsFor: 'fileIn/fileOut' stamp: 'dtl 1/2/2014 10:29'!
writeScanOn: strm

	strm nextPut: self class scanCharacter.
	alignment printOn: strm.! !

!TextAlignment methodsFor: 'comparing' stamp: 'ar 12/15/2001 23:33'!
= other 
	^ (other class == self class) 
		and: [other alignment = alignment]! !
Collection initialize!
WeakArray initialize!
String initialize!
Symbol initialize!
WideSymbol initialize!
Base64MimeConverter initialize!
SetElement initialize!
TranscriptStream initialize!
TextAction initialize!
Character initialize!
Text initialize!
